/* Definitions file for GNU Emacs running on Mach (BSD 4.3)
   Copyright (C) 1985, 1986, 1993 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Synched up with: Not in FSF.
   #### This file does not appear in FSF 19.29 and likely should be
   deleted.  There is already a mach2.h and the THIS_IS_YMAKEFILE
   stuff is obsolete. */

/* This has only been tested in the Carnegie Mellon University
 * Computer Science Department Mach environment.  If you use it
 * elsewhere, please let me (dkindred@cmu.edu) know whether it
 * works. */

#include "bsd4-3.h"

/* SYSTEM_TYPE should indicate the kind of system you are using.
 It sets the Lisp variable system-type.  */

#undef SYSTEM_TYPE
#define SYSTEM_TYPE "mach"

/* Define this macro if system defines a type `union wait'.  */

#define HAVE_UNION_WAIT

/* Don't send signals to subprocesses by "typing" special chars at them. */
#undef SIGNALS_VIA_CHARACTERS

/* XEmacs change */
/* unistd.h defines _POSIX_VERSION, which leads some things to believe 
   that _POSIX_PATH_MAX should be defined.  Unfortunately, it isn't. */
#ifndef THIS_IS_YMAKEFILE
#include <sys/param.h>
#define _POSIX_PATH_MAX MAXPATHLEN
#endif

#ifndef THIS_IS_YMAKEFILE
typedef int pid_t;
/* XEmacs change */
typedef unsigned short mode_t;
#endif THIS_IS_YMAKEFILE

#if (defined(i386) || defined(ibmrt))
  /* use drem() instead of fmod() -- this is a BUG in the compiler runtime. */
# define USE_DREM
#endif
