/* This file is the configuration file for the Linux operating system,
   starting with version 1.1.56.
   Copyright (C) 1985, 1986, 1992, 1994 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Synched up with: Not in FSF. */

#include "linux-old.h"

/* SIGIO is working as of 1.1.56 */
/* but GDB 4.13 gets all screwed up when SIGIO's are sent to it
   (at least, this is the case for me, under 1.2.4).  FMH!!!!! */
#ifndef DEBUG_XEMACS
#undef BROKEN_SIGIO
#endif
