;;; load the elisp side of Energize

;;; Remember, if you add any files here, you have to mention them in
;;; ymakefile or their docstrings won't be available.

(load "view-less")
(garbage-collect)
;;; (or (featurep 'font-lock) (load "font-lock"))
;;; (garbage-collect)
;;; (or (featurep 'fast-lock) (load "fast-lock"))
;;; (garbage-collect)
(or (featurep 'dired-extra) (load "dired-x"))
(load "userlock")
(garbage-collect)
(load "comint")
(garbage-collect)
(load "shell")
(garbage-collect)
(load "compile")
(garbage-collect)
(load "gdb")
(garbage-collect)
(load "evi")
(garbage-collect)
(load "energize/energize-init")
(garbage-collect)
(load "energize/energize-windows")
(garbage-collect)
(load "energize/energize-mode")
(garbage-collect)
(load "energize/energize-menus")
(garbage-collect)
(load "energize/energize-shell")
(garbage-collect)
(load "energize/energize-visit-use")
(garbage-collect)
(load "energize/energize-vi")
(garbage-collect)
(load "energize/energize-advice")
(garbage-collect)
(load "energize/energize-font-lock")

(autoload 'energize-annotate-print-ps "energize-annoprint" t)
(autoload 'energize-set-font-family "energize-font-size" t)
(autoload 'energize-set-font-size "energize-font-size" t)
(autoload 'energize-set-font-boldness "energize-font-size" t)
(autoload 'blink-paren-init "blink-paren" t)

(if (fboundp (intern-soft "initialize-backtrace-logging-internal"))
    (load "energize/backtrace-logging"))

(provide 'energize)
