/* Synched up with: Not in FSF. */

#include "decosf3-1.h"

/* XEmacs change from "Andrew G. Cohen, 617-353-6051" <cohen@andy.bu.edu>

/* apparently this breaks things under OSF 3.2 */
#undef GMALLOC_NEEDS_SBRK_DECL
