;;; w3-10646.el,v --- ISO 10646 Entities for Emacs-w3
;; Author: wmperry
;; Created: 1995/05/29 14:46:00
;; Version: 1.4
;; Keywords: i18n

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; This work by William M. Perry <wmperry@spry.com> is derived from
;;; copyrighted material by Erik Naggum <erik@naggum.no>.  The original
;;; files may be fetched from ftp.ifi.uio.no in the directory
;;; /pub/SGML/CHARSET.
;;;
;;; Compilation copyright 1993,1994 by Erik Naggum <erik@naggum.no>.
;;;
;;; This file is not part of GNU Emacs.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst w3-numeric-entities
  '(
    (256 . "LATIN CAPITAL LETTER A WITH MACRON")
    (257 . "LATIN SMALL LETTER A WITH MACRON")
    (258 . "LATIN CAPITAL LETTER A WITH BREVE")
    (259 . "LATIN SMALL LETTER A WITH BREVE")
    (260 . "LATIN CAPITAL LETTER A WITH OGONEK")
    (261 . "LATIN SMALL LETTER A WITH OGONEK")
    (262 . "LATIN CAPITAL LETTER C WITH ACUTE")
    (263 . "LATIN SMALL LETTER C WITH ACUTE")
    (264 . "LATIN CAPITAL LETTER C WITH CIRCUMFLEX")
    (265 . "LATIN SMALL LETTER C WITH CIRCUMFLEX")
    (266 . "LATIN CAPITAL LETTER C WITH DOT ABOVE")
    (267 . "LATIN SMALL LETTER C WITH DOT ABOVE")
    (268 . "LATIN CAPITAL LETTER C WITH CARON")
    (269 . "LATIN SMALL LETTER C WITH CARON")
    (270 . "LATIN CAPITAL LETTER D WITH CARON")
    (271 . "LATIN SMALL LETTER D WITH CARON")
    (272 . "LATIN CAPITAL LETTER D WITH STROKE")
    (273 . "LATIN SMALL LETTER D WITH STROKE")
    (274 . "LATIN CAPITAL LETTER E WITH MACRON")
    (275 . "LATIN SMALL LETTER E WITH MACRON")
    (276 . "LATIN CAPITAL LETTER E WITH BREVE")
    (277 . "LATIN SMALL LETTER E WITH BREVE")
    (278 . "LATIN CAPITAL LETTER E WITH DOT ABOVE")
    (279 . "LATIN SMALL LETTER E WITH DOT ABOVE")
    (280 . "LATIN CAPITAL LETTER E WITH OGONEK")
    (281 . "LATIN SMALL LETTER E WITH OGONEK")
    (282 . "LATIN CAPITAL LETTER E WITH CARON")
    (283 . "LATIN SMALL LETTER E WITH CARON")
    (284 . "LATIN CAPITAL LETTER G WITH CIRCUMFLEX")
    (285 . "LATIN SMALL LETTER G WITH CIRCUMFLEX")
    (286 . "LATIN CAPITAL LETTER G WITH BREVE")
    (287 . "LATIN SMALL LETTER G WITH BREVE")
    (288 . "LATIN CAPITAL LETTER G WITH DOT ABOVE")
    (289 . "LATIN SMALL LETTER G WITH DOT ABOVE")
    (290 . "LATIN CAPITAL LETTER G WITH CEDILLA")
    (291 . "LATIN SMALL LETTER G WITH CEDILLA")
    (292 . "LATIN CAPITAL LETTER H WITH CIRCUMFLEX")
    (293 . "LATIN SMALL LETTER H WITH CIRCUMFLEX")
    (294 . "LATIN CAPITAL LETTER H WITH STROKE")
    (295 . "LATIN SMALL LETTER H WITH STROKE")
    (296 . "LATIN CAPITAL LETTER I WITH TILDE")
    (297 . "LATIN SMALL LETTER I WITH TILDE")
    (298 . "LATIN CAPITAL LETTER I WITH MACRON")
    (299 . "LATIN SMALL LETTER I WITH MACRON")
    (300 . "LATIN CAPITAL LETTER I WITH BREVE")
    (301 . "LATIN SMALL LETTER I WITH BREVE")
    (302 . "LATIN CAPITAL LETTER I WITH OGONEK")
    (303 . "LATIN SMALL LETTER I WITH OGONEK")
    (304 . "LATIN CAPITAL LETTER I WITH DOT ABOVE")
    (305 . "LATIN SMALL LETTER DOTLESS I")
    (306 . "LATIN CAPITAL LETTER IJ|LATIN CAPITAL LIGATURE IJ")
    (307 . "LATIN SMALL LETTER IJ|LATIN SMALL LIGATURE IJ")
    (308 . "LATIN CAPITAL LETTER J WITH CIRCUMFLEX")
    (309 . "LATIN SMALL LETTER J WITH CIRCUMFLEX")
    (310 . "LATIN CAPITAL LETTER K WITH CEDILLA")
    (311 . "LATIN SMALL LETTER K WITH CEDILLA")
    (312 . "LATIN SMALL LETTER KRA")
    (313 . "LATIN CAPITAL LETTER L WITH ACUTE")
    (314 . "LATIN SMALL LETTER L WITH ACUTE")
    (315 . "LATIN CAPITAL LETTER L WITH CEDILLA")
    (316 . "LATIN SMALL LETTER L WITH CEDILLA")
    (317 . "LATIN CAPITAL LETTER L WITH CARON")
    (318 . "LATIN SMALL LETTER L WITH CARON")
    (319 . "LATIN CAPITAL LETTER L WITH MIDDLE DOT")
    (320 . "LATIN SMALL LETTER L WITH MIDDLE DOT")
    (321 . "LATIN CAPITAL LETTER L WITH STROKE")
    (322 . "LATIN SMALL LETTER L WITH STROKE")
    (323 . "LATIN CAPITAL LETTER N WITH ACUTE")
    (324 . "LATIN SMALL LETTER N WITH ACUTE")
    (325 . "LATIN CAPITAL LETTER N WITH CEDILLA")
    (326 . "LATIN SMALL LETTER N WITH CEDILLA")
    (327 . "LATIN CAPITAL LETTER N WITH CARON")
    (328 . "LATIN SMALL LETTER N WITH CARON")
    (329 . "LATIN SMALL LETTER N PRECEDED BY APOSTROPHE")
    (330 . "LATIN CAPITAL LETTER ENG")
    (331 . "LATIN SMALL LETTER ENG")
    (332 . "LATIN CAPITAL LETTER O WITH MACRON")
    (333 . "LATIN SMALL LETTER O WITH MACRON")
    (334 . "LATIN CAPITAL LETTER O WITH BREVE")
    (335 . "LATIN SMALL LETTER O WITH BREVE")
    (336 . "LATIN CAPITAL LETTER O WITH DOUBLE ACUTE")
    (337 . "LATIN SMALL LETTER O WITH DOUBLE ACUTE")
    (338 . "LATIN CAPITAL LETTER OE|LATIN CAPITAL LIGATURE OE")
    (339 . "LATIN SMALL LETTER OE|LATIN SMALL LIGATURE OE")
    (340 . "LATIN CAPITAL LETTER R WITH ACUTE")
    (341 . "LATIN SMALL LETTER R WITH ACUTE")
    (342 . "LATIN CAPITAL LETTER R WITH CEDILLA")
    (343 . "LATIN SMALL LETTER R WITH CEDILLA")
    (344 . "LATIN CAPITAL LETTER R WITH CARON")
    (345 . "LATIN SMALL LETTER R WITH CARON")
    (346 . "LATIN CAPITAL LETTER S WITH ACUTE")
    (347 . "LATIN SMALL LETTER S WITH ACUTE")
    (348 . "LATIN CAPITAL LETTER S WITH CIRCUMFLEX")
    (349 . "LATIN SMALL LETTER S WITH CIRCUMFLEX")
    (350 . "LATIN CAPITAL LETTER S WITH CEDILLA")
    (351 . "LATIN SMALL LETTER S WITH CEDILLA")
    (352 . "LATIN CAPITAL LETTER S WITH CARON")
    (353 . "LATIN SMALL LETTER S WITH CARON")
    (354 . "LATIN CAPITAL LETTER T WITH CEDILLA")
    (355 . "LATIN SMALL LETTER T WITH CEDILLA")
    (356 . "LATIN CAPITAL LETTER T WITH CARON")
    (357 . "LATIN SMALL LETTER T WITH CARON")
    (358 . "LATIN CAPITAL LETTER T WITH STROKE")
    (359 . "LATIN SMALL LETTER T WITH STROKE")
    (360 . "LATIN CAPITAL LETTER U WITH TILDE")
    (361 . "LATIN SMALL LETTER U WITH TILDE")
    (362 . "LATIN CAPITAL LETTER U WITH MACRON")
    (363 . "LATIN SMALL LETTER U WITH MACRON")
    (364 . "LATIN CAPITAL LETTER U WITH BREVE")
    (365 . "LATIN SMALL LETTER U WITH BREVE")
    (366 . "LATIN CAPITAL LETTER U WITH RING ABOVE")
    (367 . "LATIN SMALL LETTER U WITH RING ABOVE")
    (368 . "LATIN CAPITAL LETTER U WITH DOUBLE ACUTE")
    (369 . "LATIN SMALL LETTER U WITH DOUBLE ACUTE")
    (370 . "LATIN CAPITAL LETTER U WITH OGONEK")
    (371 . "LATIN SMALL LETTER U WITH OGONEK")
    (372 . "LATIN CAPITAL LETTER W WITH CIRCUMFLEX")
    (373 . "LATIN SMALL LETTER W WITH CIRCUMFLEX")
    (374 . "LATIN CAPITAL LETTER Y WITH CIRCUMFLEX")
    (375 . "LATIN SMALL LETTER Y WITH CIRCUMFLEX")
    (376 . "LATIN CAPITAL LETTER Y WITH DIAERESIS")
    (377 . "LATIN CAPITAL LETTER Z WITH ACUTE")
    (378 . "LATIN SMALL LETTER Z WITH ACUTE")
    (379 . "LATIN CAPITAL LETTER Z WITH DOT ABOVE")
    (380 . "LATIN SMALL LETTER Z WITH DOT ABOVE")
    (381 . "LATIN CAPITAL LETTER Z WITH CARON")
    (382 . "LATIN SMALL LETTER Z WITH CARON")
    (383 . "LATIN SMALL LETTER LONG S")
    (384 . "LATIN SMALL LETTER B WITH STROKE")
    (385 . "LATIN CAPITAL LETTER B WITH HOOK")
    (386 . "LATIN CAPITAL LETTER B WITH TOPBAR")
    (387 . "LATIN SMALL LETTER B WITH TOPBAR")
    (388 . "LATIN CAPITAL LETTER TONE SIX")
    (389 . "LATIN SMALL LETTER TONE SIX")
    (390 . "LATIN CAPITAL LETTER OPEN O")
    (391 . "LATIN CAPITAL LETTER C WITH HOOK")
    (392 . "LATIN SMALL LETTER C WITH HOOK")
    (393 . "LATIN CAPITAL LETTER AFRICAN D")
    (394 . "LATIN CAPITAL LETTER D WITH HOOK")
    (395 . "LATIN CAPITAL LETTER D WITH TOPBAR")
    (396 . "LATIN SMALL LETTER D WITH TOPBAR")
    (397 . "LATIN SMALL LETTER TURNED DELTA")
    (398 . "LATIN CAPITAL LETTER REVERSED E")
    (399 . "LATIN CAPITAL LETTER SCHWA")
    (400 . "LATIN CAPITAL LETTER OPEN E")
    (401 . "LATIN CAPITAL LETTER F WITH HOOK")
    (402 . "LATIN SMALL LETTER F WITH HOOK")
    (403 . "LATIN CAPITAL LETTER G WITH HOOK")
    (404 . "LATIN CAPITAL LETTER GAMMA")
    (405 . "LATIN SMALL LETTER HV")
    (406 . "LATIN CAPITAL LETTER IOTA")
    (407 . "LATIN CAPITAL LETTER I WITH STROKE")
    (408 . "LATIN CAPITAL LETTER K WITH HOOK")
    (409 . "LATIN SMALL LETTER K WITH HOOK")
    (410 . "LATIN SMALL LETTER L WITH BAR")
    (411 . "LATIN SMALL LETTER LAMBDA WITH STROKE")
    (412 . "LATIN CAPITAL LETTER TURNED M")
    (413 . "LATIN CAPITAL LETTER N WITH LEFT HOOK")
    (414 . "LATIN SMALL LETTER N WITH LONG RIGHT LEG")
    (415 . "LATIN CAPITAL LETTER O WITH MIDDLE TILDE")
    (416 . "LATIN CAPITAL LETTER O WITH HORN")
    (417 . "LATIN SMALL LETTER O WITH HORN")
    (418 . "LATIN CAPITAL LETTER OI")
    (419 . "LATIN SMALL LETTER OI")
    (420 . "LATIN CAPITAL LETTER P WITH HOOK")
    (421 . "LATIN SMALL LETTER P WITH HOOK")
    (422 . "LATIN LETTER YR")
    (423 . "LATIN CAPITAL LETTER TONE TWO")
    (424 . "LATIN SMALL LETTER TONE TWO")
    (425 . "LATIN CAPITAL LETTER ESH")
    (426 . "LATIN LETTER REVERSED ESH LOOP")
    (427 . "LATIN SMALL LETTER T WITH PALATAL HOOK")
    (428 . "LATIN CAPITAL LETTER T WITH HOOK")
    (429 . "LATIN SMALL LETTER T WITH HOOK")
    (430 . "LATIN CAPITAL LETTER T WITH RETROFLEX HOOK")
    (431 . "LATIN CAPITAL LETTER U WITH HORN")
    (432 . "LATIN SMALL LETTER U WITH HORN")
    (433 . "LATIN CAPITAL LETTER UPSILON")
    (434 . "LATIN CAPITAL LETTER V WITH HOOK")
    (435 . "LATIN CAPITAL LETTER Y WITH HOOK")
    (436 . "LATIN SMALL LETTER Y WITH HOOK")
    (437 . "LATIN CAPITAL LETTER Z WITH STROKE")
    (438 . "LATIN SMALL LETTER Z WITH STROKE")
    (439 . "LATIN CAPITAL LETTER EZH")
    (440 . "LATIN CAPITAL LETTER EZH REVERSED")
    (441 . "LATIN SMALL LETTER EZH REVERSED")
    (442 . "LATIN SMALL LETTER EZH WITH TAIL")
    (443 . "LATIN LETTER TWO WITH STROKE")
    (444 . "LATIN CAPITAL LETTER TONE FIVE")
    (445 . "LATIN SMALL LETTER TONE FIVE")
    (446 . "LATIN LETTER INVERTED GLOTTAL STOP WITH STROKE")
    (447 . "LATIN LETTER WYNN")
    (448 . "LATIN LETTER DENTAL CLICK")
    (449 . "LATIN LETTER LATERAL CLICK")
    (450 . "LATIN LETTER ALVEOLAR CLICK")
    (451 . "LATIN LETTER RETROFLEX CLICK")
    (452 . "LATIN CAPITAL LETTER DZ WITH CARON")
    (453 . "LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON")
    (454 . "LATIN SMALL LETTER DZ WITH CARON")
    (455 . "LATIN CAPITAL LETTER LJ")
    (456 . "LATIN CAPITAL LETTER L WITH SMALL LETTER J")
    (457 . "LATIN SMALL LETTER LJ")
    (458 . "LATIN CAPITAL LETTER NJ")
    (459 . "LATIN CAPITAL LETTER N WITH SMALL LETTER J")
    (460 . "LATIN SMALL LETTER NJ")
    (461 . "LATIN CAPITAL LETTER A WITH CARON")
    (462 . "LATIN SMALL LETTER A WITH CARON")
    (463 . "LATIN CAPITAL LETTER I WITH CARON")
    (464 . "LATIN SMALL LETTER I WITH CARON")
    (465 . "LATIN CAPITAL LETTER O WITH CARON")
    (466 . "LATIN SMALL LETTER O WITH CARON")
    (467 . "LATIN CAPITAL LETTER U WITH CARON")
    (468 . "LATIN SMALL LETTER U WITH CARON")
    (469 . "LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON")
    (470 . "LATIN SMALL LETTER U WITH DIAERESIS AND MACRON")
    (471 . "LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE")
    (472 . "LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE")
    (473 . "LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON")
    (474 . "LATIN SMALL LETTER U WITH DIAERESIS AND CARON")
    (475 . "LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE")
    (476 . "LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE")
    (477 . "LATIN SMALL LETTER TURNED E")
    (478 . "LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON")
    (479 . "LATIN SMALL LETTER A WITH DIAERESIS AND MACRON")
    (480 . "LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON")
    (481 . "LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON")
    (482 . "LATIN CAPITAL LETTER AE WITH MACRON|LATIN CAPITAL LIGATURE AE WITH MACRON")
    (483 . "LATIN SMALL LETTER AE WITH MACRON|LATIN SMALL LIGATURE AE WITH MACRON")
    (484 . "LATIN CAPITAL LETTER G WITH STROKE")
    (485 . "LATIN SMALL LETTER G WITH STROKE")
    (486 . "LATIN CAPITAL LETTER G WITH CARON")
    (487 . "LATIN SMALL LETTER G WITH CARON")
    (488 . "LATIN CAPITAL LETTER K WITH CARON")
    (489 . "LATIN SMALL LETTER K WITH CARON")
    (490 . "LATIN CAPITAL LETTER O WITH OGONEK")
    (491 . "LATIN SMALL LETTER O WITH OGONEK")
    (492 . "LATIN CAPITAL LETTER O WITH OGONEK AND MACRON")
    (493 . "LATIN SMALL LETTER O WITH OGONEK AND MACRON")
    (494 . "LATIN CAPITAL LETTER EZH WITH CARON")
    (495 . "LATIN SMALL LETTER EZH WITH CARON")
    (496 . "LATIN SMALL LETTER J WITH CARON")
    (497 . "LATIN CAPITAL LETTER DZ")
    (498 . "LATIN CAPITAL LETTER D WITH SMALL LETTER Z")
    (499 . "LATIN SMALL LETTER DZ")
    (500 . "LATIN CAPITAL LETTER G WITH ACUTE")
    (501 . "LATIN SMALL LETTER G WITH ACUTE")
    (506 . "LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE")
    (507 . "LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE")
    (508 . "LATIN CAPITAL LETTER AE WITH ACUTE|LATIN CAPITAL LIGATURE AE WITH ACUTE")
    (509 . "LATIN SMALL LETTER AE WITH ACUTE|LATIN SMALL LIGATURE AE WITH ACUTE")
    (510 . "LATIN CAPITAL LETTER O WITH STROKE AND ACUTE")
    (511 . "LATIN SMALL LETTER O WITH STROKE AND ACUTE")
    (512 . "LATIN CAPITAL LETTER A WITH DOUBLE GRAVE")
    (513 . "LATIN SMALL LETTER A WITH DOUBLE GRAVE")
    (514 . "LATIN CAPITAL LETTER A WITH INVERTED BREVE")
    (515 . "LATIN SMALL LETTER A WITH INVERTED BREVE")
    (516 . "LATIN CAPITAL LETTER E WITH DOUBLE GRAVE")
    (517 . "LATIN SMALL LETTER E WITH DOUBLE GRAVE")
    (518 . "LATIN CAPITAL LETTER E WITH INVERTED BREVE")
    (519 . "LATIN SMALL LETTER E WITH INVERTED BREVE")
    (520 . "LATIN CAPITAL LETTER I WITH DOUBLE GRAVE")
    (521 . "LATIN SMALL LETTER I WITH DOUBLE GRAVE")
    (522 . "LATIN CAPITAL LETTER I WITH INVERTED BREVE")
    (523 . "LATIN SMALL LETTER I WITH INVERTED BREVE")
    (524 . "LATIN CAPITAL LETTER O WITH DOUBLE GRAVE")
    (525 . "LATIN SMALL LETTER O WITH DOUBLE GRAVE")
    (526 . "LATIN CAPITAL LETTER O WITH INVERTED BREVE")
    (527 . "LATIN SMALL LETTER O WITH INVERTED BREVE")
    (528 . "LATIN CAPITAL LETTER R WITH DOUBLE GRAVE")
    (529 . "LATIN SMALL LETTER R WITH DOUBLE GRAVE")
    (530 . "LATIN CAPITAL LETTER R WITH INVERTED BREVE")
    (531 . "LATIN SMALL LETTER R WITH INVERTED BREVE")
    (532 . "LATIN CAPITAL LETTER U WITH DOUBLE GRAVE")
    (533 . "LATIN SMALL LETTER U WITH DOUBLE GRAVE")
    (534 . "LATIN CAPITAL LETTER U WITH INVERTED BREVE")
    (535 . "LATIN SMALL LETTER U WITH INVERTED BREVE")
    (592 . "LATIN SMALL LETTER TURNED A")
    (593 . "LATIN SMALL LETTER ALPHA")
    (594 . "LATIN SMALL LETTER TURNED ALPHA")
    (595 . "LATIN SMALL LETTER B WITH HOOK")
    (596 . "LATIN SMALL LETTER OPEN O")
    (597 . "LATIN SMALL LETTER C WITH CURL")
    (598 . "LATIN SMALL LETTER D WITH TAIL")
    (599 . "LATIN SMALL LETTER D WITH HOOK")
    (600 . "LATIN SMALL LETTER REVERSED E")
    (601 . "LATIN SMALL LETTER SCHWA")
    (602 . "LATIN SMALL LETTER SCHWA WITH HOOK")
    (603 . "LATIN SMALL LETTER OPEN E")
    (604 . "LATIN SMALL LETTER REVERSED OPEN E")
    (605 . "LATIN SMALL LETTER REVERSED OPEN E WITH HOOK")
    (606 . "LATIN SMALL LETTER CLOSED REVERSED OPEN E")
    (607 . "LATIN SMALL LETTER DOTLESS J WITH STROKE")
    (608 . "LATIN SMALL LETTER G WITH HOOK")
    (609 . "LATIN SMALL LETTER SCRIPT G")
    (610 . "LATIN LETTER SMALL CAPITAL G")
    (611 . "LATIN SMALL LETTER GAMMA")
    (612 . "LATIN SMALL LETTER RAMS HORN")
    (613 . "LATIN SMALL LETTER TURNED H")
    (614 . "LATIN SMALL LETTER H WITH HOOK")
    (615 . "LATIN SMALL LETTER HENG WITH HOOK")
    (616 . "LATIN SMALL LETTER I WITH STROKE")
    (617 . "LATIN SMALL LETTER IOTA")
    (618 . "LATIN LETTER SMALL CAPITAL I")
    (619 . "LATIN SMALL LETTER L WITH MIDDLE TILDE")
    (620 . "LATIN SMALL LETTER L WITH BELT")
    (621 . "LATIN SMALL LETTER L WITH RETROFLEX HOOK")
    (622 . "LATIN SMALL LETTER LEZH")
    (623 . "LATIN SMALL LETTER TURNED M")
    (624 . "LATIN SMALL LETTER TURNED M WITH LONG LEG")
    (625 . "LATIN SMALL LETTER M WITH HOOK")
    (626 . "LATIN SMALL LETTER N WITH LEFT HOOK")
    (627 . "LATIN SMALL LETTER N WITH RETROFLEX HOOK")
    (628 . "LATIN LETTER SMALL CAPITAL N")
    (629 . "LATIN SMALL LETTER BARRED O")
    (630 . "LATIN LETTER SMALL CAPITAL OE")
    (631 . "LATIN SMALL LETTER CLOSED OMEGA")
    (632 . "LATIN SMALL LETTER PHI")
    (633 . "LATIN SMALL LETTER TURNED R")
    (634 . "LATIN SMALL LETTER TURNED R WITH LONG LEG")
    (635 . "LATIN SMALL LETTER TURNED R WITH HOOK")
    (636 . "LATIN SMALL LETTER R WITH LONG LEG")
    (637 . "LATIN SMALL LETTER R WITH TAIL")
    (638 . "LATIN SMALL LETTER R WITH FISHHOOK")
    (639 . "LATIN SMALL LETTER REVERSED R WITH FISHHOOK")
    (640 . "LATIN LETTER SMALL CAPITAL R")
    (641 . "LATIN LETTER SMALL CAPITAL INVERTED R")
    (642 . "LATIN SMALL LETTER S WITH HOOK")
    (643 . "LATIN SMALL LETTER ESH")
    (644 . "LATIN SMALL LETTER DOTLESS J WITH STROKE AND HOOK")
    (645 . "LATIN SMALL LETTER SQUAT REVERSED ESH")
    (646 . "LATIN SMALL LETTER ESH WITH CURL")
    (647 . "LATIN SMALL LETTER TURNED T")
    (648 . "LATIN SMALL LETTER T WITH RETROFLEX HOOK")
    (649 . "LATIN SMALL LETTER U BAR")
    (650 . "LATIN SMALL LETTER UPSILON")
    (651 . "LATIN SMALL LETTER V WITH HOOK")
    (652 . "LATIN SMALL LETTER TURNED V")
    (653 . "LATIN SMALL LETTER TURNED W")
    (654 . "LATIN SMALL LETTER TURNED Y")
    (655 . "LATIN LETTER SMALL CAPITAL Y")
    (656 . "LATIN SMALL LETTER Z WITH RETROFLEX HOOK")
    (657 . "LATIN SMALL LETTER Z WITH CURL")
    (658 . "LATIN SMALL LETTER EZH")
    (659 . "LATIN SMALL LETTER EZH WITH CURL")
    (660 . "LATIN LETTER GLOTTAL STOP")
    (661 . "LATIN LETTER PHARYNGEAL VOICED FRICATIVE")
    (662 . "LATIN LETTER INVERTED GLOTTAL STOP")
    (663 . "LATIN LETTER STRETCHED C")
    (664 . "LATIN LETTER BILABIAL CLICK")
    (665 . "LATIN LETTER SMALL CAPITAL B")
    (666 . "LATIN SMALL LETTER CLOSED OPEN E")
    (667 . "LATIN LETTER SMALL CAPITAL G WITH HOOK")
    (668 . "LATIN LETTER SMALL CAPITAL H")
    (669 . "LATIN SMALL LETTER J WITH CROSSED-TAIL")
    (670 . "LATIN SMALL LETTER TURNED K")
    (671 . "LATIN LETTER SMALL CAPITAL L")
    (672 . "LATIN SMALL LETTER Q WITH HOOK")
    (673 . "LATIN LETTER GLOTTAL STOP WITH STROKE")
    (674 . "LATIN LETTER REVERSED GLOTTAL STOP WITH STROKE")
    (675 . "LATIN SMALL LETTER DZ DIGRAPH")
    (676 . "LATIN SMALL LETTER DEZH DIGRAPH")
    (677 . "LATIN SMALL LETTER DZ DIGRAPH WITH CURL")
    (678 . "LATIN SMALL LETTER TS DIGRAPH")
    (679 . "LATIN SMALL LETTER TESH DIGRAPH")
    (680 . "LATIN SMALL LETTER TC DIGRAPH WITH CURL")
    (688 . "MODIFIER LETTER SMALL H")
    (689 . "MODIFIER LETTER SMALL H WITH HOOK")
    (690 . "MODIFIER LETTER SMALL J")
    (691 . "MODIFIER LETTER SMALL R")
    (692 . "MODIFIER LETTER SMALL TURNED R")
    (693 . "MODIFIER LETTER SMALL TURNED R WITH HOOK")
    (694 . "MODIFIER LETTER SMALL CAPITAL INVERTED R")
    (695 . "MODIFIER LETTER SMALL W")
    (696 . "MODIFIER LETTER SMALL Y")
    (697 . "MODIFIER LETTER PRIME")
    (698 . "MODIFIER LETTER DOUBLE PRIME")
    (699 . "MODIFIER LETTER TURNED COMMA")
    (700 . "MODIFIER LETTER APOSTROPHE")
    (701 . "MODIFIER LETTER REVERSED COMMA")
    (702 . "MODIFIER LETTER RIGHT HALF RING")
    (703 . "MODIFIER LETTER LEFT HALF RING")
    (704 . "MODIFIER LETTER GLOTTAL STOP")
    (705 . "MODIFIER LETTER REVERSED GLOTTAL STOP")
    (706 . "MODIFIER LETTER LEFT ARROWHEAD")
    (707 . "MODIFIER LETTER RIGHT ARROWHEAD")
    (708 . "MODIFIER LETTER UP ARROWHEAD")
    (709 . "MODIFIER LETTER DOWN ARROWHEAD")
    (710 . "MODIFIER LETTER CIRCUMFLEX ACCENT")
    (711 . "CARON")
    (712 . "MODIFIER LETTER VERTICAL LINE")
    (713 . "MODIFIER LETTER MACRON")
    (714 . "MODIFIER LETTER ACUTE ACCENT")
    (715 . "MODIFIER LETTER GRAVE ACCENT")
    (716 . "MODIFIER LETTER LOW VERTICAL LINE")
    (717 . "MODIFIER LETTER LOW MACRON")
    (718 . "MODIFIER LETTER LOW GRAVE ACCENT")
    (719 . "MODIFIER LETTER LOW ACUTE ACCENT")
    (720 . "MODIFIER LETTER TRIANGULAR COLON")
    (721 . "MODIFIER LETTER HALF TRIANGULAR COLON")
    (722 . "MODIFIER LETTER CENTERED RIGHT HALF RING|MODIFIER LETTER CENTRED RIGHT HALF RING")
    (723 . "MODIFIER LETTER CENTERED LEFT HALF RING|MODIFIER LETTER CENTRED LEFT HALF RING")
    (724 . "MODIFIER LETTER UP TACK")
    (725 . "MODIFIER LETTER DOWN TACK")
    (726 . "MODIFIER LETTER PLUS SIGN")
    (727 . "MODIFIER LETTER MINUS SIGN")
    (728 . "BREVE")
    (729 . "DOT ABOVE")
    (730 . "RING ABOVE")
    (731 . "OGONEK")
    (732 . "SMALL TILDE")
    (733 . "DOUBLE ACUTE ACCENT")
    (734 . "MODIFIER LETTER RHOTIC HOOK")
    (736 . "MODIFIER LETTER SMALL GAMMA")
    (737 . "MODIFIER LETTER SMALL L")
    (738 . "MODIFIER LETTER SMALL S")
    (739 . "MODIFIER LETTER SMALL X")
    (740 . "MODIFIER LETTER SMALL REVERSED GLOTTAL STOP")
    (741 . "MODIFIER LETTER EXTRA-HIGH TONE BAR")
    (742 . "MODIFIER LETTER HIGH TONE BAR")
    (743 . "MODIFIER LETTER MID TONE BAR")
    (744 . "MODIFIER LETTER LOW TONE BAR")
    (745 . "MODIFIER LETTER EXTRA-LOW TONE BAR")
    (768 . "COMBINING GRAVE ACCENT")
    (769 . "COMBINING ACUTE ACCENT")
    (770 . "COMBINING CIRCUMFLEX ACCENT")
    (771 . "COMBINING TILDE")
    (772 . "COMBINING MACRON")
    (773 . "COMBINING OVERLINE")
    (774 . "COMBINING BREVE")
    (775 . "COMBINING DOT ABOVE")
    (776 . "COMBINING DIAERESIS")
    (777 . "COMBINING HOOK ABOVE")
    (778 . "COMBINING RING ABOVE")
    (779 . "COMBINING DOUBLE ACUTE ACCENT")
    (780 . "COMBINING CARON")
    (781 . "COMBINING VERTICAL LINE ABOVE")
    (782 . "COMBINING DOUBLE VERTICAL LINE ABOVE")
    (783 . "COMBINING DOUBLE GRAVE ACCENT")
    (784 . "COMBINING CANDRABINDU")
    (785 . "COMBINING INVERTED BREVE")
    (786 . "COMBINING TURNED COMMA ABOVE")
    (787 . "COMBINING COMMA ABOVE")
    (788 . "COMBINING REVERSED COMMA ABOVE")
    (789 . "COMBINING COMMA ABOVE RIGHT")
    (790 . "COMBINING GRAVE ACCENT BELOW")
    (791 . "COMBINING ACUTE ACCENT BELOW")
    (792 . "COMBINING LEFT TACK BELOW")
    (793 . "COMBINING RIGHT TACK BELOW")
    (794 . "COMBINING LEFT ANGLE ABOVE")
    (795 . "COMBINING HORN")
    (796 . "COMBINING LEFT HALF RING BELOW")
    (797 . "COMBINING UP TACK BELOW")
    (798 . "COMBINING DOWN TACK BELOW")
    (799 . "COMBINING PLUS SIGN BELOW")
    (800 . "COMBINING MINUS SIGN BELOW")
    (801 . "COMBINING PALATALIZED HOOK BELOW")
    (802 . "COMBINING RETROFLEX HOOK BELOW")
    (803 . "COMBINING DOT BELOW")
    (804 . "COMBINING DIAERESIS BELOW")
    (805 . "COMBINING RING BELOW")
    (806 . "COMBINING COMMA BELOW")
    (807 . "COMBINING CEDILLA")
    (808 . "COMBINING OGONEK")
    (809 . "COMBINING VERTICAL LINE BELOW")
    (810 . "COMBINING BRIDGE BELOW")
    (811 . "COMBINING INVERTED DOUBLE ARCH BELOW")
    (812 . "COMBINING CARON BELOW")
    (813 . "COMBINING CIRCUMFLEX ACCENT BELOW")
    (814 . "COMBINING BREVE BELOW")
    (815 . "COMBINING INVERTED BREVE BELOW")
    (816 . "COMBINING TILDE BELOW")
    (817 . "COMBINING MACRON BELOW")
    (818 . "COMBINING LOW LINE")
    (819 . "COMBINING DOUBLE LOW LINE")
    (820 . "COMBINING TILDE OVERLAY")
    (821 . "COMBINING SHORT STROKE OVERLAY")
    (822 . "COMBINING LONG STROKE OVERLAY")
    (823 . "COMBINING SHORT SOLIDUS OVERLAY")
    (824 . "COMBINING LONG SOLIDUS OVERLAY")
    (825 . "COMBINING RIGHT HALF RING BELOW")
    (826 . "COMBINING INVERTED BRIDGE BELOW")
    (827 . "COMBINING SQUARE BELOW")
    (828 . "COMBINING SEAGULL BELOW")
    (829 . "COMBINING X ABOVE")
    (830 . "COMBINING VERTICAL TILDE")
    (831 . "COMBINING DOUBLE OVERLINE")
    (832 . "COMBINING GRAVE TONE MARK")
    (833 . "COMBINING ACUTE TONE MARK")
    (834 . "COMBINING GREEK PERISPOMENI")
    (835 . "COMBINING GREEK KORONIS")
    (836 . "COMBINING GREEK DIALYTIKA TONOS")
    (837 . "COMBINING GREEK YPOGEGRAMMENI")
    (864 . "COMBINING DOUBLE TILDE")
    (865 . "COMBINING DOUBLE INVERTED BREVE")
    (884 . "GREEK NUMERAL SIGN")
    (885 . "GREEK LOWER NUMERAL SIGN")
    (890 . "GREEK YPOGEGRAMMENI")
    (894 . "GREEK QUESTION MARK")
    (900 . "GREEK TONOS")
    (901 . "GREEK DIALYTIKA TONOS")
    (902 . "GREEK CAPITAL LETTER ALPHA WITH TONOS")
    (903 . "GREEK ANO TELEIA")
    (904 . "GREEK CAPITAL LETTER EPSILON WITH TONOS")
    (905 . "GREEK CAPITAL LETTER ETA WITH TONOS")
    (906 . "GREEK CAPITAL LETTER IOTA WITH TONOS")
    (908 . "GREEK CAPITAL LETTER OMICRON WITH TONOS")
    (910 . "GREEK CAPITAL LETTER UPSILON WITH TONOS")
    (911 . "GREEK CAPITAL LETTER OMEGA WITH TONOS")
    (912 . "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS")
    (913 . "GREEK CAPITAL LETTER ALPHA")
    (914 . "GREEK CAPITAL LETTER BETA")
    (915 . "GREEK CAPITAL LETTER GAMMA")
    (916 . "GREEK CAPITAL LETTER DELTA")
    (917 . "GREEK CAPITAL LETTER EPSILON")
    (918 . "GREEK CAPITAL LETTER ZETA")
    (919 . "GREEK CAPITAL LETTER ETA")
    (920 . "GREEK CAPITAL LETTER THETA")
    (921 . "GREEK CAPITAL LETTER IOTA")
    (922 . "GREEK CAPITAL LETTER KAPPA")
    (923 . "GREEK CAPITAL LETTER LAMDA")
    (924 . "GREEK CAPITAL LETTER MU")
    (925 . "GREEK CAPITAL LETTER NU")
    (926 . "GREEK CAPITAL LETTER XI")
    (927 . "GREEK CAPITAL LETTER OMICRON")
    (928 . "GREEK CAPITAL LETTER PI")
    (929 . "GREEK CAPITAL LETTER RHO")
    (931 . "GREEK CAPITAL LETTER SIGMA")
    (932 . "GREEK CAPITAL LETTER TAU")
    (933 . "GREEK CAPITAL LETTER UPSILON")
    (934 . "GREEK CAPITAL LETTER PHI")
    (935 . "GREEK CAPITAL LETTER CHI")
    (936 . "GREEK CAPITAL LETTER PSI")
    (937 . "GREEK CAPITAL LETTER OMEGA")
    (938 . "GREEK CAPITAL LETTER IOTA WITH DIALYTIKA")
    (939 . "GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA")
    (940 . "GREEK SMALL LETTER ALPHA WITH TONOS")
    (941 . "GREEK SMALL LETTER EPSILON WITH TONOS")
    (942 . "GREEK SMALL LETTER ETA WITH TONOS")
    (943 . "GREEK SMALL LETTER IOTA WITH TONOS")
    (944 . "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS")
    (945 . "GREEK SMALL LETTER ALPHA")
    (946 . "GREEK SMALL LETTER BETA")
    (947 . "GREEK SMALL LETTER GAMMA")
    (948 . "GREEK SMALL LETTER DELTA")
    (949 . "GREEK SMALL LETTER EPSILON")
    (950 . "GREEK SMALL LETTER ZETA")
    (951 . "GREEK SMALL LETTER ETA")
    (952 . "GREEK SMALL LETTER THETA")
    (953 . "GREEK SMALL LETTER IOTA")
    (954 . "GREEK SMALL LETTER KAPPA")
    (955 . "GREEK SMALL LETTER LAMDA")
    (956 . "GREEK SMALL LETTER MU")
    (957 . "GREEK SMALL LETTER NU")
    (958 . "GREEK SMALL LETTER XI")
    (959 . "GREEK SMALL LETTER OMICRON")
    (960 . "GREEK SMALL LETTER PI")
    (961 . "GREEK SMALL LETTER RHO")
    (962 . "GREEK SMALL LETTER FINAL SIGMA")
    (963 . "GREEK SMALL LETTER SIGMA")
    (964 . "GREEK SMALL LETTER TAU")
    (965 . "GREEK SMALL LETTER UPSILON")
    (966 . "GREEK SMALL LETTER PHI")
    (967 . "GREEK SMALL LETTER CHI")
    (968 . "GREEK SMALL LETTER PSI")
    (969 . "GREEK SMALL LETTER OMEGA")
    (970 . "GREEK SMALL LETTER IOTA WITH DIALYTIKA")
    (971 . "GREEK SMALL LETTER UPSILON WITH DIALYTIKA")
    (972 . "GREEK SMALL LETTER OMICRON WITH TONOS")
    (973 . "GREEK SMALL LETTER UPSILON WITH TONOS")
    (974 . "GREEK SMALL LETTER OMEGA WITH TONOS")
    (976 . "GREEK BETA SYMBOL")
    (977 . "GREEK THETA SYMBOL")
    (978 . "GREEK UPSILON WITH HOOK SYMBOL")
    (979 . "GREEK UPSILON WITH ACUTE AND HOOK SYMBOL")
    (980 . "GREEK UPSILON WITH DIAERESIS AND HOOK SYMBOL")
    (981 . "GREEK PHI SYMBOL")
    (982 . "GREEK PI SYMBOL")
    (986 . "GREEK LETTER SIGMA")
    (988 . "GREEK LETTER DIGAMMA")
    (990 . "GREEK LETTER KOPPA")
    (992 . "GREEK LETTER SAMPI")
    (994 . "COPTIC CAPITAL LETTER SHEI")
    (995 . "COPTIC SMALL LETTER SHEI")
    (996 . "COPTIC CAPITAL LETTER FEI")
    (997 . "COPTIC SMALL LETTER FEI")
    (998 . "COPTIC CAPITAL LETTER KHEI")
    (999 . "COPTIC SMALL LETTER KHEI")
    (1000 . "COPTIC CAPITAL LETTER HORI")
    (1001 . "COPTIC SMALL LETTER HORI")
    (1002 . "COPTIC CAPITAL LETTER GANGIA")
    (1003 . "COPTIC SMALL LETTER GANGIA")
    (1004 . "COPTIC CAPITAL LETTER SHIMA")
    (1005 . "COPTIC SMALL LETTER SHIMA")
    (1006 . "COPTIC CAPITAL LETTER DEI")
    (1007 . "COPTIC SMALL LETTER DEI")
    (1008 . "GREEK KAPPA SYMBOL")
    (1009 . "GREEK RHO SYMBOL")
    (1010 . "GREEK LUNATE SIGMA SYMBOL")
    (1011 . "GREEK LETTER YOT")
    (1025 . "CYRILLIC CAPITAL LETTER IO")
    (1026 . "CYRILLIC CAPITAL LETTER DJE")
    (1027 . "CYRILLIC CAPITAL LETTER GJE")
    (1028 . "CYRILLIC CAPITAL LETTER UKRAINIAN IE")
    (1029 . "CYRILLIC CAPITAL LETTER DZE")
    (1030 . "CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I")
    (1031 . "CYRILLIC CAPITAL LETTER YI")
    (1032 . "CYRILLIC CAPITAL LETTER JE")
    (1033 . "CYRILLIC CAPITAL LETTER LJE")
    (1034 . "CYRILLIC CAPITAL LETTER NJE")
    (1035 . "CYRILLIC CAPITAL LETTER TSHE")
    (1036 . "CYRILLIC CAPITAL LETTER KJE")
    (1038 . "CYRILLIC CAPITAL LETTER SHORT U")
    (1039 . "CYRILLIC CAPITAL LETTER DZHE")
    (1040 . "CYRILLIC CAPITAL LETTER A")
    (1041 . "CYRILLIC CAPITAL LETTER BE")
    (1042 . "CYRILLIC CAPITAL LETTER VE")
    (1043 . "CYRILLIC CAPITAL LETTER GHE")
    (1044 . "CYRILLIC CAPITAL LETTER DE")
    (1045 . "CYRILLIC CAPITAL LETTER IE")
    (1046 . "CYRILLIC CAPITAL LETTER ZHE")
    (1047 . "CYRILLIC CAPITAL LETTER ZE")
    (1048 . "CYRILLIC CAPITAL LETTER I")
    (1049 . "CYRILLIC CAPITAL LETTER SHORT I")
    (1050 . "CYRILLIC CAPITAL LETTER KA")
    (1051 . "CYRILLIC CAPITAL LETTER EL")
    (1052 . "CYRILLIC CAPITAL LETTER EM")
    (1053 . "CYRILLIC CAPITAL LETTER EN")
    (1054 . "CYRILLIC CAPITAL LETTER O")
    (1055 . "CYRILLIC CAPITAL LETTER PE")
    (1056 . "CYRILLIC CAPITAL LETTER ER")
    (1057 . "CYRILLIC CAPITAL LETTER ES")
    (1058 . "CYRILLIC CAPITAL LETTER TE")
    (1059 . "CYRILLIC CAPITAL LETTER U")
    (1060 . "CYRILLIC CAPITAL LETTER EF")
    (1061 . "CYRILLIC CAPITAL LETTER HA")
    (1062 . "CYRILLIC CAPITAL LETTER TSE")
    (1063 . "CYRILLIC CAPITAL LETTER CHE")
    (1064 . "CYRILLIC CAPITAL LETTER SHA")
    (1065 . "CYRILLIC CAPITAL LETTER SHCHA")
    (1066 . "CYRILLIC CAPITAL LETTER HARD SIGN")
    (1067 . "CYRILLIC CAPITAL LETTER YERU")
    (1068 . "CYRILLIC CAPITAL LETTER SOFT SIGN")
    (1069 . "CYRILLIC CAPITAL LETTER E")
    (1070 . "CYRILLIC CAPITAL LETTER YU")
    (1071 . "CYRILLIC CAPITAL LETTER YA")
    (1072 . "CYRILLIC SMALL LETTER A")
    (1073 . "CYRILLIC SMALL LETTER BE")
    (1074 . "CYRILLIC SMALL LETTER VE")
    (1075 . "CYRILLIC SMALL LETTER GHE")
    (1076 . "CYRILLIC SMALL LETTER DE")
    (1077 . "CYRILLIC SMALL LETTER IE")
    (1078 . "CYRILLIC SMALL LETTER ZHE")
    (1079 . "CYRILLIC SMALL LETTER ZE")
    (1080 . "CYRILLIC SMALL LETTER I")
    (1081 . "CYRILLIC SMALL LETTER SHORT I")
    (1082 . "CYRILLIC SMALL LETTER KA")
    (1083 . "CYRILLIC SMALL LETTER EL")
    (1084 . "CYRILLIC SMALL LETTER EM")
    (1085 . "CYRILLIC SMALL LETTER EN")
    (1086 . "CYRILLIC SMALL LETTER O")
    (1087 . "CYRILLIC SMALL LETTER PE")
    (1088 . "CYRILLIC SMALL LETTER ER")
    (1089 . "CYRILLIC SMALL LETTER ES")
    (1090 . "CYRILLIC SMALL LETTER TE")
    (1091 . "CYRILLIC SMALL LETTER U")
    (1092 . "CYRILLIC SMALL LETTER EF")
    (1093 . "CYRILLIC SMALL LETTER HA")
    (1094 . "CYRILLIC SMALL LETTER TSE")
    (1095 . "CYRILLIC SMALL LETTER CHE")
    (1096 . "CYRILLIC SMALL LETTER SHA")
    (1097 . "CYRILLIC SMALL LETTER SHCHA")
    (1098 . "CYRILLIC SMALL LETTER HARD SIGN")
    (1099 . "CYRILLIC SMALL LETTER YERU")
    (1100 . "CYRILLIC SMALL LETTER SOFT SIGN")
    (1101 . "CYRILLIC SMALL LETTER E")
    (1102 . "CYRILLIC SMALL LETTER YU")
    (1103 . "CYRILLIC SMALL LETTER YA")
    (1105 . "CYRILLIC SMALL LETTER IO")
    (1106 . "CYRILLIC SMALL LETTER DJE")
    (1107 . "CYRILLIC SMALL LETTER GJE")
    (1108 . "CYRILLIC SMALL LETTER UKRAINIAN IE")
    (1109 . "CYRILLIC SMALL LETTER DZE")
    (1110 . "CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I")
    (1111 . "CYRILLIC SMALL LETTER YI")
    (1112 . "CYRILLIC SMALL LETTER JE")
    (1113 . "CYRILLIC SMALL LETTER LJE")
    (1114 . "CYRILLIC SMALL LETTER NJE")
    (1115 . "CYRILLIC SMALL LETTER TSHE")
    (1116 . "CYRILLIC SMALL LETTER KJE")
    (1118 . "CYRILLIC SMALL LETTER SHORT U")
    (1119 . "CYRILLIC SMALL LETTER DZHE")
    (1120 . "CYRILLIC CAPITAL LETTER OMEGA")
    (1121 . "CYRILLIC SMALL LETTER OMEGA")
    (1122 . "CYRILLIC CAPITAL LETTER YAT")
    (1123 . "CYRILLIC SMALL LETTER YAT")
    (1124 . "CYRILLIC CAPITAL LETTER IOTIFIED E")
    (1125 . "CYRILLIC SMALL LETTER IOTIFIED E")
    (1126 . "CYRILLIC CAPITAL LETTER LITTLE YUS")
    (1127 . "CYRILLIC SMALL LETTER LITTLE YUS")
    (1128 . "CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS")
    (1129 . "CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS")
    (1130 . "CYRILLIC CAPITAL LETTER BIG YUS")
    (1131 . "CYRILLIC SMALL LETTER BIG YUS")
    (1132 . "CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS")
    (1133 . "CYRILLIC SMALL LETTER IOTIFIED BIG YUS")
    (1134 . "CYRILLIC CAPITAL LETTER KSI")
    (1135 . "CYRILLIC SMALL LETTER KSI")
    (1136 . "CYRILLIC CAPITAL LETTER PSI")
    (1137 . "CYRILLIC SMALL LETTER PSI")
    (1138 . "CYRILLIC CAPITAL LETTER FITA")
    (1139 . "CYRILLIC SMALL LETTER FITA")
    (1140 . "CYRILLIC CAPITAL LETTER IZHITSA")
    (1141 . "CYRILLIC SMALL LETTER IZHITSA")
    (1142 . "CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT")
    (1143 . "CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT")
    (1144 . "CYRILLIC CAPITAL LETTER UK")
    (1145 . "CYRILLIC SMALL LETTER UK")
    (1146 . "CYRILLIC CAPITAL LETTER ROUND OMEGA")
    (1147 . "CYRILLIC SMALL LETTER ROUND OMEGA")
    (1148 . "CYRILLIC CAPITAL LETTER OMEGA WITH TITLO")
    (1149 . "CYRILLIC SMALL LETTER OMEGA WITH TITLO")
    (1150 . "CYRILLIC CAPITAL LETTER OT")
    (1151 . "CYRILLIC SMALL LETTER OT")
    (1152 . "CYRILLIC CAPITAL LETTER KOPPA")
    (1153 . "CYRILLIC SMALL LETTER KOPPA")
    (1154 . "CYRILLIC THOUSANDS SIGN")
    (1155 . "COMBINING CYRILLIC TITLO")
    (1156 . "COMBINING CYRILLIC PALATALIZATION")
    (1157 . "COMBINING CYRILLIC DASIA PNEUMATA")
    (1158 . "COMBINING CYRILLIC PSILI PNEUMATA")
    (1168 . "CYRILLIC CAPITAL LETTER GHE WITH UPTURN")
    (1169 . "CYRILLIC SMALL LETTER GHE WITH UPTURN")
    (1170 . "CYRILLIC CAPITAL LETTER GHE WITH STROKE")
    (1171 . "CYRILLIC SMALL LETTER GHE WITH STROKE")
    (1172 . "CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK")
    (1173 . "CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK")
    (1174 . "CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER")
    (1175 . "CYRILLIC SMALL LETTER ZHE WITH DESCENDER")
    (1176 . "CYRILLIC CAPITAL LETTER ZE WITH DESCENDER")
    (1177 . "CYRILLIC SMALL LETTER ZE WITH DESCENDER")
    (1178 . "CYRILLIC CAPITAL LETTER KA WITH DESCENDER")
    (1179 . "CYRILLIC SMALL LETTER KA WITH DESCENDER")
    (1180 . "CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE")
    (1181 . "CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE")
    (1182 . "CYRILLIC CAPITAL LETTER KA WITH STROKE")
    (1183 . "CYRILLIC SMALL LETTER KA WITH STROKE")
    (1184 . "CYRILLIC CAPITAL LETTER BASHKIR KA")
    (1185 . "CYRILLIC SMALL LETTER BASHKIR KA")
    (1186 . "CYRILLIC CAPITAL LETTER EN WITH DESCENDER")
    (1187 . "CYRILLIC SMALL LETTER EN WITH DESCENDER")
    (1188 . "CYRILLIC CAPITAL LIGATURE EN GHE")
    (1189 . "CYRILLIC SMALL LIGATURE EN GHE")
    (1190 . "CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK")
    (1191 . "CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK")
    (1192 . "CYRILLIC CAPITAL LETTER ABKHASIAN HA")
    (1193 . "CYRILLIC SMALL LETTER ABKHASIAN HA")
    (1194 . "CYRILLIC CAPITAL LETTER ES WITH DESCENDER")
    (1195 . "CYRILLIC SMALL LETTER ES WITH DESCENDER")
    (1196 . "CYRILLIC CAPITAL LETTER TE WITH DESCENDER")
    (1197 . "CYRILLIC SMALL LETTER TE WITH DESCENDER")
    (1198 . "CYRILLIC CAPITAL LETTER STRAIGHT U")
    (1199 . "CYRILLIC SMALL LETTER STRAIGHT U")
    (1200 . "CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE")
    (1201 . "CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE")
    (1202 . "CYRILLIC CAPITAL LETTER HA WITH DESCENDER")
    (1203 . "CYRILLIC SMALL LETTER HA WITH DESCENDER")
    (1204 . "CYRILLIC CAPITAL LIGATURE TE TSE")
    (1205 . "CYRILLIC SMALL LIGATURE TE TSE")
    (1206 . "CYRILLIC CAPITAL LETTER CHE WITH DESCENDER")
    (1207 . "CYRILLIC SMALL LETTER CHE WITH DESCENDER")
    (1208 . "CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE")
    (1209 . "CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE")
    (1210 . "CYRILLIC CAPITAL LETTER SHHA")
    (1211 . "CYRILLIC SMALL LETTER SHHA")
    (1212 . "CYRILLIC CAPITAL LETTER ABKHASIAN CHE")
    (1213 . "CYRILLIC SMALL LETTER ABKHASIAN CHE")
    (1214 . "CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER")
    (1215 . "CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER")
    (1216 . "CYRILLIC LETTER PALOCHKA")
    (1217 . "CYRILLIC CAPITAL LETTER ZHE WITH BREVE")
    (1218 . "CYRILLIC SMALL LETTER ZHE WITH BREVE")
    (1219 . "CYRILLIC CAPITAL LETTER KA WITH HOOK")
    (1220 . "CYRILLIC SMALL LETTER KA WITH HOOK")
    (1223 . "CYRILLIC CAPITAL LETTER EN WITH HOOK")
    (1224 . "CYRILLIC SMALL LETTER EN WITH HOOK")
    (1227 . "CYRILLIC CAPITAL LETTER KHAKASSIAN CHE")
    (1228 . "CYRILLIC SMALL LETTER KHAKASSIAN CHE")
    (1232 . "CYRILLIC CAPITAL LETTER A WITH BREVE")
    (1233 . "CYRILLIC SMALL LETTER A WITH BREVE")
    (1234 . "CYRILLIC CAPITAL LETTER A WITH DIAERESIS")
    (1235 . "CYRILLIC SMALL LETTER A WITH DIAERESIS")
    (1236 . "CYRILLIC CAPITAL LIGATURE A IE")
    (1237 . "CYRILLIC SMALL LIGATURE A IE")
    (1238 . "CYRILLIC CAPITAL LETTER IE WITH BREVE")
    (1239 . "CYRILLIC SMALL LETTER IE WITH BREVE")
    (1240 . "CYRILLIC CAPITAL LETTER SCHWA")
    (1241 . "CYRILLIC SMALL LETTER SCHWA")
    (1242 . "CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS")
    (1243 . "CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS")
    (1244 . "CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS")
    (1245 . "CYRILLIC SMALL LETTER ZHE WITH DIAERESIS")
    (1246 . "CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS")
    (1247 . "CYRILLIC SMALL LETTER ZE WITH DIAERESIS")
    (1248 . "CYRILLIC CAPITAL LETTER ABKHASIAN DZE")
    (1249 . "CYRILLIC SMALL LETTER ABKHASIAN DZE")
    (1250 . "CYRILLIC CAPITAL LETTER I WITH MACRON")
    (1251 . "CYRILLIC SMALL LETTER I WITH MACRON")
    (1252 . "CYRILLIC CAPITAL LETTER I WITH DIAERESIS")
    (1253 . "CYRILLIC SMALL LETTER I WITH DIAERESIS")
    (1254 . "CYRILLIC CAPITAL LETTER O WITH DIAERESIS")
    (1255 . "CYRILLIC SMALL LETTER O WITH DIAERESIS")
    (1256 . "CYRILLIC CAPITAL LETTER BARRED O")
    (1257 . "CYRILLIC SMALL LETTER BARRED O")
    (1258 . "CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS")
    (1259 . "CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS")
    (1262 . "CYRILLIC CAPITAL LETTER U WITH MACRON")
    (1263 . "CYRILLIC SMALL LETTER U WITH MACRON")
    (1264 . "CYRILLIC CAPITAL LETTER U WITH DIAERESIS")
    (1265 . "CYRILLIC SMALL LETTER U WITH DIAERESIS")
    (1266 . "CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE")
    (1267 . "CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE")
    (1268 . "CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS")
    (1269 . "CYRILLIC SMALL LETTER CHE WITH DIAERESIS")
    (1272 . "CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS")
    (1273 . "CYRILLIC SMALL LETTER YERU WITH DIAERESIS")
    (1329 . "ARMENIAN CAPITAL LETTER AYB")
    (1330 . "ARMENIAN CAPITAL LETTER BEN")
    (1331 . "ARMENIAN CAPITAL LETTER GIM")
    (1332 . "ARMENIAN CAPITAL LETTER DA")
    (1333 . "ARMENIAN CAPITAL LETTER ECH")
    (1334 . "ARMENIAN CAPITAL LETTER ZA")
    (1335 . "ARMENIAN CAPITAL LETTER EH")
    (1336 . "ARMENIAN CAPITAL LETTER ET")
    (1337 . "ARMENIAN CAPITAL LETTER TO")
    (1338 . "ARMENIAN CAPITAL LETTER ZHE")
    (1339 . "ARMENIAN CAPITAL LETTER INI")
    (1340 . "ARMENIAN CAPITAL LETTER LIWN")
    (1341 . "ARMENIAN CAPITAL LETTER XEH")
    (1342 . "ARMENIAN CAPITAL LETTER CA")
    (1343 . "ARMENIAN CAPITAL LETTER KEN")
    (1344 . "ARMENIAN CAPITAL LETTER HO")
    (1345 . "ARMENIAN CAPITAL LETTER JA")
    (1346 . "ARMENIAN CAPITAL LETTER GHAD")
    (1347 . "ARMENIAN CAPITAL LETTER CHEH")
    (1348 . "ARMENIAN CAPITAL LETTER MEN")
    (1349 . "ARMENIAN CAPITAL LETTER YI")
    (1350 . "ARMENIAN CAPITAL LETTER NOW")
    (1351 . "ARMENIAN CAPITAL LETTER SHA")
    (1352 . "ARMENIAN CAPITAL LETTER VO")
    (1353 . "ARMENIAN CAPITAL LETTER CHA")
    (1354 . "ARMENIAN CAPITAL LETTER PEH")
    (1355 . "ARMENIAN CAPITAL LETTER JHEH")
    (1356 . "ARMENIAN CAPITAL LETTER RA")
    (1357 . "ARMENIAN CAPITAL LETTER SEH")
    (1358 . "ARMENIAN CAPITAL LETTER VEW")
    (1359 . "ARMENIAN CAPITAL LETTER TIWN")
    (1360 . "ARMENIAN CAPITAL LETTER REH")
    (1361 . "ARMENIAN CAPITAL LETTER CO")
    (1362 . "ARMENIAN CAPITAL LETTER YIWN")
    (1363 . "ARMENIAN CAPITAL LETTER PIWR")
    (1364 . "ARMENIAN CAPITAL LETTER KEH")
    (1365 . "ARMENIAN CAPITAL LETTER OH")
    (1366 . "ARMENIAN CAPITAL LETTER FEH")
    (1369 . "ARMENIAN MODIFIER LETTER LEFT HALF RING")
    (1370 . "ARMENIAN APOSTROPHE")
    (1371 . "ARMENIAN EMPHASIS MARK")
    (1372 . "ARMENIAN EXCLAMATION MARK")
    (1373 . "ARMENIAN COMMA")
    (1374 . "ARMENIAN QUESTION MARK")
    (1375 . "ARMENIAN ABBREVIATION MARK")
    (1377 . "ARMENIAN SMALL LETTER AYB")
    (1378 . "ARMENIAN SMALL LETTER BEN")
    (1379 . "ARMENIAN SMALL LETTER GIM")
    (1380 . "ARMENIAN SMALL LETTER DA")
    (1381 . "ARMENIAN SMALL LETTER ECH")
    (1382 . "ARMENIAN SMALL LETTER ZA")
    (1383 . "ARMENIAN SMALL LETTER EH")
    (1384 . "ARMENIAN SMALL LETTER ET")
    (1385 . "ARMENIAN SMALL LETTER TO")
    (1386 . "ARMENIAN SMALL LETTER ZHE")
    (1387 . "ARMENIAN SMALL LETTER INI")
    (1388 . "ARMENIAN SMALL LETTER LIWN")
    (1389 . "ARMENIAN SMALL LETTER XEH")
    (1390 . "ARMENIAN SMALL LETTER CA")
    (1391 . "ARMENIAN SMALL LETTER KEN")
    (1392 . "ARMENIAN SMALL LETTER HO")
    (1393 . "ARMENIAN SMALL LETTER JA")
    (1394 . "ARMENIAN SMALL LETTER GHAD")
    (1395 . "ARMENIAN SMALL LETTER CHEH")
    (1396 . "ARMENIAN SMALL LETTER MEN")
    (1397 . "ARMENIAN SMALL LETTER YI")
    (1398 . "ARMENIAN SMALL LETTER NOW")
    (1399 . "ARMENIAN SMALL LETTER SHA")
    (1400 . "ARMENIAN SMALL LETTER VO")
    (1401 . "ARMENIAN SMALL LETTER CHA")
    (1402 . "ARMENIAN SMALL LETTER PEH")
    (1403 . "ARMENIAN SMALL LETTER JHEH")
    (1404 . "ARMENIAN SMALL LETTER RA")
    (1405 . "ARMENIAN SMALL LETTER SEH")
    (1406 . "ARMENIAN SMALL LETTER VEW")
    (1407 . "ARMENIAN SMALL LETTER TIWN")
    (1408 . "ARMENIAN SMALL LETTER REH")
    (1409 . "ARMENIAN SMALL LETTER CO")
    (1410 . "ARMENIAN SMALL LETTER YIWN")
    (1411 . "ARMENIAN SMALL LETTER PIWR")
    (1412 . "ARMENIAN SMALL LETTER KEH")
    (1413 . "ARMENIAN SMALL LETTER OH")
    (1414 . "ARMENIAN SMALL LETTER FEH")
    (1415 . "ARMENIAN SMALL LIGATURE ECH YIWN")
    (1417 . "ARMENIAN PERIOD|ARMENIAN FULL STOP")
    (1456 . "HEBREW POINT SHEVA")
    (1457 . "HEBREW POINT HATAF SEGOL")
    (1458 . "HEBREW POINT HATAF PATAH")
    (1459 . "HEBREW POINT HATAF QAMATS")
    (1460 . "HEBREW POINT HIRIQ")
    (1461 . "HEBREW POINT TSERE")
    (1462 . "HEBREW POINT SEGOL")
    (1463 . "HEBREW POINT PATAH")
    (1464 . "HEBREW POINT QAMATS")
    (1465 . "HEBREW POINT HOLAM")
    (1467 . "HEBREW POINT QUBUTS")
    (1468 . "HEBREW POINT DAGESH OR MAPIQ")
    (1469 . "HEBREW POINT METEG")
    (1470 . "HEBREW PUNCTUATION MAQAF")
    (1471 . "HEBREW POINT RAFE")
    (1472 . "HEBREW PUNCTUATION PASEQ")
    (1473 . "HEBREW POINT SHIN DOT")
    (1474 . "HEBREW POINT SIN DOT")
    (1475 . "HEBREW PUNCTUATION SOF PASUQ")
    (1488 . "HEBREW LETTER ALEF")
    (1489 . "HEBREW LETTER BET")
    (1490 . "HEBREW LETTER GIMEL")
    (1491 . "HEBREW LETTER DALET")
    (1492 . "HEBREW LETTER HE")
    (1493 . "HEBREW LETTER VAV")
    (1494 . "HEBREW LETTER ZAYIN")
    (1495 . "HEBREW LETTER HET")
    (1496 . "HEBREW LETTER TET")
    (1497 . "HEBREW LETTER YOD")
    (1498 . "HEBREW LETTER FINAL KAF")
    (1499 . "HEBREW LETTER KAF")
    (1500 . "HEBREW LETTER LAMED")
    (1501 . "HEBREW LETTER FINAL MEM")
    (1502 . "HEBREW LETTER MEM")
    (1503 . "HEBREW LETTER FINAL NUN")
    (1504 . "HEBREW LETTER NUN")
    (1505 . "HEBREW LETTER SAMEKH")
    (1506 . "HEBREW LETTER AYIN")
    (1507 . "HEBREW LETTER FINAL PE")
    (1508 . "HEBREW LETTER PE")
    (1509 . "HEBREW LETTER FINAL TSADI")
    (1510 . "HEBREW LETTER TSADI")
    (1511 . "HEBREW LETTER QOF")
    (1512 . "HEBREW LETTER RESH")
    (1513 . "HEBREW LETTER SHIN")
    (1514 . "HEBREW LETTER TAV")
    (1520 . "HEBREW LIGATURE YIDDISH DOUBLE VAV")
    (1521 . "HEBREW LIGATURE YIDDISH VAV YOD")
    (1522 . "HEBREW LIGATURE YIDDISH DOUBLE YOD")
    (1523 . "HEBREW PUNCTUATION GERESH")
    (1524 . "HEBREW PUNCTUATION GERSHAYIM")
    (1548 . "ARABIC COMMA")
    (1563 . "ARABIC SEMICOLON")
    (1567 . "ARABIC QUESTION MARK")
    (1569 . "ARABIC LETTER HAMZA")
    (1570 . "ARABIC LETTER ALEF WITH MADDA ABOVE")
    (1571 . "ARABIC LETTER ALEF WITH HAMZA ABOVE")
    (1572 . "ARABIC LETTER WAW WITH HAMZA ABOVE")
    (1573 . "ARABIC LETTER ALEF WITH HAMZA BELOW")
    (1574 . "ARABIC LETTER YEH WITH HAMZA ABOVE")
    (1575 . "ARABIC LETTER ALEF")
    (1576 . "ARABIC LETTER BEH")
    (1577 . "ARABIC LETTER TEH MARBUTA")
    (1578 . "ARABIC LETTER TEH")
    (1579 . "ARABIC LETTER THEH")
    (1580 . "ARABIC LETTER JEEM")
    (1581 . "ARABIC LETTER HAH")
    (1582 . "ARABIC LETTER KHAH")
    (1583 . "ARABIC LETTER DAL")
    (1584 . "ARABIC LETTER THAL")
    (1585 . "ARABIC LETTER REH")
    (1586 . "ARABIC LETTER ZAIN")
    (1587 . "ARABIC LETTER SEEN")
    (1588 . "ARABIC LETTER SHEEN")
    (1589 . "ARABIC LETTER SAD")
    (1590 . "ARABIC LETTER DAD")
    (1591 . "ARABIC LETTER TAH")
    (1592 . "ARABIC LETTER ZAH")
    (1593 . "ARABIC LETTER AIN")
    (1594 . "ARABIC LETTER GHAIN")
    (1600 . "ARABIC TATWEEL")
    (1601 . "ARABIC LETTER FEH")
    (1602 . "ARABIC LETTER QAF")
    (1603 . "ARABIC LETTER KAF")
    (1604 . "ARABIC LETTER LAM")
    (1605 . "ARABIC LETTER MEEM")
    (1606 . "ARABIC LETTER NOON")
    (1607 . "ARABIC LETTER HEH")
    (1608 . "ARABIC LETTER WAW")
    (1609 . "ARABIC LETTER ALEF MAKSURA")
    (1610 . "ARABIC LETTER YEH")
    (1611 . "ARABIC FATHATAN")
    (1612 . "ARABIC DAMMATAN")
    (1613 . "ARABIC KASRATAN")
    (1614 . "ARABIC FATHA")
    (1615 . "ARABIC DAMMA")
    (1616 . "ARABIC KASRA")
    (1617 . "ARABIC SHADDA")
    (1618 . "ARABIC SUKUN")
    (1632 . "ARABIC-INDIC DIGIT ZERO")
    (1633 . "ARABIC-INDIC DIGIT ONE")
    (1634 . "ARABIC-INDIC DIGIT TWO")
    (1635 . "ARABIC-INDIC DIGIT THREE")
    (1636 . "ARABIC-INDIC DIGIT FOUR")
    (1637 . "ARABIC-INDIC DIGIT FIVE")
    (1638 . "ARABIC-INDIC DIGIT SIX")
    (1639 . "ARABIC-INDIC DIGIT SEVEN")
    (1640 . "ARABIC-INDIC DIGIT EIGHT")
    (1641 . "ARABIC-INDIC DIGIT NINE")
    (1642 . "ARABIC PERCENT SIGN")
    (1643 . "ARABIC DECIMAL SEPARATOR")
    (1644 . "ARABIC THOUSANDS SEPARATOR")
    (1645 . "ARABIC FIVE POINTED STAR")
    (1648 . "ARABIC LETTER SUPERSCRIPT ALEF")
    (1649 . "ARABIC LETTER ALEF WASLA")
    (1650 . "ARABIC LETTER ALEF WITH WAVY HAMZA ABOVE")
    (1651 . "ARABIC LETTER ALEF WITH WAVY HAMZA BELOW")
    (1652 . "ARABIC LETTER HIGH HAMZA")
    (1653 . "ARABIC LETTER HIGH HAMZA ALEF")
    (1654 . "ARABIC LETTER HIGH HAMZA WAW")
    (1655 . "ARABIC LETTER U WITH HAMZA ABOVE")
    (1656 . "ARABIC LETTER HIGH HAMZA YEH")
    (1657 . "ARABIC LETTER TTEH")
    (1658 . "ARABIC LETTER TTEHEH")
    (1659 . "ARABIC LETTER BEEH")
    (1660 . "ARABIC LETTER TEH WITH RING")
    (1661 . "ARABIC LETTER TEH WITH THREE DOTS ABOVE DOWNWARDS")
    (1662 . "ARABIC LETTER PEH")
    (1663 . "ARABIC LETTER TEHEH")
    (1664 . "ARABIC LETTER BEHEH")
    (1665 . "ARABIC LETTER HAH WITH HAMZA ABOVE")
    (1666 . "ARABIC LETTER HAH WITH TWO DOTS VERTICAL ABOVE")
    (1667 . "ARABIC LETTER NYEH")
    (1668 . "ARABIC LETTER DYEH")
    (1669 . "ARABIC LETTER HAH WITH THREE DOTS ABOVE")
    (1670 . "ARABIC LETTER TCHEH")
    (1671 . "ARABIC LETTER TCHEHEH")
    (1672 . "ARABIC LETTER DDAL")
    (1673 . "ARABIC LETTER DAL WITH RING")
    (1674 . "ARABIC LETTER DAL WITH DOT BELOW")
    (1675 . "ARABIC LETTER DAL WITH DOT BELOW AND SMALL TAH")
    (1676 . "ARABIC LETTER DAHAL")
    (1677 . "ARABIC LETTER DDAHAL")
    (1678 . "ARABIC LETTER DUL")
    (1679 . "ARABIC LETTER DAL WITH THREE DOTS ABOVE DOWNWARDS")
    (1680 . "ARABIC LETTER DAL WITH FOUR DOTS ABOVE")
    (1681 . "ARABIC LETTER RREH")
    (1682 . "ARABIC LETTER REH WITH SMALL V")
    (1683 . "ARABIC LETTER REH WITH RING")
    (1684 . "ARABIC LETTER REH WITH DOT BELOW")
    (1685 . "ARABIC LETTER REH WITH SMALL V BELOW")
    (1686 . "ARABIC LETTER REH WITH DOT BELOW AND DOT ABOVE")
    (1687 . "ARABIC LETTER REH WITH TWO DOTS ABOVE")
    (1688 . "ARABIC LETTER JEH")
    (1689 . "ARABIC LETTER REH WITH FOUR DOTS ABOVE")
    (1690 . "ARABIC LETTER SEEN WITH DOT BELOW AND DOT ABOVE")
    (1691 . "ARABIC LETTER SEEN WITH THREE DOTS BELOW")
    (1692 . "ARABIC LETTER SEEN WITH THREE DOTS BELOW AND THREE DOTS ABOVE")
    (1693 . "ARABIC LETTER SAD WITH TWO DOTS BELOW")
    (1694 . "ARABIC LETTER SAD WITH THREE DOTS ABOVE")
    (1695 . "ARABIC LETTER TAH WITH THREE DOTS ABOVE")
    (1696 . "ARABIC LETTER AIN WITH THREE DOTS ABOVE")
    (1697 . "ARABIC LETTER DOTLESS FEH")
    (1698 . "ARABIC LETTER FEH WITH DOT MOVED BELOW")
    (1699 . "ARABIC LETTER FEH WITH DOT BELOW")
    (1700 . "ARABIC LETTER VEH")
    (1701 . "ARABIC LETTER FEH WITH THREE DOTS BELOW")
    (1702 . "ARABIC LETTER PEHEH")
    (1703 . "ARABIC LETTER QAF WITH DOT ABOVE")
    (1704 . "ARABIC LETTER QAF WITH THREE DOTS ABOVE")
    (1705 . "ARABIC LETTER KEHEH")
    (1706 . "ARABIC LETTER SWASH KAF")
    (1707 . "ARABIC LETTER KAF WITH RING")
    (1708 . "ARABIC LETTER KAF WITH DOT ABOVE")
    (1709 . "ARABIC LETTER NG")
    (1710 . "ARABIC LETTER KAF WITH THREE DOTS BELOW")
    (1711 . "ARABIC LETTER GAF")
    (1712 . "ARABIC LETTER GAF WITH RING")
    (1713 . "ARABIC LETTER NGOEH")
    (1714 . "ARABIC LETTER GAF WITH TWO DOTS BELOW")
    (1715 . "ARABIC LETTER GUEH")
    (1716 . "ARABIC LETTER GAF WITH THREE DOTS ABOVE")
    (1717 . "ARABIC LETTER LAM WITH SMALL V")
    (1718 . "ARABIC LETTER LAM WITH DOT ABOVE")
    (1719 . "ARABIC LETTER LAM WITH THREE DOTS ABOVE")
    (1722 . "ARABIC LETTER NOON GHUNNA")
    (1723 . "ARABIC LETTER RNOON")
    (1724 . "ARABIC LETTER NOON WITH RING")
    (1725 . "ARABIC LETTER NOON WITH THREE DOTS ABOVE")
    (1726 . "ARABIC LETTER HEH DOACHASHMEE")
    (1728 . "ARABIC LETTER HEH WITH YEH ABOVE")
    (1729 . "ARABIC LETTER HEH GOAL")
    (1730 . "ARABIC LETTER HEH GOAL WITH HAMZA ABOVE")
    (1731 . "ARABIC LETTER TEH MARBUTA GOAL")
    (1732 . "ARABIC LETTER WAW WITH RING")
    (1733 . "ARABIC LETTER KIRGHIZ OE")
    (1734 . "ARABIC LETTER OE")
    (1735 . "ARABIC LETTER U")
    (1736 . "ARABIC LETTER YU")
    (1737 . "ARABIC LETTER KIRGHIZ YU")
    (1738 . "ARABIC LETTER WAW WITH TWO DOTS ABOVE")
    (1739 . "ARABIC LETTER VE")
    (1740 . "ARABIC LETTER FARSI YEH")
    (1741 . "ARABIC LETTER YEH WITH TAIL")
    (1742 . "ARABIC LETTER YEH WITH SMALL V")
    (1744 . "ARABIC LETTER E")
    (1745 . "ARABIC LETTER YEH WITH THREE DOTS BELOW")
    (1746 . "ARABIC LETTER YEH BARREE")
    (1747 . "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE")
    (1748 . "ARABIC PERIOD|ARABIC FULL STOP")
    (1749 . "ARABIC LETTER AE")
    (1750 . "ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA")
    (1751 . "ARABIC SMALL HIGH LIGATURE QAF WITH LAM WITH ALEF MAKSURA")
    (1752 . "ARABIC SMALL HIGH MEEM INITIAL FORM")
    (1753 . "ARABIC SMALL HIGH LAM ALEF")
    (1754 . "ARABIC SMALL HIGH JEEM")
    (1755 . "ARABIC SMALL HIGH THREE DOTS")
    (1756 . "ARABIC SMALL HIGH SEEN")
    (1757 . "ARABIC END OF AYAH")
    (1758 . "ARABIC START OF RUB EL HIZB")
    (1759 . "ARABIC SMALL HIGH ROUNDED ZERO")
    (1760 . "ARABIC SMALL HIGH UPRIGHT RECTANGULAR ZERO")
    (1761 . "ARABIC SMALL HIGH DOTLESS HEAD OF KHAH")
    (1762 . "ARABIC SMALL HIGH MEEM ISOLATED FORM")
    (1763 . "ARABIC SMALL LOW SEEN")
    (1764 . "ARABIC SMALL HIGH MADDA")
    (1765 . "ARABIC SMALL WAW")
    (1766 . "ARABIC SMALL YEH")
    (1767 . "ARABIC SMALL HIGH YEH")
    (1768 . "ARABIC SMALL HIGH NOON")
    (1769 . "ARABIC PLACE OF SAJDAH")
    (1770 . "ARABIC EMPTY CENTER LOW STOP|ARABIC EMPTY CENTRE LOW STOP")
    (1771 . "ARABIC EMPTY CENTER HIGH STOP|ARABIC EMPTY CENTRE HIGH STOP")
    (1772 . "ARABIC ROUNDED HIGH STOP WITH FILLED CENTER|ARABIC ROUNDED HIGH STOP WITH FILLED CENTRE")
    (1773 . "ARABIC SMALL LOW MEEM")
    (1776 . "EXTENDED ARABIC-INDIC DIGIT ZERO")
    (1777 . "EXTENDED ARABIC-INDIC DIGIT ONE")
    (1778 . "EXTENDED ARABIC-INDIC DIGIT TWO")
    (1779 . "EXTENDED ARABIC-INDIC DIGIT THREE")
    (1780 . "EXTENDED ARABIC-INDIC DIGIT FOUR")
    (1781 . "EXTENDED ARABIC-INDIC DIGIT FIVE")
    (1782 . "EXTENDED ARABIC-INDIC DIGIT SIX")
    (1783 . "EXTENDED ARABIC-INDIC DIGIT SEVEN")
    (1784 . "EXTENDED ARABIC-INDIC DIGIT EIGHT")
    (1785 . "EXTENDED ARABIC-INDIC DIGIT NINE")
    (2305 . "DEVANAGARI SIGN CANDRABINDU")
    (2306 . "DEVANAGARI SIGN ANUSVARA")
    (2307 . "DEVANAGARI SIGN VISARGA")
    (2309 . "DEVANAGARI LETTER A")
    (2310 . "DEVANAGARI LETTER AA")
    (2311 . "DEVANAGARI LETTER I")
    (2312 . "DEVANAGARI LETTER II")
    (2313 . "DEVANAGARI LETTER U")
    (2314 . "DEVANAGARI LETTER UU")
    (2315 . "DEVANAGARI LETTER VOCALIC R")
    (2316 . "DEVANAGARI LETTER VOCALIC L")
    (2317 . "DEVANAGARI LETTER CANDRA E")
    (2318 . "DEVANAGARI LETTER SHORT E")
    (2319 . "DEVANAGARI LETTER E")
    (2320 . "DEVANAGARI LETTER AI")
    (2321 . "DEVANAGARI LETTER CANDRA O")
    (2322 . "DEVANAGARI LETTER SHORT O")
    (2323 . "DEVANAGARI LETTER O")
    (2324 . "DEVANAGARI LETTER AU")
    (2325 . "DEVANAGARI LETTER KA")
    (2326 . "DEVANAGARI LETTER KHA")
    (2327 . "DEVANAGARI LETTER GA")
    (2328 . "DEVANAGARI LETTER GHA")
    (2329 . "DEVANAGARI LETTER NGA")
    (2330 . "DEVANAGARI LETTER CA")
    (2331 . "DEVANAGARI LETTER CHA")
    (2332 . "DEVANAGARI LETTER JA")
    (2333 . "DEVANAGARI LETTER JHA")
    (2334 . "DEVANAGARI LETTER NYA")
    (2335 . "DEVANAGARI LETTER TTA")
    (2336 . "DEVANAGARI LETTER TTHA")
    (2337 . "DEVANAGARI LETTER DDA")
    (2338 . "DEVANAGARI LETTER DDHA")
    (2339 . "DEVANAGARI LETTER NNA")
    (2340 . "DEVANAGARI LETTER TA")
    (2341 . "DEVANAGARI LETTER THA")
    (2342 . "DEVANAGARI LETTER DA")
    (2343 . "DEVANAGARI LETTER DHA")
    (2344 . "DEVANAGARI LETTER NA")
    (2345 . "DEVANAGARI LETTER NNNA")
    (2346 . "DEVANAGARI LETTER PA")
    (2347 . "DEVANAGARI LETTER PHA")
    (2348 . "DEVANAGARI LETTER BA")
    (2349 . "DEVANAGARI LETTER BHA")
    (2350 . "DEVANAGARI LETTER MA")
    (2351 . "DEVANAGARI LETTER YA")
    (2352 . "DEVANAGARI LETTER RA")
    (2353 . "DEVANAGARI LETTER RRA")
    (2354 . "DEVANAGARI LETTER LA")
    (2355 . "DEVANAGARI LETTER LLA")
    (2356 . "DEVANAGARI LETTER LLLA")
    (2357 . "DEVANAGARI LETTER VA")
    (2358 . "DEVANAGARI LETTER SHA")
    (2359 . "DEVANAGARI LETTER SSA")
    (2360 . "DEVANAGARI LETTER SA")
    (2361 . "DEVANAGARI LETTER HA")
    (2364 . "DEVANAGARI SIGN NUKTA")
    (2365 . "DEVANAGARI SIGN AVAGRAHA")
    (2366 . "DEVANAGARI VOWEL SIGN AA")
    (2367 . "DEVANAGARI VOWEL SIGN I")
    (2368 . "DEVANAGARI VOWEL SIGN II")
    (2369 . "DEVANAGARI VOWEL SIGN U")
    (2370 . "DEVANAGARI VOWEL SIGN UU")
    (2371 . "DEVANAGARI VOWEL SIGN VOCALIC R")
    (2372 . "DEVANAGARI VOWEL SIGN VOCALIC RR")
    (2373 . "DEVANAGARI VOWEL SIGN CANDRA E")
    (2374 . "DEVANAGARI VOWEL SIGN SHORT E")
    (2375 . "DEVANAGARI VOWEL SIGN E")
    (2376 . "DEVANAGARI VOWEL SIGN AI")
    (2377 . "DEVANAGARI VOWEL SIGN CANDRA O")
    (2378 . "DEVANAGARI VOWEL SIGN SHORT O")
    (2379 . "DEVANAGARI VOWEL SIGN O")
    (2380 . "DEVANAGARI VOWEL SIGN AU")
    (2381 . "DEVANAGARI SIGN VIRAMA")
    (2384 . "DEVANAGARI OM")
    (2385 . "DEVANAGARI STRESS SIGN UDATTA")
    (2386 . "DEVANAGARI STRESS SIGN ANUDATTA")
    (2387 . "DEVANAGARI GRAVE ACCENT")
    (2388 . "DEVANAGARI ACUTE ACCENT")
    (2392 . "DEVANAGARI LETTER QA")
    (2393 . "DEVANAGARI LETTER KHHA")
    (2394 . "DEVANAGARI LETTER GHHA")
    (2395 . "DEVANAGARI LETTER ZA")
    (2396 . "DEVANAGARI LETTER DDDHA")
    (2397 . "DEVANAGARI LETTER RHA")
    (2398 . "DEVANAGARI LETTER FA")
    (2399 . "DEVANAGARI LETTER YYA")
    (2400 . "DEVANAGARI LETTER VOCALIC RR")
    (2401 . "DEVANAGARI LETTER VOCALIC LL")
    (2402 . "DEVANAGARI VOWEL SIGN VOCALIC L")
    (2403 . "DEVANAGARI VOWEL SIGN VOCALIC LL")
    (2404 . "DEVANAGARI DANDA")
    (2405 . "DEVANAGARI DOUBLE DANDA")
    (2406 . "DEVANAGARI DIGIT ZERO")
    (2407 . "DEVANAGARI DIGIT ONE")
    (2408 . "DEVANAGARI DIGIT TWO")
    (2409 . "DEVANAGARI DIGIT THREE")
    (2410 . "DEVANAGARI DIGIT FOUR")
    (2411 . "DEVANAGARI DIGIT FIVE")
    (2412 . "DEVANAGARI DIGIT SIX")
    (2413 . "DEVANAGARI DIGIT SEVEN")
    (2414 . "DEVANAGARI DIGIT EIGHT")
    (2415 . "DEVANAGARI DIGIT NINE")
    (2416 . "DEVANAGARI ABBREVIATION SIGN")
    (2433 . "BENGALI SIGN CANDRABINDU")
    (2434 . "BENGALI SIGN ANUSVARA")
    (2435 . "BENGALI SIGN VISARGA")
    (2437 . "BENGALI LETTER A")
    (2438 . "BENGALI LETTER AA")
    (2439 . "BENGALI LETTER I")
    (2440 . "BENGALI LETTER II")
    (2441 . "BENGALI LETTER U")
    (2442 . "BENGALI LETTER UU")
    (2443 . "BENGALI LETTER VOCALIC R")
    (2444 . "BENGALI LETTER VOCALIC L")
    (2447 . "BENGALI LETTER E")
    (2448 . "BENGALI LETTER AI")
    (2451 . "BENGALI LETTER O")
    (2452 . "BENGALI LETTER AU")
    (2453 . "BENGALI LETTER KA")
    (2454 . "BENGALI LETTER KHA")
    (2455 . "BENGALI LETTER GA")
    (2456 . "BENGALI LETTER GHA")
    (2457 . "BENGALI LETTER NGA")
    (2458 . "BENGALI LETTER CA")
    (2459 . "BENGALI LETTER CHA")
    (2460 . "BENGALI LETTER JA")
    (2461 . "BENGALI LETTER JHA")
    (2462 . "BENGALI LETTER NYA")
    (2463 . "BENGALI LETTER TTA")
    (2464 . "BENGALI LETTER TTHA")
    (2465 . "BENGALI LETTER DDA")
    (2466 . "BENGALI LETTER DDHA")
    (2467 . "BENGALI LETTER NNA")
    (2468 . "BENGALI LETTER TA")
    (2469 . "BENGALI LETTER THA")
    (2470 . "BENGALI LETTER DA")
    (2471 . "BENGALI LETTER DHA")
    (2472 . "BENGALI LETTER NA")
    (2474 . "BENGALI LETTER PA")
    (2475 . "BENGALI LETTER PHA")
    (2476 . "BENGALI LETTER BA")
    (2477 . "BENGALI LETTER BHA")
    (2478 . "BENGALI LETTER MA")
    (2479 . "BENGALI LETTER YA")
    (2480 . "BENGALI LETTER RA")
    (2482 . "BENGALI LETTER LA")
    (2486 . "BENGALI LETTER SHA")
    (2487 . "BENGALI LETTER SSA")
    (2488 . "BENGALI LETTER SA")
    (2489 . "BENGALI LETTER HA")
    (2492 . "BENGALI SIGN NUKTA")
    (2494 . "BENGALI VOWEL SIGN AA")
    (2495 . "BENGALI VOWEL SIGN I")
    (2496 . "BENGALI VOWEL SIGN II")
    (2497 . "BENGALI VOWEL SIGN U")
    (2498 . "BENGALI VOWEL SIGN UU")
    (2499 . "BENGALI VOWEL SIGN VOCALIC R")
    (2500 . "BENGALI VOWEL SIGN VOCALIC RR")
    (2503 . "BENGALI VOWEL SIGN E")
    (2504 . "BENGALI VOWEL SIGN AI")
    (2507 . "BENGALI VOWEL SIGN O")
    (2508 . "BENGALI VOWEL SIGN AU")
    (2509 . "BENGALI SIGN VIRAMA")
    (2519 . "BENGALI AU LENGTH MARK")
    (2524 . "BENGALI LETTER RRA")
    (2525 . "BENGALI LETTER RHA")
    (2527 . "BENGALI LETTER YYA")
    (2528 . "BENGALI LETTER VOCALIC RR")
    (2529 . "BENGALI LETTER VOCALIC LL")
    (2530 . "BENGALI VOWEL SIGN VOCALIC L")
    (2531 . "BENGALI VOWEL SIGN VOCALIC LL")
    (2534 . "BENGALI DIGIT ZERO")
    (2535 . "BENGALI DIGIT ONE")
    (2536 . "BENGALI DIGIT TWO")
    (2537 . "BENGALI DIGIT THREE")
    (2538 . "BENGALI DIGIT FOUR")
    (2539 . "BENGALI DIGIT FIVE")
    (2540 . "BENGALI DIGIT SIX")
    (2541 . "BENGALI DIGIT SEVEN")
    (2542 . "BENGALI DIGIT EIGHT")
    (2543 . "BENGALI DIGIT NINE")
    (2544 . "BENGALI LETTER RA WITH MIDDLE DIAGONAL")
    (2545 . "BENGALI LETTER RA WITH LOWER DIAGONAL")
    (2546 . "BENGALI RUPEE MARK")
    (2547 . "BENGALI RUPEE SIGN")
    (2548 . "BENGALI CURRENCY NUMERATOR ONE")
    (2549 . "BENGALI CURRENCY NUMERATOR TWO")
    (2550 . "BENGALI CURRENCY NUMERATOR THREE")
    (2551 . "BENGALI CURRENCY NUMERATOR FOUR")
    (2552 . "BENGALI CURRENCY NUMERATOR ONE LESS THAN THE DENOMINATOR")
    (2553 . "BENGALI CURRENCY DENOMINATOR SIXTEEN")
    (2554 . "BENGALI ISSHAR")
    (2562 . "GURMUKHI SIGN BINDI")
    (2565 . "GURMUKHI LETTER A")
    (2566 . "GURMUKHI LETTER AA")
    (2567 . "GURMUKHI LETTER I")
    (2568 . "GURMUKHI LETTER II")
    (2569 . "GURMUKHI LETTER U")
    (2570 . "GURMUKHI LETTER UU")
    (2575 . "GURMUKHI LETTER EE")
    (2576 . "GURMUKHI LETTER AI")
    (2579 . "GURMUKHI LETTER OO")
    (2580 . "GURMUKHI LETTER AU")
    (2581 . "GURMUKHI LETTER KA")
    (2582 . "GURMUKHI LETTER KHA")
    (2583 . "GURMUKHI LETTER GA")
    (2584 . "GURMUKHI LETTER GHA")
    (2585 . "GURMUKHI LETTER NGA")
    (2586 . "GURMUKHI LETTER CA")
    (2587 . "GURMUKHI LETTER CHA")
    (2588 . "GURMUKHI LETTER JA")
    (2589 . "GURMUKHI LETTER JHA")
    (2590 . "GURMUKHI LETTER NYA")
    (2591 . "GURMUKHI LETTER TTA")
    (2592 . "GURMUKHI LETTER TTHA")
    (2593 . "GURMUKHI LETTER DDA")
    (2594 . "GURMUKHI LETTER DDHA")
    (2595 . "GURMUKHI LETTER NNA")
    (2596 . "GURMUKHI LETTER TA")
    (2597 . "GURMUKHI LETTER THA")
    (2598 . "GURMUKHI LETTER DA")
    (2599 . "GURMUKHI LETTER DHA")
    (2600 . "GURMUKHI LETTER NA")
    (2602 . "GURMUKHI LETTER PA")
    (2603 . "GURMUKHI LETTER PHA")
    (2604 . "GURMUKHI LETTER BA")
    (2605 . "GURMUKHI LETTER BHA")
    (2606 . "GURMUKHI LETTER MA")
    (2607 . "GURMUKHI LETTER YA")
    (2608 . "GURMUKHI LETTER RA")
    (2610 . "GURMUKHI LETTER LA")
    (2611 . "GURMUKHI LETTER LLA")
    (2613 . "GURMUKHI LETTER VA")
    (2614 . "GURMUKHI LETTER SHA")
    (2616 . "GURMUKHI LETTER SA")
    (2617 . "GURMUKHI LETTER HA")
    (2620 . "GURMUKHI SIGN NUKTA")
    (2622 . "GURMUKHI VOWEL SIGN AA")
    (2623 . "GURMUKHI VOWEL SIGN I")
    (2624 . "GURMUKHI VOWEL SIGN II")
    (2625 . "GURMUKHI VOWEL SIGN U")
    (2626 . "GURMUKHI VOWEL SIGN UU")
    (2631 . "GURMUKHI VOWEL SIGN EE")
    (2632 . "GURMUKHI VOWEL SIGN AI")
    (2635 . "GURMUKHI VOWEL SIGN OO")
    (2636 . "GURMUKHI VOWEL SIGN AU")
    (2637 . "GURMUKHI SIGN VIRAMA")
    (2649 . "GURMUKHI LETTER KHHA")
    (2650 . "GURMUKHI LETTER GHHA")
    (2651 . "GURMUKHI LETTER ZA")
    (2652 . "GURMUKHI LETTER RRA")
    (2654 . "GURMUKHI LETTER FA")
    (2662 . "GURMUKHI DIGIT ZERO")
    (2663 . "GURMUKHI DIGIT ONE")
    (2664 . "GURMUKHI DIGIT TWO")
    (2665 . "GURMUKHI DIGIT THREE")
    (2666 . "GURMUKHI DIGIT FOUR")
    (2667 . "GURMUKHI DIGIT FIVE")
    (2668 . "GURMUKHI DIGIT SIX")
    (2669 . "GURMUKHI DIGIT SEVEN")
    (2670 . "GURMUKHI DIGIT EIGHT")
    (2671 . "GURMUKHI DIGIT NINE")
    (2672 . "GURMUKHI TIPPI")
    (2673 . "GURMUKHI ADDAK")
    (2674 . "GURMUKHI IRI")
    (2675 . "GURMUKHI URA")
    (2676 . "GURMUKHI EK ONKAR")
    (2689 . "GUJARATI SIGN CANDRABINDU")
    (2690 . "GUJARATI SIGN ANUSVARA")
    (2691 . "GUJARATI SIGN VISARGA")
    (2693 . "GUJARATI LETTER A")
    (2694 . "GUJARATI LETTER AA")
    (2695 . "GUJARATI LETTER I")
    (2696 . "GUJARATI LETTER II")
    (2697 . "GUJARATI LETTER U")
    (2698 . "GUJARATI LETTER UU")
    (2699 . "GUJARATI LETTER VOCALIC R")
    (2701 . "GUJARATI VOWEL CANDRA E")
    (2703 . "GUJARATI LETTER E")
    (2704 . "GUJARATI LETTER AI")
    (2705 . "GUJARATI VOWEL CANDRA O")
    (2707 . "GUJARATI LETTER O")
    (2708 . "GUJARATI LETTER AU")
    (2709 . "GUJARATI LETTER KA")
    (2710 . "GUJARATI LETTER KHA")
    (2711 . "GUJARATI LETTER GA")
    (2712 . "GUJARATI LETTER GHA")
    (2713 . "GUJARATI LETTER NGA")
    (2714 . "GUJARATI LETTER CA")
    (2715 . "GUJARATI LETTER CHA")
    (2716 . "GUJARATI LETTER JA")
    (2717 . "GUJARATI LETTER JHA")
    (2718 . "GUJARATI LETTER NYA")
    (2719 . "GUJARATI LETTER TTA")
    (2720 . "GUJARATI LETTER TTHA")
    (2721 . "GUJARATI LETTER DDA")
    (2722 . "GUJARATI LETTER DDHA")
    (2723 . "GUJARATI LETTER NNA")
    (2724 . "GUJARATI LETTER TA")
    (2725 . "GUJARATI LETTER THA")
    (2726 . "GUJARATI LETTER DA")
    (2727 . "GUJARATI LETTER DHA")
    (2728 . "GUJARATI LETTER NA")
    (2730 . "GUJARATI LETTER PA")
    (2731 . "GUJARATI LETTER PHA")
    (2732 . "GUJARATI LETTER BA")
    (2733 . "GUJARATI LETTER BHA")
    (2734 . "GUJARATI LETTER MA")
    (2735 . "GUJARATI LETTER YA")
    (2736 . "GUJARATI LETTER RA")
    (2738 . "GUJARATI LETTER LA")
    (2739 . "GUJARATI LETTER LLA")
    (2741 . "GUJARATI LETTER VA")
    (2742 . "GUJARATI LETTER SHA")
    (2743 . "GUJARATI LETTER SSA")
    (2744 . "GUJARATI LETTER SA")
    (2745 . "GUJARATI LETTER HA")
    (2748 . "GUJARATI SIGN NUKTA")
    (2749 . "GUJARATI SIGN AVAGRAHA")
    (2750 . "GUJARATI VOWEL SIGN AA")
    (2751 . "GUJARATI VOWEL SIGN I")
    (2752 . "GUJARATI VOWEL SIGN II")
    (2753 . "GUJARATI VOWEL SIGN U")
    (2754 . "GUJARATI VOWEL SIGN UU")
    (2755 . "GUJARATI VOWEL SIGN VOCALIC R")
    (2756 . "GUJARATI VOWEL SIGN VOCALIC RR")
    (2757 . "GUJARATI VOWEL SIGN CANDRA E")
    (2759 . "GUJARATI VOWEL SIGN E")
    (2760 . "GUJARATI VOWEL SIGN AI")
    (2761 . "GUJARATI VOWEL SIGN CANDRA O")
    (2763 . "GUJARATI VOWEL SIGN O")
    (2764 . "GUJARATI VOWEL SIGN AU")
    (2765 . "GUJARATI SIGN VIRAMA")
    (2768 . "GUJARATI OM")
    (2784 . "GUJARATI LETTER VOCALIC RR")
    (2790 . "GUJARATI DIGIT ZERO")
    (2791 . "GUJARATI DIGIT ONE")
    (2792 . "GUJARATI DIGIT TWO")
    (2793 . "GUJARATI DIGIT THREE")
    (2794 . "GUJARATI DIGIT FOUR")
    (2795 . "GUJARATI DIGIT FIVE")
    (2796 . "GUJARATI DIGIT SIX")
    (2797 . "GUJARATI DIGIT SEVEN")
    (2798 . "GUJARATI DIGIT EIGHT")
    (2799 . "GUJARATI DIGIT NINE")
    (2817 . "ORIYA SIGN CANDRABINDU")
    (2818 . "ORIYA SIGN ANUSVARA")
    (2819 . "ORIYA SIGN VISARGA")
    (2821 . "ORIYA LETTER A")
    (2822 . "ORIYA LETTER AA")
    (2823 . "ORIYA LETTER I")
    (2824 . "ORIYA LETTER II")
    (2825 . "ORIYA LETTER U")
    (2826 . "ORIYA LETTER UU")
    (2827 . "ORIYA LETTER VOCALIC R")
    (2828 . "ORIYA LETTER VOCALIC L")
    (2831 . "ORIYA LETTER E")
    (2832 . "ORIYA LETTER AI")
    (2835 . "ORIYA LETTER O")
    (2836 . "ORIYA LETTER AU")
    (2837 . "ORIYA LETTER KA")
    (2838 . "ORIYA LETTER KHA")
    (2839 . "ORIYA LETTER GA")
    (2840 . "ORIYA LETTER GHA")
    (2841 . "ORIYA LETTER NGA")
    (2842 . "ORIYA LETTER CA")
    (2843 . "ORIYA LETTER CHA")
    (2844 . "ORIYA LETTER JA")
    (2845 . "ORIYA LETTER JHA")
    (2846 . "ORIYA LETTER NYA")
    (2847 . "ORIYA LETTER TTA")
    (2848 . "ORIYA LETTER TTHA")
    (2849 . "ORIYA LETTER DDA")
    (2850 . "ORIYA LETTER DDHA")
    (2851 . "ORIYA LETTER NNA")
    (2852 . "ORIYA LETTER TA")
    (2853 . "ORIYA LETTER THA")
    (2854 . "ORIYA LETTER DA")
    (2855 . "ORIYA LETTER DHA")
    (2856 . "ORIYA LETTER NA")
    (2858 . "ORIYA LETTER PA")
    (2859 . "ORIYA LETTER PHA")
    (2860 . "ORIYA LETTER BA")
    (2861 . "ORIYA LETTER BHA")
    (2862 . "ORIYA LETTER MA")
    (2863 . "ORIYA LETTER YA")
    (2864 . "ORIYA LETTER RA")
    (2866 . "ORIYA LETTER LA")
    (2867 . "ORIYA LETTER LLA")
    (2870 . "ORIYA LETTER SHA")
    (2871 . "ORIYA LETTER SSA")
    (2872 . "ORIYA LETTER SA")
    (2873 . "ORIYA LETTER HA")
    (2876 . "ORIYA SIGN NUKTA")
    (2877 . "ORIYA SIGN AVAGRAHA")
    (2878 . "ORIYA VOWEL SIGN AA")
    (2879 . "ORIYA VOWEL SIGN I")
    (2880 . "ORIYA VOWEL SIGN II")
    (2881 . "ORIYA VOWEL SIGN U")
    (2882 . "ORIYA VOWEL SIGN UU")
    (2883 . "ORIYA VOWEL SIGN VOCALIC R")
    (2887 . "ORIYA VOWEL SIGN E")
    (2888 . "ORIYA VOWEL SIGN AI")
    (2891 . "ORIYA VOWEL SIGN O")
    (2892 . "ORIYA VOWEL SIGN AU")
    (2893 . "ORIYA SIGN VIRAMA")
    (2902 . "ORIYA AI LENGTH MARK")
    (2903 . "ORIYA AU LENGTH MARK")
    (2908 . "ORIYA LETTER RRA")
    (2909 . "ORIYA LETTER RHA")
    (2911 . "ORIYA LETTER YYA")
    (2912 . "ORIYA LETTER VOCALIC RR")
    (2913 . "ORIYA LETTER VOCALIC LL")
    (2918 . "ORIYA DIGIT ZERO")
    (2919 . "ORIYA DIGIT ONE")
    (2920 . "ORIYA DIGIT TWO")
    (2921 . "ORIYA DIGIT THREE")
    (2922 . "ORIYA DIGIT FOUR")
    (2923 . "ORIYA DIGIT FIVE")
    (2924 . "ORIYA DIGIT SIX")
    (2925 . "ORIYA DIGIT SEVEN")
    (2926 . "ORIYA DIGIT EIGHT")
    (2927 . "ORIYA DIGIT NINE")
    (2928 . "ORIYA ISSHAR")
    (2946 . "TAMIL SIGN ANUSVARA")
    (2947 . "TAMIL SIGN VISARGA")
    (2949 . "TAMIL LETTER A")
    (2950 . "TAMIL LETTER AA")
    (2951 . "TAMIL LETTER I")
    (2952 . "TAMIL LETTER II")
    (2953 . "TAMIL LETTER U")
    (2954 . "TAMIL LETTER UU")
    (2958 . "TAMIL LETTER E")
    (2959 . "TAMIL LETTER EE")
    (2960 . "TAMIL LETTER AI")
    (2962 . "TAMIL LETTER O")
    (2963 . "TAMIL LETTER OO")
    (2964 . "TAMIL LETTER AU")
    (2965 . "TAMIL LETTER KA")
    (2969 . "TAMIL LETTER NGA")
    (2970 . "TAMIL LETTER CA")
    (2972 . "TAMIL LETTER JA")
    (2974 . "TAMIL LETTER NYA")
    (2975 . "TAMIL LETTER TTA")
    (2979 . "TAMIL LETTER NNA")
    (2980 . "TAMIL LETTER TA")
    (2984 . "TAMIL LETTER NA")
    (2985 . "TAMIL LETTER NNNA")
    (2986 . "TAMIL LETTER PA")
    (2990 . "TAMIL LETTER MA")
    (2991 . "TAMIL LETTER YA")
    (2992 . "TAMIL LETTER RA")
    (2993 . "TAMIL LETTER RRA")
    (2994 . "TAMIL LETTER LA")
    (2995 . "TAMIL LETTER LLA")
    (2996 . "TAMIL LETTER LLLA")
    (2997 . "TAMIL LETTER VA")
    (2999 . "TAMIL LETTER SSA")
    (3000 . "TAMIL LETTER SA")
    (3001 . "TAMIL LETTER HA")
    (3006 . "TAMIL VOWEL SIGN AA")
    (3007 . "TAMIL VOWEL SIGN I")
    (3008 . "TAMIL VOWEL SIGN II")
    (3009 . "TAMIL VOWEL SIGN U")
    (3010 . "TAMIL VOWEL SIGN UU")
    (3014 . "TAMIL VOWEL SIGN E")
    (3015 . "TAMIL VOWEL SIGN EE")
    (3016 . "TAMIL VOWEL SIGN AI")
    (3018 . "TAMIL VOWEL SIGN O")
    (3019 . "TAMIL VOWEL SIGN OO")
    (3020 . "TAMIL VOWEL SIGN AU")
    (3021 . "TAMIL SIGN VIRAMA")
    (3031 . "TAMIL AU LENGTH MARK")
    (3047 . "TAMIL DIGIT ONE")
    (3048 . "TAMIL DIGIT TWO")
    (3049 . "TAMIL DIGIT THREE")
    (3050 . "TAMIL DIGIT FOUR")
    (3051 . "TAMIL DIGIT FIVE")
    (3052 . "TAMIL DIGIT SIX")
    (3053 . "TAMIL DIGIT SEVEN")
    (3054 . "TAMIL DIGIT EIGHT")
    (3055 . "TAMIL DIGIT NINE")
    (3056 . "TAMIL NUMBER TEN")
    (3057 . "TAMIL NUMBER ONE HUNDRED")
    (3058 . "TAMIL NUMBER ONE THOUSAND")
    (3073 . "TELUGU SIGN CANDRABINDU")
    (3074 . "TELUGU SIGN ANUSVARA")
    (3075 . "TELUGU SIGN VISARGA")
    (3077 . "TELUGU LETTER A")
    (3078 . "TELUGU LETTER AA")
    (3079 . "TELUGU LETTER I")
    (3080 . "TELUGU LETTER II")
    (3081 . "TELUGU LETTER U")
    (3082 . "TELUGU LETTER UU")
    (3083 . "TELUGU LETTER VOCALIC R")
    (3084 . "TELUGU LETTER VOCALIC L")
    (3086 . "TELUGU LETTER E")
    (3087 . "TELUGU LETTER EE")
    (3088 . "TELUGU LETTER AI")
    (3090 . "TELUGU LETTER O")
    (3091 . "TELUGU LETTER OO")
    (3092 . "TELUGU LETTER AU")
    (3093 . "TELUGU LETTER KA")
    (3094 . "TELUGU LETTER KHA")
    (3095 . "TELUGU LETTER GA")
    (3096 . "TELUGU LETTER GHA")
    (3097 . "TELUGU LETTER NGA")
    (3098 . "TELUGU LETTER CA")
    (3099 . "TELUGU LETTER CHA")
    (3100 . "TELUGU LETTER JA")
    (3101 . "TELUGU LETTER JHA")
    (3102 . "TELUGU LETTER NYA")
    (3103 . "TELUGU LETTER TTA")
    (3104 . "TELUGU LETTER TTHA")
    (3105 . "TELUGU LETTER DDA")
    (3106 . "TELUGU LETTER DDHA")
    (3107 . "TELUGU LETTER NNA")
    (3108 . "TELUGU LETTER TA")
    (3109 . "TELUGU LETTER THA")
    (3110 . "TELUGU LETTER DA")
    (3111 . "TELUGU LETTER DHA")
    (3112 . "TELUGU LETTER NA")
    (3114 . "TELUGU LETTER PA")
    (3115 . "TELUGU LETTER PHA")
    (3116 . "TELUGU LETTER BA")
    (3117 . "TELUGU LETTER BHA")
    (3118 . "TELUGU LETTER MA")
    (3119 . "TELUGU LETTER YA")
    (3120 . "TELUGU LETTER RA")
    (3121 . "TELUGU LETTER RRA")
    (3122 . "TELUGU LETTER LA")
    (3123 . "TELUGU LETTER LLA")
    (3125 . "TELUGU LETTER VA")
    (3126 . "TELUGU LETTER SHA")
    (3127 . "TELUGU LETTER SSA")
    (3128 . "TELUGU LETTER SA")
    (3129 . "TELUGU LETTER HA")
    (3134 . "TELUGU VOWEL SIGN AA")
    (3135 . "TELUGU VOWEL SIGN I")
    (3136 . "TELUGU VOWEL SIGN II")
    (3137 . "TELUGU VOWEL SIGN U")
    (3138 . "TELUGU VOWEL SIGN UU")
    (3139 . "TELUGU VOWEL SIGN VOCALIC R")
    (3140 . "TELUGU VOWEL SIGN VOCALIC RR")
    (3142 . "TELUGU VOWEL SIGN E")
    (3143 . "TELUGU VOWEL SIGN EE")
    (3144 . "TELUGU VOWEL SIGN AI")
    (3146 . "TELUGU VOWEL SIGN O")
    (3147 . "TELUGU VOWEL SIGN OO")
    (3148 . "TELUGU VOWEL SIGN AU")
    (3149 . "TELUGU SIGN VIRAMA")
    (3157 . "TELUGU LENGTH MARK")
    (3158 . "TELUGU AI LENGTH MARK")
    (3168 . "TELUGU LETTER VOCALIC RR")
    (3169 . "TELUGU LETTER VOCALIC LL")
    (3174 . "TELUGU DIGIT ZERO")
    (3175 . "TELUGU DIGIT ONE")
    (3176 . "TELUGU DIGIT TWO")
    (3177 . "TELUGU DIGIT THREE")
    (3178 . "TELUGU DIGIT FOUR")
    (3179 . "TELUGU DIGIT FIVE")
    (3180 . "TELUGU DIGIT SIX")
    (3181 . "TELUGU DIGIT SEVEN")
    (3182 . "TELUGU DIGIT EIGHT")
    (3183 . "TELUGU DIGIT NINE")
    (3202 . "KANNADA SIGN ANUSVARA")
    (3203 . "KANNADA SIGN VISARGA")
    (3205 . "KANNADA LETTER A")
    (3206 . "KANNADA LETTER AA")
    (3207 . "KANNADA LETTER I")
    (3208 . "KANNADA LETTER II")
    (3209 . "KANNADA LETTER U")
    (3210 . "KANNADA LETTER UU")
    (3211 . "KANNADA LETTER VOCALIC R")
    (3212 . "KANNADA LETTER VOCALIC L")
    (3214 . "KANNADA LETTER E")
    (3215 . "KANNADA LETTER EE")
    (3216 . "KANNADA LETTER AI")
    (3218 . "KANNADA LETTER O")
    (3219 . "KANNADA LETTER OO")
    (3220 . "KANNADA LETTER AU")
    (3221 . "KANNADA LETTER KA")
    (3222 . "KANNADA LETTER KHA")
    (3223 . "KANNADA LETTER GA")
    (3224 . "KANNADA LETTER GHA")
    (3225 . "KANNADA LETTER NGA")
    (3226 . "KANNADA LETTER CA")
    (3227 . "KANNADA LETTER CHA")
    (3228 . "KANNADA LETTER JA")
    (3229 . "KANNADA LETTER JHA")
    (3230 . "KANNADA LETTER NYA")
    (3231 . "KANNADA LETTER TTA")
    (3232 . "KANNADA LETTER TTHA")
    (3233 . "KANNADA LETTER DDA")
    (3234 . "KANNADA LETTER DDHA")
    (3235 . "KANNADA LETTER NNA")
    (3236 . "KANNADA LETTER TA")
    (3237 . "KANNADA LETTER THA")
    (3238 . "KANNADA LETTER DA")
    (3239 . "KANNADA LETTER DHA")
    (3240 . "KANNADA LETTER NA")
    (3242 . "KANNADA LETTER PA")
    (3243 . "KANNADA LETTER PHA")
    (3244 . "KANNADA LETTER BA")
    (3245 . "KANNADA LETTER BHA")
    (3246 . "KANNADA LETTER MA")
    (3247 . "KANNADA LETTER YA")
    (3248 . "KANNADA LETTER RA")
    (3249 . "KANNADA LETTER RRA")
    (3250 . "KANNADA LETTER LA")
    (3251 . "KANNADA LETTER LLA")
    (3253 . "KANNADA LETTER VA")
    (3254 . "KANNADA LETTER SHA")
    (3255 . "KANNADA LETTER SSA")
    (3256 . "KANNADA LETTER SA")
    (3257 . "KANNADA LETTER HA")
    (3262 . "KANNADA VOWEL SIGN AA")
    (3263 . "KANNADA VOWEL SIGN I")
    (3264 . "KANNADA VOWEL SIGN II")
    (3265 . "KANNADA VOWEL SIGN U")
    (3266 . "KANNADA VOWEL SIGN UU")
    (3267 . "KANNADA VOWEL SIGN VOCALIC R")
    (3268 . "KANNADA VOWEL SIGN VOCALIC RR")
    (3270 . "KANNADA VOWEL SIGN E")
    (3271 . "KANNADA VOWEL SIGN EE")
    (3272 . "KANNADA VOWEL SIGN AI")
    (3274 . "KANNADA VOWEL SIGN O")
    (3275 . "KANNADA VOWEL SIGN OO")
    (3276 . "KANNADA VOWEL SIGN AU")
    (3277 . "KANNADA SIGN VIRAMA")
    (3285 . "KANNADA LENGTH MARK")
    (3286 . "KANNADA AI LENGTH MARK")
    (3294 . "KANNADA LETTER FA")
    (3296 . "KANNADA LETTER VOCALIC RR")
    (3297 . "KANNADA LETTER VOCALIC LL")
    (3302 . "KANNADA DIGIT ZERO")
    (3303 . "KANNADA DIGIT ONE")
    (3304 . "KANNADA DIGIT TWO")
    (3305 . "KANNADA DIGIT THREE")
    (3306 . "KANNADA DIGIT FOUR")
    (3307 . "KANNADA DIGIT FIVE")
    (3308 . "KANNADA DIGIT SIX")
    (3309 . "KANNADA DIGIT SEVEN")
    (3310 . "KANNADA DIGIT EIGHT")
    (3311 . "KANNADA DIGIT NINE")
    (3330 . "MALAYALAM SIGN ANUSVARA")
    (3331 . "MALAYALAM SIGN VISARGA")
    (3333 . "MALAYALAM LETTER A")
    (3334 . "MALAYALAM LETTER AA")
    (3335 . "MALAYALAM LETTER I")
    (3336 . "MALAYALAM LETTER II")
    (3337 . "MALAYALAM LETTER U")
    (3338 . "MALAYALAM LETTER UU")
    (3339 . "MALAYALAM LETTER VOCALIC R")
    (3340 . "MALAYALAM LETTER VOCALIC L")
    (3342 . "MALAYALAM LETTER E")
    (3343 . "MALAYALAM LETTER EE")
    (3344 . "MALAYALAM LETTER AI")
    (3346 . "MALAYALAM LETTER O")
    (3347 . "MALAYALAM LETTER OO")
    (3348 . "MALAYALAM LETTER AU")
    (3349 . "MALAYALAM LETTER KA")
    (3350 . "MALAYALAM LETTER KHA")
    (3351 . "MALAYALAM LETTER GA")
    (3352 . "MALAYALAM LETTER GHA")
    (3353 . "MALAYALAM LETTER NGA")
    (3354 . "MALAYALAM LETTER CA")
    (3355 . "MALAYALAM LETTER CHA")
    (3356 . "MALAYALAM LETTER JA")
    (3357 . "MALAYALAM LETTER JHA")
    (3358 . "MALAYALAM LETTER NYA")
    (3359 . "MALAYALAM LETTER TTA")
    (3360 . "MALAYALAM LETTER TTHA")
    (3361 . "MALAYALAM LETTER DDA")
    (3362 . "MALAYALAM LETTER DDHA")
    (3363 . "MALAYALAM LETTER NNA")
    (3364 . "MALAYALAM LETTER TA")
    (3365 . "MALAYALAM LETTER THA")
    (3366 . "MALAYALAM LETTER DA")
    (3367 . "MALAYALAM LETTER DHA")
    (3368 . "MALAYALAM LETTER NA")
    (3370 . "MALAYALAM LETTER PA")
    (3371 . "MALAYALAM LETTER PHA")
    (3372 . "MALAYALAM LETTER BA")
    (3373 . "MALAYALAM LETTER BHA")
    (3374 . "MALAYALAM LETTER MA")
    (3375 . "MALAYALAM LETTER YA")
    (3376 . "MALAYALAM LETTER RA")
    (3377 . "MALAYALAM LETTER RRA")
    (3378 . "MALAYALAM LETTER LA")
    (3379 . "MALAYALAM LETTER LLA")
    (3380 . "MALAYALAM LETTER LLLA")
    (3381 . "MALAYALAM LETTER VA")
    (3382 . "MALAYALAM LETTER SHA")
    (3383 . "MALAYALAM LETTER SSA")
    (3384 . "MALAYALAM LETTER SA")
    (3385 . "MALAYALAM LETTER HA")
    (3390 . "MALAYALAM VOWEL SIGN AA")
    (3391 . "MALAYALAM VOWEL SIGN I")
    (3392 . "MALAYALAM VOWEL SIGN II")
    (3393 . "MALAYALAM VOWEL SIGN U")
    (3394 . "MALAYALAM VOWEL SIGN UU")
    (3395 . "MALAYALAM VOWEL SIGN VOCALIC R")
    (3398 . "MALAYALAM VOWEL SIGN E")
    (3399 . "MALAYALAM VOWEL SIGN EE")
    (3400 . "MALAYALAM VOWEL SIGN AI")
    (3402 . "MALAYALAM VOWEL SIGN O")
    (3403 . "MALAYALAM VOWEL SIGN OO")
    (3404 . "MALAYALAM VOWEL SIGN AU")
    (3405 . "MALAYALAM SIGN VIRAMA")
    (3415 . "MALAYALAM AU LENGTH MARK")
    (3424 . "MALAYALAM LETTER VOCALIC RR")
    (3425 . "MALAYALAM LETTER VOCALIC LL")
    (3430 . "MALAYALAM DIGIT ZERO")
    (3431 . "MALAYALAM DIGIT ONE")
    (3432 . "MALAYALAM DIGIT TWO")
    (3433 . "MALAYALAM DIGIT THREE")
    (3434 . "MALAYALAM DIGIT FOUR")
    (3435 . "MALAYALAM DIGIT FIVE")
    (3436 . "MALAYALAM DIGIT SIX")
    (3437 . "MALAYALAM DIGIT SEVEN")
    (3438 . "MALAYALAM DIGIT EIGHT")
    (3439 . "MALAYALAM DIGIT NINE")
    (3585 . "THAI CHARACTER KO KAI")
    (3586 . "THAI CHARACTER KHO KHAI")
    (3587 . "THAI CHARACTER KHO KHUAT")
    (3588 . "THAI CHARACTER KHO KHWAI")
    (3589 . "THAI CHARACTER KHO KHON")
    (3590 . "THAI CHARACTER KHO RAKHANG")
    (3591 . "THAI CHARACTER NGO NGU")
    (3592 . "THAI CHARACTER CHO CHAN")
    (3593 . "THAI CHARACTER CHO CHING")
    (3594 . "THAI CHARACTER CHO CHANG")
    (3595 . "THAI CHARACTER SO SO")
    (3596 . "THAI CHARACTER CHO CHOE")
    (3597 . "THAI CHARACTER YO YING")
    (3598 . "THAI CHARACTER DO CHADA")
    (3599 . "THAI CHARACTER TO PATAK")
    (3600 . "THAI CHARACTER THO THAN")
    (3601 . "THAI CHARACTER THO NANGMONTHO")
    (3602 . "THAI CHARACTER THO PHUTHAO")
    (3603 . "THAI CHARACTER NO NEN")
    (3604 . "THAI CHARACTER DO DEK")
    (3605 . "THAI CHARACTER TO TAO")
    (3606 . "THAI CHARACTER THO THUNG")
    (3607 . "THAI CHARACTER THO THAHAN")
    (3608 . "THAI CHARACTER THO THONG")
    (3609 . "THAI CHARACTER NO NU")
    (3610 . "THAI CHARACTER BO BAIMAI")
    (3611 . "THAI CHARACTER PO PLA")
    (3612 . "THAI CHARACTER PHO PHUNG")
    (3613 . "THAI CHARACTER FO FA")
    (3614 . "THAI CHARACTER PHO PHAN")
    (3615 . "THAI CHARACTER FO FAN")
    (3616 . "THAI CHARACTER PHO SAMPHAO")
    (3617 . "THAI CHARACTER MO MA")
    (3618 . "THAI CHARACTER YO YAK")
    (3619 . "THAI CHARACTER RO RUA")
    (3620 . "THAI CHARACTER RU")
    (3621 . "THAI CHARACTER LO LING")
    (3622 . "THAI CHARACTER LU")
    (3623 . "THAI CHARACTER WO WAEN")
    (3624 . "THAI CHARACTER SO SALA")
    (3625 . "THAI CHARACTER SO RUSI")
    (3626 . "THAI CHARACTER SO SUA")
    (3627 . "THAI CHARACTER HO HIP")
    (3628 . "THAI CHARACTER LO CHULA")
    (3629 . "THAI CHARACTER O ANG")
    (3630 . "THAI CHARACTER HO NOKHUK")
    (3631 . "THAI CHARACTER PAIYANNOI")
    (3632 . "THAI CHARACTER SARA A")
    (3633 . "THAI CHARACTER MAI HAN-AKAT")
    (3634 . "THAI CHARACTER SARA AA")
    (3635 . "THAI CHARACTER SARA AM")
    (3636 . "THAI CHARACTER SARA I")
    (3637 . "THAI CHARACTER SARA II")
    (3638 . "THAI CHARACTER SARA UE")
    (3639 . "THAI CHARACTER SARA UEE")
    (3640 . "THAI CHARACTER SARA U")
    (3641 . "THAI CHARACTER SARA UU")
    (3642 . "THAI CHARACTER PHINTHU")
    (3647 . "THAI CURRENCY SYMBOL BAHT")
    (3648 . "THAI CHARACTER SARA E")
    (3649 . "THAI CHARACTER SARA AE")
    (3650 . "THAI CHARACTER SARA O")
    (3651 . "THAI CHARACTER SARA AI MAIMUAN")
    (3652 . "THAI CHARACTER SARA AI MAIMALAI")
    (3653 . "THAI CHARACTER LAKKHANGYAO")
    (3654 . "THAI CHARACTER MAIYAMOK")
    (3655 . "THAI CHARACTER MAITAIKHU")
    (3656 . "THAI CHARACTER MAI EK")
    (3657 . "THAI CHARACTER MAI THO")
    (3658 . "THAI CHARACTER MAI TRI")
    (3659 . "THAI CHARACTER MAI CHATTAWA")
    (3660 . "THAI CHARACTER THANTHAKHAT")
    (3661 . "THAI CHARACTER NIKHAHIT")
    (3662 . "THAI CHARACTER YAMAKKAN")
    (3663 . "THAI CHARACTER FONGMAN")
    (3664 . "THAI DIGIT ZERO")
    (3665 . "THAI DIGIT ONE")
    (3666 . "THAI DIGIT TWO")
    (3667 . "THAI DIGIT THREE")
    (3668 . "THAI DIGIT FOUR")
    (3669 . "THAI DIGIT FIVE")
    (3670 . "THAI DIGIT SIX")
    (3671 . "THAI DIGIT SEVEN")
    (3672 . "THAI DIGIT EIGHT")
    (3673 . "THAI DIGIT NINE")
    (3674 . "THAI CHARACTER ANGKHANKHU")
    (3675 . "THAI CHARACTER KHOMUT")
    (3713 . "LAO LETTER KO")
    (3714 . "LAO LETTER KHO SUNG")
    (3716 . "LAO LETTER KHO TAM")
    (3719 . "LAO LETTER NGO")
    (3720 . "LAO LETTER CO")
    (3722 . "LAO LETTER SO TAM")
    (3725 . "LAO LETTER NYO")
    (3732 . "LAO LETTER DO")
    (3733 . "LAO LETTER TO")
    (3734 . "LAO LETTER THO SUNG")
    (3735 . "LAO LETTER THO TAM")
    (3737 . "LAO LETTER NO")
    (3738 . "LAO LETTER BO")
    (3739 . "LAO LETTER PO")
    (3740 . "LAO LETTER PHO SUNG")
    (3741 . "LAO LETTER FO TAM")
    (3742 . "LAO LETTER PHO TAM")
    (3743 . "LAO LETTER FO SUNG")
    (3745 . "LAO LETTER MO")
    (3746 . "LAO LETTER YO")
    (3747 . "LAO LETTER LO LING")
    (3749 . "LAO LETTER LO LOOT")
    (3751 . "LAO LETTER WO")
    (3754 . "LAO LETTER SO SUNG")
    (3755 . "LAO LETTER HO SUNG")
    (3757 . "LAO LETTER O")
    (3758 . "LAO LETTER HO TAM")
    (3759 . "LAO ELLIPSIS")
    (3760 . "LAO VOWEL SIGN A")
    (3761 . "LAO VOWEL SIGN MAI KAN")
    (3762 . "LAO VOWEL SIGN AA")
    (3763 . "LAO VOWEL SIGN AM")
    (3764 . "LAO VOWEL SIGN I")
    (3765 . "LAO VOWEL SIGN II")
    (3766 . "LAO VOWEL SIGN Y")
    (3767 . "LAO VOWEL SIGN YY")
    (3768 . "LAO VOWEL SIGN U")
    (3769 . "LAO VOWEL SIGN UU")
    (3771 . "LAO VOWEL SIGN MAI KON")
    (3772 . "LAO SEMIVOWEL SIGN LO")
    (3773 . "LAO SEMIVOWEL SIGN NYO")
    (3776 . "LAO VOWEL SIGN E")
    (3777 . "LAO VOWEL SIGN EI")
    (3778 . "LAO VOWEL SIGN O")
    (3779 . "LAO VOWEL SIGN AY")
    (3780 . "LAO VOWEL SIGN AI")
    (3782 . "LAO KO LA")
    (3784 . "LAO TONE MAI EK")
    (3785 . "LAO TONE MAI THO")
    (3786 . "LAO TONE MAI TI")
    (3787 . "LAO TONE MAI CATAWA")
    (3788 . "LAO CANCELLATION MARK")
    (3789 . "LAO NIGGAHITA")
    (3792 . "LAO DIGIT ZERO")
    (3793 . "LAO DIGIT ONE")
    (3794 . "LAO DIGIT TWO")
    (3795 . "LAO DIGIT THREE")
    (3796 . "LAO DIGIT FOUR")
    (3797 . "LAO DIGIT FIVE")
    (3798 . "LAO DIGIT SIX")
    (3799 . "LAO DIGIT SEVEN")
    (3800 . "LAO DIGIT EIGHT")
    (3801 . "LAO DIGIT NINE")
    (3804 . "LAO HO NO")
    (3805 . "LAO HO MO")
    (4256 . "GEORGIAN CAPITAL LETTER AN")
    (4257 . "GEORGIAN CAPITAL LETTER BAN")
    (4258 . "GEORGIAN CAPITAL LETTER GAN")
    (4259 . "GEORGIAN CAPITAL LETTER DON")
    (4260 . "GEORGIAN CAPITAL LETTER EN")
    (4261 . "GEORGIAN CAPITAL LETTER VIN")
    (4262 . "GEORGIAN CAPITAL LETTER ZEN")
    (4263 . "GEORGIAN CAPITAL LETTER TAN")
    (4264 . "GEORGIAN CAPITAL LETTER IN")
    (4265 . "GEORGIAN CAPITAL LETTER KAN")
    (4266 . "GEORGIAN CAPITAL LETTER LAS")
    (4267 . "GEORGIAN CAPITAL LETTER MAN")
    (4268 . "GEORGIAN CAPITAL LETTER NAR")
    (4269 . "GEORGIAN CAPITAL LETTER ON")
    (4270 . "GEORGIAN CAPITAL LETTER PAR")
    (4271 . "GEORGIAN CAPITAL LETTER ZHAR")
    (4272 . "GEORGIAN CAPITAL LETTER RAE")
    (4273 . "GEORGIAN CAPITAL LETTER SAN")
    (4274 . "GEORGIAN CAPITAL LETTER TAR")
    (4275 . "GEORGIAN CAPITAL LETTER UN")
    (4276 . "GEORGIAN CAPITAL LETTER PHAR")
    (4277 . "GEORGIAN CAPITAL LETTER KHAR")
    (4278 . "GEORGIAN CAPITAL LETTER GHAN")
    (4279 . "GEORGIAN CAPITAL LETTER QAR")
    (4280 . "GEORGIAN CAPITAL LETTER SHIN")
    (4281 . "GEORGIAN CAPITAL LETTER CHIN")
    (4282 . "GEORGIAN CAPITAL LETTER CAN")
    (4283 . "GEORGIAN CAPITAL LETTER JIL")
    (4284 . "GEORGIAN CAPITAL LETTER CIL")
    (4285 . "GEORGIAN CAPITAL LETTER CHAR")
    (4286 . "GEORGIAN CAPITAL LETTER XAN")
    (4287 . "GEORGIAN CAPITAL LETTER JHAN")
    (4288 . "GEORGIAN CAPITAL LETTER HAE")
    (4289 . "GEORGIAN CAPITAL LETTER HE")
    (4290 . "GEORGIAN CAPITAL LETTER HIE")
    (4291 . "GEORGIAN CAPITAL LETTER WE")
    (4292 . "GEORGIAN CAPITAL LETTER HAR")
    (4293 . "GEORGIAN CAPITAL LETTER HOE")
    (4304 . "GEORGIAN LETTER AN")
    (4305 . "GEORGIAN LETTER BAN")
    (4306 . "GEORGIAN LETTER GAN")
    (4307 . "GEORGIAN LETTER DON")
    (4308 . "GEORGIAN LETTER EN")
    (4309 . "GEORGIAN LETTER VIN")
    (4310 . "GEORGIAN LETTER ZEN")
    (4311 . "GEORGIAN LETTER TAN")
    (4312 . "GEORGIAN LETTER IN")
    (4313 . "GEORGIAN LETTER KAN")
    (4314 . "GEORGIAN LETTER LAS")
    (4315 . "GEORGIAN LETTER MAN")
    (4316 . "GEORGIAN LETTER NAR")
    (4317 . "GEORGIAN LETTER ON")
    (4318 . "GEORGIAN LETTER PAR")
    (4319 . "GEORGIAN LETTER ZHAR")
    (4320 . "GEORGIAN LETTER RAE")
    (4321 . "GEORGIAN LETTER SAN")
    (4322 . "GEORGIAN LETTER TAR")
    (4323 . "GEORGIAN LETTER UN")
    (4324 . "GEORGIAN LETTER PHAR")
    (4325 . "GEORGIAN LETTER KHAR")
    (4326 . "GEORGIAN LETTER GHAN")
    (4327 . "GEORGIAN LETTER QAR")
    (4328 . "GEORGIAN LETTER SHIN")
    (4329 . "GEORGIAN LETTER CHIN")
    (4330 . "GEORGIAN LETTER CAN")
    (4331 . "GEORGIAN LETTER JIL")
    (4332 . "GEORGIAN LETTER CIL")
    (4333 . "GEORGIAN LETTER CHAR")
    (4334 . "GEORGIAN LETTER XAN")
    (4335 . "GEORGIAN LETTER JHAN")
    (4336 . "GEORGIAN LETTER HAE")
    (4337 . "GEORGIAN LETTER HE")
    (4338 . "GEORGIAN LETTER HIE")
    (4339 . "GEORGIAN LETTER WE")
    (4340 . "GEORGIAN LETTER HAR")
    (4341 . "GEORGIAN LETTER HOE")
    (4342 . "GEORGIAN LETTER FI")
    (4347 . "GEORGIAN PARAGRAPH SEPARATOR")
    (4352 . "HANGUL CHOSEONG KIYEOK")
    (4353 . "HANGUL CHOSEONG SSANGKIYEOK")
    (4354 . "HANGUL CHOSEONG NIEUN")
    (4355 . "HANGUL CHOSEONG TIKEUT")
    (4356 . "HANGUL CHOSEONG SSANGTIKEUT")
    (4357 . "HANGUL CHOSEONG RIEUL")
    (4358 . "HANGUL CHOSEONG MIEUM")
    (4359 . "HANGUL CHOSEONG PIEUP")
    (4360 . "HANGUL CHOSEONG SSANGPIEUP")
    (4361 . "HANGUL CHOSEONG SIOS")
    (4362 . "HANGUL CHOSEONG SSANGSIOS")
    (4363 . "HANGUL CHOSEONG IEUNG")
    (4364 . "HANGUL CHOSEONG CIEUC")
    (4365 . "HANGUL CHOSEONG SSANGCIEUC")
    (4366 . "HANGUL CHOSEONG CHIEUCH")
    (4367 . "HANGUL CHOSEONG KHIEUKH")
    (4368 . "HANGUL CHOSEONG THIEUTH")
    (4369 . "HANGUL CHOSEONG PHIEUPH")
    (4370 . "HANGUL CHOSEONG HIEUH")
    (4371 . "HANGUL CHOSEONG NIEUN-KIYEOK")
    (4372 . "HANGUL CHOSEONG SSANGNIEUN")
    (4373 . "HANGUL CHOSEONG NIEUN-TIKEUT")
    (4374 . "HANGUL CHOSEONG NIEUN-PIEUP")
    (4375 . "HANGUL CHOSEONG TIKEUT-KIYEOK")
    (4376 . "HANGUL CHOSEONG RIEUL-NIEUN")
    (4377 . "HANGUL CHOSEONG SSANGRIEUL")
    (4378 . "HANGUL CHOSEONG RIEUL-HIEUH")
    (4379 . "HANGUL CHOSEONG KAPYEOUNRIEUL")
    (4380 . "HANGUL CHOSEONG MIEUM-PIEUP")
    (4381 . "HANGUL CHOSEONG KAPYEOUNMIEUM")
    (4382 . "HANGUL CHOSEONG PIEUP-KIYEOK")
    (4383 . "HANGUL CHOSEONG PIEUP-NIEUN")
    (4384 . "HANGUL CHOSEONG PIEUP-TIKEUT")
    (4385 . "HANGUL CHOSEONG PIEUP-SIOS")
    (4386 . "HANGUL CHOSEONG PIEUP-SIOS-KIYEOK")
    (4387 . "HANGUL CHOSEONG PIEUP-SIOS-TIKEUT")
    (4388 . "HANGUL CHOSEONG PIEUP-SIOS-PIEUP")
    (4389 . "HANGUL CHOSEONG PIEUP-SSANGSIOS")
    (4390 . "HANGUL CHOSEONG PIEUP-SIOS-CIEUC")
    (4391 . "HANGUL CHOSEONG PIEUP-CIEUC")
    (4392 . "HANGUL CHOSEONG PIEUP-CHIEUCH")
    (4393 . "HANGUL CHOSEONG PIEUP-THIEUTH")
    (4394 . "HANGUL CHOSEONG PIEUP-PHIEUPH")
    (4395 . "HANGUL CHOSEONG KAPYEOUNPIEUP")
    (4396 . "HANGUL CHOSEONG KAPYEOUNSSANGPIEUP")
    (4397 . "HANGUL CHOSEONG SIOS-KIYEOK")
    (4398 . "HANGUL CHOSEONG SIOS-NIEUN")
    (4399 . "HANGUL CHOSEONG SIOS-TIKEUT")
    (4400 . "HANGUL CHOSEONG SIOS-RIEUL")
    (4401 . "HANGUL CHOSEONG SIOS-MIEUM")
    (4402 . "HANGUL CHOSEONG SIOS-PIEUP")
    (4403 . "HANGUL CHOSEONG SIOS-PIEUP-KIYEOK")
    (4404 . "HANGUL CHOSEONG SIOS-SSANGSIOS")
    (4405 . "HANGUL CHOSEONG SIOS-IEUNG")
    (4406 . "HANGUL CHOSEONG SIOS-CIEUC")
    (4407 . "HANGUL CHOSEONG SIOS-CHIEUCH")
    (4408 . "HANGUL CHOSEONG SIOS-KHIEUKH")
    (4409 . "HANGUL CHOSEONG SIOS-THIEUTH")
    (4410 . "HANGUL CHOSEONG SIOS-PHIEUPH")
    (4411 . "HANGUL CHOSEONG SIOS-HIEUH")
    (4412 . "HANGUL CHOSEONG CHITUEUMSIOS")
    (4413 . "HANGUL CHOSEONG CHITUEUMSSANGSIOS")
    (4414 . "HANGUL CHOSEONG CEONGCHIEUMSIOS")
    (4415 . "HANGUL CHOSEONG CEONGCHIEUMSSANGSIOS")
    (4416 . "HANGUL CHOSEONG PANSIOS")
    (4417 . "HANGUL CHOSEONG IEUNG-KIYEOK")
    (4418 . "HANGUL CHOSEONG IEUNG-TIKEUT")
    (4419 . "HANGUL CHOSEONG IEUNG-MIEUM")
    (4420 . "HANGUL CHOSEONG IEUNG-PIEUP")
    (4421 . "HANGUL CHOSEONG IEUNG-SIOS")
    (4422 . "HANGUL CHOSEONG IEUNG-PANSIOS")
    (4423 . "HANGUL CHOSEONG SSANGIEUNG")
    (4424 . "HANGUL CHOSEONG IEUNG-CIEUC")
    (4425 . "HANGUL CHOSEONG IEUNG-CHIEUCH")
    (4426 . "HANGUL CHOSEONG IEUNG-THIEUTH")
    (4427 . "HANGUL CHOSEONG IEUNG-PHIEUPH")
    (4428 . "HANGUL CHOSEONG YESIEUNG")
    (4429 . "HANGUL CHOSEONG CIEUC-IEUNG")
    (4430 . "HANGUL CHOSEONG CHITUEUMCIEUC")
    (4431 . "HANGUL CHOSEONG CHITUEUMSSANGCIEUC")
    (4432 . "HANGUL CHOSEONG CEONGCHIEUMCIEUC")
    (4433 . "HANGUL CHOSEONG CEONGCHIEUMSSANGCIEUC")
    (4434 . "HANGUL CHOSEONG CHIEUCH-KHIEUKH")
    (4435 . "HANGUL CHOSEONG CHIEUCH-HIEUH")
    (4436 . "HANGUL CHOSEONG CHITUEUMCHIEUCH")
    (4437 . "HANGUL CHOSEONG CEONGCHIEUMCHIEUCH")
    (4438 . "HANGUL CHOSEONG PHIEUPH-PIEUP")
    (4439 . "HANGUL CHOSEONG KAPYEOUNPHIEUPH")
    (4440 . "HANGUL CHOSEONG SSANGHIEUH")
    (4441 . "HANGUL CHOSEONG YEORINHIEUH")
    (4447 . "HANGUL CHOSEONG FILLER")
    (4448 . "HANGUL JUNGSEONG FILLER")
    (4449 . "HANGUL JUNGSEONG A")
    (4450 . "HANGUL JUNGSEONG AE")
    (4451 . "HANGUL JUNGSEONG YA")
    (4452 . "HANGUL JUNGSEONG YAE")
    (4453 . "HANGUL JUNGSEONG EO")
    (4454 . "HANGUL JUNGSEONG E")
    (4455 . "HANGUL JUNGSEONG YEO")
    (4456 . "HANGUL JUNGSEONG YE")
    (4457 . "HANGUL JUNGSEONG O")
    (4458 . "HANGUL JUNGSEONG WA")
    (4459 . "HANGUL JUNGSEONG WAE")
    (4460 . "HANGUL JUNGSEONG OE")
    (4461 . "HANGUL JUNGSEONG YO")
    (4462 . "HANGUL JUNGSEONG U")
    (4463 . "HANGUL JUNGSEONG WEO")
    (4464 . "HANGUL JUNGSEONG WE")
    (4465 . "HANGUL JUNGSEONG WI")
    (4466 . "HANGUL JUNGSEONG YU")
    (4467 . "HANGUL JUNGSEONG EU")
    (4468 . "HANGUL JUNGSEONG YI")
    (4469 . "HANGUL JUNGSEONG I")
    (4470 . "HANGUL JUNGSEONG A-O")
    (4471 . "HANGUL JUNGSEONG A-U")
    (4472 . "HANGUL JUNGSEONG YA-O")
    (4473 . "HANGUL JUNGSEONG YA-YO")
    (4474 . "HANGUL JUNGSEONG EO-O")
    (4475 . "HANGUL JUNGSEONG EO-U")
    (4476 . "HANGUL JUNGSEONG EO-EU")
    (4477 . "HANGUL JUNGSEONG YEO-O")
    (4478 . "HANGUL JUNGSEONG YEO-U")
    (4479 . "HANGUL JUNGSEONG O-EO")
    (4480 . "HANGUL JUNGSEONG O-E")
    (4481 . "HANGUL JUNGSEONG O-YE")
    (4482 . "HANGUL JUNGSEONG O-O")
    (4483 . "HANGUL JUNGSEONG O-U")
    (4484 . "HANGUL JUNGSEONG YO-YA")
    (4485 . "HANGUL JUNGSEONG YO-YAE")
    (4486 . "HANGUL JUNGSEONG YO-YEO")
    (4487 . "HANGUL JUNGSEONG YO-O")
    (4488 . "HANGUL JUNGSEONG YO-I")
    (4489 . "HANGUL JUNGSEONG U-A")
    (4490 . "HANGUL JUNGSEONG U-AE")
    (4491 . "HANGUL JUNGSEONG U-EO-EU")
    (4492 . "HANGUL JUNGSEONG U-YE")
    (4493 . "HANGUL JUNGSEONG U-U")
    (4494 . "HANGUL JUNGSEONG YU-A")
    (4495 . "HANGUL JUNGSEONG YU-EO")
    (4496 . "HANGUL JUNGSEONG YU-E")
    (4497 . "HANGUL JUNGSEONG YU-YEO")
    (4498 . "HANGUL JUNGSEONG YU-YE")
    (4499 . "HANGUL JUNGSEONG YU-U")
    (4500 . "HANGUL JUNGSEONG YU-I")
    (4501 . "HANGUL JUNGSEONG EU-U")
    (4502 . "HANGUL JUNGSEONG EU-EU")
    (4503 . "HANGUL JUNGSEONG YI-U")
    (4504 . "HANGUL JUNGSEONG I-A")
    (4505 . "HANGUL JUNGSEONG I-YA")
    (4506 . "HANGUL JUNGSEONG I-O")
    (4507 . "HANGUL JUNGSEONG I-U")
    (4508 . "HANGUL JUNGSEONG I-EU")
    (4509 . "HANGUL JUNGSEONG I-ARAEA")
    (4510 . "HANGUL JUNGSEONG ARAEA")
    (4511 . "HANGUL JUNGSEONG ARAEA-EO")
    (4512 . "HANGUL JUNGSEONG ARAEA-U")
    (4513 . "HANGUL JUNGSEONG ARAEA-I")
    (4514 . "HANGUL JUNGSEONG SSANGARAEA")
    (4520 . "HANGUL JONGSEONG KIYEOK")
    (4521 . "HANGUL JONGSEONG SSANGKIYEOK")
    (4522 . "HANGUL JONGSEONG KIYEOK-SIOS")
    (4523 . "HANGUL JONGSEONG NIEUN")
    (4524 . "HANGUL JONGSEONG NIEUN-CIEUC")
    (4525 . "HANGUL JONGSEONG NIEUN-HIEUH")
    (4526 . "HANGUL JONGSEONG TIKEUT")
    (4527 . "HANGUL JONGSEONG RIEUL")
    (4528 . "HANGUL JONGSEONG RIEUL-KIYEOK")
    (4529 . "HANGUL JONGSEONG RIEUL-MIEUM")
    (4530 . "HANGUL JONGSEONG RIEUL-PIEUP")
    (4531 . "HANGUL JONGSEONG RIEUL-SIOS")
    (4532 . "HANGUL JONGSEONG RIEUL-THIEUTH")
    (4533 . "HANGUL JONGSEONG RIEUL-PHIEUPH")
    (4534 . "HANGUL JONGSEONG RIEUL-HIEUH")
    (4535 . "HANGUL JONGSEONG MIEUM")
    (4536 . "HANGUL JONGSEONG PIEUP")
    (4537 . "HANGUL JONGSEONG PIEUP-SIOS")
    (4538 . "HANGUL JONGSEONG SIOS")
    (4539 . "HANGUL JONGSEONG SSANGSIOS")
    (4540 . "HANGUL JONGSEONG IEUNG")
    (4541 . "HANGUL JONGSEONG CIEUC")
    (4542 . "HANGUL JONGSEONG CHIEUCH")
    (4543 . "HANGUL JONGSEONG KHIEUKH")
    (4544 . "HANGUL JONGSEONG THIEUTH")
    (4545 . "HANGUL JONGSEONG PHIEUPH")
    (4546 . "HANGUL JONGSEONG HIEUH")
    (4547 . "HANGUL JONGSEONG KIYEOK-RIEUL")
    (4548 . "HANGUL JONGSEONG KIYEOK-SIOS-KIYEOK")
    (4549 . "HANGUL JONGSEONG NIEUN-KIYEOK")
    (4550 . "HANGUL JONGSEONG NIEUN-TIKEUT")
    (4551 . "HANGUL JONGSEONG NIEUN-SIOS")
    (4552 . "HANGUL JONGSEONG NIEUN-PANSIOS")
    (4553 . "HANGUL JONGSEONG NIEUN-THIEUTH")
    (4554 . "HANGUL JONGSEONG TIKEUT-KIYEOK")
    (4555 . "HANGUL JONGSEONG TIKEUT-RIEUL")
    (4556 . "HANGUL JONGSEONG RIEUL-KIYEOK-SIOS")
    (4557 . "HANGUL JONGSEONG RIEUL-NIEUN")
    (4558 . "HANGUL JONGSEONG RIEUL-TIKEUT")
    (4559 . "HANGUL JONGSEONG RIEUL-TIKEUT-HIEUH")
    (4560 . "HANGUL JONGSEONG SSANGRIEUL")
    (4561 . "HANGUL JONGSEONG RIEUL-MIEUM-KIYEOK")
    (4562 . "HANGUL JONGSEONG RIEUL-MIEUM-SIOS")
    (4563 . "HANGUL JONGSEONG RIEUL-PIEUP-SIOS")
    (4564 . "HANGUL JONGSEONG RIEUL-PIEUP-HIEUH")
    (4565 . "HANGUL JONGSEONG RIEUL-KAPYEOUNPIEUP")
    (4566 . "HANGUL JONGSEONG RIEUL-SSANGSIOS")
    (4567 . "HANGUL JONGSEONG RIEUL-PANSIOS")
    (4568 . "HANGUL JONGSEONG RIEUL-KHIEUKH")
    (4569 . "HANGUL JONGSEONG RIEUL-YEORINHIEUH")
    (4570 . "HANGUL JONGSEONG MIEUM-KIYEOK")
    (4571 . "HANGUL JONGSEONG MIEUM-RIEUL")
    (4572 . "HANGUL JONGSEONG MIEUM-PIEUP")
    (4573 . "HANGUL JONGSEONG MIEUM-SIOS")
    (4574 . "HANGUL JONGSEONG MIEUM-SSANGSIOS")
    (4575 . "HANGUL JONGSEONG MIEUM-PANSIOS")
    (4576 . "HANGUL JONGSEONG MIEUM-CHIEUCH")
    (4577 . "HANGUL JONGSEONG MIEUM-HIEUH")
    (4578 . "HANGUL JONGSEONG KAPYEOUNMIEUM")
    (4579 . "HANGUL JONGSEONG PIEUP-RIEUL")
    (4580 . "HANGUL JONGSEONG PIEUP-PHIEUPH")
    (4581 . "HANGUL JONGSEONG PIEUP-HIEUH")
    (4582 . "HANGUL JONGSEONG KAPYEOUNPIEUP")
    (4583 . "HANGUL JONGSEONG SIOS-KIYEOK")
    (4584 . "HANGUL JONGSEONG SIOS-TIKEUT")
    (4585 . "HANGUL JONGSEONG SIOS-RIEUL")
    (4586 . "HANGUL JONGSEONG SIOS-PIEUP")
    (4587 . "HANGUL JONGSEONG PANSIOS")
    (4588 . "HANGUL JONGSEONG IEUNG-KIYEOK")
    (4589 . "HANGUL JONGSEONG IEUNG-SSANGKIYEOK")
    (4590 . "HANGUL JONGSEONG SSANGIEUNG")
    (4591 . "HANGUL JONGSEONG IEUNG-KHIEUKH")
    (4592 . "HANGUL JONGSEONG YESIEUNG")
    (4593 . "HANGUL JONGSEONG YESIEUNG-SIOS")
    (4594 . "HANGUL JONGSEONG YESIEUNG-PANSIOS")
    (4595 . "HANGUL JONGSEONG PHIEUPH-PIEUP")
    (4596 . "HANGUL JONGSEONG KAPYEOUNPHIEUPH")
    (4597 . "HANGUL JONGSEONG HIEUH-NIEUN")
    (4598 . "HANGUL JONGSEONG HIEUH-RIEUL")
    (4599 . "HANGUL JONGSEONG HIEUH-MIEUM")
    (4600 . "HANGUL JONGSEONG HIEUH-PIEUP")
    (4601 . "HANGUL JONGSEONG YEORINHIEUH")
    (7680 . "LATIN CAPITAL LETTER A WITH RING BELOW")
    (7681 . "LATIN SMALL LETTER A WITH RING BELOW")
    (7682 . "LATIN CAPITAL LETTER B WITH DOT ABOVE")
    (7683 . "LATIN SMALL LETTER B WITH DOT ABOVE")
    (7684 . "LATIN CAPITAL LETTER B WITH DOT BELOW")
    (7685 . "LATIN SMALL LETTER B WITH DOT BELOW")
    (7686 . "LATIN CAPITAL LETTER B WITH LINE BELOW")
    (7687 . "LATIN SMALL LETTER B WITH LINE BELOW")
    (7688 . "LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE")
    (7689 . "LATIN SMALL LETTER C WITH CEDILLA AND ACUTE")
    (7690 . "LATIN CAPITAL LETTER D WITH DOT ABOVE")
    (7691 . "LATIN SMALL LETTER D WITH DOT ABOVE")
    (7692 . "LATIN CAPITAL LETTER D WITH DOT BELOW")
    (7693 . "LATIN SMALL LETTER D WITH DOT BELOW")
    (7694 . "LATIN CAPITAL LETTER D WITH LINE BELOW")
    (7695 . "LATIN SMALL LETTER D WITH LINE BELOW")
    (7696 . "LATIN CAPITAL LETTER D WITH CEDILLA")
    (7697 . "LATIN SMALL LETTER D WITH CEDILLA")
    (7698 . "LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW")
    (7699 . "LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW")
    (7700 . "LATIN CAPITAL LETTER E WITH MACRON AND GRAVE")
    (7701 . "LATIN SMALL LETTER E WITH MACRON AND GRAVE")
    (7702 . "LATIN CAPITAL LETTER E WITH MACRON AND ACUTE")
    (7703 . "LATIN SMALL LETTER E WITH MACRON AND ACUTE")
    (7704 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW")
    (7705 . "LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW")
    (7706 . "LATIN CAPITAL LETTER E WITH TILDE BELOW")
    (7707 . "LATIN SMALL LETTER E WITH TILDE BELOW")
    (7708 . "LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE")
    (7709 . "LATIN SMALL LETTER E WITH CEDILLA AND BREVE")
    (7710 . "LATIN CAPITAL LETTER F WITH DOT ABOVE")
    (7711 . "LATIN SMALL LETTER F WITH DOT ABOVE")
    (7712 . "LATIN CAPITAL LETTER G WITH MACRON")
    (7713 . "LATIN SMALL LETTER G WITH MACRON")
    (7714 . "LATIN CAPITAL LETTER H WITH DOT ABOVE")
    (7715 . "LATIN SMALL LETTER H WITH DOT ABOVE")
    (7716 . "LATIN CAPITAL LETTER H WITH DOT BELOW")
    (7717 . "LATIN SMALL LETTER H WITH DOT BELOW")
    (7718 . "LATIN CAPITAL LETTER H WITH DIAERESIS")
    (7719 . "LATIN SMALL LETTER H WITH DIAERESIS")
    (7720 . "LATIN CAPITAL LETTER H WITH CEDILLA")
    (7721 . "LATIN SMALL LETTER H WITH CEDILLA")
    (7722 . "LATIN CAPITAL LETTER H WITH BREVE BELOW")
    (7723 . "LATIN SMALL LETTER H WITH BREVE BELOW")
    (7724 . "LATIN CAPITAL LETTER I WITH TILDE BELOW")
    (7725 . "LATIN SMALL LETTER I WITH TILDE BELOW")
    (7726 . "LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE")
    (7727 . "LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE")
    (7728 . "LATIN CAPITAL LETTER K WITH ACUTE")
    (7729 . "LATIN SMALL LETTER K WITH ACUTE")
    (7730 . "LATIN CAPITAL LETTER K WITH DOT BELOW")
    (7731 . "LATIN SMALL LETTER K WITH DOT BELOW")
    (7732 . "LATIN CAPITAL LETTER K WITH LINE BELOW")
    (7733 . "LATIN SMALL LETTER K WITH LINE BELOW")
    (7734 . "LATIN CAPITAL LETTER L WITH DOT BELOW")
    (7735 . "LATIN SMALL LETTER L WITH DOT BELOW")
    (7736 . "LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON")
    (7737 . "LATIN SMALL LETTER L WITH DOT BELOW AND MACRON")
    (7738 . "LATIN CAPITAL LETTER L WITH LINE BELOW")
    (7739 . "LATIN SMALL LETTER L WITH LINE BELOW")
    (7740 . "LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW")
    (7741 . "LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW")
    (7742 . "LATIN CAPITAL LETTER M WITH ACUTE")
    (7743 . "LATIN SMALL LETTER M WITH ACUTE")
    (7744 . "LATIN CAPITAL LETTER M WITH DOT ABOVE")
    (7745 . "LATIN SMALL LETTER M WITH DOT ABOVE")
    (7746 . "LATIN CAPITAL LETTER M WITH DOT BELOW")
    (7747 . "LATIN SMALL LETTER M WITH DOT BELOW")
    (7748 . "LATIN CAPITAL LETTER N WITH DOT ABOVE")
    (7749 . "LATIN SMALL LETTER N WITH DOT ABOVE")
    (7750 . "LATIN CAPITAL LETTER N WITH DOT BELOW")
    (7751 . "LATIN SMALL LETTER N WITH DOT BELOW")
    (7752 . "LATIN CAPITAL LETTER N WITH LINE BELOW")
    (7753 . "LATIN SMALL LETTER N WITH LINE BELOW")
    (7754 . "LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW")
    (7755 . "LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW")
    (7756 . "LATIN CAPITAL LETTER O WITH TILDE AND ACUTE")
    (7757 . "LATIN SMALL LETTER O WITH TILDE AND ACUTE")
    (7758 . "LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS")
    (7759 . "LATIN SMALL LETTER O WITH TILDE AND DIAERESIS")
    (7760 . "LATIN CAPITAL LETTER O WITH MACRON AND GRAVE")
    (7761 . "LATIN SMALL LETTER O WITH MACRON AND GRAVE")
    (7762 . "LATIN CAPITAL LETTER O WITH MACRON AND ACUTE")
    (7763 . "LATIN SMALL LETTER O WITH MACRON AND ACUTE")
    (7764 . "LATIN CAPITAL LETTER P WITH ACUTE")
    (7765 . "LATIN SMALL LETTER P WITH ACUTE")
    (7766 . "LATIN CAPITAL LETTER P WITH DOT ABOVE")
    (7767 . "LATIN SMALL LETTER P WITH DOT ABOVE")
    (7768 . "LATIN CAPITAL LETTER R WITH DOT ABOVE")
    (7769 . "LATIN SMALL LETTER R WITH DOT ABOVE")
    (7770 . "LATIN CAPITAL LETTER R WITH DOT BELOW")
    (7771 . "LATIN SMALL LETTER R WITH DOT BELOW")
    (7772 . "LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON")
    (7773 . "LATIN SMALL LETTER R WITH DOT BELOW AND MACRON")
    (7774 . "LATIN CAPITAL LETTER R WITH LINE BELOW")
    (7775 . "LATIN SMALL LETTER R WITH LINE BELOW")
    (7776 . "LATIN CAPITAL LETTER S WITH DOT ABOVE")
    (7777 . "LATIN SMALL LETTER S WITH DOT ABOVE")
    (7778 . "LATIN CAPITAL LETTER S WITH DOT BELOW")
    (7779 . "LATIN SMALL LETTER S WITH DOT BELOW")
    (7780 . "LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE")
    (7781 . "LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE")
    (7782 . "LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE")
    (7783 . "LATIN SMALL LETTER S WITH CARON AND DOT ABOVE")
    (7784 . "LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE")
    (7785 . "LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE")
    (7786 . "LATIN CAPITAL LETTER T WITH DOT ABOVE")
    (7787 . "LATIN SMALL LETTER T WITH DOT ABOVE")
    (7788 . "LATIN CAPITAL LETTER T WITH DOT BELOW")
    (7789 . "LATIN SMALL LETTER T WITH DOT BELOW")
    (7790 . "LATIN CAPITAL LETTER T WITH LINE BELOW")
    (7791 . "LATIN SMALL LETTER T WITH LINE BELOW")
    (7792 . "LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW")
    (7793 . "LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW")
    (7794 . "LATIN CAPITAL LETTER U WITH DIAERESIS BELOW")
    (7795 . "LATIN SMALL LETTER U WITH DIAERESIS BELOW")
    (7796 . "LATIN CAPITAL LETTER U WITH TILDE BELOW")
    (7797 . "LATIN SMALL LETTER U WITH TILDE BELOW")
    (7798 . "LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW")
    (7799 . "LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW")
    (7800 . "LATIN CAPITAL LETTER U WITH TILDE AND ACUTE")
    (7801 . "LATIN SMALL LETTER U WITH TILDE AND ACUTE")
    (7802 . "LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS")
    (7803 . "LATIN SMALL LETTER U WITH MACRON AND DIAERESIS")
    (7804 . "LATIN CAPITAL LETTER V WITH TILDE")
    (7805 . "LATIN SMALL LETTER V WITH TILDE")
    (7806 . "LATIN CAPITAL LETTER V WITH DOT BELOW")
    (7807 . "LATIN SMALL LETTER V WITH DOT BELOW")
    (7808 . "LATIN CAPITAL LETTER W WITH GRAVE")
    (7809 . "LATIN SMALL LETTER W WITH GRAVE")
    (7810 . "LATIN CAPITAL LETTER W WITH ACUTE")
    (7811 . "LATIN SMALL LETTER W WITH ACUTE")
    (7812 . "LATIN CAPITAL LETTER W WITH DIAERESIS")
    (7813 . "LATIN SMALL LETTER W WITH DIAERESIS")
    (7814 . "LATIN CAPITAL LETTER W WITH DOT ABOVE")
    (7815 . "LATIN SMALL LETTER W WITH DOT ABOVE")
    (7816 . "LATIN CAPITAL LETTER W WITH DOT BELOW")
    (7817 . "LATIN SMALL LETTER W WITH DOT BELOW")
    (7818 . "LATIN CAPITAL LETTER X WITH DOT ABOVE")
    (7819 . "LATIN SMALL LETTER X WITH DOT ABOVE")
    (7820 . "LATIN CAPITAL LETTER X WITH DIAERESIS")
    (7821 . "LATIN SMALL LETTER X WITH DIAERESIS")
    (7822 . "LATIN CAPITAL LETTER Y WITH DOT ABOVE")
    (7823 . "LATIN SMALL LETTER Y WITH DOT ABOVE")
    (7824 . "LATIN CAPITAL LETTER Z WITH CIRCUMFLEX")
    (7825 . "LATIN SMALL LETTER Z WITH CIRCUMFLEX")
    (7826 . "LATIN CAPITAL LETTER Z WITH DOT BELOW")
    (7827 . "LATIN SMALL LETTER Z WITH DOT BELOW")
    (7828 . "LATIN CAPITAL LETTER Z WITH LINE BELOW")
    (7829 . "LATIN SMALL LETTER Z WITH LINE BELOW")
    (7830 . "LATIN SMALL LETTER H WITH LINE BELOW")
    (7831 . "LATIN SMALL LETTER T WITH DIAERESIS")
    (7832 . "LATIN SMALL LETTER W WITH RING ABOVE")
    (7833 . "LATIN SMALL LETTER Y WITH RING ABOVE")
    (7834 . "LATIN SMALL LETTER A WITH RIGHT HALF RING")
    (7840 . "LATIN CAPITAL LETTER A WITH DOT BELOW")
    (7841 . "LATIN SMALL LETTER A WITH DOT BELOW")
    (7842 . "LATIN CAPITAL LETTER A WITH HOOK ABOVE")
    (7843 . "LATIN SMALL LETTER A WITH HOOK ABOVE")
    (7844 . "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE")
    (7845 . "LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE")
    (7846 . "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE")
    (7847 . "LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE")
    (7848 . "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE")
    (7849 . "LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE")
    (7850 . "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE")
    (7851 . "LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE")
    (7852 . "LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW")
    (7853 . "LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW")
    (7854 . "LATIN CAPITAL LETTER A WITH BREVE AND ACUTE")
    (7855 . "LATIN SMALL LETTER A WITH BREVE AND ACUTE")
    (7856 . "LATIN CAPITAL LETTER A WITH BREVE AND GRAVE")
    (7857 . "LATIN SMALL LETTER A WITH BREVE AND GRAVE")
    (7858 . "LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE")
    (7859 . "LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE")
    (7860 . "LATIN CAPITAL LETTER A WITH BREVE AND TILDE")
    (7861 . "LATIN SMALL LETTER A WITH BREVE AND TILDE")
    (7862 . "LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW")
    (7863 . "LATIN SMALL LETTER A WITH BREVE AND DOT BELOW")
    (7864 . "LATIN CAPITAL LETTER E WITH DOT BELOW")
    (7865 . "LATIN SMALL LETTER E WITH DOT BELOW")
    (7866 . "LATIN CAPITAL LETTER E WITH HOOK ABOVE")
    (7867 . "LATIN SMALL LETTER E WITH HOOK ABOVE")
    (7868 . "LATIN CAPITAL LETTER E WITH TILDE")
    (7869 . "LATIN SMALL LETTER E WITH TILDE")
    (7870 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE")
    (7871 . "LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE")
    (7872 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE")
    (7873 . "LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE")
    (7874 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE")
    (7875 . "LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE")
    (7876 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE")
    (7877 . "LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE")
    (7878 . "LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW")
    (7879 . "LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW")
    (7880 . "LATIN CAPITAL LETTER I WITH HOOK ABOVE")
    (7881 . "LATIN SMALL LETTER I WITH HOOK ABOVE")
    (7882 . "LATIN CAPITAL LETTER I WITH DOT BELOW")
    (7883 . "LATIN SMALL LETTER I WITH DOT BELOW")
    (7884 . "LATIN CAPITAL LETTER O WITH DOT BELOW")
    (7885 . "LATIN SMALL LETTER O WITH DOT BELOW")
    (7886 . "LATIN CAPITAL LETTER O WITH HOOK ABOVE")
    (7887 . "LATIN SMALL LETTER O WITH HOOK ABOVE")
    (7888 . "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE")
    (7889 . "LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE")
    (7890 . "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE")
    (7891 . "LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE")
    (7892 . "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE")
    (7893 . "LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE")
    (7894 . "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE")
    (7895 . "LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE")
    (7896 . "LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW")
    (7897 . "LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW")
    (7898 . "LATIN CAPITAL LETTER O WITH HORN AND ACUTE")
    (7899 . "LATIN SMALL LETTER O WITH HORN AND ACUTE")
    (7900 . "LATIN CAPITAL LETTER O WITH HORN AND GRAVE")
    (7901 . "LATIN SMALL LETTER O WITH HORN AND GRAVE")
    (7902 . "LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE")
    (7903 . "LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE")
    (7904 . "LATIN CAPITAL LETTER O WITH HORN AND TILDE")
    (7905 . "LATIN SMALL LETTER O WITH HORN AND TILDE")
    (7906 . "LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW")
    (7907 . "LATIN SMALL LETTER O WITH HORN AND DOT BELOW")
    (7908 . "LATIN CAPITAL LETTER U WITH DOT BELOW")
    (7909 . "LATIN SMALL LETTER U WITH DOT BELOW")
    (7910 . "LATIN CAPITAL LETTER U WITH HOOK ABOVE")
    (7911 . "LATIN SMALL LETTER U WITH HOOK ABOVE")
    (7912 . "LATIN CAPITAL LETTER U WITH HORN AND ACUTE")
    (7913 . "LATIN SMALL LETTER U WITH HORN AND ACUTE")
    (7914 . "LATIN CAPITAL LETTER U WITH HORN AND GRAVE")
    (7915 . "LATIN SMALL LETTER U WITH HORN AND GRAVE")
    (7916 . "LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE")
    (7917 . "LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE")
    (7918 . "LATIN CAPITAL LETTER U WITH HORN AND TILDE")
    (7919 . "LATIN SMALL LETTER U WITH HORN AND TILDE")
    (7920 . "LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW")
    (7921 . "LATIN SMALL LETTER U WITH HORN AND DOT BELOW")
    (7922 . "LATIN CAPITAL LETTER Y WITH GRAVE")
    (7923 . "LATIN SMALL LETTER Y WITH GRAVE")
    (7924 . "LATIN CAPITAL LETTER Y WITH DOT BELOW")
    (7925 . "LATIN SMALL LETTER Y WITH DOT BELOW")
    (7926 . "LATIN CAPITAL LETTER Y WITH HOOK ABOVE")
    (7927 . "LATIN SMALL LETTER Y WITH HOOK ABOVE")
    (7928 . "LATIN CAPITAL LETTER Y WITH TILDE")
    (7929 . "LATIN SMALL LETTER Y WITH TILDE")
    (7936 . "GREEK SMALL LETTER ALPHA WITH PSILI")
    (7937 . "GREEK SMALL LETTER ALPHA WITH DASIA")
    (7938 . "GREEK SMALL LETTER ALPHA WITH PSILI AND VARIA")
    (7939 . "GREEK SMALL LETTER ALPHA WITH DASIA AND VARIA")
    (7940 . "GREEK SMALL LETTER ALPHA WITH PSILI AND OXIA")
    (7941 . "GREEK SMALL LETTER ALPHA WITH DASIA AND OXIA")
    (7942 . "GREEK SMALL LETTER ALPHA WITH PSILI AND PERISPOMENI")
    (7943 . "GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI")
    (7944 . "GREEK CAPITAL LETTER ALPHA WITH PSILI")
    (7945 . "GREEK CAPITAL LETTER ALPHA WITH DASIA")
    (7946 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND VARIA")
    (7947 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND VARIA")
    (7948 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND OXIA")
    (7949 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND OXIA")
    (7950 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PERISPOMENI")
    (7951 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI")
    (7952 . "GREEK SMALL LETTER EPSILON WITH PSILI")
    (7953 . "GREEK SMALL LETTER EPSILON WITH DASIA")
    (7954 . "GREEK SMALL LETTER EPSILON WITH PSILI AND VARIA")
    (7955 . "GREEK SMALL LETTER EPSILON WITH DASIA AND VARIA")
    (7956 . "GREEK SMALL LETTER EPSILON WITH PSILI AND OXIA")
    (7957 . "GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA")
    (7960 . "GREEK CAPITAL LETTER EPSILON WITH PSILI")
    (7961 . "GREEK CAPITAL LETTER EPSILON WITH DASIA")
    (7962 . "GREEK CAPITAL LETTER EPSILON WITH PSILI AND VARIA")
    (7963 . "GREEK CAPITAL LETTER EPSILON WITH DASIA AND VARIA")
    (7964 . "GREEK CAPITAL LETTER EPSILON WITH PSILI AND OXIA")
    (7965 . "GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA")
    (7968 . "GREEK SMALL LETTER ETA WITH PSILI")
    (7969 . "GREEK SMALL LETTER ETA WITH DASIA")
    (7970 . "GREEK SMALL LETTER ETA WITH PSILI AND VARIA")
    (7971 . "GREEK SMALL LETTER ETA WITH DASIA AND VARIA")
    (7972 . "GREEK SMALL LETTER ETA WITH PSILI AND OXIA")
    (7973 . "GREEK SMALL LETTER ETA WITH DASIA AND OXIA")
    (7974 . "GREEK SMALL LETTER ETA WITH PSILI AND PERISPOMENI")
    (7975 . "GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI")
    (7976 . "GREEK CAPITAL LETTER ETA WITH PSILI")
    (7977 . "GREEK CAPITAL LETTER ETA WITH DASIA")
    (7978 . "GREEK CAPITAL LETTER ETA WITH PSILI AND VARIA")
    (7979 . "GREEK CAPITAL LETTER ETA WITH DASIA AND VARIA")
    (7980 . "GREEK CAPITAL LETTER ETA WITH PSILI AND OXIA")
    (7981 . "GREEK CAPITAL LETTER ETA WITH DASIA AND OXIA")
    (7982 . "GREEK CAPITAL LETTER ETA WITH PSILI AND PERISPOMENI")
    (7983 . "GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI")
    (7984 . "GREEK SMALL LETTER IOTA WITH PSILI")
    (7985 . "GREEK SMALL LETTER IOTA WITH DASIA")
    (7986 . "GREEK SMALL LETTER IOTA WITH PSILI AND VARIA")
    (7987 . "GREEK SMALL LETTER IOTA WITH DASIA AND VARIA")
    (7988 . "GREEK SMALL LETTER IOTA WITH PSILI AND OXIA")
    (7989 . "GREEK SMALL LETTER IOTA WITH DASIA AND OXIA")
    (7990 . "GREEK SMALL LETTER IOTA WITH PSILI AND PERISPOMENI")
    (7991 . "GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI")
    (7992 . "GREEK CAPITAL LETTER IOTA WITH PSILI")
    (7993 . "GREEK CAPITAL LETTER IOTA WITH DASIA")
    (7994 . "GREEK CAPITAL LETTER IOTA WITH PSILI AND VARIA")
    (7995 . "GREEK CAPITAL LETTER IOTA WITH DASIA AND VARIA")
    (7996 . "GREEK CAPITAL LETTER IOTA WITH PSILI AND OXIA")
    (7997 . "GREEK CAPITAL LETTER IOTA WITH DASIA AND OXIA")
    (7998 . "GREEK CAPITAL LETTER IOTA WITH PSILI AND PERISPOMENI")
    (7999 . "GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI")
    (8000 . "GREEK SMALL LETTER OMICRON WITH PSILI")
    (8001 . "GREEK SMALL LETTER OMICRON WITH DASIA")
    (8002 . "GREEK SMALL LETTER OMICRON WITH PSILI AND VARIA")
    (8003 . "GREEK SMALL LETTER OMICRON WITH DASIA AND VARIA")
    (8004 . "GREEK SMALL LETTER OMICRON WITH PSILI AND OXIA")
    (8005 . "GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA")
    (8008 . "GREEK CAPITAL LETTER OMICRON WITH PSILI")
    (8009 . "GREEK CAPITAL LETTER OMICRON WITH DASIA")
    (8010 . "GREEK CAPITAL LETTER OMICRON WITH PSILI AND VARIA")
    (8011 . "GREEK CAPITAL LETTER OMICRON WITH DASIA AND VARIA")
    (8012 . "GREEK CAPITAL LETTER OMICRON WITH PSILI AND OXIA")
    (8013 . "GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA")
    (8016 . "GREEK SMALL LETTER UPSILON WITH PSILI")
    (8017 . "GREEK SMALL LETTER UPSILON WITH DASIA")
    (8018 . "GREEK SMALL LETTER UPSILON WITH PSILI AND VARIA")
    (8019 . "GREEK SMALL LETTER UPSILON WITH DASIA AND VARIA")
    (8020 . "GREEK SMALL LETTER UPSILON WITH PSILI AND OXIA")
    (8021 . "GREEK SMALL LETTER UPSILON WITH DASIA AND OXIA")
    (8022 . "GREEK SMALL LETTER UPSILON WITH PSILI AND PERISPOMENI")
    (8023 . "GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI")
    (8025 . "GREEK CAPITAL LETTER UPSILON WITH DASIA")
    (8027 . "GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA")
    (8029 . "GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA")
    (8031 . "GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI")
    (8032 . "GREEK SMALL LETTER OMEGA WITH PSILI")
    (8033 . "GREEK SMALL LETTER OMEGA WITH DASIA")
    (8034 . "GREEK SMALL LETTER OMEGA WITH PSILI AND VARIA")
    (8035 . "GREEK SMALL LETTER OMEGA WITH DASIA AND VARIA")
    (8036 . "GREEK SMALL LETTER OMEGA WITH PSILI AND OXIA")
    (8037 . "GREEK SMALL LETTER OMEGA WITH DASIA AND OXIA")
    (8038 . "GREEK SMALL LETTER OMEGA WITH PSILI AND PERISPOMENI")
    (8039 . "GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI")
    (8040 . "GREEK CAPITAL LETTER OMEGA WITH PSILI")
    (8041 . "GREEK CAPITAL LETTER OMEGA WITH DASIA")
    (8042 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND VARIA")
    (8043 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND VARIA")
    (8044 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND OXIA")
    (8045 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND OXIA")
    (8046 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PERISPOMENI")
    (8047 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI")
    (8048 . "GREEK SMALL LETTER ALPHA WITH VARIA")
    (8049 . "GREEK SMALL LETTER ALPHA WITH OXIA")
    (8050 . "GREEK SMALL LETTER EPSILON WITH VARIA")
    (8051 . "GREEK SMALL LETTER EPSILON WITH OXIA")
    (8052 . "GREEK SMALL LETTER ETA WITH VARIA")
    (8053 . "GREEK SMALL LETTER ETA WITH OXIA")
    (8054 . "GREEK SMALL LETTER IOTA WITH VARIA")
    (8055 . "GREEK SMALL LETTER IOTA WITH OXIA")
    (8056 . "GREEK SMALL LETTER OMICRON WITH VARIA")
    (8057 . "GREEK SMALL LETTER OMICRON WITH OXIA")
    (8058 . "GREEK SMALL LETTER UPSILON WITH VARIA")
    (8059 . "GREEK SMALL LETTER UPSILON WITH OXIA")
    (8060 . "GREEK SMALL LETTER OMEGA WITH VARIA")
    (8061 . "GREEK SMALL LETTER OMEGA WITH OXIA")
    (8064 . "GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI")
    (8065 . "GREEK SMALL LETTER ALPHA WITH DASIA AND YPOGEGRAMMENI")
    (8066 . "GREEK SMALL LETTER ALPHA WITH PSILI AND VARIA AND YPOGEGRAMMENI")
    (8067 . "GREEK SMALL LETTER ALPHA WITH DASIA AND VARIA AND YPOGEGRAMMENI")
    (8068 . "GREEK SMALL LETTER ALPHA WITH PSILI AND OXIA AND YPOGEGRAMMENI")
    (8069 . "GREEK SMALL LETTER ALPHA WITH DASIA AND OXIA AND YPOGEGRAMMENI")
    (8070 . "GREEK SMALL LETTER ALPHA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI")
    (8071 . "GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI")
    (8072 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PROSGEGRAMMENI")
    (8073 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PROSGEGRAMMENI")
    (8074 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND VARIA AND PROSGEGRAMMENI")
    (8075 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND VARIA AND PROSGEGRAMMENI")
    (8076 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND OXIA AND PROSGEGRAMMENI")
    (8077 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND OXIA AND PROSGEGRAMMENI")
    (8078 . "GREEK CAPITAL LETTER ALPHA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI")
    (8079 . "GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI")
    (8080 . "GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI")
    (8081 . "GREEK SMALL LETTER ETA WITH DASIA AND YPOGEGRAMMENI")
    (8082 . "GREEK SMALL LETTER ETA WITH PSILI AND VARIA AND YPOGEGRAMMENI")
    (8083 . "GREEK SMALL LETTER ETA WITH DASIA AND VARIA AND YPOGEGRAMMENI")
    (8084 . "GREEK SMALL LETTER ETA WITH PSILI AND OXIA AND YPOGEGRAMMENI")
    (8085 . "GREEK SMALL LETTER ETA WITH DASIA AND OXIA AND YPOGEGRAMMENI")
    (8086 . "GREEK SMALL LETTER ETA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI")
    (8087 . "GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI")
    (8088 . "GREEK CAPITAL LETTER ETA WITH PSILI AND PROSGEGRAMMENI")
    (8089 . "GREEK CAPITAL LETTER ETA WITH DASIA AND PROSGEGRAMMENI")
    (8090 . "GREEK CAPITAL LETTER ETA WITH PSILI AND VARIA AND PROSGEGRAMMENI")
    (8091 . "GREEK CAPITAL LETTER ETA WITH DASIA AND VARIA AND PROSGEGRAMMENI")
    (8092 . "GREEK CAPITAL LETTER ETA WITH PSILI AND OXIA AND PROSGEGRAMMENI")
    (8093 . "GREEK CAPITAL LETTER ETA WITH DASIA AND OXIA AND PROSGEGRAMMENI")
    (8094 . "GREEK CAPITAL LETTER ETA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI")
    (8095 . "GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI")
    (8096 . "GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI")
    (8097 . "GREEK SMALL LETTER OMEGA WITH DASIA AND YPOGEGRAMMENI")
    (8098 . "GREEK SMALL LETTER OMEGA WITH PSILI AND VARIA AND YPOGEGRAMMENI")
    (8099 . "GREEK SMALL LETTER OMEGA WITH DASIA AND VARIA AND YPOGEGRAMMENI")
    (8100 . "GREEK SMALL LETTER OMEGA WITH PSILI AND OXIA AND YPOGEGRAMMENI")
    (8101 . "GREEK SMALL LETTER OMEGA WITH DASIA AND OXIA AND YPOGEGRAMMENI")
    (8102 . "GREEK SMALL LETTER OMEGA WITH PSILI AND PERISPOMENI AND YPOGEGRAMMENI")
    (8103 . "GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI")
    (8104 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PROSGEGRAMMENI")
    (8105 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PROSGEGRAMMENI")
    (8106 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND VARIA AND PROSGEGRAMMENI")
    (8107 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND VARIA AND PROSGEGRAMMENI")
    (8108 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND OXIA AND PROSGEGRAMMENI")
    (8109 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND OXIA AND PROSGEGRAMMENI")
    (8110 . "GREEK CAPITAL LETTER OMEGA WITH PSILI AND PERISPOMENI AND PROSGEGRAMMENI")
    (8111 . "GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI")
    (8112 . "GREEK SMALL LETTER ALPHA WITH VRACHY")
    (8113 . "GREEK SMALL LETTER ALPHA WITH MACRON")
    (8114 . "GREEK SMALL LETTER ALPHA WITH VARIA AND YPOGEGRAMMENI")
    (8115 . "GREEK SMALL LETTER ALPHA WITH YPOGEGRAMMENI")
    (8116 . "GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI")
    (8118 . "GREEK SMALL LETTER ALPHA WITH PERISPOMENI")
    (8119 . "GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI")
    (8120 . "GREEK CAPITAL LETTER ALPHA WITH VRACHY")
    (8121 . "GREEK CAPITAL LETTER ALPHA WITH MACRON")
    (8122 . "GREEK CAPITAL LETTER ALPHA WITH VARIA")
    (8123 . "GREEK CAPITAL LETTER ALPHA WITH OXIA")
    (8124 . "GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI")
    (8125 . "GREEK KORONIS")
    (8126 . "GREEK PROSGEGRAMMENI")
    (8127 . "GREEK PSILI")
    (8128 . "GREEK PERISPOMENI")
    (8129 . "GREEK DIALYTIKA AND PERISPOMENI")
    (8130 . "GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI")
    (8131 . "GREEK SMALL LETTER ETA WITH YPOGEGRAMMENI")
    (8132 . "GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI")
    (8134 . "GREEK SMALL LETTER ETA WITH PERISPOMENI")
    (8135 . "GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI")
    (8136 . "GREEK CAPITAL LETTER EPSILON WITH VARIA")
    (8137 . "GREEK CAPITAL LETTER EPSILON WITH OXIA")
    (8138 . "GREEK CAPITAL LETTER ETA WITH VARIA")
    (8139 . "GREEK CAPITAL LETTER ETA WITH OXIA")
    (8140 . "GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI")
    (8141 . "GREEK PSILI AND VARIA")
    (8142 . "GREEK PSILI AND OXIA")
    (8143 . "GREEK PSILI AND PERISPOMENI")
    (8144 . "GREEK SMALL LETTER IOTA WITH VRACHY")
    (8145 . "GREEK SMALL LETTER IOTA WITH MACRON")
    (8146 . "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND VARIA")
    (8147 . "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA")
    (8150 . "GREEK SMALL LETTER IOTA WITH PERISPOMENI")
    (8151 . "GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI")
    (8152 . "GREEK CAPITAL LETTER IOTA WITH VRACHY")
    (8153 . "GREEK CAPITAL LETTER IOTA WITH MACRON")
    (8154 . "GREEK CAPITAL LETTER IOTA WITH VARIA")
    (8155 . "GREEK CAPITAL LETTER IOTA WITH OXIA")
    (8157 . "GREEK DASIA AND VARIA")
    (8158 . "GREEK DASIA AND OXIA")
    (8159 . "GREEK DASIA AND PERISPOMENI")
    (8160 . "GREEK SMALL LETTER UPSILON WITH VRACHY")
    (8161 . "GREEK SMALL LETTER UPSILON WITH MACRON")
    (8162 . "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND VARIA")
    (8163 . "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND OXIA")
    (8164 . "GREEK SMALL LETTER RHO WITH PSILI")
    (8165 . "GREEK SMALL LETTER RHO WITH DASIA")
    (8166 . "GREEK SMALL LETTER UPSILON WITH PERISPOMENI")
    (8167 . "GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI")
    (8168 . "GREEK CAPITAL LETTER UPSILON WITH VRACHY")
    (8169 . "GREEK CAPITAL LETTER UPSILON WITH MACRON")
    (8170 . "GREEK CAPITAL LETTER UPSILON WITH VARIA")
    (8171 . "GREEK CAPITAL LETTER UPSILON WITH OXIA")
    (8172 . "GREEK CAPITAL LETTER RHO WITH DASIA")
    (8173 . "GREEK DIALYTIKA AND VARIA")
    (8174 . "GREEK DIALYTIKA AND OXIA")
    (8175 . "GREEK VARIA")
    (8178 . "GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI")
    (8179 . "GREEK SMALL LETTER OMEGA WITH YPOGEGRAMMENI")
    (8180 . "GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI")
    (8182 . "GREEK SMALL LETTER OMEGA WITH PERISPOMENI")
    (8183 . "GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI")
    (8184 . "GREEK CAPITAL LETTER OMICRON WITH VARIA")
    (8185 . "GREEK CAPITAL LETTER OMICRON WITH OXIA")
    (8186 . "GREEK CAPITAL LETTER OMEGA WITH VARIA")
    (8187 . "GREEK CAPITAL LETTER OMEGA WITH OXIA")
    (8188 . "GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI")
    (8189 . "GREEK OXIA")
    (8190 . "GREEK DASIA")
    (8192 . "EN QUAD")
    (8193 . "EM QUAD")
    (8194 . "EN SPACE")
    (8195 . "EM SPACE")
    (8196 . "THREE-PER-EM SPACE")
    (8197 . "FOUR-PER-EM SPACE")
    (8198 . "SIX-PER-EM SPACE")
    (8199 . "FIGURE SPACE")
    (8200 . "PUNCTUATION SPACE")
    (8201 . "THIN SPACE")
    (8202 . "HAIR SPACE")
    (8203 . "ZERO WIDTH SPACE")
    (8204 . "ZERO WIDTH NON-JOINER")
    (8205 . "ZERO WIDTH JOINER")
    (8206 . "LEFT-TO-RIGHT MARK")
    (8207 . "RIGHT-TO-LEFT MARK")
    (8208 . "HYPHEN")
    (8209 . "NON-BREAKING HYPHEN")
    (8210 . "FIGURE DASH")
    (8211 . "EN DASH")
    (8212 . "EM DASH")
    (8213 . "HORIZONTAL BAR")
    (8214 . "DOUBLE VERTICAL LINE")
    (8215 . "DOUBLE LOW LINE")
    (8216 . "LEFT SINGLE QUOTATION MARK")
    (8217 . "RIGHT SINGLE QUOTATION MARK")
    (8218 . "SINGLE LOW-NINE QUOTATION MARK")
    (8219 . "SINGLE HIGH-REVERSED-NINE QUOTATION MARK")
    (8220 . "LEFT DOUBLE QUOTATION MARK")
    (8221 . "RIGHT DOUBLE QUOTATION MARK")
    (8222 . "DOUBLE LOW-NINE QUOTATION MARK")
    (8223 . "DOUBLE HIGH-REVERSED-NINE QUOTATION MARK")
    (8224 . "DAGGER")
    (8225 . "DOUBLE DAGGER")
    (8226 . "BULLET")
    (8227 . "TRIANGULAR BULLET")
    (8228 . "ONE DOT LEADER")
    (8229 . "TWO DOT LEADER")
    (8230 . "HORIZONTAL ELLIPSIS")
    (8231 . "HYPHENATION POINT")
    (8232 . "LINE SEPARATOR")
    (8233 . "PARAGRAPH SEPARATOR")
    (8234 . "LEFT-TO-RIGHT EMBEDDING")
    (8235 . "RIGHT-TO-LEFT EMBEDDING")
    (8236 . "POP DIRECTIONAL FORMATTING")
    (8237 . "LEFT-TO-RIGHT OVERRIDE")
    (8238 . "RIGHT-TO-LEFT OVERRIDE")
    (8240 . "PER MILLE SIGN")
    (8241 . "PER TEN THOUSAND SIGN")
    (8242 . "PRIME")
    (8243 . "DOUBLE PRIME")
    (8244 . "TRIPLE PRIME")
    (8245 . "REVERSED PRIME")
    (8246 . "REVERSED DOUBLE PRIME")
    (8247 . "REVERSED TRIPLE PRIME")
    (8248 . "CARET")
    (8249 . "SINGLE LEFT-POINTING ANGLE QUOTATION MARK")
    (8250 . "SINGLE RIGHT-POINTING ANGLE QUOTATION MARK")
    (8251 . "REFERENCE MARK")
    (8252 . "DOUBLE EXCLAMATION MARK")
    (8253 . "INTERROBANG")
    (8254 . "OVERLINE")
    (8255 . "UNDERTIE")
    (8256 . "CHARACTER TIE")
    (8257 . "CARET INSERTION POINT")
    (8258 . "ASTERISM")
    (8259 . "HYPHEN BULLET")
    (8260 . "FRACTION SLASH")
    (8261 . "LEFT SQUARE BRACKET WITH QUILL")
    (8262 . "RIGHT SQUARE BRACKET WITH QUILL")
    (8298 . "INHIBIT SYMMETRIC SWAPPING")
    (8299 . "ACTIVATE SYMMETRIC SWAPPING")
    (8300 . "INHIBIT ARABIC FORM SHAPING")
    (8301 . "ACTIVATE ARABIC FORM SHAPING")
    (8302 . "NATIONAL DIGIT SHAPES")
    (8303 . "NOMINAL DIGIT SHAPES")
    (8304 . "SUPERSCRIPT ZERO")
    (8308 . "SUPERSCRIPT FOUR")
    (8309 . "SUPERSCRIPT FIVE")
    (8310 . "SUPERSCRIPT SIX")
    (8311 . "SUPERSCRIPT SEVEN")
    (8312 . "SUPERSCRIPT EIGHT")
    (8313 . "SUPERSCRIPT NINE")
    (8314 . "SUPERSCRIPT PLUS SIGN")
    (8315 . "SUPERSCRIPT MINUS")
    (8316 . "SUPERSCRIPT EQUALS SIGN")
    (8317 . "SUPERSCRIPT LEFT PARENTHESIS")
    (8318 . "SUPERSCRIPT RIGHT PARENTHESIS")
    (8319 . "SUPERSCRIPT LATIN SMALL LETTER N")
    (8320 . "SUBSCRIPT ZERO")
    (8321 . "SUBSCRIPT ONE")
    (8322 . "SUBSCRIPT TWO")
    (8323 . "SUBSCRIPT THREE")
    (8324 . "SUBSCRIPT FOUR")
    (8325 . "SUBSCRIPT FIVE")
    (8326 . "SUBSCRIPT SIX")
    (8327 . "SUBSCRIPT SEVEN")
    (8328 . "SUBSCRIPT EIGHT")
    (8329 . "SUBSCRIPT NINE")
    (8330 . "SUBSCRIPT PLUS SIGN")
    (8331 . "SUBSCRIPT MINUS")
    (8332 . "SUBSCRIPT EQUALS SIGN")
    (8333 . "SUBSCRIPT LEFT PARENTHESIS")
    (8334 . "SUBSCRIPT RIGHT PARENTHESIS")
    (8352 . "EURO-CURRENCY SIGN")
    (8353 . "COLON SIGN")
    (8354 . "CRUZEIRO SIGN")
    (8355 . "FRENCH FRANC SIGN")
    (8356 . "LIRA SIGN")
    (8357 . "MILL SIGN")
    (8358 . "NAIRA SIGN")
    (8359 . "PESETA SIGN")
    (8360 . "RUPEE SIGN")
    (8361 . "WON SIGN")
    (8362 . "NEW SHEQEL SIGN")
    (8400 . "COMBINING LEFT HARPOON ABOVE")
    (8401 . "COMBINING RIGHT HARPOON ABOVE")
    (8402 . "COMBINING LONG VERTICAL LINE OVERLAY")
    (8403 . "COMBINING SHORT VERTICAL LINE OVERLAY")
    (8404 . "COMBINING ANTICLOCKWISE ARROW ABOVE")
    (8405 . "COMBINING CLOCKWISE ARROW ABOVE")
    (8406 . "COMBINING LEFT ARROW ABOVE")
    (8407 . "COMBINING RIGHT ARROW ABOVE")
    (8408 . "COMBINING RING OVERLAY")
    (8409 . "COMBINING CLOCKWISE RING OVERLAY")
    (8410 . "COMBINING ANTICLOCKWISE RING OVERLAY")
    (8411 . "COMBINING THREE DOTS ABOVE")
    (8412 . "COMBINING FOUR DOTS ABOVE")
    (8413 . "COMBINING ENCLOSING CIRCLE")
    (8414 . "COMBINING ENCLOSING SQUARE")
    (8415 . "COMBINING ENCLOSING DIAMOND")
    (8416 . "COMBINING ENCLOSING CIRCLE BACKSLASH")
    (8417 . "COMBINING LEFT RIGHT ARROW ABOVE")
    (8448 . "ACCOUNT OF")
    (8449 . "ADDRESSED TO THE SUBJECT")
    (8450 . "DOUBLE-STRUCK CAPITAL C")
    (8451 . "DEGREE CELSIUS")
    (8452 . "CENTER LINE SYMBOL|CENTRE LINE SYMBOL")
    (8453 . "CARE OF")
    (8454 . "CADA UNA")
    (8455 . "EULER CONSTANT")
    (8456 . "SCRUPLE")
    (8457 . "DEGREE FAHRENHEIT")
    (8458 . "SCRIPT SMALL G")
    (8459 . "SCRIPT CAPITAL H")
    (8460 . "BLACK-LETTER CAPITAL H")
    (8461 . "DOUBLE-STRUCK CAPITAL H")
    (8462 . "PLANCK CONSTANT")
    (8463 . "PLANCK CONSTANT OVER TWO PI")
    (8464 . "SCRIPT CAPITAL I")
    (8465 . "BLACK-LETTER CAPITAL I")
    (8466 . "SCRIPT CAPITAL L")
    (8467 . "SCRIPT SMALL L")
    (8468 . "L B BAR SYMBOL")
    (8469 . "DOUBLE-STRUCK CAPITAL N")
    (8470 . "NUMERO SIGN")
    (8471 . "SOUND RECORDING COPYRIGHT")
    (8472 . "SCRIPT CAPITAL P")
    (8473 . "DOUBLE-STRUCK CAPITAL P")
    (8474 . "DOUBLE-STRUCK CAPITAL Q")
    (8475 . "SCRIPT CAPITAL R")
    (8476 . "BLACK-LETTER CAPITAL R")
    (8477 . "DOUBLE-STRUCK CAPITAL R")
    (8478 . "PRESCRIPTION TAKE")
    (8479 . "RESPONSE")
    (8480 . "SERVICE MARK")
    (8481 . "TELEPHONE SIGN")
    (8482 . "TRADE MARK SIGN")
    (8483 . "VERSICLE")
    (8484 . "DOUBLE-STRUCK CAPITAL Z")
    (8485 . "OUNCE SIGN")
    (8486 . "OHM SIGN")
    (8487 . "INVERTED OHM SIGN")
    (8488 . "BLACK-LETTER CAPITAL Z")
    (8489 . "TURNED GREEK SMALL LETTER IOTA")
    (8490 . "KELVIN SIGN")
    (8491 . "ANGSTROM SIGN")
    (8492 . "SCRIPT CAPITAL B")
    (8493 . "BLACK-LETTER CAPITAL C")
    (8494 . "ESTIMATED SYMBOL")
    (8495 . "SCRIPT SMALL E")
    (8496 . "SCRIPT CAPITAL E")
    (8497 . "SCRIPT CAPITAL F")
    (8498 . "TURNED CAPITAL F")
    (8499 . "SCRIPT CAPITAL M")
    (8500 . "SCRIPT SMALL O")
    (8501 . "ALEF SYMBOL")
    (8502 . "BET SYMBOL")
    (8503 . "GIMEL SYMBOL")
    (8504 . "DALET SYMBOL")
    (8531 . "VULGAR FRACTION ONE THIRD")
    (8532 . "VULGAR FRACTION TWO THIRDS")
    (8533 . "VULGAR FRACTION ONE FIFTH")
    (8534 . "VULGAR FRACTION TWO FIFTHS")
    (8535 . "VULGAR FRACTION THREE FIFTHS")
    (8536 . "VULGAR FRACTION FOUR FIFTHS")
    (8537 . "VULGAR FRACTION ONE SIXTH")
    (8538 . "VULGAR FRACTION FIVE SIXTHS")
    (8539 . "VULGAR FRACTION ONE EIGHTH")
    (8540 . "VULGAR FRACTION THREE EIGHTHS")
    (8541 . "VULGAR FRACTION FIVE EIGHTHS")
    (8542 . "VULGAR FRACTION SEVEN EIGHTHS")
    (8543 . "FRACTION NUMERATOR ONE")
    (8544 . "ROMAN NUMERAL ONE")
    (8545 . "ROMAN NUMERAL TWO")
    (8546 . "ROMAN NUMERAL THREE")
    (8547 . "ROMAN NUMERAL FOUR")
    (8548 . "ROMAN NUMERAL FIVE")
    (8549 . "ROMAN NUMERAL SIX")
    (8550 . "ROMAN NUMERAL SEVEN")
    (8551 . "ROMAN NUMERAL EIGHT")
    (8552 . "ROMAN NUMERAL NINE")
    (8553 . "ROMAN NUMERAL TEN")
    (8554 . "ROMAN NUMERAL ELEVEN")
    (8555 . "ROMAN NUMERAL TWELVE")
    (8556 . "ROMAN NUMERAL FIFTY")
    (8557 . "ROMAN NUMERAL ONE HUNDRED")
    (8558 . "ROMAN NUMERAL FIVE HUNDRED")
    (8559 . "ROMAN NUMERAL ONE THOUSAND")
    (8560 . "SMALL ROMAN NUMERAL ONE")
    (8561 . "SMALL ROMAN NUMERAL TWO")
    (8562 . "SMALL ROMAN NUMERAL THREE")
    (8563 . "SMALL ROMAN NUMERAL FOUR")
    (8564 . "SMALL ROMAN NUMERAL FIVE")
    (8565 . "SMALL ROMAN NUMERAL SIX")
    (8566 . "SMALL ROMAN NUMERAL SEVEN")
    (8567 . "SMALL ROMAN NUMERAL EIGHT")
    (8568 . "SMALL ROMAN NUMERAL NINE")
    (8569 . "SMALL ROMAN NUMERAL TEN")
    (8570 . "SMALL ROMAN NUMERAL ELEVEN")
    (8571 . "SMALL ROMAN NUMERAL TWELVE")
    (8572 . "SMALL ROMAN NUMERAL FIFTY")
    (8573 . "SMALL ROMAN NUMERAL ONE HUNDRED")
    (8574 . "SMALL ROMAN NUMERAL FIVE HUNDRED")
    (8575 . "SMALL ROMAN NUMERAL ONE THOUSAND")
    (8576 . "ROMAN NUMERAL ONE THOUSAND C D")
    (8577 . "ROMAN NUMERAL FIVE THOUSAND")
    (8578 . "ROMAN NUMERAL TEN THOUSAND")
    (8592 . "LEFTWARDS ARROW")
    (8593 . "UPWARDS ARROW")
    (8594 . "RIGHTWARDS ARROW")
    (8595 . "DOWNWARDS ARROW")
    (8596 . "LEFT RIGHT ARROW")
    (8597 . "UP DOWN ARROW")
    (8598 . "NORTH WEST ARROW")
    (8599 . "NORTH EAST ARROW")
    (8600 . "SOUTH EAST ARROW")
    (8601 . "SOUTH WEST ARROW")
    (8602 . "LEFTWARDS ARROW WITH STROKE")
    (8603 . "RIGHTWARDS ARROW WITH STROKE")
    (8604 . "LEFTWARDS WAVE ARROW")
    (8605 . "RIGHTWARDS WAVE ARROW")
    (8606 . "LEFTWARDS TWO HEADED ARROW")
    (8607 . "UPWARDS TWO HEADED ARROW")
    (8608 . "RIGHTWARDS TWO HEADED ARROW")
    (8609 . "DOWNWARDS TWO HEADED ARROW")
    (8610 . "LEFTWARDS ARROW WITH TAIL")
    (8611 . "RIGHTWARDS ARROW WITH TAIL")
    (8612 . "LEFTWARDS ARROW FROM BAR")
    (8613 . "UPWARDS ARROW FROM BAR")
    (8614 . "RIGHTWARDS ARROW FROM BAR")
    (8615 . "DOWNWARDS ARROW FROM BAR")
    (8616 . "UP DOWN ARROW WITH BASE")
    (8617 . "LEFTWARDS ARROW WITH HOOK")
    (8618 . "RIGHTWARDS ARROW WITH HOOK")
    (8619 . "LEFTWARDS ARROW WITH LOOP")
    (8620 . "RIGHTWARDS ARROW WITH LOOP")
    (8621 . "LEFT RIGHT WAVE ARROW")
    (8622 . "LEFT RIGHT ARROW WITH STROKE")
    (8623 . "DOWNWARDS ZIGZAG ARROW")
    (8624 . "UPWARDS ARROW WITH TIP LEFTWARDS")
    (8625 . "UPWARDS ARROW WITH TIP RIGHTWARDS")
    (8626 . "DOWNWARDS ARROW WITH TIP LEFTWARDS")
    (8627 . "DOWNWARDS ARROW WITH TIP RIGHTWARDS")
    (8628 . "RIGHTWARDS ARROW WITH CORNER DOWNWARDS")
    (8629 . "DOWNWARDS ARROW WITH CORNER LEFTWARDS")
    (8630 . "ANTICLOCKWISE TOP SEMICIRCLE ARROW")
    (8631 . "CLOCKWISE TOP SEMICIRCLE ARROW")
    (8632 . "NORTH WEST ARROW TO LONG BAR")
    (8633 . "LEFTWARDS ARROW TO BAR OVER RIGHTWARDS ARROW TO BAR")
    (8634 . "ANTICLOCKWISE OPEN CIRCLE ARROW")
    (8635 . "CLOCKWISE OPEN CIRCLE ARROW")
    (8636 . "LEFTWARDS HARPOON WITH BARB UPWARDS")
    (8637 . "LEFTWARDS HARPOON WITH BARB DOWNWARDS")
    (8638 . "UPWARDS HARPOON WITH BARB RIGHTWARDS")
    (8639 . "UPWARDS HARPOON WITH BARB LEFTWARDS")
    (8640 . "RIGHTWARDS HARPOON WITH BARB UPWARDS")
    (8641 . "RIGHTWARDS HARPOON WITH BARB DOWNWARDS")
    (8642 . "DOWNWARDS HARPOON WITH BARB RIGHTWARDS")
    (8643 . "DOWNWARDS HARPOON WITH BARB LEFTWARDS")
    (8644 . "RIGHTWARDS ARROW OVER LEFTWARDS ARROW")
    (8645 . "UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW")
    (8646 . "LEFTWARDS ARROW OVER RIGHTWARDS ARROW")
    (8647 . "LEFTWARDS PAIRED ARROWS")
    (8648 . "UPWARDS PAIRED ARROWS")
    (8649 . "RIGHTWARDS PAIRED ARROWS")
    (8650 . "DOWNWARDS PAIRED ARROWS")
    (8651 . "LEFTWARDS HARPOON OVER RIGHTWARDS HARPOON")
    (8652 . "RIGHTWARDS HARPOON OVER LEFTWARDS HARPOON")
    (8653 . "LEFTWARDS DOUBLE ARROW WITH STROKE")
    (8654 . "LEFT RIGHT DOUBLE ARROW WITH STROKE")
    (8655 . "RIGHTWARDS DOUBLE ARROW WITH STROKE")
    (8656 . "LEFTWARDS DOUBLE ARROW")
    (8657 . "UPWARDS DOUBLE ARROW")
    (8658 . "RIGHTWARDS DOUBLE ARROW")
    (8659 . "DOWNWARDS DOUBLE ARROW")
    (8660 . "LEFT RIGHT DOUBLE ARROW")
    (8661 . "UP DOWN DOUBLE ARROW")
    (8662 . "NORTH WEST DOUBLE ARROW")
    (8663 . "NORTH EAST DOUBLE ARROW")
    (8664 . "SOUTH EAST DOUBLE ARROW")
    (8665 . "SOUTH WEST DOUBLE ARROW")
    (8666 . "LEFTWARDS TRIPLE ARROW")
    (8667 . "RIGHTWARDS TRIPLE ARROW")
    (8668 . "LEFTWARDS SQUIGGLE ARROW")
    (8669 . "RIGHTWARDS SQUIGGLE ARROW")
    (8670 . "UPWARDS ARROW WITH DOUBLE STROKE")
    (8671 . "DOWNWARDS ARROW WITH DOUBLE STROKE")
    (8672 . "LEFTWARDS DASHED ARROW")
    (8673 . "UPWARDS DASHED ARROW")
    (8674 . "RIGHTWARDS DASHED ARROW")
    (8675 . "DOWNWARDS DASHED ARROW")
    (8676 . "LEFTWARDS ARROW TO BAR")
    (8677 . "RIGHTWARDS ARROW TO BAR")
    (8678 . "LEFTWARDS WHITE ARROW")
    (8679 . "UPWARDS WHITE ARROW")
    (8680 . "RIGHTWARDS WHITE ARROW")
    (8681 . "DOWNWARDS WHITE ARROW")
    (8682 . "UPWARDS WHITE ARROW FROM BAR")
    (8704 . "FOR ALL")
    (8705 . "COMPLEMENT")
    (8706 . "PARTIAL DIFFERENTIAL")
    (8707 . "THERE EXISTS")
    (8708 . "THERE DOES NOT EXIST")
    (8709 . "EMPTY SET")
    (8710 . "INCREMENT")
    (8711 . "NABLA")
    (8712 . "ELEMENT OF")
    (8713 . "NOT AN ELEMENT OF")
    (8714 . "SMALL ELEMENT OF")
    (8715 . "CONTAINS AS MEMBER")
    (8716 . "DOES NOT CONTAIN AS MEMBER")
    (8717 . "SMALL CONTAINS AS MEMBER")
    (8718 . "END OF PROOF")
    (8719 . "N-ARY PRODUCT")
    (8720 . "N-ARY COPRODUCT")
    (8721 . "N-ARY SUMMATION")
    (8722 . "MINUS SIGN")
    (8723 . "MINUS-OR-PLUS SIGN")
    (8724 . "DOT PLUS")
    (8725 . "DIVISION SLASH")
    (8726 . "SET MINUS")
    (8727 . "ASTERISK OPERATOR")
    (8728 . "RING OPERATOR")
    (8729 . "BULLET OPERATOR")
    (8730 . "SQUARE ROOT")
    (8731 . "CUBE ROOT")
    (8732 . "FOURTH ROOT")
    (8733 . "PROPORTIONAL TO")
    (8734 . "INFINITY")
    (8735 . "RIGHT ANGLE")
    (8736 . "ANGLE")
    (8737 . "MEASURED ANGLE")
    (8738 . "SPHERICAL ANGLE")
    (8739 . "DIVIDES")
    (8740 . "DOES NOT DIVIDE")
    (8741 . "PARALLEL TO")
    (8742 . "NOT PARALLEL TO")
    (8743 . "LOGICAL AND")
    (8744 . "LOGICAL OR")
    (8745 . "INTERSECTION")
    (8746 . "UNION")
    (8747 . "INTEGRAL")
    (8748 . "DOUBLE INTEGRAL")
    (8749 . "TRIPLE INTEGRAL")
    (8750 . "CONTOUR INTEGRAL")
    (8751 . "SURFACE INTEGRAL")
    (8752 . "VOLUME INTEGRAL")
    (8753 . "CLOCKWISE INTEGRAL")
    (8754 . "CLOCKWISE CONTOUR INTEGRAL")
    (8755 . "ANTICLOCKWISE CONTOUR INTEGRAL")
    (8756 . "THEREFORE")
    (8757 . "BECAUSE")
    (8758 . "RATIO")
    (8759 . "PROPORTION")
    (8760 . "DOT MINUS")
    (8761 . "EXCESS")
    (8762 . "GEOMETRIC PROPORTION")
    (8763 . "HOMOTHETIC")
    (8764 . "TILDE OPERATOR")
    (8765 . "REVERSED TILDE")
    (8766 . "INVERTED LAZY S")
    (8767 . "SINE WAVE")
    (8768 . "WREATH PRODUCT")
    (8769 . "NOT TILDE")
    (8770 . "MINUS TILDE")
    (8771 . "ASYMPTOTICALLY EQUAL TO")
    (8772 . "NOT ASYMPTOTICALLY EQUAL TO")
    (8773 . "APPROXIMATELY EQUAL TO")
    (8774 . "APPROXIMATELY BUT NOT ACTUALLY EQUAL TO")
    (8775 . "NEITHER APPROXIMATELY NOR ACTUALLY EQUAL TO")
    (8776 . "ALMOST EQUAL TO")
    (8777 . "NOT ALMOST EQUAL TO")
    (8778 . "ALMOST EQUAL OR EQUAL TO")
    (8779 . "TRIPLE TILDE")
    (8780 . "ALL EQUAL TO")
    (8781 . "EQUIVALENT TO")
    (8782 . "GEOMETRICALLY EQUIVALENT TO")
    (8783 . "DIFFERENCE BETWEEN")
    (8784 . "APPROACHES THE LIMIT")
    (8785 . "GEOMETRICALLY EQUAL TO")
    (8786 . "APPROXIMATELY EQUAL TO OR THE IMAGE OF")
    (8787 . "IMAGE OF OR APPROXIMATELY EQUAL TO")
    (8788 . "COLON EQUALS")
    (8789 . "EQUALS COLON")
    (8790 . "RING IN EQUAL TO")
    (8791 . "RING EQUAL TO")
    (8792 . "CORRESPONDS TO")
    (8793 . "ESTIMATES")
    (8794 . "EQUIANGULAR TO")
    (8795 . "STAR EQUALS")
    (8796 . "DELTA EQUAL TO")
    (8797 . "EQUAL TO BY DEFINITION")
    (8798 . "MEASURED BY")
    (8799 . "QUESTIONED EQUAL TO")
    (8800 . "NOT EQUAL TO")
    (8801 . "IDENTICAL TO")
    (8802 . "NOT IDENTICAL TO")
    (8803 . "STRICTLY EQUIVALENT TO")
    (8804 . "LESS-THAN OR EQUAL TO")
    (8805 . "GREATER-THAN OR EQUAL TO")
    (8806 . "LESS-THAN OVER EQUAL TO")
    (8807 . "GREATER-THAN OVER EQUAL TO")
    (8808 . "LESS-THAN BUT NOT EQUAL TO")
    (8809 . "GREATER-THAN BUT NOT EQUAL TO")
    (8810 . "MUCH LESS-THAN")
    (8811 . "MUCH GREATER-THAN")
    (8812 . "BETWEEN")
    (8813 . "NOT EQUIVALENT TO")
    (8814 . "NOT LESS-THAN")
    (8815 . "NOT GREATER-THAN")
    (8816 . "NEITHER LESS-THAN NOR EQUAL TO")
    (8817 . "NEITHER GREATER-THAN NOR EQUAL TO")
    (8818 . "LESS-THAN OR EQUIVALENT TO")
    (8819 . "GREATER-THAN OR EQUIVALENT TO")
    (8820 . "NEITHER LESS-THAN NOR EQUIVALENT TO")
    (8821 . "NEITHER GREATER-THAN NOR EQUIVALENT TO")
    (8822 . "LESS-THAN OR GREATER-THAN")
    (8823 . "GREATER-THAN OR LESS-THAN")
    (8824 . "NEITHER LESS-THAN NOR GREATER-THAN")
    (8825 . "NEITHER GREATER-THAN NOR LESS-THAN")
    (8826 . "PRECEDES")
    (8827 . "SUCCEEDS")
    (8828 . "PRECEDES OR EQUAL TO")
    (8829 . "SUCCEEDS OR EQUAL TO")
    (8830 . "PRECEDES OR EQUIVALENT TO")
    (8831 . "SUCCEEDS OR EQUIVALENT TO")
    (8832 . "DOES NOT PRECEDE")
    (8833 . "DOES NOT SUCCEED")
    (8834 . "SUBSET OF")
    (8835 . "SUPERSET OF")
    (8836 . "NOT A SUBSET OF")
    (8837 . "NOT A SUPERSET OF")
    (8838 . "SUBSET OF OR EQUAL TO")
    (8839 . "SUPERSET OF OR EQUAL TO")
    (8840 . "NEITHER A SUBSET OF NOR EQUAL TO")
    (8841 . "NEITHER A SUPERSET OF NOR EQUAL TO")
    (8842 . "SUBSET OF WITH NOT EQUAL TO")
    (8843 . "SUPERSET OF WITH NOT EQUAL TO")
    (8844 . "MULTISET")
    (8845 . "MULTISET MULTIPLICATION")
    (8846 . "MULTISET UNION")
    (8847 . "SQUARE IMAGE OF")
    (8848 . "SQUARE ORIGINAL OF")
    (8849 . "SQUARE IMAGE OF OR EQUAL TO")
    (8850 . "SQUARE ORIGINAL OF OR EQUAL TO")
    (8851 . "SQUARE CAP")
    (8852 . "SQUARE CUP")
    (8853 . "CIRCLED PLUS")
    (8854 . "CIRCLED MINUS")
    (8855 . "CIRCLED TIMES")
    (8856 . "CIRCLED DIVISION SLASH")
    (8857 . "CIRCLED DOT OPERATOR")
    (8858 . "CIRCLED RING OPERATOR")
    (8859 . "CIRCLED ASTERISK OPERATOR")
    (8860 . "CIRCLED EQUALS")
    (8861 . "CIRCLED DASH")
    (8862 . "SQUARED PLUS")
    (8863 . "SQUARED MINUS")
    (8864 . "SQUARED TIMES")
    (8865 . "SQUARED DOT OPERATOR")
    (8866 . "RIGHT TACK")
    (8867 . "LEFT TACK")
    (8868 . "DOWN TACK")
    (8869 . "UP TACK")
    (8870 . "ASSERTION")
    (8871 . "MODELS")
    (8872 . "TRUE")
    (8873 . "FORCES")
    (8874 . "TRIPLE VERTICAL BAR RIGHT TURNSTILE")
    (8875 . "DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE")
    (8876 . "DOES NOT PROVE")
    (8877 . "NOT TRUE")
    (8878 . "DOES NOT FORCE")
    (8879 . "NEGATED DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE")
    (8880 . "PRECEDES UNDER RELATION")
    (8881 . "SUCCEEDS UNDER RELATION")
    (8882 . "NORMAL SUBGROUP OF")
    (8883 . "CONTAINS AS NORMAL SUBGROUP")
    (8884 . "NORMAL SUBGROUP OF OR EQUAL TO")
    (8885 . "CONTAINS AS NORMAL SUBGROUP OR EQUAL TO")
    (8886 . "ORIGINAL OF")
    (8887 . "IMAGE OF")
    (8888 . "MULTIMAP")
    (8889 . "HERMITIAN CONJUGATE MATRIX")
    (8890 . "INTERCALATE")
    (8891 . "XOR")
    (8892 . "NAND")
    (8893 . "NOR")
    (8894 . "RIGHT ANGLE WITH ARC")
    (8895 . "RIGHT TRIANGLE")
    (8896 . "N-ARY LOGICAL AND")
    (8897 . "N-ARY LOGICAL OR")
    (8898 . "N-ARY INTERSECTION")
    (8899 . "N-ARY UNION")
    (8900 . "DIAMOND OPERATOR")
    (8901 . "DOT OPERATOR")
    (8902 . "STAR OPERATOR")
    (8903 . "DIVISION TIMES")
    (8904 . "BOWTIE")
    (8905 . "LEFT NORMAL FACTOR SEMIDIRECT PRODUCT")
    (8906 . "RIGHT NORMAL FACTOR SEMIDIRECT PRODUCT")
    (8907 . "LEFT SEMIDIRECT PRODUCT")
    (8908 . "RIGHT SEMIDIRECT PRODUCT")
    (8909 . "REVERSED TILDE EQUALS")
    (8910 . "CURLY LOGICAL OR")
    (8911 . "CURLY LOGICAL AND")
    (8912 . "DOUBLE SUBSET")
    (8913 . "DOUBLE SUPERSET")
    (8914 . "DOUBLE INTERSECTION")
    (8915 . "DOUBLE UNION")
    (8916 . "PITCHFORK")
    (8917 . "EQUAL AND PARALLEL TO")
    (8918 . "LESS-THAN WITH DOT")
    (8919 . "GREATER-THAN WITH DOT")
    (8920 . "VERY MUCH LESS-THAN")
    (8921 . "VERY MUCH GREATER-THAN")
    (8922 . "LESS-THAN EQUAL TO OR GREATER-THAN")
    (8923 . "GREATER-THAN EQUAL TO OR LESS-THAN")
    (8924 . "EQUAL TO OR LESS-THAN")
    (8925 . "EQUAL TO OR GREATER-THAN")
    (8926 . "EQUAL TO OR PRECEDES")
    (8927 . "EQUAL TO OR SUCCEEDS")
    (8928 . "DOES NOT PRECEDE OR EQUAL")
    (8929 . "DOES NOT SUCCEED OR EQUAL")
    (8930 . "NOT SQUARE IMAGE OF OR EQUAL TO")
    (8931 . "NOT SQUARE ORIGINAL OF OR EQUAL TO")
    (8932 . "SQUARE IMAGE OF OR NOT EQUAL TO")
    (8933 . "SQUARE ORIGINAL OF OR NOT EQUAL TO")
    (8934 . "LESS-THAN BUT NOT EQUIVALENT TO")
    (8935 . "GREATER-THAN BUT NOT EQUIVALENT TO")
    (8936 . "PRECEDES BUT NOT EQUIVALENT TO")
    (8937 . "SUCCEEDS BUT NOT EQUIVALENT TO")
    (8938 . "NOT NORMAL SUBGROUP OF")
    (8939 . "DOES NOT CONTAIN AS NORMAL SUBGROUP")
    (8940 . "NOT NORMAL SUBGROUP OF OR EQUAL TO")
    (8941 . "DOES NOT CONTAIN AS NORMAL SUBGROUP OR EQUAL")
    (8942 . "VERTICAL ELLIPSIS")
    (8943 . "MIDLINE HORIZONTAL ELLIPSIS")
    (8944 . "UP RIGHT DIAGONAL ELLIPSIS")
    (8945 . "DOWN RIGHT DIAGONAL ELLIPSIS")
    (8960 . "DIAMETER SIGN")
    (8962 . "HOUSE")
    (8963 . "UP ARROWHEAD")
    (8964 . "DOWN ARROWHEAD")
    (8965 . "PROJECTIVE")
    (8966 . "PERSPECTIVE")
    (8967 . "WAVY LINE")
    (8968 . "LEFT CEILING")
    (8969 . "RIGHT CEILING")
    (8970 . "LEFT FLOOR")
    (8971 . "RIGHT FLOOR")
    (8972 . "BOTTOM RIGHT CROP")
    (8973 . "BOTTOM LEFT CROP")
    (8974 . "TOP RIGHT CROP")
    (8975 . "TOP LEFT CROP")
    (8976 . "REVERSED NOT SIGN")
    (8977 . "SQUARE LOZENGE")
    (8978 . "ARC")
    (8979 . "SEGMENT")
    (8980 . "SECTOR")
    (8981 . "TELEPHONE RECORDER")
    (8982 . "POSITION INDICATOR")
    (8983 . "VIEWDATA SQUARE")
    (8984 . "PLACE OF INTEREST SIGN")
    (8985 . "TURNED NOT SIGN")
    (8986 . "WATCH")
    (8987 . "HOURGLASS")
    (8988 . "TOP LEFT CORNER")
    (8989 . "TOP RIGHT CORNER")
    (8990 . "BOTTOM LEFT CORNER")
    (8991 . "BOTTOM RIGHT CORNER")
    (8992 . "TOP HALF INTEGRAL")
    (8993 . "BOTTOM HALF INTEGRAL")
    (8994 . "FROWN")
    (8995 . "SMILE")
    (8996 . "UP ARROWHEAD BETWEEN TWO HORIZONTAL BARS")
    (8997 . "OPTION KEY")
    (8998 . "ERASE TO THE RIGHT")
    (8999 . "X IN A RECTANGLE BOX")
    (9000 . "KEYBOARD")
    (9001 . "LEFT-POINTING ANGLE BRACKET")
    (9002 . "RIGHT-POINTING ANGLE BRACKET")
    (9003 . "ERASE TO THE LEFT")
    (9004 . "BENZENE RING")
    (9005 . "CYLINDRICITY")
    (9006 . "ALL AROUND-PROFILE")
    (9007 . "SYMMETRY")
    (9008 . "TOTAL RUNOUT")
    (9009 . "DIMENSION ORIGIN")
    (9010 . "CONICAL TAPER")
    (9011 . "SLOPE")
    (9012 . "COUNTERBORE")
    (9013 . "COUNTERSINK")
    (9014 . "APL FUNCTIONAL SYMBOL I-BEAM")
    (9015 . "APL FUNCTIONAL SYMBOL SQUISH QUAD")
    (9016 . "APL FUNCTIONAL SYMBOL QUAD EQUAL")
    (9017 . "APL FUNCTIONAL SYMBOL QUAD DIVIDE")
    (9018 . "APL FUNCTIONAL SYMBOL QUAD DIAMOND")
    (9019 . "APL FUNCTIONAL SYMBOL QUAD JOT")
    (9020 . "APL FUNCTIONAL SYMBOL QUAD CIRCLE")
    (9021 . "APL FUNCTIONAL SYMBOL CIRCLE STILE")
    (9022 . "APL FUNCTIONAL SYMBOL CIRCLE JOT")
    (9023 . "APL FUNCTIONAL SYMBOL SLASH BAR")
    (9024 . "APL FUNCTIONAL SYMBOL BACKSLASH BAR")
    (9025 . "APL FUNCTIONAL SYMBOL QUAD SLASH")
    (9026 . "APL FUNCTIONAL SYMBOL QUAD BACKSLASH")
    (9027 . "APL FUNCTIONAL SYMBOL QUAD LESS-THAN")
    (9028 . "APL FUNCTIONAL SYMBOL QUAD GREATER-THAN")
    (9029 . "APL FUNCTIONAL SYMBOL LEFTWARDS VANE")
    (9030 . "APL FUNCTIONAL SYMBOL RIGHTWARDS VANE")
    (9031 . "APL FUNCTIONAL SYMBOL QUAD LEFTWARDS ARROW")
    (9032 . "APL FUNCTIONAL SYMBOL QUAD RIGHTWARDS ARROW")
    (9033 . "APL FUNCTIONAL SYMBOL CIRCLE BACKSLASH")
    (9034 . "APL FUNCTIONAL SYMBOL DOWN TACK UNDERBAR")
    (9035 . "APL FUNCTIONAL SYMBOL DELTA STILE")
    (9036 . "APL FUNCTIONAL SYMBOL QUAD DOWN CARET")
    (9037 . "APL FUNCTIONAL SYMBOL QUAD DELTA")
    (9038 . "APL FUNCTIONAL SYMBOL DOWN TACK JOT")
    (9039 . "APL FUNCTIONAL SYMBOL UPWARDS VANE")
    (9040 . "APL FUNCTIONAL SYMBOL QUAD UPWARDS ARROW")
    (9041 . "APL FUNCTIONAL SYMBOL UP TACK OVERBAR")
    (9042 . "APL FUNCTIONAL SYMBOL DEL STILE")
    (9043 . "APL FUNCTIONAL SYMBOL QUAD UP CARET")
    (9044 . "APL FUNCTIONAL SYMBOL QUAD DEL")
    (9045 . "APL FUNCTIONAL SYMBOL UP TACK JOT")
    (9046 . "APL FUNCTIONAL SYMBOL DOWNWARDS VANE")
    (9047 . "APL FUNCTIONAL SYMBOL QUAD DOWNWARDS ARROW")
    (9048 . "APL FUNCTIONAL SYMBOL QUOTE UNDERBAR")
    (9049 . "APL FUNCTIONAL SYMBOL DELTA UNDERBAR")
    (9050 . "APL FUNCTIONAL SYMBOL DIAMOND UNDERBAR")
    (9051 . "APL FUNCTIONAL SYMBOL JOT UNDERBAR")
    (9052 . "APL FUNCTIONAL SYMBOL CIRCLE UNDERBAR")
    (9053 . "APL FUNCTIONAL SYMBOL UP SHOE JOT")
    (9054 . "APL FUNCTIONAL SYMBOL QUOTE QUAD")
    (9055 . "APL FUNCTIONAL SYMBOL CIRCLE STAR")
    (9056 . "APL FUNCTIONAL SYMBOL QUAD COLON")
    (9057 . "APL FUNCTIONAL SYMBOL UP TACK DIAERESIS")
    (9058 . "APL FUNCTIONAL SYMBOL DEL DIAERESIS")
    (9059 . "APL FUNCTIONAL SYMBOL STAR DIAERESIS")
    (9060 . "APL FUNCTIONAL SYMBOL JOT DIAERESIS")
    (9061 . "APL FUNCTIONAL SYMBOL CIRCLE DIAERESIS")
    (9062 . "APL FUNCTIONAL SYMBOL DOWN SHOE STILE")
    (9063 . "APL FUNCTIONAL SYMBOL LEFT SHOE STILE")
    (9064 . "APL FUNCTIONAL SYMBOL TILDE DIAERESIS")
    (9065 . "APL FUNCTIONAL SYMBOL GREATER-THAN DIAERESIS")
    (9066 . "APL FUNCTIONAL SYMBOL COMMA BAR")
    (9067 . "APL FUNCTIONAL SYMBOL DEL TILDE")
    (9068 . "APL FUNCTIONAL SYMBOL ZILDE")
    (9069 . "APL FUNCTIONAL SYMBOL STILE TILDE")
    (9070 . "APL FUNCTIONAL SYMBOL SEMICOLON UNDERBAR")
    (9071 . "APL FUNCTIONAL SYMBOL QUAD NOT EQUAL")
    (9072 . "APL FUNCTIONAL SYMBOL QUAD QUESTION")
    (9073 . "APL FUNCTIONAL SYMBOL DOWN CARET TILDE")
    (9074 . "APL FUNCTIONAL SYMBOL UP CARET TILDE")
    (9075 . "APL FUNCTIONAL SYMBOL IOTA")
    (9076 . "APL FUNCTIONAL SYMBOL RHO")
    (9077 . "APL FUNCTIONAL SYMBOL OMEGA")
    (9078 . "APL FUNCTIONAL SYMBOL ALPHA UNDERBAR")
    (9079 . "APL FUNCTIONAL SYMBOL EPSILON UNDERBAR")
    (9080 . "APL FUNCTIONAL SYMBOL IOTA UNDERBAR")
    (9081 . "APL FUNCTIONAL SYMBOL OMEGA UNDERBAR")
    (9082 . "APL FUNCTIONAL SYMBOL ALPHA")
    (9216 . "SYMBOL FOR NULL")
    (9217 . "SYMBOL FOR START OF HEADING")
    (9218 . "SYMBOL FOR START OF TEXT")
    (9219 . "SYMBOL FOR END OF TEXT")
    (9220 . "SYMBOL FOR END OF TRANSMISSION")
    (9221 . "SYMBOL FOR ENQUIRY")
    (9222 . "SYMBOL FOR ACKNOWLEDGE")
    (9223 . "SYMBOL FOR BELL")
    (9224 . "SYMBOL FOR BACKSPACE")
    (9225 . "SYMBOL FOR HORIZONTAL TABULATION")
    (9226 . "SYMBOL FOR LINE FEED")
    (9227 . "SYMBOL FOR VERTICAL TABULATION")
    (9228 . "SYMBOL FOR FORM FEED")
    (9229 . "SYMBOL FOR CARRIAGE RETURN")
    (9230 . "SYMBOL FOR SHIFT OUT")
    (9231 . "SYMBOL FOR SHIFT IN")
    (9232 . "SYMBOL FOR DATA LINK ESCAPE")
    (9233 . "SYMBOL FOR DEVICE CONTROL ONE")
    (9234 . "SYMBOL FOR DEVICE CONTROL TWO")
    (9235 . "SYMBOL FOR DEVICE CONTROL THREE")
    (9236 . "SYMBOL FOR DEVICE CONTROL FOUR")
    (9237 . "SYMBOL FOR NEGATIVE ACKNOWLEDGE")
    (9238 . "SYMBOL FOR SYNCHRONOUS IDLE")
    (9239 . "SYMBOL FOR END OF TRANSMISSION BLOCK")
    (9240 . "SYMBOL FOR CANCEL")
    (9241 . "SYMBOL FOR END OF MEDIUM")
    (9242 . "SYMBOL FOR SUBSTITUTE")
    (9243 . "SYMBOL FOR ESCAPE")
    (9244 . "SYMBOL FOR FILE SEPARATOR")
    (9245 . "SYMBOL FOR GROUP SEPARATOR")
    (9246 . "SYMBOL FOR RECORD SEPARATOR")
    (9247 . "SYMBOL FOR UNIT SEPARATOR")
    (9248 . "SYMBOL FOR SPACE")
    (9249 . "SYMBOL FOR DELETE")
    (9250 . "BLANK SYMBOL")
    (9251 . "OPEN BOX")
    (9252 . "SYMBOL FOR NEWLINE")
    (9280 . "OCR HOOK")
    (9281 . "OCR CHAIR")
    (9282 . "OCR FORK")
    (9283 . "OCR INVERTED FORK")
    (9284 . "OCR BELT BUCKLE")
    (9285 . "OCR BOW TIE")
    (9286 . "OCR BRANCH BANK IDENTIFICATION")
    (9287 . "OCR AMOUNT OF CHECK")
    (9288 . "OCR DASH")
    (9289 . "OCR CUSTOMER ACCOUNT NUMBER")
    (9290 . "OCR DOUBLE BACKSLASH")
    (9312 . "CIRCLED DIGIT ONE")
    (9313 . "CIRCLED DIGIT TWO")
    (9314 . "CIRCLED DIGIT THREE")
    (9315 . "CIRCLED DIGIT FOUR")
    (9316 . "CIRCLED DIGIT FIVE")
    (9317 . "CIRCLED DIGIT SIX")
    (9318 . "CIRCLED DIGIT SEVEN")
    (9319 . "CIRCLED DIGIT EIGHT")
    (9320 . "CIRCLED DIGIT NINE")
    (9321 . "CIRCLED NUMBER TEN")
    (9322 . "CIRCLED NUMBER ELEVEN")
    (9323 . "CIRCLED NUMBER TWELVE")
    (9324 . "CIRCLED NUMBER THIRTEEN")
    (9325 . "CIRCLED NUMBER FOURTEEN")
    (9326 . "CIRCLED NUMBER FIFTEEN")
    (9327 . "CIRCLED NUMBER SIXTEEN")
    (9328 . "CIRCLED NUMBER SEVENTEEN")
    (9329 . "CIRCLED NUMBER EIGHTEEN")
    (9330 . "CIRCLED NUMBER NINETEEN")
    (9331 . "CIRCLED NUMBER TWENTY")
    (9332 . "PARENTHESIZED DIGIT ONE")
    (9333 . "PARENTHESIZED DIGIT TWO")
    (9334 . "PARENTHESIZED DIGIT THREE")
    (9335 . "PARENTHESIZED DIGIT FOUR")
    (9336 . "PARENTHESIZED DIGIT FIVE")
    (9337 . "PARENTHESIZED DIGIT SIX")
    (9338 . "PARENTHESIZED DIGIT SEVEN")
    (9339 . "PARENTHESIZED DIGIT EIGHT")
    (9340 . "PARENTHESIZED DIGIT NINE")
    (9341 . "PARENTHESIZED NUMBER TEN")
    (9342 . "PARENTHESIZED NUMBER ELEVEN")
    (9343 . "PARENTHESIZED NUMBER TWELVE")
    (9344 . "PARENTHESIZED NUMBER THIRTEEN")
    (9345 . "PARENTHESIZED NUMBER FOURTEEN")
    (9346 . "PARENTHESIZED NUMBER FIFTEEN")
    (9347 . "PARENTHESIZED NUMBER SIXTEEN")
    (9348 . "PARENTHESIZED NUMBER SEVENTEEN")
    (9349 . "PARENTHESIZED NUMBER EIGHTEEN")
    (9350 . "PARENTHESIZED NUMBER NINETEEN")
    (9351 . "PARENTHESIZED NUMBER TWENTY")
    (9352 . "DIGIT ONE PERIOD|DIGIT ONE FULL STOP")
    (9353 . "DIGIT TWO PERIOD|DIGIT TWO FULL STOP")
    (9354 . "DIGIT THREE PERIOD|DIGIT THREE FULL STOP")
    (9355 . "DIGIT FOUR PERIOD|DIGIT FOUR FULL STOP")
    (9356 . "DIGIT FIVE PERIOD|DIGIT FIVE FULL STOP")
    (9357 . "DIGIT SIX PERIOD|DIGIT SIX FULL STOP")
    (9358 . "DIGIT SEVEN PERIOD|DIGIT SEVEN FULL STOP")
    (9359 . "DIGIT EIGHT PERIOD|DIGIT EIGHT FULL STOP")
    (9360 . "DIGIT NINE PERIOD|DIGIT NINE FULL STOP")
    (9361 . "NUMBER TEN PERIOD|NUMBER TEN FULL STOP")
    (9362 . "NUMBER ELEVEN PERIOD|NUMBER ELEVEN FULL STOP")
    (9363 . "NUMBER TWELVE PERIOD|NUMBER TWELVE FULL STOP")
    (9364 . "NUMBER THIRTEEN PERIOD|NUMBER THIRTEEN FULL STOP")
    (9365 . "NUMBER FOURTEEN PERIOD|NUMBER FOURTEEN FULL STOP")
    (9366 . "NUMBER FIFTEEN PERIOD|NUMBER FIFTEEN FULL STOP")
    (9367 . "NUMBER SIXTEEN PERIOD|NUMBER SIXTEEN FULL STOP")
    (9368 . "NUMBER SEVENTEEN PERIOD|NUMBER SEVENTEEN FULL STOP")
    (9369 . "NUMBER EIGHTEEN PERIOD|NUMBER EIGHTEEN FULL STOP")
    (9370 . "NUMBER NINETEEN PERIOD|NUMBER NINETEEN FULL STOP")
    (9371 . "NUMBER TWENTY PERIOD|NUMBER TWENTY FULL STOP")
    (9372 . "PARENTHESIZED LATIN SMALL LETTER A")
    (9373 . "PARENTHESIZED LATIN SMALL LETTER B")
    (9374 . "PARENTHESIZED LATIN SMALL LETTER C")
    (9375 . "PARENTHESIZED LATIN SMALL LETTER D")
    (9376 . "PARENTHESIZED LATIN SMALL LETTER E")
    (9377 . "PARENTHESIZED LATIN SMALL LETTER F")
    (9378 . "PARENTHESIZED LATIN SMALL LETTER G")
    (9379 . "PARENTHESIZED LATIN SMALL LETTER H")
    (9380 . "PARENTHESIZED LATIN SMALL LETTER I")
    (9381 . "PARENTHESIZED LATIN SMALL LETTER J")
    (9382 . "PARENTHESIZED LATIN SMALL LETTER K")
    (9383 . "PARENTHESIZED LATIN SMALL LETTER L")
    (9384 . "PARENTHESIZED LATIN SMALL LETTER M")
    (9385 . "PARENTHESIZED LATIN SMALL LETTER N")
    (9386 . "PARENTHESIZED LATIN SMALL LETTER O")
    (9387 . "PARENTHESIZED LATIN SMALL LETTER P")
    (9388 . "PARENTHESIZED LATIN SMALL LETTER Q")
    (9389 . "PARENTHESIZED LATIN SMALL LETTER R")
    (9390 . "PARENTHESIZED LATIN SMALL LETTER S")
    (9391 . "PARENTHESIZED LATIN SMALL LETTER T")
    (9392 . "PARENTHESIZED LATIN SMALL LETTER U")
    (9393 . "PARENTHESIZED LATIN SMALL LETTER V")
    (9394 . "PARENTHESIZED LATIN SMALL LETTER W")
    (9395 . "PARENTHESIZED LATIN SMALL LETTER X")
    (9396 . "PARENTHESIZED LATIN SMALL LETTER Y")
    (9397 . "PARENTHESIZED LATIN SMALL LETTER Z")
    (9398 . "CIRCLED LATIN CAPITAL LETTER A")
    (9399 . "CIRCLED LATIN CAPITAL LETTER B")
    (9400 . "CIRCLED LATIN CAPITAL LETTER C")
    (9401 . "CIRCLED LATIN CAPITAL LETTER D")
    (9402 . "CIRCLED LATIN CAPITAL LETTER E")
    (9403 . "CIRCLED LATIN CAPITAL LETTER F")
    (9404 . "CIRCLED LATIN CAPITAL LETTER G")
    (9405 . "CIRCLED LATIN CAPITAL LETTER H")
    (9406 . "CIRCLED LATIN CAPITAL LETTER I")
    (9407 . "CIRCLED LATIN CAPITAL LETTER J")
    (9408 . "CIRCLED LATIN CAPITAL LETTER K")
    (9409 . "CIRCLED LATIN CAPITAL LETTER L")
    (9410 . "CIRCLED LATIN CAPITAL LETTER M")
    (9411 . "CIRCLED LATIN CAPITAL LETTER N")
    (9412 . "CIRCLED LATIN CAPITAL LETTER O")
    (9413 . "CIRCLED LATIN CAPITAL LETTER P")
    (9414 . "CIRCLED LATIN CAPITAL LETTER Q")
    (9415 . "CIRCLED LATIN CAPITAL LETTER R")
    (9416 . "CIRCLED LATIN CAPITAL LETTER S")
    (9417 . "CIRCLED LATIN CAPITAL LETTER T")
    (9418 . "CIRCLED LATIN CAPITAL LETTER U")
    (9419 . "CIRCLED LATIN CAPITAL LETTER V")
    (9420 . "CIRCLED LATIN CAPITAL LETTER W")
    (9421 . "CIRCLED LATIN CAPITAL LETTER X")
    (9422 . "CIRCLED LATIN CAPITAL LETTER Y")
    (9423 . "CIRCLED LATIN CAPITAL LETTER Z")
    (9424 . "CIRCLED LATIN SMALL LETTER A")
    (9425 . "CIRCLED LATIN SMALL LETTER B")
    (9426 . "CIRCLED LATIN SMALL LETTER C")
    (9427 . "CIRCLED LATIN SMALL LETTER D")
    (9428 . "CIRCLED LATIN SMALL LETTER E")
    (9429 . "CIRCLED LATIN SMALL LETTER F")
    (9430 . "CIRCLED LATIN SMALL LETTER G")
    (9431 . "CIRCLED LATIN SMALL LETTER H")
    (9432 . "CIRCLED LATIN SMALL LETTER I")
    (9433 . "CIRCLED LATIN SMALL LETTER J")
    (9434 . "CIRCLED LATIN SMALL LETTER K")
    (9435 . "CIRCLED LATIN SMALL LETTER L")
    (9436 . "CIRCLED LATIN SMALL LETTER M")
    (9437 . "CIRCLED LATIN SMALL LETTER N")
    (9438 . "CIRCLED LATIN SMALL LETTER O")
    (9439 . "CIRCLED LATIN SMALL LETTER P")
    (9440 . "CIRCLED LATIN SMALL LETTER Q")
    (9441 . "CIRCLED LATIN SMALL LETTER R")
    (9442 . "CIRCLED LATIN SMALL LETTER S")
    (9443 . "CIRCLED LATIN SMALL LETTER T")
    (9444 . "CIRCLED LATIN SMALL LETTER U")
    (9445 . "CIRCLED LATIN SMALL LETTER V")
    (9446 . "CIRCLED LATIN SMALL LETTER W")
    (9447 . "CIRCLED LATIN SMALL LETTER X")
    (9448 . "CIRCLED LATIN SMALL LETTER Y")
    (9449 . "CIRCLED LATIN SMALL LETTER Z")
    (9450 . "CIRCLED DIGIT ZERO")
    (9472 . "BOX DRAWINGS LIGHT HORIZONTAL")
    (9473 . "BOX DRAWINGS HEAVY HORIZONTAL")
    (9474 . "BOX DRAWINGS LIGHT VERTICAL")
    (9475 . "BOX DRAWINGS HEAVY VERTICAL")
    (9476 . "BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL")
    (9477 . "BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL")
    (9478 . "BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL")
    (9479 . "BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL")
    (9480 . "BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL")
    (9481 . "BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL")
    (9482 . "BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL")
    (9483 . "BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL")
    (9484 . "BOX DRAWINGS LIGHT DOWN AND RIGHT")
    (9485 . "BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY")
    (9486 . "BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT")
    (9487 . "BOX DRAWINGS HEAVY DOWN AND RIGHT")
    (9488 . "BOX DRAWINGS LIGHT DOWN AND LEFT")
    (9489 . "BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY")
    (9490 . "BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT")
    (9491 . "BOX DRAWINGS HEAVY DOWN AND LEFT")
    (9492 . "BOX DRAWINGS LIGHT UP AND RIGHT")
    (9493 . "BOX DRAWINGS UP LIGHT AND RIGHT HEAVY")
    (9494 . "BOX DRAWINGS UP HEAVY AND RIGHT LIGHT")
    (9495 . "BOX DRAWINGS HEAVY UP AND RIGHT")
    (9496 . "BOX DRAWINGS LIGHT UP AND LEFT")
    (9497 . "BOX DRAWINGS UP LIGHT AND LEFT HEAVY")
    (9498 . "BOX DRAWINGS UP HEAVY AND LEFT LIGHT")
    (9499 . "BOX DRAWINGS HEAVY UP AND LEFT")
    (9500 . "BOX DRAWINGS LIGHT VERTICAL AND RIGHT")
    (9501 . "BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY")
    (9502 . "BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT")
    (9503 . "BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT")
    (9504 . "BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT")
    (9505 . "BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY")
    (9506 . "BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY")
    (9507 . "BOX DRAWINGS HEAVY VERTICAL AND RIGHT")
    (9508 . "BOX DRAWINGS LIGHT VERTICAL AND LEFT")
    (9509 . "BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY")
    (9510 . "BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT")
    (9511 . "BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT")
    (9512 . "BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT")
    (9513 . "BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY")
    (9514 . "BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY")
    (9515 . "BOX DRAWINGS HEAVY VERTICAL AND LEFT")
    (9516 . "BOX DRAWINGS LIGHT DOWN AND HORIZONTAL")
    (9517 . "BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT")
    (9518 . "BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT")
    (9519 . "BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY")
    (9520 . "BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT")
    (9521 . "BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY")
    (9522 . "BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY")
    (9523 . "BOX DRAWINGS HEAVY DOWN AND HORIZONTAL")
    (9524 . "BOX DRAWINGS LIGHT UP AND HORIZONTAL")
    (9525 . "BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT")
    (9526 . "BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT")
    (9527 . "BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY")
    (9528 . "BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT")
    (9529 . "BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY")
    (9530 . "BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY")
    (9531 . "BOX DRAWINGS HEAVY UP AND HORIZONTAL")
    (9532 . "BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL")
    (9533 . "BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT")
    (9534 . "BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT")
    (9535 . "BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY")
    (9536 . "BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT")
    (9537 . "BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT")
    (9538 . "BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT")
    (9539 . "BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT")
    (9540 . "BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT")
    (9541 . "BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT")
    (9542 . "BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT")
    (9543 . "BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY")
    (9544 . "BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY")
    (9545 . "BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY")
    (9546 . "BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY")
    (9547 . "BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL")
    (9548 . "BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL")
    (9549 . "BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL")
    (9550 . "BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL")
    (9551 . "BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL")
    (9552 . "BOX DRAWINGS DOUBLE HORIZONTAL")
    (9553 . "BOX DRAWINGS DOUBLE VERTICAL")
    (9554 . "BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE")
    (9555 . "BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE")
    (9556 . "BOX DRAWINGS DOUBLE DOWN AND RIGHT")
    (9557 . "BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE")
    (9558 . "BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE")
    (9559 . "BOX DRAWINGS DOUBLE DOWN AND LEFT")
    (9560 . "BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE")
    (9561 . "BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE")
    (9562 . "BOX DRAWINGS DOUBLE UP AND RIGHT")
    (9563 . "BOX DRAWINGS UP SINGLE AND LEFT DOUBLE")
    (9564 . "BOX DRAWINGS UP DOUBLE AND LEFT SINGLE")
    (9565 . "BOX DRAWINGS DOUBLE UP AND LEFT")
    (9566 . "BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE")
    (9567 . "BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE")
    (9568 . "BOX DRAWINGS DOUBLE VERTICAL AND RIGHT")
    (9569 . "BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE")
    (9570 . "BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE")
    (9571 . "BOX DRAWINGS DOUBLE VERTICAL AND LEFT")
    (9572 . "BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE")
    (9573 . "BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE")
    (9574 . "BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL")
    (9575 . "BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE")
    (9576 . "BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE")
    (9577 . "BOX DRAWINGS DOUBLE UP AND HORIZONTAL")
    (9578 . "BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE")
    (9579 . "BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE")
    (9580 . "BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL")
    (9581 . "BOX DRAWINGS LIGHT ARC DOWN AND RIGHT")
    (9582 . "BOX DRAWINGS LIGHT ARC DOWN AND LEFT")
    (9583 . "BOX DRAWINGS LIGHT ARC UP AND LEFT")
    (9584 . "BOX DRAWINGS LIGHT ARC UP AND RIGHT")
    (9585 . "BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT")
    (9586 . "BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT")
    (9587 . "BOX DRAWINGS LIGHT DIAGONAL CROSS")
    (9588 . "BOX DRAWINGS LIGHT LEFT")
    (9589 . "BOX DRAWINGS LIGHT UP")
    (9590 . "BOX DRAWINGS LIGHT RIGHT")
    (9591 . "BOX DRAWINGS LIGHT DOWN")
    (9592 . "BOX DRAWINGS HEAVY LEFT")
    (9593 . "BOX DRAWINGS HEAVY UP")
    (9594 . "BOX DRAWINGS HEAVY RIGHT")
    (9595 . "BOX DRAWINGS HEAVY DOWN")
    (9596 . "BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT")
    (9597 . "BOX DRAWINGS LIGHT UP AND HEAVY DOWN")
    (9598 . "BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT")
    (9599 . "BOX DRAWINGS HEAVY UP AND LIGHT DOWN")
    (9600 . "UPPER HALF BLOCK")
    (9601 . "LOWER ONE EIGHTH BLOCK")
    (9602 . "LOWER ONE QUARTER BLOCK")
    (9603 . "LOWER THREE EIGHTHS BLOCK")
    (9604 . "LOWER HALF BLOCK")
    (9605 . "LOWER FIVE EIGHTHS BLOCK")
    (9606 . "LOWER THREE QUARTERS BLOCK")
    (9607 . "LOWER SEVEN EIGHTHS BLOCK")
    (9608 . "FULL BLOCK")
    (9609 . "LEFT SEVEN EIGHTHS BLOCK")
    (9610 . "LEFT THREE QUARTERS BLOCK")
    (9611 . "LEFT FIVE EIGHTHS BLOCK")
    (9612 . "LEFT HALF BLOCK")
    (9613 . "LEFT THREE EIGHTHS BLOCK")
    (9614 . "LEFT ONE QUARTER BLOCK")
    (9615 . "LEFT ONE EIGHTH BLOCK")
    (9616 . "RIGHT HALF BLOCK")
    (9617 . "LIGHT SHADE")
    (9618 . "MEDIUM SHADE")
    (9619 . "DARK SHADE")
    (9620 . "UPPER ONE EIGHTH BLOCK")
    (9621 . "RIGHT ONE EIGHTH BLOCK")
    (9632 . "BLACK SQUARE")
    (9633 . "WHITE SQUARE")
    (9634 . "WHITE SQUARE WITH ROUNDED CORNERS")
    (9635 . "WHITE SQUARE CONTAINING BLACK SMALL SQUARE")
    (9636 . "SQUARE WITH HORIZONTAL FILL")
    (9637 . "SQUARE WITH VERTICAL FILL")
    (9638 . "SQUARE WITH ORTHOGONAL CROSSHATCH FILL")
    (9639 . "SQUARE WITH UPPER LEFT TO LOWER RIGHT FILL")
    (9640 . "SQUARE WITH UPPER RIGHT TO LOWER LEFT FILL")
    (9641 . "SQUARE WITH DIAGONAL CROSSHATCH FILL")
    (9642 . "BLACK SMALL SQUARE")
    (9643 . "WHITE SMALL SQUARE")
    (9644 . "BLACK RECTANGLE")
    (9645 . "WHITE RECTANGLE")
    (9646 . "BLACK VERTICAL RECTANGLE")
    (9647 . "WHITE VERTICAL RECTANGLE")
    (9648 . "BLACK PARALLELOGRAM")
    (9649 . "WHITE PARALLELOGRAM")
    (9650 . "BLACK UP-POINTING TRIANGLE")
    (9651 . "WHITE UP-POINTING TRIANGLE")
    (9652 . "BLACK UP-POINTING SMALL TRIANGLE")
    (9653 . "WHITE UP-POINTING SMALL TRIANGLE")
    (9654 . "BLACK RIGHT-POINTING TRIANGLE")
    (9655 . "WHITE RIGHT-POINTING TRIANGLE")
    (9656 . "BLACK RIGHT-POINTING SMALL TRIANGLE")
    (9657 . "WHITE RIGHT-POINTING SMALL TRIANGLE")
    (9658 . "BLACK RIGHT-POINTING POINTER")
    (9659 . "WHITE RIGHT-POINTING POINTER")
    (9660 . "BLACK DOWN-POINTING TRIANGLE")
    (9661 . "WHITE DOWN-POINTING TRIANGLE")
    (9662 . "BLACK DOWN-POINTING SMALL TRIANGLE")
    (9663 . "WHITE DOWN-POINTING SMALL TRIANGLE")
    (9664 . "BLACK LEFT-POINTING TRIANGLE")
    (9665 . "WHITE LEFT-POINTING TRIANGLE")
    (9666 . "BLACK LEFT-POINTING SMALL TRIANGLE")
    (9667 . "WHITE LEFT-POINTING SMALL TRIANGLE")
    (9668 . "BLACK LEFT-POINTING POINTER")
    (9669 . "WHITE LEFT-POINTING POINTER")
    (9670 . "BLACK DIAMOND")
    (9671 . "WHITE DIAMOND")
    (9672 . "WHITE DIAMOND CONTAINING BLACK SMALL DIAMOND")
    (9673 . "FISHEYE")
    (9674 . "LOZENGE")
    (9675 . "WHITE CIRCLE")
    (9676 . "DOTTED CIRCLE")
    (9677 . "CIRCLE WITH VERTICAL FILL")
    (9678 . "BULLSEYE")
    (9679 . "BLACK CIRCLE")
    (9680 . "CIRCLE WITH LEFT HALF BLACK")
    (9681 . "CIRCLE WITH RIGHT HALF BLACK")
    (9682 . "CIRCLE WITH LOWER HALF BLACK")
    (9683 . "CIRCLE WITH UPPER HALF BLACK")
    (9684 . "CIRCLE WITH UPPER RIGHT QUADRANT BLACK")
    (9685 . "CIRCLE WITH ALL BUT UPPER LEFT QUADRANT BLACK")
    (9686 . "LEFT HALF BLACK CIRCLE")
    (9687 . "RIGHT HALF BLACK CIRCLE")
    (9688 . "INVERSE BULLET")
    (9689 . "INVERSE WHITE CIRCLE")
    (9690 . "UPPER HALF INVERSE WHITE CIRCLE")
    (9691 . "LOWER HALF INVERSE WHITE CIRCLE")
    (9692 . "UPPER LEFT QUADRANT CIRCULAR ARC")
    (9693 . "UPPER RIGHT QUADRANT CIRCULAR ARC")
    (9694 . "LOWER RIGHT QUADRANT CIRCULAR ARC")
    (9695 . "LOWER LEFT QUADRANT CIRCULAR ARC")
    (9696 . "UPPER HALF CIRCLE")
    (9697 . "LOWER HALF CIRCLE")
    (9698 . "BLACK LOWER RIGHT TRIANGLE")
    (9699 . "BLACK LOWER LEFT TRIANGLE")
    (9700 . "BLACK UPPER LEFT TRIANGLE")
    (9701 . "BLACK UPPER RIGHT TRIANGLE")
    (9702 . "WHITE BULLET")
    (9703 . "SQUARE WITH LEFT HALF BLACK")
    (9704 . "SQUARE WITH RIGHT HALF BLACK")
    (9705 . "SQUARE WITH UPPER LEFT DIAGONAL HALF BLACK")
    (9706 . "SQUARE WITH LOWER RIGHT DIAGONAL HALF BLACK")
    (9707 . "WHITE SQUARE WITH VERTICAL BISECTING LINE")
    (9708 . "WHITE UP-POINTING TRIANGLE WITH DOT")
    (9709 . "UP-POINTING TRIANGLE WITH LEFT HALF BLACK")
    (9710 . "UP-POINTING TRIANGLE WITH RIGHT HALF BLACK")
    (9711 . "LARGE CIRCLE")
    (9728 . "BLACK SUN WITH RAYS")
    (9729 . "CLOUD")
    (9730 . "UMBRELLA")
    (9731 . "SNOWMAN")
    (9732 . "COMET")
    (9733 . "BLACK STAR")
    (9734 . "WHITE STAR")
    (9735 . "LIGHTNING")
    (9736 . "THUNDERSTORM")
    (9737 . "SUN")
    (9738 . "ASCENDING NODE")
    (9739 . "DESCENDING NODE")
    (9740 . "CONJUNCTION")
    (9741 . "OPPOSITION")
    (9742 . "BLACK TELEPHONE")
    (9743 . "WHITE TELEPHONE")
    (9744 . "BALLOT BOX")
    (9745 . "BALLOT BOX WITH CHECK")
    (9746 . "BALLOT BOX WITH X")
    (9747 . "SALTIRE")
    (9754 . "BLACK LEFT POINTING INDEX")
    (9755 . "BLACK RIGHT POINTING INDEX")
    (9756 . "WHITE LEFT POINTING INDEX")
    (9757 . "WHITE UP POINTING INDEX")
    (9758 . "WHITE RIGHT POINTING INDEX")
    (9759 . "WHITE DOWN POINTING INDEX")
    (9760 . "SKULL AND CROSSBONES")
    (9761 . "CAUTION SIGN")
    (9762 . "RADIOACTIVE SIGN")
    (9763 . "BIOHAZARD SIGN")
    (9764 . "CADUCEUS")
    (9765 . "ANKH")
    (9766 . "ORTHODOX CROSS")
    (9767 . "CHI RHO")
    (9768 . "CROSS OF LORRAINE")
    (9769 . "CROSS OF JERUSALEM")
    (9770 . "STAR AND CRESCENT")
    (9771 . "FARSI SYMBOL")
    (9772 . "ADI SHAKTI")
    (9773 . "HAMMER AND SICKLE")
    (9774 . "PEACE SYMBOL")
    (9775 . "YIN YANG")
    (9776 . "TRIGRAM FOR HEAVEN")
    (9777 . "TRIGRAM FOR LAKE")
    (9778 . "TRIGRAM FOR FIRE")
    (9779 . "TRIGRAM FOR THUNDER")
    (9780 . "TRIGRAM FOR WIND")
    (9781 . "TRIGRAM FOR WATER")
    (9782 . "TRIGRAM FOR MOUNTAIN")
    (9783 . "TRIGRAM FOR EARTH")
    (9784 . "WHEEL OF DHARMA")
    (9785 . "WHITE FROWNING FACE")
    (9786 . "WHITE SMILING FACE")
    (9787 . "BLACK SMILING FACE")
    (9788 . "WHITE SUN WITH RAYS")
    (9789 . "FIRST QUARTER MOON")
    (9790 . "LAST QUARTER MOON")
    (9791 . "MERCURY")
    (9792 . "FEMALE SIGN")
    (9793 . "EARTH")
    (9794 . "MALE SIGN")
    (9795 . "JUPITER")
    (9796 . "SATURN")
    (9797 . "URANUS")
    (9798 . "NEPTUNE")
    (9799 . "PLUTO")
    (9800 . "ARIES")
    (9801 . "TAURUS")
    (9802 . "GEMINI")
    (9803 . "CANCER")
    (9804 . "LEO")
    (9805 . "VIRGO")
    (9806 . "LIBRA")
    (9807 . "SCORPIUS")
    (9808 . "SAGITTARIUS")
    (9809 . "CAPRICORN")
    (9810 . "AQUARIUS")
    (9811 . "PISCES")
    (9812 . "WHITE CHESS KING")
    (9813 . "WHITE CHESS QUEEN")
    (9814 . "WHITE CHESS ROOK")
    (9815 . "WHITE CHESS BISHOP")
    (9816 . "WHITE CHESS KNIGHT")
    (9817 . "WHITE CHESS PAWN")
    (9818 . "BLACK CHESS KING")
    (9819 . "BLACK CHESS QUEEN")
    (9820 . "BLACK CHESS ROOK")
    (9821 . "BLACK CHESS BISHOP")
    (9822 . "BLACK CHESS KNIGHT")
    (9823 . "BLACK CHESS PAWN")
    (9824 . "BLACK SPADE SUIT")
    (9825 . "WHITE HEART SUIT")
    (9826 . "WHITE DIAMOND SUIT")
    (9827 . "BLACK CLUB SUIT")
    (9828 . "WHITE SPADE SUIT")
    (9829 . "BLACK HEART SUIT")
    (9830 . "BLACK DIAMOND SUIT")
    (9831 . "WHITE CLUB SUIT")
    (9832 . "HOT SPRINGS")
    (9833 . "QUARTER NOTE")
    (9834 . "EIGHTH NOTE")
    (9835 . "BEAMED EIGHTH NOTES")
    (9836 . "BEAMED SIXTEENTH NOTES")
    (9837 . "MUSIC FLAT SIGN")
    (9838 . "MUSIC NATURAL SIGN")
    (9839 . "MUSIC SHARP SIGN")
    (9985 . "UPPER BLADE SCISSORS")
    (9986 . "BLACK SCISSORS")
    (9987 . "LOWER BLADE SCISSORS")
    (9988 . "WHITE SCISSORS")
    (9990 . "TELEPHONE LOCATION SIGN")
    (9991 . "TAPE DRIVE")
    (9992 . "AIRPLANE")
    (9993 . "ENVELOPE")
    (9996 . "VICTORY HAND")
    (9997 . "WRITING HAND")
    (9998 . "LOWER RIGHT PENCIL")
    (9999 . "PENCIL")
    (10000 . "UPPER RIGHT PENCIL")
    (10001 . "WHITE NIB")
    (10002 . "BLACK NIB")
    (10003 . "CHECK MARK")
    (10004 . "HEAVY CHECK MARK")
    (10005 . "MULTIPLICATION X")
    (10006 . "HEAVY MULTIPLICATION X")
    (10007 . "BALLOT X")
    (10008 . "HEAVY BALLOT X")
    (10009 . "OUTLINED GREEK CROSS")
    (10010 . "HEAVY GREEK CROSS")
    (10011 . "OPEN CENTER CROSS|OPEN CENTRE CROSS")
    (10012 . "HEAVY OPEN CENTER CROSS|HEAVY OPEN CENTRE CROSS")
    (10013 . "LATIN CROSS")
    (10014 . "SHADOWED WHITE LATIN CROSS")
    (10015 . "OUTLINED LATIN CROSS")
    (10016 . "MALTESE CROSS")
    (10017 . "STAR OF DAVID")
    (10018 . "FOUR TEARDROP-SPOKED ASTERISK")
    (10019 . "FOUR BALLOON-SPOKED ASTERISK")
    (10020 . "HEAVY FOUR BALLOON-SPOKED ASTERISK")
    (10021 . "FOUR CLUB-SPOKED ASTERISK")
    (10022 . "BLACK FOUR-POINTED STAR")
    (10023 . "WHITE FOUR-POINTED STAR")
    (10025 . "STRESS OUTLINED WHITE STAR")
    (10026 . "CIRCLED WHITE STAR")
    (10027 . "OPEN CENTER BLACK STAR|OPEN CENTRE BLACK STAR")
    (10028 . "BLACK CENTER WHITE STAR|BLACK CENTRE WHITE STAR")
    (10029 . "OUTLINED BLACK STAR")
    (10030 . "HEAVY OUTLINED BLACK STAR")
    (10031 . "PINWHEEL STAR")
    (10032 . "SHADOWED WHITE STAR")
    (10033 . "HEAVY ASTERISK")
    (10034 . "OPEN CENTER ASTERISK|OPEN CENTRE ASTERISK")
    (10035 . "EIGHT SPOKED ASTERISK")
    (10036 . "EIGHT POINTED BLACK STAR")
    (10037 . "EIGHT POINTED PINWHEEL STAR")
    (10038 . "SIX POINTED BLACK STAR")
    (10039 . "EIGHT POINTED RECTILINEAR BLACK STAR")
    (10040 . "HEAVY EIGHT POINTED RECTILINEAR BLACK STAR")
    (10041 . "TWELVE POINTED BLACK STAR")
    (10042 . "SIXTEEN POINTED ASTERISK")
    (10043 . "TEARDROP-SPOKED ASTERISK")
    (10044 . "OPEN CENTER TEARDROP-SPOKED ASTERISK|OPEN CENTRE TEARDROP-SPOKED ASTERISK")
    (10045 . "HEAVY TEARDROP-SPOKED ASTERISK")
    (10046 . "SIX PETALLED BLACK AND WHITE FLORETTE")
    (10047 . "BLACK FLORETTE")
    (10048 . "WHITE FLORETTE")
    (10049 . "EIGHT PETALLED OUTLINED BLACK FLORETTE")
    (10050 . "CIRCLED OPEN CENTER EIGHT POINTED STAR|CIRCLED OPEN CENTRE EIGHT POINTED STAR")
    (10051 . "HEAVY TEARDROP-SPOKED PINWHEEL ASTERISK")
    (10052 . "SNOWFLAKE")
    (10053 . "TIGHT TRIFOLIATE SNOWFLAKE")
    (10054 . "HEAVY CHEVRON SNOWFLAKE")
    (10055 . "SPARKLE")
    (10056 . "HEAVY SPARKLE")
    (10057 . "BALLOON-SPOKED ASTERISK")
    (10058 . "EIGHT TEARDROP-SPOKED PROPELLER ASTERISK")
    (10059 . "HEAVY EIGHT TEARDROP-SPOKED PROPELLER ASTERISK")
    (10061 . "SHADOWED WHITE CIRCLE")
    (10063 . "LOWER RIGHT DROP-SHADOWED WHITE SQUARE")
    (10064 . "UPPER RIGHT DROP-SHADOWED WHITE SQUARE")
    (10065 . "LOWER RIGHT SHADOWED WHITE SQUARE")
    (10066 . "UPPER RIGHT SHADOWED WHITE SQUARE")
    (10070 . "BLACK DIAMOND MINUS WHITE X")
    (10072 . "LIGHT VERTICAL BAR")
    (10073 . "MEDIUM VERTICAL BAR")
    (10074 . "HEAVY VERTICAL BAR")
    (10075 . "HEAVY SINGLE TURNED COMMA QUOTATION MARK ORNAMENT")
    (10076 . "HEAVY SINGLE COMMA QUOTATION MARK ORNAMENT")
    (10077 . "HEAVY DOUBLE TURNED COMMA QUOTATION MARK ORNAMENT")
    (10078 . "HEAVY DOUBLE COMMA QUOTATION MARK ORNAMENT")
    (10081 . "CURVED STEM PARAGRAPH SIGN ORNAMENT")
    (10082 . "HEAVY EXCLAMATION MARK ORNAMENT")
    (10083 . "HEAVY HEART EXCLAMATION MARK ORNAMENT")
    (10084 . "HEAVY BLACK HEART")
    (10085 . "ROTATED HEAVY BLACK HEART BULLET")
    (10086 . "FLORAL HEART")
    (10087 . "ROTATED FLORAL HEART BULLET")
    (10102 . "DINGBAT NEGATIVE CIRCLED DIGIT ONE")
    (10103 . "DINGBAT NEGATIVE CIRCLED DIGIT TWO")
    (10104 . "DINGBAT NEGATIVE CIRCLED DIGIT THREE")
    (10105 . "DINGBAT NEGATIVE CIRCLED DIGIT FOUR")
    (10106 . "DINGBAT NEGATIVE CIRCLED DIGIT FIVE")
    (10107 . "DINGBAT NEGATIVE CIRCLED DIGIT SIX")
    (10108 . "DINGBAT NEGATIVE CIRCLED DIGIT SEVEN")
    (10109 . "DINGBAT NEGATIVE CIRCLED DIGIT EIGHT")
    (10110 . "DINGBAT NEGATIVE CIRCLED DIGIT NINE")
    (10111 . "DINGBAT NEGATIVE CIRCLED NUMBER TEN")
    (10112 . "DINGBAT CIRCLED SANS-SERIF DIGIT ONE")
    (10113 . "DINGBAT CIRCLED SANS-SERIF DIGIT TWO")
    (10114 . "DINGBAT CIRCLED SANS-SERIF DIGIT THREE")
    (10115 . "DINGBAT CIRCLED SANS-SERIF DIGIT FOUR")
    (10116 . "DINGBAT CIRCLED SANS-SERIF DIGIT FIVE")
    (10117 . "DINGBAT CIRCLED SANS-SERIF DIGIT SIX")
    (10118 . "DINGBAT CIRCLED SANS-SERIF DIGIT SEVEN")
    (10119 . "DINGBAT CIRCLED SANS-SERIF DIGIT EIGHT")
    (10120 . "DINGBAT CIRCLED SANS-SERIF DIGIT NINE")
    (10121 . "DINGBAT CIRCLED SANS-SERIF NUMBER TEN")
    (10122 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE")
    (10123 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO")
    (10124 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE")
    (10125 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR")
    (10126 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE")
    (10127 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX")
    (10128 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN")
    (10129 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT")
    (10130 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE")
    (10131 . "DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN")
    (10132 . "HEAVY WIDE-HEADED RIGHTWARDS ARROW")
    (10136 . "HEAVY SOUTH EAST ARROW")
    (10137 . "HEAVY RIGHTWARDS ARROW")
    (10138 . "HEAVY NORTH EAST ARROW")
    (10139 . "DRAFTING POINT RIGHTWARDS ARROW")
    (10140 . "HEAVY ROUND-TIPPED RIGHTWARDS ARROW")
    (10141 . "TRIANGLE-HEADED RIGHTWARDS ARROW")
    (10142 . "HEAVY TRIANGLE-HEADED RIGHTWARDS ARROW")
    (10143 . "DASHED TRIANGLE-HEADED RIGHTWARDS ARROW")
    (10144 . "HEAVY DASHED TRIANGLE-HEADED RIGHTWARDS ARROW")
    (10145 . "BLACK RIGHTWARDS ARROW")
    (10146 . "THREE-D TOP-LIGHTED RIGHTWARDS ARROWHEAD")
    (10147 . "THREE-D BOTTOM-LIGHTED RIGHTWARDS ARROWHEAD")
    (10148 . "BLACK RIGHTWARDS ARROWHEAD")
    (10149 . "HEAVY BLACK CURVED DOWNWARDS AND RIGHTWARDS ARROW")
    (10150 . "HEAVY BLACK CURVED UPWARDS AND RIGHTWARDS ARROW")
    (10151 . "SQUAT BLACK RIGHTWARDS ARROW")
    (10152 . "HEAVY CONCAVE-POINTED BLACK RIGHTWARDS ARROW")
    (10153 . "RIGHT-SHADED WHITE RIGHTWARDS ARROW")
    (10154 . "LEFT-SHADED WHITE RIGHTWARDS ARROW")
    (10155 . "BACK-TILTED SHADOWED WHITE RIGHTWARDS ARROW")
    (10156 . "FRONT-TILTED SHADOWED WHITE RIGHTWARDS ARROW")
    (10157 . "HEAVY LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW")
    (10158 . "HEAVY UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW")
    (10159 . "NOTCHED LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW")
    (10161 . "NOTCHED UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW")
    (10162 . "CIRCLED HEAVY WHITE RIGHTWARDS ARROW")
    (10163 . "WHITE-FEATHERED RIGHTWARDS ARROW")
    (10164 . "BLACK-FEATHERED SOUTH EAST ARROW")
    (10165 . "BLACK-FEATHERED RIGHTWARDS ARROW")
    (10166 . "BLACK-FEATHERED NORTH EAST ARROW")
    (10167 . "HEAVY BLACK-FEATHERED SOUTH EAST ARROW")
    (10168 . "HEAVY BLACK-FEATHERED RIGHTWARDS ARROW")
    (10169 . "HEAVY BLACK-FEATHERED NORTH EAST ARROW")
    (10170 . "TEARDROP-BARBED RIGHTWARDS ARROW")
    (10171 . "HEAVY TEARDROP-SHANKED RIGHTWARDS ARROW")
    (10172 . "WEDGE-TAILED RIGHTWARDS ARROW")
    (10173 . "HEAVY WEDGE-TAILED RIGHTWARDS ARROW")
    (10174 . "OPEN-OUTLINED RIGHTWARDS ARROW")
    (12288 . "IDEOGRAPHIC SPACE")
    (12289 . "IDEOGRAPHIC COMMA")
    (12290 . "IDEOGRAPHIC PERIOD|IDEOGRAPHIC FULL STOP")
    (12291 . "DITTO MARK")
    (12292 . "JAPANESE INDUSTRIAL STANDARD SYMBOL")
    (12293 . "IDEOGRAPHIC ITERATION MARK")
    (12294 . "IDEOGRAPHIC CLOSING MARK")
    (12295 . "IDEOGRAPHIC NUMBER ZERO")
    (12296 . "LEFT ANGLE BRACKET")
    (12297 . "RIGHT ANGLE BRACKET")
    (12298 . "LEFT DOUBLE ANGLE BRACKET")
    (12299 . "RIGHT DOUBLE ANGLE BRACKET")
    (12300 . "LEFT CORNER BRACKET")
    (12301 . "RIGHT CORNER BRACKET")
    (12302 . "LEFT WHITE CORNER BRACKET")
    (12303 . "RIGHT WHITE CORNER BRACKET")
    (12304 . "LEFT BLACK LENTICULAR BRACKET")
    (12305 . "RIGHT BLACK LENTICULAR BRACKET")
    (12306 . "POSTAL MARK")
    (12307 . "GETA MARK")
    (12308 . "LEFT TORTOISE SHELL BRACKET")
    (12309 . "RIGHT TORTOISE SHELL BRACKET")
    (12310 . "LEFT WHITE LENTICULAR BRACKET")
    (12311 . "RIGHT WHITE LENTICULAR BRACKET")
    (12312 . "LEFT WHITE TORTOISE SHELL BRACKET")
    (12313 . "RIGHT WHITE TORTOISE SHELL BRACKET")
    (12314 . "LEFT WHITE SQUARE BRACKET")
    (12315 . "RIGHT WHITE SQUARE BRACKET")
    (12316 . "WAVE DASH")
    (12317 . "REVERSED DOUBLE PRIME QUOTATION MARK")
    (12318 . "DOUBLE PRIME QUOTATION MARK")
    (12319 . "LOW DOUBLE PRIME QUOTATION MARK")
    (12320 . "POSTAL MARK FACE")
    (12321 . "HANGZHOU NUMERAL ONE")
    (12322 . "HANGZHOU NUMERAL TWO")
    (12323 . "HANGZHOU NUMERAL THREE")
    (12324 . "HANGZHOU NUMERAL FOUR")
    (12325 . "HANGZHOU NUMERAL FIVE")
    (12326 . "HANGZHOU NUMERAL SIX")
    (12327 . "HANGZHOU NUMERAL SEVEN")
    (12328 . "HANGZHOU NUMERAL EIGHT")
    (12329 . "HANGZHOU NUMERAL NINE")
    (12330 . "IDEOGRAPHIC LEVEL TONE MARK")
    (12331 . "IDEOGRAPHIC RISING TONE MARK")
    (12332 . "IDEOGRAPHIC DEPARTING TONE MARK")
    (12333 . "IDEOGRAPHIC ENTERING TONE MARK")
    (12334 . "HANGUL SINGLE DOT TONE MARK")
    (12335 . "HANGUL DOUBLE DOT TONE MARK")
    (12336 . "WAVY DASH")
    (12337 . "VERTICAL KANA REPEAT MARK")
    (12338 . "VERTICAL KANA REPEAT WITH VOICED SOUND MARK")
    (12339 . "VERTICAL KANA REPEAT MARK UPPER HALF")
    (12340 . "VERTICAL KANA REPEAT WITH VOICED SOUND MARK UPPER HALF")
    (12341 . "VERTICAL KANA REPEAT MARK LOWER HALF")
    (12342 . "CIRCLED POSTAL MARK")
    (12343 . "IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL")
    (12351 . "IDEOGRAPHIC HALF FILL SPACE")
    (12353 . "HIRAGANA LETTER SMALL A")
    (12354 . "HIRAGANA LETTER A")
    (12355 . "HIRAGANA LETTER SMALL I")
    (12356 . "HIRAGANA LETTER I")
    (12357 . "HIRAGANA LETTER SMALL U")
    (12358 . "HIRAGANA LETTER U")
    (12359 . "HIRAGANA LETTER SMALL E")
    (12360 . "HIRAGANA LETTER E")
    (12361 . "HIRAGANA LETTER SMALL O")
    (12362 . "HIRAGANA LETTER O")
    (12363 . "HIRAGANA LETTER KA")
    (12364 . "HIRAGANA LETTER GA")
    (12365 . "HIRAGANA LETTER KI")
    (12366 . "HIRAGANA LETTER GI")
    (12367 . "HIRAGANA LETTER KU")
    (12368 . "HIRAGANA LETTER GU")
    (12369 . "HIRAGANA LETTER KE")
    (12370 . "HIRAGANA LETTER GE")
    (12371 . "HIRAGANA LETTER KO")
    (12372 . "HIRAGANA LETTER GO")
    (12373 . "HIRAGANA LETTER SA")
    (12374 . "HIRAGANA LETTER ZA")
    (12375 . "HIRAGANA LETTER SI")
    (12376 . "HIRAGANA LETTER ZI")
    (12377 . "HIRAGANA LETTER SU")
    (12378 . "HIRAGANA LETTER ZU")
    (12379 . "HIRAGANA LETTER SE")
    (12380 . "HIRAGANA LETTER ZE")
    (12381 . "HIRAGANA LETTER SO")
    (12382 . "HIRAGANA LETTER ZO")
    (12383 . "HIRAGANA LETTER TA")
    (12384 . "HIRAGANA LETTER DA")
    (12385 . "HIRAGANA LETTER TI")
    (12386 . "HIRAGANA LETTER DI")
    (12387 . "HIRAGANA LETTER SMALL TU")
    (12388 . "HIRAGANA LETTER TU")
    (12389 . "HIRAGANA LETTER DU")
    (12390 . "HIRAGANA LETTER TE")
    (12391 . "HIRAGANA LETTER DE")
    (12392 . "HIRAGANA LETTER TO")
    (12393 . "HIRAGANA LETTER DO")
    (12394 . "HIRAGANA LETTER NA")
    (12395 . "HIRAGANA LETTER NI")
    (12396 . "HIRAGANA LETTER NU")
    (12397 . "HIRAGANA LETTER NE")
    (12398 . "HIRAGANA LETTER NO")
    (12399 . "HIRAGANA LETTER HA")
    (12400 . "HIRAGANA LETTER BA")
    (12401 . "HIRAGANA LETTER PA")
    (12402 . "HIRAGANA LETTER HI")
    (12403 . "HIRAGANA LETTER BI")
    (12404 . "HIRAGANA LETTER PI")
    (12405 . "HIRAGANA LETTER HU")
    (12406 . "HIRAGANA LETTER BU")
    (12407 . "HIRAGANA LETTER PU")
    (12408 . "HIRAGANA LETTER HE")
    (12409 . "HIRAGANA LETTER BE")
    (12410 . "HIRAGANA LETTER PE")
    (12411 . "HIRAGANA LETTER HO")
    (12412 . "HIRAGANA LETTER BO")
    (12413 . "HIRAGANA LETTER PO")
    (12414 . "HIRAGANA LETTER MA")
    (12415 . "HIRAGANA LETTER MI")
    (12416 . "HIRAGANA LETTER MU")
    (12417 . "HIRAGANA LETTER ME")
    (12418 . "HIRAGANA LETTER MO")
    (12419 . "HIRAGANA LETTER SMALL YA")
    (12420 . "HIRAGANA LETTER YA")
    (12421 . "HIRAGANA LETTER SMALL YU")
    (12422 . "HIRAGANA LETTER YU")
    (12423 . "HIRAGANA LETTER SMALL YO")
    (12424 . "HIRAGANA LETTER YO")
    (12425 . "HIRAGANA LETTER RA")
    (12426 . "HIRAGANA LETTER RI")
    (12427 . "HIRAGANA LETTER RU")
    (12428 . "HIRAGANA LETTER RE")
    (12429 . "HIRAGANA LETTER RO")
    (12430 . "HIRAGANA LETTER SMALL WA")
    (12431 . "HIRAGANA LETTER WA")
    (12432 . "HIRAGANA LETTER WI")
    (12433 . "HIRAGANA LETTER WE")
    (12434 . "HIRAGANA LETTER WO")
    (12435 . "HIRAGANA LETTER N")
    (12436 . "HIRAGANA LETTER VU")
    (12441 . "COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK")
    (12442 . "COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK")
    (12443 . "KATAKANA-HIRAGANA VOICED SOUND MARK")
    (12444 . "KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK")
    (12445 . "HIRAGANA ITERATION MARK")
    (12446 . "HIRAGANA VOICED ITERATION MARK")
    (12449 . "KATAKANA LETTER SMALL A")
    (12450 . "KATAKANA LETTER A")
    (12451 . "KATAKANA LETTER SMALL I")
    (12452 . "KATAKANA LETTER I")
    (12453 . "KATAKANA LETTER SMALL U")
    (12454 . "KATAKANA LETTER U")
    (12455 . "KATAKANA LETTER SMALL E")
    (12456 . "KATAKANA LETTER E")
    (12457 . "KATAKANA LETTER SMALL O")
    (12458 . "KATAKANA LETTER O")
    (12459 . "KATAKANA LETTER KA")
    (12460 . "KATAKANA LETTER GA")
    (12461 . "KATAKANA LETTER KI")
    (12462 . "KATAKANA LETTER GI")
    (12463 . "KATAKANA LETTER KU")
    (12464 . "KATAKANA LETTER GU")
    (12465 . "KATAKANA LETTER KE")
    (12466 . "KATAKANA LETTER GE")
    (12467 . "KATAKANA LETTER KO")
    (12468 . "KATAKANA LETTER GO")
    (12469 . "KATAKANA LETTER SA")
    (12470 . "KATAKANA LETTER ZA")
    (12471 . "KATAKANA LETTER SI")
    (12472 . "KATAKANA LETTER ZI")
    (12473 . "KATAKANA LETTER SU")
    (12474 . "KATAKANA LETTER ZU")
    (12475 . "KATAKANA LETTER SE")
    (12476 . "KATAKANA LETTER ZE")
    (12477 . "KATAKANA LETTER SO")
    (12478 . "KATAKANA LETTER ZO")
    (12479 . "KATAKANA LETTER TA")
    (12480 . "KATAKANA LETTER DA")
    (12481 . "KATAKANA LETTER TI")
    (12482 . "KATAKANA LETTER DI")
    (12483 . "KATAKANA LETTER SMALL TU")
    (12484 . "KATAKANA LETTER TU")
    (12485 . "KATAKANA LETTER DU")
    (12486 . "KATAKANA LETTER TE")
    (12487 . "KATAKANA LETTER DE")
    (12488 . "KATAKANA LETTER TO")
    (12489 . "KATAKANA LETTER DO")
    (12490 . "KATAKANA LETTER NA")
    (12491 . "KATAKANA LETTER NI")
    (12492 . "KATAKANA LETTER NU")
    (12493 . "KATAKANA LETTER NE")
    (12494 . "KATAKANA LETTER NO")
    (12495 . "KATAKANA LETTER HA")
    (12496 . "KATAKANA LETTER BA")
    (12497 . "KATAKANA LETTER PA")
    (12498 . "KATAKANA LETTER HI")
    (12499 . "KATAKANA LETTER BI")
    (12500 . "KATAKANA LETTER PI")
    (12501 . "KATAKANA LETTER HU")
    (12502 . "KATAKANA LETTER BU")
    (12503 . "KATAKANA LETTER PU")
    (12504 . "KATAKANA LETTER HE")
    (12505 . "KATAKANA LETTER BE")
    (12506 . "KATAKANA LETTER PE")
    (12507 . "KATAKANA LETTER HO")
    (12508 . "KATAKANA LETTER BO")
    (12509 . "KATAKANA LETTER PO")
    (12510 . "KATAKANA LETTER MA")
    (12511 . "KATAKANA LETTER MI")
    (12512 . "KATAKANA LETTER MU")
    (12513 . "KATAKANA LETTER ME")
    (12514 . "KATAKANA LETTER MO")
    (12515 . "KATAKANA LETTER SMALL YA")
    (12516 . "KATAKANA LETTER YA")
    (12517 . "KATAKANA LETTER SMALL YU")
    (12518 . "KATAKANA LETTER YU")
    (12519 . "KATAKANA LETTER SMALL YO")
    (12520 . "KATAKANA LETTER YO")
    (12521 . "KATAKANA LETTER RA")
    (12522 . "KATAKANA LETTER RI")
    (12523 . "KATAKANA LETTER RU")
    (12524 . "KATAKANA LETTER RE")
    (12525 . "KATAKANA LETTER RO")
    (12526 . "KATAKANA LETTER SMALL WA")
    (12527 . "KATAKANA LETTER WA")
    (12528 . "KATAKANA LETTER WI")
    (12529 . "KATAKANA LETTER WE")
    (12530 . "KATAKANA LETTER WO")
    (12531 . "KATAKANA LETTER N")
    (12532 . "KATAKANA LETTER VU")
    (12533 . "KATAKANA LETTER SMALL KA")
    (12534 . "KATAKANA LETTER SMALL KE")
    (12535 . "KATAKANA LETTER VA")
    (12536 . "KATAKANA LETTER VI")
    (12537 . "KATAKANA LETTER VE")
    (12538 . "KATAKANA LETTER VO")
    (12539 . "KATAKANA MIDDLE DOT")
    (12540 . "KATAKANA-HIRAGANA PROLONGED SOUND MARK")
    (12541 . "KATAKANA ITERATION MARK")
    (12542 . "KATAKANA VOICED ITERATION MARK")
    (12549 . "BOPOMOFO LETTER B")
    (12550 . "BOPOMOFO LETTER P")
    (12551 . "BOPOMOFO LETTER M")
    (12552 . "BOPOMOFO LETTER F")
    (12553 . "BOPOMOFO LETTER D")
    (12554 . "BOPOMOFO LETTER T")
    (12555 . "BOPOMOFO LETTER N")
    (12556 . "BOPOMOFO LETTER L")
    (12557 . "BOPOMOFO LETTER G")
    (12558 . "BOPOMOFO LETTER K")
    (12559 . "BOPOMOFO LETTER H")
    (12560 . "BOPOMOFO LETTER J")
    (12561 . "BOPOMOFO LETTER Q")
    (12562 . "BOPOMOFO LETTER X")
    (12563 . "BOPOMOFO LETTER ZH")
    (12564 . "BOPOMOFO LETTER CH")
    (12565 . "BOPOMOFO LETTER SH")
    (12566 . "BOPOMOFO LETTER R")
    (12567 . "BOPOMOFO LETTER Z")
    (12568 . "BOPOMOFO LETTER C")
    (12569 . "BOPOMOFO LETTER S")
    (12570 . "BOPOMOFO LETTER A")
    (12571 . "BOPOMOFO LETTER O")
    (12572 . "BOPOMOFO LETTER E")
    (12573 . "BOPOMOFO LETTER EH")
    (12574 . "BOPOMOFO LETTER AI")
    (12575 . "BOPOMOFO LETTER EI")
    (12576 . "BOPOMOFO LETTER AU")
    (12577 . "BOPOMOFO LETTER OU")
    (12578 . "BOPOMOFO LETTER AN")
    (12579 . "BOPOMOFO LETTER EN")
    (12580 . "BOPOMOFO LETTER ANG")
    (12581 . "BOPOMOFO LETTER ENG")
    (12582 . "BOPOMOFO LETTER ER")
    (12583 . "BOPOMOFO LETTER I")
    (12584 . "BOPOMOFO LETTER U")
    (12585 . "BOPOMOFO LETTER IU")
    (12586 . "BOPOMOFO LETTER V")
    (12587 . "BOPOMOFO LETTER NG")
    (12588 . "BOPOMOFO LETTER GN")
    (12593 . "HANGUL LETTER KIYEOK")
    (12594 . "HANGUL LETTER SSANGKIYEOK")
    (12595 . "HANGUL LETTER KIYEOK-SIOS")
    (12596 . "HANGUL LETTER NIEUN")
    (12597 . "HANGUL LETTER NIEUN-CIEUC")
    (12598 . "HANGUL LETTER NIEUN-HIEUH")
    (12599 . "HANGUL LETTER TIKEUT")
    (12600 . "HANGUL LETTER SSANGTIKEUT")
    (12601 . "HANGUL LETTER RIEUL")
    (12602 . "HANGUL LETTER RIEUL-KIYEOK")
    (12603 . "HANGUL LETTER RIEUL-MIEUM")
    (12604 . "HANGUL LETTER RIEUL-PIEUP")
    (12605 . "HANGUL LETTER RIEUL-SIOS")
    (12606 . "HANGUL LETTER RIEUL-THIEUTH")
    (12607 . "HANGUL LETTER RIEUL-PHIEUPH")
    (12608 . "HANGUL LETTER RIEUL-HIEUH")
    (12609 . "HANGUL LETTER MIEUM")
    (12610 . "HANGUL LETTER PIEUP")
    (12611 . "HANGUL LETTER SSANGPIEUP")
    (12612 . "HANGUL LETTER PIEUP-SIOS")
    (12613 . "HANGUL LETTER SIOS")
    (12614 . "HANGUL LETTER SSANGSIOS")
    (12615 . "HANGUL LETTER IEUNG")
    (12616 . "HANGUL LETTER CIEUC")
    (12617 . "HANGUL LETTER SSANGCIEUC")
    (12618 . "HANGUL LETTER CHIEUCH")
    (12619 . "HANGUL LETTER KHIEUKH")
    (12620 . "HANGUL LETTER THIEUTH")
    (12621 . "HANGUL LETTER PHIEUPH")
    (12622 . "HANGUL LETTER HIEUH")
    (12623 . "HANGUL LETTER A")
    (12624 . "HANGUL LETTER AE")
    (12625 . "HANGUL LETTER YA")
    (12626 . "HANGUL LETTER YAE")
    (12627 . "HANGUL LETTER EO")
    (12628 . "HANGUL LETTER E")
    (12629 . "HANGUL LETTER YEO")
    (12630 . "HANGUL LETTER YE")
    (12631 . "HANGUL LETTER O")
    (12632 . "HANGUL LETTER WA")
    (12633 . "HANGUL LETTER WAE")
    (12634 . "HANGUL LETTER OE")
    (12635 . "HANGUL LETTER YO")
    (12636 . "HANGUL LETTER U")
    (12637 . "HANGUL LETTER WEO")
    (12638 . "HANGUL LETTER WE")
    (12639 . "HANGUL LETTER WI")
    (12640 . "HANGUL LETTER YU")
    (12641 . "HANGUL LETTER EU")
    (12642 . "HANGUL LETTER YI")
    (12643 . "HANGUL LETTER I")
    (12644 . "HANGUL FILLER")
    (12645 . "HANGUL LETTER SSANGNIEUN")
    (12646 . "HANGUL LETTER NIEUN-TIKEUT")
    (12647 . "HANGUL LETTER NIEUN-SIOS")
    (12648 . "HANGUL LETTER NIEUN-PANSIOS")
    (12649 . "HANGUL LETTER RIEUL-KIYEOK-SIOS")
    (12650 . "HANGUL LETTER RIEUL-TIKEUT")
    (12651 . "HANGUL LETTER RIEUL-PIEUP-SIOS")
    (12652 . "HANGUL LETTER RIEUL-PANSIOS")
    (12653 . "HANGUL LETTER RIEUL-YEORINHIEUH")
    (12654 . "HANGUL LETTER MIEUM-PIEUP")
    (12655 . "HANGUL LETTER MIEUM-SIOS")
    (12656 . "HANGUL LETTER MIEUM-PANSIOS")
    (12657 . "HANGUL LETTER KAPYEOUNMIEUM")
    (12658 . "HANGUL LETTER PIEUP-KIYEOK")
    (12659 . "HANGUL LETTER PIEUP-TIKEUT")
    (12660 . "HANGUL LETTER PIEUP-SIOS-KIYEOK")
    (12661 . "HANGUL LETTER PIEUP-SIOS-TIKEUT")
    (12662 . "HANGUL LETTER PIEUP-CIEUC")
    (12663 . "HANGUL LETTER PIEUP-THIEUTH")
    (12664 . "HANGUL LETTER KAPYEOUNPIEUP")
    (12665 . "HANGUL LETTER KAPYEOUNSSANGPIEUP")
    (12666 . "HANGUL LETTER SIOS-KIYEOK")
    (12667 . "HANGUL LETTER SIOS-NIEUN")
    (12668 . "HANGUL LETTER SIOS-TIKEUT")
    (12669 . "HANGUL LETTER SIOS-PIEUP")
    (12670 . "HANGUL LETTER SIOS-CIEUC")
    (12671 . "HANGUL LETTER PANSIOS")
    (12672 . "HANGUL LETTER SSANGIEUNG")
    (12673 . "HANGUL LETTER YESIEUNG")
    (12674 . "HANGUL LETTER YESIEUNG-SIOS")
    (12675 . "HANGUL LETTER YESIEUNG-PANSIOS")
    (12676 . "HANGUL LETTER KAPYEOUNPHIEUPH")
    (12677 . "HANGUL LETTER SSANGHIEUH")
    (12678 . "HANGUL LETTER YEORINHIEUH")
    (12679 . "HANGUL LETTER YO-YA")
    (12680 . "HANGUL LETTER YO-YAE")
    (12681 . "HANGUL LETTER YO-I")
    (12682 . "HANGUL LETTER YU-YEO")
    (12683 . "HANGUL LETTER YU-YE")
    (12684 . "HANGUL LETTER YU-I")
    (12685 . "HANGUL LETTER ARAEA")
    (12686 . "HANGUL LETTER ARAEAE")
    (12688 . "IDEOGRAPHIC ANNOTATION LINKING MARK")
    (12689 . "IDEOGRAPHIC ANNOTATION REVERSE MARK")
    (12690 . "IDEOGRAPHIC ANNOTATION ONE MARK")
    (12691 . "IDEOGRAPHIC ANNOTATION TWO MARK")
    (12692 . "IDEOGRAPHIC ANNOTATION THREE MARK")
    (12693 . "IDEOGRAPHIC ANNOTATION FOUR MARK")
    (12694 . "IDEOGRAPHIC ANNOTATION TOP MARK")
    (12695 . "IDEOGRAPHIC ANNOTATION MIDDLE MARK")
    (12696 . "IDEOGRAPHIC ANNOTATION BOTTOM MARK")
    (12697 . "IDEOGRAPHIC ANNOTATION FIRST MARK")
    (12698 . "IDEOGRAPHIC ANNOTATION SECOND MARK")
    (12699 . "IDEOGRAPHIC ANNOTATION THIRD MARK")
    (12700 . "IDEOGRAPHIC ANNOTATION FOURTH MARK")
    (12701 . "IDEOGRAPHIC ANNOTATION HEAVEN MARK")
    (12702 . "IDEOGRAPHIC ANNOTATION EARTH MARK")
    (12703 . "IDEOGRAPHIC ANNOTATION MAN MARK")
    (12800 . "PARENTHESIZED HANGUL KIYEOK")
    (12801 . "PARENTHESIZED HANGUL NIEUN")
    (12802 . "PARENTHESIZED HANGUL TIKEUT")
    (12803 . "PARENTHESIZED HANGUL RIEUL")
    (12804 . "PARENTHESIZED HANGUL MIEUM")
    (12805 . "PARENTHESIZED HANGUL PIEUP")
    (12806 . "PARENTHESIZED HANGUL SIOS")
    (12807 . "PARENTHESIZED HANGUL IEUNG")
    (12808 . "PARENTHESIZED HANGUL CIEUC")
    (12809 . "PARENTHESIZED HANGUL CHIEUCH")
    (12810 . "PARENTHESIZED HANGUL KHIEUKH")
    (12811 . "PARENTHESIZED HANGUL THIEUTH")
    (12812 . "PARENTHESIZED HANGUL PHIEUPH")
    (12813 . "PARENTHESIZED HANGUL HIEUH")
    (12814 . "PARENTHESIZED HANGUL KIYEOK A")
    (12815 . "PARENTHESIZED HANGUL NIEUN A")
    (12816 . "PARENTHESIZED HANGUL TIKEUT A")
    (12817 . "PARENTHESIZED HANGUL RIEUL A")
    (12818 . "PARENTHESIZED HANGUL MIEUM A")
    (12819 . "PARENTHESIZED HANGUL PIEUP A")
    (12820 . "PARENTHESIZED HANGUL SIOS A")
    (12821 . "PARENTHESIZED HANGUL IEUNG A")
    (12822 . "PARENTHESIZED HANGUL CIEUC A")
    (12823 . "PARENTHESIZED HANGUL CHIEUCH A")
    (12824 . "PARENTHESIZED HANGUL KHIEUKH A")
    (12825 . "PARENTHESIZED HANGUL THIEUTH A")
    (12826 . "PARENTHESIZED HANGUL PHIEUPH A")
    (12827 . "PARENTHESIZED HANGUL HIEUH A")
    (12828 . "PARENTHESIZED HANGUL CIEUC U")
    (12832 . "PARENTHESIZED IDEOGRAPH ONE")
    (12833 . "PARENTHESIZED IDEOGRAPH TWO")
    (12834 . "PARENTHESIZED IDEOGRAPH THREE")
    (12835 . "PARENTHESIZED IDEOGRAPH FOUR")
    (12836 . "PARENTHESIZED IDEOGRAPH FIVE")
    (12837 . "PARENTHESIZED IDEOGRAPH SIX")
    (12838 . "PARENTHESIZED IDEOGRAPH SEVEN")
    (12839 . "PARENTHESIZED IDEOGRAPH EIGHT")
    (12840 . "PARENTHESIZED IDEOGRAPH NINE")
    (12841 . "PARENTHESIZED IDEOGRAPH TEN")
    (12842 . "PARENTHESIZED IDEOGRAPH MOON")
    (12843 . "PARENTHESIZED IDEOGRAPH FIRE")
    (12844 . "PARENTHESIZED IDEOGRAPH WATER")
    (12845 . "PARENTHESIZED IDEOGRAPH WOOD")
    (12846 . "PARENTHESIZED IDEOGRAPH METAL")
    (12847 . "PARENTHESIZED IDEOGRAPH EARTH")
    (12848 . "PARENTHESIZED IDEOGRAPH SUN")
    (12849 . "PARENTHESIZED IDEOGRAPH STOCK")
    (12850 . "PARENTHESIZED IDEOGRAPH HAVE")
    (12851 . "PARENTHESIZED IDEOGRAPH SOCIETY")
    (12852 . "PARENTHESIZED IDEOGRAPH NAME")
    (12853 . "PARENTHESIZED IDEOGRAPH SPECIAL")
    (12854 . "PARENTHESIZED IDEOGRAPH FINANCIAL")
    (12855 . "PARENTHESIZED IDEOGRAPH CONGRATULATION")
    (12856 . "PARENTHESIZED IDEOGRAPH LABOR")
    (12857 . "PARENTHESIZED IDEOGRAPH REPRESENT")
    (12858 . "PARENTHESIZED IDEOGRAPH CALL")
    (12859 . "PARENTHESIZED IDEOGRAPH STUDY")
    (12860 . "PARENTHESIZED IDEOGRAPH SUPERVISE")
    (12861 . "PARENTHESIZED IDEOGRAPH ENTERPRISE")
    (12862 . "PARENTHESIZED IDEOGRAPH RESOURCE")
    (12863 . "PARENTHESIZED IDEOGRAPH ALLIANCE")
    (12864 . "PARENTHESIZED IDEOGRAPH FESTIVAL")
    (12865 . "PARENTHESIZED IDEOGRAPH REST")
    (12866 . "PARENTHESIZED IDEOGRAPH SELF")
    (12867 . "PARENTHESIZED IDEOGRAPH REACH")
    (12896 . "CIRCLED HANGUL KIYEOK")
    (12897 . "CIRCLED HANGUL NIEUN")
    (12898 . "CIRCLED HANGUL TIKEUT")
    (12899 . "CIRCLED HANGUL RIEUL")
    (12900 . "CIRCLED HANGUL MIEUM")
    (12901 . "CIRCLED HANGUL PIEUP")
    (12902 . "CIRCLED HANGUL SIOS")
    (12903 . "CIRCLED HANGUL IEUNG")
    (12904 . "CIRCLED HANGUL CIEUC")
    (12905 . "CIRCLED HANGUL CHIEUCH")
    (12906 . "CIRCLED HANGUL KHIEUKH")
    (12907 . "CIRCLED HANGUL THIEUTH")
    (12908 . "CIRCLED HANGUL PHIEUPH")
    (12909 . "CIRCLED HANGUL HIEUH")
    (12910 . "CIRCLED HANGUL KIYEOK A")
    (12911 . "CIRCLED HANGUL NIEUN A")
    (12912 . "CIRCLED HANGUL TIKEUT A")
    (12913 . "CIRCLED HANGUL RIEUL A")
    (12914 . "CIRCLED HANGUL MIEUM A")
    (12915 . "CIRCLED HANGUL PIEUP A")
    (12916 . "CIRCLED HANGUL SIOS A")
    (12917 . "CIRCLED HANGUL IEUNG A")
    (12918 . "CIRCLED HANGUL CIEUC A")
    (12919 . "CIRCLED HANGUL CHIEUCH A")
    (12920 . "CIRCLED HANGUL KHIEUKH A")
    (12921 . "CIRCLED HANGUL THIEUTH A")
    (12922 . "CIRCLED HANGUL PHIEUPH A")
    (12923 . "CIRCLED HANGUL HIEUH A")
    (12927 . "KOREAN STANDARD SYMBOL")
    (12928 . "CIRCLED IDEOGRAPH ONE")
    (12929 . "CIRCLED IDEOGRAPH TWO")
    (12930 . "CIRCLED IDEOGRAPH THREE")
    (12931 . "CIRCLED IDEOGRAPH FOUR")
    (12932 . "CIRCLED IDEOGRAPH FIVE")
    (12933 . "CIRCLED IDEOGRAPH SIX")
    (12934 . "CIRCLED IDEOGRAPH SEVEN")
    (12935 . "CIRCLED IDEOGRAPH EIGHT")
    (12936 . "CIRCLED IDEOGRAPH NINE")
    (12937 . "CIRCLED IDEOGRAPH TEN")
    (12938 . "CIRCLED IDEOGRAPH MOON")
    (12939 . "CIRCLED IDEOGRAPH FIRE")
    (12940 . "CIRCLED IDEOGRAPH WATER")
    (12941 . "CIRCLED IDEOGRAPH WOOD")
    (12942 . "CIRCLED IDEOGRAPH METAL")
    (12943 . "CIRCLED IDEOGRAPH EARTH")
    (12944 . "CIRCLED IDEOGRAPH SUN")
    (12945 . "CIRCLED IDEOGRAPH STOCK")
    (12946 . "CIRCLED IDEOGRAPH HAVE")
    (12947 . "CIRCLED IDEOGRAPH SOCIETY")
    (12948 . "CIRCLED IDEOGRAPH NAME")
    (12949 . "CIRCLED IDEOGRAPH SPECIAL")
    (12950 . "CIRCLED IDEOGRAPH FINANCIAL")
    (12951 . "CIRCLED IDEOGRAPH CONGRATULATION")
    (12952 . "CIRCLED IDEOGRAPH LABOR")
    (12953 . "CIRCLED IDEOGRAPH SECRET")
    (12954 . "CIRCLED IDEOGRAPH MALE")
    (12955 . "CIRCLED IDEOGRAPH FEMALE")
    (12956 . "CIRCLED IDEOGRAPH SUITABLE")
    (12957 . "CIRCLED IDEOGRAPH EXCELLENT")
    (12958 . "CIRCLED IDEOGRAPH PRINT")
    (12959 . "CIRCLED IDEOGRAPH ATTENTION")
    (12960 . "CIRCLED IDEOGRAPH ITEM")
    (12961 . "CIRCLED IDEOGRAPH REST")
    (12962 . "CIRCLED IDEOGRAPH COPY")
    (12963 . "CIRCLED IDEOGRAPH CORRECT")
    (12964 . "CIRCLED IDEOGRAPH HIGH")
    (12965 . "CIRCLED IDEOGRAPH CENTER|CIRCLED IDEOGRAPH CENTRE")
    (12966 . "CIRCLED IDEOGRAPH LOW")
    (12967 . "CIRCLED IDEOGRAPH LEFT")
    (12968 . "CIRCLED IDEOGRAPH RIGHT")
    (12969 . "CIRCLED IDEOGRAPH MEDICINE")
    (12970 . "CIRCLED IDEOGRAPH RELIGION")
    (12971 . "CIRCLED IDEOGRAPH STUDY")
    (12972 . "CIRCLED IDEOGRAPH SUPERVISE")
    (12973 . "CIRCLED IDEOGRAPH ENTERPRISE")
    (12974 . "CIRCLED IDEOGRAPH RESOURCE")
    (12975 . "CIRCLED IDEOGRAPH ALLIANCE")
    (12976 . "CIRCLED IDEOGRAPH NIGHT")
    (12992 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY")
    (12993 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR FEBRUARY")
    (12994 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR MARCH")
    (12995 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR APRIL")
    (12996 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR MAY")
    (12997 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JUNE")
    (12998 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR JULY")
    (12999 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR AUGUST")
    (13000 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR SEPTEMBER")
    (13001 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR OCTOBER")
    (13002 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR NOVEMBER")
    (13003 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DECEMBER")
    (13008 . "CIRCLED KATAKANA A")
    (13009 . "CIRCLED KATAKANA I")
    (13010 . "CIRCLED KATAKANA U")
    (13011 . "CIRCLED KATAKANA E")
    (13012 . "CIRCLED KATAKANA O")
    (13013 . "CIRCLED KATAKANA KA")
    (13014 . "CIRCLED KATAKANA KI")
    (13015 . "CIRCLED KATAKANA KU")
    (13016 . "CIRCLED KATAKANA KE")
    (13017 . "CIRCLED KATAKANA KO")
    (13018 . "CIRCLED KATAKANA SA")
    (13019 . "CIRCLED KATAKANA SI")
    (13020 . "CIRCLED KATAKANA SU")
    (13021 . "CIRCLED KATAKANA SE")
    (13022 . "CIRCLED KATAKANA SO")
    (13023 . "CIRCLED KATAKANA TA")
    (13024 . "CIRCLED KATAKANA TI")
    (13025 . "CIRCLED KATAKANA TU")
    (13026 . "CIRCLED KATAKANA TE")
    (13027 . "CIRCLED KATAKANA TO")
    (13028 . "CIRCLED KATAKANA NA")
    (13029 . "CIRCLED KATAKANA NI")
    (13030 . "CIRCLED KATAKANA NU")
    (13031 . "CIRCLED KATAKANA NE")
    (13032 . "CIRCLED KATAKANA NO")
    (13033 . "CIRCLED KATAKANA HA")
    (13034 . "CIRCLED KATAKANA HI")
    (13035 . "CIRCLED KATAKANA HU")
    (13036 . "CIRCLED KATAKANA HE")
    (13037 . "CIRCLED KATAKANA HO")
    (13038 . "CIRCLED KATAKANA MA")
    (13039 . "CIRCLED KATAKANA MI")
    (13040 . "CIRCLED KATAKANA MU")
    (13041 . "CIRCLED KATAKANA ME")
    (13042 . "CIRCLED KATAKANA MO")
    (13043 . "CIRCLED KATAKANA YA")
    (13044 . "CIRCLED KATAKANA YU")
    (13045 . "CIRCLED KATAKANA YO")
    (13046 . "CIRCLED KATAKANA RA")
    (13047 . "CIRCLED KATAKANA RI")
    (13048 . "CIRCLED KATAKANA RU")
    (13049 . "CIRCLED KATAKANA RE")
    (13050 . "CIRCLED KATAKANA RO")
    (13051 . "CIRCLED KATAKANA WA")
    (13052 . "CIRCLED KATAKANA WI")
    (13053 . "CIRCLED KATAKANA WE")
    (13054 . "CIRCLED KATAKANA WO")
    (13056 . "SQUARE APAATO")
    (13057 . "SQUARE ARUHUA")
    (13058 . "SQUARE ANPEA")
    (13059 . "SQUARE AARU")
    (13060 . "SQUARE ININGU")
    (13061 . "SQUARE INTI")
    (13062 . "SQUARE UON")
    (13063 . "SQUARE ESUKUUDO")
    (13064 . "SQUARE EEKAA")
    (13065 . "SQUARE ONSU")
    (13066 . "SQUARE OOMU")
    (13067 . "SQUARE KAIRI")
    (13068 . "SQUARE KARATTO")
    (13069 . "SQUARE KARORII")
    (13070 . "SQUARE GARON")
    (13071 . "SQUARE GANMA")
    (13072 . "SQUARE GIGA")
    (13073 . "SQUARE GINII")
    (13074 . "SQUARE KYURII")
    (13075 . "SQUARE GIRUDAA")
    (13076 . "SQUARE KIRO")
    (13077 . "SQUARE KIROGURAMU")
    (13078 . "SQUARE KIROMEETORU")
    (13079 . "SQUARE KIROWATTO")
    (13080 . "SQUARE GURAMU")
    (13081 . "SQUARE GURAMUTON")
    (13082 . "SQUARE KURUZEIRO")
    (13083 . "SQUARE KUROONE")
    (13084 . "SQUARE KEESU")
    (13085 . "SQUARE KORUNA")
    (13086 . "SQUARE KOOPO")
    (13087 . "SQUARE SAIKURU")
    (13088 . "SQUARE SANTIIMU")
    (13089 . "SQUARE SIRINGU")
    (13090 . "SQUARE SENTI")
    (13091 . "SQUARE SENTO")
    (13092 . "SQUARE DAASU")
    (13093 . "SQUARE DESI")
    (13094 . "SQUARE DORU")
    (13095 . "SQUARE TON")
    (13096 . "SQUARE NANO")
    (13097 . "SQUARE NOTTO")
    (13098 . "SQUARE HAITU")
    (13099 . "SQUARE PAASENTO")
    (13100 . "SQUARE PAATU")
    (13101 . "SQUARE BAARERU")
    (13102 . "SQUARE PIASUTORU")
    (13103 . "SQUARE PIKURU")
    (13104 . "SQUARE PIKO")
    (13105 . "SQUARE BIRU")
    (13106 . "SQUARE HUARADDO")
    (13107 . "SQUARE HUIITO")
    (13108 . "SQUARE BUSSYERU")
    (13109 . "SQUARE HURAN")
    (13110 . "SQUARE HEKUTAARU")
    (13111 . "SQUARE PESO")
    (13112 . "SQUARE PENIHI")
    (13113 . "SQUARE HERUTU")
    (13114 . "SQUARE PENSU")
    (13115 . "SQUARE PEEZI")
    (13116 . "SQUARE BEETA")
    (13117 . "SQUARE POINTO")
    (13118 . "SQUARE BORUTO")
    (13119 . "SQUARE HON")
    (13120 . "SQUARE PONDO")
    (13121 . "SQUARE HOORU")
    (13122 . "SQUARE HOON")
    (13123 . "SQUARE MAIKURO")
    (13124 . "SQUARE MAIRU")
    (13125 . "SQUARE MAHHA")
    (13126 . "SQUARE MARUKU")
    (13127 . "SQUARE MANSYON")
    (13128 . "SQUARE MIKURON")
    (13129 . "SQUARE MIRI")
    (13130 . "SQUARE MIRIBAARU")
    (13131 . "SQUARE MEGA")
    (13132 . "SQUARE MEGATON")
    (13133 . "SQUARE MEETORU")
    (13134 . "SQUARE YAADO")
    (13135 . "SQUARE YAARU")
    (13136 . "SQUARE YUAN")
    (13137 . "SQUARE RITTORU")
    (13138 . "SQUARE RIRA")
    (13139 . "SQUARE RUPII")
    (13140 . "SQUARE RUUBURU")
    (13141 . "SQUARE REMU")
    (13142 . "SQUARE RENTOGEN")
    (13143 . "SQUARE WATTO")
    (13144 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ZERO")
    (13145 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ONE")
    (13146 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWO")
    (13147 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR THREE")
    (13148 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FOUR")
    (13149 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FIVE")
    (13150 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SIX")
    (13151 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SEVEN")
    (13152 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR EIGHT")
    (13153 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR NINE")
    (13154 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TEN")
    (13155 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ELEVEN")
    (13156 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWELVE")
    (13157 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR THIRTEEN")
    (13158 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FOURTEEN")
    (13159 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR FIFTEEN")
    (13160 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SIXTEEN")
    (13161 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR SEVENTEEN")
    (13162 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR EIGHTEEN")
    (13163 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR NINETEEN")
    (13164 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY")
    (13165 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-ONE")
    (13166 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-TWO")
    (13167 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-THREE")
    (13168 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-FOUR")
    (13169 . "SQUARE HPA")
    (13170 . "SQUARE DA")
    (13171 . "SQUARE AU")
    (13172 . "SQUARE BAR")
    (13173 . "SQUARE OV")
    (13174 . "SQUARE PC")
    (13179 . "SQUARE ERA NAME HEISEI")
    (13180 . "SQUARE ERA NAME SYOUWA")
    (13181 . "SQUARE ERA NAME TAISYOU")
    (13182 . "SQUARE ERA NAME MEIZI")
    (13183 . "SQUARE CORPORATION")
    (13184 . "SQUARE PA AMPS")
    (13185 . "SQUARE NA")
    (13186 . "SQUARE MU A")
    (13187 . "SQUARE MA")
    (13188 . "SQUARE KA")
    (13189 . "SQUARE KB")
    (13190 . "SQUARE MB")
    (13191 . "SQUARE GB")
    (13192 . "SQUARE CAL")
    (13193 . "SQUARE KCAL")
    (13194 . "SQUARE PF")
    (13195 . "SQUARE NF")
    (13196 . "SQUARE MU F")
    (13197 . "SQUARE MU G")
    (13198 . "SQUARE MG")
    (13199 . "SQUARE KG")
    (13200 . "SQUARE HZ")
    (13201 . "SQUARE KHZ")
    (13202 . "SQUARE MHZ")
    (13203 . "SQUARE GHZ")
    (13204 . "SQUARE THZ")
    (13205 . "SQUARE MU L")
    (13206 . "SQUARE ML")
    (13207 . "SQUARE DL")
    (13208 . "SQUARE KL")
    (13209 . "SQUARE FM")
    (13210 . "SQUARE NM")
    (13211 . "SQUARE MU M")
    (13212 . "SQUARE MM")
    (13213 . "SQUARE CM")
    (13214 . "SQUARE KM")
    (13215 . "SQUARE MM SQUARED")
    (13216 . "SQUARE CM SQUARED")
    (13217 . "SQUARE M SQUARED")
    (13218 . "SQUARE KM SQUARED")
    (13219 . "SQUARE MM CUBED")
    (13220 . "SQUARE CM CUBED")
    (13221 . "SQUARE M CUBED")
    (13222 . "SQUARE KM CUBED")
    (13223 . "SQUARE M OVER S")
    (13224 . "SQUARE M OVER S SQUARED")
    (13225 . "SQUARE PA")
    (13226 . "SQUARE KPA")
    (13227 . "SQUARE MPA")
    (13228 . "SQUARE GPA")
    (13229 . "SQUARE RAD")
    (13230 . "SQUARE RAD OVER S")
    (13231 . "SQUARE RAD OVER S SQUARED")
    (13232 . "SQUARE PS")
    (13233 . "SQUARE NS")
    (13234 . "SQUARE MU S")
    (13235 . "SQUARE MS")
    (13236 . "SQUARE PV")
    (13237 . "SQUARE NV")
    (13238 . "SQUARE MU V")
    (13239 . "SQUARE MV")
    (13240 . "SQUARE KV")
    (13241 . "SQUARE MV MEGA")
    (13242 . "SQUARE PW")
    (13243 . "SQUARE NW")
    (13244 . "SQUARE MU W")
    (13245 . "SQUARE MW")
    (13246 . "SQUARE KW")
    (13247 . "SQUARE MW MEGA")
    (13248 . "SQUARE K OHM")
    (13249 . "SQUARE M OHM")
    (13250 . "SQUARE AM")
    (13251 . "SQUARE BQ")
    (13252 . "SQUARE CC")
    (13253 . "SQUARE CD")
    (13254 . "SQUARE C OVER KG")
    (13255 . "SQUARE CO")
    (13256 . "SQUARE DB")
    (13257 . "SQUARE GY")
    (13258 . "SQUARE HA")
    (13259 . "SQUARE HP")
    (13260 . "SQUARE IN")
    (13261 . "SQUARE KK")
    (13262 . "SQUARE KM CAPITAL")
    (13263 . "SQUARE KT")
    (13264 . "SQUARE LM")
    (13265 . "SQUARE LN")
    (13266 . "SQUARE LOG")
    (13267 . "SQUARE LX")
    (13268 . "SQUARE MB SMALL")
    (13269 . "SQUARE MIL")
    (13270 . "SQUARE MOL")
    (13271 . "SQUARE PH")
    (13272 . "SQUARE PM")
    (13273 . "SQUARE PPM")
    (13274 . "SQUARE PR")
    (13275 . "SQUARE SR")
    (13276 . "SQUARE SV")
    (13277 . "SQUARE WB")
    (13280 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY ONE")
    (13281 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWO")
    (13282 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THREE")
    (13283 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FOUR")
    (13284 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FIVE")
    (13285 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SIX")
    (13286 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SEVEN")
    (13287 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY EIGHT")
    (13288 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY NINE")
    (13289 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TEN")
    (13290 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY ELEVEN")
    (13291 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWELVE")
    (13292 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTEEN")
    (13293 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FOURTEEN")
    (13294 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY FIFTEEN")
    (13295 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SIXTEEN")
    (13296 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY SEVENTEEN")
    (13297 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY EIGHTEEN")
    (13298 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY NINETEEN")
    (13299 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY")
    (13300 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-ONE")
    (13301 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-TWO")
    (13302 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-THREE")
    (13303 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-FOUR")
    (13304 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-FIVE")
    (13305 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-SIX")
    (13306 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-SEVEN")
    (13307 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-EIGHT")
    (13308 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY TWENTY-NINE")
    (13309 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTY")
    (13310 . "IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTY-ONE")
    (13312 . "HANGUL SYLLABLE KIYEOK A")
    (13313 . "HANGUL SYLLABLE KIYEOK A KIYEOK")
    (13314 . "HANGUL SYLLABLE KIYEOK A NIEUN")
    (13315 . "HANGUL SYLLABLE KIYEOK A TIKEUT")
    (13316 . "HANGUL SYLLABLE KIYEOK A RIEUL")
    (13317 . "HANGUL SYLLABLE KIYEOK A RIEUL-KIYEOK")
    (13318 . "HANGUL SYLLABLE KIYEOK A RIEUL-MIEUM")
    (13319 . "HANGUL SYLLABLE KIYEOK A MIEUM")
    (13320 . "HANGUL SYLLABLE KIYEOK A PIEUP")
    (13321 . "HANGUL SYLLABLE KIYEOK A PIEUP-SIOS")
    (13322 . "HANGUL SYLLABLE KIYEOK A SIOS")
    (13323 . "HANGUL SYLLABLE KIYEOK A SSANGSIOS")
    (13324 . "HANGUL SYLLABLE KIYEOK A IEUNG")
    (13325 . "HANGUL SYLLABLE KIYEOK A CIEUC")
    (13326 . "HANGUL SYLLABLE KIYEOK A CHIEUCH")
    (13327 . "HANGUL SYLLABLE KIYEOK A THIEUTH")
    (13328 . "HANGUL SYLLABLE KIYEOK A PHIEUPH")
    (13329 . "HANGUL SYLLABLE KIYEOK A HIEUH")
    (13330 . "HANGUL SYLLABLE KIYEOK AE")
    (13331 . "HANGUL SYLLABLE KIYEOK AE KIYEOK")
    (13332 . "HANGUL SYLLABLE KIYEOK AE NIEUN")
    (13333 . "HANGUL SYLLABLE KIYEOK AE RIEUL")
    (13334 . "HANGUL SYLLABLE KIYEOK AE MIEUM")
    (13335 . "HANGUL SYLLABLE KIYEOK AE PIEUP")
    (13336 . "HANGUL SYLLABLE KIYEOK AE SIOS")
    (13337 . "HANGUL SYLLABLE KIYEOK AE SSANGSIOS")
    (13338 . "HANGUL SYLLABLE KIYEOK AE IEUNG")
    (13339 . "HANGUL SYLLABLE KIYEOK YA")
    (13340 . "HANGUL SYLLABLE KIYEOK YA KIYEOK")
    (13341 . "HANGUL SYLLABLE KIYEOK YA NIEUN")
    (13342 . "HANGUL SYLLABLE KIYEOK YA RIEUL")
    (13343 . "HANGUL SYLLABLE KIYEOK YA SIOS")
    (13344 . "HANGUL SYLLABLE KIYEOK YA IEUNG")
    (13345 . "HANGUL SYLLABLE KIYEOK YAE")
    (13346 . "HANGUL SYLLABLE KIYEOK YAE NIEUN")
    (13347 . "HANGUL SYLLABLE KIYEOK YAE RIEUL")
    (13348 . "HANGUL SYLLABLE KIYEOK EO")
    (13349 . "HANGUL SYLLABLE KIYEOK EO KIYEOK")
    (13350 . "HANGUL SYLLABLE KIYEOK EO NIEUN")
    (13351 . "HANGUL SYLLABLE KIYEOK EO TIKEUT")
    (13352 . "HANGUL SYLLABLE KIYEOK EO RIEUL")
    (13353 . "HANGUL SYLLABLE KIYEOK EO RIEUL-MIEUM")
    (13354 . "HANGUL SYLLABLE KIYEOK EO MIEUM")
    (13355 . "HANGUL SYLLABLE KIYEOK EO PIEUP")
    (13356 . "HANGUL SYLLABLE KIYEOK EO SIOS")
    (13357 . "HANGUL SYLLABLE KIYEOK EO SSANGSIOS")
    (13358 . "HANGUL SYLLABLE KIYEOK EO IEUNG")
    (13359 . "HANGUL SYLLABLE KIYEOK EO CIEUC")
    (13360 . "HANGUL SYLLABLE KIYEOK EO THIEUTH")
    (13361 . "HANGUL SYLLABLE KIYEOK EO PHIEUPH")
    (13362 . "HANGUL SYLLABLE KIYEOK EO HIEUH")
    (13363 . "HANGUL SYLLABLE KIYEOK E")
    (13364 . "HANGUL SYLLABLE KIYEOK E NIEUN")
    (13365 . "HANGUL SYLLABLE KIYEOK E RIEUL")
    (13366 . "HANGUL SYLLABLE KIYEOK E MIEUM")
    (13367 . "HANGUL SYLLABLE KIYEOK E PIEUP")
    (13368 . "HANGUL SYLLABLE KIYEOK E SIOS")
    (13369 . "HANGUL SYLLABLE KIYEOK E SSANGSIOS")
    (13370 . "HANGUL SYLLABLE KIYEOK E IEUNG")
    (13371 . "HANGUL SYLLABLE KIYEOK YEO")
    (13372 . "HANGUL SYLLABLE KIYEOK YEO KIYEOK")
    (13373 . "HANGUL SYLLABLE KIYEOK YEO SSANGKIYEOK")
    (13374 . "HANGUL SYLLABLE KIYEOK YEO NIEUN")
    (13375 . "HANGUL SYLLABLE KIYEOK YEO TIKEUT")
    (13376 . "HANGUL SYLLABLE KIYEOK YEO RIEUL")
    (13377 . "HANGUL SYLLABLE KIYEOK YEO MIEUM")
    (13378 . "HANGUL SYLLABLE KIYEOK YEO PIEUP")
    (13379 . "HANGUL SYLLABLE KIYEOK YEO SIOS")
    (13380 . "HANGUL SYLLABLE KIYEOK YEO SSANGSIOS")
    (13381 . "HANGUL SYLLABLE KIYEOK YEO IEUNG")
    (13382 . "HANGUL SYLLABLE KIYEOK YEO THIEUTH")
    (13383 . "HANGUL SYLLABLE KIYEOK YE")
    (13384 . "HANGUL SYLLABLE KIYEOK YE NIEUN")
    (13385 . "HANGUL SYLLABLE KIYEOK YE RIEUL")
    (13386 . "HANGUL SYLLABLE KIYEOK YE PIEUP")
    (13387 . "HANGUL SYLLABLE KIYEOK YE SIOS")
    (13388 . "HANGUL SYLLABLE KIYEOK O")
    (13389 . "HANGUL SYLLABLE KIYEOK O KIYEOK")
    (13390 . "HANGUL SYLLABLE KIYEOK O NIEUN")
    (13391 . "HANGUL SYLLABLE KIYEOK O TIKEUT")
    (13392 . "HANGUL SYLLABLE KIYEOK O RIEUL")
    (13393 . "HANGUL SYLLABLE KIYEOK O RIEUL-MIEUM")
    (13394 . "HANGUL SYLLABLE KIYEOK O RIEUL-SIOS")
    (13395 . "HANGUL SYLLABLE KIYEOK O RIEUL-HIEUH")
    (13396 . "HANGUL SYLLABLE KIYEOK O MIEUM")
    (13397 . "HANGUL SYLLABLE KIYEOK O PIEUP")
    (13398 . "HANGUL SYLLABLE KIYEOK O SIOS")
    (13399 . "HANGUL SYLLABLE KIYEOK O IEUNG")
    (13400 . "HANGUL SYLLABLE KIYEOK O CIEUC")
    (13401 . "HANGUL SYLLABLE KIYEOK WA")
    (13402 . "HANGUL SYLLABLE KIYEOK WA KIYEOK")
    (13403 . "HANGUL SYLLABLE KIYEOK WA NIEUN")
    (13404 . "HANGUL SYLLABLE KIYEOK WA RIEUL")
    (13405 . "HANGUL SYLLABLE KIYEOK WA RIEUL-MIEUM")
    (13406 . "HANGUL SYLLABLE KIYEOK WA MIEUM")
    (13407 . "HANGUL SYLLABLE KIYEOK WA PIEUP")
    (13408 . "HANGUL SYLLABLE KIYEOK WA SIOS")
    (13409 . "HANGUL SYLLABLE KIYEOK WA IEUNG")
    (13410 . "HANGUL SYLLABLE KIYEOK WAE")
    (13411 . "HANGUL SYLLABLE KIYEOK WAE NIEUN")
    (13412 . "HANGUL SYLLABLE KIYEOK WAE RIEUL")
    (13413 . "HANGUL SYLLABLE KIYEOK WAE PIEUP")
    (13414 . "HANGUL SYLLABLE KIYEOK WAE SSANGSIOS")
    (13415 . "HANGUL SYLLABLE KIYEOK WAE IEUNG")
    (13416 . "HANGUL SYLLABLE KIYEOK OE")
    (13417 . "HANGUL SYLLABLE KIYEOK OE KIYEOK")
    (13418 . "HANGUL SYLLABLE KIYEOK OE NIEUN")
    (13419 . "HANGUL SYLLABLE KIYEOK OE RIEUL")
    (13420 . "HANGUL SYLLABLE KIYEOK OE MIEUM")
    (13421 . "HANGUL SYLLABLE KIYEOK OE PIEUP")
    (13422 . "HANGUL SYLLABLE KIYEOK OE SIOS")
    (13423 . "HANGUL SYLLABLE KIYEOK OE IEUNG")
    (13424 . "HANGUL SYLLABLE KIYEOK YO")
    (13425 . "HANGUL SYLLABLE KIYEOK YO NIEUN")
    (13426 . "HANGUL SYLLABLE KIYEOK YO RIEUL")
    (13427 . "HANGUL SYLLABLE KIYEOK YO PIEUP")
    (13428 . "HANGUL SYLLABLE KIYEOK YO SIOS")
    (13429 . "HANGUL SYLLABLE KIYEOK U")
    (13430 . "HANGUL SYLLABLE KIYEOK U KIYEOK")
    (13431 . "HANGUL SYLLABLE KIYEOK U NIEUN")
    (13432 . "HANGUL SYLLABLE KIYEOK U TIKEUT")
    (13433 . "HANGUL SYLLABLE KIYEOK U RIEUL")
    (13434 . "HANGUL SYLLABLE KIYEOK U RIEUL-KIYEOK")
    (13435 . "HANGUL SYLLABLE KIYEOK U RIEUL-MIEUM")
    (13436 . "HANGUL SYLLABLE KIYEOK U RIEUL-HIEUH")
    (13437 . "HANGUL SYLLABLE KIYEOK U MIEUM")
    (13438 . "HANGUL SYLLABLE KIYEOK U PIEUP")
    (13439 . "HANGUL SYLLABLE KIYEOK U SIOS")
    (13440 . "HANGUL SYLLABLE KIYEOK U IEUNG")
    (13441 . "HANGUL SYLLABLE KIYEOK U CIEUC")
    (13442 . "HANGUL SYLLABLE KIYEOK WEO")
    (13443 . "HANGUL SYLLABLE KIYEOK WEO KIYEOK")
    (13444 . "HANGUL SYLLABLE KIYEOK WEO NIEUN")
    (13445 . "HANGUL SYLLABLE KIYEOK WEO RIEUL")
    (13446 . "HANGUL SYLLABLE KIYEOK WEO SSANGSIOS")
    (13447 . "HANGUL SYLLABLE KIYEOK WEO IEUNG")
    (13448 . "HANGUL SYLLABLE KIYEOK WE")
    (13449 . "HANGUL SYLLABLE KIYEOK WE SIOS")
    (13450 . "HANGUL SYLLABLE KIYEOK WI")
    (13451 . "HANGUL SYLLABLE KIYEOK WI KIYEOK")
    (13452 . "HANGUL SYLLABLE KIYEOK WI NIEUN")
    (13453 . "HANGUL SYLLABLE KIYEOK WI RIEUL")
    (13454 . "HANGUL SYLLABLE KIYEOK WI MIEUM")
    (13455 . "HANGUL SYLLABLE KIYEOK WI PIEUP")
    (13456 . "HANGUL SYLLABLE KIYEOK WI SIOS")
    (13457 . "HANGUL SYLLABLE KIYEOK YU")
    (13458 . "HANGUL SYLLABLE KIYEOK YU NIEUN")
    (13459 . "HANGUL SYLLABLE KIYEOK YU RIEUL")
    (13460 . "HANGUL SYLLABLE KIYEOK EU")
    (13461 . "HANGUL SYLLABLE KIYEOK EU KIYEOK")
    (13462 . "HANGUL SYLLABLE KIYEOK EU NIEUN")
    (13463 . "HANGUL SYLLABLE KIYEOK EU TIKEUT")
    (13464 . "HANGUL SYLLABLE KIYEOK EU RIEUL")
    (13465 . "HANGUL SYLLABLE KIYEOK EU RIEUL-KIYEOK")
    (13466 . "HANGUL SYLLABLE KIYEOK EU MIEUM")
    (13467 . "HANGUL SYLLABLE KIYEOK EU PIEUP")
    (13468 . "HANGUL SYLLABLE KIYEOK EU SIOS")
    (13469 . "HANGUL SYLLABLE KIYEOK EU IEUNG")
    (13470 . "HANGUL SYLLABLE KIYEOK YI")
    (13471 . "HANGUL SYLLABLE KIYEOK I")
    (13472 . "HANGUL SYLLABLE KIYEOK I KIYEOK")
    (13473 . "HANGUL SYLLABLE KIYEOK I NIEUN")
    (13474 . "HANGUL SYLLABLE KIYEOK I TIKEUT")
    (13475 . "HANGUL SYLLABLE KIYEOK I RIEUL")
    (13476 . "HANGUL SYLLABLE KIYEOK I RIEUL-MIEUM")
    (13477 . "HANGUL SYLLABLE KIYEOK I MIEUM")
    (13478 . "HANGUL SYLLABLE KIYEOK I PIEUP")
    (13479 . "HANGUL SYLLABLE KIYEOK I SIOS")
    (13480 . "HANGUL SYLLABLE KIYEOK I IEUNG")
    (13481 . "HANGUL SYLLABLE KIYEOK I CIEUC")
    (13482 . "HANGUL SYLLABLE KIYEOK I PHIEUPH")
    (13483 . "HANGUL SYLLABLE SSANGKIYEOK A")
    (13484 . "HANGUL SYLLABLE SSANGKIYEOK A KIYEOK")
    (13485 . "HANGUL SYLLABLE SSANGKIYEOK A SSANGKIYEOK")
    (13486 . "HANGUL SYLLABLE SSANGKIYEOK A NIEUN")
    (13487 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL")
    (13488 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-MIEUM")
    (13489 . "HANGUL SYLLABLE SSANGKIYEOK A MIEUM")
    (13490 . "HANGUL SYLLABLE SSANGKIYEOK A PIEUP")
    (13491 . "HANGUL SYLLABLE SSANGKIYEOK A SIOS")
    (13492 . "HANGUL SYLLABLE SSANGKIYEOK A SSANGSIOS")
    (13493 . "HANGUL SYLLABLE SSANGKIYEOK A IEUNG")
    (13494 . "HANGUL SYLLABLE SSANGKIYEOK A THIEUTH")
    (13495 . "HANGUL SYLLABLE SSANGKIYEOK AE")
    (13496 . "HANGUL SYLLABLE SSANGKIYEOK AE KIYEOK")
    (13497 . "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN")
    (13498 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL")
    (13499 . "HANGUL SYLLABLE SSANGKIYEOK AE MIEUM")
    (13500 . "HANGUL SYLLABLE SSANGKIYEOK AE PIEUP")
    (13501 . "HANGUL SYLLABLE SSANGKIYEOK AE SIOS")
    (13502 . "HANGUL SYLLABLE SSANGKIYEOK AE SSANGSIOS")
    (13503 . "HANGUL SYLLABLE SSANGKIYEOK AE IEUNG")
    (13504 . "HANGUL SYLLABLE SSANGKIYEOK YA")
    (13505 . "HANGUL SYLLABLE SSANGKIYEOK YA KIYEOK")
    (13506 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL")
    (13507 . "HANGUL SYLLABLE SSANGKIYEOK EO")
    (13508 . "HANGUL SYLLABLE SSANGKIYEOK EO KIYEOK")
    (13509 . "HANGUL SYLLABLE SSANGKIYEOK EO SSANGKIYEOK")
    (13510 . "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN")
    (13511 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL")
    (13512 . "HANGUL SYLLABLE SSANGKIYEOK EO MIEUM")
    (13513 . "HANGUL SYLLABLE SSANGKIYEOK EO PIEUP")
    (13514 . "HANGUL SYLLABLE SSANGKIYEOK EO SIOS")
    (13515 . "HANGUL SYLLABLE SSANGKIYEOK EO SSANGSIOS")
    (13516 . "HANGUL SYLLABLE SSANGKIYEOK EO IEUNG")
    (13517 . "HANGUL SYLLABLE SSANGKIYEOK E")
    (13518 . "HANGUL SYLLABLE SSANGKIYEOK E KIYEOK")
    (13519 . "HANGUL SYLLABLE SSANGKIYEOK E NIEUN")
    (13520 . "HANGUL SYLLABLE SSANGKIYEOK E MIEUM")
    (13521 . "HANGUL SYLLABLE SSANGKIYEOK E SIOS")
    (13522 . "HANGUL SYLLABLE SSANGKIYEOK E IEUNG")
    (13523 . "HANGUL SYLLABLE SSANGKIYEOK YEO")
    (13524 . "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN")
    (13525 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL")
    (13526 . "HANGUL SYLLABLE SSANGKIYEOK YEO SIOS")
    (13527 . "HANGUL SYLLABLE SSANGKIYEOK YEO SSANGSIOS")
    (13528 . "HANGUL SYLLABLE SSANGKIYEOK YEO THIEUTH")
    (13529 . "HANGUL SYLLABLE SSANGKIYEOK YE")
    (13530 . "HANGUL SYLLABLE SSANGKIYEOK O")
    (13531 . "HANGUL SYLLABLE SSANGKIYEOK O KIYEOK")
    (13532 . "HANGUL SYLLABLE SSANGKIYEOK O NIEUN")
    (13533 . "HANGUL SYLLABLE SSANGKIYEOK O NIEUN-HIEUH")
    (13534 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL")
    (13535 . "HANGUL SYLLABLE SSANGKIYEOK O MIEUM")
    (13536 . "HANGUL SYLLABLE SSANGKIYEOK O PIEUP")
    (13537 . "HANGUL SYLLABLE SSANGKIYEOK O SIOS")
    (13538 . "HANGUL SYLLABLE SSANGKIYEOK O IEUNG")
    (13539 . "HANGUL SYLLABLE SSANGKIYEOK O CIEUC")
    (13540 . "HANGUL SYLLABLE SSANGKIYEOK O CHIEUCH")
    (13541 . "HANGUL SYLLABLE SSANGKIYEOK WA")
    (13542 . "HANGUL SYLLABLE SSANGKIYEOK WA KIYEOK")
    (13543 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL")
    (13544 . "HANGUL SYLLABLE SSANGKIYEOK WA SSANGSIOS")
    (13545 . "HANGUL SYLLABLE SSANGKIYEOK WA IEUNG")
    (13546 . "HANGUL SYLLABLE SSANGKIYEOK WAE")
    (13547 . "HANGUL SYLLABLE SSANGKIYEOK WAE KIYEOK")
    (13548 . "HANGUL SYLLABLE SSANGKIYEOK WAE IEUNG")
    (13549 . "HANGUL SYLLABLE SSANGKIYEOK OE")
    (13550 . "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN")
    (13551 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL")
    (13552 . "HANGUL SYLLABLE SSANGKIYEOK OE MIEUM")
    (13553 . "HANGUL SYLLABLE SSANGKIYEOK OE PIEUP")
    (13554 . "HANGUL SYLLABLE SSANGKIYEOK OE IEUNG")
    (13555 . "HANGUL SYLLABLE SSANGKIYEOK YO")
    (13556 . "HANGUL SYLLABLE SSANGKIYEOK U")
    (13557 . "HANGUL SYLLABLE SSANGKIYEOK U KIYEOK")
    (13558 . "HANGUL SYLLABLE SSANGKIYEOK U NIEUN")
    (13559 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL")
    (13560 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-HIEUH")
    (13561 . "HANGUL SYLLABLE SSANGKIYEOK U MIEUM")
    (13562 . "HANGUL SYLLABLE SSANGKIYEOK U PIEUP")
    (13563 . "HANGUL SYLLABLE SSANGKIYEOK U SIOS")
    (13564 . "HANGUL SYLLABLE SSANGKIYEOK U IEUNG")
    (13565 . "HANGUL SYLLABLE SSANGKIYEOK U CIEUC")
    (13566 . "HANGUL SYLLABLE SSANGKIYEOK WEO")
    (13567 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL")
    (13568 . "HANGUL SYLLABLE SSANGKIYEOK WEO SSANGSIOS")
    (13569 . "HANGUL SYLLABLE SSANGKIYEOK WEO IEUNG")
    (13570 . "HANGUL SYLLABLE SSANGKIYEOK WE")
    (13571 . "HANGUL SYLLABLE SSANGKIYEOK WE KIYEOK")
    (13572 . "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN")
    (13573 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL")
    (13574 . "HANGUL SYLLABLE SSANGKIYEOK WE MIEUM")
    (13575 . "HANGUL SYLLABLE SSANGKIYEOK WE PIEUP")
    (13576 . "HANGUL SYLLABLE SSANGKIYEOK WE SSANGSIOS")
    (13577 . "HANGUL SYLLABLE SSANGKIYEOK WI")
    (13578 . "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN")
    (13579 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL")
    (13580 . "HANGUL SYLLABLE SSANGKIYEOK WI MIEUM")
    (13581 . "HANGUL SYLLABLE SSANGKIYEOK WI PIEUP")
    (13582 . "HANGUL SYLLABLE SSANGKIYEOK YU")
    (13583 . "HANGUL SYLLABLE SSANGKIYEOK EU")
    (13584 . "HANGUL SYLLABLE SSANGKIYEOK EU KIYEOK")
    (13585 . "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN")
    (13586 . "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN-HIEUH")
    (13587 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL")
    (13588 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-MIEUM")
    (13589 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-HIEUH")
    (13590 . "HANGUL SYLLABLE SSANGKIYEOK EU MIEUM")
    (13591 . "HANGUL SYLLABLE SSANGKIYEOK EU PIEUP")
    (13592 . "HANGUL SYLLABLE SSANGKIYEOK EU SIOS")
    (13593 . "HANGUL SYLLABLE SSANGKIYEOK EU IEUNG")
    (13594 . "HANGUL SYLLABLE SSANGKIYEOK EU THIEUTH")
    (13595 . "HANGUL SYLLABLE SSANGKIYEOK I")
    (13596 . "HANGUL SYLLABLE SSANGKIYEOK I KIYEOK")
    (13597 . "HANGUL SYLLABLE SSANGKIYEOK I NIEUN")
    (13598 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL")
    (13599 . "HANGUL SYLLABLE SSANGKIYEOK I MIEUM")
    (13600 . "HANGUL SYLLABLE SSANGKIYEOK I PIEUP")
    (13601 . "HANGUL SYLLABLE SSANGKIYEOK I SIOS")
    (13602 . "HANGUL SYLLABLE SSANGKIYEOK I IEUNG")
    (13603 . "HANGUL SYLLABLE NIEUN A")
    (13604 . "HANGUL SYLLABLE NIEUN A KIYEOK")
    (13605 . "HANGUL SYLLABLE NIEUN A SSANGKIYEOK")
    (13606 . "HANGUL SYLLABLE NIEUN A NIEUN")
    (13607 . "HANGUL SYLLABLE NIEUN A TIKEUT")
    (13608 . "HANGUL SYLLABLE NIEUN A RIEUL")
    (13609 . "HANGUL SYLLABLE NIEUN A RIEUL-KIYEOK")
    (13610 . "HANGUL SYLLABLE NIEUN A RIEUL-MIEUM")
    (13611 . "HANGUL SYLLABLE NIEUN A MIEUM")
    (13612 . "HANGUL SYLLABLE NIEUN A PIEUP")
    (13613 . "HANGUL SYLLABLE NIEUN A SIOS")
    (13614 . "HANGUL SYLLABLE NIEUN A SSANGSIOS")
    (13615 . "HANGUL SYLLABLE NIEUN A IEUNG")
    (13616 . "HANGUL SYLLABLE NIEUN A CIEUC")
    (13617 . "HANGUL SYLLABLE NIEUN A CHIEUCH")
    (13618 . "HANGUL SYLLABLE NIEUN A THIEUTH")
    (13619 . "HANGUL SYLLABLE NIEUN A HIEUH")
    (13620 . "HANGUL SYLLABLE NIEUN AE")
    (13621 . "HANGUL SYLLABLE NIEUN AE KIYEOK")
    (13622 . "HANGUL SYLLABLE NIEUN AE NIEUN")
    (13623 . "HANGUL SYLLABLE NIEUN AE RIEUL")
    (13624 . "HANGUL SYLLABLE NIEUN AE MIEUM")
    (13625 . "HANGUL SYLLABLE NIEUN AE PIEUP")
    (13626 . "HANGUL SYLLABLE NIEUN AE SIOS")
    (13627 . "HANGUL SYLLABLE NIEUN AE SSANGSIOS")
    (13628 . "HANGUL SYLLABLE NIEUN AE IEUNG")
    (13629 . "HANGUL SYLLABLE NIEUN YA")
    (13630 . "HANGUL SYLLABLE NIEUN YA KIYEOK")
    (13631 . "HANGUL SYLLABLE NIEUN YA NIEUN")
    (13632 . "HANGUL SYLLABLE NIEUN YA RIEUL")
    (13633 . "HANGUL SYLLABLE NIEUN YA MIEUM")
    (13634 . "HANGUL SYLLABLE NIEUN YA IEUNG")
    (13635 . "HANGUL SYLLABLE NIEUN EO")
    (13636 . "HANGUL SYLLABLE NIEUN EO KIYEOK")
    (13637 . "HANGUL SYLLABLE NIEUN EO KIYEOK-SIOS")
    (13638 . "HANGUL SYLLABLE NIEUN EO NIEUN")
    (13639 . "HANGUL SYLLABLE NIEUN EO RIEUL")
    (13640 . "HANGUL SYLLABLE NIEUN EO RIEUL-MIEUM")
    (13641 . "HANGUL SYLLABLE NIEUN EO RIEUL-PIEUP")
    (13642 . "HANGUL SYLLABLE NIEUN EO MIEUM")
    (13643 . "HANGUL SYLLABLE NIEUN EO PIEUP")
    (13644 . "HANGUL SYLLABLE NIEUN EO SIOS")
    (13645 . "HANGUL SYLLABLE NIEUN EO SSANGSIOS")
    (13646 . "HANGUL SYLLABLE NIEUN EO IEUNG")
    (13647 . "HANGUL SYLLABLE NIEUN EO HIEUH")
    (13648 . "HANGUL SYLLABLE NIEUN E")
    (13649 . "HANGUL SYLLABLE NIEUN E KIYEOK")
    (13650 . "HANGUL SYLLABLE NIEUN E NIEUN")
    (13651 . "HANGUL SYLLABLE NIEUN E RIEUL")
    (13652 . "HANGUL SYLLABLE NIEUN E MIEUM")
    (13653 . "HANGUL SYLLABLE NIEUN E PIEUP")
    (13654 . "HANGUL SYLLABLE NIEUN E SIOS")
    (13655 . "HANGUL SYLLABLE NIEUN E SSANGSIOS")
    (13656 . "HANGUL SYLLABLE NIEUN E IEUNG")
    (13657 . "HANGUL SYLLABLE NIEUN YEO")
    (13658 . "HANGUL SYLLABLE NIEUN YEO KIYEOK")
    (13659 . "HANGUL SYLLABLE NIEUN YEO NIEUN")
    (13660 . "HANGUL SYLLABLE NIEUN YEO RIEUL")
    (13661 . "HANGUL SYLLABLE NIEUN YEO MIEUM")
    (13662 . "HANGUL SYLLABLE NIEUN YEO PIEUP")
    (13663 . "HANGUL SYLLABLE NIEUN YEO SSANGSIOS")
    (13664 . "HANGUL SYLLABLE NIEUN YEO IEUNG")
    (13665 . "HANGUL SYLLABLE NIEUN YEO KHIEUKH")
    (13666 . "HANGUL SYLLABLE NIEUN YE")
    (13667 . "HANGUL SYLLABLE NIEUN YE NIEUN")
    (13668 . "HANGUL SYLLABLE NIEUN O")
    (13669 . "HANGUL SYLLABLE NIEUN O KIYEOK")
    (13670 . "HANGUL SYLLABLE NIEUN O NIEUN")
    (13671 . "HANGUL SYLLABLE NIEUN O RIEUL")
    (13672 . "HANGUL SYLLABLE NIEUN O RIEUL-MIEUM")
    (13673 . "HANGUL SYLLABLE NIEUN O MIEUM")
    (13674 . "HANGUL SYLLABLE NIEUN O PIEUP")
    (13675 . "HANGUL SYLLABLE NIEUN O SIOS")
    (13676 . "HANGUL SYLLABLE NIEUN O IEUNG")
    (13677 . "HANGUL SYLLABLE NIEUN O PHIEUPH")
    (13678 . "HANGUL SYLLABLE NIEUN O HIEUH")
    (13679 . "HANGUL SYLLABLE NIEUN WA")
    (13680 . "HANGUL SYLLABLE NIEUN WA NIEUN")
    (13681 . "HANGUL SYLLABLE NIEUN WA RIEUL")
    (13682 . "HANGUL SYLLABLE NIEUN WA SSANGSIOS")
    (13683 . "HANGUL SYLLABLE NIEUN OE")
    (13684 . "HANGUL SYLLABLE NIEUN OE NIEUN")
    (13685 . "HANGUL SYLLABLE NIEUN OE RIEUL")
    (13686 . "HANGUL SYLLABLE NIEUN OE MIEUM")
    (13687 . "HANGUL SYLLABLE NIEUN OE PIEUP")
    (13688 . "HANGUL SYLLABLE NIEUN OE SIOS")
    (13689 . "HANGUL SYLLABLE NIEUN YO")
    (13690 . "HANGUL SYLLABLE NIEUN YO KIYEOK")
    (13691 . "HANGUL SYLLABLE NIEUN YO NIEUN")
    (13692 . "HANGUL SYLLABLE NIEUN YO RIEUL")
    (13693 . "HANGUL SYLLABLE NIEUN YO PIEUP")
    (13694 . "HANGUL SYLLABLE NIEUN YO SIOS")
    (13695 . "HANGUL SYLLABLE NIEUN YO IEUNG")
    (13696 . "HANGUL SYLLABLE NIEUN U")
    (13697 . "HANGUL SYLLABLE NIEUN U KIYEOK")
    (13698 . "HANGUL SYLLABLE NIEUN U NIEUN")
    (13699 . "HANGUL SYLLABLE NIEUN U TIKEUT")
    (13700 . "HANGUL SYLLABLE NIEUN U RIEUL")
    (13701 . "HANGUL SYLLABLE NIEUN U MIEUM")
    (13702 . "HANGUL SYLLABLE NIEUN U PIEUP")
    (13703 . "HANGUL SYLLABLE NIEUN U SIOS")
    (13704 . "HANGUL SYLLABLE NIEUN U IEUNG")
    (13705 . "HANGUL SYLLABLE NIEUN WEO")
    (13706 . "HANGUL SYLLABLE NIEUN WEO SSANGSIOS")
    (13707 . "HANGUL SYLLABLE NIEUN WE")
    (13708 . "HANGUL SYLLABLE NIEUN WI")
    (13709 . "HANGUL SYLLABLE NIEUN WI NIEUN")
    (13710 . "HANGUL SYLLABLE NIEUN WI RIEUL")
    (13711 . "HANGUL SYLLABLE NIEUN WI MIEUM")
    (13712 . "HANGUL SYLLABLE NIEUN WI PIEUP")
    (13713 . "HANGUL SYLLABLE NIEUN YU")
    (13714 . "HANGUL SYLLABLE NIEUN YU KIYEOK")
    (13715 . "HANGUL SYLLABLE NIEUN YU RIEUL")
    (13716 . "HANGUL SYLLABLE NIEUN YU MIEUM")
    (13717 . "HANGUL SYLLABLE NIEUN YU PIEUP")
    (13718 . "HANGUL SYLLABLE NIEUN YU IEUNG")
    (13719 . "HANGUL SYLLABLE NIEUN EU")
    (13720 . "HANGUL SYLLABLE NIEUN EU KIYEOK")
    (13721 . "HANGUL SYLLABLE NIEUN EU NIEUN")
    (13722 . "HANGUL SYLLABLE NIEUN EU RIEUL")
    (13723 . "HANGUL SYLLABLE NIEUN EU RIEUL-KIYEOK")
    (13724 . "HANGUL SYLLABLE NIEUN EU RIEUL-MIEUM")
    (13725 . "HANGUL SYLLABLE NIEUN EU MIEUM")
    (13726 . "HANGUL SYLLABLE NIEUN EU PIEUP")
    (13727 . "HANGUL SYLLABLE NIEUN EU SIOS")
    (13728 . "HANGUL SYLLABLE NIEUN EU IEUNG")
    (13729 . "HANGUL SYLLABLE NIEUN EU CIEUC")
    (13730 . "HANGUL SYLLABLE NIEUN EU PHIEUPH")
    (13731 . "HANGUL SYLLABLE NIEUN YI")
    (13732 . "HANGUL SYLLABLE NIEUN YI NIEUN")
    (13733 . "HANGUL SYLLABLE NIEUN YI RIEUL")
    (13734 . "HANGUL SYLLABLE NIEUN I")
    (13735 . "HANGUL SYLLABLE NIEUN I KIYEOK")
    (13736 . "HANGUL SYLLABLE NIEUN I NIEUN")
    (13737 . "HANGUL SYLLABLE NIEUN I RIEUL")
    (13738 . "HANGUL SYLLABLE NIEUN I RIEUL-MIEUM")
    (13739 . "HANGUL SYLLABLE NIEUN I MIEUM")
    (13740 . "HANGUL SYLLABLE NIEUN I PIEUP")
    (13741 . "HANGUL SYLLABLE NIEUN I SIOS")
    (13742 . "HANGUL SYLLABLE NIEUN I IEUNG")
    (13743 . "HANGUL SYLLABLE NIEUN I PHIEUPH")
    (13744 . "HANGUL SYLLABLE TIKEUT A")
    (13745 . "HANGUL SYLLABLE TIKEUT A KIYEOK")
    (13746 . "HANGUL SYLLABLE TIKEUT A SSANGKIYEOK")
    (13747 . "HANGUL SYLLABLE TIKEUT A NIEUN")
    (13748 . "HANGUL SYLLABLE TIKEUT A TIKEUT")
    (13749 . "HANGUL SYLLABLE TIKEUT A RIEUL")
    (13750 . "HANGUL SYLLABLE TIKEUT A RIEUL-KIYEOK")
    (13751 . "HANGUL SYLLABLE TIKEUT A RIEUL-MIEUM")
    (13752 . "HANGUL SYLLABLE TIKEUT A RIEUL-PIEUP")
    (13753 . "HANGUL SYLLABLE TIKEUT A RIEUL-HIEUH")
    (13754 . "HANGUL SYLLABLE TIKEUT A MIEUM")
    (13755 . "HANGUL SYLLABLE TIKEUT A PIEUP")
    (13756 . "HANGUL SYLLABLE TIKEUT A SIOS")
    (13757 . "HANGUL SYLLABLE TIKEUT A SSANGSIOS")
    (13758 . "HANGUL SYLLABLE TIKEUT A IEUNG")
    (13759 . "HANGUL SYLLABLE TIKEUT A CIEUC")
    (13760 . "HANGUL SYLLABLE TIKEUT A CHIEUCH")
    (13761 . "HANGUL SYLLABLE TIKEUT A HIEUH")
    (13762 . "HANGUL SYLLABLE TIKEUT AE")
    (13763 . "HANGUL SYLLABLE TIKEUT AE KIYEOK")
    (13764 . "HANGUL SYLLABLE TIKEUT AE NIEUN")
    (13765 . "HANGUL SYLLABLE TIKEUT AE RIEUL")
    (13766 . "HANGUL SYLLABLE TIKEUT AE MIEUM")
    (13767 . "HANGUL SYLLABLE TIKEUT AE PIEUP")
    (13768 . "HANGUL SYLLABLE TIKEUT AE SIOS")
    (13769 . "HANGUL SYLLABLE TIKEUT AE SSANGSIOS")
    (13770 . "HANGUL SYLLABLE TIKEUT AE IEUNG")
    (13771 . "HANGUL SYLLABLE TIKEUT YA")
    (13772 . "HANGUL SYLLABLE TIKEUT EO")
    (13773 . "HANGUL SYLLABLE TIKEUT EO KIYEOK")
    (13774 . "HANGUL SYLLABLE TIKEUT EO SSANGKIYEOK")
    (13775 . "HANGUL SYLLABLE TIKEUT EO NIEUN")
    (13776 . "HANGUL SYLLABLE TIKEUT EO TIKEUT")
    (13777 . "HANGUL SYLLABLE TIKEUT EO RIEUL")
    (13778 . "HANGUL SYLLABLE TIKEUT EO RIEUL-MIEUM")
    (13779 . "HANGUL SYLLABLE TIKEUT EO RIEUL-PIEUP")
    (13780 . "HANGUL SYLLABLE TIKEUT EO MIEUM")
    (13781 . "HANGUL SYLLABLE TIKEUT EO PIEUP")
    (13782 . "HANGUL SYLLABLE TIKEUT EO SIOS")
    (13783 . "HANGUL SYLLABLE TIKEUT EO IEUNG")
    (13784 . "HANGUL SYLLABLE TIKEUT EO CHIEUCH")
    (13785 . "HANGUL SYLLABLE TIKEUT EO PHIEUPH")
    (13786 . "HANGUL SYLLABLE TIKEUT E")
    (13787 . "HANGUL SYLLABLE TIKEUT E KIYEOK")
    (13788 . "HANGUL SYLLABLE TIKEUT E NIEUN")
    (13789 . "HANGUL SYLLABLE TIKEUT E RIEUL")
    (13790 . "HANGUL SYLLABLE TIKEUT E MIEUM")
    (13791 . "HANGUL SYLLABLE TIKEUT E PIEUP")
    (13792 . "HANGUL SYLLABLE TIKEUT E SIOS")
    (13793 . "HANGUL SYLLABLE TIKEUT E SSANGSIOS")
    (13794 . "HANGUL SYLLABLE TIKEUT E IEUNG")
    (13795 . "HANGUL SYLLABLE TIKEUT YEO")
    (13796 . "HANGUL SYLLABLE TIKEUT YEO NIEUN")
    (13797 . "HANGUL SYLLABLE TIKEUT YEO RIEUL")
    (13798 . "HANGUL SYLLABLE TIKEUT YEO SSANGSIOS")
    (13799 . "HANGUL SYLLABLE TIKEUT YEO IEUNG")
    (13800 . "HANGUL SYLLABLE TIKEUT YE")
    (13801 . "HANGUL SYLLABLE TIKEUT YE NIEUN")
    (13802 . "HANGUL SYLLABLE TIKEUT O")
    (13803 . "HANGUL SYLLABLE TIKEUT O KIYEOK")
    (13804 . "HANGUL SYLLABLE TIKEUT O NIEUN")
    (13805 . "HANGUL SYLLABLE TIKEUT O TIKEUT")
    (13806 . "HANGUL SYLLABLE TIKEUT O RIEUL")
    (13807 . "HANGUL SYLLABLE TIKEUT O RIEUL-MIEUM")
    (13808 . "HANGUL SYLLABLE TIKEUT O RIEUL-SIOS")
    (13809 . "HANGUL SYLLABLE TIKEUT O MIEUM")
    (13810 . "HANGUL SYLLABLE TIKEUT O PIEUP")
    (13811 . "HANGUL SYLLABLE TIKEUT O SIOS")
    (13812 . "HANGUL SYLLABLE TIKEUT O IEUNG")
    (13813 . "HANGUL SYLLABLE TIKEUT O CHIEUCH")
    (13814 . "HANGUL SYLLABLE TIKEUT O THIEUTH")
    (13815 . "HANGUL SYLLABLE TIKEUT WA")
    (13816 . "HANGUL SYLLABLE TIKEUT WA NIEUN")
    (13817 . "HANGUL SYLLABLE TIKEUT WA RIEUL")
    (13818 . "HANGUL SYLLABLE TIKEUT WAE")
    (13819 . "HANGUL SYLLABLE TIKEUT WAE SSANGSIOS")
    (13820 . "HANGUL SYLLABLE TIKEUT OE")
    (13821 . "HANGUL SYLLABLE TIKEUT OE NIEUN")
    (13822 . "HANGUL SYLLABLE TIKEUT OE RIEUL")
    (13823 . "HANGUL SYLLABLE TIKEUT OE MIEUM")
    (13824 . "HANGUL SYLLABLE TIKEUT OE PIEUP")
    (13825 . "HANGUL SYLLABLE TIKEUT OE SIOS")
    (13826 . "HANGUL SYLLABLE TIKEUT YO")
    (13827 . "HANGUL SYLLABLE TIKEUT U")
    (13828 . "HANGUL SYLLABLE TIKEUT U KIYEOK")
    (13829 . "HANGUL SYLLABLE TIKEUT U NIEUN")
    (13830 . "HANGUL SYLLABLE TIKEUT U RIEUL")
    (13831 . "HANGUL SYLLABLE TIKEUT U MIEUM")
    (13832 . "HANGUL SYLLABLE TIKEUT U PIEUP")
    (13833 . "HANGUL SYLLABLE TIKEUT U SIOS")
    (13834 . "HANGUL SYLLABLE TIKEUT U IEUNG")
    (13835 . "HANGUL SYLLABLE TIKEUT WEO")
    (13836 . "HANGUL SYLLABLE TIKEUT WEO SSANGSIOS")
    (13837 . "HANGUL SYLLABLE TIKEUT WE")
    (13838 . "HANGUL SYLLABLE TIKEUT WE IEUNG")
    (13839 . "HANGUL SYLLABLE TIKEUT WI")
    (13840 . "HANGUL SYLLABLE TIKEUT WI NIEUN")
    (13841 . "HANGUL SYLLABLE TIKEUT WI RIEUL")
    (13842 . "HANGUL SYLLABLE TIKEUT WI PIEUP")
    (13843 . "HANGUL SYLLABLE TIKEUT WI SIOS")
    (13844 . "HANGUL SYLLABLE TIKEUT WI IEUNG")
    (13845 . "HANGUL SYLLABLE TIKEUT YU")
    (13846 . "HANGUL SYLLABLE TIKEUT YU NIEUN")
    (13847 . "HANGUL SYLLABLE TIKEUT YU RIEUL")
    (13848 . "HANGUL SYLLABLE TIKEUT YU MIEUM")
    (13849 . "HANGUL SYLLABLE TIKEUT YU IEUNG")
    (13850 . "HANGUL SYLLABLE TIKEUT EU")
    (13851 . "HANGUL SYLLABLE TIKEUT EU KIYEOK")
    (13852 . "HANGUL SYLLABLE TIKEUT EU NIEUN")
    (13853 . "HANGUL SYLLABLE TIKEUT EU TIKEUT")
    (13854 . "HANGUL SYLLABLE TIKEUT EU RIEUL")
    (13855 . "HANGUL SYLLABLE TIKEUT EU RIEUL-MIEUM")
    (13856 . "HANGUL SYLLABLE TIKEUT EU MIEUM")
    (13857 . "HANGUL SYLLABLE TIKEUT EU PIEUP")
    (13858 . "HANGUL SYLLABLE TIKEUT EU SIOS")
    (13859 . "HANGUL SYLLABLE TIKEUT EU IEUNG")
    (13860 . "HANGUL SYLLABLE TIKEUT YI")
    (13861 . "HANGUL SYLLABLE TIKEUT I")
    (13862 . "HANGUL SYLLABLE TIKEUT I KIYEOK")
    (13863 . "HANGUL SYLLABLE TIKEUT I NIEUN")
    (13864 . "HANGUL SYLLABLE TIKEUT I TIKEUT")
    (13865 . "HANGUL SYLLABLE TIKEUT I RIEUL")
    (13866 . "HANGUL SYLLABLE TIKEUT I MIEUM")
    (13867 . "HANGUL SYLLABLE TIKEUT I PIEUP")
    (13868 . "HANGUL SYLLABLE TIKEUT I SIOS")
    (13869 . "HANGUL SYLLABLE TIKEUT I SSANGSIOS")
    (13870 . "HANGUL SYLLABLE TIKEUT I IEUNG")
    (13871 . "HANGUL SYLLABLE TIKEUT I CIEUC")
    (13872 . "HANGUL SYLLABLE SSANGTIKEUT A")
    (13873 . "HANGUL SYLLABLE SSANGTIKEUT A KIYEOK")
    (13874 . "HANGUL SYLLABLE SSANGTIKEUT A NIEUN")
    (13875 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL")
    (13876 . "HANGUL SYLLABLE SSANGTIKEUT A MIEUM")
    (13877 . "HANGUL SYLLABLE SSANGTIKEUT A PIEUP")
    (13878 . "HANGUL SYLLABLE SSANGTIKEUT A SIOS")
    (13879 . "HANGUL SYLLABLE SSANGTIKEUT A SSANGSIOS")
    (13880 . "HANGUL SYLLABLE SSANGTIKEUT A IEUNG")
    (13881 . "HANGUL SYLLABLE SSANGTIKEUT A HIEUH")
    (13882 . "HANGUL SYLLABLE SSANGTIKEUT AE")
    (13883 . "HANGUL SYLLABLE SSANGTIKEUT AE KIYEOK")
    (13884 . "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN")
    (13885 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL")
    (13886 . "HANGUL SYLLABLE SSANGTIKEUT AE MIEUM")
    (13887 . "HANGUL SYLLABLE SSANGTIKEUT AE PIEUP")
    (13888 . "HANGUL SYLLABLE SSANGTIKEUT AE SIOS")
    (13889 . "HANGUL SYLLABLE SSANGTIKEUT AE SSANGSIOS")
    (13890 . "HANGUL SYLLABLE SSANGTIKEUT AE IEUNG")
    (13891 . "HANGUL SYLLABLE SSANGTIKEUT EO")
    (13892 . "HANGUL SYLLABLE SSANGTIKEUT EO KIYEOK")
    (13893 . "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN")
    (13894 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL")
    (13895 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-MIEUM")
    (13896 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-PIEUP")
    (13897 . "HANGUL SYLLABLE SSANGTIKEUT EO MIEUM")
    (13898 . "HANGUL SYLLABLE SSANGTIKEUT EO PIEUP")
    (13899 . "HANGUL SYLLABLE SSANGTIKEUT EO SIOS")
    (13900 . "HANGUL SYLLABLE SSANGTIKEUT EO SSANGSIOS")
    (13901 . "HANGUL SYLLABLE SSANGTIKEUT EO IEUNG")
    (13902 . "HANGUL SYLLABLE SSANGTIKEUT EO HIEUH")
    (13903 . "HANGUL SYLLABLE SSANGTIKEUT E")
    (13904 . "HANGUL SYLLABLE SSANGTIKEUT E KIYEOK")
    (13905 . "HANGUL SYLLABLE SSANGTIKEUT E NIEUN")
    (13906 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL")
    (13907 . "HANGUL SYLLABLE SSANGTIKEUT E MIEUM")
    (13908 . "HANGUL SYLLABLE SSANGTIKEUT E PIEUP")
    (13909 . "HANGUL SYLLABLE SSANGTIKEUT E SIOS")
    (13910 . "HANGUL SYLLABLE SSANGTIKEUT E SSANGSIOS")
    (13911 . "HANGUL SYLLABLE SSANGTIKEUT E IEUNG")
    (13912 . "HANGUL SYLLABLE SSANGTIKEUT YEO")
    (13913 . "HANGUL SYLLABLE SSANGTIKEUT YEO SSANGSIOS")
    (13914 . "HANGUL SYLLABLE SSANGTIKEUT O")
    (13915 . "HANGUL SYLLABLE SSANGTIKEUT O KIYEOK")
    (13916 . "HANGUL SYLLABLE SSANGTIKEUT O NIEUN")
    (13917 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL")
    (13918 . "HANGUL SYLLABLE SSANGTIKEUT O IEUNG")
    (13919 . "HANGUL SYLLABLE SSANGTIKEUT WA")
    (13920 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL")
    (13921 . "HANGUL SYLLABLE SSANGTIKEUT WAE")
    (13922 . "HANGUL SYLLABLE SSANGTIKEUT OE")
    (13923 . "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN")
    (13924 . "HANGUL SYLLABLE SSANGTIKEUT U")
    (13925 . "HANGUL SYLLABLE SSANGTIKEUT U KIYEOK")
    (13926 . "HANGUL SYLLABLE SSANGTIKEUT U NIEUN")
    (13927 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL")
    (13928 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-HIEUH")
    (13929 . "HANGUL SYLLABLE SSANGTIKEUT U MIEUM")
    (13930 . "HANGUL SYLLABLE SSANGTIKEUT U IEUNG")
    (13931 . "HANGUL SYLLABLE SSANGTIKEUT WE")
    (13932 . "HANGUL SYLLABLE SSANGTIKEUT WI")
    (13933 . "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN")
    (13934 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL")
    (13935 . "HANGUL SYLLABLE SSANGTIKEUT WI MIEUM")
    (13936 . "HANGUL SYLLABLE SSANGTIKEUT WI PIEUP")
    (13937 . "HANGUL SYLLABLE SSANGTIKEUT WI IEUNG")
    (13938 . "HANGUL SYLLABLE SSANGTIKEUT EU")
    (13939 . "HANGUL SYLLABLE SSANGTIKEUT EU KIYEOK")
    (13940 . "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN")
    (13941 . "HANGUL SYLLABLE SSANGTIKEUT EU TIKEUT")
    (13942 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL")
    (13943 . "HANGUL SYLLABLE SSANGTIKEUT EU MIEUM")
    (13944 . "HANGUL SYLLABLE SSANGTIKEUT EU PIEUP")
    (13945 . "HANGUL SYLLABLE SSANGTIKEUT EU SIOS")
    (13946 . "HANGUL SYLLABLE SSANGTIKEUT YI")
    (13947 . "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN")
    (13948 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL")
    (13949 . "HANGUL SYLLABLE SSANGTIKEUT YI MIEUM")
    (13950 . "HANGUL SYLLABLE SSANGTIKEUT YI PIEUP")
    (13951 . "HANGUL SYLLABLE SSANGTIKEUT I")
    (13952 . "HANGUL SYLLABLE SSANGTIKEUT I NIEUN")
    (13953 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL")
    (13954 . "HANGUL SYLLABLE SSANGTIKEUT I MIEUM")
    (13955 . "HANGUL SYLLABLE SSANGTIKEUT I PIEUP")
    (13956 . "HANGUL SYLLABLE SSANGTIKEUT I SIOS")
    (13957 . "HANGUL SYLLABLE SSANGTIKEUT I IEUNG")
    (13958 . "HANGUL SYLLABLE RIEUL A")
    (13959 . "HANGUL SYLLABLE RIEUL A KIYEOK")
    (13960 . "HANGUL SYLLABLE RIEUL A NIEUN")
    (13961 . "HANGUL SYLLABLE RIEUL A RIEUL")
    (13962 . "HANGUL SYLLABLE RIEUL A MIEUM")
    (13963 . "HANGUL SYLLABLE RIEUL A PIEUP")
    (13964 . "HANGUL SYLLABLE RIEUL A SIOS")
    (13965 . "HANGUL SYLLABLE RIEUL A SSANGSIOS")
    (13966 . "HANGUL SYLLABLE RIEUL A IEUNG")
    (13967 . "HANGUL SYLLABLE RIEUL A CIEUC")
    (13968 . "HANGUL SYLLABLE RIEUL A PHIEUPH")
    (13969 . "HANGUL SYLLABLE RIEUL A HIEUH")
    (13970 . "HANGUL SYLLABLE RIEUL AE")
    (13971 . "HANGUL SYLLABLE RIEUL AE KIYEOK")
    (13972 . "HANGUL SYLLABLE RIEUL AE NIEUN")
    (13973 . "HANGUL SYLLABLE RIEUL AE RIEUL")
    (13974 . "HANGUL SYLLABLE RIEUL AE MIEUM")
    (13975 . "HANGUL SYLLABLE RIEUL AE PIEUP")
    (13976 . "HANGUL SYLLABLE RIEUL AE SIOS")
    (13977 . "HANGUL SYLLABLE RIEUL AE SSANGSIOS")
    (13978 . "HANGUL SYLLABLE RIEUL AE IEUNG")
    (13979 . "HANGUL SYLLABLE RIEUL YA")
    (13980 . "HANGUL SYLLABLE RIEUL YA KIYEOK")
    (13981 . "HANGUL SYLLABLE RIEUL YA NIEUN")
    (13982 . "HANGUL SYLLABLE RIEUL YA SIOS")
    (13983 . "HANGUL SYLLABLE RIEUL YA IEUNG")
    (13984 . "HANGUL SYLLABLE RIEUL EO")
    (13985 . "HANGUL SYLLABLE RIEUL EO KIYEOK")
    (13986 . "HANGUL SYLLABLE RIEUL EO NIEUN")
    (13987 . "HANGUL SYLLABLE RIEUL EO RIEUL")
    (13988 . "HANGUL SYLLABLE RIEUL EO MIEUM")
    (13989 . "HANGUL SYLLABLE RIEUL EO PIEUP")
    (13990 . "HANGUL SYLLABLE RIEUL EO SIOS")
    (13991 . "HANGUL SYLLABLE RIEUL EO SSANGSIOS")
    (13992 . "HANGUL SYLLABLE RIEUL EO IEUNG")
    (13993 . "HANGUL SYLLABLE RIEUL EO HIEUH")
    (13994 . "HANGUL SYLLABLE RIEUL E")
    (13995 . "HANGUL SYLLABLE RIEUL E KIYEOK")
    (13996 . "HANGUL SYLLABLE RIEUL E NIEUN")
    (13997 . "HANGUL SYLLABLE RIEUL E RIEUL")
    (13998 . "HANGUL SYLLABLE RIEUL E MIEUM")
    (13999 . "HANGUL SYLLABLE RIEUL E PIEUP")
    (14000 . "HANGUL SYLLABLE RIEUL E SIOS")
    (14001 . "HANGUL SYLLABLE RIEUL E IEUNG")
    (14002 . "HANGUL SYLLABLE RIEUL YEO")
    (14003 . "HANGUL SYLLABLE RIEUL YEO KIYEOK")
    (14004 . "HANGUL SYLLABLE RIEUL YEO NIEUN")
    (14005 . "HANGUL SYLLABLE RIEUL YEO RIEUL")
    (14006 . "HANGUL SYLLABLE RIEUL YEO MIEUM")
    (14007 . "HANGUL SYLLABLE RIEUL YEO PIEUP")
    (14008 . "HANGUL SYLLABLE RIEUL YEO SIOS")
    (14009 . "HANGUL SYLLABLE RIEUL YEO SSANGSIOS")
    (14010 . "HANGUL SYLLABLE RIEUL YEO IEUNG")
    (14011 . "HANGUL SYLLABLE RIEUL YE")
    (14012 . "HANGUL SYLLABLE RIEUL YE NIEUN")
    (14013 . "HANGUL SYLLABLE RIEUL YE PIEUP")
    (14014 . "HANGUL SYLLABLE RIEUL YE SIOS")
    (14015 . "HANGUL SYLLABLE RIEUL O")
    (14016 . "HANGUL SYLLABLE RIEUL O KIYEOK")
    (14017 . "HANGUL SYLLABLE RIEUL O NIEUN")
    (14018 . "HANGUL SYLLABLE RIEUL O RIEUL")
    (14019 . "HANGUL SYLLABLE RIEUL O MIEUM")
    (14020 . "HANGUL SYLLABLE RIEUL O PIEUP")
    (14021 . "HANGUL SYLLABLE RIEUL O SIOS")
    (14022 . "HANGUL SYLLABLE RIEUL O IEUNG")
    (14023 . "HANGUL SYLLABLE RIEUL WA")
    (14024 . "HANGUL SYLLABLE RIEUL WA NIEUN")
    (14025 . "HANGUL SYLLABLE RIEUL WA IEUNG")
    (14026 . "HANGUL SYLLABLE RIEUL WAE SSANGSIOS")
    (14027 . "HANGUL SYLLABLE RIEUL OE")
    (14028 . "HANGUL SYLLABLE RIEUL OE NIEUN")
    (14029 . "HANGUL SYLLABLE RIEUL OE RIEUL")
    (14030 . "HANGUL SYLLABLE RIEUL OE MIEUM")
    (14031 . "HANGUL SYLLABLE RIEUL OE PIEUP")
    (14032 . "HANGUL SYLLABLE RIEUL OE SIOS")
    (14033 . "HANGUL SYLLABLE RIEUL OE IEUNG")
    (14034 . "HANGUL SYLLABLE RIEUL YO")
    (14035 . "HANGUL SYLLABLE RIEUL YO NIEUN")
    (14036 . "HANGUL SYLLABLE RIEUL YO RIEUL")
    (14037 . "HANGUL SYLLABLE RIEUL YO PIEUP")
    (14038 . "HANGUL SYLLABLE RIEUL YO SIOS")
    (14039 . "HANGUL SYLLABLE RIEUL YO IEUNG")
    (14040 . "HANGUL SYLLABLE RIEUL U")
    (14041 . "HANGUL SYLLABLE RIEUL U KIYEOK")
    (14042 . "HANGUL SYLLABLE RIEUL U NIEUN")
    (14043 . "HANGUL SYLLABLE RIEUL U RIEUL")
    (14044 . "HANGUL SYLLABLE RIEUL U MIEUM")
    (14045 . "HANGUL SYLLABLE RIEUL U PIEUP")
    (14046 . "HANGUL SYLLABLE RIEUL U SIOS")
    (14047 . "HANGUL SYLLABLE RIEUL U IEUNG")
    (14048 . "HANGUL SYLLABLE RIEUL WEO")
    (14049 . "HANGUL SYLLABLE RIEUL WEO SSANGSIOS")
    (14050 . "HANGUL SYLLABLE RIEUL WE")
    (14051 . "HANGUL SYLLABLE RIEUL WI")
    (14052 . "HANGUL SYLLABLE RIEUL WI KIYEOK")
    (14053 . "HANGUL SYLLABLE RIEUL WI NIEUN")
    (14054 . "HANGUL SYLLABLE RIEUL WI RIEUL")
    (14055 . "HANGUL SYLLABLE RIEUL WI MIEUM")
    (14056 . "HANGUL SYLLABLE RIEUL WI SIOS")
    (14057 . "HANGUL SYLLABLE RIEUL WI IEUNG")
    (14058 . "HANGUL SYLLABLE RIEUL YU")
    (14059 . "HANGUL SYLLABLE RIEUL YU KIYEOK")
    (14060 . "HANGUL SYLLABLE RIEUL YU NIEUN")
    (14061 . "HANGUL SYLLABLE RIEUL YU RIEUL")
    (14062 . "HANGUL SYLLABLE RIEUL YU MIEUM")
    (14063 . "HANGUL SYLLABLE RIEUL YU PIEUP")
    (14064 . "HANGUL SYLLABLE RIEUL YU SIOS")
    (14065 . "HANGUL SYLLABLE RIEUL YU IEUNG")
    (14066 . "HANGUL SYLLABLE RIEUL EU")
    (14067 . "HANGUL SYLLABLE RIEUL EU KIYEOK")
    (14068 . "HANGUL SYLLABLE RIEUL EU NIEUN")
    (14069 . "HANGUL SYLLABLE RIEUL EU RIEUL")
    (14070 . "HANGUL SYLLABLE RIEUL EU MIEUM")
    (14071 . "HANGUL SYLLABLE RIEUL EU PIEUP")
    (14072 . "HANGUL SYLLABLE RIEUL EU SIOS")
    (14073 . "HANGUL SYLLABLE RIEUL EU IEUNG")
    (14074 . "HANGUL SYLLABLE RIEUL EU CIEUC")
    (14075 . "HANGUL SYLLABLE RIEUL EU THIEUTH")
    (14076 . "HANGUL SYLLABLE RIEUL EU PHIEUPH")
    (14077 . "HANGUL SYLLABLE RIEUL I")
    (14078 . "HANGUL SYLLABLE RIEUL I KIYEOK")
    (14079 . "HANGUL SYLLABLE RIEUL I NIEUN")
    (14080 . "HANGUL SYLLABLE RIEUL I RIEUL")
    (14081 . "HANGUL SYLLABLE RIEUL I MIEUM")
    (14082 . "HANGUL SYLLABLE RIEUL I PIEUP")
    (14083 . "HANGUL SYLLABLE RIEUL I SIOS")
    (14084 . "HANGUL SYLLABLE RIEUL I IEUNG")
    (14085 . "HANGUL SYLLABLE MIEUM A")
    (14086 . "HANGUL SYLLABLE MIEUM A KIYEOK")
    (14087 . "HANGUL SYLLABLE MIEUM A NIEUN")
    (14088 . "HANGUL SYLLABLE MIEUM A NIEUN-HIEUH")
    (14089 . "HANGUL SYLLABLE MIEUM A TIKEUT")
    (14090 . "HANGUL SYLLABLE MIEUM A RIEUL")
    (14091 . "HANGUL SYLLABLE MIEUM A RIEUL-KIYEOK")
    (14092 . "HANGUL SYLLABLE MIEUM A RIEUL-MIEUM")
    (14093 . "HANGUL SYLLABLE MIEUM A MIEUM")
    (14094 . "HANGUL SYLLABLE MIEUM A PIEUP")
    (14095 . "HANGUL SYLLABLE MIEUM A SIOS")
    (14096 . "HANGUL SYLLABLE MIEUM A IEUNG")
    (14097 . "HANGUL SYLLABLE MIEUM A CIEUC")
    (14098 . "HANGUL SYLLABLE MIEUM A THIEUTH")
    (14099 . "HANGUL SYLLABLE MIEUM A HIEUH")
    (14100 . "HANGUL SYLLABLE MIEUM AE")
    (14101 . "HANGUL SYLLABLE MIEUM AE KIYEOK")
    (14102 . "HANGUL SYLLABLE MIEUM AE NIEUN")
    (14103 . "HANGUL SYLLABLE MIEUM AE RIEUL")
    (14104 . "HANGUL SYLLABLE MIEUM AE MIEUM")
    (14105 . "HANGUL SYLLABLE MIEUM AE PIEUP")
    (14106 . "HANGUL SYLLABLE MIEUM AE SIOS")
    (14107 . "HANGUL SYLLABLE MIEUM AE SSANGSIOS")
    (14108 . "HANGUL SYLLABLE MIEUM AE IEUNG")
    (14109 . "HANGUL SYLLABLE MIEUM AE CIEUC")
    (14110 . "HANGUL SYLLABLE MIEUM YA")
    (14111 . "HANGUL SYLLABLE MIEUM YA KIYEOK")
    (14112 . "HANGUL SYLLABLE MIEUM YA RIEUL")
    (14113 . "HANGUL SYLLABLE MIEUM YA IEUNG")
    (14114 . "HANGUL SYLLABLE MIEUM EO")
    (14115 . "HANGUL SYLLABLE MIEUM EO KIYEOK")
    (14116 . "HANGUL SYLLABLE MIEUM EO NIEUN")
    (14117 . "HANGUL SYLLABLE MIEUM EO RIEUL")
    (14118 . "HANGUL SYLLABLE MIEUM EO RIEUL-MIEUM")
    (14119 . "HANGUL SYLLABLE MIEUM EO MIEUM")
    (14120 . "HANGUL SYLLABLE MIEUM EO PIEUP")
    (14121 . "HANGUL SYLLABLE MIEUM EO SIOS")
    (14122 . "HANGUL SYLLABLE MIEUM EO IEUNG")
    (14123 . "HANGUL SYLLABLE MIEUM EO CIEUC")
    (14124 . "HANGUL SYLLABLE MIEUM EO HIEUH")
    (14125 . "HANGUL SYLLABLE MIEUM E")
    (14126 . "HANGUL SYLLABLE MIEUM E KIYEOK")
    (14127 . "HANGUL SYLLABLE MIEUM E NIEUN")
    (14128 . "HANGUL SYLLABLE MIEUM E RIEUL")
    (14129 . "HANGUL SYLLABLE MIEUM E MIEUM")
    (14130 . "HANGUL SYLLABLE MIEUM E PIEUP")
    (14131 . "HANGUL SYLLABLE MIEUM E SIOS")
    (14132 . "HANGUL SYLLABLE MIEUM E SSANGSIOS")
    (14133 . "HANGUL SYLLABLE MIEUM E IEUNG")
    (14134 . "HANGUL SYLLABLE MIEUM YEO")
    (14135 . "HANGUL SYLLABLE MIEUM YEO KIYEOK")
    (14136 . "HANGUL SYLLABLE MIEUM YEO NIEUN")
    (14137 . "HANGUL SYLLABLE MIEUM YEO RIEUL")
    (14138 . "HANGUL SYLLABLE MIEUM YEO SIOS")
    (14139 . "HANGUL SYLLABLE MIEUM YEO SSANGSIOS")
    (14140 . "HANGUL SYLLABLE MIEUM YEO IEUNG")
    (14141 . "HANGUL SYLLABLE MIEUM YEO CHIEUCH")
    (14142 . "HANGUL SYLLABLE MIEUM YE")
    (14143 . "HANGUL SYLLABLE MIEUM O")
    (14144 . "HANGUL SYLLABLE MIEUM O KIYEOK")
    (14145 . "HANGUL SYLLABLE MIEUM O KIYEOK-SIOS")
    (14146 . "HANGUL SYLLABLE MIEUM O NIEUN")
    (14147 . "HANGUL SYLLABLE MIEUM O RIEUL")
    (14148 . "HANGUL SYLLABLE MIEUM O RIEUL-MIEUM")
    (14149 . "HANGUL SYLLABLE MIEUM O MIEUM")
    (14150 . "HANGUL SYLLABLE MIEUM O PIEUP")
    (14151 . "HANGUL SYLLABLE MIEUM O SIOS")
    (14152 . "HANGUL SYLLABLE MIEUM O IEUNG")
    (14153 . "HANGUL SYLLABLE MIEUM WA")
    (14154 . "HANGUL SYLLABLE MIEUM WA NIEUN")
    (14155 . "HANGUL SYLLABLE MIEUM WA SSANGSIOS")
    (14156 . "HANGUL SYLLABLE MIEUM WA IEUNG")
    (14157 . "HANGUL SYLLABLE MIEUM OE")
    (14158 . "HANGUL SYLLABLE MIEUM OE NIEUN")
    (14159 . "HANGUL SYLLABLE MIEUM OE RIEUL")
    (14160 . "HANGUL SYLLABLE MIEUM OE PIEUP")
    (14161 . "HANGUL SYLLABLE MIEUM OE SIOS")
    (14162 . "HANGUL SYLLABLE MIEUM OE IEUNG")
    (14163 . "HANGUL SYLLABLE MIEUM YO")
    (14164 . "HANGUL SYLLABLE MIEUM YO NIEUN")
    (14165 . "HANGUL SYLLABLE MIEUM YO RIEUL")
    (14166 . "HANGUL SYLLABLE MIEUM YO PIEUP")
    (14167 . "HANGUL SYLLABLE MIEUM YO SIOS")
    (14168 . "HANGUL SYLLABLE MIEUM U")
    (14169 . "HANGUL SYLLABLE MIEUM U KIYEOK")
    (14170 . "HANGUL SYLLABLE MIEUM U SSANGKIYEOK")
    (14171 . "HANGUL SYLLABLE MIEUM U NIEUN")
    (14172 . "HANGUL SYLLABLE MIEUM U TIKEUT")
    (14173 . "HANGUL SYLLABLE MIEUM U RIEUL")
    (14174 . "HANGUL SYLLABLE MIEUM U RIEUL-KIYEOK")
    (14175 . "HANGUL SYLLABLE MIEUM U RIEUL-MIEUM")
    (14176 . "HANGUL SYLLABLE MIEUM U MIEUM")
    (14177 . "HANGUL SYLLABLE MIEUM U PIEUP")
    (14178 . "HANGUL SYLLABLE MIEUM U SIOS")
    (14179 . "HANGUL SYLLABLE MIEUM U IEUNG")
    (14180 . "HANGUL SYLLABLE MIEUM U THIEUTH")
    (14181 . "HANGUL SYLLABLE MIEUM U HIEUH")
    (14182 . "HANGUL SYLLABLE MIEUM WEO")
    (14183 . "HANGUL SYLLABLE MIEUM WEO NIEUN")
    (14184 . "HANGUL SYLLABLE MIEUM WEO RIEUL")
    (14185 . "HANGUL SYLLABLE MIEUM WEO PIEUP")
    (14186 . "HANGUL SYLLABLE MIEUM WEO SIOS")
    (14187 . "HANGUL SYLLABLE MIEUM WE")
    (14188 . "HANGUL SYLLABLE MIEUM WI")
    (14189 . "HANGUL SYLLABLE MIEUM WI NIEUN")
    (14190 . "HANGUL SYLLABLE MIEUM WI RIEUL")
    (14191 . "HANGUL SYLLABLE MIEUM YU")
    (14192 . "HANGUL SYLLABLE MIEUM YU NIEUN")
    (14193 . "HANGUL SYLLABLE MIEUM YU RIEUL")
    (14194 . "HANGUL SYLLABLE MIEUM YU MIEUM")
    (14195 . "HANGUL SYLLABLE MIEUM YU SIOS")
    (14196 . "HANGUL SYLLABLE MIEUM EU")
    (14197 . "HANGUL SYLLABLE MIEUM EU NIEUN")
    (14198 . "HANGUL SYLLABLE MIEUM EU RIEUL")
    (14199 . "HANGUL SYLLABLE MIEUM EU MIEUM")
    (14200 . "HANGUL SYLLABLE MIEUM EU SIOS")
    (14201 . "HANGUL SYLLABLE MIEUM I")
    (14202 . "HANGUL SYLLABLE MIEUM I KIYEOK")
    (14203 . "HANGUL SYLLABLE MIEUM I NIEUN")
    (14204 . "HANGUL SYLLABLE MIEUM I TIKEUT")
    (14205 . "HANGUL SYLLABLE MIEUM I RIEUL")
    (14206 . "HANGUL SYLLABLE MIEUM I RIEUL-MIEUM")
    (14207 . "HANGUL SYLLABLE MIEUM I MIEUM")
    (14208 . "HANGUL SYLLABLE MIEUM I PIEUP")
    (14209 . "HANGUL SYLLABLE MIEUM I SIOS")
    (14210 . "HANGUL SYLLABLE MIEUM I SSANGSIOS")
    (14211 . "HANGUL SYLLABLE MIEUM I IEUNG")
    (14212 . "HANGUL SYLLABLE MIEUM I CHIEUCH")
    (14213 . "HANGUL SYLLABLE MIEUM I THIEUTH")
    (14214 . "HANGUL SYLLABLE PIEUP A")
    (14215 . "HANGUL SYLLABLE PIEUP A KIYEOK")
    (14216 . "HANGUL SYLLABLE PIEUP A SSANGKIYEOK")
    (14217 . "HANGUL SYLLABLE PIEUP A KIYEOK-SIOS")
    (14218 . "HANGUL SYLLABLE PIEUP A NIEUN")
    (14219 . "HANGUL SYLLABLE PIEUP A TIKEUT")
    (14220 . "HANGUL SYLLABLE PIEUP A RIEUL")
    (14221 . "HANGUL SYLLABLE PIEUP A RIEUL-KIYEOK")
    (14222 . "HANGUL SYLLABLE PIEUP A RIEUL-MIEUM")
    (14223 . "HANGUL SYLLABLE PIEUP A RIEUL-PIEUP")
    (14224 . "HANGUL SYLLABLE PIEUP A MIEUM")
    (14225 . "HANGUL SYLLABLE PIEUP A PIEUP")
    (14226 . "HANGUL SYLLABLE PIEUP A SIOS")
    (14227 . "HANGUL SYLLABLE PIEUP A IEUNG")
    (14228 . "HANGUL SYLLABLE PIEUP A THIEUTH")
    (14229 . "HANGUL SYLLABLE PIEUP AE")
    (14230 . "HANGUL SYLLABLE PIEUP AE KIYEOK")
    (14231 . "HANGUL SYLLABLE PIEUP AE NIEUN")
    (14232 . "HANGUL SYLLABLE PIEUP AE RIEUL")
    (14233 . "HANGUL SYLLABLE PIEUP AE MIEUM")
    (14234 . "HANGUL SYLLABLE PIEUP AE PIEUP")
    (14235 . "HANGUL SYLLABLE PIEUP AE SIOS")
    (14236 . "HANGUL SYLLABLE PIEUP AE SSANGSIOS")
    (14237 . "HANGUL SYLLABLE PIEUP AE IEUNG")
    (14238 . "HANGUL SYLLABLE PIEUP AE THIEUTH")
    (14239 . "HANGUL SYLLABLE PIEUP YA")
    (14240 . "HANGUL SYLLABLE PIEUP YA KIYEOK")
    (14241 . "HANGUL SYLLABLE PIEUP YA NIEUN")
    (14242 . "HANGUL SYLLABLE PIEUP YA PIEUP")
    (14243 . "HANGUL SYLLABLE PIEUP EO")
    (14244 . "HANGUL SYLLABLE PIEUP EO KIYEOK")
    (14245 . "HANGUL SYLLABLE PIEUP EO NIEUN")
    (14246 . "HANGUL SYLLABLE PIEUP EO TIKEUT")
    (14247 . "HANGUL SYLLABLE PIEUP EO RIEUL")
    (14248 . "HANGUL SYLLABLE PIEUP EO RIEUL-MIEUM")
    (14249 . "HANGUL SYLLABLE PIEUP EO MIEUM")
    (14250 . "HANGUL SYLLABLE PIEUP EO PIEUP")
    (14251 . "HANGUL SYLLABLE PIEUP EO SIOS")
    (14252 . "HANGUL SYLLABLE PIEUP EO IEUNG")
    (14253 . "HANGUL SYLLABLE PIEUP EO CIEUC")
    (14254 . "HANGUL SYLLABLE PIEUP E")
    (14255 . "HANGUL SYLLABLE PIEUP E KIYEOK")
    (14256 . "HANGUL SYLLABLE PIEUP E NIEUN")
    (14257 . "HANGUL SYLLABLE PIEUP E TIKEUT")
    (14258 . "HANGUL SYLLABLE PIEUP E RIEUL")
    (14259 . "HANGUL SYLLABLE PIEUP E MIEUM")
    (14260 . "HANGUL SYLLABLE PIEUP E PIEUP")
    (14261 . "HANGUL SYLLABLE PIEUP E SIOS")
    (14262 . "HANGUL SYLLABLE PIEUP E SSANGSIOS")
    (14263 . "HANGUL SYLLABLE PIEUP E IEUNG")
    (14264 . "HANGUL SYLLABLE PIEUP YEO")
    (14265 . "HANGUL SYLLABLE PIEUP YEO KIYEOK")
    (14266 . "HANGUL SYLLABLE PIEUP YEO NIEUN")
    (14267 . "HANGUL SYLLABLE PIEUP YEO RIEUL")
    (14268 . "HANGUL SYLLABLE PIEUP YEO PIEUP")
    (14269 . "HANGUL SYLLABLE PIEUP YEO SIOS")
    (14270 . "HANGUL SYLLABLE PIEUP YEO SSANGSIOS")
    (14271 . "HANGUL SYLLABLE PIEUP YEO IEUNG")
    (14272 . "HANGUL SYLLABLE PIEUP YEO THIEUTH")
    (14273 . "HANGUL SYLLABLE PIEUP YE")
    (14274 . "HANGUL SYLLABLE PIEUP YE NIEUN")
    (14275 . "HANGUL SYLLABLE PIEUP O")
    (14276 . "HANGUL SYLLABLE PIEUP O KIYEOK")
    (14277 . "HANGUL SYLLABLE PIEUP O SSANGKIYEOK")
    (14278 . "HANGUL SYLLABLE PIEUP O NIEUN")
    (14279 . "HANGUL SYLLABLE PIEUP O RIEUL")
    (14280 . "HANGUL SYLLABLE PIEUP O MIEUM")
    (14281 . "HANGUL SYLLABLE PIEUP O PIEUP")
    (14282 . "HANGUL SYLLABLE PIEUP O SIOS")
    (14283 . "HANGUL SYLLABLE PIEUP O IEUNG")
    (14284 . "HANGUL SYLLABLE PIEUP WA")
    (14285 . "HANGUL SYLLABLE PIEUP WA NIEUN")
    (14286 . "HANGUL SYLLABLE PIEUP WA SSANGSIOS")
    (14287 . "HANGUL SYLLABLE PIEUP WAE")
    (14288 . "HANGUL SYLLABLE PIEUP WAE SSANGSIOS")
    (14289 . "HANGUL SYLLABLE PIEUP OE")
    (14290 . "HANGUL SYLLABLE PIEUP OE KIYEOK")
    (14291 . "HANGUL SYLLABLE PIEUP OE NIEUN")
    (14292 . "HANGUL SYLLABLE PIEUP OE RIEUL")
    (14293 . "HANGUL SYLLABLE PIEUP OE MIEUM")
    (14294 . "HANGUL SYLLABLE PIEUP OE PIEUP")
    (14295 . "HANGUL SYLLABLE PIEUP YO")
    (14296 . "HANGUL SYLLABLE PIEUP YO NIEUN")
    (14297 . "HANGUL SYLLABLE PIEUP U")
    (14298 . "HANGUL SYLLABLE PIEUP U KIYEOK")
    (14299 . "HANGUL SYLLABLE PIEUP U NIEUN")
    (14300 . "HANGUL SYLLABLE PIEUP U TIKEUT")
    (14301 . "HANGUL SYLLABLE PIEUP U RIEUL")
    (14302 . "HANGUL SYLLABLE PIEUP U RIEUL-KIYEOK")
    (14303 . "HANGUL SYLLABLE PIEUP U RIEUL-MIEUM")
    (14304 . "HANGUL SYLLABLE PIEUP U MIEUM")
    (14305 . "HANGUL SYLLABLE PIEUP U PIEUP")
    (14306 . "HANGUL SYLLABLE PIEUP U SIOS")
    (14307 . "HANGUL SYLLABLE PIEUP U IEUNG")
    (14308 . "HANGUL SYLLABLE PIEUP U THIEUTH")
    (14309 . "HANGUL SYLLABLE PIEUP U PHIEUPH")
    (14310 . "HANGUL SYLLABLE PIEUP WEO")
    (14311 . "HANGUL SYLLABLE PIEUP WEO RIEUL")
    (14312 . "HANGUL SYLLABLE PIEUP WEO SSANGSIOS")
    (14313 . "HANGUL SYLLABLE PIEUP WE")
    (14314 . "HANGUL SYLLABLE PIEUP WI")
    (14315 . "HANGUL SYLLABLE PIEUP WI KIYEOK")
    (14316 . "HANGUL SYLLABLE PIEUP WI NIEUN")
    (14317 . "HANGUL SYLLABLE PIEUP WI RIEUL")
    (14318 . "HANGUL SYLLABLE PIEUP WI IEUNG")
    (14319 . "HANGUL SYLLABLE PIEUP YU")
    (14320 . "HANGUL SYLLABLE PIEUP YU NIEUN")
    (14321 . "HANGUL SYLLABLE PIEUP YU RIEUL")
    (14322 . "HANGUL SYLLABLE PIEUP YU MIEUM")
    (14323 . "HANGUL SYLLABLE PIEUP YU SIOS")
    (14324 . "HANGUL SYLLABLE PIEUP YU IEUNG")
    (14325 . "HANGUL SYLLABLE PIEUP EU")
    (14326 . "HANGUL SYLLABLE PIEUP EU KIYEOK")
    (14327 . "HANGUL SYLLABLE PIEUP EU NIEUN")
    (14328 . "HANGUL SYLLABLE PIEUP EU RIEUL")
    (14329 . "HANGUL SYLLABLE PIEUP EU MIEUM")
    (14330 . "HANGUL SYLLABLE PIEUP EU PIEUP")
    (14331 . "HANGUL SYLLABLE PIEUP EU SIOS")
    (14332 . "HANGUL SYLLABLE PIEUP I")
    (14333 . "HANGUL SYLLABLE PIEUP I KIYEOK")
    (14334 . "HANGUL SYLLABLE PIEUP I NIEUN")
    (14335 . "HANGUL SYLLABLE PIEUP I RIEUL")
    (14336 . "HANGUL SYLLABLE PIEUP I RIEUL-MIEUM")
    (14337 . "HANGUL SYLLABLE PIEUP I MIEUM")
    (14338 . "HANGUL SYLLABLE PIEUP I PIEUP")
    (14339 . "HANGUL SYLLABLE PIEUP I SIOS")
    (14340 . "HANGUL SYLLABLE PIEUP I IEUNG")
    (14341 . "HANGUL SYLLABLE PIEUP I CIEUC")
    (14342 . "HANGUL SYLLABLE PIEUP I CHIEUCH")
    (14343 . "HANGUL SYLLABLE SSANGPIEUP A")
    (14344 . "HANGUL SYLLABLE SSANGPIEUP A KIYEOK")
    (14345 . "HANGUL SYLLABLE SSANGPIEUP A NIEUN")
    (14346 . "HANGUL SYLLABLE SSANGPIEUP A RIEUL")
    (14347 . "HANGUL SYLLABLE SSANGPIEUP A RIEUL-MIEUM")
    (14348 . "HANGUL SYLLABLE SSANGPIEUP A MIEUM")
    (14349 . "HANGUL SYLLABLE SSANGPIEUP A PIEUP")
    (14350 . "HANGUL SYLLABLE SSANGPIEUP A SIOS")
    (14351 . "HANGUL SYLLABLE SSANGPIEUP A SSANGSIOS")
    (14352 . "HANGUL SYLLABLE SSANGPIEUP A IEUNG")
    (14353 . "HANGUL SYLLABLE SSANGPIEUP A HIEUH")
    (14354 . "HANGUL SYLLABLE SSANGPIEUP AE")
    (14355 . "HANGUL SYLLABLE SSANGPIEUP AE KIYEOK")
    (14356 . "HANGUL SYLLABLE SSANGPIEUP AE NIEUN")
    (14357 . "HANGUL SYLLABLE SSANGPIEUP AE RIEUL")
    (14358 . "HANGUL SYLLABLE SSANGPIEUP AE MIEUM")
    (14359 . "HANGUL SYLLABLE SSANGPIEUP AE PIEUP")
    (14360 . "HANGUL SYLLABLE SSANGPIEUP AE SIOS")
    (14361 . "HANGUL SYLLABLE SSANGPIEUP AE SSANGSIOS")
    (14362 . "HANGUL SYLLABLE SSANGPIEUP AE IEUNG")
    (14363 . "HANGUL SYLLABLE SSANGPIEUP YA")
    (14364 . "HANGUL SYLLABLE SSANGPIEUP YA KIYEOK")
    (14365 . "HANGUL SYLLABLE SSANGPIEUP YA MIEUM")
    (14366 . "HANGUL SYLLABLE SSANGPIEUP EO")
    (14367 . "HANGUL SYLLABLE SSANGPIEUP EO KIYEOK")
    (14368 . "HANGUL SYLLABLE SSANGPIEUP EO NIEUN")
    (14369 . "HANGUL SYLLABLE SSANGPIEUP EO TIKEUT")
    (14370 . "HANGUL SYLLABLE SSANGPIEUP EO RIEUL")
    (14371 . "HANGUL SYLLABLE SSANGPIEUP EO MIEUM")
    (14372 . "HANGUL SYLLABLE SSANGPIEUP EO SIOS")
    (14373 . "HANGUL SYLLABLE SSANGPIEUP EO SSANGSIOS")
    (14374 . "HANGUL SYLLABLE SSANGPIEUP EO IEUNG")
    (14375 . "HANGUL SYLLABLE SSANGPIEUP E")
    (14376 . "HANGUL SYLLABLE SSANGPIEUP E IEUNG")
    (14377 . "HANGUL SYLLABLE SSANGPIEUP YEO")
    (14378 . "HANGUL SYLLABLE SSANGPIEUP YEO KIYEOK")
    (14379 . "HANGUL SYLLABLE SSANGPIEUP YEO MIEUM")
    (14380 . "HANGUL SYLLABLE SSANGPIEUP YEO PIEUP")
    (14381 . "HANGUL SYLLABLE SSANGPIEUP YEO SIOS")
    (14382 . "HANGUL SYLLABLE SSANGPIEUP YEO SSANGSIOS")
    (14383 . "HANGUL SYLLABLE SSANGPIEUP YEO IEUNG")
    (14384 . "HANGUL SYLLABLE SSANGPIEUP O")
    (14385 . "HANGUL SYLLABLE SSANGPIEUP O KIYEOK")
    (14386 . "HANGUL SYLLABLE SSANGPIEUP O NIEUN")
    (14387 . "HANGUL SYLLABLE SSANGPIEUP O RIEUL")
    (14388 . "HANGUL SYLLABLE SSANGPIEUP O MIEUM")
    (14389 . "HANGUL SYLLABLE SSANGPIEUP O PIEUP")
    (14390 . "HANGUL SYLLABLE SSANGPIEUP O IEUNG")
    (14391 . "HANGUL SYLLABLE SSANGPIEUP OE")
    (14392 . "HANGUL SYLLABLE SSANGPIEUP YO")
    (14393 . "HANGUL SYLLABLE SSANGPIEUP YO IEUNG")
    (14394 . "HANGUL SYLLABLE SSANGPIEUP U")
    (14395 . "HANGUL SYLLABLE SSANGPIEUP U KIYEOK")
    (14396 . "HANGUL SYLLABLE SSANGPIEUP U NIEUN")
    (14397 . "HANGUL SYLLABLE SSANGPIEUP U RIEUL")
    (14398 . "HANGUL SYLLABLE SSANGPIEUP U MIEUM")
    (14399 . "HANGUL SYLLABLE SSANGPIEUP U SIOS")
    (14400 . "HANGUL SYLLABLE SSANGPIEUP U IEUNG")
    (14401 . "HANGUL SYLLABLE SSANGPIEUP YU")
    (14402 . "HANGUL SYLLABLE SSANGPIEUP YU IEUNG")
    (14403 . "HANGUL SYLLABLE SSANGPIEUP EU")
    (14404 . "HANGUL SYLLABLE SSANGPIEUP EU NIEUN")
    (14405 . "HANGUL SYLLABLE SSANGPIEUP EU RIEUL")
    (14406 . "HANGUL SYLLABLE SSANGPIEUP EU MIEUM")
    (14407 . "HANGUL SYLLABLE SSANGPIEUP EU PIEUP")
    (14408 . "HANGUL SYLLABLE SSANGPIEUP I")
    (14409 . "HANGUL SYLLABLE SSANGPIEUP I KIYEOK")
    (14410 . "HANGUL SYLLABLE SSANGPIEUP I NIEUN")
    (14411 . "HANGUL SYLLABLE SSANGPIEUP I RIEUL")
    (14412 . "HANGUL SYLLABLE SSANGPIEUP I MIEUM")
    (14413 . "HANGUL SYLLABLE SSANGPIEUP I PIEUP")
    (14414 . "HANGUL SYLLABLE SSANGPIEUP I SSANGSIOS")
    (14415 . "HANGUL SYLLABLE SSANGPIEUP I IEUNG")
    (14416 . "HANGUL SYLLABLE SIOS A")
    (14417 . "HANGUL SYLLABLE SIOS A KIYEOK")
    (14418 . "HANGUL SYLLABLE SIOS A KIYEOK-SIOS")
    (14419 . "HANGUL SYLLABLE SIOS A NIEUN")
    (14420 . "HANGUL SYLLABLE SIOS A TIKEUT")
    (14421 . "HANGUL SYLLABLE SIOS A RIEUL")
    (14422 . "HANGUL SYLLABLE SIOS A RIEUL-KIYEOK")
    (14423 . "HANGUL SYLLABLE SIOS A RIEUL-MIEUM")
    (14424 . "HANGUL SYLLABLE SIOS A MIEUM")
    (14425 . "HANGUL SYLLABLE SIOS A PIEUP")
    (14426 . "HANGUL SYLLABLE SIOS A SIOS")
    (14427 . "HANGUL SYLLABLE SIOS A SSANGSIOS")
    (14428 . "HANGUL SYLLABLE SIOS A IEUNG")
    (14429 . "HANGUL SYLLABLE SIOS A THIEUTH")
    (14430 . "HANGUL SYLLABLE SIOS AE")
    (14431 . "HANGUL SYLLABLE SIOS AE KIYEOK")
    (14432 . "HANGUL SYLLABLE SIOS AE NIEUN")
    (14433 . "HANGUL SYLLABLE SIOS AE RIEUL")
    (14434 . "HANGUL SYLLABLE SIOS AE MIEUM")
    (14435 . "HANGUL SYLLABLE SIOS AE PIEUP")
    (14436 . "HANGUL SYLLABLE SIOS AE SIOS")
    (14437 . "HANGUL SYLLABLE SIOS AE SSANGSIOS")
    (14438 . "HANGUL SYLLABLE SIOS AE IEUNG")
    (14439 . "HANGUL SYLLABLE SIOS YA")
    (14440 . "HANGUL SYLLABLE SIOS YA KIYEOK")
    (14441 . "HANGUL SYLLABLE SIOS YA NIEUN")
    (14442 . "HANGUL SYLLABLE SIOS YA RIEUL")
    (14443 . "HANGUL SYLLABLE SIOS YA MIEUM")
    (14444 . "HANGUL SYLLABLE SIOS YA PIEUP")
    (14445 . "HANGUL SYLLABLE SIOS YA SIOS")
    (14446 . "HANGUL SYLLABLE SIOS YA IEUNG")
    (14447 . "HANGUL SYLLABLE SIOS YAE")
    (14448 . "HANGUL SYLLABLE SIOS YAE NIEUN")
    (14449 . "HANGUL SYLLABLE SIOS YAE RIEUL")
    (14450 . "HANGUL SYLLABLE SIOS YAE MIEUM")
    (14451 . "HANGUL SYLLABLE SIOS YAE IEUNG")
    (14452 . "HANGUL SYLLABLE SIOS EO")
    (14453 . "HANGUL SYLLABLE SIOS EO KIYEOK")
    (14454 . "HANGUL SYLLABLE SIOS EO SSANGKIYEOK")
    (14455 . "HANGUL SYLLABLE SIOS EO KIYEOK-SIOS")
    (14456 . "HANGUL SYLLABLE SIOS EO NIEUN")
    (14457 . "HANGUL SYLLABLE SIOS EO TIKEUT")
    (14458 . "HANGUL SYLLABLE SIOS EO RIEUL")
    (14459 . "HANGUL SYLLABLE SIOS EO RIEUL-MIEUM")
    (14460 . "HANGUL SYLLABLE SIOS EO RIEUL-PIEUP")
    (14461 . "HANGUL SYLLABLE SIOS EO MIEUM")
    (14462 . "HANGUL SYLLABLE SIOS EO PIEUP")
    (14463 . "HANGUL SYLLABLE SIOS EO SIOS")
    (14464 . "HANGUL SYLLABLE SIOS EO SSANGSIOS")
    (14465 . "HANGUL SYLLABLE SIOS EO IEUNG")
    (14466 . "HANGUL SYLLABLE SIOS EO PHIEUPH")
    (14467 . "HANGUL SYLLABLE SIOS E")
    (14468 . "HANGUL SYLLABLE SIOS E KIYEOK")
    (14469 . "HANGUL SYLLABLE SIOS E NIEUN")
    (14470 . "HANGUL SYLLABLE SIOS E RIEUL")
    (14471 . "HANGUL SYLLABLE SIOS E MIEUM")
    (14472 . "HANGUL SYLLABLE SIOS E PIEUP")
    (14473 . "HANGUL SYLLABLE SIOS E SIOS")
    (14474 . "HANGUL SYLLABLE SIOS E SSANGSIOS")
    (14475 . "HANGUL SYLLABLE SIOS E IEUNG")
    (14476 . "HANGUL SYLLABLE SIOS YEO")
    (14477 . "HANGUL SYLLABLE SIOS YEO KIYEOK")
    (14478 . "HANGUL SYLLABLE SIOS YEO NIEUN")
    (14479 . "HANGUL SYLLABLE SIOS YEO RIEUL")
    (14480 . "HANGUL SYLLABLE SIOS YEO MIEUM")
    (14481 . "HANGUL SYLLABLE SIOS YEO PIEUP")
    (14482 . "HANGUL SYLLABLE SIOS YEO SIOS")
    (14483 . "HANGUL SYLLABLE SIOS YEO SSANGSIOS")
    (14484 . "HANGUL SYLLABLE SIOS YEO IEUNG")
    (14485 . "HANGUL SYLLABLE SIOS YE")
    (14486 . "HANGUL SYLLABLE SIOS YE NIEUN")
    (14487 . "HANGUL SYLLABLE SIOS YE RIEUL")
    (14488 . "HANGUL SYLLABLE SIOS YE IEUNG")
    (14489 . "HANGUL SYLLABLE SIOS O")
    (14490 . "HANGUL SYLLABLE SIOS O KIYEOK")
    (14491 . "HANGUL SYLLABLE SIOS O SSANGKIYEOK")
    (14492 . "HANGUL SYLLABLE SIOS O NIEUN")
    (14493 . "HANGUL SYLLABLE SIOS O RIEUL")
    (14494 . "HANGUL SYLLABLE SIOS O RIEUL-MIEUM")
    (14495 . "HANGUL SYLLABLE SIOS O MIEUM")
    (14496 . "HANGUL SYLLABLE SIOS O PIEUP")
    (14497 . "HANGUL SYLLABLE SIOS O SIOS")
    (14498 . "HANGUL SYLLABLE SIOS O IEUNG")
    (14499 . "HANGUL SYLLABLE SIOS O THIEUTH")
    (14500 . "HANGUL SYLLABLE SIOS WA")
    (14501 . "HANGUL SYLLABLE SIOS WA KIYEOK")
    (14502 . "HANGUL SYLLABLE SIOS WA NIEUN")
    (14503 . "HANGUL SYLLABLE SIOS WA RIEUL")
    (14504 . "HANGUL SYLLABLE SIOS WA IEUNG")
    (14505 . "HANGUL SYLLABLE SIOS WAE")
    (14506 . "HANGUL SYLLABLE SIOS WAE NIEUN")
    (14507 . "HANGUL SYLLABLE SIOS WAE RIEUL")
    (14508 . "HANGUL SYLLABLE SIOS WAE MIEUM")
    (14509 . "HANGUL SYLLABLE SIOS WAE SIOS")
    (14510 . "HANGUL SYLLABLE SIOS WAE SSANGSIOS")
    (14511 . "HANGUL SYLLABLE SIOS OE")
    (14512 . "HANGUL SYLLABLE SIOS OE NIEUN")
    (14513 . "HANGUL SYLLABLE SIOS OE RIEUL")
    (14514 . "HANGUL SYLLABLE SIOS OE MIEUM")
    (14515 . "HANGUL SYLLABLE SIOS OE PIEUP")
    (14516 . "HANGUL SYLLABLE SIOS OE SIOS")
    (14517 . "HANGUL SYLLABLE SIOS YO")
    (14518 . "HANGUL SYLLABLE SIOS YO KIYEOK")
    (14519 . "HANGUL SYLLABLE SIOS YO NIEUN")
    (14520 . "HANGUL SYLLABLE SIOS YO RIEUL")
    (14521 . "HANGUL SYLLABLE SIOS YO MIEUM")
    (14522 . "HANGUL SYLLABLE SIOS YO PIEUP")
    (14523 . "HANGUL SYLLABLE SIOS YO SIOS")
    (14524 . "HANGUL SYLLABLE SIOS YO IEUNG")
    (14525 . "HANGUL SYLLABLE SIOS U")
    (14526 . "HANGUL SYLLABLE SIOS U KIYEOK")
    (14527 . "HANGUL SYLLABLE SIOS U NIEUN")
    (14528 . "HANGUL SYLLABLE SIOS U TIKEUT")
    (14529 . "HANGUL SYLLABLE SIOS U RIEUL")
    (14530 . "HANGUL SYLLABLE SIOS U MIEUM")
    (14531 . "HANGUL SYLLABLE SIOS U PIEUP")
    (14532 . "HANGUL SYLLABLE SIOS U SIOS")
    (14533 . "HANGUL SYLLABLE SIOS U IEUNG")
    (14534 . "HANGUL SYLLABLE SIOS U CHIEUCH")
    (14535 . "HANGUL SYLLABLE SIOS U THIEUTH")
    (14536 . "HANGUL SYLLABLE SIOS U PHIEUPH")
    (14537 . "HANGUL SYLLABLE SIOS WEO")
    (14538 . "HANGUL SYLLABLE SIOS WEO SSANGSIOS")
    (14539 . "HANGUL SYLLABLE SIOS WE")
    (14540 . "HANGUL SYLLABLE SIOS WE KIYEOK")
    (14541 . "HANGUL SYLLABLE SIOS WE NIEUN")
    (14542 . "HANGUL SYLLABLE SIOS WE RIEUL")
    (14543 . "HANGUL SYLLABLE SIOS WE MIEUM")
    (14544 . "HANGUL SYLLABLE SIOS WE IEUNG")
    (14545 . "HANGUL SYLLABLE SIOS WI")
    (14546 . "HANGUL SYLLABLE SIOS WI KIYEOK")
    (14547 . "HANGUL SYLLABLE SIOS WI NIEUN")
    (14548 . "HANGUL SYLLABLE SIOS WI RIEUL")
    (14549 . "HANGUL SYLLABLE SIOS WI MIEUM")
    (14550 . "HANGUL SYLLABLE SIOS WI PIEUP")
    (14551 . "HANGUL SYLLABLE SIOS WI SIOS")
    (14552 . "HANGUL SYLLABLE SIOS WI IEUNG")
    (14553 . "HANGUL SYLLABLE SIOS YU")
    (14554 . "HANGUL SYLLABLE SIOS YU KIYEOK")
    (14555 . "HANGUL SYLLABLE SIOS YU RIEUL")
    (14556 . "HANGUL SYLLABLE SIOS YU MIEUM")
    (14557 . "HANGUL SYLLABLE SIOS YU SIOS")
    (14558 . "HANGUL SYLLABLE SIOS YU IEUNG")
    (14559 . "HANGUL SYLLABLE SIOS EU")
    (14560 . "HANGUL SYLLABLE SIOS EU KIYEOK")
    (14561 . "HANGUL SYLLABLE SIOS EU NIEUN")
    (14562 . "HANGUL SYLLABLE SIOS EU RIEUL")
    (14563 . "HANGUL SYLLABLE SIOS EU RIEUL-KIYEOK")
    (14564 . "HANGUL SYLLABLE SIOS EU MIEUM")
    (14565 . "HANGUL SYLLABLE SIOS EU PIEUP")
    (14566 . "HANGUL SYLLABLE SIOS EU SIOS")
    (14567 . "HANGUL SYLLABLE SIOS EU IEUNG")
    (14568 . "HANGUL SYLLABLE SIOS I")
    (14569 . "HANGUL SYLLABLE SIOS I KIYEOK")
    (14570 . "HANGUL SYLLABLE SIOS I NIEUN")
    (14571 . "HANGUL SYLLABLE SIOS I TIKEUT")
    (14572 . "HANGUL SYLLABLE SIOS I RIEUL")
    (14573 . "HANGUL SYLLABLE SIOS I RIEUL-HIEUH")
    (14574 . "HANGUL SYLLABLE SIOS I MIEUM")
    (14575 . "HANGUL SYLLABLE SIOS I PIEUP")
    (14576 . "HANGUL SYLLABLE SIOS I SIOS")
    (14577 . "HANGUL SYLLABLE SIOS I IEUNG")
    (14578 . "HANGUL SYLLABLE SIOS I PHIEUPH")
    (14579 . "HANGUL SYLLABLE SSANGSIOS A")
    (14580 . "HANGUL SYLLABLE SSANGSIOS A KIYEOK")
    (14581 . "HANGUL SYLLABLE SSANGSIOS A KIYEOK-SIOS")
    (14582 . "HANGUL SYLLABLE SSANGSIOS A NIEUN")
    (14583 . "HANGUL SYLLABLE SSANGSIOS A RIEUL")
    (14584 . "HANGUL SYLLABLE SSANGSIOS A MIEUM")
    (14585 . "HANGUL SYLLABLE SSANGSIOS A PIEUP")
    (14586 . "HANGUL SYLLABLE SSANGSIOS A SSANGSIOS")
    (14587 . "HANGUL SYLLABLE SSANGSIOS A IEUNG")
    (14588 . "HANGUL SYLLABLE SSANGSIOS A HIEUH")
    (14589 . "HANGUL SYLLABLE SSANGSIOS AE")
    (14590 . "HANGUL SYLLABLE SSANGSIOS AE KIYEOK")
    (14591 . "HANGUL SYLLABLE SSANGSIOS AE NIEUN")
    (14592 . "HANGUL SYLLABLE SSANGSIOS AE RIEUL")
    (14593 . "HANGUL SYLLABLE SSANGSIOS AE MIEUM")
    (14594 . "HANGUL SYLLABLE SSANGSIOS AE PIEUP")
    (14595 . "HANGUL SYLLABLE SSANGSIOS AE SSANGSIOS")
    (14596 . "HANGUL SYLLABLE SSANGSIOS AE IEUNG")
    (14597 . "HANGUL SYLLABLE SSANGSIOS YA IEUNG")
    (14598 . "HANGUL SYLLABLE SSANGSIOS EO")
    (14599 . "HANGUL SYLLABLE SSANGSIOS EO KIYEOK")
    (14600 . "HANGUL SYLLABLE SSANGSIOS EO NIEUN")
    (14601 . "HANGUL SYLLABLE SSANGSIOS EO RIEUL")
    (14602 . "HANGUL SYLLABLE SSANGSIOS EO RIEUL-MIEUM")
    (14603 . "HANGUL SYLLABLE SSANGSIOS EO MIEUM")
    (14604 . "HANGUL SYLLABLE SSANGSIOS EO PIEUP")
    (14605 . "HANGUL SYLLABLE SSANGSIOS EO SSANGSIOS")
    (14606 . "HANGUL SYLLABLE SSANGSIOS EO IEUNG")
    (14607 . "HANGUL SYLLABLE SSANGSIOS E")
    (14608 . "HANGUL SYLLABLE SSANGSIOS E NIEUN")
    (14609 . "HANGUL SYLLABLE SSANGSIOS E RIEUL")
    (14610 . "HANGUL SYLLABLE SSANGSIOS YE NIEUN")
    (14611 . "HANGUL SYLLABLE SSANGSIOS O")
    (14612 . "HANGUL SYLLABLE SSANGSIOS O KIYEOK")
    (14613 . "HANGUL SYLLABLE SSANGSIOS O NIEUN")
    (14614 . "HANGUL SYLLABLE SSANGSIOS O TIKEUT")
    (14615 . "HANGUL SYLLABLE SSANGSIOS O RIEUL")
    (14616 . "HANGUL SYLLABLE SSANGSIOS O RIEUL-MIEUM")
    (14617 . "HANGUL SYLLABLE SSANGSIOS O MIEUM")
    (14618 . "HANGUL SYLLABLE SSANGSIOS O PIEUP")
    (14619 . "HANGUL SYLLABLE SSANGSIOS O IEUNG")
    (14620 . "HANGUL SYLLABLE SSANGSIOS WA")
    (14621 . "HANGUL SYLLABLE SSANGSIOS WA KIYEOK")
    (14622 . "HANGUL SYLLABLE SSANGSIOS WA NIEUN")
    (14623 . "HANGUL SYLLABLE SSANGSIOS WA SSANGSIOS")
    (14624 . "HANGUL SYLLABLE SSANGSIOS WAE")
    (14625 . "HANGUL SYLLABLE SSANGSIOS WAE SSANGSIOS")
    (14626 . "HANGUL SYLLABLE SSANGSIOS OE")
    (14627 . "HANGUL SYLLABLE SSANGSIOS OE NIEUN")
    (14628 . "HANGUL SYLLABLE SSANGSIOS OE RIEUL")
    (14629 . "HANGUL SYLLABLE SSANGSIOS OE MIEUM")
    (14630 . "HANGUL SYLLABLE SSANGSIOS OE PIEUP")
    (14631 . "HANGUL SYLLABLE SSANGSIOS YO")
    (14632 . "HANGUL SYLLABLE SSANGSIOS U")
    (14633 . "HANGUL SYLLABLE SSANGSIOS U KIYEOK")
    (14634 . "HANGUL SYLLABLE SSANGSIOS U NIEUN")
    (14635 . "HANGUL SYLLABLE SSANGSIOS U RIEUL")
    (14636 . "HANGUL SYLLABLE SSANGSIOS U MIEUM")
    (14637 . "HANGUL SYLLABLE SSANGSIOS U PIEUP")
    (14638 . "HANGUL SYLLABLE SSANGSIOS U IEUNG")
    (14639 . "HANGUL SYLLABLE SSANGSIOS WEO")
    (14640 . "HANGUL SYLLABLE SSANGSIOS WEO SSANGSIOS")
    (14641 . "HANGUL SYLLABLE SSANGSIOS WE")
    (14642 . "HANGUL SYLLABLE SSANGSIOS WI")
    (14643 . "HANGUL SYLLABLE SSANGSIOS WI NIEUN")
    (14644 . "HANGUL SYLLABLE SSANGSIOS YU IEUNG")
    (14645 . "HANGUL SYLLABLE SSANGSIOS EU")
    (14646 . "HANGUL SYLLABLE SSANGSIOS EU KIYEOK")
    (14647 . "HANGUL SYLLABLE SSANGSIOS EU NIEUN")
    (14648 . "HANGUL SYLLABLE SSANGSIOS EU RIEUL")
    (14649 . "HANGUL SYLLABLE SSANGSIOS EU RIEUL-MIEUM")
    (14650 . "HANGUL SYLLABLE SSANGSIOS EU RIEUL-HIEUH")
    (14651 . "HANGUL SYLLABLE SSANGSIOS EU MIEUM")
    (14652 . "HANGUL SYLLABLE SSANGSIOS EU PIEUP")
    (14653 . "HANGUL SYLLABLE SSANGSIOS YI")
    (14654 . "HANGUL SYLLABLE SSANGSIOS YI NIEUN")
    (14655 . "HANGUL SYLLABLE SSANGSIOS YI RIEUL")
    (14656 . "HANGUL SYLLABLE SSANGSIOS YI MIEUM")
    (14657 . "HANGUL SYLLABLE SSANGSIOS I")
    (14658 . "HANGUL SYLLABLE SSANGSIOS I KIYEOK")
    (14659 . "HANGUL SYLLABLE SSANGSIOS I NIEUN")
    (14660 . "HANGUL SYLLABLE SSANGSIOS I RIEUL")
    (14661 . "HANGUL SYLLABLE SSANGSIOS I MIEUM")
    (14662 . "HANGUL SYLLABLE SSANGSIOS I PIEUP")
    (14663 . "HANGUL SYLLABLE SSANGSIOS I SIOS")
    (14664 . "HANGUL SYLLABLE SSANGSIOS I IEUNG")
    (14665 . "HANGUL SYLLABLE IEUNG A")
    (14666 . "HANGUL SYLLABLE IEUNG A KIYEOK")
    (14667 . "HANGUL SYLLABLE IEUNG A NIEUN")
    (14668 . "HANGUL SYLLABLE IEUNG A NIEUN-CIEUC")
    (14669 . "HANGUL SYLLABLE IEUNG A NIEUN-HIEUH")
    (14670 . "HANGUL SYLLABLE IEUNG A RIEUL")
    (14671 . "HANGUL SYLLABLE IEUNG A RIEUL-KIYEOK")
    (14672 . "HANGUL SYLLABLE IEUNG A RIEUL-MIEUM")
    (14673 . "HANGUL SYLLABLE IEUNG A RIEUL-HIEUH")
    (14674 . "HANGUL SYLLABLE IEUNG A MIEUM")
    (14675 . "HANGUL SYLLABLE IEUNG A PIEUP")
    (14676 . "HANGUL SYLLABLE IEUNG A SIOS")
    (14677 . "HANGUL SYLLABLE IEUNG A SSANGSIOS")
    (14678 . "HANGUL SYLLABLE IEUNG A IEUNG")
    (14679 . "HANGUL SYLLABLE IEUNG A THIEUTH")
    (14680 . "HANGUL SYLLABLE IEUNG A PHIEUPH")
    (14681 . "HANGUL SYLLABLE IEUNG AE")
    (14682 . "HANGUL SYLLABLE IEUNG AE KIYEOK")
    (14683 . "HANGUL SYLLABLE IEUNG AE NIEUN")
    (14684 . "HANGUL SYLLABLE IEUNG AE RIEUL")
    (14685 . "HANGUL SYLLABLE IEUNG AE MIEUM")
    (14686 . "HANGUL SYLLABLE IEUNG AE PIEUP")
    (14687 . "HANGUL SYLLABLE IEUNG AE SIOS")
    (14688 . "HANGUL SYLLABLE IEUNG AE SSANGSIOS")
    (14689 . "HANGUL SYLLABLE IEUNG AE IEUNG")
    (14690 . "HANGUL SYLLABLE IEUNG YA")
    (14691 . "HANGUL SYLLABLE IEUNG YA KIYEOK")
    (14692 . "HANGUL SYLLABLE IEUNG YA NIEUN")
    (14693 . "HANGUL SYLLABLE IEUNG YA RIEUL")
    (14694 . "HANGUL SYLLABLE IEUNG YA RIEUL-PIEUP")
    (14695 . "HANGUL SYLLABLE IEUNG YA MIEUM")
    (14696 . "HANGUL SYLLABLE IEUNG YA PIEUP")
    (14697 . "HANGUL SYLLABLE IEUNG YA SIOS")
    (14698 . "HANGUL SYLLABLE IEUNG YA IEUNG")
    (14699 . "HANGUL SYLLABLE IEUNG YA THIEUTH")
    (14700 . "HANGUL SYLLABLE IEUNG YA HIEUH")
    (14701 . "HANGUL SYLLABLE IEUNG YAE")
    (14702 . "HANGUL SYLLABLE IEUNG YAE NIEUN")
    (14703 . "HANGUL SYLLABLE IEUNG YAE RIEUL")
    (14704 . "HANGUL SYLLABLE IEUNG YAE PIEUP")
    (14705 . "HANGUL SYLLABLE IEUNG EO")
    (14706 . "HANGUL SYLLABLE IEUNG EO KIYEOK")
    (14707 . "HANGUL SYLLABLE IEUNG EO NIEUN")
    (14708 . "HANGUL SYLLABLE IEUNG EO NIEUN-CIEUC")
    (14709 . "HANGUL SYLLABLE IEUNG EO TIKEUT")
    (14710 . "HANGUL SYLLABLE IEUNG EO RIEUL")
    (14711 . "HANGUL SYLLABLE IEUNG EO RIEUL-KIYEOK")
    (14712 . "HANGUL SYLLABLE IEUNG EO RIEUL-MIEUM")
    (14713 . "HANGUL SYLLABLE IEUNG EO MIEUM")
    (14714 . "HANGUL SYLLABLE IEUNG EO PIEUP")
    (14715 . "HANGUL SYLLABLE IEUNG EO PIEUP-SIOS")
    (14716 . "HANGUL SYLLABLE IEUNG EO SIOS")
    (14717 . "HANGUL SYLLABLE IEUNG EO SSANGSIOS")
    (14718 . "HANGUL SYLLABLE IEUNG EO IEUNG")
    (14719 . "HANGUL SYLLABLE IEUNG EO CIEUC")
    (14720 . "HANGUL SYLLABLE IEUNG EO KHIEUKH")
    (14721 . "HANGUL SYLLABLE IEUNG EO PHIEUPH")
    (14722 . "HANGUL SYLLABLE IEUNG E")
    (14723 . "HANGUL SYLLABLE IEUNG E KIYEOK")
    (14724 . "HANGUL SYLLABLE IEUNG E NIEUN")
    (14725 . "HANGUL SYLLABLE IEUNG E RIEUL")
    (14726 . "HANGUL SYLLABLE IEUNG E MIEUM")
    (14727 . "HANGUL SYLLABLE IEUNG E PIEUP")
    (14728 . "HANGUL SYLLABLE IEUNG E SIOS")
    (14729 . "HANGUL SYLLABLE IEUNG E IEUNG")
    (14730 . "HANGUL SYLLABLE IEUNG YEO")
    (14731 . "HANGUL SYLLABLE IEUNG YEO KIYEOK")
    (14732 . "HANGUL SYLLABLE IEUNG YEO SSANGKIYEOK")
    (14733 . "HANGUL SYLLABLE IEUNG YEO NIEUN")
    (14734 . "HANGUL SYLLABLE IEUNG YEO RIEUL")
    (14735 . "HANGUL SYLLABLE IEUNG YEO RIEUL-MIEUM")
    (14736 . "HANGUL SYLLABLE IEUNG YEO RIEUL-PIEUP")
    (14737 . "HANGUL SYLLABLE IEUNG YEO MIEUM")
    (14738 . "HANGUL SYLLABLE IEUNG YEO PIEUP")
    (14739 . "HANGUL SYLLABLE IEUNG YEO PIEUP-SIOS")
    (14740 . "HANGUL SYLLABLE IEUNG YEO SIOS")
    (14741 . "HANGUL SYLLABLE IEUNG YEO SSANGSIOS")
    (14742 . "HANGUL SYLLABLE IEUNG YEO IEUNG")
    (14743 . "HANGUL SYLLABLE IEUNG YEO THIEUTH")
    (14744 . "HANGUL SYLLABLE IEUNG YEO PHIEUPH")
    (14745 . "HANGUL SYLLABLE IEUNG YEO HIEUH")
    (14746 . "HANGUL SYLLABLE IEUNG YE")
    (14747 . "HANGUL SYLLABLE IEUNG YE NIEUN")
    (14748 . "HANGUL SYLLABLE IEUNG YE RIEUL")
    (14749 . "HANGUL SYLLABLE IEUNG YE MIEUM")
    (14750 . "HANGUL SYLLABLE IEUNG YE PIEUP")
    (14751 . "HANGUL SYLLABLE IEUNG YE SIOS")
    (14752 . "HANGUL SYLLABLE IEUNG YE SSANGSIOS")
    (14753 . "HANGUL SYLLABLE IEUNG O")
    (14754 . "HANGUL SYLLABLE IEUNG O KIYEOK")
    (14755 . "HANGUL SYLLABLE IEUNG O NIEUN")
    (14756 . "HANGUL SYLLABLE IEUNG O RIEUL")
    (14757 . "HANGUL SYLLABLE IEUNG O RIEUL-KIYEOK")
    (14758 . "HANGUL SYLLABLE IEUNG O RIEUL-MIEUM")
    (14759 . "HANGUL SYLLABLE IEUNG O RIEUL-SIOS")
    (14760 . "HANGUL SYLLABLE IEUNG O RIEUL-HIEUH")
    (14761 . "HANGUL SYLLABLE IEUNG O MIEUM")
    (14762 . "HANGUL SYLLABLE IEUNG O PIEUP")
    (14763 . "HANGUL SYLLABLE IEUNG O SIOS")
    (14764 . "HANGUL SYLLABLE IEUNG O IEUNG")
    (14765 . "HANGUL SYLLABLE IEUNG O CHIEUCH")
    (14766 . "HANGUL SYLLABLE IEUNG WA")
    (14767 . "HANGUL SYLLABLE IEUNG WA KIYEOK")
    (14768 . "HANGUL SYLLABLE IEUNG WA NIEUN")
    (14769 . "HANGUL SYLLABLE IEUNG WA RIEUL")
    (14770 . "HANGUL SYLLABLE IEUNG WA MIEUM")
    (14771 . "HANGUL SYLLABLE IEUNG WA PIEUP")
    (14772 . "HANGUL SYLLABLE IEUNG WA SIOS")
    (14773 . "HANGUL SYLLABLE IEUNG WA SSANGSIOS")
    (14774 . "HANGUL SYLLABLE IEUNG WA IEUNG")
    (14775 . "HANGUL SYLLABLE IEUNG WAE")
    (14776 . "HANGUL SYLLABLE IEUNG WAE KIYEOK")
    (14777 . "HANGUL SYLLABLE IEUNG WAE NIEUN")
    (14778 . "HANGUL SYLLABLE IEUNG WAE MIEUM")
    (14779 . "HANGUL SYLLABLE IEUNG WAE SIOS")
    (14780 . "HANGUL SYLLABLE IEUNG WAE IEUNG")
    (14781 . "HANGUL SYLLABLE IEUNG OE")
    (14782 . "HANGUL SYLLABLE IEUNG OE KIYEOK")
    (14783 . "HANGUL SYLLABLE IEUNG OE NIEUN")
    (14784 . "HANGUL SYLLABLE IEUNG OE RIEUL")
    (14785 . "HANGUL SYLLABLE IEUNG OE MIEUM")
    (14786 . "HANGUL SYLLABLE IEUNG OE PIEUP")
    (14787 . "HANGUL SYLLABLE IEUNG OE SIOS")
    (14788 . "HANGUL SYLLABLE IEUNG OE IEUNG")
    (14789 . "HANGUL SYLLABLE IEUNG YO")
    (14790 . "HANGUL SYLLABLE IEUNG YO KIYEOK")
    (14791 . "HANGUL SYLLABLE IEUNG YO NIEUN")
    (14792 . "HANGUL SYLLABLE IEUNG YO RIEUL")
    (14793 . "HANGUL SYLLABLE IEUNG YO MIEUM")
    (14794 . "HANGUL SYLLABLE IEUNG YO PIEUP")
    (14795 . "HANGUL SYLLABLE IEUNG YO SIOS")
    (14796 . "HANGUL SYLLABLE IEUNG YO IEUNG")
    (14797 . "HANGUL SYLLABLE IEUNG U")
    (14798 . "HANGUL SYLLABLE IEUNG U KIYEOK")
    (14799 . "HANGUL SYLLABLE IEUNG U NIEUN")
    (14800 . "HANGUL SYLLABLE IEUNG U RIEUL")
    (14801 . "HANGUL SYLLABLE IEUNG U RIEUL-KIYEOK")
    (14802 . "HANGUL SYLLABLE IEUNG U RIEUL-MIEUM")
    (14803 . "HANGUL SYLLABLE IEUNG U MIEUM")
    (14804 . "HANGUL SYLLABLE IEUNG U PIEUP")
    (14805 . "HANGUL SYLLABLE IEUNG U SIOS")
    (14806 . "HANGUL SYLLABLE IEUNG U IEUNG")
    (14807 . "HANGUL SYLLABLE IEUNG WEO")
    (14808 . "HANGUL SYLLABLE IEUNG WEO KIYEOK")
    (14809 . "HANGUL SYLLABLE IEUNG WEO NIEUN")
    (14810 . "HANGUL SYLLABLE IEUNG WEO RIEUL")
    (14811 . "HANGUL SYLLABLE IEUNG WEO MIEUM")
    (14812 . "HANGUL SYLLABLE IEUNG WEO PIEUP")
    (14813 . "HANGUL SYLLABLE IEUNG WEO SSANGSIOS")
    (14814 . "HANGUL SYLLABLE IEUNG WEO IEUNG")
    (14815 . "HANGUL SYLLABLE IEUNG WE")
    (14816 . "HANGUL SYLLABLE IEUNG WE KIYEOK")
    (14817 . "HANGUL SYLLABLE IEUNG WE NIEUN")
    (14818 . "HANGUL SYLLABLE IEUNG WE RIEUL")
    (14819 . "HANGUL SYLLABLE IEUNG WE MIEUM")
    (14820 . "HANGUL SYLLABLE IEUNG WE PIEUP")
    (14821 . "HANGUL SYLLABLE IEUNG WE IEUNG")
    (14822 . "HANGUL SYLLABLE IEUNG WI")
    (14823 . "HANGUL SYLLABLE IEUNG WI KIYEOK")
    (14824 . "HANGUL SYLLABLE IEUNG WI NIEUN")
    (14825 . "HANGUL SYLLABLE IEUNG WI RIEUL")
    (14826 . "HANGUL SYLLABLE IEUNG WI MIEUM")
    (14827 . "HANGUL SYLLABLE IEUNG WI PIEUP")
    (14828 . "HANGUL SYLLABLE IEUNG WI SIOS")
    (14829 . "HANGUL SYLLABLE IEUNG WI IEUNG")
    (14830 . "HANGUL SYLLABLE IEUNG YU")
    (14831 . "HANGUL SYLLABLE IEUNG YU KIYEOK")
    (14832 . "HANGUL SYLLABLE IEUNG YU NIEUN")
    (14833 . "HANGUL SYLLABLE IEUNG YU RIEUL")
    (14834 . "HANGUL SYLLABLE IEUNG YU MIEUM")
    (14835 . "HANGUL SYLLABLE IEUNG YU PIEUP")
    (14836 . "HANGUL SYLLABLE IEUNG YU SIOS")
    (14837 . "HANGUL SYLLABLE IEUNG YU IEUNG")
    (14838 . "HANGUL SYLLABLE IEUNG YU CHIEUCH")
    (14839 . "HANGUL SYLLABLE IEUNG EU")
    (14840 . "HANGUL SYLLABLE IEUNG EU KIYEOK")
    (14841 . "HANGUL SYLLABLE IEUNG EU NIEUN")
    (14842 . "HANGUL SYLLABLE IEUNG EU RIEUL")
    (14843 . "HANGUL SYLLABLE IEUNG EU RIEUL-PHIEUPH")
    (14844 . "HANGUL SYLLABLE IEUNG EU MIEUM")
    (14845 . "HANGUL SYLLABLE IEUNG EU PIEUP")
    (14846 . "HANGUL SYLLABLE IEUNG EU SIOS")
    (14847 . "HANGUL SYLLABLE IEUNG EU IEUNG")
    (14848 . "HANGUL SYLLABLE IEUNG EU CIEUC")
    (14849 . "HANGUL SYLLABLE IEUNG EU CHIEUCH")
    (14850 . "HANGUL SYLLABLE IEUNG EU KHIEUKH")
    (14851 . "HANGUL SYLLABLE IEUNG EU THIEUTH")
    (14852 . "HANGUL SYLLABLE IEUNG EU PHIEUPH")
    (14853 . "HANGUL SYLLABLE IEUNG EU HIEUH")
    (14854 . "HANGUL SYLLABLE IEUNG YI")
    (14855 . "HANGUL SYLLABLE IEUNG YI NIEUN")
    (14856 . "HANGUL SYLLABLE IEUNG YI RIEUL")
    (14857 . "HANGUL SYLLABLE IEUNG YI MIEUM")
    (14858 . "HANGUL SYLLABLE IEUNG YI SIOS")
    (14859 . "HANGUL SYLLABLE IEUNG I")
    (14860 . "HANGUL SYLLABLE IEUNG I KIYEOK")
    (14861 . "HANGUL SYLLABLE IEUNG I NIEUN")
    (14862 . "HANGUL SYLLABLE IEUNG I RIEUL")
    (14863 . "HANGUL SYLLABLE IEUNG I RIEUL-KIYEOK")
    (14864 . "HANGUL SYLLABLE IEUNG I RIEUL-MIEUM")
    (14865 . "HANGUL SYLLABLE IEUNG I RIEUL-HIEUH")
    (14866 . "HANGUL SYLLABLE IEUNG I MIEUM")
    (14867 . "HANGUL SYLLABLE IEUNG I PIEUP")
    (14868 . "HANGUL SYLLABLE IEUNG I SIOS")
    (14869 . "HANGUL SYLLABLE IEUNG I SSANGSIOS")
    (14870 . "HANGUL SYLLABLE IEUNG I IEUNG")
    (14871 . "HANGUL SYLLABLE IEUNG I CIEUC")
    (14872 . "HANGUL SYLLABLE IEUNG I PHIEUPH")
    (14873 . "HANGUL SYLLABLE CIEUC A")
    (14874 . "HANGUL SYLLABLE CIEUC A KIYEOK")
    (14875 . "HANGUL SYLLABLE CIEUC A NIEUN")
    (14876 . "HANGUL SYLLABLE CIEUC A NIEUN-HIEUH")
    (14877 . "HANGUL SYLLABLE CIEUC A TIKEUT")
    (14878 . "HANGUL SYLLABLE CIEUC A RIEUL")
    (14879 . "HANGUL SYLLABLE CIEUC A RIEUL-MIEUM")
    (14880 . "HANGUL SYLLABLE CIEUC A MIEUM")
    (14881 . "HANGUL SYLLABLE CIEUC A PIEUP")
    (14882 . "HANGUL SYLLABLE CIEUC A SIOS")
    (14883 . "HANGUL SYLLABLE CIEUC A SSANGSIOS")
    (14884 . "HANGUL SYLLABLE CIEUC A IEUNG")
    (14885 . "HANGUL SYLLABLE CIEUC A CIEUC")
    (14886 . "HANGUL SYLLABLE CIEUC AE")
    (14887 . "HANGUL SYLLABLE CIEUC AE KIYEOK")
    (14888 . "HANGUL SYLLABLE CIEUC AE NIEUN")
    (14889 . "HANGUL SYLLABLE CIEUC AE RIEUL")
    (14890 . "HANGUL SYLLABLE CIEUC AE MIEUM")
    (14891 . "HANGUL SYLLABLE CIEUC AE PIEUP")
    (14892 . "HANGUL SYLLABLE CIEUC AE SIOS")
    (14893 . "HANGUL SYLLABLE CIEUC AE SSANGSIOS")
    (14894 . "HANGUL SYLLABLE CIEUC AE IEUNG")
    (14895 . "HANGUL SYLLABLE CIEUC YA")
    (14896 . "HANGUL SYLLABLE CIEUC YA KIYEOK")
    (14897 . "HANGUL SYLLABLE CIEUC YA NIEUN")
    (14898 . "HANGUL SYLLABLE CIEUC YA NIEUN-HIEUH")
    (14899 . "HANGUL SYLLABLE CIEUC YA RIEUL")
    (14900 . "HANGUL SYLLABLE CIEUC YA MIEUM")
    (14901 . "HANGUL SYLLABLE CIEUC YA IEUNG")
    (14902 . "HANGUL SYLLABLE CIEUC YAE")
    (14903 . "HANGUL SYLLABLE CIEUC YAE NIEUN")
    (14904 . "HANGUL SYLLABLE CIEUC YAE RIEUL")
    (14905 . "HANGUL SYLLABLE CIEUC EO")
    (14906 . "HANGUL SYLLABLE CIEUC EO KIYEOK")
    (14907 . "HANGUL SYLLABLE CIEUC EO NIEUN")
    (14908 . "HANGUL SYLLABLE CIEUC EO RIEUL")
    (14909 . "HANGUL SYLLABLE CIEUC EO RIEUL-MIEUM")
    (14910 . "HANGUL SYLLABLE CIEUC EO MIEUM")
    (14911 . "HANGUL SYLLABLE CIEUC EO PIEUP")
    (14912 . "HANGUL SYLLABLE CIEUC EO SIOS")
    (14913 . "HANGUL SYLLABLE CIEUC EO IEUNG")
    (14914 . "HANGUL SYLLABLE CIEUC EO CIEUC")
    (14915 . "HANGUL SYLLABLE CIEUC E")
    (14916 . "HANGUL SYLLABLE CIEUC E KIYEOK")
    (14917 . "HANGUL SYLLABLE CIEUC E NIEUN")
    (14918 . "HANGUL SYLLABLE CIEUC E RIEUL")
    (14919 . "HANGUL SYLLABLE CIEUC E MIEUM")
    (14920 . "HANGUL SYLLABLE CIEUC E PIEUP")
    (14921 . "HANGUL SYLLABLE CIEUC E SIOS")
    (14922 . "HANGUL SYLLABLE CIEUC E IEUNG")
    (14923 . "HANGUL SYLLABLE CIEUC YEO")
    (14924 . "HANGUL SYLLABLE CIEUC YEO NIEUN")
    (14925 . "HANGUL SYLLABLE CIEUC YEO RIEUL")
    (14926 . "HANGUL SYLLABLE CIEUC YEO MIEUM")
    (14927 . "HANGUL SYLLABLE CIEUC YEO PIEUP")
    (14928 . "HANGUL SYLLABLE CIEUC YEO SSANGSIOS")
    (14929 . "HANGUL SYLLABLE CIEUC YEO IEUNG")
    (14930 . "HANGUL SYLLABLE CIEUC YE")
    (14931 . "HANGUL SYLLABLE CIEUC O")
    (14932 . "HANGUL SYLLABLE CIEUC O KIYEOK")
    (14933 . "HANGUL SYLLABLE CIEUC O NIEUN")
    (14934 . "HANGUL SYLLABLE CIEUC O RIEUL")
    (14935 . "HANGUL SYLLABLE CIEUC O RIEUL-MIEUM")
    (14936 . "HANGUL SYLLABLE CIEUC O MIEUM")
    (14937 . "HANGUL SYLLABLE CIEUC O PIEUP")
    (14938 . "HANGUL SYLLABLE CIEUC O SIOS")
    (14939 . "HANGUL SYLLABLE CIEUC O IEUNG")
    (14940 . "HANGUL SYLLABLE CIEUC O CIEUC")
    (14941 . "HANGUL SYLLABLE CIEUC O CHIEUCH")
    (14942 . "HANGUL SYLLABLE CIEUC O HIEUH")
    (14943 . "HANGUL SYLLABLE CIEUC WA")
    (14944 . "HANGUL SYLLABLE CIEUC WA KIYEOK")
    (14945 . "HANGUL SYLLABLE CIEUC WA RIEUL")
    (14946 . "HANGUL SYLLABLE CIEUC WA PIEUP")
    (14947 . "HANGUL SYLLABLE CIEUC WA SIOS")
    (14948 . "HANGUL SYLLABLE CIEUC WA IEUNG")
    (14949 . "HANGUL SYLLABLE CIEUC WAE")
    (14950 . "HANGUL SYLLABLE CIEUC WAE SSANGSIOS")
    (14951 . "HANGUL SYLLABLE CIEUC WAE IEUNG")
    (14952 . "HANGUL SYLLABLE CIEUC OE")
    (14953 . "HANGUL SYLLABLE CIEUC OE NIEUN")
    (14954 . "HANGUL SYLLABLE CIEUC OE RIEUL")
    (14955 . "HANGUL SYLLABLE CIEUC OE MIEUM")
    (14956 . "HANGUL SYLLABLE CIEUC OE PIEUP")
    (14957 . "HANGUL SYLLABLE CIEUC OE SIOS")
    (14958 . "HANGUL SYLLABLE CIEUC OE IEUNG")
    (14959 . "HANGUL SYLLABLE CIEUC YO")
    (14960 . "HANGUL SYLLABLE CIEUC YO KIYEOK")
    (14961 . "HANGUL SYLLABLE CIEUC YO NIEUN")
    (14962 . "HANGUL SYLLABLE CIEUC YO IEUNG")
    (14963 . "HANGUL SYLLABLE CIEUC U")
    (14964 . "HANGUL SYLLABLE CIEUC U KIYEOK")
    (14965 . "HANGUL SYLLABLE CIEUC U NIEUN")
    (14966 . "HANGUL SYLLABLE CIEUC U RIEUL")
    (14967 . "HANGUL SYLLABLE CIEUC U RIEUL-KIYEOK")
    (14968 . "HANGUL SYLLABLE CIEUC U RIEUL-MIEUM")
    (14969 . "HANGUL SYLLABLE CIEUC U MIEUM")
    (14970 . "HANGUL SYLLABLE CIEUC U PIEUP")
    (14971 . "HANGUL SYLLABLE CIEUC U SIOS")
    (14972 . "HANGUL SYLLABLE CIEUC U IEUNG")
    (14973 . "HANGUL SYLLABLE CIEUC WEO")
    (14974 . "HANGUL SYLLABLE CIEUC WEO SSANGSIOS")
    (14975 . "HANGUL SYLLABLE CIEUC WE")
    (14976 . "HANGUL SYLLABLE CIEUC WI")
    (14977 . "HANGUL SYLLABLE CIEUC WI KIYEOK")
    (14978 . "HANGUL SYLLABLE CIEUC WI NIEUN")
    (14979 . "HANGUL SYLLABLE CIEUC WI RIEUL")
    (14980 . "HANGUL SYLLABLE CIEUC WI MIEUM")
    (14981 . "HANGUL SYLLABLE CIEUC WI PIEUP")
    (14982 . "HANGUL SYLLABLE CIEUC WI SIOS")
    (14983 . "HANGUL SYLLABLE CIEUC YU")
    (14984 . "HANGUL SYLLABLE CIEUC YU NIEUN")
    (14985 . "HANGUL SYLLABLE CIEUC YU RIEUL")
    (14986 . "HANGUL SYLLABLE CIEUC YU MIEUM")
    (14987 . "HANGUL SYLLABLE CIEUC EU")
    (14988 . "HANGUL SYLLABLE CIEUC EU KIYEOK")
    (14989 . "HANGUL SYLLABLE CIEUC EU NIEUN")
    (14990 . "HANGUL SYLLABLE CIEUC EU RIEUL")
    (14991 . "HANGUL SYLLABLE CIEUC EU MIEUM")
    (14992 . "HANGUL SYLLABLE CIEUC EU PIEUP")
    (14993 . "HANGUL SYLLABLE CIEUC EU SIOS")
    (14994 . "HANGUL SYLLABLE CIEUC EU IEUNG")
    (14995 . "HANGUL SYLLABLE CIEUC I")
    (14996 . "HANGUL SYLLABLE CIEUC I KIYEOK")
    (14997 . "HANGUL SYLLABLE CIEUC I NIEUN")
    (14998 . "HANGUL SYLLABLE CIEUC I TIKEUT")
    (14999 . "HANGUL SYLLABLE CIEUC I RIEUL")
    (15000 . "HANGUL SYLLABLE CIEUC I RIEUL-MIEUM")
    (15001 . "HANGUL SYLLABLE CIEUC I MIEUM")
    (15002 . "HANGUL SYLLABLE CIEUC I PIEUP")
    (15003 . "HANGUL SYLLABLE CIEUC I SIOS")
    (15004 . "HANGUL SYLLABLE CIEUC I IEUNG")
    (15005 . "HANGUL SYLLABLE CIEUC I CIEUC")
    (15006 . "HANGUL SYLLABLE CIEUC I THIEUTH")
    (15007 . "HANGUL SYLLABLE CIEUC I PHIEUPH")
    (15008 . "HANGUL SYLLABLE SSANGCIEUC A")
    (15009 . "HANGUL SYLLABLE SSANGCIEUC A KIYEOK")
    (15010 . "HANGUL SYLLABLE SSANGCIEUC A NIEUN")
    (15011 . "HANGUL SYLLABLE SSANGCIEUC A NIEUN-HIEUH")
    (15012 . "HANGUL SYLLABLE SSANGCIEUC A RIEUL")
    (15013 . "HANGUL SYLLABLE SSANGCIEUC A RIEUL-PIEUP")
    (15014 . "HANGUL SYLLABLE SSANGCIEUC A MIEUM")
    (15015 . "HANGUL SYLLABLE SSANGCIEUC A PIEUP")
    (15016 . "HANGUL SYLLABLE SSANGCIEUC A SIOS")
    (15017 . "HANGUL SYLLABLE SSANGCIEUC A SSANGSIOS")
    (15018 . "HANGUL SYLLABLE SSANGCIEUC A IEUNG")
    (15019 . "HANGUL SYLLABLE SSANGCIEUC AE")
    (15020 . "HANGUL SYLLABLE SSANGCIEUC AE KIYEOK")
    (15021 . "HANGUL SYLLABLE SSANGCIEUC AE NIEUN")
    (15022 . "HANGUL SYLLABLE SSANGCIEUC AE RIEUL")
    (15023 . "HANGUL SYLLABLE SSANGCIEUC AE MIEUM")
    (15024 . "HANGUL SYLLABLE SSANGCIEUC AE PIEUP")
    (15025 . "HANGUL SYLLABLE SSANGCIEUC AE SIOS")
    (15026 . "HANGUL SYLLABLE SSANGCIEUC AE SSANGSIOS")
    (15027 . "HANGUL SYLLABLE SSANGCIEUC AE IEUNG")
    (15028 . "HANGUL SYLLABLE SSANGCIEUC YA")
    (15029 . "HANGUL SYLLABLE SSANGCIEUC YA NIEUN")
    (15030 . "HANGUL SYLLABLE SSANGCIEUC YA IEUNG")
    (15031 . "HANGUL SYLLABLE SSANGCIEUC EO")
    (15032 . "HANGUL SYLLABLE SSANGCIEUC EO KIYEOK")
    (15033 . "HANGUL SYLLABLE SSANGCIEUC EO NIEUN")
    (15034 . "HANGUL SYLLABLE SSANGCIEUC EO RIEUL")
    (15035 . "HANGUL SYLLABLE SSANGCIEUC EO MIEUM")
    (15036 . "HANGUL SYLLABLE SSANGCIEUC EO PIEUP")
    (15037 . "HANGUL SYLLABLE SSANGCIEUC EO SIOS")
    (15038 . "HANGUL SYLLABLE SSANGCIEUC EO SSANGSIOS")
    (15039 . "HANGUL SYLLABLE SSANGCIEUC EO IEUNG")
    (15040 . "HANGUL SYLLABLE SSANGCIEUC E")
    (15041 . "HANGUL SYLLABLE SSANGCIEUC E IEUNG")
    (15042 . "HANGUL SYLLABLE SSANGCIEUC YEO")
    (15043 . "HANGUL SYLLABLE SSANGCIEUC YEO SSANGSIOS")
    (15044 . "HANGUL SYLLABLE SSANGCIEUC O")
    (15045 . "HANGUL SYLLABLE SSANGCIEUC O KIYEOK")
    (15046 . "HANGUL SYLLABLE SSANGCIEUC O NIEUN")
    (15047 . "HANGUL SYLLABLE SSANGCIEUC O RIEUL")
    (15048 . "HANGUL SYLLABLE SSANGCIEUC O MIEUM")
    (15049 . "HANGUL SYLLABLE SSANGCIEUC O PIEUP")
    (15050 . "HANGUL SYLLABLE SSANGCIEUC O SIOS")
    (15051 . "HANGUL SYLLABLE SSANGCIEUC O IEUNG")
    (15052 . "HANGUL SYLLABLE SSANGCIEUC O CHIEUCH")
    (15053 . "HANGUL SYLLABLE SSANGCIEUC WA")
    (15054 . "HANGUL SYLLABLE SSANGCIEUC WA KIYEOK")
    (15055 . "HANGUL SYLLABLE SSANGCIEUC WA RIEUL")
    (15056 . "HANGUL SYLLABLE SSANGCIEUC WA SSANGSIOS")
    (15057 . "HANGUL SYLLABLE SSANGCIEUC WAE")
    (15058 . "HANGUL SYLLABLE SSANGCIEUC WAE SSANGSIOS")
    (15059 . "HANGUL SYLLABLE SSANGCIEUC OE")
    (15060 . "HANGUL SYLLABLE SSANGCIEUC OE NIEUN")
    (15061 . "HANGUL SYLLABLE SSANGCIEUC OE RIEUL")
    (15062 . "HANGUL SYLLABLE SSANGCIEUC OE MIEUM")
    (15063 . "HANGUL SYLLABLE SSANGCIEUC OE PIEUP")
    (15064 . "HANGUL SYLLABLE SSANGCIEUC YO IEUNG")
    (15065 . "HANGUL SYLLABLE SSANGCIEUC U")
    (15066 . "HANGUL SYLLABLE SSANGCIEUC U KIYEOK")
    (15067 . "HANGUL SYLLABLE SSANGCIEUC U NIEUN")
    (15068 . "HANGUL SYLLABLE SSANGCIEUC U RIEUL")
    (15069 . "HANGUL SYLLABLE SSANGCIEUC U MIEUM")
    (15070 . "HANGUL SYLLABLE SSANGCIEUC U PIEUP")
    (15071 . "HANGUL SYLLABLE SSANGCIEUC U IEUNG")
    (15072 . "HANGUL SYLLABLE SSANGCIEUC WEO")
    (15073 . "HANGUL SYLLABLE SSANGCIEUC WEO SSANGSIOS")
    (15074 . "HANGUL SYLLABLE SSANGCIEUC WEO IEUNG")
    (15075 . "HANGUL SYLLABLE SSANGCIEUC WI")
    (15076 . "HANGUL SYLLABLE SSANGCIEUC YU")
    (15077 . "HANGUL SYLLABLE SSANGCIEUC EU")
    (15078 . "HANGUL SYLLABLE SSANGCIEUC EU MIEUM")
    (15079 . "HANGUL SYLLABLE SSANGCIEUC EU SIOS")
    (15080 . "HANGUL SYLLABLE SSANGCIEUC EU IEUNG")
    (15081 . "HANGUL SYLLABLE SSANGCIEUC I")
    (15082 . "HANGUL SYLLABLE SSANGCIEUC I KIYEOK")
    (15083 . "HANGUL SYLLABLE SSANGCIEUC I NIEUN")
    (15084 . "HANGUL SYLLABLE SSANGCIEUC I RIEUL")
    (15085 . "HANGUL SYLLABLE SSANGCIEUC I MIEUM")
    (15086 . "HANGUL SYLLABLE SSANGCIEUC I PIEUP")
    (15087 . "HANGUL SYLLABLE SSANGCIEUC I IEUNG")
    (15088 . "HANGUL SYLLABLE SSANGCIEUC I CIEUC")
    (15089 . "HANGUL SYLLABLE SSANGCIEUC I HIEUH")
    (15090 . "HANGUL SYLLABLE CHIEUCH A")
    (15091 . "HANGUL SYLLABLE CHIEUCH A KIYEOK")
    (15092 . "HANGUL SYLLABLE CHIEUCH A NIEUN")
    (15093 . "HANGUL SYLLABLE CHIEUCH A NIEUN-HIEUH")
    (15094 . "HANGUL SYLLABLE CHIEUCH A RIEUL")
    (15095 . "HANGUL SYLLABLE CHIEUCH A MIEUM")
    (15096 . "HANGUL SYLLABLE CHIEUCH A PIEUP")
    (15097 . "HANGUL SYLLABLE CHIEUCH A SIOS")
    (15098 . "HANGUL SYLLABLE CHIEUCH A SSANGSIOS")
    (15099 . "HANGUL SYLLABLE CHIEUCH A IEUNG")
    (15100 . "HANGUL SYLLABLE CHIEUCH A CIEUC")
    (15101 . "HANGUL SYLLABLE CHIEUCH AE")
    (15102 . "HANGUL SYLLABLE CHIEUCH AE KIYEOK")
    (15103 . "HANGUL SYLLABLE CHIEUCH AE NIEUN")
    (15104 . "HANGUL SYLLABLE CHIEUCH AE RIEUL")
    (15105 . "HANGUL SYLLABLE CHIEUCH AE MIEUM")
    (15106 . "HANGUL SYLLABLE CHIEUCH AE PIEUP")
    (15107 . "HANGUL SYLLABLE CHIEUCH AE SIOS")
    (15108 . "HANGUL SYLLABLE CHIEUCH AE SSANGSIOS")
    (15109 . "HANGUL SYLLABLE CHIEUCH AE IEUNG")
    (15110 . "HANGUL SYLLABLE CHIEUCH YA")
    (15111 . "HANGUL SYLLABLE CHIEUCH YA NIEUN")
    (15112 . "HANGUL SYLLABLE CHIEUCH YA NIEUN-HIEUH")
    (15113 . "HANGUL SYLLABLE CHIEUCH YA RIEUL")
    (15114 . "HANGUL SYLLABLE CHIEUCH YA MIEUM")
    (15115 . "HANGUL SYLLABLE CHIEUCH YA IEUNG")
    (15116 . "HANGUL SYLLABLE CHIEUCH EO")
    (15117 . "HANGUL SYLLABLE CHIEUCH EO KIYEOK")
    (15118 . "HANGUL SYLLABLE CHIEUCH EO NIEUN")
    (15119 . "HANGUL SYLLABLE CHIEUCH EO RIEUL")
    (15120 . "HANGUL SYLLABLE CHIEUCH EO MIEUM")
    (15121 . "HANGUL SYLLABLE CHIEUCH EO PIEUP")
    (15122 . "HANGUL SYLLABLE CHIEUCH EO SIOS")
    (15123 . "HANGUL SYLLABLE CHIEUCH EO SSANGSIOS")
    (15124 . "HANGUL SYLLABLE CHIEUCH EO IEUNG")
    (15125 . "HANGUL SYLLABLE CHIEUCH E")
    (15126 . "HANGUL SYLLABLE CHIEUCH E KIYEOK")
    (15127 . "HANGUL SYLLABLE CHIEUCH E NIEUN")
    (15128 . "HANGUL SYLLABLE CHIEUCH E RIEUL")
    (15129 . "HANGUL SYLLABLE CHIEUCH E MIEUM")
    (15130 . "HANGUL SYLLABLE CHIEUCH E PIEUP")
    (15131 . "HANGUL SYLLABLE CHIEUCH E SIOS")
    (15132 . "HANGUL SYLLABLE CHIEUCH E IEUNG")
    (15133 . "HANGUL SYLLABLE CHIEUCH YEO")
    (15134 . "HANGUL SYLLABLE CHIEUCH YEO NIEUN")
    (15135 . "HANGUL SYLLABLE CHIEUCH YEO SSANGSIOS")
    (15136 . "HANGUL SYLLABLE CHIEUCH YE")
    (15137 . "HANGUL SYLLABLE CHIEUCH YE NIEUN")
    (15138 . "HANGUL SYLLABLE CHIEUCH YE IEUNG")
    (15139 . "HANGUL SYLLABLE CHIEUCH O")
    (15140 . "HANGUL SYLLABLE CHIEUCH O KIYEOK")
    (15141 . "HANGUL SYLLABLE CHIEUCH O NIEUN")
    (15142 . "HANGUL SYLLABLE CHIEUCH O RIEUL")
    (15143 . "HANGUL SYLLABLE CHIEUCH O MIEUM")
    (15144 . "HANGUL SYLLABLE CHIEUCH O PIEUP")
    (15145 . "HANGUL SYLLABLE CHIEUCH O SIOS")
    (15146 . "HANGUL SYLLABLE CHIEUCH O IEUNG")
    (15147 . "HANGUL SYLLABLE CHIEUCH WA")
    (15148 . "HANGUL SYLLABLE CHIEUCH WA NIEUN")
    (15149 . "HANGUL SYLLABLE CHIEUCH WA RIEUL")
    (15150 . "HANGUL SYLLABLE CHIEUCH WA IEUNG")
    (15151 . "HANGUL SYLLABLE CHIEUCH OE")
    (15152 . "HANGUL SYLLABLE CHIEUCH OE NIEUN")
    (15153 . "HANGUL SYLLABLE CHIEUCH OE RIEUL")
    (15154 . "HANGUL SYLLABLE CHIEUCH OE MIEUM")
    (15155 . "HANGUL SYLLABLE CHIEUCH OE PIEUP")
    (15156 . "HANGUL SYLLABLE CHIEUCH OE SIOS")
    (15157 . "HANGUL SYLLABLE CHIEUCH OE IEUNG")
    (15158 . "HANGUL SYLLABLE CHIEUCH YO")
    (15159 . "HANGUL SYLLABLE CHIEUCH YO MIEUM")
    (15160 . "HANGUL SYLLABLE CHIEUCH U")
    (15161 . "HANGUL SYLLABLE CHIEUCH U KIYEOK")
    (15162 . "HANGUL SYLLABLE CHIEUCH U NIEUN")
    (15163 . "HANGUL SYLLABLE CHIEUCH U RIEUL")
    (15164 . "HANGUL SYLLABLE CHIEUCH U MIEUM")
    (15165 . "HANGUL SYLLABLE CHIEUCH U PIEUP")
    (15166 . "HANGUL SYLLABLE CHIEUCH U SIOS")
    (15167 . "HANGUL SYLLABLE CHIEUCH U IEUNG")
    (15168 . "HANGUL SYLLABLE CHIEUCH WEO")
    (15169 . "HANGUL SYLLABLE CHIEUCH WEO SSANGSIOS")
    (15170 . "HANGUL SYLLABLE CHIEUCH WE")
    (15171 . "HANGUL SYLLABLE CHIEUCH WE NIEUN")
    (15172 . "HANGUL SYLLABLE CHIEUCH WI")
    (15173 . "HANGUL SYLLABLE CHIEUCH WI NIEUN")
    (15174 . "HANGUL SYLLABLE CHIEUCH WI RIEUL")
    (15175 . "HANGUL SYLLABLE CHIEUCH WI MIEUM")
    (15176 . "HANGUL SYLLABLE CHIEUCH WI PIEUP")
    (15177 . "HANGUL SYLLABLE CHIEUCH WI SIOS")
    (15178 . "HANGUL SYLLABLE CHIEUCH WI IEUNG")
    (15179 . "HANGUL SYLLABLE CHIEUCH YU")
    (15180 . "HANGUL SYLLABLE CHIEUCH YU NIEUN")
    (15181 . "HANGUL SYLLABLE CHIEUCH YU RIEUL")
    (15182 . "HANGUL SYLLABLE CHIEUCH YU MIEUM")
    (15183 . "HANGUL SYLLABLE CHIEUCH YU IEUNG")
    (15184 . "HANGUL SYLLABLE CHIEUCH EU")
    (15185 . "HANGUL SYLLABLE CHIEUCH EU KIYEOK")
    (15186 . "HANGUL SYLLABLE CHIEUCH EU NIEUN")
    (15187 . "HANGUL SYLLABLE CHIEUCH EU RIEUL")
    (15188 . "HANGUL SYLLABLE CHIEUCH EU MIEUM")
    (15189 . "HANGUL SYLLABLE CHIEUCH EU PIEUP")
    (15190 . "HANGUL SYLLABLE CHIEUCH EU SIOS")
    (15191 . "HANGUL SYLLABLE CHIEUCH EU IEUNG")
    (15192 . "HANGUL SYLLABLE CHIEUCH I")
    (15193 . "HANGUL SYLLABLE CHIEUCH I KIYEOK")
    (15194 . "HANGUL SYLLABLE CHIEUCH I NIEUN")
    (15195 . "HANGUL SYLLABLE CHIEUCH I TIKEUT")
    (15196 . "HANGUL SYLLABLE CHIEUCH I RIEUL")
    (15197 . "HANGUL SYLLABLE CHIEUCH I RIEUL-KIYEOK")
    (15198 . "HANGUL SYLLABLE CHIEUCH I MIEUM")
    (15199 . "HANGUL SYLLABLE CHIEUCH I PIEUP")
    (15200 . "HANGUL SYLLABLE CHIEUCH I SIOS")
    (15201 . "HANGUL SYLLABLE CHIEUCH I IEUNG")
    (15202 . "HANGUL SYLLABLE KHIEUKH A")
    (15203 . "HANGUL SYLLABLE KHIEUKH A KIYEOK")
    (15204 . "HANGUL SYLLABLE KHIEUKH A NIEUN")
    (15205 . "HANGUL SYLLABLE KHIEUKH A RIEUL")
    (15206 . "HANGUL SYLLABLE KHIEUKH A MIEUM")
    (15207 . "HANGUL SYLLABLE KHIEUKH A PIEUP")
    (15208 . "HANGUL SYLLABLE KHIEUKH A SIOS")
    (15209 . "HANGUL SYLLABLE KHIEUKH A IEUNG")
    (15210 . "HANGUL SYLLABLE KHIEUKH AE")
    (15211 . "HANGUL SYLLABLE KHIEUKH AE KIYEOK")
    (15212 . "HANGUL SYLLABLE KHIEUKH AE NIEUN")
    (15213 . "HANGUL SYLLABLE KHIEUKH AE RIEUL")
    (15214 . "HANGUL SYLLABLE KHIEUKH AE MIEUM")
    (15215 . "HANGUL SYLLABLE KHIEUKH AE PIEUP")
    (15216 . "HANGUL SYLLABLE KHIEUKH AE SIOS")
    (15217 . "HANGUL SYLLABLE KHIEUKH AE SSANGSIOS")
    (15218 . "HANGUL SYLLABLE KHIEUKH AE IEUNG")
    (15219 . "HANGUL SYLLABLE KHIEUKH YA")
    (15220 . "HANGUL SYLLABLE KHIEUKH YA KIYEOK")
    (15221 . "HANGUL SYLLABLE KHIEUKH YA IEUNG")
    (15222 . "HANGUL SYLLABLE KHIEUKH EO")
    (15223 . "HANGUL SYLLABLE KHIEUKH EO KIYEOK")
    (15224 . "HANGUL SYLLABLE KHIEUKH EO NIEUN")
    (15225 . "HANGUL SYLLABLE KHIEUKH EO TIKEUT")
    (15226 . "HANGUL SYLLABLE KHIEUKH EO RIEUL")
    (15227 . "HANGUL SYLLABLE KHIEUKH EO MIEUM")
    (15228 . "HANGUL SYLLABLE KHIEUKH EO PIEUP")
    (15229 . "HANGUL SYLLABLE KHIEUKH EO SIOS")
    (15230 . "HANGUL SYLLABLE KHIEUKH EO SSANGSIOS")
    (15231 . "HANGUL SYLLABLE KHIEUKH EO IEUNG")
    (15232 . "HANGUL SYLLABLE KHIEUKH E")
    (15233 . "HANGUL SYLLABLE KHIEUKH E KIYEOK")
    (15234 . "HANGUL SYLLABLE KHIEUKH E NIEUN")
    (15235 . "HANGUL SYLLABLE KHIEUKH E RIEUL")
    (15236 . "HANGUL SYLLABLE KHIEUKH E MIEUM")
    (15237 . "HANGUL SYLLABLE KHIEUKH E PIEUP")
    (15238 . "HANGUL SYLLABLE KHIEUKH E SIOS")
    (15239 . "HANGUL SYLLABLE KHIEUKH E IEUNG")
    (15240 . "HANGUL SYLLABLE KHIEUKH YEO")
    (15241 . "HANGUL SYLLABLE KHIEUKH YEO NIEUN")
    (15242 . "HANGUL SYLLABLE KHIEUKH YEO RIEUL")
    (15243 . "HANGUL SYLLABLE KHIEUKH YEO MIEUM")
    (15244 . "HANGUL SYLLABLE KHIEUKH YEO PIEUP")
    (15245 . "HANGUL SYLLABLE KHIEUKH YEO SIOS")
    (15246 . "HANGUL SYLLABLE KHIEUKH YEO SSANGSIOS")
    (15247 . "HANGUL SYLLABLE KHIEUKH YEO IEUNG")
    (15248 . "HANGUL SYLLABLE KHIEUKH YE")
    (15249 . "HANGUL SYLLABLE KHIEUKH O")
    (15250 . "HANGUL SYLLABLE KHIEUKH O KIYEOK")
    (15251 . "HANGUL SYLLABLE KHIEUKH O NIEUN")
    (15252 . "HANGUL SYLLABLE KHIEUKH O RIEUL")
    (15253 . "HANGUL SYLLABLE KHIEUKH O MIEUM")
    (15254 . "HANGUL SYLLABLE KHIEUKH O PIEUP")
    (15255 . "HANGUL SYLLABLE KHIEUKH O SIOS")
    (15256 . "HANGUL SYLLABLE KHIEUKH O IEUNG")
    (15257 . "HANGUL SYLLABLE KHIEUKH WA")
    (15258 . "HANGUL SYLLABLE KHIEUKH WA KIYEOK")
    (15259 . "HANGUL SYLLABLE KHIEUKH WA NIEUN")
    (15260 . "HANGUL SYLLABLE KHIEUKH WA RIEUL")
    (15261 . "HANGUL SYLLABLE KHIEUKH WA MIEUM")
    (15262 . "HANGUL SYLLABLE KHIEUKH WA IEUNG")
    (15263 . "HANGUL SYLLABLE KHIEUKH WAE")
    (15264 . "HANGUL SYLLABLE KHIEUKH WAE IEUNG")
    (15265 . "HANGUL SYLLABLE KHIEUKH OE")
    (15266 . "HANGUL SYLLABLE KHIEUKH OE RIEUL")
    (15267 . "HANGUL SYLLABLE KHIEUKH YO")
    (15268 . "HANGUL SYLLABLE KHIEUKH U")
    (15269 . "HANGUL SYLLABLE KHIEUKH U KIYEOK")
    (15270 . "HANGUL SYLLABLE KHIEUKH U NIEUN")
    (15271 . "HANGUL SYLLABLE KHIEUKH U RIEUL")
    (15272 . "HANGUL SYLLABLE KHIEUKH U MIEUM")
    (15273 . "HANGUL SYLLABLE KHIEUKH U PIEUP")
    (15274 . "HANGUL SYLLABLE KHIEUKH U SIOS")
    (15275 . "HANGUL SYLLABLE KHIEUKH U IEUNG")
    (15276 . "HANGUL SYLLABLE KHIEUKH WEO")
    (15277 . "HANGUL SYLLABLE KHIEUKH WEO NIEUN")
    (15278 . "HANGUL SYLLABLE KHIEUKH WEO RIEUL")
    (15279 . "HANGUL SYLLABLE KHIEUKH WEO IEUNG")
    (15280 . "HANGUL SYLLABLE KHIEUKH WE")
    (15281 . "HANGUL SYLLABLE KHIEUKH WE IEUNG")
    (15282 . "HANGUL SYLLABLE KHIEUKH WI")
    (15283 . "HANGUL SYLLABLE KHIEUKH WI KIYEOK")
    (15284 . "HANGUL SYLLABLE KHIEUKH WI NIEUN")
    (15285 . "HANGUL SYLLABLE KHIEUKH WI RIEUL")
    (15286 . "HANGUL SYLLABLE KHIEUKH WI MIEUM")
    (15287 . "HANGUL SYLLABLE KHIEUKH WI PIEUP")
    (15288 . "HANGUL SYLLABLE KHIEUKH WI SIOS")
    (15289 . "HANGUL SYLLABLE KHIEUKH WI IEUNG")
    (15290 . "HANGUL SYLLABLE KHIEUKH YU")
    (15291 . "HANGUL SYLLABLE KHIEUKH YU NIEUN")
    (15292 . "HANGUL SYLLABLE KHIEUKH YU RIEUL")
    (15293 . "HANGUL SYLLABLE KHIEUKH YU MIEUM")
    (15294 . "HANGUL SYLLABLE KHIEUKH EU")
    (15295 . "HANGUL SYLLABLE KHIEUKH EU KIYEOK")
    (15296 . "HANGUL SYLLABLE KHIEUKH EU NIEUN")
    (15297 . "HANGUL SYLLABLE KHIEUKH EU RIEUL")
    (15298 . "HANGUL SYLLABLE KHIEUKH EU MIEUM")
    (15299 . "HANGUL SYLLABLE KHIEUKH EU PIEUP")
    (15300 . "HANGUL SYLLABLE KHIEUKH EU IEUNG")
    (15301 . "HANGUL SYLLABLE KHIEUKH I")
    (15302 . "HANGUL SYLLABLE KHIEUKH I KIYEOK")
    (15303 . "HANGUL SYLLABLE KHIEUKH I NIEUN")
    (15304 . "HANGUL SYLLABLE KHIEUKH I RIEUL")
    (15305 . "HANGUL SYLLABLE KHIEUKH I MIEUM")
    (15306 . "HANGUL SYLLABLE KHIEUKH I PIEUP")
    (15307 . "HANGUL SYLLABLE KHIEUKH I SIOS")
    (15308 . "HANGUL SYLLABLE KHIEUKH I IEUNG")
    (15309 . "HANGUL SYLLABLE THIEUTH A")
    (15310 . "HANGUL SYLLABLE THIEUTH A KIYEOK")
    (15311 . "HANGUL SYLLABLE THIEUTH A NIEUN")
    (15312 . "HANGUL SYLLABLE THIEUTH A RIEUL")
    (15313 . "HANGUL SYLLABLE THIEUTH A RIEUL-KIYEOK")
    (15314 . "HANGUL SYLLABLE THIEUTH A MIEUM")
    (15315 . "HANGUL SYLLABLE THIEUTH A PIEUP")
    (15316 . "HANGUL SYLLABLE THIEUTH A SIOS")
    (15317 . "HANGUL SYLLABLE THIEUTH A SSANGSIOS")
    (15318 . "HANGUL SYLLABLE THIEUTH A IEUNG")
    (15319 . "HANGUL SYLLABLE THIEUTH AE")
    (15320 . "HANGUL SYLLABLE THIEUTH AE KIYEOK")
    (15321 . "HANGUL SYLLABLE THIEUTH AE NIEUN")
    (15322 . "HANGUL SYLLABLE THIEUTH AE RIEUL")
    (15323 . "HANGUL SYLLABLE THIEUTH AE MIEUM")
    (15324 . "HANGUL SYLLABLE THIEUTH AE PIEUP")
    (15325 . "HANGUL SYLLABLE THIEUTH AE SIOS")
    (15326 . "HANGUL SYLLABLE THIEUTH AE SSANGSIOS")
    (15327 . "HANGUL SYLLABLE THIEUTH AE IEUNG")
    (15328 . "HANGUL SYLLABLE THIEUTH YA")
    (15329 . "HANGUL SYLLABLE THIEUTH YA IEUNG")
    (15330 . "HANGUL SYLLABLE THIEUTH EO")
    (15331 . "HANGUL SYLLABLE THIEUTH EO KIYEOK")
    (15332 . "HANGUL SYLLABLE THIEUTH EO NIEUN")
    (15333 . "HANGUL SYLLABLE THIEUTH EO RIEUL")
    (15334 . "HANGUL SYLLABLE THIEUTH EO RIEUL-MIEUM")
    (15335 . "HANGUL SYLLABLE THIEUTH EO MIEUM")
    (15336 . "HANGUL SYLLABLE THIEUTH EO PIEUP")
    (15337 . "HANGUL SYLLABLE THIEUTH EO SIOS")
    (15338 . "HANGUL SYLLABLE THIEUTH EO SSANGSIOS")
    (15339 . "HANGUL SYLLABLE THIEUTH EO IEUNG")
    (15340 . "HANGUL SYLLABLE THIEUTH E")
    (15341 . "HANGUL SYLLABLE THIEUTH E KIYEOK")
    (15342 . "HANGUL SYLLABLE THIEUTH E NIEUN")
    (15343 . "HANGUL SYLLABLE THIEUTH E RIEUL")
    (15344 . "HANGUL SYLLABLE THIEUTH E MIEUM")
    (15345 . "HANGUL SYLLABLE THIEUTH E PIEUP")
    (15346 . "HANGUL SYLLABLE THIEUTH E SIOS")
    (15347 . "HANGUL SYLLABLE THIEUTH E IEUNG")
    (15348 . "HANGUL SYLLABLE THIEUTH YEO")
    (15349 . "HANGUL SYLLABLE THIEUTH YEO NIEUN")
    (15350 . "HANGUL SYLLABLE THIEUTH YEO SSANGSIOS")
    (15351 . "HANGUL SYLLABLE THIEUTH YE")
    (15352 . "HANGUL SYLLABLE THIEUTH YE NIEUN")
    (15353 . "HANGUL SYLLABLE THIEUTH O")
    (15354 . "HANGUL SYLLABLE THIEUTH O KIYEOK")
    (15355 . "HANGUL SYLLABLE THIEUTH O NIEUN")
    (15356 . "HANGUL SYLLABLE THIEUTH O RIEUL")
    (15357 . "HANGUL SYLLABLE THIEUTH O MIEUM")
    (15358 . "HANGUL SYLLABLE THIEUTH O PIEUP")
    (15359 . "HANGUL SYLLABLE THIEUTH O SIOS")
    (15360 . "HANGUL SYLLABLE THIEUTH O IEUNG")
    (15361 . "HANGUL SYLLABLE THIEUTH O PHIEUPH")
    (15362 . "HANGUL SYLLABLE THIEUTH WA")
    (15363 . "HANGUL SYLLABLE THIEUTH WA NIEUN")
    (15364 . "HANGUL SYLLABLE THIEUTH WAE")
    (15365 . "HANGUL SYLLABLE THIEUTH OE")
    (15366 . "HANGUL SYLLABLE THIEUTH OE NIEUN")
    (15367 . "HANGUL SYLLABLE THIEUTH OE SIOS")
    (15368 . "HANGUL SYLLABLE THIEUTH OE IEUNG")
    (15369 . "HANGUL SYLLABLE THIEUTH YO")
    (15370 . "HANGUL SYLLABLE THIEUTH U")
    (15371 . "HANGUL SYLLABLE THIEUTH U KIYEOK")
    (15372 . "HANGUL SYLLABLE THIEUTH U NIEUN")
    (15373 . "HANGUL SYLLABLE THIEUTH U RIEUL")
    (15374 . "HANGUL SYLLABLE THIEUTH U MIEUM")
    (15375 . "HANGUL SYLLABLE THIEUTH U PIEUP")
    (15376 . "HANGUL SYLLABLE THIEUTH U SIOS")
    (15377 . "HANGUL SYLLABLE THIEUTH U IEUNG")
    (15378 . "HANGUL SYLLABLE THIEUTH WEO")
    (15379 . "HANGUL SYLLABLE THIEUTH WEO SSANGSIOS")
    (15380 . "HANGUL SYLLABLE THIEUTH WE")
    (15381 . "HANGUL SYLLABLE THIEUTH WI")
    (15382 . "HANGUL SYLLABLE THIEUTH WI KIYEOK")
    (15383 . "HANGUL SYLLABLE THIEUTH WI NIEUN")
    (15384 . "HANGUL SYLLABLE THIEUTH WI RIEUL")
    (15385 . "HANGUL SYLLABLE THIEUTH WI MIEUM")
    (15386 . "HANGUL SYLLABLE THIEUTH WI PIEUP")
    (15387 . "HANGUL SYLLABLE THIEUTH WI IEUNG")
    (15388 . "HANGUL SYLLABLE THIEUTH YU")
    (15389 . "HANGUL SYLLABLE THIEUTH YU NIEUN")
    (15390 . "HANGUL SYLLABLE THIEUTH YU RIEUL")
    (15391 . "HANGUL SYLLABLE THIEUTH YU MIEUM")
    (15392 . "HANGUL SYLLABLE THIEUTH YU IEUNG")
    (15393 . "HANGUL SYLLABLE THIEUTH EU")
    (15394 . "HANGUL SYLLABLE THIEUTH EU KIYEOK")
    (15395 . "HANGUL SYLLABLE THIEUTH EU NIEUN")
    (15396 . "HANGUL SYLLABLE THIEUTH EU TIKEUT")
    (15397 . "HANGUL SYLLABLE THIEUTH EU RIEUL")
    (15398 . "HANGUL SYLLABLE THIEUTH EU RIEUL-MIEUM")
    (15399 . "HANGUL SYLLABLE THIEUTH EU MIEUM")
    (15400 . "HANGUL SYLLABLE THIEUTH EU PIEUP")
    (15401 . "HANGUL SYLLABLE THIEUTH EU SIOS")
    (15402 . "HANGUL SYLLABLE THIEUTH YI")
    (15403 . "HANGUL SYLLABLE THIEUTH YI NIEUN")
    (15404 . "HANGUL SYLLABLE THIEUTH YI RIEUL")
    (15405 . "HANGUL SYLLABLE THIEUTH YI MIEUM")
    (15406 . "HANGUL SYLLABLE THIEUTH YI PIEUP")
    (15407 . "HANGUL SYLLABLE THIEUTH I")
    (15408 . "HANGUL SYLLABLE THIEUTH I KIYEOK")
    (15409 . "HANGUL SYLLABLE THIEUTH I NIEUN")
    (15410 . "HANGUL SYLLABLE THIEUTH I RIEUL")
    (15411 . "HANGUL SYLLABLE THIEUTH I MIEUM")
    (15412 . "HANGUL SYLLABLE THIEUTH I PIEUP")
    (15413 . "HANGUL SYLLABLE THIEUTH I SIOS")
    (15414 . "HANGUL SYLLABLE THIEUTH I IEUNG")
    (15415 . "HANGUL SYLLABLE PHIEUPH A")
    (15416 . "HANGUL SYLLABLE PHIEUPH A KIYEOK")
    (15417 . "HANGUL SYLLABLE PHIEUPH A SSANGKIYEOK")
    (15418 . "HANGUL SYLLABLE PHIEUPH A NIEUN")
    (15419 . "HANGUL SYLLABLE PHIEUPH A RIEUL")
    (15420 . "HANGUL SYLLABLE PHIEUPH A RIEUL-MIEUM")
    (15421 . "HANGUL SYLLABLE PHIEUPH A MIEUM")
    (15422 . "HANGUL SYLLABLE PHIEUPH A PIEUP")
    (15423 . "HANGUL SYLLABLE PHIEUPH A SIOS")
    (15424 . "HANGUL SYLLABLE PHIEUPH A SSANGSIOS")
    (15425 . "HANGUL SYLLABLE PHIEUPH A IEUNG")
    (15426 . "HANGUL SYLLABLE PHIEUPH A THIEUTH")
    (15427 . "HANGUL SYLLABLE PHIEUPH AE")
    (15428 . "HANGUL SYLLABLE PHIEUPH AE KIYEOK")
    (15429 . "HANGUL SYLLABLE PHIEUPH AE NIEUN")
    (15430 . "HANGUL SYLLABLE PHIEUPH AE RIEUL")
    (15431 . "HANGUL SYLLABLE PHIEUPH AE MIEUM")
    (15432 . "HANGUL SYLLABLE PHIEUPH AE PIEUP")
    (15433 . "HANGUL SYLLABLE PHIEUPH AE SIOS")
    (15434 . "HANGUL SYLLABLE PHIEUPH AE SSANGSIOS")
    (15435 . "HANGUL SYLLABLE PHIEUPH AE IEUNG")
    (15436 . "HANGUL SYLLABLE PHIEUPH YA")
    (15437 . "HANGUL SYLLABLE PHIEUPH YA KIYEOK")
    (15438 . "HANGUL SYLLABLE PHIEUPH EO")
    (15439 . "HANGUL SYLLABLE PHIEUPH EO KIYEOK")
    (15440 . "HANGUL SYLLABLE PHIEUPH EO NIEUN")
    (15441 . "HANGUL SYLLABLE PHIEUPH EO RIEUL")
    (15442 . "HANGUL SYLLABLE PHIEUPH EO MIEUM")
    (15443 . "HANGUL SYLLABLE PHIEUPH EO PIEUP")
    (15444 . "HANGUL SYLLABLE PHIEUPH EO SIOS")
    (15445 . "HANGUL SYLLABLE PHIEUPH EO SSANGSIOS")
    (15446 . "HANGUL SYLLABLE PHIEUPH EO IEUNG")
    (15447 . "HANGUL SYLLABLE PHIEUPH E")
    (15448 . "HANGUL SYLLABLE PHIEUPH E KIYEOK")
    (15449 . "HANGUL SYLLABLE PHIEUPH E NIEUN")
    (15450 . "HANGUL SYLLABLE PHIEUPH E RIEUL")
    (15451 . "HANGUL SYLLABLE PHIEUPH E MIEUM")
    (15452 . "HANGUL SYLLABLE PHIEUPH E PIEUP")
    (15453 . "HANGUL SYLLABLE PHIEUPH E SIOS")
    (15454 . "HANGUL SYLLABLE PHIEUPH E IEUNG")
    (15455 . "HANGUL SYLLABLE PHIEUPH YEO")
    (15456 . "HANGUL SYLLABLE PHIEUPH YEO NIEUN")
    (15457 . "HANGUL SYLLABLE PHIEUPH YEO RIEUL")
    (15458 . "HANGUL SYLLABLE PHIEUPH YEO MIEUM")
    (15459 . "HANGUL SYLLABLE PHIEUPH YEO PIEUP")
    (15460 . "HANGUL SYLLABLE PHIEUPH YEO SSANGSIOS")
    (15461 . "HANGUL SYLLABLE PHIEUPH YEO IEUNG")
    (15462 . "HANGUL SYLLABLE PHIEUPH YE")
    (15463 . "HANGUL SYLLABLE PHIEUPH YE RIEUL")
    (15464 . "HANGUL SYLLABLE PHIEUPH YE PIEUP")
    (15465 . "HANGUL SYLLABLE PHIEUPH YE SIOS")
    (15466 . "HANGUL SYLLABLE PHIEUPH O")
    (15467 . "HANGUL SYLLABLE PHIEUPH O KIYEOK")
    (15468 . "HANGUL SYLLABLE PHIEUPH O NIEUN")
    (15469 . "HANGUL SYLLABLE PHIEUPH O RIEUL")
    (15470 . "HANGUL SYLLABLE PHIEUPH O MIEUM")
    (15471 . "HANGUL SYLLABLE PHIEUPH O PIEUP")
    (15472 . "HANGUL SYLLABLE PHIEUPH O SIOS")
    (15473 . "HANGUL SYLLABLE PHIEUPH O IEUNG")
    (15474 . "HANGUL SYLLABLE PHIEUPH WA")
    (15475 . "HANGUL SYLLABLE PHIEUPH WA IEUNG")
    (15476 . "HANGUL SYLLABLE PHIEUPH OE")
    (15477 . "HANGUL SYLLABLE PHIEUPH OE NIEUN")
    (15478 . "HANGUL SYLLABLE PHIEUPH YO")
    (15479 . "HANGUL SYLLABLE PHIEUPH YO NIEUN")
    (15480 . "HANGUL SYLLABLE PHIEUPH YO RIEUL")
    (15481 . "HANGUL SYLLABLE PHIEUPH YO PIEUP")
    (15482 . "HANGUL SYLLABLE PHIEUPH YO SIOS")
    (15483 . "HANGUL SYLLABLE PHIEUPH U")
    (15484 . "HANGUL SYLLABLE PHIEUPH U KIYEOK")
    (15485 . "HANGUL SYLLABLE PHIEUPH U NIEUN")
    (15486 . "HANGUL SYLLABLE PHIEUPH U TIKEUT")
    (15487 . "HANGUL SYLLABLE PHIEUPH U RIEUL")
    (15488 . "HANGUL SYLLABLE PHIEUPH U RIEUL-MIEUM")
    (15489 . "HANGUL SYLLABLE PHIEUPH U MIEUM")
    (15490 . "HANGUL SYLLABLE PHIEUPH U PIEUP")
    (15491 . "HANGUL SYLLABLE PHIEUPH U SIOS")
    (15492 . "HANGUL SYLLABLE PHIEUPH U IEUNG")
    (15493 . "HANGUL SYLLABLE PHIEUPH WEO")
    (15494 . "HANGUL SYLLABLE PHIEUPH WEO IEUNG")
    (15495 . "HANGUL SYLLABLE PHIEUPH WI")
    (15496 . "HANGUL SYLLABLE PHIEUPH WI NIEUN")
    (15497 . "HANGUL SYLLABLE PHIEUPH WI RIEUL")
    (15498 . "HANGUL SYLLABLE PHIEUPH WI MIEUM")
    (15499 . "HANGUL SYLLABLE PHIEUPH WI SIOS")
    (15500 . "HANGUL SYLLABLE PHIEUPH YU")
    (15501 . "HANGUL SYLLABLE PHIEUPH YU NIEUN")
    (15502 . "HANGUL SYLLABLE PHIEUPH YU RIEUL")
    (15503 . "HANGUL SYLLABLE PHIEUPH YU MIEUM")
    (15504 . "HANGUL SYLLABLE PHIEUPH YU SIOS")
    (15505 . "HANGUL SYLLABLE PHIEUPH YU IEUNG")
    (15506 . "HANGUL SYLLABLE PHIEUPH EU")
    (15507 . "HANGUL SYLLABLE PHIEUPH EU NIEUN")
    (15508 . "HANGUL SYLLABLE PHIEUPH EU RIEUL")
    (15509 . "HANGUL SYLLABLE PHIEUPH EU MIEUM")
    (15510 . "HANGUL SYLLABLE PHIEUPH EU PIEUP")
    (15511 . "HANGUL SYLLABLE PHIEUPH EU SIOS")
    (15512 . "HANGUL SYLLABLE PHIEUPH I")
    (15513 . "HANGUL SYLLABLE PHIEUPH I KIYEOK")
    (15514 . "HANGUL SYLLABLE PHIEUPH I NIEUN")
    (15515 . "HANGUL SYLLABLE PHIEUPH I RIEUL")
    (15516 . "HANGUL SYLLABLE PHIEUPH I MIEUM")
    (15517 . "HANGUL SYLLABLE PHIEUPH I PIEUP")
    (15518 . "HANGUL SYLLABLE PHIEUPH I SIOS")
    (15519 . "HANGUL SYLLABLE PHIEUPH I IEUNG")
    (15520 . "HANGUL SYLLABLE HIEUH A")
    (15521 . "HANGUL SYLLABLE HIEUH A KIYEOK")
    (15522 . "HANGUL SYLLABLE HIEUH A NIEUN")
    (15523 . "HANGUL SYLLABLE HIEUH A RIEUL")
    (15524 . "HANGUL SYLLABLE HIEUH A RIEUL-THIEUTH")
    (15525 . "HANGUL SYLLABLE HIEUH A MIEUM")
    (15526 . "HANGUL SYLLABLE HIEUH A PIEUP")
    (15527 . "HANGUL SYLLABLE HIEUH A SIOS")
    (15528 . "HANGUL SYLLABLE HIEUH A IEUNG")
    (15529 . "HANGUL SYLLABLE HIEUH AE")
    (15530 . "HANGUL SYLLABLE HIEUH AE KIYEOK")
    (15531 . "HANGUL SYLLABLE HIEUH AE NIEUN")
    (15532 . "HANGUL SYLLABLE HIEUH AE RIEUL")
    (15533 . "HANGUL SYLLABLE HIEUH AE MIEUM")
    (15534 . "HANGUL SYLLABLE HIEUH AE PIEUP")
    (15535 . "HANGUL SYLLABLE HIEUH AE SIOS")
    (15536 . "HANGUL SYLLABLE HIEUH AE SSANGSIOS")
    (15537 . "HANGUL SYLLABLE HIEUH AE IEUNG")
    (15538 . "HANGUL SYLLABLE HIEUH YA")
    (15539 . "HANGUL SYLLABLE HIEUH YA IEUNG")
    (15540 . "HANGUL SYLLABLE HIEUH EO")
    (15541 . "HANGUL SYLLABLE HIEUH EO KIYEOK")
    (15542 . "HANGUL SYLLABLE HIEUH EO NIEUN")
    (15543 . "HANGUL SYLLABLE HIEUH EO RIEUL")
    (15544 . "HANGUL SYLLABLE HIEUH EO RIEUL-MIEUM")
    (15545 . "HANGUL SYLLABLE HIEUH EO MIEUM")
    (15546 . "HANGUL SYLLABLE HIEUH EO PIEUP")
    (15547 . "HANGUL SYLLABLE HIEUH EO SIOS")
    (15548 . "HANGUL SYLLABLE HIEUH EO IEUNG")
    (15549 . "HANGUL SYLLABLE HIEUH E")
    (15550 . "HANGUL SYLLABLE HIEUH E KIYEOK")
    (15551 . "HANGUL SYLLABLE HIEUH E NIEUN")
    (15552 . "HANGUL SYLLABLE HIEUH E RIEUL")
    (15553 . "HANGUL SYLLABLE HIEUH E MIEUM")
    (15554 . "HANGUL SYLLABLE HIEUH E PIEUP")
    (15555 . "HANGUL SYLLABLE HIEUH E SIOS")
    (15556 . "HANGUL SYLLABLE HIEUH E IEUNG")
    (15557 . "HANGUL SYLLABLE HIEUH YEO")
    (15558 . "HANGUL SYLLABLE HIEUH YEO KIYEOK")
    (15559 . "HANGUL SYLLABLE HIEUH YEO NIEUN")
    (15560 . "HANGUL SYLLABLE HIEUH YEO RIEUL")
    (15561 . "HANGUL SYLLABLE HIEUH YEO MIEUM")
    (15562 . "HANGUL SYLLABLE HIEUH YEO PIEUP")
    (15563 . "HANGUL SYLLABLE HIEUH YEO SIOS")
    (15564 . "HANGUL SYLLABLE HIEUH YEO SSANGSIOS")
    (15565 . "HANGUL SYLLABLE HIEUH YEO IEUNG")
    (15566 . "HANGUL SYLLABLE HIEUH YE")
    (15567 . "HANGUL SYLLABLE HIEUH YE NIEUN")
    (15568 . "HANGUL SYLLABLE HIEUH YE RIEUL")
    (15569 . "HANGUL SYLLABLE HIEUH YE PIEUP")
    (15570 . "HANGUL SYLLABLE HIEUH O")
    (15571 . "HANGUL SYLLABLE HIEUH O KIYEOK")
    (15572 . "HANGUL SYLLABLE HIEUH O NIEUN")
    (15573 . "HANGUL SYLLABLE HIEUH O RIEUL")
    (15574 . "HANGUL SYLLABLE HIEUH O RIEUL-THIEUTH")
    (15575 . "HANGUL SYLLABLE HIEUH O MIEUM")
    (15576 . "HANGUL SYLLABLE HIEUH O PIEUP")
    (15577 . "HANGUL SYLLABLE HIEUH O SIOS")
    (15578 . "HANGUL SYLLABLE HIEUH O IEUNG")
    (15579 . "HANGUL SYLLABLE HIEUH O THIEUTH")
    (15580 . "HANGUL SYLLABLE HIEUH WA")
    (15581 . "HANGUL SYLLABLE HIEUH WA KIYEOK")
    (15582 . "HANGUL SYLLABLE HIEUH WA NIEUN")
    (15583 . "HANGUL SYLLABLE HIEUH WA RIEUL")
    (15584 . "HANGUL SYLLABLE HIEUH WA SIOS")
    (15585 . "HANGUL SYLLABLE HIEUH WA IEUNG")
    (15586 . "HANGUL SYLLABLE HIEUH WAE")
    (15587 . "HANGUL SYLLABLE HIEUH WAE KIYEOK")
    (15588 . "HANGUL SYLLABLE HIEUH WAE NIEUN")
    (15589 . "HANGUL SYLLABLE HIEUH WAE SIOS")
    (15590 . "HANGUL SYLLABLE HIEUH WAE IEUNG")
    (15591 . "HANGUL SYLLABLE HIEUH OE")
    (15592 . "HANGUL SYLLABLE HIEUH OE KIYEOK")
    (15593 . "HANGUL SYLLABLE HIEUH OE NIEUN")
    (15594 . "HANGUL SYLLABLE HIEUH OE RIEUL")
    (15595 . "HANGUL SYLLABLE HIEUH OE PIEUP")
    (15596 . "HANGUL SYLLABLE HIEUH OE SIOS")
    (15597 . "HANGUL SYLLABLE HIEUH OE IEUNG")
    (15598 . "HANGUL SYLLABLE HIEUH YO")
    (15599 . "HANGUL SYLLABLE HIEUH YO NIEUN")
    (15600 . "HANGUL SYLLABLE HIEUH YO RIEUL")
    (15601 . "HANGUL SYLLABLE HIEUH YO PIEUP")
    (15602 . "HANGUL SYLLABLE HIEUH YO SIOS")
    (15603 . "HANGUL SYLLABLE HIEUH U")
    (15604 . "HANGUL SYLLABLE HIEUH U KIYEOK")
    (15605 . "HANGUL SYLLABLE HIEUH U NIEUN")
    (15606 . "HANGUL SYLLABLE HIEUH U RIEUL")
    (15607 . "HANGUL SYLLABLE HIEUH U RIEUL-THIEUTH")
    (15608 . "HANGUL SYLLABLE HIEUH U MIEUM")
    (15609 . "HANGUL SYLLABLE HIEUH U SIOS")
    (15610 . "HANGUL SYLLABLE HIEUH U IEUNG")
    (15611 . "HANGUL SYLLABLE HIEUH WEO")
    (15612 . "HANGUL SYLLABLE HIEUH WEO NIEUN")
    (15613 . "HANGUL SYLLABLE HIEUH WEO RIEUL")
    (15614 . "HANGUL SYLLABLE HIEUH WEO MIEUM")
    (15615 . "HANGUL SYLLABLE HIEUH WEO IEUNG")
    (15616 . "HANGUL SYLLABLE HIEUH WE")
    (15617 . "HANGUL SYLLABLE HIEUH WE KIYEOK")
    (15618 . "HANGUL SYLLABLE HIEUH WE NIEUN")
    (15619 . "HANGUL SYLLABLE HIEUH WE RIEUL")
    (15620 . "HANGUL SYLLABLE HIEUH WE IEUNG")
    (15621 . "HANGUL SYLLABLE HIEUH WI")
    (15622 . "HANGUL SYLLABLE HIEUH WI KIYEOK")
    (15623 . "HANGUL SYLLABLE HIEUH WI NIEUN")
    (15624 . "HANGUL SYLLABLE HIEUH WI RIEUL")
    (15625 . "HANGUL SYLLABLE HIEUH WI MIEUM")
    (15626 . "HANGUL SYLLABLE HIEUH WI PIEUP")
    (15627 . "HANGUL SYLLABLE HIEUH WI SIOS")
    (15628 . "HANGUL SYLLABLE HIEUH WI IEUNG")
    (15629 . "HANGUL SYLLABLE HIEUH YU")
    (15630 . "HANGUL SYLLABLE HIEUH YU KIYEOK")
    (15631 . "HANGUL SYLLABLE HIEUH YU NIEUN")
    (15632 . "HANGUL SYLLABLE HIEUH YU RIEUL")
    (15633 . "HANGUL SYLLABLE HIEUH YU MIEUM")
    (15634 . "HANGUL SYLLABLE HIEUH YU SIOS")
    (15635 . "HANGUL SYLLABLE HIEUH YU IEUNG")
    (15636 . "HANGUL SYLLABLE HIEUH EU")
    (15637 . "HANGUL SYLLABLE HIEUH EU KIYEOK")
    (15638 . "HANGUL SYLLABLE HIEUH EU NIEUN")
    (15639 . "HANGUL SYLLABLE HIEUH EU NIEUN-HIEUH")
    (15640 . "HANGUL SYLLABLE HIEUH EU TIKEUT")
    (15641 . "HANGUL SYLLABLE HIEUH EU RIEUL")
    (15642 . "HANGUL SYLLABLE HIEUH EU RIEUL-KIYEOK")
    (15643 . "HANGUL SYLLABLE HIEUH EU MIEUM")
    (15644 . "HANGUL SYLLABLE HIEUH EU PIEUP")
    (15645 . "HANGUL SYLLABLE HIEUH EU SIOS")
    (15646 . "HANGUL SYLLABLE HIEUH EU IEUNG")
    (15647 . "HANGUL SYLLABLE HIEUH EU THIEUTH")
    (15648 . "HANGUL SYLLABLE HIEUH YI")
    (15649 . "HANGUL SYLLABLE HIEUH YI NIEUN")
    (15650 . "HANGUL SYLLABLE HIEUH YI RIEUL")
    (15651 . "HANGUL SYLLABLE HIEUH YI MIEUM")
    (15652 . "HANGUL SYLLABLE HIEUH YI PIEUP")
    (15653 . "HANGUL SYLLABLE HIEUH YI IEUNG")
    (15654 . "HANGUL SYLLABLE HIEUH I")
    (15655 . "HANGUL SYLLABLE HIEUH I KIYEOK")
    (15656 . "HANGUL SYLLABLE HIEUH I NIEUN")
    (15657 . "HANGUL SYLLABLE HIEUH I RIEUL")
    (15658 . "HANGUL SYLLABLE HIEUH I MIEUM")
    (15659 . "HANGUL SYLLABLE HIEUH I PIEUP")
    (15660 . "HANGUL SYLLABLE HIEUH I SIOS")
    (15661 . "HANGUL SYLLABLE HIEUH I IEUNG")
    (15662 . "HANGUL SYLLABLE KIYEOK A SSANGKIYEOK")
    (15663 . "HANGUL SYLLABLE KIYEOK A RIEUL-PIEUP")
    (15664 . "HANGUL SYLLABLE KIYEOK A RIEUL-SIOS")
    (15665 . "HANGUL SYLLABLE KIYEOK AE NIEUN-HIEUH")
    (15666 . "HANGUL SYLLABLE KIYEOK AE TIKEUT")
    (15667 . "HANGUL SYLLABLE KIYEOK AE CIEUC")
    (15668 . "HANGUL SYLLABLE KIYEOK AE THIEUTH")
    (15669 . "HANGUL SYLLABLE KIYEOK AE PHIEUPH")
    (15670 . "HANGUL SYLLABLE KIYEOK YA TIKEUT")
    (15671 . "HANGUL SYLLABLE KIYEOK YA RIEUL-KIYEOK")
    (15672 . "HANGUL SYLLABLE KIYEOK YA RIEUL-HIEUH")
    (15673 . "HANGUL SYLLABLE KIYEOK YA MIEUM")
    (15674 . "HANGUL SYLLABLE KIYEOK YA PIEUP")
    (15675 . "HANGUL SYLLABLE KIYEOK YA SSANGSIOS")
    (15676 . "HANGUL SYLLABLE KIYEOK YAE MIEUM")
    (15677 . "HANGUL SYLLABLE KIYEOK YAE PIEUP")
    (15678 . "HANGUL SYLLABLE KIYEOK EO KIYEOK-SIOS")
    (15679 . "HANGUL SYLLABLE KIYEOK EO NIEUN-CIEUC")
    (15680 . "HANGUL SYLLABLE KIYEOK EO RIEUL-KIYEOK")
    (15681 . "HANGUL SYLLABLE KIYEOK EO CHIEUCH")
    (15682 . "HANGUL SYLLABLE KIYEOK E KIYEOK")
    (15683 . "HANGUL SYLLABLE KIYEOK E TIKEUT")
    (15684 . "HANGUL SYLLABLE KIYEOK E THIEUTH")
    (15685 . "HANGUL SYLLABLE KIYEOK E HIEUH")
    (15686 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-KIYEOK")
    (15687 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-SIOS")
    (15688 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-HIEUH")
    (15689 . "HANGUL SYLLABLE KIYEOK YEO CIEUC")
    (15690 . "HANGUL SYLLABLE KIYEOK YEO CHIEUCH")
    (15691 . "HANGUL SYLLABLE KIYEOK YEO PHIEUPH")
    (15692 . "HANGUL SYLLABLE KIYEOK YE KIYEOK")
    (15693 . "HANGUL SYLLABLE KIYEOK YE TIKEUT")
    (15694 . "HANGUL SYLLABLE KIYEOK YE MIEUM")
    (15695 . "HANGUL SYLLABLE KIYEOK YE SSANGSIOS")
    (15696 . "HANGUL SYLLABLE KIYEOK YE IEUNG")
    (15697 . "HANGUL SYLLABLE KIYEOK O RIEUL-KIYEOK")
    (15698 . "HANGUL SYLLABLE KIYEOK O RIEUL-PIEUP")
    (15699 . "HANGUL SYLLABLE KIYEOK O RIEUL-PHIEUPH")
    (15700 . "HANGUL SYLLABLE KIYEOK O CHIEUCH")
    (15701 . "HANGUL SYLLABLE KIYEOK O THIEUTH")
    (15702 . "HANGUL SYLLABLE KIYEOK O PHIEUPH")
    (15703 . "HANGUL SYLLABLE KIYEOK O HIEUH")
    (15704 . "HANGUL SYLLABLE KIYEOK WA TIKEUT")
    (15705 . "HANGUL SYLLABLE KIYEOK WA SSANGSIOS")
    (15706 . "HANGUL SYLLABLE KIYEOK WAE KIYEOK")
    (15707 . "HANGUL SYLLABLE KIYEOK WAE TIKEUT")
    (15708 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-MIEUM")
    (15709 . "HANGUL SYLLABLE KIYEOK WAE MIEUM")
    (15710 . "HANGUL SYLLABLE KIYEOK WAE SIOS")
    (15711 . "HANGUL SYLLABLE KIYEOK OE TIKEUT")
    (15712 . "HANGUL SYLLABLE KIYEOK OE RIEUL-MIEUM")
    (15713 . "HANGUL SYLLABLE KIYEOK OE SSANGSIOS")
    (15714 . "HANGUL SYLLABLE KIYEOK YO KIYEOK")
    (15715 . "HANGUL SYLLABLE KIYEOK YO TIKEUT")
    (15716 . "HANGUL SYLLABLE KIYEOK YO MIEUM")
    (15717 . "HANGUL SYLLABLE KIYEOK YO IEUNG")
    (15718 . "HANGUL SYLLABLE KIYEOK U RIEUL-SIOS")
    (15719 . "HANGUL SYLLABLE KIYEOK U CHIEUCH")
    (15720 . "HANGUL SYLLABLE KIYEOK U PHIEUPH")
    (15721 . "HANGUL SYLLABLE KIYEOK WEO TIKEUT")
    (15722 . "HANGUL SYLLABLE KIYEOK WEO PIEUP")
    (15723 . "HANGUL SYLLABLE KIYEOK WEO SIOS")
    (15724 . "HANGUL SYLLABLE KIYEOK WE KIYEOK")
    (15725 . "HANGUL SYLLABLE KIYEOK WE NIEUN")
    (15726 . "HANGUL SYLLABLE KIYEOK WE TIKEUT")
    (15727 . "HANGUL SYLLABLE KIYEOK WE RIEUL")
    (15728 . "HANGUL SYLLABLE KIYEOK WE MIEUM")
    (15729 . "HANGUL SYLLABLE KIYEOK WE PIEUP")
    (15730 . "HANGUL SYLLABLE KIYEOK WE SSANGSIOS")
    (15731 . "HANGUL SYLLABLE KIYEOK WE IEUNG")
    (15732 . "HANGUL SYLLABLE KIYEOK WI TIKEUT")
    (15733 . "HANGUL SYLLABLE KIYEOK WI RIEUL-MIEUM")
    (15734 . "HANGUL SYLLABLE KIYEOK WI SSANGSIOS")
    (15735 . "HANGUL SYLLABLE KIYEOK WI IEUNG")
    (15736 . "HANGUL SYLLABLE KIYEOK YU KIYEOK")
    (15737 . "HANGUL SYLLABLE KIYEOK YU RIEUL-SIOS")
    (15738 . "HANGUL SYLLABLE KIYEOK YU MIEUM")
    (15739 . "HANGUL SYLLABLE KIYEOK YU PIEUP")
    (15740 . "HANGUL SYLLABLE KIYEOK YU SIOS")
    (15741 . "HANGUL SYLLABLE KIYEOK YU IEUNG")
    (15742 . "HANGUL SYLLABLE KIYEOK EU NIEUN-HIEUH")
    (15743 . "HANGUL SYLLABLE KIYEOK EU RIEUL-MIEUM")
    (15744 . "HANGUL SYLLABLE KIYEOK EU RIEUL-PIEUP")
    (15745 . "HANGUL SYLLABLE KIYEOK EU RIEUL-SIOS")
    (15746 . "HANGUL SYLLABLE KIYEOK EU RIEUL-HIEUH")
    (15747 . "HANGUL SYLLABLE KIYEOK EU CIEUC")
    (15748 . "HANGUL SYLLABLE KIYEOK EU CHIEUCH")
    (15749 . "HANGUL SYLLABLE KIYEOK EU THIEUTH")
    (15750 . "HANGUL SYLLABLE KIYEOK EU PHIEUPH")
    (15751 . "HANGUL SYLLABLE KIYEOK EU HIEUH")
    (15752 . "HANGUL SYLLABLE KIYEOK YI KIYEOK")
    (15753 . "HANGUL SYLLABLE KIYEOK YI NIEUN")
    (15754 . "HANGUL SYLLABLE KIYEOK YI RIEUL")
    (15755 . "HANGUL SYLLABLE KIYEOK YI RIEUL-SIOS")
    (15756 . "HANGUL SYLLABLE KIYEOK YI MIEUM")
    (15757 . "HANGUL SYLLABLE KIYEOK YI PIEUP")
    (15758 . "HANGUL SYLLABLE KIYEOK YI SIOS")
    (15759 . "HANGUL SYLLABLE KIYEOK YI IEUNG")
    (15760 . "HANGUL SYLLABLE KIYEOK I RIEUL-KIYEOK")
    (15761 . "HANGUL SYLLABLE KIYEOK I RIEUL-SIOS")
    (15762 . "HANGUL SYLLABLE KIYEOK I SSANGSIOS")
    (15763 . "HANGUL SYLLABLE KIYEOK I CHIEUCH")
    (15764 . "HANGUL SYLLABLE KIYEOK I THIEUTH")
    (15765 . "HANGUL SYLLABLE KIYEOK I HIEUH")
    (15766 . "HANGUL SYLLABLE SSANGKIYEOK A TIKEUT")
    (15767 . "HANGUL SYLLABLE SSANGKIYEOK A CIEUC")
    (15768 . "HANGUL SYLLABLE SSANGKIYEOK A CHIEUCH")
    (15769 . "HANGUL SYLLABLE SSANGKIYEOK AE SSANGKIYEOK")
    (15770 . "HANGUL SYLLABLE SSANGKIYEOK AE TIKEUT")
    (15771 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-PIEUP")
    (15772 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-PHIEUPH")
    (15773 . "HANGUL SYLLABLE SSANGKIYEOK AE THIEUTH")
    (15774 . "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN")
    (15775 . "HANGUL SYLLABLE SSANGKIYEOK YA TIKEUT")
    (15776 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-KIYEOK")
    (15777 . "HANGUL SYLLABLE SSANGKIYEOK YA SIOS")
    (15778 . "HANGUL SYLLABLE SSANGKIYEOK YA IEUNG")
    (15779 . "HANGUL SYLLABLE SSANGKIYEOK YAE")
    (15780 . "HANGUL SYLLABLE SSANGKIYEOK YAE IEUNG")
    (15781 . "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN-HIEUH")
    (15782 . "HANGUL SYLLABLE SSANGKIYEOK EO TIKEUT")
    (15783 . "HANGUL SYLLABLE SSANGKIYEOK EO CIEUC")
    (15784 . "HANGUL SYLLABLE SSANGKIYEOK EO CHIEUCH")
    (15785 . "HANGUL SYLLABLE SSANGKIYEOK EO THIEUTH")
    (15786 . "HANGUL SYLLABLE SSANGKIYEOK E SSANGKIYEOK")
    (15787 . "HANGUL SYLLABLE SSANGKIYEOK E TIKEUT")
    (15788 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL")
    (15789 . "HANGUL SYLLABLE SSANGKIYEOK E PIEUP")
    (15790 . "HANGUL SYLLABLE SSANGKIYEOK E SSANGSIOS")
    (15791 . "HANGUL SYLLABLE SSANGKIYEOK E THIEUTH")
    (15792 . "HANGUL SYLLABLE SSANGKIYEOK YEO KIYEOK")
    (15793 . "HANGUL SYLLABLE SSANGKIYEOK YEO TIKEUT")
    (15794 . "HANGUL SYLLABLE SSANGKIYEOK YEO MIEUM")
    (15795 . "HANGUL SYLLABLE SSANGKIYEOK YEO PIEUP")
    (15796 . "HANGUL SYLLABLE SSANGKIYEOK YEO IEUNG")
    (15797 . "HANGUL SYLLABLE SSANGKIYEOK YE TIKEUT")
    (15798 . "HANGUL SYLLABLE SSANGKIYEOK YE IEUNG")
    (15799 . "HANGUL SYLLABLE SSANGKIYEOK O TIKEUT")
    (15800 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-MIEUM")
    (15801 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-SIOS")
    (15802 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-HIEUH")
    (15803 . "HANGUL SYLLABLE SSANGKIYEOK O THIEUTH")
    (15804 . "HANGUL SYLLABLE SSANGKIYEOK O HIEUH")
    (15805 . "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN")
    (15806 . "HANGUL SYLLABLE SSANGKIYEOK WA TIKEUT")
    (15807 . "HANGUL SYLLABLE SSANGKIYEOK WA MIEUM")
    (15808 . "HANGUL SYLLABLE SSANGKIYEOK WA PIEUP")
    (15809 . "HANGUL SYLLABLE SSANGKIYEOK WA SIOS")
    (15810 . "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN")
    (15811 . "HANGUL SYLLABLE SSANGKIYEOK WAE TIKEUT")
    (15812 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL")
    (15813 . "HANGUL SYLLABLE SSANGKIYEOK WAE MIEUM")
    (15814 . "HANGUL SYLLABLE SSANGKIYEOK WAE PIEUP")
    (15815 . "HANGUL SYLLABLE SSANGKIYEOK WAE SSANGSIOS")
    (15816 . "HANGUL SYLLABLE SSANGKIYEOK OE KIYEOK")
    (15817 . "HANGUL SYLLABLE SSANGKIYEOK OE TIKEUT")
    (15818 . "HANGUL SYLLABLE SSANGKIYEOK OE SIOS")
    (15819 . "HANGUL SYLLABLE SSANGKIYEOK OE SSANGSIOS")
    (15820 . "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN")
    (15821 . "HANGUL SYLLABLE SSANGKIYEOK YO TIKEUT")
    (15822 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL")
    (15823 . "HANGUL SYLLABLE SSANGKIYEOK YO MIEUM")
    (15824 . "HANGUL SYLLABLE SSANGKIYEOK YO PIEUP")
    (15825 . "HANGUL SYLLABLE SSANGKIYEOK YO CIEUC")
    (15826 . "HANGUL SYLLABLE SSANGKIYEOK U TIKEUT")
    (15827 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-KIYEOK")
    (15828 . "HANGUL SYLLABLE SSANGKIYEOK U CHIEUCH")
    (15829 . "HANGUL SYLLABLE SSANGKIYEOK WEO KIYEOK")
    (15830 . "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN")
    (15831 . "HANGUL SYLLABLE SSANGKIYEOK WEO TIKEUT")
    (15832 . "HANGUL SYLLABLE SSANGKIYEOK WEO MIEUM")
    (15833 . "HANGUL SYLLABLE SSANGKIYEOK WEO SIOS")
    (15834 . "HANGUL SYLLABLE SSANGKIYEOK WE TIKEUT")
    (15835 . "HANGUL SYLLABLE SSANGKIYEOK WE SIOS")
    (15836 . "HANGUL SYLLABLE SSANGKIYEOK WE IEUNG")
    (15837 . "HANGUL SYLLABLE SSANGKIYEOK WI KIYEOK")
    (15838 . "HANGUL SYLLABLE SSANGKIYEOK WI TIKEUT")
    (15839 . "HANGUL SYLLABLE SSANGKIYEOK WI SIOS")
    (15840 . "HANGUL SYLLABLE SSANGKIYEOK WI IEUNG")
    (15841 . "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN")
    (15842 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL")
    (15843 . "HANGUL SYLLABLE SSANGKIYEOK YU MIEUM")
    (15844 . "HANGUL SYLLABLE SSANGKIYEOK YU PIEUP")
    (15845 . "HANGUL SYLLABLE SSANGKIYEOK EU TIKEUT")
    (15846 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-KIYEOK")
    (15847 . "HANGUL SYLLABLE SSANGKIYEOK EU CHIEUCH")
    (15848 . "HANGUL SYLLABLE SSANGKIYEOK EU HIEUH")
    (15849 . "HANGUL SYLLABLE SSANGKIYEOK YI")
    (15850 . "HANGUL SYLLABLE SSANGKIYEOK YI KIYEOK")
    (15851 . "HANGUL SYLLABLE SSANGKIYEOK YI TIKEUT")
    (15852 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL")
    (15853 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-PIEUP")
    (15854 . "HANGUL SYLLABLE SSANGKIYEOK YI SIOS")
    (15855 . "HANGUL SYLLABLE SSANGKIYEOK YI IEUNG")
    (15856 . "HANGUL SYLLABLE SSANGKIYEOK I TIKEUT")
    (15857 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-HIEUH")
    (15858 . "HANGUL SYLLABLE SSANGKIYEOK I SSANGSIOS")
    (15859 . "HANGUL SYLLABLE SSANGKIYEOK I THIEUTH")
    (15860 . "HANGUL SYLLABLE NIEUN A KIYEOK-SIOS")
    (15861 . "HANGUL SYLLABLE NIEUN A RIEUL-SIOS")
    (15862 . "HANGUL SYLLABLE NIEUN A PIEUP-SIOS")
    (15863 . "HANGUL SYLLABLE NIEUN A KHIEUKH")
    (15864 . "HANGUL SYLLABLE NIEUN A PHIEUPH")
    (15865 . "HANGUL SYLLABLE NIEUN AE TIKEUT")
    (15866 . "HANGUL SYLLABLE NIEUN AE PHIEUPH")
    (15867 . "HANGUL SYLLABLE NIEUN YA TIKEUT")
    (15868 . "HANGUL SYLLABLE NIEUN YA PIEUP")
    (15869 . "HANGUL SYLLABLE NIEUN YA SIOS")
    (15870 . "HANGUL SYLLABLE NIEUN YA CIEUC")
    (15871 . "HANGUL SYLLABLE NIEUN YA THIEUTH")
    (15872 . "HANGUL SYLLABLE NIEUN YAE")
    (15873 . "HANGUL SYLLABLE NIEUN YAE IEUNG")
    (15874 . "HANGUL SYLLABLE NIEUN EO SSANGKIYEOK")
    (15875 . "HANGUL SYLLABLE NIEUN EO TIKEUT")
    (15876 . "HANGUL SYLLABLE NIEUN EO RIEUL-HIEUH")
    (15877 . "HANGUL SYLLABLE NIEUN EO CIEUC")
    (15878 . "HANGUL SYLLABLE NIEUN EO KHIEUKH")
    (15879 . "HANGUL SYLLABLE NIEUN EO THIEUTH")
    (15880 . "HANGUL SYLLABLE NIEUN EO PHIEUPH")
    (15881 . "HANGUL SYLLABLE NIEUN E TIKEUT")
    (15882 . "HANGUL SYLLABLE NIEUN E KHIEUKH")
    (15883 . "HANGUL SYLLABLE NIEUN E THIEUTH")
    (15884 . "HANGUL SYLLABLE NIEUN E PHIEUPH")
    (15885 . "HANGUL SYLLABLE NIEUN E HIEUH")
    (15886 . "HANGUL SYLLABLE NIEUN YEO KIYEOK-SIOS")
    (15887 . "HANGUL SYLLABLE NIEUN YEO TIKEUT")
    (15888 . "HANGUL SYLLABLE NIEUN YEO RIEUL-PIEUP")
    (15889 . "HANGUL SYLLABLE NIEUN YEO SIOS")
    (15890 . "HANGUL SYLLABLE NIEUN YEO THIEUTH")
    (15891 . "HANGUL SYLLABLE NIEUN YEO PHIEUPH")
    (15892 . "HANGUL SYLLABLE NIEUN YEO HIEUH")
    (15893 . "HANGUL SYLLABLE NIEUN YE KIYEOK")
    (15894 . "HANGUL SYLLABLE NIEUN YE TIKEUT")
    (15895 . "HANGUL SYLLABLE NIEUN YE RIEUL")
    (15896 . "HANGUL SYLLABLE NIEUN YE MIEUM")
    (15897 . "HANGUL SYLLABLE NIEUN YE PIEUP")
    (15898 . "HANGUL SYLLABLE NIEUN YE SIOS")
    (15899 . "HANGUL SYLLABLE NIEUN YE IEUNG")
    (15900 . "HANGUL SYLLABLE NIEUN O SSANGKIYEOK")
    (15901 . "HANGUL SYLLABLE NIEUN O KIYEOK-SIOS")
    (15902 . "HANGUL SYLLABLE NIEUN O NIEUN-HIEUH")
    (15903 . "HANGUL SYLLABLE NIEUN O TIKEUT")
    (15904 . "HANGUL SYLLABLE NIEUN O CIEUC")
    (15905 . "HANGUL SYLLABLE NIEUN O KHIEUKH")
    (15906 . "HANGUL SYLLABLE NIEUN O THIEUTH")
    (15907 . "HANGUL SYLLABLE NIEUN WA KIYEOK")
    (15908 . "HANGUL SYLLABLE NIEUN WA TIKEUT")
    (15909 . "HANGUL SYLLABLE NIEUN WA MIEUM")
    (15910 . "HANGUL SYLLABLE NIEUN WA PIEUP")
    (15911 . "HANGUL SYLLABLE NIEUN WA SIOS")
    (15912 . "HANGUL SYLLABLE NIEUN WA IEUNG")
    (15913 . "HANGUL SYLLABLE NIEUN WAE")
    (15914 . "HANGUL SYLLABLE NIEUN WAE NIEUN")
    (15915 . "HANGUL SYLLABLE NIEUN WAE RIEUL")
    (15916 . "HANGUL SYLLABLE NIEUN WAE SSANGSIOS")
    (15917 . "HANGUL SYLLABLE NIEUN OE KIYEOK")
    (15918 . "HANGUL SYLLABLE NIEUN OE TIKEUT")
    (15919 . "HANGUL SYLLABLE NIEUN OE SSANGSIOS")
    (15920 . "HANGUL SYLLABLE NIEUN OE IEUNG")
    (15921 . "HANGUL SYLLABLE NIEUN OE PHIEUPH")
    (15922 . "HANGUL SYLLABLE NIEUN YO TIKEUT")
    (15923 . "HANGUL SYLLABLE NIEUN YO MIEUM")
    (15924 . "HANGUL SYLLABLE NIEUN U RIEUL-KIYEOK")
    (15925 . "HANGUL SYLLABLE NIEUN U RIEUL-SIOS")
    (15926 . "HANGUL SYLLABLE NIEUN U RIEUL-HIEUH")
    (15927 . "HANGUL SYLLABLE NIEUN U CHIEUCH")
    (15928 . "HANGUL SYLLABLE NIEUN U PHIEUPH")
    (15929 . "HANGUL SYLLABLE NIEUN WEO KIYEOK")
    (15930 . "HANGUL SYLLABLE NIEUN WEO NIEUN")
    (15931 . "HANGUL SYLLABLE NIEUN WEO TIKEUT")
    (15932 . "HANGUL SYLLABLE NIEUN WEO RIEUL")
    (15933 . "HANGUL SYLLABLE NIEUN WEO MIEUM")
    (15934 . "HANGUL SYLLABLE NIEUN WEO PIEUP")
    (15935 . "HANGUL SYLLABLE NIEUN WEO SIOS")
    (15936 . "HANGUL SYLLABLE NIEUN WEO IEUNG")
    (15937 . "HANGUL SYLLABLE NIEUN WE KIYEOK")
    (15938 . "HANGUL SYLLABLE NIEUN WE NIEUN")
    (15939 . "HANGUL SYLLABLE NIEUN WE TIKEUT")
    (15940 . "HANGUL SYLLABLE NIEUN WE RIEUL")
    (15941 . "HANGUL SYLLABLE NIEUN WE MIEUM")
    (15942 . "HANGUL SYLLABLE NIEUN WE PIEUP")
    (15943 . "HANGUL SYLLABLE NIEUN WE SIOS")
    (15944 . "HANGUL SYLLABLE NIEUN WE SSANGSIOS")
    (15945 . "HANGUL SYLLABLE NIEUN WE IEUNG")
    (15946 . "HANGUL SYLLABLE NIEUN WI KIYEOK")
    (15947 . "HANGUL SYLLABLE NIEUN WI TIKEUT")
    (15948 . "HANGUL SYLLABLE NIEUN WI SIOS")
    (15949 . "HANGUL SYLLABLE NIEUN WI IEUNG")
    (15950 . "HANGUL SYLLABLE NIEUN WI CHIEUCH")
    (15951 . "HANGUL SYLLABLE NIEUN YU NIEUN")
    (15952 . "HANGUL SYLLABLE NIEUN YU TIKEUT")
    (15953 . "HANGUL SYLLABLE NIEUN YU SIOS")
    (15954 . "HANGUL SYLLABLE NIEUN YU CIEUC")
    (15955 . "HANGUL SYLLABLE NIEUN YU CHIEUCH")
    (15956 . "HANGUL SYLLABLE NIEUN EU TIKEUT")
    (15957 . "HANGUL SYLLABLE NIEUN EU RIEUL-SIOS")
    (15958 . "HANGUL SYLLABLE NIEUN EU CHIEUCH")
    (15959 . "HANGUL SYLLABLE NIEUN EU HIEUH")
    (15960 . "HANGUL SYLLABLE NIEUN YI KIYEOK")
    (15961 . "HANGUL SYLLABLE NIEUN YI TIKEUT")
    (15962 . "HANGUL SYLLABLE NIEUN YI MIEUM")
    (15963 . "HANGUL SYLLABLE NIEUN YI PIEUP")
    (15964 . "HANGUL SYLLABLE NIEUN YI SIOS")
    (15965 . "HANGUL SYLLABLE NIEUN YI SSANGSIOS")
    (15966 . "HANGUL SYLLABLE NIEUN YI IEUNG")
    (15967 . "HANGUL SYLLABLE NIEUN I TIKEUT")
    (15968 . "HANGUL SYLLABLE NIEUN I RIEUL-KIYEOK")
    (15969 . "HANGUL SYLLABLE NIEUN I RIEUL-PIEUP")
    (15970 . "HANGUL SYLLABLE NIEUN I RIEUL-SIOS")
    (15971 . "HANGUL SYLLABLE NIEUN I CIEUC")
    (15972 . "HANGUL SYLLABLE NIEUN I KHIEUKH")
    (15973 . "HANGUL SYLLABLE NIEUN I HIEUH")
    (15974 . "HANGUL SYLLABLE TIKEUT A RIEUL-SIOS")
    (15975 . "HANGUL SYLLABLE TIKEUT A RIEUL-PHIEUPH")
    (15976 . "HANGUL SYLLABLE TIKEUT A PIEUP-SIOS")
    (15977 . "HANGUL SYLLABLE TIKEUT A KHIEUKH")
    (15978 . "HANGUL SYLLABLE TIKEUT A THIEUTH")
    (15979 . "HANGUL SYLLABLE TIKEUT AE TIKEUT")
    (15980 . "HANGUL SYLLABLE TIKEUT AE THIEUTH")
    (15981 . "HANGUL SYLLABLE TIKEUT YA KIYEOK")
    (15982 . "HANGUL SYLLABLE TIKEUT YA NIEUN")
    (15983 . "HANGUL SYLLABLE TIKEUT YA RIEUL")
    (15984 . "HANGUL SYLLABLE TIKEUT YA RIEUL-PIEUP")
    (15985 . "HANGUL SYLLABLE TIKEUT YA MIEUM")
    (15986 . "HANGUL SYLLABLE TIKEUT YA PIEUP")
    (15987 . "HANGUL SYLLABLE TIKEUT YA SIOS")
    (15988 . "HANGUL SYLLABLE TIKEUT YA IEUNG")
    (15989 . "HANGUL SYLLABLE TIKEUT YAE")
    (15990 . "HANGUL SYLLABLE TIKEUT YAE NIEUN")
    (15991 . "HANGUL SYLLABLE TIKEUT YAE IEUNG")
    (15992 . "HANGUL SYLLABLE TIKEUT EO NIEUN-CIEUC")
    (15993 . "HANGUL SYLLABLE TIKEUT EO PIEUP-SIOS")
    (15994 . "HANGUL SYLLABLE TIKEUT EO SSANGSIOS")
    (15995 . "HANGUL SYLLABLE TIKEUT EO CIEUC")
    (15996 . "HANGUL SYLLABLE TIKEUT EO THIEUTH")
    (15997 . "HANGUL SYLLABLE TIKEUT EO HIEUH")
    (15998 . "HANGUL SYLLABLE TIKEUT E TIKEUT")
    (15999 . "HANGUL SYLLABLE TIKEUT E CIEUC")
    (16000 . "HANGUL SYLLABLE TIKEUT E PHIEUPH")
    (16001 . "HANGUL SYLLABLE TIKEUT YEO KIYEOK")
    (16002 . "HANGUL SYLLABLE TIKEUT YEO KIYEOK-SIOS")
    (16003 . "HANGUL SYLLABLE TIKEUT YEO TIKEUT")
    (16004 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-SIOS")
    (16005 . "HANGUL SYLLABLE TIKEUT YEO MIEUM")
    (16006 . "HANGUL SYLLABLE TIKEUT YEO PIEUP")
    (16007 . "HANGUL SYLLABLE TIKEUT YEO SIOS")
    (16008 . "HANGUL SYLLABLE TIKEUT YE KIYEOK")
    (16009 . "HANGUL SYLLABLE TIKEUT YE RIEUL")
    (16010 . "HANGUL SYLLABLE TIKEUT YE MIEUM")
    (16011 . "HANGUL SYLLABLE TIKEUT YE PIEUP")
    (16012 . "HANGUL SYLLABLE TIKEUT YE SIOS")
    (16013 . "HANGUL SYLLABLE TIKEUT YE IEUNG")
    (16014 . "HANGUL SYLLABLE TIKEUT O SSANGKIYEOK")
    (16015 . "HANGUL SYLLABLE TIKEUT O KIYEOK-SIOS")
    (16016 . "HANGUL SYLLABLE TIKEUT O RIEUL-PIEUP")
    (16017 . "HANGUL SYLLABLE TIKEUT O RIEUL-HIEUH")
    (16018 . "HANGUL SYLLABLE TIKEUT O CIEUC")
    (16019 . "HANGUL SYLLABLE TIKEUT O KHIEUKH")
    (16020 . "HANGUL SYLLABLE TIKEUT O PHIEUPH")
    (16021 . "HANGUL SYLLABLE TIKEUT O HIEUH")
    (16022 . "HANGUL SYLLABLE TIKEUT WA KIYEOK")
    (16023 . "HANGUL SYLLABLE TIKEUT WA MIEUM")
    (16024 . "HANGUL SYLLABLE TIKEUT WA PIEUP")
    (16025 . "HANGUL SYLLABLE TIKEUT WA SIOS")
    (16026 . "HANGUL SYLLABLE TIKEUT WA SSANGSIOS")
    (16027 . "HANGUL SYLLABLE TIKEUT WA IEUNG")
    (16028 . "HANGUL SYLLABLE TIKEUT WAE NIEUN")
    (16029 . "HANGUL SYLLABLE TIKEUT WAE TIKEUT")
    (16030 . "HANGUL SYLLABLE TIKEUT WAE RIEUL")
    (16031 . "HANGUL SYLLABLE TIKEUT WAE MIEUM")
    (16032 . "HANGUL SYLLABLE TIKEUT WAE PIEUP")
    (16033 . "HANGUL SYLLABLE TIKEUT WAE SIOS")
    (16034 . "HANGUL SYLLABLE TIKEUT OE KIYEOK")
    (16035 . "HANGUL SYLLABLE TIKEUT OE TIKEUT")
    (16036 . "HANGUL SYLLABLE TIKEUT OE RIEUL-SIOS")
    (16037 . "HANGUL SYLLABLE TIKEUT OE SSANGSIOS")
    (16038 . "HANGUL SYLLABLE TIKEUT OE IEUNG")
    (16039 . "HANGUL SYLLABLE TIKEUT YO KIYEOK")
    (16040 . "HANGUL SYLLABLE TIKEUT YO NIEUN")
    (16041 . "HANGUL SYLLABLE TIKEUT YO TIKEUT")
    (16042 . "HANGUL SYLLABLE TIKEUT YO RIEUL")
    (16043 . "HANGUL SYLLABLE TIKEUT YO MIEUM")
    (16044 . "HANGUL SYLLABLE TIKEUT YO PIEUP")
    (16045 . "HANGUL SYLLABLE TIKEUT YO SIOS")
    (16046 . "HANGUL SYLLABLE TIKEUT YO IEUNG")
    (16047 . "HANGUL SYLLABLE TIKEUT YO HIEUH")
    (16048 . "HANGUL SYLLABLE TIKEUT U TIKEUT")
    (16049 . "HANGUL SYLLABLE TIKEUT U RIEUL-KIYEOK")
    (16050 . "HANGUL SYLLABLE TIKEUT U RIEUL-MIEUM")
    (16051 . "HANGUL SYLLABLE TIKEUT U RIEUL-PIEUP")
    (16052 . "HANGUL SYLLABLE TIKEUT U PHIEUPH")
    (16053 . "HANGUL SYLLABLE TIKEUT WEO KIYEOK")
    (16054 . "HANGUL SYLLABLE TIKEUT WEO NIEUN")
    (16055 . "HANGUL SYLLABLE TIKEUT WEO TIKEUT")
    (16056 . "HANGUL SYLLABLE TIKEUT WEO RIEUL")
    (16057 . "HANGUL SYLLABLE TIKEUT WEO MIEUM")
    (16058 . "HANGUL SYLLABLE TIKEUT WEO PIEUP")
    (16059 . "HANGUL SYLLABLE TIKEUT WEO SIOS")
    (16060 . "HANGUL SYLLABLE TIKEUT WEO IEUNG")
    (16061 . "HANGUL SYLLABLE TIKEUT WE KIYEOK")
    (16062 . "HANGUL SYLLABLE TIKEUT WE NIEUN")
    (16063 . "HANGUL SYLLABLE TIKEUT WE TIKEUT")
    (16064 . "HANGUL SYLLABLE TIKEUT WE RIEUL")
    (16065 . "HANGUL SYLLABLE TIKEUT WE MIEUM")
    (16066 . "HANGUL SYLLABLE TIKEUT WE PIEUP")
    (16067 . "HANGUL SYLLABLE TIKEUT WE SIOS")
    (16068 . "HANGUL SYLLABLE TIKEUT WE SSANGSIOS")
    (16069 . "HANGUL SYLLABLE TIKEUT WI KIYEOK")
    (16070 . "HANGUL SYLLABLE TIKEUT WI TIKEUT")
    (16071 . "HANGUL SYLLABLE TIKEUT WI MIEUM")
    (16072 . "HANGUL SYLLABLE TIKEUT WI SSANGSIOS")
    (16073 . "HANGUL SYLLABLE TIKEUT YU KIYEOK")
    (16074 . "HANGUL SYLLABLE TIKEUT YU PIEUP")
    (16075 . "HANGUL SYLLABLE TIKEUT YU SIOS")
    (16076 . "HANGUL SYLLABLE TIKEUT EU RIEUL-KIYEOK")
    (16077 . "HANGUL SYLLABLE TIKEUT EU RIEUL-PIEUP")
    (16078 . "HANGUL SYLLABLE TIKEUT EU RIEUL-SIOS")
    (16079 . "HANGUL SYLLABLE TIKEUT YI KIYEOK")
    (16080 . "HANGUL SYLLABLE TIKEUT YI NIEUN")
    (16081 . "HANGUL SYLLABLE TIKEUT YI TIKEUT")
    (16082 . "HANGUL SYLLABLE TIKEUT YI RIEUL")
    (16083 . "HANGUL SYLLABLE TIKEUT YI MIEUM")
    (16084 . "HANGUL SYLLABLE TIKEUT YI PIEUP")
    (16085 . "HANGUL SYLLABLE TIKEUT YI SIOS")
    (16086 . "HANGUL SYLLABLE TIKEUT YI IEUNG")
    (16087 . "HANGUL SYLLABLE TIKEUT I CHIEUCH")
    (16088 . "HANGUL SYLLABLE TIKEUT I THIEUTH")
    (16089 . "HANGUL SYLLABLE TIKEUT I PHIEUPH")
    (16090 . "HANGUL SYLLABLE TIKEUT I HIEUH")
    (16091 . "HANGUL SYLLABLE SSANGTIKEUT A SSANGKIYEOK")
    (16092 . "HANGUL SYLLABLE SSANGTIKEUT A TIKEUT")
    (16093 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-KIYEOK")
    (16094 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-MIEUM")
    (16095 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-PIEUP")
    (16096 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-HIEUH")
    (16097 . "HANGUL SYLLABLE SSANGTIKEUT AE SSANGKIYEOK")
    (16098 . "HANGUL SYLLABLE SSANGTIKEUT AE TIKEUT")
    (16099 . "HANGUL SYLLABLE SSANGTIKEUT AE HIEUH")
    (16100 . "HANGUL SYLLABLE SSANGTIKEUT YA")
    (16101 . "HANGUL SYLLABLE SSANGTIKEUT YA KIYEOK")
    (16102 . "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN")
    (16103 . "HANGUL SYLLABLE SSANGTIKEUT YA IEUNG")
    (16104 . "HANGUL SYLLABLE SSANGTIKEUT YAE")
    (16105 . "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN")
    (16106 . "HANGUL SYLLABLE SSANGTIKEUT EO TIKEUT")
    (16107 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-HIEUH")
    (16108 . "HANGUL SYLLABLE SSANGTIKEUT E TIKEUT")
    (16109 . "HANGUL SYLLABLE SSANGTIKEUT YEO KIYEOK")
    (16110 . "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN")
    (16111 . "HANGUL SYLLABLE SSANGTIKEUT YEO TIKEUT")
    (16112 . "HANGUL SYLLABLE SSANGTIKEUT YEO MIEUM")
    (16113 . "HANGUL SYLLABLE SSANGTIKEUT YEO PIEUP")
    (16114 . "HANGUL SYLLABLE SSANGTIKEUT YEO SIOS")
    (16115 . "HANGUL SYLLABLE SSANGTIKEUT YEO IEUNG")
    (16116 . "HANGUL SYLLABLE SSANGTIKEUT YE")
    (16117 . "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN")
    (16118 . "HANGUL SYLLABLE SSANGTIKEUT YE PIEUP")
    (16119 . "HANGUL SYLLABLE SSANGTIKEUT O SSANGKIYEOK")
    (16120 . "HANGUL SYLLABLE SSANGTIKEUT O TIKEUT")
    (16121 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-KIYEOK")
    (16122 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-MIEUM")
    (16123 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-HIEUH")
    (16124 . "HANGUL SYLLABLE SSANGTIKEUT O MIEUM")
    (16125 . "HANGUL SYLLABLE SSANGTIKEUT O PIEUP")
    (16126 . "HANGUL SYLLABLE SSANGTIKEUT O SIOS")
    (16127 . "HANGUL SYLLABLE SSANGTIKEUT WA KIYEOK")
    (16128 . "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN")
    (16129 . "HANGUL SYLLABLE SSANGTIKEUT WA IEUNG")
    (16130 . "HANGUL SYLLABLE SSANGTIKEUT WAE KIYEOK")
    (16131 . "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN")
    (16132 . "HANGUL SYLLABLE SSANGTIKEUT WAE TIKEUT")
    (16133 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL")
    (16134 . "HANGUL SYLLABLE SSANGTIKEUT WAE MIEUM")
    (16135 . "HANGUL SYLLABLE SSANGTIKEUT WAE PIEUP")
    (16136 . "HANGUL SYLLABLE SSANGTIKEUT WAE SIOS")
    (16137 . "HANGUL SYLLABLE SSANGTIKEUT WAE SSANGSIOS")
    (16138 . "HANGUL SYLLABLE SSANGTIKEUT OE KIYEOK")
    (16139 . "HANGUL SYLLABLE SSANGTIKEUT OE TIKEUT")
    (16140 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL")
    (16141 . "HANGUL SYLLABLE SSANGTIKEUT OE MIEUM")
    (16142 . "HANGUL SYLLABLE SSANGTIKEUT OE PIEUP")
    (16143 . "HANGUL SYLLABLE SSANGTIKEUT OE SIOS")
    (16144 . "HANGUL SYLLABLE SSANGTIKEUT OE SSANGSIOS")
    (16145 . "HANGUL SYLLABLE SSANGTIKEUT OE IEUNG")
    (16146 . "HANGUL SYLLABLE SSANGTIKEUT YO")
    (16147 . "HANGUL SYLLABLE SSANGTIKEUT YO KIYEOK")
    (16148 . "HANGUL SYLLABLE SSANGTIKEUT U TIKEUT")
    (16149 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-MIEUM")
    (16150 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-PIEUP")
    (16151 . "HANGUL SYLLABLE SSANGTIKEUT U PIEUP")
    (16152 . "HANGUL SYLLABLE SSANGTIKEUT U SIOS")
    (16153 . "HANGUL SYLLABLE SSANGTIKEUT U THIEUTH")
    (16154 . "HANGUL SYLLABLE SSANGTIKEUT WEO")
    (16155 . "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN")
    (16156 . "HANGUL SYLLABLE SSANGTIKEUT WEO TIKEUT")
    (16157 . "HANGUL SYLLABLE SSANGTIKEUT WEO SIOS")
    (16158 . "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN")
    (16159 . "HANGUL SYLLABLE SSANGTIKEUT WE TIKEUT")
    (16160 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL")
    (16161 . "HANGUL SYLLABLE SSANGTIKEUT WE MIEUM")
    (16162 . "HANGUL SYLLABLE SSANGTIKEUT WE PIEUP")
    (16163 . "HANGUL SYLLABLE SSANGTIKEUT WE SSANGSIOS")
    (16164 . "HANGUL SYLLABLE SSANGTIKEUT WE IEUNG")
    (16165 . "HANGUL SYLLABLE SSANGTIKEUT WI TIKEUT")
    (16166 . "HANGUL SYLLABLE SSANGTIKEUT WI SIOS")
    (16167 . "HANGUL SYLLABLE SSANGTIKEUT YU")
    (16168 . "HANGUL SYLLABLE SSANGTIKEUT YU KIYEOK")
    (16169 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL")
    (16170 . "HANGUL SYLLABLE SSANGTIKEUT YU MIEUM")
    (16171 . "HANGUL SYLLABLE SSANGTIKEUT YU IEUNG")
    (16172 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-MIEUM")
    (16173 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-PIEUP")
    (16174 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-HIEUH")
    (16175 . "HANGUL SYLLABLE SSANGTIKEUT EU IEUNG")
    (16176 . "HANGUL SYLLABLE SSANGTIKEUT YI IEUNG")
    (16177 . "HANGUL SYLLABLE SSANGTIKEUT I KIYEOK")
    (16178 . "HANGUL SYLLABLE SSANGTIKEUT I TIKEUT")
    (16179 . "HANGUL SYLLABLE SSANGTIKEUT I HIEUH")
    (16180 . "HANGUL SYLLABLE RIEUL A TIKEUT")
    (16181 . "HANGUL SYLLABLE RIEUL A RIEUL-SIOS")
    (16182 . "HANGUL SYLLABLE RIEUL A CHIEUCH")
    (16183 . "HANGUL SYLLABLE RIEUL A KHIEUKH")
    (16184 . "HANGUL SYLLABLE RIEUL A THIEUTH")
    (16185 . "HANGUL SYLLABLE RIEUL AE TIKEUT")
    (16186 . "HANGUL SYLLABLE RIEUL AE KHIEUKH")
    (16187 . "HANGUL SYLLABLE RIEUL AE THIEUTH")
    (16188 . "HANGUL SYLLABLE RIEUL AE PHIEUPH")
    (16189 . "HANGUL SYLLABLE RIEUL YA TIKEUT")
    (16190 . "HANGUL SYLLABLE RIEUL YA RIEUL")
    (16191 . "HANGUL SYLLABLE RIEUL YA MIEUM")
    (16192 . "HANGUL SYLLABLE RIEUL YA PIEUP")
    (16193 . "HANGUL SYLLABLE RIEUL YAE")
    (16194 . "HANGUL SYLLABLE RIEUL YAE SIOS")
    (16195 . "HANGUL SYLLABLE RIEUL EO NIEUN-HIEUH")
    (16196 . "HANGUL SYLLABLE RIEUL EO TIKEUT")
    (16197 . "HANGUL SYLLABLE RIEUL EO PIEUP-SIOS")
    (16198 . "HANGUL SYLLABLE RIEUL EO CIEUC")
    (16199 . "HANGUL SYLLABLE RIEUL EO KHIEUKH")
    (16200 . "HANGUL SYLLABLE RIEUL EO PHIEUPH")
    (16201 . "HANGUL SYLLABLE RIEUL E TIKEUT")
    (16202 . "HANGUL SYLLABLE RIEUL E SSANGSIOS")
    (16203 . "HANGUL SYLLABLE RIEUL E THIEUTH")
    (16204 . "HANGUL SYLLABLE RIEUL E PHIEUPH")
    (16205 . "HANGUL SYLLABLE RIEUL E HIEUH")
    (16206 . "HANGUL SYLLABLE RIEUL YEO TIKEUT")
    (16207 . "HANGUL SYLLABLE RIEUL YEO RIEUL-SIOS")
    (16208 . "HANGUL SYLLABLE RIEUL YEO KHIEUKH")
    (16209 . "HANGUL SYLLABLE RIEUL YEO PHIEUPH")
    (16210 . "HANGUL SYLLABLE RIEUL YE KIYEOK")
    (16211 . "HANGUL SYLLABLE RIEUL YE TIKEUT")
    (16212 . "HANGUL SYLLABLE RIEUL YE RIEUL")
    (16213 . "HANGUL SYLLABLE RIEUL YE MIEUM")
    (16214 . "HANGUL SYLLABLE RIEUL YE IEUNG")
    (16215 . "HANGUL SYLLABLE RIEUL O TIKEUT")
    (16216 . "HANGUL SYLLABLE RIEUL O RIEUL-SIOS")
    (16217 . "HANGUL SYLLABLE RIEUL O RIEUL-HIEUH")
    (16218 . "HANGUL SYLLABLE RIEUL O KHIEUKH")
    (16219 . "HANGUL SYLLABLE RIEUL O PHIEUPH")
    (16220 . "HANGUL SYLLABLE RIEUL O HIEUH")
    (16221 . "HANGUL SYLLABLE RIEUL WA KIYEOK")
    (16222 . "HANGUL SYLLABLE RIEUL WA RIEUL")
    (16223 . "HANGUL SYLLABLE RIEUL WA MIEUM")
    (16224 . "HANGUL SYLLABLE RIEUL WA PIEUP")
    (16225 . "HANGUL SYLLABLE RIEUL WA SIOS")
    (16226 . "HANGUL SYLLABLE RIEUL WA SSANGSIOS")
    (16227 . "HANGUL SYLLABLE RIEUL WAE")
    (16228 . "HANGUL SYLLABLE RIEUL WAE NIEUN")
    (16229 . "HANGUL SYLLABLE RIEUL WAE TIKEUT")
    (16230 . "HANGUL SYLLABLE RIEUL WAE RIEUL")
    (16231 . "HANGUL SYLLABLE RIEUL WAE SIOS")
    (16232 . "HANGUL SYLLABLE RIEUL OE KIYEOK")
    (16233 . "HANGUL SYLLABLE RIEUL OE NIEUN-CIEUC")
    (16234 . "HANGUL SYLLABLE RIEUL OE TIKEUT")
    (16235 . "HANGUL SYLLABLE RIEUL OE SSANGSIOS")
    (16236 . "HANGUL SYLLABLE RIEUL YO KIYEOK")
    (16237 . "HANGUL SYLLABLE RIEUL YO TIKEUT")
    (16238 . "HANGUL SYLLABLE RIEUL YO MIEUM")
    (16239 . "HANGUL SYLLABLE RIEUL U TIKEUT")
    (16240 . "HANGUL SYLLABLE RIEUL U RIEUL-PIEUP")
    (16241 . "HANGUL SYLLABLE RIEUL U KHIEUKH")
    (16242 . "HANGUL SYLLABLE RIEUL U PHIEUPH")
    (16243 . "HANGUL SYLLABLE RIEUL WEO KIYEOK")
    (16244 . "HANGUL SYLLABLE RIEUL WEO NIEUN")
    (16245 . "HANGUL SYLLABLE RIEUL WEO TIKEUT")
    (16246 . "HANGUL SYLLABLE RIEUL WEO RIEUL")
    (16247 . "HANGUL SYLLABLE RIEUL WEO MIEUM")
    (16248 . "HANGUL SYLLABLE RIEUL WEO PIEUP")
    (16249 . "HANGUL SYLLABLE RIEUL WEO SIOS")
    (16250 . "HANGUL SYLLABLE RIEUL WEO IEUNG")
    (16251 . "HANGUL SYLLABLE RIEUL WE KIYEOK")
    (16252 . "HANGUL SYLLABLE RIEUL WE NIEUN")
    (16253 . "HANGUL SYLLABLE RIEUL WE RIEUL")
    (16254 . "HANGUL SYLLABLE RIEUL WE MIEUM")
    (16255 . "HANGUL SYLLABLE RIEUL WE PIEUP")
    (16256 . "HANGUL SYLLABLE RIEUL WE SIOS")
    (16257 . "HANGUL SYLLABLE RIEUL WE SSANGSIOS")
    (16258 . "HANGUL SYLLABLE RIEUL WE IEUNG")
    (16259 . "HANGUL SYLLABLE RIEUL WI TIKEUT")
    (16260 . "HANGUL SYLLABLE RIEUL WI PIEUP")
    (16261 . "HANGUL SYLLABLE RIEUL WI SSANGSIOS")
    (16262 . "HANGUL SYLLABLE RIEUL YU TIKEUT")
    (16263 . "HANGUL SYLLABLE RIEUL EU TIKEUT")
    (16264 . "HANGUL SYLLABLE RIEUL EU RIEUL-KIYEOK")
    (16265 . "HANGUL SYLLABLE RIEUL EU RIEUL-SIOS")
    (16266 . "HANGUL SYLLABLE RIEUL EU CHIEUCH")
    (16267 . "HANGUL SYLLABLE RIEUL EU HIEUH")
    (16268 . "HANGUL SYLLABLE RIEUL YI")
    (16269 . "HANGUL SYLLABLE RIEUL YI KIYEOK")
    (16270 . "HANGUL SYLLABLE RIEUL YI NIEUN")
    (16271 . "HANGUL SYLLABLE RIEUL YI RIEUL")
    (16272 . "HANGUL SYLLABLE RIEUL YI RIEUL-PHIEUPH")
    (16273 . "HANGUL SYLLABLE RIEUL YI MIEUM")
    (16274 . "HANGUL SYLLABLE RIEUL YI PIEUP")
    (16275 . "HANGUL SYLLABLE RIEUL YI SIOS")
    (16276 . "HANGUL SYLLABLE RIEUL YI IEUNG")
    (16277 . "HANGUL SYLLABLE RIEUL I TIKEUT")
    (16278 . "HANGUL SYLLABLE RIEUL I PIEUP-SIOS")
    (16279 . "HANGUL SYLLABLE RIEUL I SSANGSIOS")
    (16280 . "HANGUL SYLLABLE RIEUL I KHIEUKH")
    (16281 . "HANGUL SYLLABLE RIEUL I PHIEUPH")
    (16282 . "HANGUL SYLLABLE MIEUM A SSANGKIYEOK")
    (16283 . "HANGUL SYLLABLE MIEUM A RIEUL-SIOS")
    (16284 . "HANGUL SYLLABLE MIEUM A SSANGSIOS")
    (16285 . "HANGUL SYLLABLE MIEUM A CHIEUCH")
    (16286 . "HANGUL SYLLABLE MIEUM A KHIEUKH")
    (16287 . "HANGUL SYLLABLE MIEUM A PHIEUPH")
    (16288 . "HANGUL SYLLABLE MIEUM AE TIKEUT")
    (16289 . "HANGUL SYLLABLE MIEUM AE RIEUL-KIYEOK")
    (16290 . "HANGUL SYLLABLE MIEUM AE CHIEUCH")
    (16291 . "HANGUL SYLLABLE MIEUM AE THIEUTH")
    (16292 . "HANGUL SYLLABLE MIEUM AE PHIEUPH")
    (16293 . "HANGUL SYLLABLE MIEUM YA NIEUN")
    (16294 . "HANGUL SYLLABLE MIEUM YA MIEUM")
    (16295 . "HANGUL SYLLABLE MIEUM YA PIEUP")
    (16296 . "HANGUL SYLLABLE MIEUM YA SIOS")
    (16297 . "HANGUL SYLLABLE MIEUM YA KHIEUKH")
    (16298 . "HANGUL SYLLABLE MIEUM YAE")
    (16299 . "HANGUL SYLLABLE MIEUM EO KIYEOK-SIOS")
    (16300 . "HANGUL SYLLABLE MIEUM EO TIKEUT")
    (16301 . "HANGUL SYLLABLE MIEUM EO RIEUL-KIYEOK")
    (16302 . "HANGUL SYLLABLE MIEUM EO SSANGSIOS")
    (16303 . "HANGUL SYLLABLE MIEUM EO CHIEUCH")
    (16304 . "HANGUL SYLLABLE MIEUM E TIKEUT")
    (16305 . "HANGUL SYLLABLE MIEUM E RIEUL-SIOS")
    (16306 . "HANGUL SYLLABLE MIEUM E CIEUC")
    (16307 . "HANGUL SYLLABLE MIEUM E CHIEUCH")
    (16308 . "HANGUL SYLLABLE MIEUM E THIEUTH")
    (16309 . "HANGUL SYLLABLE MIEUM YEO TIKEUT")
    (16310 . "HANGUL SYLLABLE MIEUM YEO RIEUL-MIEUM")
    (16311 . "HANGUL SYLLABLE MIEUM YEO MIEUM")
    (16312 . "HANGUL SYLLABLE MIEUM YEO PIEUP")
    (16313 . "HANGUL SYLLABLE MIEUM YEO THIEUTH")
    (16314 . "HANGUL SYLLABLE MIEUM YE KIYEOK")
    (16315 . "HANGUL SYLLABLE MIEUM YE NIEUN")
    (16316 . "HANGUL SYLLABLE MIEUM YE TIKEUT")
    (16317 . "HANGUL SYLLABLE MIEUM YE RIEUL")
    (16318 . "HANGUL SYLLABLE MIEUM YE MIEUM")
    (16319 . "HANGUL SYLLABLE MIEUM YE PIEUP")
    (16320 . "HANGUL SYLLABLE MIEUM YE SIOS")
    (16321 . "HANGUL SYLLABLE MIEUM YE IEUNG")
    (16322 . "HANGUL SYLLABLE MIEUM YE CHIEUCH")
    (16323 . "HANGUL SYLLABLE MIEUM YE THIEUTH")
    (16324 . "HANGUL SYLLABLE MIEUM YE PHIEUPH")
    (16325 . "HANGUL SYLLABLE MIEUM O TIKEUT")
    (16326 . "HANGUL SYLLABLE MIEUM O RIEUL-KIYEOK")
    (16327 . "HANGUL SYLLABLE MIEUM O RIEUL-SIOS")
    (16328 . "HANGUL SYLLABLE MIEUM O CHIEUCH")
    (16329 . "HANGUL SYLLABLE MIEUM O HIEUH")
    (16330 . "HANGUL SYLLABLE MIEUM WA KIYEOK")
    (16331 . "HANGUL SYLLABLE MIEUM WA TIKEUT")
    (16332 . "HANGUL SYLLABLE MIEUM WA RIEUL")
    (16333 . "HANGUL SYLLABLE MIEUM WA MIEUM")
    (16334 . "HANGUL SYLLABLE MIEUM WA PIEUP")
    (16335 . "HANGUL SYLLABLE MIEUM WA SIOS")
    (16336 . "HANGUL SYLLABLE MIEUM WAE")
    (16337 . "HANGUL SYLLABLE MIEUM WAE NIEUN")
    (16338 . "HANGUL SYLLABLE MIEUM WAE RIEUL")
    (16339 . "HANGUL SYLLABLE MIEUM WAE PIEUP")
    (16340 . "HANGUL SYLLABLE MIEUM WAE SSANGSIOS")
    (16341 . "HANGUL SYLLABLE MIEUM OE KIYEOK")
    (16342 . "HANGUL SYLLABLE MIEUM OE TIKEUT")
    (16343 . "HANGUL SYLLABLE MIEUM OE MIEUM")
    (16344 . "HANGUL SYLLABLE MIEUM YO KIYEOK")
    (16345 . "HANGUL SYLLABLE MIEUM YO TIKEUT")
    (16346 . "HANGUL SYLLABLE MIEUM YO MIEUM")
    (16347 . "HANGUL SYLLABLE MIEUM YO IEUNG")
    (16348 . "HANGUL SYLLABLE MIEUM U NIEUN-HIEUH")
    (16349 . "HANGUL SYLLABLE MIEUM U RIEUL-SIOS")
    (16350 . "HANGUL SYLLABLE MIEUM U CHIEUCH")
    (16351 . "HANGUL SYLLABLE MIEUM WEO KIYEOK")
    (16352 . "HANGUL SYLLABLE MIEUM WEO TIKEUT")
    (16353 . "HANGUL SYLLABLE MIEUM WEO MIEUM")
    (16354 . "HANGUL SYLLABLE MIEUM WEO SSANGSIOS")
    (16355 . "HANGUL SYLLABLE MIEUM WEO IEUNG")
    (16356 . "HANGUL SYLLABLE MIEUM WE KIYEOK")
    (16357 . "HANGUL SYLLABLE MIEUM WE NIEUN")
    (16358 . "HANGUL SYLLABLE MIEUM WE RIEUL")
    (16359 . "HANGUL SYLLABLE MIEUM WE MIEUM")
    (16360 . "HANGUL SYLLABLE MIEUM WE PIEUP")
    (16361 . "HANGUL SYLLABLE MIEUM WE SIOS")
    (16362 . "HANGUL SYLLABLE MIEUM WE SSANGSIOS")
    (16363 . "HANGUL SYLLABLE MIEUM WE IEUNG")
    (16364 . "HANGUL SYLLABLE MIEUM WI KIYEOK")
    (16365 . "HANGUL SYLLABLE MIEUM WI SSANGKIYEOK")
    (16366 . "HANGUL SYLLABLE MIEUM WI MIEUM")
    (16367 . "HANGUL SYLLABLE MIEUM WI PIEUP")
    (16368 . "HANGUL SYLLABLE MIEUM WI SIOS")
    (16369 . "HANGUL SYLLABLE MIEUM WI SSANGSIOS")
    (16370 . "HANGUL SYLLABLE MIEUM WI IEUNG")
    (16371 . "HANGUL SYLLABLE MIEUM YU KIYEOK")
    (16372 . "HANGUL SYLLABLE MIEUM YU TIKEUT")
    (16373 . "HANGUL SYLLABLE MIEUM YU PIEUP")
    (16374 . "HANGUL SYLLABLE MIEUM YU IEUNG")
    (16375 . "HANGUL SYLLABLE MIEUM EU KIYEOK")
    (16376 . "HANGUL SYLLABLE MIEUM EU TIKEUT")
    (16377 . "HANGUL SYLLABLE MIEUM EU RIEUL-KIYEOK")
    (16378 . "HANGUL SYLLABLE MIEUM EU RIEUL-SIOS")
    (16379 . "HANGUL SYLLABLE MIEUM EU RIEUL-HIEUH")
    (16380 . "HANGUL SYLLABLE MIEUM EU PIEUP")
    (16381 . "HANGUL SYLLABLE MIEUM EU IEUNG")
    (16382 . "HANGUL SYLLABLE MIEUM EU THIEUTH")
    (16383 . "HANGUL SYLLABLE MIEUM YI")
    (16384 . "HANGUL SYLLABLE MIEUM YI KIYEOK")
    (16385 . "HANGUL SYLLABLE MIEUM YI NIEUN")
    (16386 . "HANGUL SYLLABLE MIEUM YI RIEUL")
    (16387 . "HANGUL SYLLABLE MIEUM YI MIEUM")
    (16388 . "HANGUL SYLLABLE MIEUM YI PIEUP")
    (16389 . "HANGUL SYLLABLE MIEUM YI SIOS")
    (16390 . "HANGUL SYLLABLE MIEUM YI IEUNG")
    (16391 . "HANGUL SYLLABLE MIEUM YI CIEUC")
    (16392 . "HANGUL SYLLABLE MIEUM I RIEUL-KIYEOK")
    (16393 . "HANGUL SYLLABLE MIEUM I RIEUL-SIOS")
    (16394 . "HANGUL SYLLABLE MIEUM I CIEUC")
    (16395 . "HANGUL SYLLABLE MIEUM I KHIEUKH")
    (16396 . "HANGUL SYLLABLE PIEUP A RIEUL-SIOS")
    (16397 . "HANGUL SYLLABLE PIEUP A RIEUL-HIEUH")
    (16398 . "HANGUL SYLLABLE PIEUP A SSANGSIOS")
    (16399 . "HANGUL SYLLABLE PIEUP A CHIEUCH")
    (16400 . "HANGUL SYLLABLE PIEUP A KHIEUKH")
    (16401 . "HANGUL SYLLABLE PIEUP A HIEUH")
    (16402 . "HANGUL SYLLABLE PIEUP AE TIKEUT")
    (16403 . "HANGUL SYLLABLE PIEUP AE CIEUC")
    (16404 . "HANGUL SYLLABLE PIEUP YA RIEUL")
    (16405 . "HANGUL SYLLABLE PIEUP YA MIEUM")
    (16406 . "HANGUL SYLLABLE PIEUP YA SIOS")
    (16407 . "HANGUL SYLLABLE PIEUP YA IEUNG")
    (16408 . "HANGUL SYLLABLE PIEUP YA HIEUH")
    (16409 . "HANGUL SYLLABLE PIEUP YAE")
    (16410 . "HANGUL SYLLABLE PIEUP YAE RIEUL")
    (16411 . "HANGUL SYLLABLE PIEUP YAE RIEUL-HIEUH")
    (16412 . "HANGUL SYLLABLE PIEUP YAE IEUNG")
    (16413 . "HANGUL SYLLABLE PIEUP EO SSANGKIYEOK")
    (16414 . "HANGUL SYLLABLE PIEUP EO RIEUL-KIYEOK")
    (16415 . "HANGUL SYLLABLE PIEUP EO RIEUL-SIOS")
    (16416 . "HANGUL SYLLABLE PIEUP EO SSANGSIOS")
    (16417 . "HANGUL SYLLABLE PIEUP EO KHIEUKH")
    (16418 . "HANGUL SYLLABLE PIEUP EO THIEUTH")
    (16419 . "HANGUL SYLLABLE PIEUP E SSANGKIYEOK")
    (16420 . "HANGUL SYLLABLE PIEUP E KHIEUKH")
    (16421 . "HANGUL SYLLABLE PIEUP E THIEUTH")
    (16422 . "HANGUL SYLLABLE PIEUP YEO TIKEUT")
    (16423 . "HANGUL SYLLABLE PIEUP YEO RIEUL-SIOS")
    (16424 . "HANGUL SYLLABLE PIEUP YEO MIEUM")
    (16425 . "HANGUL SYLLABLE PIEUP YEO CIEUC")
    (16426 . "HANGUL SYLLABLE PIEUP YEO CHIEUCH")
    (16427 . "HANGUL SYLLABLE PIEUP YEO KHIEUKH")
    (16428 . "HANGUL SYLLABLE PIEUP YE KIYEOK")
    (16429 . "HANGUL SYLLABLE PIEUP YE RIEUL")
    (16430 . "HANGUL SYLLABLE PIEUP YE MIEUM")
    (16431 . "HANGUL SYLLABLE PIEUP YE PIEUP")
    (16432 . "HANGUL SYLLABLE PIEUP YE SIOS")
    (16433 . "HANGUL SYLLABLE PIEUP YE IEUNG")
    (16434 . "HANGUL SYLLABLE PIEUP O KIYEOK-SIOS")
    (16435 . "HANGUL SYLLABLE PIEUP O TIKEUT")
    (16436 . "HANGUL SYLLABLE PIEUP O RIEUL-KIYEOK")
    (16437 . "HANGUL SYLLABLE PIEUP O RIEUL-PIEUP")
    (16438 . "HANGUL SYLLABLE PIEUP O CIEUC")
    (16439 . "HANGUL SYLLABLE PIEUP O CHIEUCH")
    (16440 . "HANGUL SYLLABLE PIEUP O THIEUTH")
    (16441 . "HANGUL SYLLABLE PIEUP O HIEUH")
    (16442 . "HANGUL SYLLABLE PIEUP WA KIYEOK")
    (16443 . "HANGUL SYLLABLE PIEUP WA TIKEUT")
    (16444 . "HANGUL SYLLABLE PIEUP WA RIEUL")
    (16445 . "HANGUL SYLLABLE PIEUP WA MIEUM")
    (16446 . "HANGUL SYLLABLE PIEUP WA PIEUP")
    (16447 . "HANGUL SYLLABLE PIEUP WA SIOS")
    (16448 . "HANGUL SYLLABLE PIEUP WA IEUNG")
    (16449 . "HANGUL SYLLABLE PIEUP WAE NIEUN")
    (16450 . "HANGUL SYLLABLE PIEUP WAE TIKEUT")
    (16451 . "HANGUL SYLLABLE PIEUP WAE RIEUL")
    (16452 . "HANGUL SYLLABLE PIEUP OE SSANGKIYEOK")
    (16453 . "HANGUL SYLLABLE PIEUP OE TIKEUT")
    (16454 . "HANGUL SYLLABLE PIEUP OE SIOS")
    (16455 . "HANGUL SYLLABLE PIEUP OE SSANGSIOS")
    (16456 . "HANGUL SYLLABLE PIEUP OE IEUNG")
    (16457 . "HANGUL SYLLABLE PIEUP YO KIYEOK")
    (16458 . "HANGUL SYLLABLE PIEUP YO RIEUL")
    (16459 . "HANGUL SYLLABLE PIEUP YO MIEUM")
    (16460 . "HANGUL SYLLABLE PIEUP YO PIEUP")
    (16461 . "HANGUL SYLLABLE PIEUP YO SIOS")
    (16462 . "HANGUL SYLLABLE PIEUP YO IEUNG")
    (16463 . "HANGUL SYLLABLE PIEUP U SSANGKIYEOK")
    (16464 . "HANGUL SYLLABLE PIEUP U RIEUL-PIEUP")
    (16465 . "HANGUL SYLLABLE PIEUP U RIEUL-PHIEUPH")
    (16466 . "HANGUL SYLLABLE PIEUP U CIEUC")
    (16467 . "HANGUL SYLLABLE PIEUP U CHIEUCH")
    (16468 . "HANGUL SYLLABLE PIEUP U KHIEUKH")
    (16469 . "HANGUL SYLLABLE PIEUP U HIEUH")
    (16470 . "HANGUL SYLLABLE PIEUP WEO KIYEOK")
    (16471 . "HANGUL SYLLABLE PIEUP WEO NIEUN")
    (16472 . "HANGUL SYLLABLE PIEUP WEO TIKEUT")
    (16473 . "HANGUL SYLLABLE PIEUP WEO MIEUM")
    (16474 . "HANGUL SYLLABLE PIEUP WEO PIEUP")
    (16475 . "HANGUL SYLLABLE PIEUP WEO SIOS")
    (16476 . "HANGUL SYLLABLE PIEUP WEO IEUNG")
    (16477 . "HANGUL SYLLABLE PIEUP WEO KHIEUKH")
    (16478 . "HANGUL SYLLABLE PIEUP WE KIYEOK")
    (16479 . "HANGUL SYLLABLE PIEUP WE NIEUN")
    (16480 . "HANGUL SYLLABLE PIEUP WE TIKEUT")
    (16481 . "HANGUL SYLLABLE PIEUP WE RIEUL")
    (16482 . "HANGUL SYLLABLE PIEUP WE MIEUM")
    (16483 . "HANGUL SYLLABLE PIEUP WE PIEUP")
    (16484 . "HANGUL SYLLABLE PIEUP WE SIOS")
    (16485 . "HANGUL SYLLABLE PIEUP WE SSANGSIOS")
    (16486 . "HANGUL SYLLABLE PIEUP WE IEUNG")
    (16487 . "HANGUL SYLLABLE PIEUP WI TIKEUT")
    (16488 . "HANGUL SYLLABLE PIEUP WI MIEUM")
    (16489 . "HANGUL SYLLABLE PIEUP WI PIEUP")
    (16490 . "HANGUL SYLLABLE PIEUP WI SIOS")
    (16491 . "HANGUL SYLLABLE PIEUP YU KIYEOK")
    (16492 . "HANGUL SYLLABLE PIEUP YU TIKEUT")
    (16493 . "HANGUL SYLLABLE PIEUP YU PIEUP")
    (16494 . "HANGUL SYLLABLE PIEUP EU TIKEUT")
    (16495 . "HANGUL SYLLABLE PIEUP EU RIEUL-KIYEOK")
    (16496 . "HANGUL SYLLABLE PIEUP EU RIEUL-PIEUP")
    (16497 . "HANGUL SYLLABLE PIEUP EU RIEUL-SIOS")
    (16498 . "HANGUL SYLLABLE PIEUP EU RIEUL-HIEUH")
    (16499 . "HANGUL SYLLABLE PIEUP EU IEUNG")
    (16500 . "HANGUL SYLLABLE PIEUP EU CHIEUCH")
    (16501 . "HANGUL SYLLABLE PIEUP EU THIEUTH")
    (16502 . "HANGUL SYLLABLE PIEUP EU HIEUH")
    (16503 . "HANGUL SYLLABLE PIEUP YI")
    (16504 . "HANGUL SYLLABLE PIEUP YI KIYEOK")
    (16505 . "HANGUL SYLLABLE PIEUP YI NIEUN")
    (16506 . "HANGUL SYLLABLE PIEUP YI RIEUL")
    (16507 . "HANGUL SYLLABLE PIEUP YI MIEUM")
    (16508 . "HANGUL SYLLABLE PIEUP YI PIEUP")
    (16509 . "HANGUL SYLLABLE PIEUP YI SIOS")
    (16510 . "HANGUL SYLLABLE PIEUP YI IEUNG")
    (16511 . "HANGUL SYLLABLE PIEUP I TIKEUT")
    (16512 . "HANGUL SYLLABLE PIEUP I SSANGSIOS")
    (16513 . "HANGUL SYLLABLE PIEUP I KHIEUKH")
    (16514 . "HANGUL SYLLABLE PIEUP I THIEUTH")
    (16515 . "HANGUL SYLLABLE PIEUP I HIEUH")
    (16516 . "HANGUL SYLLABLE SSANGPIEUP A SSANGKIYEOK")
    (16517 . "HANGUL SYLLABLE SSANGPIEUP A TIKEUT")
    (16518 . "HANGUL SYLLABLE SSANGPIEUP A RIEUL-KIYEOK")
    (16519 . "HANGUL SYLLABLE SSANGPIEUP A CIEUC")
    (16520 . "HANGUL SYLLABLE SSANGPIEUP A THIEUTH")
    (16521 . "HANGUL SYLLABLE SSANGPIEUP AE SSANGKIYEOK")
    (16522 . "HANGUL SYLLABLE SSANGPIEUP AE TIKEUT")
    (16523 . "HANGUL SYLLABLE SSANGPIEUP YA NIEUN")
    (16524 . "HANGUL SYLLABLE SSANGPIEUP YAE")
    (16525 . "HANGUL SYLLABLE SSANGPIEUP YAE NIEUN")
    (16526 . "HANGUL SYLLABLE SSANGPIEUP EO PIEUP")
    (16527 . "HANGUL SYLLABLE SSANGPIEUP EO CIEUC")
    (16528 . "HANGUL SYLLABLE SSANGPIEUP E KIYEOK")
    (16529 . "HANGUL SYLLABLE SSANGPIEUP E NIEUN")
    (16530 . "HANGUL SYLLABLE SSANGPIEUP E TIKEUT")
    (16531 . "HANGUL SYLLABLE SSANGPIEUP E RIEUL")
    (16532 . "HANGUL SYLLABLE SSANGPIEUP E MIEUM")
    (16533 . "HANGUL SYLLABLE SSANGPIEUP E PIEUP")
    (16534 . "HANGUL SYLLABLE SSANGPIEUP E SIOS")
    (16535 . "HANGUL SYLLABLE SSANGPIEUP YEO NIEUN")
    (16536 . "HANGUL SYLLABLE SSANGPIEUP YEO TIKEUT")
    (16537 . "HANGUL SYLLABLE SSANGPIEUP YEO RIEUL")
    (16538 . "HANGUL SYLLABLE SSANGPIEUP YEO CHIEUCH")
    (16539 . "HANGUL SYLLABLE SSANGPIEUP YEO THIEUTH")
    (16540 . "HANGUL SYLLABLE SSANGPIEUP YE")
    (16541 . "HANGUL SYLLABLE SSANGPIEUP YE SIOS")
    (16542 . "HANGUL SYLLABLE SSANGPIEUP YE SSANGSIOS")
    (16543 . "HANGUL SYLLABLE SSANGPIEUP YE IEUNG")
    (16544 . "HANGUL SYLLABLE SSANGPIEUP O TIKEUT")
    (16545 . "HANGUL SYLLABLE SSANGPIEUP O SIOS")
    (16546 . "HANGUL SYLLABLE SSANGPIEUP O HIEUH")
    (16547 . "HANGUL SYLLABLE SSANGPIEUP WA")
    (16548 . "HANGUL SYLLABLE SSANGPIEUP WA NIEUN")
    (16549 . "HANGUL SYLLABLE SSANGPIEUP WA TIKEUT")
    (16550 . "HANGUL SYLLABLE SSANGPIEUP WAE")
    (16551 . "HANGUL SYLLABLE SSANGPIEUP WAE TIKEUT")
    (16552 . "HANGUL SYLLABLE SSANGPIEUP OE MIEUM")
    (16553 . "HANGUL SYLLABLE SSANGPIEUP OE PIEUP")
    (16554 . "HANGUL SYLLABLE SSANGPIEUP OE KHIEUKH")
    (16555 . "HANGUL SYLLABLE SSANGPIEUP YO MIEUM")
    (16556 . "HANGUL SYLLABLE SSANGPIEUP YO PIEUP")
    (16557 . "HANGUL SYLLABLE SSANGPIEUP U TIKEUT")
    (16558 . "HANGUL SYLLABLE SSANGPIEUP U RIEUL-KIYEOK")
    (16559 . "HANGUL SYLLABLE SSANGPIEUP U PIEUP")
    (16560 . "HANGUL SYLLABLE SSANGPIEUP WEO")
    (16561 . "HANGUL SYLLABLE SSANGPIEUP WEO TIKEUT")
    (16562 . "HANGUL SYLLABLE SSANGPIEUP WE")
    (16563 . "HANGUL SYLLABLE SSANGPIEUP WI")
    (16564 . "HANGUL SYLLABLE SSANGPIEUP WI KIYEOK")
    (16565 . "HANGUL SYLLABLE SSANGPIEUP YU TIKEUT")
    (16566 . "HANGUL SYLLABLE SSANGPIEUP YU RIEUL")
    (16567 . "HANGUL SYLLABLE SSANGPIEUP EU KIYEOK")
    (16568 . "HANGUL SYLLABLE SSANGPIEUP EU TIKEUT")
    (16569 . "HANGUL SYLLABLE SSANGPIEUP EU SIOS")
    (16570 . "HANGUL SYLLABLE SSANGPIEUP YI")
    (16571 . "HANGUL SYLLABLE SSANGPIEUP I TIKEUT")
    (16572 . "HANGUL SYLLABLE SSANGPIEUP I SIOS")
    (16573 . "HANGUL SYLLABLE SSANGPIEUP I CIEUC")
    (16574 . "HANGUL SYLLABLE SSANGPIEUP I CHIEUCH")
    (16575 . "HANGUL SYLLABLE SSANGPIEUP I HIEUH")
    (16576 . "HANGUL SYLLABLE SIOS A SSANGKIYEOK")
    (16577 . "HANGUL SYLLABLE SIOS A RIEUL-PIEUP")
    (16578 . "HANGUL SYLLABLE SIOS A RIEUL-SIOS")
    (16579 . "HANGUL SYLLABLE SIOS A RIEUL-PHIEUPH")
    (16580 . "HANGUL SYLLABLE SIOS A RIEUL-HIEUH")
    (16581 . "HANGUL SYLLABLE SIOS A CIEUC")
    (16582 . "HANGUL SYLLABLE SIOS A CHIEUCH")
    (16583 . "HANGUL SYLLABLE SIOS A KHIEUKH")
    (16584 . "HANGUL SYLLABLE SIOS A PHIEUPH")
    (16585 . "HANGUL SYLLABLE SIOS A HIEUH")
    (16586 . "HANGUL SYLLABLE SIOS AE KIYEOK-SIOS")
    (16587 . "HANGUL SYLLABLE SIOS AE TIKEUT")
    (16588 . "HANGUL SYLLABLE SIOS AE HIEUH")
    (16589 . "HANGUL SYLLABLE SIOS YA TIKEUT")
    (16590 . "HANGUL SYLLABLE SIOS YA SSANGSIOS")
    (16591 . "HANGUL SYLLABLE SIOS YA CHIEUCH")
    (16592 . "HANGUL SYLLABLE SIOS YA PHIEUPH")
    (16593 . "HANGUL SYLLABLE SIOS YA HIEUH")
    (16594 . "HANGUL SYLLABLE SIOS EO RIEUL-KIYEOK")
    (16595 . "HANGUL SYLLABLE SIOS EO RIEUL-SIOS")
    (16596 . "HANGUL SYLLABLE SIOS EO RIEUL-PHIEUPH")
    (16597 . "HANGUL SYLLABLE SIOS EO KHIEUKH")
    (16598 . "HANGUL SYLLABLE SIOS E TIKEUT")
    (16599 . "HANGUL SYLLABLE SIOS E CIEUC")
    (16600 . "HANGUL SYLLABLE SIOS E THIEUTH")
    (16601 . "HANGUL SYLLABLE SIOS E PHIEUPH")
    (16602 . "HANGUL SYLLABLE SIOS YEO KIYEOK-SIOS")
    (16603 . "HANGUL SYLLABLE SIOS YEO TIKEUT")
    (16604 . "HANGUL SYLLABLE SIOS YEO RIEUL-PIEUP")
    (16605 . "HANGUL SYLLABLE SIOS YEO RIEUL-SIOS")
    (16606 . "HANGUL SYLLABLE SIOS YE KIYEOK")
    (16607 . "HANGUL SYLLABLE SIOS YE KIYEOK-SIOS")
    (16608 . "HANGUL SYLLABLE SIOS YE MIEUM")
    (16609 . "HANGUL SYLLABLE SIOS YE PIEUP")
    (16610 . "HANGUL SYLLABLE SIOS YE SIOS")
    (16611 . "HANGUL SYLLABLE SIOS YE SSANGSIOS")
    (16612 . "HANGUL SYLLABLE SIOS O TIKEUT")
    (16613 . "HANGUL SYLLABLE SIOS O RIEUL-KIYEOK")
    (16614 . "HANGUL SYLLABLE SIOS O RIEUL-PIEUP")
    (16615 . "HANGUL SYLLABLE SIOS O RIEUL-SIOS")
    (16616 . "HANGUL SYLLABLE SIOS O CHIEUCH")
    (16617 . "HANGUL SYLLABLE SIOS O PHIEUPH")
    (16618 . "HANGUL SYLLABLE SIOS WA MIEUM")
    (16619 . "HANGUL SYLLABLE SIOS WA PIEUP")
    (16620 . "HANGUL SYLLABLE SIOS WA SIOS")
    (16621 . "HANGUL SYLLABLE SIOS WA SSANGSIOS")
    (16622 . "HANGUL SYLLABLE SIOS WAE KIYEOK")
    (16623 . "HANGUL SYLLABLE SIOS WAE TIKEUT")
    (16624 . "HANGUL SYLLABLE SIOS WAE PIEUP")
    (16625 . "HANGUL SYLLABLE SIOS WAE IEUNG")
    (16626 . "HANGUL SYLLABLE SIOS OE KIYEOK")
    (16627 . "HANGUL SYLLABLE SIOS OE TIKEUT")
    (16628 . "HANGUL SYLLABLE SIOS OE SSANGSIOS")
    (16629 . "HANGUL SYLLABLE SIOS OE IEUNG")
    (16630 . "HANGUL SYLLABLE SIOS YO TIKEUT")
    (16631 . "HANGUL SYLLABLE SIOS YO PHIEUPH")
    (16632 . "HANGUL SYLLABLE SIOS U RIEUL-KIYEOK")
    (16633 . "HANGUL SYLLABLE SIOS U RIEUL-SIOS")
    (16634 . "HANGUL SYLLABLE SIOS U RIEUL-HIEUH")
    (16635 . "HANGUL SYLLABLE SIOS U CIEUC")
    (16636 . "HANGUL SYLLABLE SIOS U HIEUH")
    (16637 . "HANGUL SYLLABLE SIOS WEO KIYEOK")
    (16638 . "HANGUL SYLLABLE SIOS WEO NIEUN")
    (16639 . "HANGUL SYLLABLE SIOS WEO TIKEUT")
    (16640 . "HANGUL SYLLABLE SIOS WEO RIEUL")
    (16641 . "HANGUL SYLLABLE SIOS WEO MIEUM")
    (16642 . "HANGUL SYLLABLE SIOS WEO PIEUP")
    (16643 . "HANGUL SYLLABLE SIOS WEO SIOS")
    (16644 . "HANGUL SYLLABLE SIOS WEO IEUNG")
    (16645 . "HANGUL SYLLABLE SIOS WE TIKEUT")
    (16646 . "HANGUL SYLLABLE SIOS WE PIEUP")
    (16647 . "HANGUL SYLLABLE SIOS WE SIOS")
    (16648 . "HANGUL SYLLABLE SIOS WE SSANGSIOS")
    (16649 . "HANGUL SYLLABLE SIOS WI TIKEUT")
    (16650 . "HANGUL SYLLABLE SIOS WI PHIEUPH")
    (16651 . "HANGUL SYLLABLE SIOS YU NIEUN")
    (16652 . "HANGUL SYLLABLE SIOS YU TIKEUT")
    (16653 . "HANGUL SYLLABLE SIOS YU PIEUP")
    (16654 . "HANGUL SYLLABLE SIOS EU TIKEUT")
    (16655 . "HANGUL SYLLABLE SIOS EU RIEUL-MIEUM")
    (16656 . "HANGUL SYLLABLE SIOS EU RIEUL-PIEUP")
    (16657 . "HANGUL SYLLABLE SIOS EU RIEUL-SIOS")
    (16658 . "HANGUL SYLLABLE SIOS EU RIEUL-PHIEUPH")
    (16659 . "HANGUL SYLLABLE SIOS EU RIEUL-HIEUH")
    (16660 . "HANGUL SYLLABLE SIOS EU CIEUC")
    (16661 . "HANGUL SYLLABLE SIOS EU CHIEUCH")
    (16662 . "HANGUL SYLLABLE SIOS EU PHIEUPH")
    (16663 . "HANGUL SYLLABLE SIOS YI")
    (16664 . "HANGUL SYLLABLE SIOS YI KIYEOK")
    (16665 . "HANGUL SYLLABLE SIOS YI NIEUN")
    (16666 . "HANGUL SYLLABLE SIOS YI RIEUL")
    (16667 . "HANGUL SYLLABLE SIOS YI MIEUM")
    (16668 . "HANGUL SYLLABLE SIOS YI PIEUP")
    (16669 . "HANGUL SYLLABLE SIOS YI SIOS")
    (16670 . "HANGUL SYLLABLE SIOS YI IEUNG")
    (16671 . "HANGUL SYLLABLE SIOS I SSANGKIYEOK")
    (16672 . "HANGUL SYLLABLE SIOS I NIEUN-HIEUH")
    (16673 . "HANGUL SYLLABLE SIOS I RIEUL-KIYEOK")
    (16674 . "HANGUL SYLLABLE SIOS I RIEUL-MIEUM")
    (16675 . "HANGUL SYLLABLE SIOS I RIEUL-SIOS")
    (16676 . "HANGUL SYLLABLE SIOS I SSANGSIOS")
    (16677 . "HANGUL SYLLABLE SIOS I CHIEUCH")
    (16678 . "HANGUL SYLLABLE SIOS I KHIEUKH")
    (16679 . "HANGUL SYLLABLE SSANGSIOS A TIKEUT")
    (16680 . "HANGUL SYLLABLE SSANGSIOS A RIEUL-KIYEOK")
    (16681 . "HANGUL SYLLABLE SSANGSIOS A RIEUL-MIEUM")
    (16682 . "HANGUL SYLLABLE SSANGSIOS A SIOS")
    (16683 . "HANGUL SYLLABLE SSANGSIOS A CIEUC")
    (16684 . "HANGUL SYLLABLE SSANGSIOS A THIEUTH")
    (16685 . "HANGUL SYLLABLE SSANGSIOS AE TIKEUT")
    (16686 . "HANGUL SYLLABLE SSANGSIOS AE SIOS")
    (16687 . "HANGUL SYLLABLE SSANGSIOS AE HIEUH")
    (16688 . "HANGUL SYLLABLE SSANGSIOS YA")
    (16689 . "HANGUL SYLLABLE SSANGSIOS YA KIYEOK")
    (16690 . "HANGUL SYLLABLE SSANGSIOS YA NIEUN")
    (16691 . "HANGUL SYLLABLE SSANGSIOS YA TIKEUT")
    (16692 . "HANGUL SYLLABLE SSANGSIOS YA RIEUL")
    (16693 . "HANGUL SYLLABLE SSANGSIOS YA MIEUM")
    (16694 . "HANGUL SYLLABLE SSANGSIOS YA PIEUP")
    (16695 . "HANGUL SYLLABLE SSANGSIOS YA SIOS")
    (16696 . "HANGUL SYLLABLE SSANGSIOS YAE")
    (16697 . "HANGUL SYLLABLE SSANGSIOS YAE NIEUN")
    (16698 . "HANGUL SYLLABLE SSANGSIOS YAE RIEUL")
    (16699 . "HANGUL SYLLABLE SSANGSIOS YAE MIEUM")
    (16700 . "HANGUL SYLLABLE SSANGSIOS YAE IEUNG")
    (16701 . "HANGUL SYLLABLE SSANGSIOS EO SSANGKIYEOK")
    (16702 . "HANGUL SYLLABLE SSANGSIOS EO TIKEUT")
    (16703 . "HANGUL SYLLABLE SSANGSIOS EO SIOS")
    (16704 . "HANGUL SYLLABLE SSANGSIOS EO PHIEUPH")
    (16705 . "HANGUL SYLLABLE SSANGSIOS E KIYEOK")
    (16706 . "HANGUL SYLLABLE SSANGSIOS E TIKEUT")
    (16707 . "HANGUL SYLLABLE SSANGSIOS E MIEUM")
    (16708 . "HANGUL SYLLABLE SSANGSIOS E PIEUP")
    (16709 . "HANGUL SYLLABLE SSANGSIOS E SIOS")
    (16710 . "HANGUL SYLLABLE SSANGSIOS E SSANGSIOS")
    (16711 . "HANGUL SYLLABLE SSANGSIOS E IEUNG")
    (16712 . "HANGUL SYLLABLE SSANGSIOS E THIEUTH")
    (16713 . "HANGUL SYLLABLE SSANGSIOS YEO")
    (16714 . "HANGUL SYLLABLE SSANGSIOS YEO KIYEOK")
    (16715 . "HANGUL SYLLABLE SSANGSIOS YEO NIEUN")
    (16716 . "HANGUL SYLLABLE SSANGSIOS YEO TIKEUT")
    (16717 . "HANGUL SYLLABLE SSANGSIOS YEO RIEUL")
    (16718 . "HANGUL SYLLABLE SSANGSIOS YEO MIEUM")
    (16719 . "HANGUL SYLLABLE SSANGSIOS YEO PIEUP")
    (16720 . "HANGUL SYLLABLE SSANGSIOS YEO SIOS")
    (16721 . "HANGUL SYLLABLE SSANGSIOS YEO SSANGSIOS")
    (16722 . "HANGUL SYLLABLE SSANGSIOS YEO IEUNG")
    (16723 . "HANGUL SYLLABLE SSANGSIOS YE")
    (16724 . "HANGUL SYLLABLE SSANGSIOS YE KIYEOK")
    (16725 . "HANGUL SYLLABLE SSANGSIOS YE MIEUM")
    (16726 . "HANGUL SYLLABLE SSANGSIOS YE PIEUP")
    (16727 . "HANGUL SYLLABLE SSANGSIOS YE SIOS")
    (16728 . "HANGUL SYLLABLE SSANGSIOS YE SSANGSIOS")
    (16729 . "HANGUL SYLLABLE SSANGSIOS YE IEUNG")
    (16730 . "HANGUL SYLLABLE SSANGSIOS O SIOS")
    (16731 . "HANGUL SYLLABLE SSANGSIOS O THIEUTH")
    (16732 . "HANGUL SYLLABLE SSANGSIOS WA TIKEUT")
    (16733 . "HANGUL SYLLABLE SSANGSIOS WA RIEUL")
    (16734 . "HANGUL SYLLABLE SSANGSIOS WA MIEUM")
    (16735 . "HANGUL SYLLABLE SSANGSIOS WA PIEUP")
    (16736 . "HANGUL SYLLABLE SSANGSIOS WA SIOS")
    (16737 . "HANGUL SYLLABLE SSANGSIOS WA IEUNG")
    (16738 . "HANGUL SYLLABLE SSANGSIOS WAE KIYEOK")
    (16739 . "HANGUL SYLLABLE SSANGSIOS WAE NIEUN")
    (16740 . "HANGUL SYLLABLE SSANGSIOS WAE TIKEUT")
    (16741 . "HANGUL SYLLABLE SSANGSIOS WAE RIEUL")
    (16742 . "HANGUL SYLLABLE SSANGSIOS WAE SIOS")
    (16743 . "HANGUL SYLLABLE SSANGSIOS OE KIYEOK")
    (16744 . "HANGUL SYLLABLE SSANGSIOS OE TIKEUT")
    (16745 . "HANGUL SYLLABLE SSANGSIOS OE SIOS")
    (16746 . "HANGUL SYLLABLE SSANGSIOS OE SSANGSIOS")
    (16747 . "HANGUL SYLLABLE SSANGSIOS OE IEUNG")
    (16748 . "HANGUL SYLLABLE SSANGSIOS YO KIYEOK")
    (16749 . "HANGUL SYLLABLE SSANGSIOS YO NIEUN")
    (16750 . "HANGUL SYLLABLE SSANGSIOS YO TIKEUT")
    (16751 . "HANGUL SYLLABLE SSANGSIOS YO RIEUL")
    (16752 . "HANGUL SYLLABLE SSANGSIOS YO MIEUM")
    (16753 . "HANGUL SYLLABLE SSANGSIOS YO PIEUP")
    (16754 . "HANGUL SYLLABLE SSANGSIOS YO SIOS")
    (16755 . "HANGUL SYLLABLE SSANGSIOS YO IEUNG")
    (16756 . "HANGUL SYLLABLE SSANGSIOS U TIKEUT")
    (16757 . "HANGUL SYLLABLE SSANGSIOS U SIOS")
    (16758 . "HANGUL SYLLABLE SSANGSIOS U PHIEUPH")
    (16759 . "HANGUL SYLLABLE SSANGSIOS WEO KIYEOK")
    (16760 . "HANGUL SYLLABLE SSANGSIOS WEO NIEUN")
    (16761 . "HANGUL SYLLABLE SSANGSIOS WEO TIKEUT")
    (16762 . "HANGUL SYLLABLE SSANGSIOS WEO RIEUL")
    (16763 . "HANGUL SYLLABLE SSANGSIOS WEO MIEUM")
    (16764 . "HANGUL SYLLABLE SSANGSIOS WEO PIEUP")
    (16765 . "HANGUL SYLLABLE SSANGSIOS WEO SIOS")
    (16766 . "HANGUL SYLLABLE SSANGSIOS WEO IEUNG")
    (16767 . "HANGUL SYLLABLE SSANGSIOS WE KIYEOK")
    (16768 . "HANGUL SYLLABLE SSANGSIOS WE NIEUN")
    (16769 . "HANGUL SYLLABLE SSANGSIOS WE TIKEUT")
    (16770 . "HANGUL SYLLABLE SSANGSIOS WE RIEUL")
    (16771 . "HANGUL SYLLABLE SSANGSIOS WE MIEUM")
    (16772 . "HANGUL SYLLABLE SSANGSIOS WE PIEUP")
    (16773 . "HANGUL SYLLABLE SSANGSIOS WE SIOS")
    (16774 . "HANGUL SYLLABLE SSANGSIOS WE SSANGSIOS")
    (16775 . "HANGUL SYLLABLE SSANGSIOS WE IEUNG")
    (16776 . "HANGUL SYLLABLE SSANGSIOS WI KIYEOK")
    (16777 . "HANGUL SYLLABLE SSANGSIOS WI TIKEUT")
    (16778 . "HANGUL SYLLABLE SSANGSIOS WI RIEUL")
    (16779 . "HANGUL SYLLABLE SSANGSIOS WI MIEUM")
    (16780 . "HANGUL SYLLABLE SSANGSIOS WI PIEUP")
    (16781 . "HANGUL SYLLABLE SSANGSIOS WI SIOS")
    (16782 . "HANGUL SYLLABLE SSANGSIOS WI IEUNG")
    (16783 . "HANGUL SYLLABLE SSANGSIOS WI HIEUH")
    (16784 . "HANGUL SYLLABLE SSANGSIOS YU")
    (16785 . "HANGUL SYLLABLE SSANGSIOS YU KIYEOK")
    (16786 . "HANGUL SYLLABLE SSANGSIOS YU NIEUN")
    (16787 . "HANGUL SYLLABLE SSANGSIOS YU TIKEUT")
    (16788 . "HANGUL SYLLABLE SSANGSIOS YU RIEUL")
    (16789 . "HANGUL SYLLABLE SSANGSIOS YU MIEUM")
    (16790 . "HANGUL SYLLABLE SSANGSIOS YU PIEUP")
    (16791 . "HANGUL SYLLABLE SSANGSIOS YU SIOS")
    (16792 . "HANGUL SYLLABLE SSANGSIOS EU TIKEUT")
    (16793 . "HANGUL SYLLABLE SSANGSIOS EU SIOS")
    (16794 . "HANGUL SYLLABLE SSANGSIOS EU IEUNG")
    (16795 . "HANGUL SYLLABLE SSANGSIOS YI KIYEOK")
    (16796 . "HANGUL SYLLABLE SSANGSIOS YI PIEUP")
    (16797 . "HANGUL SYLLABLE SSANGSIOS YI SIOS")
    (16798 . "HANGUL SYLLABLE SSANGSIOS YI IEUNG")
    (16799 . "HANGUL SYLLABLE SSANGSIOS I TIKEUT")
    (16800 . "HANGUL SYLLABLE SSANGSIOS I RIEUL-KIYEOK")
    (16801 . "HANGUL SYLLABLE SSANGSIOS I SSANGSIOS")
    (16802 . "HANGUL SYLLABLE SSANGSIOS I CHIEUCH")
    (16803 . "HANGUL SYLLABLE SSANGSIOS I KHIEUKH")
    (16804 . "HANGUL SYLLABLE SSANGSIOS I HIEUH")
    (16805 . "HANGUL SYLLABLE IEUNG A TIKEUT")
    (16806 . "HANGUL SYLLABLE IEUNG A RIEUL-PIEUP")
    (16807 . "HANGUL SYLLABLE IEUNG A RIEUL-PHIEUPH")
    (16808 . "HANGUL SYLLABLE IEUNG A PIEUP-SIOS")
    (16809 . "HANGUL SYLLABLE IEUNG A CIEUC")
    (16810 . "HANGUL SYLLABLE IEUNG A CHIEUCH")
    (16811 . "HANGUL SYLLABLE IEUNG A HIEUH")
    (16812 . "HANGUL SYLLABLE IEUNG AE TIKEUT")
    (16813 . "HANGUL SYLLABLE IEUNG AE THIEUTH")
    (16814 . "HANGUL SYLLABLE IEUNG AE PHIEUPH")
    (16815 . "HANGUL SYLLABLE IEUNG YA SSANGKIYEOK")
    (16816 . "HANGUL SYLLABLE IEUNG YA TIKEUT")
    (16817 . "HANGUL SYLLABLE IEUNG YA SSANGSIOS")
    (16818 . "HANGUL SYLLABLE IEUNG YA CIEUC")
    (16819 . "HANGUL SYLLABLE IEUNG YA KHIEUKH")
    (16820 . "HANGUL SYLLABLE IEUNG YAE KIYEOK")
    (16821 . "HANGUL SYLLABLE IEUNG YAE TIKEUT")
    (16822 . "HANGUL SYLLABLE IEUNG YAE MIEUM")
    (16823 . "HANGUL SYLLABLE IEUNG YAE SIOS")
    (16824 . "HANGUL SYLLABLE IEUNG YAE SSANGSIOS")
    (16825 . "HANGUL SYLLABLE IEUNG YAE IEUNG")
    (16826 . "HANGUL SYLLABLE IEUNG EO SSANGKIYEOK")
    (16827 . "HANGUL SYLLABLE IEUNG EO NIEUN-HIEUH")
    (16828 . "HANGUL SYLLABLE IEUNG EO RIEUL-PIEUP")
    (16829 . "HANGUL SYLLABLE IEUNG EO HIEUH")
    (16830 . "HANGUL SYLLABLE IEUNG E TIKEUT")
    (16831 . "HANGUL SYLLABLE IEUNG E SSANGSIOS")
    (16832 . "HANGUL SYLLABLE IEUNG E THIEUTH")
    (16833 . "HANGUL SYLLABLE IEUNG E PHIEUPH")
    (16834 . "HANGUL SYLLABLE IEUNG YEO NIEUN-CIEUC")
    (16835 . "HANGUL SYLLABLE IEUNG YEO TIKEUT")
    (16836 . "HANGUL SYLLABLE IEUNG YEO RIEUL-SIOS")
    (16837 . "HANGUL SYLLABLE IEUNG YEO KHIEUKH")
    (16838 . "HANGUL SYLLABLE IEUNG YE KIYEOK")
    (16839 . "HANGUL SYLLABLE IEUNG YE TIKEUT")
    (16840 . "HANGUL SYLLABLE IEUNG YE IEUNG")
    (16841 . "HANGUL SYLLABLE IEUNG YE KHIEUKH")
    (16842 . "HANGUL SYLLABLE IEUNG O SSANGKIYEOK")
    (16843 . "HANGUL SYLLABLE IEUNG O NIEUN-HIEUH")
    (16844 . "HANGUL SYLLABLE IEUNG O TIKEUT")
    (16845 . "HANGUL SYLLABLE IEUNG O RIEUL-PIEUP")
    (16846 . "HANGUL SYLLABLE IEUNG O RIEUL-PHIEUPH")
    (16847 . "HANGUL SYLLABLE IEUNG O CIEUC")
    (16848 . "HANGUL SYLLABLE IEUNG O THIEUTH")
    (16849 . "HANGUL SYLLABLE IEUNG O PHIEUPH")
    (16850 . "HANGUL SYLLABLE IEUNG WA TIKEUT")
    (16851 . "HANGUL SYLLABLE IEUNG WA KHIEUKH")
    (16852 . "HANGUL SYLLABLE IEUNG WA THIEUTH")
    (16853 . "HANGUL SYLLABLE IEUNG WAE TIKEUT")
    (16854 . "HANGUL SYLLABLE IEUNG WAE RIEUL")
    (16855 . "HANGUL SYLLABLE IEUNG WAE PIEUP")
    (16856 . "HANGUL SYLLABLE IEUNG WAE SSANGSIOS")
    (16857 . "HANGUL SYLLABLE IEUNG OE TIKEUT")
    (16858 . "HANGUL SYLLABLE IEUNG OE RIEUL-MIEUM")
    (16859 . "HANGUL SYLLABLE IEUNG OE SSANGSIOS")
    (16860 . "HANGUL SYLLABLE IEUNG OE PHIEUPH")
    (16861 . "HANGUL SYLLABLE IEUNG YO TIKEUT")
    (16862 . "HANGUL SYLLABLE IEUNG YO RIEUL-KIYEOK")
    (16863 . "HANGUL SYLLABLE IEUNG YO KHIEUKH")
    (16864 . "HANGUL SYLLABLE IEUNG U TIKEUT")
    (16865 . "HANGUL SYLLABLE IEUNG U RIEUL-SIOS")
    (16866 . "HANGUL SYLLABLE IEUNG U PIEUP-SIOS")
    (16867 . "HANGUL SYLLABLE IEUNG U CIEUC")
    (16868 . "HANGUL SYLLABLE IEUNG U CHIEUCH")
    (16869 . "HANGUL SYLLABLE IEUNG U THIEUTH")
    (16870 . "HANGUL SYLLABLE IEUNG WEO NIEUN-HIEUH")
    (16871 . "HANGUL SYLLABLE IEUNG WEO TIKEUT")
    (16872 . "HANGUL SYLLABLE IEUNG WEO SIOS")
    (16873 . "HANGUL SYLLABLE IEUNG WEO KHIEUKH")
    (16874 . "HANGUL SYLLABLE IEUNG WEO THIEUTH")
    (16875 . "HANGUL SYLLABLE IEUNG WE TIKEUT")
    (16876 . "HANGUL SYLLABLE IEUNG WE SIOS")
    (16877 . "HANGUL SYLLABLE IEUNG WE SSANGSIOS")
    (16878 . "HANGUL SYLLABLE IEUNG WE THIEUTH")
    (16879 . "HANGUL SYLLABLE IEUNG WI TIKEUT")
    (16880 . "HANGUL SYLLABLE IEUNG WI RIEUL-MIEUM")
    (16881 . "HANGUL SYLLABLE IEUNG WI RIEUL-HIEUH")
    (16882 . "HANGUL SYLLABLE IEUNG WI SSANGSIOS")
    (16883 . "HANGUL SYLLABLE IEUNG WI KHIEUKH")
    (16884 . "HANGUL SYLLABLE IEUNG WI THIEUTH")
    (16885 . "HANGUL SYLLABLE IEUNG YU TIKEUT")
    (16886 . "HANGUL SYLLABLE IEUNG YU CIEUC")
    (16887 . "HANGUL SYLLABLE IEUNG YU KHIEUKH")
    (16888 . "HANGUL SYLLABLE IEUNG YU THIEUTH")
    (16889 . "HANGUL SYLLABLE IEUNG EU TIKEUT")
    (16890 . "HANGUL SYLLABLE IEUNG EU RIEUL-KIYEOK")
    (16891 . "HANGUL SYLLABLE IEUNG EU RIEUL-MIEUM")
    (16892 . "HANGUL SYLLABLE IEUNG EU RIEUL-PIEUP")
    (16893 . "HANGUL SYLLABLE IEUNG EU PIEUP-SIOS")
    (16894 . "HANGUL SYLLABLE IEUNG YI KIYEOK")
    (16895 . "HANGUL SYLLABLE IEUNG YI TIKEUT")
    (16896 . "HANGUL SYLLABLE IEUNG YI RIEUL-PHIEUPH")
    (16897 . "HANGUL SYLLABLE IEUNG YI PIEUP")
    (16898 . "HANGUL SYLLABLE IEUNG YI IEUNG")
    (16899 . "HANGUL SYLLABLE IEUNG I TIKEUT")
    (16900 . "HANGUL SYLLABLE IEUNG I RIEUL-SIOS")
    (16901 . "HANGUL SYLLABLE IEUNG I RIEUL-PHIEUPH")
    (16902 . "HANGUL SYLLABLE IEUNG I PIEUP-SIOS")
    (16903 . "HANGUL SYLLABLE IEUNG I CHIEUCH")
    (16904 . "HANGUL SYLLABLE IEUNG I KHIEUKH")
    (16905 . "HANGUL SYLLABLE IEUNG I THIEUTH")
    (16906 . "HANGUL SYLLABLE IEUNG I HIEUH")
    (16907 . "HANGUL SYLLABLE CIEUC A KIYEOK-SIOS")
    (16908 . "HANGUL SYLLABLE CIEUC A RIEUL-KIYEOK")
    (16909 . "HANGUL SYLLABLE CIEUC A CHIEUCH")
    (16910 . "HANGUL SYLLABLE CIEUC A THIEUTH")
    (16911 . "HANGUL SYLLABLE CIEUC A PHIEUPH")
    (16912 . "HANGUL SYLLABLE CIEUC A HIEUH")
    (16913 . "HANGUL SYLLABLE CIEUC AE NIEUN-HIEUH")
    (16914 . "HANGUL SYLLABLE CIEUC AE TIKEUT")
    (16915 . "HANGUL SYLLABLE CIEUC AE CIEUC")
    (16916 . "HANGUL SYLLABLE CIEUC YA TIKEUT")
    (16917 . "HANGUL SYLLABLE CIEUC YA PIEUP")
    (16918 . "HANGUL SYLLABLE CIEUC YA SIOS")
    (16919 . "HANGUL SYLLABLE CIEUC YAE TIKEUT")
    (16920 . "HANGUL SYLLABLE CIEUC YAE MIEUM")
    (16921 . "HANGUL SYLLABLE CIEUC YAE PIEUP")
    (16922 . "HANGUL SYLLABLE CIEUC YAE SIOS")
    (16923 . "HANGUL SYLLABLE CIEUC YAE IEUNG")
    (16924 . "HANGUL SYLLABLE CIEUC EO SSANGKIYEOK")
    (16925 . "HANGUL SYLLABLE CIEUC EO NIEUN-HIEUH")
    (16926 . "HANGUL SYLLABLE CIEUC EO TIKEUT")
    (16927 . "HANGUL SYLLABLE CIEUC EO RIEUL-KIYEOK")
    (16928 . "HANGUL SYLLABLE CIEUC EO SSANGSIOS")
    (16929 . "HANGUL SYLLABLE CIEUC EO THIEUTH")
    (16930 . "HANGUL SYLLABLE CIEUC EO HIEUH")
    (16931 . "HANGUL SYLLABLE CIEUC E TIKEUT")
    (16932 . "HANGUL SYLLABLE CIEUC E SSANGSIOS")
    (16933 . "HANGUL SYLLABLE CIEUC E CIEUC")
    (16934 . "HANGUL SYLLABLE CIEUC YEO KIYEOK")
    (16935 . "HANGUL SYLLABLE CIEUC YEO TIKEUT")
    (16936 . "HANGUL SYLLABLE CIEUC YEO RIEUL-KIYEOK")
    (16937 . "HANGUL SYLLABLE CIEUC YEO RIEUL-MIEUM")
    (16938 . "HANGUL SYLLABLE CIEUC YEO RIEUL-PIEUP")
    (16939 . "HANGUL SYLLABLE CIEUC YEO RIEUL-HIEUH")
    (16940 . "HANGUL SYLLABLE CIEUC YEO SIOS")
    (16941 . "HANGUL SYLLABLE CIEUC YEO CIEUC")
    (16942 . "HANGUL SYLLABLE CIEUC YEO THIEUTH")
    (16943 . "HANGUL SYLLABLE CIEUC YEO HIEUH")
    (16944 . "HANGUL SYLLABLE CIEUC YE KIYEOK")
    (16945 . "HANGUL SYLLABLE CIEUC YE NIEUN")
    (16946 . "HANGUL SYLLABLE CIEUC YE RIEUL")
    (16947 . "HANGUL SYLLABLE CIEUC YE MIEUM")
    (16948 . "HANGUL SYLLABLE CIEUC YE PIEUP")
    (16949 . "HANGUL SYLLABLE CIEUC YE SIOS")
    (16950 . "HANGUL SYLLABLE CIEUC YE IEUNG")
    (16951 . "HANGUL SYLLABLE CIEUC O TIKEUT")
    (16952 . "HANGUL SYLLABLE CIEUC WA NIEUN")
    (16953 . "HANGUL SYLLABLE CIEUC WA NIEUN-HIEUH")
    (16954 . "HANGUL SYLLABLE CIEUC WA TIKEUT")
    (16955 . "HANGUL SYLLABLE CIEUC WA RIEUL-KIYEOK")
    (16956 . "HANGUL SYLLABLE CIEUC WA MIEUM")
    (16957 . "HANGUL SYLLABLE CIEUC WA SSANGSIOS")
    (16958 . "HANGUL SYLLABLE CIEUC WAE KIYEOK")
    (16959 . "HANGUL SYLLABLE CIEUC WAE NIEUN")
    (16960 . "HANGUL SYLLABLE CIEUC WAE TIKEUT")
    (16961 . "HANGUL SYLLABLE CIEUC WAE RIEUL")
    (16962 . "HANGUL SYLLABLE CIEUC WAE MIEUM")
    (16963 . "HANGUL SYLLABLE CIEUC WAE SIOS")
    (16964 . "HANGUL SYLLABLE CIEUC OE KIYEOK")
    (16965 . "HANGUL SYLLABLE CIEUC OE TIKEUT")
    (16966 . "HANGUL SYLLABLE CIEUC OE SSANGSIOS")
    (16967 . "HANGUL SYLLABLE CIEUC YO TIKEUT")
    (16968 . "HANGUL SYLLABLE CIEUC YO RIEUL")
    (16969 . "HANGUL SYLLABLE CIEUC YO MIEUM")
    (16970 . "HANGUL SYLLABLE CIEUC YO PIEUP")
    (16971 . "HANGUL SYLLABLE CIEUC YO SIOS")
    (16972 . "HANGUL SYLLABLE CIEUC YO HIEUH")
    (16973 . "HANGUL SYLLABLE CIEUC U TIKEUT")
    (16974 . "HANGUL SYLLABLE CIEUC U RIEUL-SIOS")
    (16975 . "HANGUL SYLLABLE CIEUC U HIEUH")
    (16976 . "HANGUL SYLLABLE CIEUC WEO KIYEOK")
    (16977 . "HANGUL SYLLABLE CIEUC WEO NIEUN")
    (16978 . "HANGUL SYLLABLE CIEUC WEO TIKEUT")
    (16979 . "HANGUL SYLLABLE CIEUC WEO RIEUL")
    (16980 . "HANGUL SYLLABLE CIEUC WEO MIEUM")
    (16981 . "HANGUL SYLLABLE CIEUC WEO PIEUP")
    (16982 . "HANGUL SYLLABLE CIEUC WEO SIOS")
    (16983 . "HANGUL SYLLABLE CIEUC WEO IEUNG")
    (16984 . "HANGUL SYLLABLE CIEUC WE KIYEOK")
    (16985 . "HANGUL SYLLABLE CIEUC WE NIEUN")
    (16986 . "HANGUL SYLLABLE CIEUC WE TIKEUT")
    (16987 . "HANGUL SYLLABLE CIEUC WE RIEUL")
    (16988 . "HANGUL SYLLABLE CIEUC WE MIEUM")
    (16989 . "HANGUL SYLLABLE CIEUC WE PIEUP")
    (16990 . "HANGUL SYLLABLE CIEUC WE SIOS")
    (16991 . "HANGUL SYLLABLE CIEUC WE SSANGSIOS")
    (16992 . "HANGUL SYLLABLE CIEUC WE IEUNG")
    (16993 . "HANGUL SYLLABLE CIEUC WI TIKEUT")
    (16994 . "HANGUL SYLLABLE CIEUC WI IEUNG")
    (16995 . "HANGUL SYLLABLE CIEUC YU KIYEOK")
    (16996 . "HANGUL SYLLABLE CIEUC YU PIEUP")
    (16997 . "HANGUL SYLLABLE CIEUC YU SIOS")
    (16998 . "HANGUL SYLLABLE CIEUC YU IEUNG")
    (16999 . "HANGUL SYLLABLE CIEUC EU TIKEUT")
    (17000 . "HANGUL SYLLABLE CIEUC EU RIEUL-KIYEOK")
    (17001 . "HANGUL SYLLABLE CIEUC EU RIEUL-MIEUM")
    (17002 . "HANGUL SYLLABLE CIEUC EU RIEUL-SIOS")
    (17003 . "HANGUL SYLLABLE CIEUC EU CIEUC")
    (17004 . "HANGUL SYLLABLE CIEUC YI")
    (17005 . "HANGUL SYLLABLE CIEUC YI KIYEOK")
    (17006 . "HANGUL SYLLABLE CIEUC YI NIEUN")
    (17007 . "HANGUL SYLLABLE CIEUC YI RIEUL")
    (17008 . "HANGUL SYLLABLE CIEUC YI MIEUM")
    (17009 . "HANGUL SYLLABLE CIEUC YI PIEUP")
    (17010 . "HANGUL SYLLABLE CIEUC YI SIOS")
    (17011 . "HANGUL SYLLABLE CIEUC YI IEUNG")
    (17012 . "HANGUL SYLLABLE CIEUC I RIEUL-HIEUH")
    (17013 . "HANGUL SYLLABLE CIEUC I PIEUP-SIOS")
    (17014 . "HANGUL SYLLABLE CIEUC I SSANGSIOS")
    (17015 . "HANGUL SYLLABLE CIEUC I CHIEUCH")
    (17016 . "HANGUL SYLLABLE CIEUC I HIEUH")
    (17017 . "HANGUL SYLLABLE SSANGCIEUC A SSANGKIYEOK")
    (17018 . "HANGUL SYLLABLE SSANGCIEUC A TIKEUT")
    (17019 . "HANGUL SYLLABLE SSANGCIEUC A RIEUL-SIOS")
    (17020 . "HANGUL SYLLABLE SSANGCIEUC A HIEUH")
    (17021 . "HANGUL SYLLABLE SSANGCIEUC AE TIKEUT")
    (17022 . "HANGUL SYLLABLE SSANGCIEUC YA KIYEOK")
    (17023 . "HANGUL SYLLABLE SSANGCIEUC YA NIEUN-HIEUH")
    (17024 . "HANGUL SYLLABLE SSANGCIEUC YA RIEUL")
    (17025 . "HANGUL SYLLABLE SSANGCIEUC YA MIEUM")
    (17026 . "HANGUL SYLLABLE SSANGCIEUC YA PIEUP")
    (17027 . "HANGUL SYLLABLE SSANGCIEUC YA SIOS")
    (17028 . "HANGUL SYLLABLE SSANGCIEUC YA THIEUTH")
    (17029 . "HANGUL SYLLABLE SSANGCIEUC YAE")
    (17030 . "HANGUL SYLLABLE SSANGCIEUC EO TIKEUT")
    (17031 . "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-PIEUP")
    (17032 . "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-SIOS")
    (17033 . "HANGUL SYLLABLE SSANGCIEUC EO RIEUL-HIEUH")
    (17034 . "HANGUL SYLLABLE SSANGCIEUC EO HIEUH")
    (17035 . "HANGUL SYLLABLE SSANGCIEUC E KIYEOK")
    (17036 . "HANGUL SYLLABLE SSANGCIEUC E NIEUN")
    (17037 . "HANGUL SYLLABLE SSANGCIEUC E TIKEUT")
    (17038 . "HANGUL SYLLABLE SSANGCIEUC E RIEUL")
    (17039 . "HANGUL SYLLABLE SSANGCIEUC E MIEUM")
    (17040 . "HANGUL SYLLABLE SSANGCIEUC E PIEUP")
    (17041 . "HANGUL SYLLABLE SSANGCIEUC E SIOS")
    (17042 . "HANGUL SYLLABLE SSANGCIEUC E SSANGSIOS")
    (17043 . "HANGUL SYLLABLE SSANGCIEUC E THIEUTH")
    (17044 . "HANGUL SYLLABLE SSANGCIEUC YEO KIYEOK")
    (17045 . "HANGUL SYLLABLE SSANGCIEUC YEO NIEUN")
    (17046 . "HANGUL SYLLABLE SSANGCIEUC YEO TIKEUT")
    (17047 . "HANGUL SYLLABLE SSANGCIEUC YEO RIEUL")
    (17048 . "HANGUL SYLLABLE SSANGCIEUC YEO MIEUM")
    (17049 . "HANGUL SYLLABLE SSANGCIEUC YEO PIEUP")
    (17050 . "HANGUL SYLLABLE SSANGCIEUC YEO SIOS")
    (17051 . "HANGUL SYLLABLE SSANGCIEUC YEO IEUNG")
    (17052 . "HANGUL SYLLABLE SSANGCIEUC YE")
    (17053 . "HANGUL SYLLABLE SSANGCIEUC YE KIYEOK")
    (17054 . "HANGUL SYLLABLE SSANGCIEUC YE NIEUN")
    (17055 . "HANGUL SYLLABLE SSANGCIEUC YE RIEUL")
    (17056 . "HANGUL SYLLABLE SSANGCIEUC YE MIEUM")
    (17057 . "HANGUL SYLLABLE SSANGCIEUC YE PIEUP")
    (17058 . "HANGUL SYLLABLE SSANGCIEUC YE SIOS")
    (17059 . "HANGUL SYLLABLE SSANGCIEUC YE IEUNG")
    (17060 . "HANGUL SYLLABLE SSANGCIEUC O SSANGKIYEOK")
    (17061 . "HANGUL SYLLABLE SSANGCIEUC O TIKEUT")
    (17062 . "HANGUL SYLLABLE SSANGCIEUC O RIEUL-MIEUM")
    (17063 . "HANGUL SYLLABLE SSANGCIEUC O CIEUC")
    (17064 . "HANGUL SYLLABLE SSANGCIEUC O HIEUH")
    (17065 . "HANGUL SYLLABLE SSANGCIEUC WA NIEUN")
    (17066 . "HANGUL SYLLABLE SSANGCIEUC WA TIKEUT")
    (17067 . "HANGUL SYLLABLE SSANGCIEUC WA MIEUM")
    (17068 . "HANGUL SYLLABLE SSANGCIEUC WA PIEUP")
    (17069 . "HANGUL SYLLABLE SSANGCIEUC WA SIOS")
    (17070 . "HANGUL SYLLABLE SSANGCIEUC WA IEUNG")
    (17071 . "HANGUL SYLLABLE SSANGCIEUC WAE KIYEOK")
    (17072 . "HANGUL SYLLABLE SSANGCIEUC WAE NIEUN")
    (17073 . "HANGUL SYLLABLE SSANGCIEUC WAE TIKEUT")
    (17074 . "HANGUL SYLLABLE SSANGCIEUC WAE RIEUL")
    (17075 . "HANGUL SYLLABLE SSANGCIEUC OE KIYEOK")
    (17076 . "HANGUL SYLLABLE SSANGCIEUC OE TIKEUT")
    (17077 . "HANGUL SYLLABLE SSANGCIEUC OE SIOS")
    (17078 . "HANGUL SYLLABLE SSANGCIEUC OE SSANGSIOS")
    (17079 . "HANGUL SYLLABLE SSANGCIEUC OE IEUNG")
    (17080 . "HANGUL SYLLABLE SSANGCIEUC OE CHIEUCH")
    (17081 . "HANGUL SYLLABLE SSANGCIEUC YO")
    (17082 . "HANGUL SYLLABLE SSANGCIEUC YO KIYEOK")
    (17083 . "HANGUL SYLLABLE SSANGCIEUC YO NIEUN")
    (17084 . "HANGUL SYLLABLE SSANGCIEUC YO RIEUL")
    (17085 . "HANGUL SYLLABLE SSANGCIEUC YO MIEUM")
    (17086 . "HANGUL SYLLABLE SSANGCIEUC YO PIEUP")
    (17087 . "HANGUL SYLLABLE SSANGCIEUC YO SIOS")
    (17088 . "HANGUL SYLLABLE SSANGCIEUC U SSANGKIYEOK")
    (17089 . "HANGUL SYLLABLE SSANGCIEUC U TIKEUT")
    (17090 . "HANGUL SYLLABLE SSANGCIEUC U RIEUL-MIEUM")
    (17091 . "HANGUL SYLLABLE SSANGCIEUC U SIOS")
    (17092 . "HANGUL SYLLABLE SSANGCIEUC WEO KIYEOK")
    (17093 . "HANGUL SYLLABLE SSANGCIEUC WEO NIEUN")
    (17094 . "HANGUL SYLLABLE SSANGCIEUC WEO TIKEUT")
    (17095 . "HANGUL SYLLABLE SSANGCIEUC WEO RIEUL")
    (17096 . "HANGUL SYLLABLE SSANGCIEUC WEO MIEUM")
    (17097 . "HANGUL SYLLABLE SSANGCIEUC WEO PIEUP")
    (17098 . "HANGUL SYLLABLE SSANGCIEUC WEO SIOS")
    (17099 . "HANGUL SYLLABLE SSANGCIEUC WE")
    (17100 . "HANGUL SYLLABLE SSANGCIEUC WE KIYEOK")
    (17101 . "HANGUL SYLLABLE SSANGCIEUC WE NIEUN")
    (17102 . "HANGUL SYLLABLE SSANGCIEUC WE TIKEUT")
    (17103 . "HANGUL SYLLABLE SSANGCIEUC WE RIEUL")
    (17104 . "HANGUL SYLLABLE SSANGCIEUC WE MIEUM")
    (17105 . "HANGUL SYLLABLE SSANGCIEUC WE PIEUP")
    (17106 . "HANGUL SYLLABLE SSANGCIEUC WE SIOS")
    (17107 . "HANGUL SYLLABLE SSANGCIEUC WE IEUNG")
    (17108 . "HANGUL SYLLABLE SSANGCIEUC WI KIYEOK")
    (17109 . "HANGUL SYLLABLE SSANGCIEUC WI NIEUN")
    (17110 . "HANGUL SYLLABLE SSANGCIEUC WI TIKEUT")
    (17111 . "HANGUL SYLLABLE SSANGCIEUC WI RIEUL")
    (17112 . "HANGUL SYLLABLE SSANGCIEUC WI MIEUM")
    (17113 . "HANGUL SYLLABLE SSANGCIEUC WI PIEUP")
    (17114 . "HANGUL SYLLABLE SSANGCIEUC WI SIOS")
    (17115 . "HANGUL SYLLABLE SSANGCIEUC WI IEUNG")
    (17116 . "HANGUL SYLLABLE SSANGCIEUC YU KIYEOK")
    (17117 . "HANGUL SYLLABLE SSANGCIEUC YU NIEUN")
    (17118 . "HANGUL SYLLABLE SSANGCIEUC YU RIEUL")
    (17119 . "HANGUL SYLLABLE SSANGCIEUC YU MIEUM")
    (17120 . "HANGUL SYLLABLE SSANGCIEUC YU PIEUP")
    (17121 . "HANGUL SYLLABLE SSANGCIEUC YU SIOS")
    (17122 . "HANGUL SYLLABLE SSANGCIEUC YU IEUNG")
    (17123 . "HANGUL SYLLABLE SSANGCIEUC EU KIYEOK")
    (17124 . "HANGUL SYLLABLE SSANGCIEUC EU NIEUN")
    (17125 . "HANGUL SYLLABLE SSANGCIEUC EU TIKEUT")
    (17126 . "HANGUL SYLLABLE SSANGCIEUC EU RIEUL")
    (17127 . "HANGUL SYLLABLE SSANGCIEUC EU PIEUP")
    (17128 . "HANGUL SYLLABLE SSANGCIEUC EU CIEUC")
    (17129 . "HANGUL SYLLABLE SSANGCIEUC YI")
    (17130 . "HANGUL SYLLABLE SSANGCIEUC YI KIYEOK")
    (17131 . "HANGUL SYLLABLE SSANGCIEUC YI NIEUN")
    (17132 . "HANGUL SYLLABLE SSANGCIEUC YI RIEUL")
    (17133 . "HANGUL SYLLABLE SSANGCIEUC YI MIEUM")
    (17134 . "HANGUL SYLLABLE SSANGCIEUC YI PIEUP")
    (17135 . "HANGUL SYLLABLE SSANGCIEUC YI SIOS")
    (17136 . "HANGUL SYLLABLE SSANGCIEUC YI IEUNG")
    (17137 . "HANGUL SYLLABLE SSANGCIEUC YI CIEUC")
    (17138 . "HANGUL SYLLABLE SSANGCIEUC I TIKEUT")
    (17139 . "HANGUL SYLLABLE SSANGCIEUC I SIOS")
    (17140 . "HANGUL SYLLABLE SSANGCIEUC I PHIEUPH")
    (17141 . "HANGUL SYLLABLE CHIEUCH A TIKEUT")
    (17142 . "HANGUL SYLLABLE CHIEUCH A RIEUL-KIYEOK")
    (17143 . "HANGUL SYLLABLE CHIEUCH A CHIEUCH")
    (17144 . "HANGUL SYLLABLE CHIEUCH A PHIEUPH")
    (17145 . "HANGUL SYLLABLE CHIEUCH AE TIKEUT")
    (17146 . "HANGUL SYLLABLE CHIEUCH AE CHIEUCH")
    (17147 . "HANGUL SYLLABLE CHIEUCH AE PHIEUPH")
    (17148 . "HANGUL SYLLABLE CHIEUCH YA KIYEOK")
    (17149 . "HANGUL SYLLABLE CHIEUCH YA PIEUP")
    (17150 . "HANGUL SYLLABLE CHIEUCH YA SIOS")
    (17151 . "HANGUL SYLLABLE CHIEUCH YA PHIEUPH")
    (17152 . "HANGUL SYLLABLE CHIEUCH YAE")
    (17153 . "HANGUL SYLLABLE CHIEUCH YAE IEUNG")
    (17154 . "HANGUL SYLLABLE CHIEUCH EO TIKEUT")
    (17155 . "HANGUL SYLLABLE CHIEUCH EO RIEUL-HIEUH")
    (17156 . "HANGUL SYLLABLE CHIEUCH EO CIEUC")
    (17157 . "HANGUL SYLLABLE CHIEUCH EO PHIEUPH")
    (17158 . "HANGUL SYLLABLE CHIEUCH E TIKEUT")
    (17159 . "HANGUL SYLLABLE CHIEUCH E RIEUL-HIEUH")
    (17160 . "HANGUL SYLLABLE CHIEUCH E SSANGSIOS")
    (17161 . "HANGUL SYLLABLE CHIEUCH YEO KIYEOK")
    (17162 . "HANGUL SYLLABLE CHIEUCH YEO TIKEUT")
    (17163 . "HANGUL SYLLABLE CHIEUCH YEO RIEUL")
    (17164 . "HANGUL SYLLABLE CHIEUCH YEO MIEUM")
    (17165 . "HANGUL SYLLABLE CHIEUCH YEO PIEUP")
    (17166 . "HANGUL SYLLABLE CHIEUCH YEO SIOS")
    (17167 . "HANGUL SYLLABLE CHIEUCH YEO IEUNG")
    (17168 . "HANGUL SYLLABLE CHIEUCH YE KIYEOK")
    (17169 . "HANGUL SYLLABLE CHIEUCH YE RIEUL")
    (17170 . "HANGUL SYLLABLE CHIEUCH YE MIEUM")
    (17171 . "HANGUL SYLLABLE CHIEUCH YE PIEUP")
    (17172 . "HANGUL SYLLABLE CHIEUCH YE SIOS")
    (17173 . "HANGUL SYLLABLE CHIEUCH O TIKEUT")
    (17174 . "HANGUL SYLLABLE CHIEUCH WA KIYEOK")
    (17175 . "HANGUL SYLLABLE CHIEUCH WA TIKEUT")
    (17176 . "HANGUL SYLLABLE CHIEUCH WA MIEUM")
    (17177 . "HANGUL SYLLABLE CHIEUCH WA PIEUP")
    (17178 . "HANGUL SYLLABLE CHIEUCH WA SIOS")
    (17179 . "HANGUL SYLLABLE CHIEUCH WAE")
    (17180 . "HANGUL SYLLABLE CHIEUCH WAE SIOS")
    (17181 . "HANGUL SYLLABLE CHIEUCH WAE SSANGSIOS")
    (17182 . "HANGUL SYLLABLE CHIEUCH OE KIYEOK")
    (17183 . "HANGUL SYLLABLE CHIEUCH OE TIKEUT")
    (17184 . "HANGUL SYLLABLE CHIEUCH OE SSANGSIOS")
    (17185 . "HANGUL SYLLABLE CHIEUCH YO KIYEOK")
    (17186 . "HANGUL SYLLABLE CHIEUCH YO NIEUN")
    (17187 . "HANGUL SYLLABLE CHIEUCH YO RIEUL")
    (17188 . "HANGUL SYLLABLE CHIEUCH YO PIEUP")
    (17189 . "HANGUL SYLLABLE CHIEUCH YO SIOS")
    (17190 . "HANGUL SYLLABLE CHIEUCH YO IEUNG")
    (17191 . "HANGUL SYLLABLE CHIEUCH U TIKEUT")
    (17192 . "HANGUL SYLLABLE CHIEUCH WEO KIYEOK")
    (17193 . "HANGUL SYLLABLE CHIEUCH WEO NIEUN")
    (17194 . "HANGUL SYLLABLE CHIEUCH WEO TIKEUT")
    (17195 . "HANGUL SYLLABLE CHIEUCH WEO RIEUL")
    (17196 . "HANGUL SYLLABLE CHIEUCH WEO MIEUM")
    (17197 . "HANGUL SYLLABLE CHIEUCH WEO PIEUP")
    (17198 . "HANGUL SYLLABLE CHIEUCH WEO SIOS")
    (17199 . "HANGUL SYLLABLE CHIEUCH WEO IEUNG")
    (17200 . "HANGUL SYLLABLE CHIEUCH WE KIYEOK")
    (17201 . "HANGUL SYLLABLE CHIEUCH WE RIEUL")
    (17202 . "HANGUL SYLLABLE CHIEUCH WE MIEUM")
    (17203 . "HANGUL SYLLABLE CHIEUCH WE PIEUP")
    (17204 . "HANGUL SYLLABLE CHIEUCH WE SIOS")
    (17205 . "HANGUL SYLLABLE CHIEUCH WE SSANGSIOS")
    (17206 . "HANGUL SYLLABLE CHIEUCH WE IEUNG")
    (17207 . "HANGUL SYLLABLE CHIEUCH WI KIYEOK")
    (17208 . "HANGUL SYLLABLE CHIEUCH WI TIKEUT")
    (17209 . "HANGUL SYLLABLE CHIEUCH YU KIYEOK")
    (17210 . "HANGUL SYLLABLE CHIEUCH YU PIEUP")
    (17211 . "HANGUL SYLLABLE CHIEUCH YU SIOS")
    (17212 . "HANGUL SYLLABLE CHIEUCH EU TIKEUT")
    (17213 . "HANGUL SYLLABLE CHIEUCH EU RIEUL-KIYEOK")
    (17214 . "HANGUL SYLLABLE CHIEUCH EU RIEUL-SIOS")
    (17215 . "HANGUL SYLLABLE CHIEUCH YI")
    (17216 . "HANGUL SYLLABLE CHIEUCH YI KIYEOK")
    (17217 . "HANGUL SYLLABLE CHIEUCH YI NIEUN")
    (17218 . "HANGUL SYLLABLE CHIEUCH YI RIEUL")
    (17219 . "HANGUL SYLLABLE CHIEUCH YI MIEUM")
    (17220 . "HANGUL SYLLABLE CHIEUCH YI PIEUP")
    (17221 . "HANGUL SYLLABLE CHIEUCH YI SIOS")
    (17222 . "HANGUL SYLLABLE CHIEUCH YI IEUNG")
    (17223 . "HANGUL SYLLABLE CHIEUCH I RIEUL-MIEUM")
    (17224 . "HANGUL SYLLABLE CHIEUCH I SSANGSIOS")
    (17225 . "HANGUL SYLLABLE CHIEUCH I CIEUC")
    (17226 . "HANGUL SYLLABLE CHIEUCH I KHIEUKH")
    (17227 . "HANGUL SYLLABLE CHIEUCH I PHIEUPH")
    (17228 . "HANGUL SYLLABLE KHIEUKH A TIKEUT")
    (17229 . "HANGUL SYLLABLE KHIEUKH A SSANGSIOS")
    (17230 . "HANGUL SYLLABLE KHIEUKH A THIEUTH")
    (17231 . "HANGUL SYLLABLE KHIEUKH A PHIEUPH")
    (17232 . "HANGUL SYLLABLE KHIEUKH AE TIKEUT")
    (17233 . "HANGUL SYLLABLE KHIEUKH AE THIEUTH")
    (17234 . "HANGUL SYLLABLE KHIEUKH AE PHIEUPH")
    (17235 . "HANGUL SYLLABLE KHIEUKH YA NIEUN")
    (17236 . "HANGUL SYLLABLE KHIEUKH YA RIEUL")
    (17237 . "HANGUL SYLLABLE KHIEUKH YA MIEUM")
    (17238 . "HANGUL SYLLABLE KHIEUKH YA PIEUP")
    (17239 . "HANGUL SYLLABLE KHIEUKH YA SIOS")
    (17240 . "HANGUL SYLLABLE KHIEUKH YAE")
    (17241 . "HANGUL SYLLABLE KHIEUKH YAE NIEUN")
    (17242 . "HANGUL SYLLABLE KHIEUKH EO CHIEUCH")
    (17243 . "HANGUL SYLLABLE KHIEUKH EO THIEUTH")
    (17244 . "HANGUL SYLLABLE KHIEUKH EO PHIEUPH")
    (17245 . "HANGUL SYLLABLE KHIEUKH E TIKEUT")
    (17246 . "HANGUL SYLLABLE KHIEUKH E SSANGSIOS")
    (17247 . "HANGUL SYLLABLE KHIEUKH E THIEUTH")
    (17248 . "HANGUL SYLLABLE KHIEUKH E PHIEUPH")
    (17249 . "HANGUL SYLLABLE KHIEUKH YEO KIYEOK")
    (17250 . "HANGUL SYLLABLE KHIEUKH YEO TIKEUT")
    (17251 . "HANGUL SYLLABLE KHIEUKH YE KIYEOK")
    (17252 . "HANGUL SYLLABLE KHIEUKH YE NIEUN")
    (17253 . "HANGUL SYLLABLE KHIEUKH YE RIEUL")
    (17254 . "HANGUL SYLLABLE KHIEUKH YE MIEUM")
    (17255 . "HANGUL SYLLABLE KHIEUKH YE PIEUP")
    (17256 . "HANGUL SYLLABLE KHIEUKH YE SIOS")
    (17257 . "HANGUL SYLLABLE KHIEUKH YE SSANGSIOS")
    (17258 . "HANGUL SYLLABLE KHIEUKH YE IEUNG")
    (17259 . "HANGUL SYLLABLE KHIEUKH O TIKEUT")
    (17260 . "HANGUL SYLLABLE KHIEUKH O CHIEUCH")
    (17261 . "HANGUL SYLLABLE KHIEUKH O PHIEUPH")
    (17262 . "HANGUL SYLLABLE KHIEUKH WA PIEUP")
    (17263 . "HANGUL SYLLABLE KHIEUKH WA SIOS")
    (17264 . "HANGUL SYLLABLE KHIEUKH WAE KIYEOK")
    (17265 . "HANGUL SYLLABLE KHIEUKH WAE NIEUN")
    (17266 . "HANGUL SYLLABLE KHIEUKH WAE TIKEUT")
    (17267 . "HANGUL SYLLABLE KHIEUKH WAE RIEUL")
    (17268 . "HANGUL SYLLABLE KHIEUKH WAE MIEUM")
    (17269 . "HANGUL SYLLABLE KHIEUKH WAE PIEUP")
    (17270 . "HANGUL SYLLABLE KHIEUKH WAE SIOS")
    (17271 . "HANGUL SYLLABLE KHIEUKH OE KIYEOK")
    (17272 . "HANGUL SYLLABLE KHIEUKH OE NIEUN")
    (17273 . "HANGUL SYLLABLE KHIEUKH OE MIEUM")
    (17274 . "HANGUL SYLLABLE KHIEUKH OE PIEUP")
    (17275 . "HANGUL SYLLABLE KHIEUKH OE SIOS")
    (17276 . "HANGUL SYLLABLE KHIEUKH OE IEUNG")
    (17277 . "HANGUL SYLLABLE KHIEUKH YO KIYEOK")
    (17278 . "HANGUL SYLLABLE KHIEUKH YO NIEUN")
    (17279 . "HANGUL SYLLABLE KHIEUKH YO RIEUL")
    (17280 . "HANGUL SYLLABLE KHIEUKH YO MIEUM")
    (17281 . "HANGUL SYLLABLE KHIEUKH YO PIEUP")
    (17282 . "HANGUL SYLLABLE KHIEUKH YO SIOS")
    (17283 . "HANGUL SYLLABLE KHIEUKH YO IEUNG")
    (17284 . "HANGUL SYLLABLE KHIEUKH U TIKEUT")
    (17285 . "HANGUL SYLLABLE KHIEUKH U PHIEUPH")
    (17286 . "HANGUL SYLLABLE KHIEUKH WEO KIYEOK")
    (17287 . "HANGUL SYLLABLE KHIEUKH WEO TIKEUT")
    (17288 . "HANGUL SYLLABLE KHIEUKH WEO MIEUM")
    (17289 . "HANGUL SYLLABLE KHIEUKH WEO PIEUP")
    (17290 . "HANGUL SYLLABLE KHIEUKH WEO SSANGSIOS")
    (17291 . "HANGUL SYLLABLE KHIEUKH WE KIYEOK")
    (17292 . "HANGUL SYLLABLE KHIEUKH WE NIEUN")
    (17293 . "HANGUL SYLLABLE KHIEUKH WE RIEUL")
    (17294 . "HANGUL SYLLABLE KHIEUKH WE MIEUM")
    (17295 . "HANGUL SYLLABLE KHIEUKH WE PIEUP")
    (17296 . "HANGUL SYLLABLE KHIEUKH WE SIOS")
    (17297 . "HANGUL SYLLABLE KHIEUKH WE SSANGSIOS")
    (17298 . "HANGUL SYLLABLE KHIEUKH WI TIKEUT")
    (17299 . "HANGUL SYLLABLE KHIEUKH YU KIYEOK")
    (17300 . "HANGUL SYLLABLE KHIEUKH YU PIEUP")
    (17301 . "HANGUL SYLLABLE KHIEUKH YU SIOS")
    (17302 . "HANGUL SYLLABLE KHIEUKH YU IEUNG")
    (17303 . "HANGUL SYLLABLE KHIEUKH YU KHIEUKH")
    (17304 . "HANGUL SYLLABLE KHIEUKH EU NIEUN-HIEUH")
    (17305 . "HANGUL SYLLABLE KHIEUKH EU RIEUL-MIEUM")
    (17306 . "HANGUL SYLLABLE KHIEUKH EU SIOS")
    (17307 . "HANGUL SYLLABLE KHIEUKH YI")
    (17308 . "HANGUL SYLLABLE KHIEUKH YI KIYEOK")
    (17309 . "HANGUL SYLLABLE KHIEUKH YI NIEUN")
    (17310 . "HANGUL SYLLABLE KHIEUKH YI RIEUL")
    (17311 . "HANGUL SYLLABLE KHIEUKH YI MIEUM")
    (17312 . "HANGUL SYLLABLE KHIEUKH YI PIEUP")
    (17313 . "HANGUL SYLLABLE KHIEUKH YI SIOS")
    (17314 . "HANGUL SYLLABLE KHIEUKH YI IEUNG")
    (17315 . "HANGUL SYLLABLE KHIEUKH I TIKEUT")
    (17316 . "HANGUL SYLLABLE KHIEUKH I SSANGSIOS")
    (17317 . "HANGUL SYLLABLE KHIEUKH I PHIEUPH")
    (17318 . "HANGUL SYLLABLE THIEUTH A TIKEUT")
    (17319 . "HANGUL SYLLABLE THIEUTH A RIEUL-MIEUM")
    (17320 . "HANGUL SYLLABLE THIEUTH A RIEUL-HIEUH")
    (17321 . "HANGUL SYLLABLE THIEUTH A PHIEUPH")
    (17322 . "HANGUL SYLLABLE THIEUTH AE NIEUN-HIEUH")
    (17323 . "HANGUL SYLLABLE THIEUTH AE TIKEUT")
    (17324 . "HANGUL SYLLABLE THIEUTH AE RIEUL-KIYEOK")
    (17325 . "HANGUL SYLLABLE THIEUTH AE PHIEUPH")
    (17326 . "HANGUL SYLLABLE THIEUTH YA KIYEOK")
    (17327 . "HANGUL SYLLABLE THIEUTH YA NIEUN")
    (17328 . "HANGUL SYLLABLE THIEUTH YA RIEUL")
    (17329 . "HANGUL SYLLABLE THIEUTH YA MIEUM")
    (17330 . "HANGUL SYLLABLE THIEUTH YA PIEUP")
    (17331 . "HANGUL SYLLABLE THIEUTH YA SIOS")
    (17332 . "HANGUL SYLLABLE THIEUTH YAE")
    (17333 . "HANGUL SYLLABLE THIEUTH YAE NIEUN")
    (17334 . "HANGUL SYLLABLE THIEUTH YAE NIEUN-HIEUH")
    (17335 . "HANGUL SYLLABLE THIEUTH EO TIKEUT")
    (17336 . "HANGUL SYLLABLE THIEUTH EO RIEUL-KIYEOK")
    (17337 . "HANGUL SYLLABLE THIEUTH EO RIEUL-PIEUP")
    (17338 . "HANGUL SYLLABLE THIEUTH E TIKEUT")
    (17339 . "HANGUL SYLLABLE THIEUTH E SSANGSIOS")
    (17340 . "HANGUL SYLLABLE THIEUTH E PHIEUPH")
    (17341 . "HANGUL SYLLABLE THIEUTH YEO KIYEOK")
    (17342 . "HANGUL SYLLABLE THIEUTH YEO TIKEUT")
    (17343 . "HANGUL SYLLABLE THIEUTH YEO RIEUL")
    (17344 . "HANGUL SYLLABLE THIEUTH YEO MIEUM")
    (17345 . "HANGUL SYLLABLE THIEUTH YEO PIEUP")
    (17346 . "HANGUL SYLLABLE THIEUTH YEO SIOS")
    (17347 . "HANGUL SYLLABLE THIEUTH YEO IEUNG")
    (17348 . "HANGUL SYLLABLE THIEUTH YE KIYEOK")
    (17349 . "HANGUL SYLLABLE THIEUTH YE RIEUL")
    (17350 . "HANGUL SYLLABLE THIEUTH YE MIEUM")
    (17351 . "HANGUL SYLLABLE THIEUTH YE SIOS")
    (17352 . "HANGUL SYLLABLE THIEUTH YE IEUNG")
    (17353 . "HANGUL SYLLABLE THIEUTH O TIKEUT")
    (17354 . "HANGUL SYLLABLE THIEUTH O CHIEUCH")
    (17355 . "HANGUL SYLLABLE THIEUTH O KHIEUKH")
    (17356 . "HANGUL SYLLABLE THIEUTH O THIEUTH")
    (17357 . "HANGUL SYLLABLE THIEUTH WA KIYEOK")
    (17358 . "HANGUL SYLLABLE THIEUTH WA RIEUL")
    (17359 . "HANGUL SYLLABLE THIEUTH WA MIEUM")
    (17360 . "HANGUL SYLLABLE THIEUTH WA PIEUP")
    (17361 . "HANGUL SYLLABLE THIEUTH WA SIOS")
    (17362 . "HANGUL SYLLABLE THIEUTH WA IEUNG")
    (17363 . "HANGUL SYLLABLE THIEUTH WAE SSANGSIOS")
    (17364 . "HANGUL SYLLABLE THIEUTH WAE IEUNG")
    (17365 . "HANGUL SYLLABLE THIEUTH OE KIYEOK")
    (17366 . "HANGUL SYLLABLE THIEUTH OE TIKEUT")
    (17367 . "HANGUL SYLLABLE THIEUTH OE RIEUL")
    (17368 . "HANGUL SYLLABLE THIEUTH OE MIEUM")
    (17369 . "HANGUL SYLLABLE THIEUTH OE PIEUP")
    (17370 . "HANGUL SYLLABLE THIEUTH OE SSANGSIOS")
    (17371 . "HANGUL SYLLABLE THIEUTH YO KIYEOK")
    (17372 . "HANGUL SYLLABLE THIEUTH YO NIEUN")
    (17373 . "HANGUL SYLLABLE THIEUTH YO RIEUL")
    (17374 . "HANGUL SYLLABLE THIEUTH YO MIEUM")
    (17375 . "HANGUL SYLLABLE THIEUTH YO PIEUP")
    (17376 . "HANGUL SYLLABLE THIEUTH YO SIOS")
    (17377 . "HANGUL SYLLABLE THIEUTH YO IEUNG")
    (17378 . "HANGUL SYLLABLE THIEUTH U TIKEUT")
    (17379 . "HANGUL SYLLABLE THIEUTH U RIEUL-MIEUM")
    (17380 . "HANGUL SYLLABLE THIEUTH WEO KIYEOK")
    (17381 . "HANGUL SYLLABLE THIEUTH WEO NIEUN")
    (17382 . "HANGUL SYLLABLE THIEUTH WEO TIKEUT")
    (17383 . "HANGUL SYLLABLE THIEUTH WEO RIEUL")
    (17384 . "HANGUL SYLLABLE THIEUTH WEO RIEUL-SIOS")
    (17385 . "HANGUL SYLLABLE THIEUTH WEO MIEUM")
    (17386 . "HANGUL SYLLABLE THIEUTH WEO PIEUP")
    (17387 . "HANGUL SYLLABLE THIEUTH WEO SIOS")
    (17388 . "HANGUL SYLLABLE THIEUTH WEO IEUNG")
    (17389 . "HANGUL SYLLABLE THIEUTH WE KIYEOK")
    (17390 . "HANGUL SYLLABLE THIEUTH WE NIEUN")
    (17391 . "HANGUL SYLLABLE THIEUTH WE TIKEUT")
    (17392 . "HANGUL SYLLABLE THIEUTH WE RIEUL")
    (17393 . "HANGUL SYLLABLE THIEUTH WE MIEUM")
    (17394 . "HANGUL SYLLABLE THIEUTH WE PIEUP")
    (17395 . "HANGUL SYLLABLE THIEUTH WE SIOS")
    (17396 . "HANGUL SYLLABLE THIEUTH WE SSANGSIOS")
    (17397 . "HANGUL SYLLABLE THIEUTH WE IEUNG")
    (17398 . "HANGUL SYLLABLE THIEUTH WI TIKEUT")
    (17399 . "HANGUL SYLLABLE THIEUTH WI SIOS")
    (17400 . "HANGUL SYLLABLE THIEUTH YU KIYEOK")
    (17401 . "HANGUL SYLLABLE THIEUTH YU PIEUP")
    (17402 . "HANGUL SYLLABLE THIEUTH YU SIOS")
    (17403 . "HANGUL SYLLABLE THIEUTH EU KIYEOK-SIOS")
    (17404 . "HANGUL SYLLABLE THIEUTH EU RIEUL-KIYEOK")
    (17405 . "HANGUL SYLLABLE THIEUTH EU RIEUL-SIOS")
    (17406 . "HANGUL SYLLABLE THIEUTH EU IEUNG")
    (17407 . "HANGUL SYLLABLE THIEUTH YI KIYEOK")
    (17408 . "HANGUL SYLLABLE THIEUTH YI RIEUL-SIOS")
    (17409 . "HANGUL SYLLABLE THIEUTH YI SIOS")
    (17410 . "HANGUL SYLLABLE THIEUTH YI IEUNG")
    (17411 . "HANGUL SYLLABLE THIEUTH I TIKEUT")
    (17412 . "HANGUL SYLLABLE THIEUTH I KHIEUKH")
    (17413 . "HANGUL SYLLABLE THIEUTH I PHIEUPH")
    (17414 . "HANGUL SYLLABLE PHIEUPH A TIKEUT")
    (17415 . "HANGUL SYLLABLE PHIEUPH A CHIEUCH")
    (17416 . "HANGUL SYLLABLE PHIEUPH A PHIEUPH")
    (17417 . "HANGUL SYLLABLE PHIEUPH AE TIKEUT")
    (17418 . "HANGUL SYLLABLE PHIEUPH AE THIEUTH")
    (17419 . "HANGUL SYLLABLE PHIEUPH YA NIEUN")
    (17420 . "HANGUL SYLLABLE PHIEUPH YA RIEUL")
    (17421 . "HANGUL SYLLABLE PHIEUPH YA MIEUM")
    (17422 . "HANGUL SYLLABLE PHIEUPH YA PIEUP")
    (17423 . "HANGUL SYLLABLE PHIEUPH YA SIOS")
    (17424 . "HANGUL SYLLABLE PHIEUPH YA IEUNG")
    (17425 . "HANGUL SYLLABLE PHIEUPH YAE")
    (17426 . "HANGUL SYLLABLE PHIEUPH EO TIKEUT")
    (17427 . "HANGUL SYLLABLE PHIEUPH EO PIEUP-SIOS")
    (17428 . "HANGUL SYLLABLE PHIEUPH EO THIEUTH")
    (17429 . "HANGUL SYLLABLE PHIEUPH E TIKEUT")
    (17430 . "HANGUL SYLLABLE PHIEUPH E SSANGSIOS")
    (17431 . "HANGUL SYLLABLE PHIEUPH E PHIEUPH")
    (17432 . "HANGUL SYLLABLE PHIEUPH YEO KIYEOK")
    (17433 . "HANGUL SYLLABLE PHIEUPH YEO NIEUN-CIEUC")
    (17434 . "HANGUL SYLLABLE PHIEUPH YEO TIKEUT")
    (17435 . "HANGUL SYLLABLE PHIEUPH YEO SIOS")
    (17436 . "HANGUL SYLLABLE PHIEUPH YE KIYEOK")
    (17437 . "HANGUL SYLLABLE PHIEUPH YE NIEUN")
    (17438 . "HANGUL SYLLABLE PHIEUPH YE TIKEUT")
    (17439 . "HANGUL SYLLABLE PHIEUPH YE MIEUM")
    (17440 . "HANGUL SYLLABLE PHIEUPH YE IEUNG")
    (17441 . "HANGUL SYLLABLE PHIEUPH O SSANGKIYEOK")
    (17442 . "HANGUL SYLLABLE PHIEUPH O TIKEUT")
    (17443 . "HANGUL SYLLABLE PHIEUPH O KHIEUKH")
    (17444 . "HANGUL SYLLABLE PHIEUPH O THIEUTH")
    (17445 . "HANGUL SYLLABLE PHIEUPH O PHIEUPH")
    (17446 . "HANGUL SYLLABLE PHIEUPH WA KIYEOK")
    (17447 . "HANGUL SYLLABLE PHIEUPH WA NIEUN")
    (17448 . "HANGUL SYLLABLE PHIEUPH WA RIEUL")
    (17449 . "HANGUL SYLLABLE PHIEUPH WA MIEUM")
    (17450 . "HANGUL SYLLABLE PHIEUPH WA PIEUP")
    (17451 . "HANGUL SYLLABLE PHIEUPH WA SIOS")
    (17452 . "HANGUL SYLLABLE PHIEUPH WAE")
    (17453 . "HANGUL SYLLABLE PHIEUPH OE KIYEOK")
    (17454 . "HANGUL SYLLABLE PHIEUPH OE RIEUL")
    (17455 . "HANGUL SYLLABLE PHIEUPH OE MIEUM")
    (17456 . "HANGUL SYLLABLE PHIEUPH OE PIEUP")
    (17457 . "HANGUL SYLLABLE PHIEUPH OE SIOS")
    (17458 . "HANGUL SYLLABLE PHIEUPH OE IEUNG")
    (17459 . "HANGUL SYLLABLE PHIEUPH YO KIYEOK")
    (17460 . "HANGUL SYLLABLE PHIEUPH YO TIKEUT")
    (17461 . "HANGUL SYLLABLE PHIEUPH YO MIEUM")
    (17462 . "HANGUL SYLLABLE PHIEUPH YO IEUNG")
    (17463 . "HANGUL SYLLABLE PHIEUPH WEO KIYEOK")
    (17464 . "HANGUL SYLLABLE PHIEUPH WEO NIEUN")
    (17465 . "HANGUL SYLLABLE PHIEUPH WEO TIKEUT")
    (17466 . "HANGUL SYLLABLE PHIEUPH WEO RIEUL")
    (17467 . "HANGUL SYLLABLE PHIEUPH WEO MIEUM")
    (17468 . "HANGUL SYLLABLE PHIEUPH WEO PIEUP")
    (17469 . "HANGUL SYLLABLE PHIEUPH WEO SIOS")
    (17470 . "HANGUL SYLLABLE PHIEUPH WEO SSANGSIOS")
    (17471 . "HANGUL SYLLABLE PHIEUPH WE")
    (17472 . "HANGUL SYLLABLE PHIEUPH WE KIYEOK")
    (17473 . "HANGUL SYLLABLE PHIEUPH WE NIEUN")
    (17474 . "HANGUL SYLLABLE PHIEUPH WE RIEUL")
    (17475 . "HANGUL SYLLABLE PHIEUPH WE MIEUM")
    (17476 . "HANGUL SYLLABLE PHIEUPH WE PIEUP")
    (17477 . "HANGUL SYLLABLE PHIEUPH WE SIOS")
    (17478 . "HANGUL SYLLABLE PHIEUPH WE IEUNG")
    (17479 . "HANGUL SYLLABLE PHIEUPH WI KIYEOK")
    (17480 . "HANGUL SYLLABLE PHIEUPH WI PIEUP")
    (17481 . "HANGUL SYLLABLE PHIEUPH WI IEUNG")
    (17482 . "HANGUL SYLLABLE PHIEUPH YU KIYEOK")
    (17483 . "HANGUL SYLLABLE PHIEUPH YU TIKEUT")
    (17484 . "HANGUL SYLLABLE PHIEUPH YU PIEUP")
    (17485 . "HANGUL SYLLABLE PHIEUPH EU KIYEOK")
    (17486 . "HANGUL SYLLABLE PHIEUPH EU TIKEUT")
    (17487 . "HANGUL SYLLABLE PHIEUPH EU RIEUL-SIOS")
    (17488 . "HANGUL SYLLABLE PHIEUPH EU IEUNG")
    (17489 . "HANGUL SYLLABLE PHIEUPH YI")
    (17490 . "HANGUL SYLLABLE PHIEUPH YI KIYEOK")
    (17491 . "HANGUL SYLLABLE PHIEUPH YI NIEUN")
    (17492 . "HANGUL SYLLABLE PHIEUPH YI RIEUL")
    (17493 . "HANGUL SYLLABLE PHIEUPH YI MIEUM")
    (17494 . "HANGUL SYLLABLE PHIEUPH YI PIEUP")
    (17495 . "HANGUL SYLLABLE PHIEUPH YI SIOS")
    (17496 . "HANGUL SYLLABLE PHIEUPH YI IEUNG")
    (17497 . "HANGUL SYLLABLE PHIEUPH I TIKEUT")
    (17498 . "HANGUL SYLLABLE PHIEUPH I THIEUTH")
    (17499 . "HANGUL SYLLABLE PHIEUPH I PHIEUPH")
    (17500 . "HANGUL SYLLABLE HIEUH A TIKEUT")
    (17501 . "HANGUL SYLLABLE HIEUH A RIEUL-KIYEOK")
    (17502 . "HANGUL SYLLABLE HIEUH A RIEUL-PIEUP")
    (17503 . "HANGUL SYLLABLE HIEUH A RIEUL-SIOS")
    (17504 . "HANGUL SYLLABLE HIEUH A RIEUL-HIEUH")
    (17505 . "HANGUL SYLLABLE HIEUH A SSANGSIOS")
    (17506 . "HANGUL SYLLABLE HIEUH A CIEUC")
    (17507 . "HANGUL SYLLABLE HIEUH A THIEUTH")
    (17508 . "HANGUL SYLLABLE HIEUH AE TIKEUT")
    (17509 . "HANGUL SYLLABLE HIEUH AE THIEUTH")
    (17510 . "HANGUL SYLLABLE HIEUH YA KIYEOK")
    (17511 . "HANGUL SYLLABLE HIEUH YA NIEUN")
    (17512 . "HANGUL SYLLABLE HIEUH YA RIEUL")
    (17513 . "HANGUL SYLLABLE HIEUH YA MIEUM")
    (17514 . "HANGUL SYLLABLE HIEUH YA PIEUP")
    (17515 . "HANGUL SYLLABLE HIEUH YA SIOS")
    (17516 . "HANGUL SYLLABLE HIEUH YA HIEUH")
    (17517 . "HANGUL SYLLABLE HIEUH YAE")
    (17518 . "HANGUL SYLLABLE HIEUH YAE SSANGSIOS")
    (17519 . "HANGUL SYLLABLE HIEUH EO TIKEUT")
    (17520 . "HANGUL SYLLABLE HIEUH EO RIEUL-KIYEOK")
    (17521 . "HANGUL SYLLABLE HIEUH EO RIEUL-HIEUH")
    (17522 . "HANGUL SYLLABLE HIEUH EO SSANGSIOS")
    (17523 . "HANGUL SYLLABLE HIEUH EO THIEUTH")
    (17524 . "HANGUL SYLLABLE HIEUH EO PHIEUPH")
    (17525 . "HANGUL SYLLABLE HIEUH E TIKEUT")
    (17526 . "HANGUL SYLLABLE HIEUH E SSANGSIOS")
    (17527 . "HANGUL SYLLABLE HIEUH E PHIEUPH")
    (17528 . "HANGUL SYLLABLE HIEUH YEO TIKEUT")
    (17529 . "HANGUL SYLLABLE HIEUH YE KIYEOK")
    (17530 . "HANGUL SYLLABLE HIEUH YE MIEUM")
    (17531 . "HANGUL SYLLABLE HIEUH YE SIOS")
    (17532 . "HANGUL SYLLABLE HIEUH YE SSANGSIOS")
    (17533 . "HANGUL SYLLABLE HIEUH YE IEUNG")
    (17534 . "HANGUL SYLLABLE HIEUH O NIEUN-CIEUC")
    (17535 . "HANGUL SYLLABLE HIEUH O TIKEUT")
    (17536 . "HANGUL SYLLABLE HIEUH O RIEUL-KIYEOK")
    (17537 . "HANGUL SYLLABLE HIEUH O RIEUL-SIOS")
    (17538 . "HANGUL SYLLABLE HIEUH O RIEUL-HIEUH")
    (17539 . "HANGUL SYLLABLE HIEUH WA TIKEUT")
    (17540 . "HANGUL SYLLABLE HIEUH WA RIEUL-SIOS")
    (17541 . "HANGUL SYLLABLE HIEUH WA MIEUM")
    (17542 . "HANGUL SYLLABLE HIEUH WA PIEUP")
    (17543 . "HANGUL SYLLABLE HIEUH WA PIEUP-SIOS")
    (17544 . "HANGUL SYLLABLE HIEUH WA SSANGSIOS")
    (17545 . "HANGUL SYLLABLE HIEUH WAE TIKEUT")
    (17546 . "HANGUL SYLLABLE HIEUH WAE RIEUL")
    (17547 . "HANGUL SYLLABLE HIEUH WAE MIEUM")
    (17548 . "HANGUL SYLLABLE HIEUH WAE PIEUP")
    (17549 . "HANGUL SYLLABLE HIEUH WAE SSANGSIOS")
    (17550 . "HANGUL SYLLABLE HIEUH OE TIKEUT")
    (17551 . "HANGUL SYLLABLE HIEUH OE MIEUM")
    (17552 . "HANGUL SYLLABLE HIEUH YO KIYEOK")
    (17553 . "HANGUL SYLLABLE HIEUH YO TIKEUT")
    (17554 . "HANGUL SYLLABLE HIEUH YO MIEUM")
    (17555 . "HANGUL SYLLABLE HIEUH YO IEUNG")
    (17556 . "HANGUL SYLLABLE HIEUH U TIKEUT")
    (17557 . "HANGUL SYLLABLE HIEUH U RIEUL-KIYEOK")
    (17558 . "HANGUL SYLLABLE HIEUH U RIEUL-MIEUM")
    (17559 . "HANGUL SYLLABLE HIEUH U RIEUL-PHIEUPH")
    (17560 . "HANGUL SYLLABLE HIEUH U RIEUL-HIEUH")
    (17561 . "HANGUL SYLLABLE HIEUH U PIEUP")
    (17562 . "HANGUL SYLLABLE HIEUH U KHIEUKH")
    (17563 . "HANGUL SYLLABLE HIEUH U THIEUTH")
    (17564 . "HANGUL SYLLABLE HIEUH WEO KIYEOK")
    (17565 . "HANGUL SYLLABLE HIEUH WEO PIEUP")
    (17566 . "HANGUL SYLLABLE HIEUH WEO SIOS")
    (17567 . "HANGUL SYLLABLE HIEUH WEO SSANGSIOS")
    (17568 . "HANGUL SYLLABLE HIEUH WE TIKEUT")
    (17569 . "HANGUL SYLLABLE HIEUH WE MIEUM")
    (17570 . "HANGUL SYLLABLE HIEUH WE PIEUP")
    (17571 . "HANGUL SYLLABLE HIEUH WE SIOS")
    (17572 . "HANGUL SYLLABLE HIEUH WE SSANGSIOS")
    (17573 . "HANGUL SYLLABLE HIEUH WI TIKEUT")
    (17574 . "HANGUL SYLLABLE HIEUH YU NIEUN-HIEUH")
    (17575 . "HANGUL SYLLABLE HIEUH YU TIKEUT")
    (17576 . "HANGUL SYLLABLE HIEUH YU RIEUL-HIEUH")
    (17577 . "HANGUL SYLLABLE HIEUH YU PIEUP")
    (17578 . "HANGUL SYLLABLE HIEUH YU THIEUTH")
    (17579 . "HANGUL SYLLABLE HIEUH EU NIEUN-CIEUC")
    (17580 . "HANGUL SYLLABLE HIEUH EU RIEUL-THIEUTH")
    (17581 . "HANGUL SYLLABLE HIEUH EU RIEUL-HIEUH")
    (17582 . "HANGUL SYLLABLE HIEUH EU KHIEUKH")
    (17583 . "HANGUL SYLLABLE HIEUH EU PHIEUPH")
    (17584 . "HANGUL SYLLABLE HIEUH EU HIEUH")
    (17585 . "HANGUL SYLLABLE HIEUH YI KIYEOK")
    (17586 . "HANGUL SYLLABLE HIEUH YI TIKEUT")
    (17587 . "HANGUL SYLLABLE HIEUH YI SIOS")
    (17588 . "HANGUL SYLLABLE HIEUH I TIKEUT")
    (17589 . "HANGUL SYLLABLE HIEUH I RIEUL-HIEUH")
    (17590 . "HANGUL SYLLABLE HIEUH I SSANGSIOS")
    (17591 . "HANGUL SYLLABLE HIEUH I KHIEUKH")
    (17592 . "HANGUL SYLLABLE MIEUM EU KIYEOK-SIOS")
    (17593 . "HANGUL SYLLABLE HIEUH O KIYEOK-SIOS")
    (17594 . "HANGUL SYLLABLE SSANGPIEUP O THIEUTH")
    (17595 . "HANGUL SYLLABLE SSANGPIEUP U THIEUTH")
    (17596 . "HANGUL SYLLABLE SSANGCIEUC WE SSANGSIOS")
    (17597 . "HANGUL SYLLABLE IEUNG WEO RIEUL-SIOS")
    (17598 . "HANGUL SYLLABLE KIYEOK A KIYEOK-SIOS")
    (17599 . "HANGUL SYLLABLE KIYEOK A NIEUN-CIEUC")
    (17600 . "HANGUL SYLLABLE KIYEOK A NIEUN-HIEUH")
    (17601 . "HANGUL SYLLABLE KIYEOK A RIEUL-THIEUTH")
    (17602 . "HANGUL SYLLABLE KIYEOK A RIEUL-PHIEUPH")
    (17603 . "HANGUL SYLLABLE KIYEOK A RIEUL-HIEUH")
    (17604 . "HANGUL SYLLABLE KIYEOK A KHIEUKH")
    (17605 . "HANGUL SYLLABLE KIYEOK AE SSANGKIYEOK")
    (17606 . "HANGUL SYLLABLE KIYEOK AE KIYEOK-SIOS")
    (17607 . "HANGUL SYLLABLE KIYEOK AE NIEUN-CIEUC")
    (17608 . "HANGUL SYLLABLE KIYEOK AE RIEUL-KIYEOK")
    (17609 . "HANGUL SYLLABLE KIYEOK AE RIEUL-MIEUM")
    (17610 . "HANGUL SYLLABLE KIYEOK AE RIEUL-PIEUP")
    (17611 . "HANGUL SYLLABLE KIYEOK AE RIEUL-SIOS")
    (17612 . "HANGUL SYLLABLE KIYEOK AE RIEUL-THIEUTH")
    (17613 . "HANGUL SYLLABLE KIYEOK AE RIEUL-PHIEUPH")
    (17614 . "HANGUL SYLLABLE KIYEOK AE RIEUL-HIEUH")
    (17615 . "HANGUL SYLLABLE KIYEOK AE PIEUP-SIOS")
    (17616 . "HANGUL SYLLABLE KIYEOK AE CHIEUCH")
    (17617 . "HANGUL SYLLABLE KIYEOK AE KHIEUKH")
    (17618 . "HANGUL SYLLABLE KIYEOK AE HIEUH")
    (17619 . "HANGUL SYLLABLE KIYEOK YA SSANGKIYEOK")
    (17620 . "HANGUL SYLLABLE KIYEOK YA KIYEOK-SIOS")
    (17621 . "HANGUL SYLLABLE KIYEOK YA NIEUN-CIEUC")
    (17622 . "HANGUL SYLLABLE KIYEOK YA NIEUN-HIEUH")
    (17623 . "HANGUL SYLLABLE KIYEOK YA RIEUL-MIEUM")
    (17624 . "HANGUL SYLLABLE KIYEOK YA RIEUL-PIEUP")
    (17625 . "HANGUL SYLLABLE KIYEOK YA RIEUL-SIOS")
    (17626 . "HANGUL SYLLABLE KIYEOK YA RIEUL-THIEUTH")
    (17627 . "HANGUL SYLLABLE KIYEOK YA RIEUL-PHIEUPH")
    (17628 . "HANGUL SYLLABLE KIYEOK YA PIEUP-SIOS")
    (17629 . "HANGUL SYLLABLE KIYEOK YA CIEUC")
    (17630 . "HANGUL SYLLABLE KIYEOK YA CHIEUCH")
    (17631 . "HANGUL SYLLABLE KIYEOK YA KHIEUKH")
    (17632 . "HANGUL SYLLABLE KIYEOK YA THIEUTH")
    (17633 . "HANGUL SYLLABLE KIYEOK YA PHIEUPH")
    (17634 . "HANGUL SYLLABLE KIYEOK YA HIEUH")
    (17635 . "HANGUL SYLLABLE KIYEOK YAE KIYEOK")
    (17636 . "HANGUL SYLLABLE KIYEOK YAE SSANGKIYEOK")
    (17637 . "HANGUL SYLLABLE KIYEOK YAE KIYEOK-SIOS")
    (17638 . "HANGUL SYLLABLE KIYEOK YAE NIEUN-CIEUC")
    (17639 . "HANGUL SYLLABLE KIYEOK YAE NIEUN-HIEUH")
    (17640 . "HANGUL SYLLABLE KIYEOK YAE TIKEUT")
    (17641 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-KIYEOK")
    (17642 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-MIEUM")
    (17643 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-PIEUP")
    (17644 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-SIOS")
    (17645 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-THIEUTH")
    (17646 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-PHIEUPH")
    (17647 . "HANGUL SYLLABLE KIYEOK YAE RIEUL-HIEUH")
    (17648 . "HANGUL SYLLABLE KIYEOK YAE PIEUP-SIOS")
    (17649 . "HANGUL SYLLABLE KIYEOK YAE SIOS")
    (17650 . "HANGUL SYLLABLE KIYEOK YAE SSANGSIOS")
    (17651 . "HANGUL SYLLABLE KIYEOK YAE IEUNG")
    (17652 . "HANGUL SYLLABLE KIYEOK YAE CIEUC")
    (17653 . "HANGUL SYLLABLE KIYEOK YAE CHIEUCH")
    (17654 . "HANGUL SYLLABLE KIYEOK YAE KHIEUKH")
    (17655 . "HANGUL SYLLABLE KIYEOK YAE THIEUTH")
    (17656 . "HANGUL SYLLABLE KIYEOK YAE PHIEUPH")
    (17657 . "HANGUL SYLLABLE KIYEOK YAE HIEUH")
    (17658 . "HANGUL SYLLABLE KIYEOK EO SSANGKIYEOK")
    (17659 . "HANGUL SYLLABLE KIYEOK EO NIEUN-HIEUH")
    (17660 . "HANGUL SYLLABLE KIYEOK EO RIEUL-PIEUP")
    (17661 . "HANGUL SYLLABLE KIYEOK EO RIEUL-SIOS")
    (17662 . "HANGUL SYLLABLE KIYEOK EO RIEUL-THIEUTH")
    (17663 . "HANGUL SYLLABLE KIYEOK EO RIEUL-PHIEUPH")
    (17664 . "HANGUL SYLLABLE KIYEOK EO RIEUL-HIEUH")
    (17665 . "HANGUL SYLLABLE KIYEOK EO PIEUP-SIOS")
    (17666 . "HANGUL SYLLABLE KIYEOK EO KHIEUKH")
    (17667 . "HANGUL SYLLABLE KIYEOK E SSANGKIYEOK")
    (17668 . "HANGUL SYLLABLE KIYEOK E KIYEOK-SIOS")
    (17669 . "HANGUL SYLLABLE KIYEOK E NIEUN-CIEUC")
    (17670 . "HANGUL SYLLABLE KIYEOK E NIEUN-HIEUH")
    (17671 . "HANGUL SYLLABLE KIYEOK E RIEUL-KIYEOK")
    (17672 . "HANGUL SYLLABLE KIYEOK E RIEUL-MIEUM")
    (17673 . "HANGUL SYLLABLE KIYEOK E RIEUL-PIEUP")
    (17674 . "HANGUL SYLLABLE KIYEOK E RIEUL-SIOS")
    (17675 . "HANGUL SYLLABLE KIYEOK E RIEUL-THIEUTH")
    (17676 . "HANGUL SYLLABLE KIYEOK E RIEUL-PHIEUPH")
    (17677 . "HANGUL SYLLABLE KIYEOK E RIEUL-HIEUH")
    (17678 . "HANGUL SYLLABLE KIYEOK E PIEUP-SIOS")
    (17679 . "HANGUL SYLLABLE KIYEOK E CIEUC")
    (17680 . "HANGUL SYLLABLE KIYEOK E CHIEUCH")
    (17681 . "HANGUL SYLLABLE KIYEOK E KHIEUKH")
    (17682 . "HANGUL SYLLABLE KIYEOK E PHIEUPH")
    (17683 . "HANGUL SYLLABLE KIYEOK YEO KIYEOK-SIOS")
    (17684 . "HANGUL SYLLABLE KIYEOK YEO NIEUN-CIEUC")
    (17685 . "HANGUL SYLLABLE KIYEOK YEO NIEUN-HIEUH")
    (17686 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-MIEUM")
    (17687 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-PIEUP")
    (17688 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-THIEUTH")
    (17689 . "HANGUL SYLLABLE KIYEOK YEO RIEUL-PHIEUPH")
    (17690 . "HANGUL SYLLABLE KIYEOK YEO PIEUP-SIOS")
    (17691 . "HANGUL SYLLABLE KIYEOK YEO KHIEUKH")
    (17692 . "HANGUL SYLLABLE KIYEOK YEO HIEUH")
    (17693 . "HANGUL SYLLABLE KIYEOK YE SSANGKIYEOK")
    (17694 . "HANGUL SYLLABLE KIYEOK YE KIYEOK-SIOS")
    (17695 . "HANGUL SYLLABLE KIYEOK YE NIEUN-CIEUC")
    (17696 . "HANGUL SYLLABLE KIYEOK YE NIEUN-HIEUH")
    (17697 . "HANGUL SYLLABLE KIYEOK YE RIEUL-KIYEOK")
    (17698 . "HANGUL SYLLABLE KIYEOK YE RIEUL-MIEUM")
    (17699 . "HANGUL SYLLABLE KIYEOK YE RIEUL-PIEUP")
    (17700 . "HANGUL SYLLABLE KIYEOK YE RIEUL-SIOS")
    (17701 . "HANGUL SYLLABLE KIYEOK YE RIEUL-THIEUTH")
    (17702 . "HANGUL SYLLABLE KIYEOK YE RIEUL-PHIEUPH")
    (17703 . "HANGUL SYLLABLE KIYEOK YE RIEUL-HIEUH")
    (17704 . "HANGUL SYLLABLE KIYEOK YE PIEUP-SIOS")
    (17705 . "HANGUL SYLLABLE KIYEOK YE CIEUC")
    (17706 . "HANGUL SYLLABLE KIYEOK YE CHIEUCH")
    (17707 . "HANGUL SYLLABLE KIYEOK YE KHIEUKH")
    (17708 . "HANGUL SYLLABLE KIYEOK YE THIEUTH")
    (17709 . "HANGUL SYLLABLE KIYEOK YE PHIEUPH")
    (17710 . "HANGUL SYLLABLE KIYEOK YE HIEUH")
    (17711 . "HANGUL SYLLABLE KIYEOK O SSANGKIYEOK")
    (17712 . "HANGUL SYLLABLE KIYEOK O KIYEOK-SIOS")
    (17713 . "HANGUL SYLLABLE KIYEOK O NIEUN-CIEUC")
    (17714 . "HANGUL SYLLABLE KIYEOK O NIEUN-HIEUH")
    (17715 . "HANGUL SYLLABLE KIYEOK O RIEUL-THIEUTH")
    (17716 . "HANGUL SYLLABLE KIYEOK O PIEUP-SIOS")
    (17717 . "HANGUL SYLLABLE KIYEOK O SSANGSIOS")
    (17718 . "HANGUL SYLLABLE KIYEOK O KHIEUKH")
    (17719 . "HANGUL SYLLABLE KIYEOK WA SSANGKIYEOK")
    (17720 . "HANGUL SYLLABLE KIYEOK WA KIYEOK-SIOS")
    (17721 . "HANGUL SYLLABLE KIYEOK WA NIEUN-CIEUC")
    (17722 . "HANGUL SYLLABLE KIYEOK WA NIEUN-HIEUH")
    (17723 . "HANGUL SYLLABLE KIYEOK WA RIEUL-KIYEOK")
    (17724 . "HANGUL SYLLABLE KIYEOK WA RIEUL-PIEUP")
    (17725 . "HANGUL SYLLABLE KIYEOK WA RIEUL-SIOS")
    (17726 . "HANGUL SYLLABLE KIYEOK WA RIEUL-THIEUTH")
    (17727 . "HANGUL SYLLABLE KIYEOK WA RIEUL-PHIEUPH")
    (17728 . "HANGUL SYLLABLE KIYEOK WA RIEUL-HIEUH")
    (17729 . "HANGUL SYLLABLE KIYEOK WA PIEUP-SIOS")
    (17730 . "HANGUL SYLLABLE KIYEOK WA CIEUC")
    (17731 . "HANGUL SYLLABLE KIYEOK WA CHIEUCH")
    (17732 . "HANGUL SYLLABLE KIYEOK WA KHIEUKH")
    (17733 . "HANGUL SYLLABLE KIYEOK WA THIEUTH")
    (17734 . "HANGUL SYLLABLE KIYEOK WA PHIEUPH")
    (17735 . "HANGUL SYLLABLE KIYEOK WA HIEUH")
    (17736 . "HANGUL SYLLABLE KIYEOK WAE SSANGKIYEOK")
    (17737 . "HANGUL SYLLABLE KIYEOK WAE KIYEOK-SIOS")
    (17738 . "HANGUL SYLLABLE KIYEOK WAE NIEUN-CIEUC")
    (17739 . "HANGUL SYLLABLE KIYEOK WAE NIEUN-HIEUH")
    (17740 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-KIYEOK")
    (17741 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-PIEUP")
    (17742 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-SIOS")
    (17743 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-THIEUTH")
    (17744 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-PHIEUPH")
    (17745 . "HANGUL SYLLABLE KIYEOK WAE RIEUL-HIEUH")
    (17746 . "HANGUL SYLLABLE KIYEOK WAE PIEUP-SIOS")
    (17747 . "HANGUL SYLLABLE KIYEOK WAE CIEUC")
    (17748 . "HANGUL SYLLABLE KIYEOK WAE CHIEUCH")
    (17749 . "HANGUL SYLLABLE KIYEOK WAE KHIEUKH")
    (17750 . "HANGUL SYLLABLE KIYEOK WAE THIEUTH")
    (17751 . "HANGUL SYLLABLE KIYEOK WAE PHIEUPH")
    (17752 . "HANGUL SYLLABLE KIYEOK WAE HIEUH")
    (17753 . "HANGUL SYLLABLE KIYEOK OE SSANGKIYEOK")
    (17754 . "HANGUL SYLLABLE KIYEOK OE KIYEOK-SIOS")
    (17755 . "HANGUL SYLLABLE KIYEOK OE NIEUN-CIEUC")
    (17756 . "HANGUL SYLLABLE KIYEOK OE NIEUN-HIEUH")
    (17757 . "HANGUL SYLLABLE KIYEOK OE RIEUL-KIYEOK")
    (17758 . "HANGUL SYLLABLE KIYEOK OE RIEUL-PIEUP")
    (17759 . "HANGUL SYLLABLE KIYEOK OE RIEUL-SIOS")
    (17760 . "HANGUL SYLLABLE KIYEOK OE RIEUL-THIEUTH")
    (17761 . "HANGUL SYLLABLE KIYEOK OE RIEUL-PHIEUPH")
    (17762 . "HANGUL SYLLABLE KIYEOK OE RIEUL-HIEUH")
    (17763 . "HANGUL SYLLABLE KIYEOK OE PIEUP-SIOS")
    (17764 . "HANGUL SYLLABLE KIYEOK OE CIEUC")
    (17765 . "HANGUL SYLLABLE KIYEOK OE CHIEUCH")
    (17766 . "HANGUL SYLLABLE KIYEOK OE KHIEUKH")
    (17767 . "HANGUL SYLLABLE KIYEOK OE THIEUTH")
    (17768 . "HANGUL SYLLABLE KIYEOK OE PHIEUPH")
    (17769 . "HANGUL SYLLABLE KIYEOK OE HIEUH")
    (17770 . "HANGUL SYLLABLE KIYEOK YO SSANGKIYEOK")
    (17771 . "HANGUL SYLLABLE KIYEOK YO KIYEOK-SIOS")
    (17772 . "HANGUL SYLLABLE KIYEOK YO NIEUN-CIEUC")
    (17773 . "HANGUL SYLLABLE KIYEOK YO NIEUN-HIEUH")
    (17774 . "HANGUL SYLLABLE KIYEOK YO RIEUL-KIYEOK")
    (17775 . "HANGUL SYLLABLE KIYEOK YO RIEUL-MIEUM")
    (17776 . "HANGUL SYLLABLE KIYEOK YO RIEUL-PIEUP")
    (17777 . "HANGUL SYLLABLE KIYEOK YO RIEUL-SIOS")
    (17778 . "HANGUL SYLLABLE KIYEOK YO RIEUL-THIEUTH")
    (17779 . "HANGUL SYLLABLE KIYEOK YO RIEUL-PHIEUPH")
    (17780 . "HANGUL SYLLABLE KIYEOK YO RIEUL-HIEUH")
    (17781 . "HANGUL SYLLABLE KIYEOK YO PIEUP-SIOS")
    (17782 . "HANGUL SYLLABLE KIYEOK YO SSANGSIOS")
    (17783 . "HANGUL SYLLABLE KIYEOK YO CIEUC")
    (17784 . "HANGUL SYLLABLE KIYEOK YO CHIEUCH")
    (17785 . "HANGUL SYLLABLE KIYEOK YO KHIEUKH")
    (17786 . "HANGUL SYLLABLE KIYEOK YO THIEUTH")
    (17787 . "HANGUL SYLLABLE KIYEOK YO PHIEUPH")
    (17788 . "HANGUL SYLLABLE KIYEOK YO HIEUH")
    (17789 . "HANGUL SYLLABLE KIYEOK U SSANGKIYEOK")
    (17790 . "HANGUL SYLLABLE KIYEOK U KIYEOK-SIOS")
    (17791 . "HANGUL SYLLABLE KIYEOK U NIEUN-CIEUC")
    (17792 . "HANGUL SYLLABLE KIYEOK U NIEUN-HIEUH")
    (17793 . "HANGUL SYLLABLE KIYEOK U RIEUL-PIEUP")
    (17794 . "HANGUL SYLLABLE KIYEOK U RIEUL-THIEUTH")
    (17795 . "HANGUL SYLLABLE KIYEOK U RIEUL-PHIEUPH")
    (17796 . "HANGUL SYLLABLE KIYEOK U PIEUP-SIOS")
    (17797 . "HANGUL SYLLABLE KIYEOK U SSANGSIOS")
    (17798 . "HANGUL SYLLABLE KIYEOK U KHIEUKH")
    (17799 . "HANGUL SYLLABLE KIYEOK U THIEUTH")
    (17800 . "HANGUL SYLLABLE KIYEOK U HIEUH")
    (17801 . "HANGUL SYLLABLE KIYEOK WEO SSANGKIYEOK")
    (17802 . "HANGUL SYLLABLE KIYEOK WEO KIYEOK-SIOS")
    (17803 . "HANGUL SYLLABLE KIYEOK WEO NIEUN-CIEUC")
    (17804 . "HANGUL SYLLABLE KIYEOK WEO NIEUN-HIEUH")
    (17805 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-KIYEOK")
    (17806 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-MIEUM")
    (17807 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-PIEUP")
    (17808 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-SIOS")
    (17809 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-THIEUTH")
    (17810 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-PHIEUPH")
    (17811 . "HANGUL SYLLABLE KIYEOK WEO RIEUL-HIEUH")
    (17812 . "HANGUL SYLLABLE KIYEOK WEO MIEUM")
    (17813 . "HANGUL SYLLABLE KIYEOK WEO PIEUP-SIOS")
    (17814 . "HANGUL SYLLABLE KIYEOK WEO CIEUC")
    (17815 . "HANGUL SYLLABLE KIYEOK WEO CHIEUCH")
    (17816 . "HANGUL SYLLABLE KIYEOK WEO KHIEUKH")
    (17817 . "HANGUL SYLLABLE KIYEOK WEO THIEUTH")
    (17818 . "HANGUL SYLLABLE KIYEOK WEO PHIEUPH")
    (17819 . "HANGUL SYLLABLE KIYEOK WEO HIEUH")
    (17820 . "HANGUL SYLLABLE KIYEOK WE SSANGKIYEOK")
    (17821 . "HANGUL SYLLABLE KIYEOK WE KIYEOK-SIOS")
    (17822 . "HANGUL SYLLABLE KIYEOK WE NIEUN-CIEUC")
    (17823 . "HANGUL SYLLABLE KIYEOK WE NIEUN-HIEUH")
    (17824 . "HANGUL SYLLABLE KIYEOK WE RIEUL-KIYEOK")
    (17825 . "HANGUL SYLLABLE KIYEOK WE RIEUL-MIEUM")
    (17826 . "HANGUL SYLLABLE KIYEOK WE RIEUL-PIEUP")
    (17827 . "HANGUL SYLLABLE KIYEOK WE RIEUL-SIOS")
    (17828 . "HANGUL SYLLABLE KIYEOK WE RIEUL-THIEUTH")
    (17829 . "HANGUL SYLLABLE KIYEOK WE RIEUL-PHIEUPH")
    (17830 . "HANGUL SYLLABLE KIYEOK WE RIEUL-HIEUH")
    (17831 . "HANGUL SYLLABLE KIYEOK WE PIEUP-SIOS")
    (17832 . "HANGUL SYLLABLE KIYEOK WE CIEUC")
    (17833 . "HANGUL SYLLABLE KIYEOK WE CHIEUCH")
    (17834 . "HANGUL SYLLABLE KIYEOK WE KHIEUKH")
    (17835 . "HANGUL SYLLABLE KIYEOK WE THIEUTH")
    (17836 . "HANGUL SYLLABLE KIYEOK WE PHIEUPH")
    (17837 . "HANGUL SYLLABLE KIYEOK WE HIEUH")
    (17838 . "HANGUL SYLLABLE KIYEOK WI SSANGKIYEOK")
    (17839 . "HANGUL SYLLABLE KIYEOK WI KIYEOK-SIOS")
    (17840 . "HANGUL SYLLABLE KIYEOK WI NIEUN-CIEUC")
    (17841 . "HANGUL SYLLABLE KIYEOK WI NIEUN-HIEUH")
    (17842 . "HANGUL SYLLABLE KIYEOK WI RIEUL-KIYEOK")
    (17843 . "HANGUL SYLLABLE KIYEOK WI RIEUL-PIEUP")
    (17844 . "HANGUL SYLLABLE KIYEOK WI RIEUL-SIOS")
    (17845 . "HANGUL SYLLABLE KIYEOK WI RIEUL-THIEUTH")
    (17846 . "HANGUL SYLLABLE KIYEOK WI RIEUL-PHIEUPH")
    (17847 . "HANGUL SYLLABLE KIYEOK WI RIEUL-HIEUH")
    (17848 . "HANGUL SYLLABLE KIYEOK WI PIEUP-SIOS")
    (17849 . "HANGUL SYLLABLE KIYEOK WI CIEUC")
    (17850 . "HANGUL SYLLABLE KIYEOK WI CHIEUCH")
    (17851 . "HANGUL SYLLABLE KIYEOK WI KHIEUKH")
    (17852 . "HANGUL SYLLABLE KIYEOK WI THIEUTH")
    (17853 . "HANGUL SYLLABLE KIYEOK WI PHIEUPH")
    (17854 . "HANGUL SYLLABLE KIYEOK WI HIEUH")
    (17855 . "HANGUL SYLLABLE KIYEOK YU SSANGKIYEOK")
    (17856 . "HANGUL SYLLABLE KIYEOK YU KIYEOK-SIOS")
    (17857 . "HANGUL SYLLABLE KIYEOK YU NIEUN-CIEUC")
    (17858 . "HANGUL SYLLABLE KIYEOK YU NIEUN-HIEUH")
    (17859 . "HANGUL SYLLABLE KIYEOK YU TIKEUT")
    (17860 . "HANGUL SYLLABLE KIYEOK YU RIEUL-KIYEOK")
    (17861 . "HANGUL SYLLABLE KIYEOK YU RIEUL-MIEUM")
    (17862 . "HANGUL SYLLABLE KIYEOK YU RIEUL-PIEUP")
    (17863 . "HANGUL SYLLABLE KIYEOK YU RIEUL-THIEUTH")
    (17864 . "HANGUL SYLLABLE KIYEOK YU RIEUL-PHIEUPH")
    (17865 . "HANGUL SYLLABLE KIYEOK YU RIEUL-HIEUH")
    (17866 . "HANGUL SYLLABLE KIYEOK YU PIEUP-SIOS")
    (17867 . "HANGUL SYLLABLE KIYEOK YU SSANGSIOS")
    (17868 . "HANGUL SYLLABLE KIYEOK YU CIEUC")
    (17869 . "HANGUL SYLLABLE KIYEOK YU CHIEUCH")
    (17870 . "HANGUL SYLLABLE KIYEOK YU KHIEUKH")
    (17871 . "HANGUL SYLLABLE KIYEOK YU THIEUTH")
    (17872 . "HANGUL SYLLABLE KIYEOK YU PHIEUPH")
    (17873 . "HANGUL SYLLABLE KIYEOK YU HIEUH")
    (17874 . "HANGUL SYLLABLE KIYEOK EU SSANGKIYEOK")
    (17875 . "HANGUL SYLLABLE KIYEOK EU KIYEOK-SIOS")
    (17876 . "HANGUL SYLLABLE KIYEOK EU NIEUN-CIEUC")
    (17877 . "HANGUL SYLLABLE KIYEOK EU RIEUL-THIEUTH")
    (17878 . "HANGUL SYLLABLE KIYEOK EU RIEUL-PHIEUPH")
    (17879 . "HANGUL SYLLABLE KIYEOK EU PIEUP-SIOS")
    (17880 . "HANGUL SYLLABLE KIYEOK EU SSANGSIOS")
    (17881 . "HANGUL SYLLABLE KIYEOK EU KHIEUKH")
    (17882 . "HANGUL SYLLABLE KIYEOK YI SSANGKIYEOK")
    (17883 . "HANGUL SYLLABLE KIYEOK YI KIYEOK-SIOS")
    (17884 . "HANGUL SYLLABLE KIYEOK YI NIEUN-CIEUC")
    (17885 . "HANGUL SYLLABLE KIYEOK YI NIEUN-HIEUH")
    (17886 . "HANGUL SYLLABLE KIYEOK YI TIKEUT")
    (17887 . "HANGUL SYLLABLE KIYEOK YI RIEUL-KIYEOK")
    (17888 . "HANGUL SYLLABLE KIYEOK YI RIEUL-MIEUM")
    (17889 . "HANGUL SYLLABLE KIYEOK YI RIEUL-PIEUP")
    (17890 . "HANGUL SYLLABLE KIYEOK YI RIEUL-THIEUTH")
    (17891 . "HANGUL SYLLABLE KIYEOK YI RIEUL-PHIEUPH")
    (17892 . "HANGUL SYLLABLE KIYEOK YI RIEUL-HIEUH")
    (17893 . "HANGUL SYLLABLE KIYEOK YI PIEUP-SIOS")
    (17894 . "HANGUL SYLLABLE KIYEOK YI SSANGSIOS")
    (17895 . "HANGUL SYLLABLE KIYEOK YI CIEUC")
    (17896 . "HANGUL SYLLABLE KIYEOK YI CHIEUCH")
    (17897 . "HANGUL SYLLABLE KIYEOK YI KHIEUKH")
    (17898 . "HANGUL SYLLABLE KIYEOK YI THIEUTH")
    (17899 . "HANGUL SYLLABLE KIYEOK YI PHIEUPH")
    (17900 . "HANGUL SYLLABLE KIYEOK YI HIEUH")
    (17901 . "HANGUL SYLLABLE KIYEOK I SSANGKIYEOK")
    (17902 . "HANGUL SYLLABLE KIYEOK I KIYEOK-SIOS")
    (17903 . "HANGUL SYLLABLE KIYEOK I NIEUN-CIEUC")
    (17904 . "HANGUL SYLLABLE KIYEOK I NIEUN-HIEUH")
    (17905 . "HANGUL SYLLABLE KIYEOK I RIEUL-PIEUP")
    (17906 . "HANGUL SYLLABLE KIYEOK I RIEUL-THIEUTH")
    (17907 . "HANGUL SYLLABLE KIYEOK I RIEUL-PHIEUPH")
    (17908 . "HANGUL SYLLABLE KIYEOK I RIEUL-HIEUH")
    (17909 . "HANGUL SYLLABLE KIYEOK I PIEUP-SIOS")
    (17910 . "HANGUL SYLLABLE KIYEOK I KHIEUKH")
    (17911 . "HANGUL SYLLABLE SSANGKIYEOK A KIYEOK-SIOS")
    (17912 . "HANGUL SYLLABLE SSANGKIYEOK A NIEUN-CIEUC")
    (17913 . "HANGUL SYLLABLE SSANGKIYEOK A NIEUN-HIEUH")
    (17914 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-KIYEOK")
    (17915 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-PIEUP")
    (17916 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-SIOS")
    (17917 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-THIEUTH")
    (17918 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-PHIEUPH")
    (17919 . "HANGUL SYLLABLE SSANGKIYEOK A RIEUL-HIEUH")
    (17920 . "HANGUL SYLLABLE SSANGKIYEOK A PIEUP-SIOS")
    (17921 . "HANGUL SYLLABLE SSANGKIYEOK A KHIEUKH")
    (17922 . "HANGUL SYLLABLE SSANGKIYEOK A PHIEUPH")
    (17923 . "HANGUL SYLLABLE SSANGKIYEOK A HIEUH")
    (17924 . "HANGUL SYLLABLE SSANGKIYEOK AE KIYEOK-SIOS")
    (17925 . "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN-CIEUC")
    (17926 . "HANGUL SYLLABLE SSANGKIYEOK AE NIEUN-HIEUH")
    (17927 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-KIYEOK")
    (17928 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-MIEUM")
    (17929 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-SIOS")
    (17930 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-THIEUTH")
    (17931 . "HANGUL SYLLABLE SSANGKIYEOK AE RIEUL-HIEUH")
    (17932 . "HANGUL SYLLABLE SSANGKIYEOK AE PIEUP-SIOS")
    (17933 . "HANGUL SYLLABLE SSANGKIYEOK AE CIEUC")
    (17934 . "HANGUL SYLLABLE SSANGKIYEOK AE CHIEUCH")
    (17935 . "HANGUL SYLLABLE SSANGKIYEOK AE KHIEUKH")
    (17936 . "HANGUL SYLLABLE SSANGKIYEOK AE PHIEUPH")
    (17937 . "HANGUL SYLLABLE SSANGKIYEOK AE HIEUH")
    (17938 . "HANGUL SYLLABLE SSANGKIYEOK YA SSANGKIYEOK")
    (17939 . "HANGUL SYLLABLE SSANGKIYEOK YA KIYEOK-SIOS")
    (17940 . "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN-CIEUC")
    (17941 . "HANGUL SYLLABLE SSANGKIYEOK YA NIEUN-HIEUH")
    (17942 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-MIEUM")
    (17943 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-PIEUP")
    (17944 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-SIOS")
    (17945 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-THIEUTH")
    (17946 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-PHIEUPH")
    (17947 . "HANGUL SYLLABLE SSANGKIYEOK YA RIEUL-HIEUH")
    (17948 . "HANGUL SYLLABLE SSANGKIYEOK YA MIEUM")
    (17949 . "HANGUL SYLLABLE SSANGKIYEOK YA PIEUP")
    (17950 . "HANGUL SYLLABLE SSANGKIYEOK YA PIEUP-SIOS")
    (17951 . "HANGUL SYLLABLE SSANGKIYEOK YA SSANGSIOS")
    (17952 . "HANGUL SYLLABLE SSANGKIYEOK YA CIEUC")
    (17953 . "HANGUL SYLLABLE SSANGKIYEOK YA CHIEUCH")
    (17954 . "HANGUL SYLLABLE SSANGKIYEOK YA KHIEUKH")
    (17955 . "HANGUL SYLLABLE SSANGKIYEOK YA THIEUTH")
    (17956 . "HANGUL SYLLABLE SSANGKIYEOK YA PHIEUPH")
    (17957 . "HANGUL SYLLABLE SSANGKIYEOK YA HIEUH")
    (17958 . "HANGUL SYLLABLE SSANGKIYEOK YAE KIYEOK")
    (17959 . "HANGUL SYLLABLE SSANGKIYEOK YAE SSANGKIYEOK")
    (17960 . "HANGUL SYLLABLE SSANGKIYEOK YAE KIYEOK-SIOS")
    (17961 . "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN")
    (17962 . "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN-CIEUC")
    (17963 . "HANGUL SYLLABLE SSANGKIYEOK YAE NIEUN-HIEUH")
    (17964 . "HANGUL SYLLABLE SSANGKIYEOK YAE TIKEUT")
    (17965 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL")
    (17966 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-KIYEOK")
    (17967 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-MIEUM")
    (17968 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-PIEUP")
    (17969 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-SIOS")
    (17970 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-THIEUTH")
    (17971 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-PHIEUPH")
    (17972 . "HANGUL SYLLABLE SSANGKIYEOK YAE RIEUL-HIEUH")
    (17973 . "HANGUL SYLLABLE SSANGKIYEOK YAE MIEUM")
    (17974 . "HANGUL SYLLABLE SSANGKIYEOK YAE PIEUP")
    (17975 . "HANGUL SYLLABLE SSANGKIYEOK YAE PIEUP-SIOS")
    (17976 . "HANGUL SYLLABLE SSANGKIYEOK YAE SIOS")
    (17977 . "HANGUL SYLLABLE SSANGKIYEOK YAE SSANGSIOS")
    (17978 . "HANGUL SYLLABLE SSANGKIYEOK YAE CIEUC")
    (17979 . "HANGUL SYLLABLE SSANGKIYEOK YAE CHIEUCH")
    (17980 . "HANGUL SYLLABLE SSANGKIYEOK YAE KHIEUKH")
    (17981 . "HANGUL SYLLABLE SSANGKIYEOK YAE THIEUTH")
    (17982 . "HANGUL SYLLABLE SSANGKIYEOK YAE PHIEUPH")
    (17983 . "HANGUL SYLLABLE SSANGKIYEOK YAE HIEUH")
    (17984 . "HANGUL SYLLABLE SSANGKIYEOK EO KIYEOK-SIOS")
    (17985 . "HANGUL SYLLABLE SSANGKIYEOK EO NIEUN-CIEUC")
    (17986 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-KIYEOK")
    (17987 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-MIEUM")
    (17988 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-PIEUP")
    (17989 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-SIOS")
    (17990 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-THIEUTH")
    (17991 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-PHIEUPH")
    (17992 . "HANGUL SYLLABLE SSANGKIYEOK EO RIEUL-HIEUH")
    (17993 . "HANGUL SYLLABLE SSANGKIYEOK EO PIEUP-SIOS")
    (17994 . "HANGUL SYLLABLE SSANGKIYEOK EO KHIEUKH")
    (17995 . "HANGUL SYLLABLE SSANGKIYEOK EO PHIEUPH")
    (17996 . "HANGUL SYLLABLE SSANGKIYEOK EO HIEUH")
    (17997 . "HANGUL SYLLABLE SSANGKIYEOK E KIYEOK-SIOS")
    (17998 . "HANGUL SYLLABLE SSANGKIYEOK E NIEUN-CIEUC")
    (17999 . "HANGUL SYLLABLE SSANGKIYEOK E NIEUN-HIEUH")
    (18000 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-KIYEOK")
    (18001 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-MIEUM")
    (18002 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-PIEUP")
    (18003 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-SIOS")
    (18004 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-THIEUTH")
    (18005 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-PHIEUPH")
    (18006 . "HANGUL SYLLABLE SSANGKIYEOK E RIEUL-HIEUH")
    (18007 . "HANGUL SYLLABLE SSANGKIYEOK E PIEUP-SIOS")
    (18008 . "HANGUL SYLLABLE SSANGKIYEOK E CIEUC")
    (18009 . "HANGUL SYLLABLE SSANGKIYEOK E CHIEUCH")
    (18010 . "HANGUL SYLLABLE SSANGKIYEOK E KHIEUKH")
    (18011 . "HANGUL SYLLABLE SSANGKIYEOK E PHIEUPH")
    (18012 . "HANGUL SYLLABLE SSANGKIYEOK E HIEUH")
    (18013 . "HANGUL SYLLABLE SSANGKIYEOK YEO SSANGKIYEOK")
    (18014 . "HANGUL SYLLABLE SSANGKIYEOK YEO KIYEOK-SIOS")
    (18015 . "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN-CIEUC")
    (18016 . "HANGUL SYLLABLE SSANGKIYEOK YEO NIEUN-HIEUH")
    (18017 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-KIYEOK")
    (18018 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-MIEUM")
    (18019 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-PIEUP")
    (18020 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-SIOS")
    (18021 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-THIEUTH")
    (18022 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-PHIEUPH")
    (18023 . "HANGUL SYLLABLE SSANGKIYEOK YEO RIEUL-HIEUH")
    (18024 . "HANGUL SYLLABLE SSANGKIYEOK YEO PIEUP-SIOS")
    (18025 . "HANGUL SYLLABLE SSANGKIYEOK YEO CIEUC")
    (18026 . "HANGUL SYLLABLE SSANGKIYEOK YEO CHIEUCH")
    (18027 . "HANGUL SYLLABLE SSANGKIYEOK YEO KHIEUKH")
    (18028 . "HANGUL SYLLABLE SSANGKIYEOK YEO PHIEUPH")
    (18029 . "HANGUL SYLLABLE SSANGKIYEOK YEO HIEUH")
    (18030 . "HANGUL SYLLABLE SSANGKIYEOK YE KIYEOK")
    (18031 . "HANGUL SYLLABLE SSANGKIYEOK YE SSANGKIYEOK")
    (18032 . "HANGUL SYLLABLE SSANGKIYEOK YE KIYEOK-SIOS")
    (18033 . "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN")
    (18034 . "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN-CIEUC")
    (18035 . "HANGUL SYLLABLE SSANGKIYEOK YE NIEUN-HIEUH")
    (18036 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL")
    (18037 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-KIYEOK")
    (18038 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-MIEUM")
    (18039 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-PIEUP")
    (18040 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-SIOS")
    (18041 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-THIEUTH")
    (18042 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-PHIEUPH")
    (18043 . "HANGUL SYLLABLE SSANGKIYEOK YE RIEUL-HIEUH")
    (18044 . "HANGUL SYLLABLE SSANGKIYEOK YE MIEUM")
    (18045 . "HANGUL SYLLABLE SSANGKIYEOK YE PIEUP")
    (18046 . "HANGUL SYLLABLE SSANGKIYEOK YE PIEUP-SIOS")
    (18047 . "HANGUL SYLLABLE SSANGKIYEOK YE SIOS")
    (18048 . "HANGUL SYLLABLE SSANGKIYEOK YE SSANGSIOS")
    (18049 . "HANGUL SYLLABLE SSANGKIYEOK YE CIEUC")
    (18050 . "HANGUL SYLLABLE SSANGKIYEOK YE CHIEUCH")
    (18051 . "HANGUL SYLLABLE SSANGKIYEOK YE KHIEUKH")
    (18052 . "HANGUL SYLLABLE SSANGKIYEOK YE THIEUTH")
    (18053 . "HANGUL SYLLABLE SSANGKIYEOK YE PHIEUPH")
    (18054 . "HANGUL SYLLABLE SSANGKIYEOK YE HIEUH")
    (18055 . "HANGUL SYLLABLE SSANGKIYEOK O SSANGKIYEOK")
    (18056 . "HANGUL SYLLABLE SSANGKIYEOK O KIYEOK-SIOS")
    (18057 . "HANGUL SYLLABLE SSANGKIYEOK O NIEUN-CIEUC")
    (18058 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-KIYEOK")
    (18059 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-PIEUP")
    (18060 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-THIEUTH")
    (18061 . "HANGUL SYLLABLE SSANGKIYEOK O RIEUL-PHIEUPH")
    (18062 . "HANGUL SYLLABLE SSANGKIYEOK O PIEUP-SIOS")
    (18063 . "HANGUL SYLLABLE SSANGKIYEOK O SSANGSIOS")
    (18064 . "HANGUL SYLLABLE SSANGKIYEOK O KHIEUKH")
    (18065 . "HANGUL SYLLABLE SSANGKIYEOK O PHIEUPH")
    (18066 . "HANGUL SYLLABLE SSANGKIYEOK WA SSANGKIYEOK")
    (18067 . "HANGUL SYLLABLE SSANGKIYEOK WA KIYEOK-SIOS")
    (18068 . "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN-CIEUC")
    (18069 . "HANGUL SYLLABLE SSANGKIYEOK WA NIEUN-HIEUH")
    (18070 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-KIYEOK")
    (18071 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-MIEUM")
    (18072 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-PIEUP")
    (18073 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-SIOS")
    (18074 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-THIEUTH")
    (18075 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-PHIEUPH")
    (18076 . "HANGUL SYLLABLE SSANGKIYEOK WA RIEUL-HIEUH")
    (18077 . "HANGUL SYLLABLE SSANGKIYEOK WA PIEUP-SIOS")
    (18078 . "HANGUL SYLLABLE SSANGKIYEOK WA CIEUC")
    (18079 . "HANGUL SYLLABLE SSANGKIYEOK WA CHIEUCH")
    (18080 . "HANGUL SYLLABLE SSANGKIYEOK WA KHIEUKH")
    (18081 . "HANGUL SYLLABLE SSANGKIYEOK WA THIEUTH")
    (18082 . "HANGUL SYLLABLE SSANGKIYEOK WA PHIEUPH")
    (18083 . "HANGUL SYLLABLE SSANGKIYEOK WA HIEUH")
    (18084 . "HANGUL SYLLABLE SSANGKIYEOK WAE SSANGKIYEOK")
    (18085 . "HANGUL SYLLABLE SSANGKIYEOK WAE KIYEOK-SIOS")
    (18086 . "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN-CIEUC")
    (18087 . "HANGUL SYLLABLE SSANGKIYEOK WAE NIEUN-HIEUH")
    (18088 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-KIYEOK")
    (18089 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-MIEUM")
    (18090 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-PIEUP")
    (18091 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-SIOS")
    (18092 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-THIEUTH")
    (18093 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-PHIEUPH")
    (18094 . "HANGUL SYLLABLE SSANGKIYEOK WAE RIEUL-HIEUH")
    (18095 . "HANGUL SYLLABLE SSANGKIYEOK WAE PIEUP-SIOS")
    (18096 . "HANGUL SYLLABLE SSANGKIYEOK WAE SIOS")
    (18097 . "HANGUL SYLLABLE SSANGKIYEOK WAE CIEUC")
    (18098 . "HANGUL SYLLABLE SSANGKIYEOK WAE CHIEUCH")
    (18099 . "HANGUL SYLLABLE SSANGKIYEOK WAE KHIEUKH")
    (18100 . "HANGUL SYLLABLE SSANGKIYEOK WAE THIEUTH")
    (18101 . "HANGUL SYLLABLE SSANGKIYEOK WAE PHIEUPH")
    (18102 . "HANGUL SYLLABLE SSANGKIYEOK WAE HIEUH")
    (18103 . "HANGUL SYLLABLE SSANGKIYEOK OE SSANGKIYEOK")
    (18104 . "HANGUL SYLLABLE SSANGKIYEOK OE KIYEOK-SIOS")
    (18105 . "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN-CIEUC")
    (18106 . "HANGUL SYLLABLE SSANGKIYEOK OE NIEUN-HIEUH")
    (18107 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-KIYEOK")
    (18108 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-MIEUM")
    (18109 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-PIEUP")
    (18110 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-SIOS")
    (18111 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-THIEUTH")
    (18112 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-PHIEUPH")
    (18113 . "HANGUL SYLLABLE SSANGKIYEOK OE RIEUL-HIEUH")
    (18114 . "HANGUL SYLLABLE SSANGKIYEOK OE PIEUP-SIOS")
    (18115 . "HANGUL SYLLABLE SSANGKIYEOK OE CIEUC")
    (18116 . "HANGUL SYLLABLE SSANGKIYEOK OE CHIEUCH")
    (18117 . "HANGUL SYLLABLE SSANGKIYEOK OE KHIEUKH")
    (18118 . "HANGUL SYLLABLE SSANGKIYEOK OE THIEUTH")
    (18119 . "HANGUL SYLLABLE SSANGKIYEOK OE PHIEUPH")
    (18120 . "HANGUL SYLLABLE SSANGKIYEOK OE HIEUH")
    (18121 . "HANGUL SYLLABLE SSANGKIYEOK YO KIYEOK")
    (18122 . "HANGUL SYLLABLE SSANGKIYEOK YO SSANGKIYEOK")
    (18123 . "HANGUL SYLLABLE SSANGKIYEOK YO KIYEOK-SIOS")
    (18124 . "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN-CIEUC")
    (18125 . "HANGUL SYLLABLE SSANGKIYEOK YO NIEUN-HIEUH")
    (18126 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-KIYEOK")
    (18127 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-MIEUM")
    (18128 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-PIEUP")
    (18129 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-SIOS")
    (18130 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-THIEUTH")
    (18131 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-PHIEUPH")
    (18132 . "HANGUL SYLLABLE SSANGKIYEOK YO RIEUL-HIEUH")
    (18133 . "HANGUL SYLLABLE SSANGKIYEOK YO PIEUP-SIOS")
    (18134 . "HANGUL SYLLABLE SSANGKIYEOK YO SIOS")
    (18135 . "HANGUL SYLLABLE SSANGKIYEOK YO SSANGSIOS")
    (18136 . "HANGUL SYLLABLE SSANGKIYEOK YO IEUNG")
    (18137 . "HANGUL SYLLABLE SSANGKIYEOK YO CHIEUCH")
    (18138 . "HANGUL SYLLABLE SSANGKIYEOK YO KHIEUKH")
    (18139 . "HANGUL SYLLABLE SSANGKIYEOK YO THIEUTH")
    (18140 . "HANGUL SYLLABLE SSANGKIYEOK YO PHIEUPH")
    (18141 . "HANGUL SYLLABLE SSANGKIYEOK YO HIEUH")
    (18142 . "HANGUL SYLLABLE SSANGKIYEOK U SSANGKIYEOK")
    (18143 . "HANGUL SYLLABLE SSANGKIYEOK U KIYEOK-SIOS")
    (18144 . "HANGUL SYLLABLE SSANGKIYEOK U NIEUN-CIEUC")
    (18145 . "HANGUL SYLLABLE SSANGKIYEOK U NIEUN-HIEUH")
    (18146 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-MIEUM")
    (18147 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-PIEUP")
    (18148 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-SIOS")
    (18149 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-THIEUTH")
    (18150 . "HANGUL SYLLABLE SSANGKIYEOK U RIEUL-PHIEUPH")
    (18151 . "HANGUL SYLLABLE SSANGKIYEOK U PIEUP-SIOS")
    (18152 . "HANGUL SYLLABLE SSANGKIYEOK U SSANGSIOS")
    (18153 . "HANGUL SYLLABLE SSANGKIYEOK U KHIEUKH")
    (18154 . "HANGUL SYLLABLE SSANGKIYEOK U THIEUTH")
    (18155 . "HANGUL SYLLABLE SSANGKIYEOK U PHIEUPH")
    (18156 . "HANGUL SYLLABLE SSANGKIYEOK U HIEUH")
    (18157 . "HANGUL SYLLABLE SSANGKIYEOK WEO SSANGKIYEOK")
    (18158 . "HANGUL SYLLABLE SSANGKIYEOK WEO KIYEOK-SIOS")
    (18159 . "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN-CIEUC")
    (18160 . "HANGUL SYLLABLE SSANGKIYEOK WEO NIEUN-HIEUH")
    (18161 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-KIYEOK")
    (18162 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-MIEUM")
    (18163 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-PIEUP")
    (18164 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-SIOS")
    (18165 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-THIEUTH")
    (18166 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-PHIEUPH")
    (18167 . "HANGUL SYLLABLE SSANGKIYEOK WEO RIEUL-HIEUH")
    (18168 . "HANGUL SYLLABLE SSANGKIYEOK WEO PIEUP")
    (18169 . "HANGUL SYLLABLE SSANGKIYEOK WEO PIEUP-SIOS")
    (18170 . "HANGUL SYLLABLE SSANGKIYEOK WEO CIEUC")
    (18171 . "HANGUL SYLLABLE SSANGKIYEOK WEO CHIEUCH")
    (18172 . "HANGUL SYLLABLE SSANGKIYEOK WEO KHIEUKH")
    (18173 . "HANGUL SYLLABLE SSANGKIYEOK WEO THIEUTH")
    (18174 . "HANGUL SYLLABLE SSANGKIYEOK WEO PHIEUPH")
    (18175 . "HANGUL SYLLABLE SSANGKIYEOK WEO HIEUH")
    (18176 . "HANGUL SYLLABLE SSANGKIYEOK WE SSANGKIYEOK")
    (18177 . "HANGUL SYLLABLE SSANGKIYEOK WE KIYEOK-SIOS")
    (18178 . "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN-CIEUC")
    (18179 . "HANGUL SYLLABLE SSANGKIYEOK WE NIEUN-HIEUH")
    (18180 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-KIYEOK")
    (18181 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-MIEUM")
    (18182 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-PIEUP")
    (18183 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-SIOS")
    (18184 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-THIEUTH")
    (18185 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-PHIEUPH")
    (18186 . "HANGUL SYLLABLE SSANGKIYEOK WE RIEUL-HIEUH")
    (18187 . "HANGUL SYLLABLE SSANGKIYEOK WE PIEUP-SIOS")
    (18188 . "HANGUL SYLLABLE SSANGKIYEOK WE CIEUC")
    (18189 . "HANGUL SYLLABLE SSANGKIYEOK WE CHIEUCH")
    (18190 . "HANGUL SYLLABLE SSANGKIYEOK WE KHIEUKH")
    (18191 . "HANGUL SYLLABLE SSANGKIYEOK WE THIEUTH")
    (18192 . "HANGUL SYLLABLE SSANGKIYEOK WE PHIEUPH")
    (18193 . "HANGUL SYLLABLE SSANGKIYEOK WE HIEUH")
    (18194 . "HANGUL SYLLABLE SSANGKIYEOK WI SSANGKIYEOK")
    (18195 . "HANGUL SYLLABLE SSANGKIYEOK WI KIYEOK-SIOS")
    (18196 . "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN-CIEUC")
    (18197 . "HANGUL SYLLABLE SSANGKIYEOK WI NIEUN-HIEUH")
    (18198 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-KIYEOK")
    (18199 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-MIEUM")
    (18200 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-PIEUP")
    (18201 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-SIOS")
    (18202 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-THIEUTH")
    (18203 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-PHIEUPH")
    (18204 . "HANGUL SYLLABLE SSANGKIYEOK WI RIEUL-HIEUH")
    (18205 . "HANGUL SYLLABLE SSANGKIYEOK WI PIEUP-SIOS")
    (18206 . "HANGUL SYLLABLE SSANGKIYEOK WI SSANGSIOS")
    (18207 . "HANGUL SYLLABLE SSANGKIYEOK WI CIEUC")
    (18208 . "HANGUL SYLLABLE SSANGKIYEOK WI CHIEUCH")
    (18209 . "HANGUL SYLLABLE SSANGKIYEOK WI KHIEUKH")
    (18210 . "HANGUL SYLLABLE SSANGKIYEOK WI THIEUTH")
    (18211 . "HANGUL SYLLABLE SSANGKIYEOK WI PHIEUPH")
    (18212 . "HANGUL SYLLABLE SSANGKIYEOK WI HIEUH")
    (18213 . "HANGUL SYLLABLE SSANGKIYEOK YU KIYEOK")
    (18214 . "HANGUL SYLLABLE SSANGKIYEOK YU SSANGKIYEOK")
    (18215 . "HANGUL SYLLABLE SSANGKIYEOK YU KIYEOK-SIOS")
    (18216 . "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN-CIEUC")
    (18217 . "HANGUL SYLLABLE SSANGKIYEOK YU NIEUN-HIEUH")
    (18218 . "HANGUL SYLLABLE SSANGKIYEOK YU TIKEUT")
    (18219 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-KIYEOK")
    (18220 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-MIEUM")
    (18221 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-PIEUP")
    (18222 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-SIOS")
    (18223 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-THIEUTH")
    (18224 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-PHIEUPH")
    (18225 . "HANGUL SYLLABLE SSANGKIYEOK YU RIEUL-HIEUH")
    (18226 . "HANGUL SYLLABLE SSANGKIYEOK YU PIEUP-SIOS")
    (18227 . "HANGUL SYLLABLE SSANGKIYEOK YU SIOS")
    (18228 . "HANGUL SYLLABLE SSANGKIYEOK YU SSANGSIOS")
    (18229 . "HANGUL SYLLABLE SSANGKIYEOK YU IEUNG")
    (18230 . "HANGUL SYLLABLE SSANGKIYEOK YU CIEUC")
    (18231 . "HANGUL SYLLABLE SSANGKIYEOK YU CHIEUCH")
    (18232 . "HANGUL SYLLABLE SSANGKIYEOK YU KHIEUKH")
    (18233 . "HANGUL SYLLABLE SSANGKIYEOK YU THIEUTH")
    (18234 . "HANGUL SYLLABLE SSANGKIYEOK YU PHIEUPH")
    (18235 . "HANGUL SYLLABLE SSANGKIYEOK YU HIEUH")
    (18236 . "HANGUL SYLLABLE SSANGKIYEOK EU SSANGKIYEOK")
    (18237 . "HANGUL SYLLABLE SSANGKIYEOK EU KIYEOK-SIOS")
    (18238 . "HANGUL SYLLABLE SSANGKIYEOK EU NIEUN-CIEUC")
    (18239 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-PIEUP")
    (18240 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-SIOS")
    (18241 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-THIEUTH")
    (18242 . "HANGUL SYLLABLE SSANGKIYEOK EU RIEUL-PHIEUPH")
    (18243 . "HANGUL SYLLABLE SSANGKIYEOK EU PIEUP-SIOS")
    (18244 . "HANGUL SYLLABLE SSANGKIYEOK EU SSANGSIOS")
    (18245 . "HANGUL SYLLABLE SSANGKIYEOK EU CIEUC")
    (18246 . "HANGUL SYLLABLE SSANGKIYEOK EU KHIEUKH")
    (18247 . "HANGUL SYLLABLE SSANGKIYEOK EU PHIEUPH")
    (18248 . "HANGUL SYLLABLE SSANGKIYEOK YI SSANGKIYEOK")
    (18249 . "HANGUL SYLLABLE SSANGKIYEOK YI KIYEOK-SIOS")
    (18250 . "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN")
    (18251 . "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN-CIEUC")
    (18252 . "HANGUL SYLLABLE SSANGKIYEOK YI NIEUN-HIEUH")
    (18253 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-KIYEOK")
    (18254 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-MIEUM")
    (18255 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-SIOS")
    (18256 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-THIEUTH")
    (18257 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-PHIEUPH")
    (18258 . "HANGUL SYLLABLE SSANGKIYEOK YI RIEUL-HIEUH")
    (18259 . "HANGUL SYLLABLE SSANGKIYEOK YI MIEUM")
    (18260 . "HANGUL SYLLABLE SSANGKIYEOK YI PIEUP")
    (18261 . "HANGUL SYLLABLE SSANGKIYEOK YI PIEUP-SIOS")
    (18262 . "HANGUL SYLLABLE SSANGKIYEOK YI SSANGSIOS")
    (18263 . "HANGUL SYLLABLE SSANGKIYEOK YI CIEUC")
    (18264 . "HANGUL SYLLABLE SSANGKIYEOK YI CHIEUCH")
    (18265 . "HANGUL SYLLABLE SSANGKIYEOK YI KHIEUKH")
    (18266 . "HANGUL SYLLABLE SSANGKIYEOK YI THIEUTH")
    (18267 . "HANGUL SYLLABLE SSANGKIYEOK YI PHIEUPH")
    (18268 . "HANGUL SYLLABLE SSANGKIYEOK YI HIEUH")
    (18269 . "HANGUL SYLLABLE SSANGKIYEOK I SSANGKIYEOK")
    (18270 . "HANGUL SYLLABLE SSANGKIYEOK I KIYEOK-SIOS")
    (18271 . "HANGUL SYLLABLE SSANGKIYEOK I NIEUN-CIEUC")
    (18272 . "HANGUL SYLLABLE SSANGKIYEOK I NIEUN-HIEUH")
    (18273 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-KIYEOK")
    (18274 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-MIEUM")
    (18275 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-PIEUP")
    (18276 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-SIOS")
    (18277 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-THIEUTH")
    (18278 . "HANGUL SYLLABLE SSANGKIYEOK I RIEUL-PHIEUPH")
    (18279 . "HANGUL SYLLABLE SSANGKIYEOK I PIEUP-SIOS")
    (18280 . "HANGUL SYLLABLE SSANGKIYEOK I CIEUC")
    (18281 . "HANGUL SYLLABLE SSANGKIYEOK I CHIEUCH")
    (18282 . "HANGUL SYLLABLE SSANGKIYEOK I KHIEUKH")
    (18283 . "HANGUL SYLLABLE SSANGKIYEOK I PHIEUPH")
    (18284 . "HANGUL SYLLABLE SSANGKIYEOK I HIEUH")
    (18285 . "HANGUL SYLLABLE NIEUN A NIEUN-CIEUC")
    (18286 . "HANGUL SYLLABLE NIEUN A NIEUN-HIEUH")
    (18287 . "HANGUL SYLLABLE NIEUN A RIEUL-PIEUP")
    (18288 . "HANGUL SYLLABLE NIEUN A RIEUL-THIEUTH")
    (18289 . "HANGUL SYLLABLE NIEUN A RIEUL-PHIEUPH")
    (18290 . "HANGUL SYLLABLE NIEUN A RIEUL-HIEUH")
    (18291 . "HANGUL SYLLABLE NIEUN AE SSANGKIYEOK")
    (18292 . "HANGUL SYLLABLE NIEUN AE KIYEOK-SIOS")
    (18293 . "HANGUL SYLLABLE NIEUN AE NIEUN-CIEUC")
    (18294 . "HANGUL SYLLABLE NIEUN AE NIEUN-HIEUH")
    (18295 . "HANGUL SYLLABLE NIEUN AE RIEUL-KIYEOK")
    (18296 . "HANGUL SYLLABLE NIEUN AE RIEUL-MIEUM")
    (18297 . "HANGUL SYLLABLE NIEUN AE RIEUL-PIEUP")
    (18298 . "HANGUL SYLLABLE NIEUN AE RIEUL-SIOS")
    (18299 . "HANGUL SYLLABLE NIEUN AE RIEUL-THIEUTH")
    (18300 . "HANGUL SYLLABLE NIEUN AE RIEUL-PHIEUPH")
    (18301 . "HANGUL SYLLABLE NIEUN AE RIEUL-HIEUH")
    (18302 . "HANGUL SYLLABLE NIEUN AE PIEUP-SIOS")
    (18303 . "HANGUL SYLLABLE NIEUN AE CIEUC")
    (18304 . "HANGUL SYLLABLE NIEUN AE CHIEUCH")
    (18305 . "HANGUL SYLLABLE NIEUN AE KHIEUKH")
    (18306 . "HANGUL SYLLABLE NIEUN AE THIEUTH")
    (18307 . "HANGUL SYLLABLE NIEUN AE HIEUH")
    (18308 . "HANGUL SYLLABLE NIEUN YA SSANGKIYEOK")
    (18309 . "HANGUL SYLLABLE NIEUN YA KIYEOK-SIOS")
    (18310 . "HANGUL SYLLABLE NIEUN YA NIEUN-CIEUC")
    (18311 . "HANGUL SYLLABLE NIEUN YA NIEUN-HIEUH")
    (18312 . "HANGUL SYLLABLE NIEUN YA RIEUL-KIYEOK")
    (18313 . "HANGUL SYLLABLE NIEUN YA RIEUL-MIEUM")
    (18314 . "HANGUL SYLLABLE NIEUN YA RIEUL-PIEUP")
    (18315 . "HANGUL SYLLABLE NIEUN YA RIEUL-SIOS")
    (18316 . "HANGUL SYLLABLE NIEUN YA RIEUL-THIEUTH")
    (18317 . "HANGUL SYLLABLE NIEUN YA RIEUL-PHIEUPH")
    (18318 . "HANGUL SYLLABLE NIEUN YA RIEUL-HIEUH")
    (18319 . "HANGUL SYLLABLE NIEUN YA PIEUP-SIOS")
    (18320 . "HANGUL SYLLABLE NIEUN YA SSANGSIOS")
    (18321 . "HANGUL SYLLABLE NIEUN YA CHIEUCH")
    (18322 . "HANGUL SYLLABLE NIEUN YA KHIEUKH")
    (18323 . "HANGUL SYLLABLE NIEUN YA PHIEUPH")
    (18324 . "HANGUL SYLLABLE NIEUN YA HIEUH")
    (18325 . "HANGUL SYLLABLE NIEUN YAE KIYEOK")
    (18326 . "HANGUL SYLLABLE NIEUN YAE SSANGKIYEOK")
    (18327 . "HANGUL SYLLABLE NIEUN YAE KIYEOK-SIOS")
    (18328 . "HANGUL SYLLABLE NIEUN YAE NIEUN")
    (18329 . "HANGUL SYLLABLE NIEUN YAE NIEUN-CIEUC")
    (18330 . "HANGUL SYLLABLE NIEUN YAE NIEUN-HIEUH")
    (18331 . "HANGUL SYLLABLE NIEUN YAE TIKEUT")
    (18332 . "HANGUL SYLLABLE NIEUN YAE RIEUL")
    (18333 . "HANGUL SYLLABLE NIEUN YAE RIEUL-KIYEOK")
    (18334 . "HANGUL SYLLABLE NIEUN YAE RIEUL-MIEUM")
    (18335 . "HANGUL SYLLABLE NIEUN YAE RIEUL-PIEUP")
    (18336 . "HANGUL SYLLABLE NIEUN YAE RIEUL-SIOS")
    (18337 . "HANGUL SYLLABLE NIEUN YAE RIEUL-THIEUTH")
    (18338 . "HANGUL SYLLABLE NIEUN YAE RIEUL-PHIEUPH")
    (18339 . "HANGUL SYLLABLE NIEUN YAE RIEUL-HIEUH")
    (18340 . "HANGUL SYLLABLE NIEUN YAE MIEUM")
    (18341 . "HANGUL SYLLABLE NIEUN YAE PIEUP")
    (18342 . "HANGUL SYLLABLE NIEUN YAE PIEUP-SIOS")
    (18343 . "HANGUL SYLLABLE NIEUN YAE SIOS")
    (18344 . "HANGUL SYLLABLE NIEUN YAE SSANGSIOS")
    (18345 . "HANGUL SYLLABLE NIEUN YAE CIEUC")
    (18346 . "HANGUL SYLLABLE NIEUN YAE CHIEUCH")
    (18347 . "HANGUL SYLLABLE NIEUN YAE KHIEUKH")
    (18348 . "HANGUL SYLLABLE NIEUN YAE THIEUTH")
    (18349 . "HANGUL SYLLABLE NIEUN YAE PHIEUPH")
    (18350 . "HANGUL SYLLABLE NIEUN YAE HIEUH")
    (18351 . "HANGUL SYLLABLE NIEUN EO NIEUN-CIEUC")
    (18352 . "HANGUL SYLLABLE NIEUN EO NIEUN-HIEUH")
    (18353 . "HANGUL SYLLABLE NIEUN EO RIEUL-KIYEOK")
    (18354 . "HANGUL SYLLABLE NIEUN EO RIEUL-SIOS")
    (18355 . "HANGUL SYLLABLE NIEUN EO RIEUL-THIEUTH")
    (18356 . "HANGUL SYLLABLE NIEUN EO RIEUL-PHIEUPH")
    (18357 . "HANGUL SYLLABLE NIEUN EO PIEUP-SIOS")
    (18358 . "HANGUL SYLLABLE NIEUN EO CHIEUCH")
    (18359 . "HANGUL SYLLABLE NIEUN E SSANGKIYEOK")
    (18360 . "HANGUL SYLLABLE NIEUN E KIYEOK-SIOS")
    (18361 . "HANGUL SYLLABLE NIEUN E NIEUN-CIEUC")
    (18362 . "HANGUL SYLLABLE NIEUN E NIEUN-HIEUH")
    (18363 . "HANGUL SYLLABLE NIEUN E RIEUL-KIYEOK")
    (18364 . "HANGUL SYLLABLE NIEUN E RIEUL-MIEUM")
    (18365 . "HANGUL SYLLABLE NIEUN E RIEUL-PIEUP")
    (18366 . "HANGUL SYLLABLE NIEUN E RIEUL-SIOS")
    (18367 . "HANGUL SYLLABLE NIEUN E RIEUL-THIEUTH")
    (18368 . "HANGUL SYLLABLE NIEUN E RIEUL-PHIEUPH")
    (18369 . "HANGUL SYLLABLE NIEUN E RIEUL-HIEUH")
    (18370 . "HANGUL SYLLABLE NIEUN E PIEUP-SIOS")
    (18371 . "HANGUL SYLLABLE NIEUN E CIEUC")
    (18372 . "HANGUL SYLLABLE NIEUN E CHIEUCH")
    (18373 . "HANGUL SYLLABLE NIEUN YEO SSANGKIYEOK")
    (18374 . "HANGUL SYLLABLE NIEUN YEO NIEUN-CIEUC")
    (18375 . "HANGUL SYLLABLE NIEUN YEO NIEUN-HIEUH")
    (18376 . "HANGUL SYLLABLE NIEUN YEO RIEUL-KIYEOK")
    (18377 . "HANGUL SYLLABLE NIEUN YEO RIEUL-MIEUM")
    (18378 . "HANGUL SYLLABLE NIEUN YEO RIEUL-SIOS")
    (18379 . "HANGUL SYLLABLE NIEUN YEO RIEUL-THIEUTH")
    (18380 . "HANGUL SYLLABLE NIEUN YEO RIEUL-PHIEUPH")
    (18381 . "HANGUL SYLLABLE NIEUN YEO RIEUL-HIEUH")
    (18382 . "HANGUL SYLLABLE NIEUN YEO PIEUP-SIOS")
    (18383 . "HANGUL SYLLABLE NIEUN YEO CIEUC")
    (18384 . "HANGUL SYLLABLE NIEUN YEO CHIEUCH")
    (18385 . "HANGUL SYLLABLE NIEUN YE SSANGKIYEOK")
    (18386 . "HANGUL SYLLABLE NIEUN YE KIYEOK-SIOS")
    (18387 . "HANGUL SYLLABLE NIEUN YE NIEUN-CIEUC")
    (18388 . "HANGUL SYLLABLE NIEUN YE NIEUN-HIEUH")
    (18389 . "HANGUL SYLLABLE NIEUN YE RIEUL-KIYEOK")
    (18390 . "HANGUL SYLLABLE NIEUN YE RIEUL-MIEUM")
    (18391 . "HANGUL SYLLABLE NIEUN YE RIEUL-PIEUP")
    (18392 . "HANGUL SYLLABLE NIEUN YE RIEUL-SIOS")
    (18393 . "HANGUL SYLLABLE NIEUN YE RIEUL-THIEUTH")
    (18394 . "HANGUL SYLLABLE NIEUN YE RIEUL-PHIEUPH")
    (18395 . "HANGUL SYLLABLE NIEUN YE RIEUL-HIEUH")
    (18396 . "HANGUL SYLLABLE NIEUN YE PIEUP-SIOS")
    (18397 . "HANGUL SYLLABLE NIEUN YE SSANGSIOS")
    (18398 . "HANGUL SYLLABLE NIEUN YE CIEUC")
    (18399 . "HANGUL SYLLABLE NIEUN YE CHIEUCH")
    (18400 . "HANGUL SYLLABLE NIEUN YE KHIEUKH")
    (18401 . "HANGUL SYLLABLE NIEUN YE THIEUTH")
    (18402 . "HANGUL SYLLABLE NIEUN YE PHIEUPH")
    (18403 . "HANGUL SYLLABLE NIEUN YE HIEUH")
    (18404 . "HANGUL SYLLABLE NIEUN O NIEUN-CIEUC")
    (18405 . "HANGUL SYLLABLE NIEUN O RIEUL-KIYEOK")
    (18406 . "HANGUL SYLLABLE NIEUN O RIEUL-PIEUP")
    (18407 . "HANGUL SYLLABLE NIEUN O RIEUL-SIOS")
    (18408 . "HANGUL SYLLABLE NIEUN O RIEUL-THIEUTH")
    (18409 . "HANGUL SYLLABLE NIEUN O RIEUL-PHIEUPH")
    (18410 . "HANGUL SYLLABLE NIEUN O RIEUL-HIEUH")
    (18411 . "HANGUL SYLLABLE NIEUN O PIEUP-SIOS")
    (18412 . "HANGUL SYLLABLE NIEUN O SSANGSIOS")
    (18413 . "HANGUL SYLLABLE NIEUN O CHIEUCH")
    (18414 . "HANGUL SYLLABLE NIEUN WA SSANGKIYEOK")
    (18415 . "HANGUL SYLLABLE NIEUN WA KIYEOK-SIOS")
    (18416 . "HANGUL SYLLABLE NIEUN WA NIEUN-CIEUC")
    (18417 . "HANGUL SYLLABLE NIEUN WA NIEUN-HIEUH")
    (18418 . "HANGUL SYLLABLE NIEUN WA RIEUL-KIYEOK")
    (18419 . "HANGUL SYLLABLE NIEUN WA RIEUL-MIEUM")
    (18420 . "HANGUL SYLLABLE NIEUN WA RIEUL-PIEUP")
    (18421 . "HANGUL SYLLABLE NIEUN WA RIEUL-SIOS")
    (18422 . "HANGUL SYLLABLE NIEUN WA RIEUL-THIEUTH")
    (18423 . "HANGUL SYLLABLE NIEUN WA RIEUL-PHIEUPH")
    (18424 . "HANGUL SYLLABLE NIEUN WA RIEUL-HIEUH")
    (18425 . "HANGUL SYLLABLE NIEUN WA PIEUP-SIOS")
    (18426 . "HANGUL SYLLABLE NIEUN WA CIEUC")
    (18427 . "HANGUL SYLLABLE NIEUN WA CHIEUCH")
    (18428 . "HANGUL SYLLABLE NIEUN WA KHIEUKH")
    (18429 . "HANGUL SYLLABLE NIEUN WA THIEUTH")
    (18430 . "HANGUL SYLLABLE NIEUN WA PHIEUPH")
    (18431 . "HANGUL SYLLABLE NIEUN WA HIEUH")
    (18432 . "HANGUL SYLLABLE NIEUN WAE KIYEOK")
    (18433 . "HANGUL SYLLABLE NIEUN WAE SSANGKIYEOK")
    (18434 . "HANGUL SYLLABLE NIEUN WAE KIYEOK-SIOS")
    (18435 . "HANGUL SYLLABLE NIEUN WAE NIEUN-CIEUC")
    (18436 . "HANGUL SYLLABLE NIEUN WAE NIEUN-HIEUH")
    (18437 . "HANGUL SYLLABLE NIEUN WAE TIKEUT")
    (18438 . "HANGUL SYLLABLE NIEUN WAE RIEUL-KIYEOK")
    (18439 . "HANGUL SYLLABLE NIEUN WAE RIEUL-MIEUM")
    (18440 . "HANGUL SYLLABLE NIEUN WAE RIEUL-PIEUP")
    (18441 . "HANGUL SYLLABLE NIEUN WAE RIEUL-SIOS")
    (18442 . "HANGUL SYLLABLE NIEUN WAE RIEUL-THIEUTH")
    (18443 . "HANGUL SYLLABLE NIEUN WAE RIEUL-PHIEUPH")
    (18444 . "HANGUL SYLLABLE NIEUN WAE RIEUL-HIEUH")
    (18445 . "HANGUL SYLLABLE NIEUN WAE MIEUM")
    (18446 . "HANGUL SYLLABLE NIEUN WAE PIEUP")
    (18447 . "HANGUL SYLLABLE NIEUN WAE PIEUP-SIOS")
    (18448 . "HANGUL SYLLABLE NIEUN WAE SIOS")
    (18449 . "HANGUL SYLLABLE NIEUN WAE IEUNG")
    (18450 . "HANGUL SYLLABLE NIEUN WAE CIEUC")
    (18451 . "HANGUL SYLLABLE NIEUN WAE CHIEUCH")
    (18452 . "HANGUL SYLLABLE NIEUN WAE KHIEUKH")
    (18453 . "HANGUL SYLLABLE NIEUN WAE THIEUTH")
    (18454 . "HANGUL SYLLABLE NIEUN WAE PHIEUPH")
    (18455 . "HANGUL SYLLABLE NIEUN WAE HIEUH")
    (18456 . "HANGUL SYLLABLE NIEUN OE SSANGKIYEOK")
    (18457 . "HANGUL SYLLABLE NIEUN OE KIYEOK-SIOS")
    (18458 . "HANGUL SYLLABLE NIEUN OE NIEUN-CIEUC")
    (18459 . "HANGUL SYLLABLE NIEUN OE NIEUN-HIEUH")
    (18460 . "HANGUL SYLLABLE NIEUN OE RIEUL-KIYEOK")
    (18461 . "HANGUL SYLLABLE NIEUN OE RIEUL-MIEUM")
    (18462 . "HANGUL SYLLABLE NIEUN OE RIEUL-PIEUP")
    (18463 . "HANGUL SYLLABLE NIEUN OE RIEUL-SIOS")
    (18464 . "HANGUL SYLLABLE NIEUN OE RIEUL-THIEUTH")
    (18465 . "HANGUL SYLLABLE NIEUN OE RIEUL-PHIEUPH")
    (18466 . "HANGUL SYLLABLE NIEUN OE RIEUL-HIEUH")
    (18467 . "HANGUL SYLLABLE NIEUN OE PIEUP-SIOS")
    (18468 . "HANGUL SYLLABLE NIEUN OE CIEUC")
    (18469 . "HANGUL SYLLABLE NIEUN OE CHIEUCH")
    (18470 . "HANGUL SYLLABLE NIEUN OE KHIEUKH")
    (18471 . "HANGUL SYLLABLE NIEUN OE THIEUTH")
    (18472 . "HANGUL SYLLABLE NIEUN OE HIEUH")
    (18473 . "HANGUL SYLLABLE NIEUN YO SSANGKIYEOK")
    (18474 . "HANGUL SYLLABLE NIEUN YO KIYEOK-SIOS")
    (18475 . "HANGUL SYLLABLE NIEUN YO NIEUN-CIEUC")
    (18476 . "HANGUL SYLLABLE NIEUN YO NIEUN-HIEUH")
    (18477 . "HANGUL SYLLABLE NIEUN YO RIEUL-KIYEOK")
    (18478 . "HANGUL SYLLABLE NIEUN YO RIEUL-MIEUM")
    (18479 . "HANGUL SYLLABLE NIEUN YO RIEUL-PIEUP")
    (18480 . "HANGUL SYLLABLE NIEUN YO RIEUL-SIOS")
    (18481 . "HANGUL SYLLABLE NIEUN YO RIEUL-THIEUTH")
    (18482 . "HANGUL SYLLABLE NIEUN YO RIEUL-PHIEUPH")
    (18483 . "HANGUL SYLLABLE NIEUN YO RIEUL-HIEUH")
    (18484 . "HANGUL SYLLABLE NIEUN YO PIEUP-SIOS")
    (18485 . "HANGUL SYLLABLE NIEUN YO SSANGSIOS")
    (18486 . "HANGUL SYLLABLE NIEUN YO CIEUC")
    (18487 . "HANGUL SYLLABLE NIEUN YO CHIEUCH")
    (18488 . "HANGUL SYLLABLE NIEUN YO KHIEUKH")
    (18489 . "HANGUL SYLLABLE NIEUN YO THIEUTH")
    (18490 . "HANGUL SYLLABLE NIEUN YO PHIEUPH")
    (18491 . "HANGUL SYLLABLE NIEUN YO HIEUH")
    (18492 . "HANGUL SYLLABLE NIEUN U SSANGKIYEOK")
    (18493 . "HANGUL SYLLABLE NIEUN U KIYEOK-SIOS")
    (18494 . "HANGUL SYLLABLE NIEUN U NIEUN-CIEUC")
    (18495 . "HANGUL SYLLABLE NIEUN U NIEUN-HIEUH")
    (18496 . "HANGUL SYLLABLE NIEUN U RIEUL-MIEUM")
    (18497 . "HANGUL SYLLABLE NIEUN U RIEUL-PIEUP")
    (18498 . "HANGUL SYLLABLE NIEUN U RIEUL-THIEUTH")
    (18499 . "HANGUL SYLLABLE NIEUN U RIEUL-PHIEUPH")
    (18500 . "HANGUL SYLLABLE NIEUN U PIEUP-SIOS")
    (18501 . "HANGUL SYLLABLE NIEUN U SSANGSIOS")
    (18502 . "HANGUL SYLLABLE NIEUN U CIEUC")
    (18503 . "HANGUL SYLLABLE NIEUN U KHIEUKH")
    (18504 . "HANGUL SYLLABLE NIEUN U THIEUTH")
    (18505 . "HANGUL SYLLABLE NIEUN U HIEUH")
    (18506 . "HANGUL SYLLABLE NIEUN WEO SSANGKIYEOK")
    (18507 . "HANGUL SYLLABLE NIEUN WEO KIYEOK-SIOS")
    (18508 . "HANGUL SYLLABLE NIEUN WEO NIEUN-CIEUC")
    (18509 . "HANGUL SYLLABLE NIEUN WEO NIEUN-HIEUH")
    (18510 . "HANGUL SYLLABLE NIEUN WEO RIEUL-KIYEOK")
    (18511 . "HANGUL SYLLABLE NIEUN WEO RIEUL-MIEUM")
    (18512 . "HANGUL SYLLABLE NIEUN WEO RIEUL-PIEUP")
    (18513 . "HANGUL SYLLABLE NIEUN WEO RIEUL-SIOS")
    (18514 . "HANGUL SYLLABLE NIEUN WEO RIEUL-THIEUTH")
    (18515 . "HANGUL SYLLABLE NIEUN WEO RIEUL-PHIEUPH")
    (18516 . "HANGUL SYLLABLE NIEUN WEO RIEUL-HIEUH")
    (18517 . "HANGUL SYLLABLE NIEUN WEO PIEUP-SIOS")
    (18518 . "HANGUL SYLLABLE NIEUN WEO CIEUC")
    (18519 . "HANGUL SYLLABLE NIEUN WEO CHIEUCH")
    (18520 . "HANGUL SYLLABLE NIEUN WEO KHIEUKH")
    (18521 . "HANGUL SYLLABLE NIEUN WEO THIEUTH")
    (18522 . "HANGUL SYLLABLE NIEUN WEO PHIEUPH")
    (18523 . "HANGUL SYLLABLE NIEUN WEO HIEUH")
    (18524 . "HANGUL SYLLABLE NIEUN WE SSANGKIYEOK")
    (18525 . "HANGUL SYLLABLE NIEUN WE KIYEOK-SIOS")
    (18526 . "HANGUL SYLLABLE NIEUN WE NIEUN-CIEUC")
    (18527 . "HANGUL SYLLABLE NIEUN WE NIEUN-HIEUH")
    (18528 . "HANGUL SYLLABLE NIEUN WE RIEUL-KIYEOK")
    (18529 . "HANGUL SYLLABLE NIEUN WE RIEUL-MIEUM")
    (18530 . "HANGUL SYLLABLE NIEUN WE RIEUL-PIEUP")
    (18531 . "HANGUL SYLLABLE NIEUN WE RIEUL-SIOS")
    (18532 . "HANGUL SYLLABLE NIEUN WE RIEUL-THIEUTH")
    (18533 . "HANGUL SYLLABLE NIEUN WE RIEUL-PHIEUPH")
    (18534 . "HANGUL SYLLABLE NIEUN WE RIEUL-HIEUH")
    (18535 . "HANGUL SYLLABLE NIEUN WE PIEUP-SIOS")
    (18536 . "HANGUL SYLLABLE NIEUN WE CIEUC")
    (18537 . "HANGUL SYLLABLE NIEUN WE CHIEUCH")
    (18538 . "HANGUL SYLLABLE NIEUN WE KHIEUKH")
    (18539 . "HANGUL SYLLABLE NIEUN WE THIEUTH")
    (18540 . "HANGUL SYLLABLE NIEUN WE PHIEUPH")
    (18541 . "HANGUL SYLLABLE NIEUN WE HIEUH")
    (18542 . "HANGUL SYLLABLE NIEUN WI SSANGKIYEOK")
    (18543 . "HANGUL SYLLABLE NIEUN WI KIYEOK-SIOS")
    (18544 . "HANGUL SYLLABLE NIEUN WI NIEUN-CIEUC")
    (18545 . "HANGUL SYLLABLE NIEUN WI NIEUN-HIEUH")
    (18546 . "HANGUL SYLLABLE NIEUN WI RIEUL-KIYEOK")
    (18547 . "HANGUL SYLLABLE NIEUN WI RIEUL-MIEUM")
    (18548 . "HANGUL SYLLABLE NIEUN WI RIEUL-PIEUP")
    (18549 . "HANGUL SYLLABLE NIEUN WI RIEUL-SIOS")
    (18550 . "HANGUL SYLLABLE NIEUN WI RIEUL-THIEUTH")
    (18551 . "HANGUL SYLLABLE NIEUN WI RIEUL-PHIEUPH")
    (18552 . "HANGUL SYLLABLE NIEUN WI RIEUL-HIEUH")
    (18553 . "HANGUL SYLLABLE NIEUN WI PIEUP-SIOS")
    (18554 . "HANGUL SYLLABLE NIEUN WI SSANGSIOS")
    (18555 . "HANGUL SYLLABLE NIEUN WI CIEUC")
    (18556 . "HANGUL SYLLABLE NIEUN WI KHIEUKH")
    (18557 . "HANGUL SYLLABLE NIEUN WI THIEUTH")
    (18558 . "HANGUL SYLLABLE NIEUN WI PHIEUPH")
    (18559 . "HANGUL SYLLABLE NIEUN WI HIEUH")
    (18560 . "HANGUL SYLLABLE NIEUN YU SSANGKIYEOK")
    (18561 . "HANGUL SYLLABLE NIEUN YU KIYEOK-SIOS")
    (18562 . "HANGUL SYLLABLE NIEUN YU NIEUN-CIEUC")
    (18563 . "HANGUL SYLLABLE NIEUN YU NIEUN-HIEUH")
    (18564 . "HANGUL SYLLABLE NIEUN YU RIEUL-KIYEOK")
    (18565 . "HANGUL SYLLABLE NIEUN YU RIEUL-MIEUM")
    (18566 . "HANGUL SYLLABLE NIEUN YU RIEUL-PIEUP")
    (18567 . "HANGUL SYLLABLE NIEUN YU RIEUL-SIOS")
    (18568 . "HANGUL SYLLABLE NIEUN YU RIEUL-THIEUTH")
    (18569 . "HANGUL SYLLABLE NIEUN YU RIEUL-PHIEUPH")
    (18570 . "HANGUL SYLLABLE NIEUN YU RIEUL-HIEUH")
    (18571 . "HANGUL SYLLABLE NIEUN YU PIEUP-SIOS")
    (18572 . "HANGUL SYLLABLE NIEUN YU SSANGSIOS")
    (18573 . "HANGUL SYLLABLE NIEUN YU KHIEUKH")
    (18574 . "HANGUL SYLLABLE NIEUN YU THIEUTH")
    (18575 . "HANGUL SYLLABLE NIEUN YU PHIEUPH")
    (18576 . "HANGUL SYLLABLE NIEUN YU HIEUH")
    (18577 . "HANGUL SYLLABLE NIEUN EU SSANGKIYEOK")
    (18578 . "HANGUL SYLLABLE NIEUN EU KIYEOK-SIOS")
    (18579 . "HANGUL SYLLABLE NIEUN EU NIEUN-CIEUC")
    (18580 . "HANGUL SYLLABLE NIEUN EU NIEUN-HIEUH")
    (18581 . "HANGUL SYLLABLE NIEUN EU RIEUL-PIEUP")
    (18582 . "HANGUL SYLLABLE NIEUN EU RIEUL-THIEUTH")
    (18583 . "HANGUL SYLLABLE NIEUN EU RIEUL-PHIEUPH")
    (18584 . "HANGUL SYLLABLE NIEUN EU RIEUL-HIEUH")
    (18585 . "HANGUL SYLLABLE NIEUN EU PIEUP-SIOS")
    (18586 . "HANGUL SYLLABLE NIEUN EU SSANGSIOS")
    (18587 . "HANGUL SYLLABLE NIEUN EU KHIEUKH")
    (18588 . "HANGUL SYLLABLE NIEUN EU THIEUTH")
    (18589 . "HANGUL SYLLABLE NIEUN YI SSANGKIYEOK")
    (18590 . "HANGUL SYLLABLE NIEUN YI KIYEOK-SIOS")
    (18591 . "HANGUL SYLLABLE NIEUN YI NIEUN-CIEUC")
    (18592 . "HANGUL SYLLABLE NIEUN YI NIEUN-HIEUH")
    (18593 . "HANGUL SYLLABLE NIEUN YI RIEUL-KIYEOK")
    (18594 . "HANGUL SYLLABLE NIEUN YI RIEUL-MIEUM")
    (18595 . "HANGUL SYLLABLE NIEUN YI RIEUL-PIEUP")
    (18596 . "HANGUL SYLLABLE NIEUN YI RIEUL-SIOS")
    (18597 . "HANGUL SYLLABLE NIEUN YI RIEUL-THIEUTH")
    (18598 . "HANGUL SYLLABLE NIEUN YI RIEUL-PHIEUPH")
    (18599 . "HANGUL SYLLABLE NIEUN YI RIEUL-HIEUH")
    (18600 . "HANGUL SYLLABLE NIEUN YI PIEUP-SIOS")
    (18601 . "HANGUL SYLLABLE NIEUN YI CIEUC")
    (18602 . "HANGUL SYLLABLE NIEUN YI CHIEUCH")
    (18603 . "HANGUL SYLLABLE NIEUN YI KHIEUKH")
    (18604 . "HANGUL SYLLABLE NIEUN YI THIEUTH")
    (18605 . "HANGUL SYLLABLE NIEUN YI PHIEUPH")
    (18606 . "HANGUL SYLLABLE NIEUN YI HIEUH")
    (18607 . "HANGUL SYLLABLE NIEUN I SSANGKIYEOK")
    (18608 . "HANGUL SYLLABLE NIEUN I KIYEOK-SIOS")
    (18609 . "HANGUL SYLLABLE NIEUN I NIEUN-CIEUC")
    (18610 . "HANGUL SYLLABLE NIEUN I NIEUN-HIEUH")
    (18611 . "HANGUL SYLLABLE NIEUN I RIEUL-THIEUTH")
    (18612 . "HANGUL SYLLABLE NIEUN I RIEUL-PHIEUPH")
    (18613 . "HANGUL SYLLABLE NIEUN I RIEUL-HIEUH")
    (18614 . "HANGUL SYLLABLE NIEUN I PIEUP-SIOS")
    (18615 . "HANGUL SYLLABLE NIEUN I SSANGSIOS")
    (18616 . "HANGUL SYLLABLE NIEUN I CHIEUCH")
    (18617 . "HANGUL SYLLABLE NIEUN I THIEUTH")
    (18618 . "HANGUL SYLLABLE TIKEUT A KIYEOK-SIOS")
    (18619 . "HANGUL SYLLABLE TIKEUT A NIEUN-CIEUC")
    (18620 . "HANGUL SYLLABLE TIKEUT A NIEUN-HIEUH")
    (18621 . "HANGUL SYLLABLE TIKEUT A RIEUL-THIEUTH")
    (18622 . "HANGUL SYLLABLE TIKEUT A PHIEUPH")
    (18623 . "HANGUL SYLLABLE TIKEUT AE SSANGKIYEOK")
    (18624 . "HANGUL SYLLABLE TIKEUT AE KIYEOK-SIOS")
    (18625 . "HANGUL SYLLABLE TIKEUT AE NIEUN-CIEUC")
    (18626 . "HANGUL SYLLABLE TIKEUT AE NIEUN-HIEUH")
    (18627 . "HANGUL SYLLABLE TIKEUT AE RIEUL-KIYEOK")
    (18628 . "HANGUL SYLLABLE TIKEUT AE RIEUL-MIEUM")
    (18629 . "HANGUL SYLLABLE TIKEUT AE RIEUL-PIEUP")
    (18630 . "HANGUL SYLLABLE TIKEUT AE RIEUL-SIOS")
    (18631 . "HANGUL SYLLABLE TIKEUT AE RIEUL-THIEUTH")
    (18632 . "HANGUL SYLLABLE TIKEUT AE RIEUL-PHIEUPH")
    (18633 . "HANGUL SYLLABLE TIKEUT AE RIEUL-HIEUH")
    (18634 . "HANGUL SYLLABLE TIKEUT AE PIEUP-SIOS")
    (18635 . "HANGUL SYLLABLE TIKEUT AE CIEUC")
    (18636 . "HANGUL SYLLABLE TIKEUT AE CHIEUCH")
    (18637 . "HANGUL SYLLABLE TIKEUT AE KHIEUKH")
    (18638 . "HANGUL SYLLABLE TIKEUT AE PHIEUPH")
    (18639 . "HANGUL SYLLABLE TIKEUT AE HIEUH")
    (18640 . "HANGUL SYLLABLE TIKEUT YA SSANGKIYEOK")
    (18641 . "HANGUL SYLLABLE TIKEUT YA KIYEOK-SIOS")
    (18642 . "HANGUL SYLLABLE TIKEUT YA NIEUN-CIEUC")
    (18643 . "HANGUL SYLLABLE TIKEUT YA NIEUN-HIEUH")
    (18644 . "HANGUL SYLLABLE TIKEUT YA TIKEUT")
    (18645 . "HANGUL SYLLABLE TIKEUT YA RIEUL-KIYEOK")
    (18646 . "HANGUL SYLLABLE TIKEUT YA RIEUL-MIEUM")
    (18647 . "HANGUL SYLLABLE TIKEUT YA RIEUL-SIOS")
    (18648 . "HANGUL SYLLABLE TIKEUT YA RIEUL-THIEUTH")
    (18649 . "HANGUL SYLLABLE TIKEUT YA RIEUL-PHIEUPH")
    (18650 . "HANGUL SYLLABLE TIKEUT YA RIEUL-HIEUH")
    (18651 . "HANGUL SYLLABLE TIKEUT YA PIEUP-SIOS")
    (18652 . "HANGUL SYLLABLE TIKEUT YA SSANGSIOS")
    (18653 . "HANGUL SYLLABLE TIKEUT YA CIEUC")
    (18654 . "HANGUL SYLLABLE TIKEUT YA CHIEUCH")
    (18655 . "HANGUL SYLLABLE TIKEUT YA KHIEUKH")
    (18656 . "HANGUL SYLLABLE TIKEUT YA THIEUTH")
    (18657 . "HANGUL SYLLABLE TIKEUT YA PHIEUPH")
    (18658 . "HANGUL SYLLABLE TIKEUT YA HIEUH")
    (18659 . "HANGUL SYLLABLE TIKEUT YAE KIYEOK")
    (18660 . "HANGUL SYLLABLE TIKEUT YAE SSANGKIYEOK")
    (18661 . "HANGUL SYLLABLE TIKEUT YAE KIYEOK-SIOS")
    (18662 . "HANGUL SYLLABLE TIKEUT YAE NIEUN-CIEUC")
    (18663 . "HANGUL SYLLABLE TIKEUT YAE NIEUN-HIEUH")
    (18664 . "HANGUL SYLLABLE TIKEUT YAE TIKEUT")
    (18665 . "HANGUL SYLLABLE TIKEUT YAE RIEUL")
    (18666 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-KIYEOK")
    (18667 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-MIEUM")
    (18668 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-PIEUP")
    (18669 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-SIOS")
    (18670 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-THIEUTH")
    (18671 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-PHIEUPH")
    (18672 . "HANGUL SYLLABLE TIKEUT YAE RIEUL-HIEUH")
    (18673 . "HANGUL SYLLABLE TIKEUT YAE MIEUM")
    (18674 . "HANGUL SYLLABLE TIKEUT YAE PIEUP")
    (18675 . "HANGUL SYLLABLE TIKEUT YAE PIEUP-SIOS")
    (18676 . "HANGUL SYLLABLE TIKEUT YAE SIOS")
    (18677 . "HANGUL SYLLABLE TIKEUT YAE SSANGSIOS")
    (18678 . "HANGUL SYLLABLE TIKEUT YAE CIEUC")
    (18679 . "HANGUL SYLLABLE TIKEUT YAE CHIEUCH")
    (18680 . "HANGUL SYLLABLE TIKEUT YAE KHIEUKH")
    (18681 . "HANGUL SYLLABLE TIKEUT YAE THIEUTH")
    (18682 . "HANGUL SYLLABLE TIKEUT YAE PHIEUPH")
    (18683 . "HANGUL SYLLABLE TIKEUT YAE HIEUH")
    (18684 . "HANGUL SYLLABLE TIKEUT EO KIYEOK-SIOS")
    (18685 . "HANGUL SYLLABLE TIKEUT EO NIEUN-HIEUH")
    (18686 . "HANGUL SYLLABLE TIKEUT EO RIEUL-KIYEOK")
    (18687 . "HANGUL SYLLABLE TIKEUT EO RIEUL-SIOS")
    (18688 . "HANGUL SYLLABLE TIKEUT EO RIEUL-THIEUTH")
    (18689 . "HANGUL SYLLABLE TIKEUT EO RIEUL-PHIEUPH")
    (18690 . "HANGUL SYLLABLE TIKEUT EO RIEUL-HIEUH")
    (18691 . "HANGUL SYLLABLE TIKEUT EO KHIEUKH")
    (18692 . "HANGUL SYLLABLE TIKEUT E SSANGKIYEOK")
    (18693 . "HANGUL SYLLABLE TIKEUT E KIYEOK-SIOS")
    (18694 . "HANGUL SYLLABLE TIKEUT E NIEUN-CIEUC")
    (18695 . "HANGUL SYLLABLE TIKEUT E NIEUN-HIEUH")
    (18696 . "HANGUL SYLLABLE TIKEUT E RIEUL-KIYEOK")
    (18697 . "HANGUL SYLLABLE TIKEUT E RIEUL-MIEUM")
    (18698 . "HANGUL SYLLABLE TIKEUT E RIEUL-PIEUP")
    (18699 . "HANGUL SYLLABLE TIKEUT E RIEUL-SIOS")
    (18700 . "HANGUL SYLLABLE TIKEUT E RIEUL-THIEUTH")
    (18701 . "HANGUL SYLLABLE TIKEUT E RIEUL-PHIEUPH")
    (18702 . "HANGUL SYLLABLE TIKEUT E RIEUL-HIEUH")
    (18703 . "HANGUL SYLLABLE TIKEUT E PIEUP-SIOS")
    (18704 . "HANGUL SYLLABLE TIKEUT E CHIEUCH")
    (18705 . "HANGUL SYLLABLE TIKEUT E KHIEUKH")
    (18706 . "HANGUL SYLLABLE TIKEUT E THIEUTH")
    (18707 . "HANGUL SYLLABLE TIKEUT E HIEUH")
    (18708 . "HANGUL SYLLABLE TIKEUT YEO SSANGKIYEOK")
    (18709 . "HANGUL SYLLABLE TIKEUT YEO NIEUN-CIEUC")
    (18710 . "HANGUL SYLLABLE TIKEUT YEO NIEUN-HIEUH")
    (18711 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-KIYEOK")
    (18712 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-MIEUM")
    (18713 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-PIEUP")
    (18714 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-THIEUTH")
    (18715 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-PHIEUPH")
    (18716 . "HANGUL SYLLABLE TIKEUT YEO RIEUL-HIEUH")
    (18717 . "HANGUL SYLLABLE TIKEUT YEO PIEUP-SIOS")
    (18718 . "HANGUL SYLLABLE TIKEUT YEO CIEUC")
    (18719 . "HANGUL SYLLABLE TIKEUT YEO CHIEUCH")
    (18720 . "HANGUL SYLLABLE TIKEUT YEO KHIEUKH")
    (18721 . "HANGUL SYLLABLE TIKEUT YEO THIEUTH")
    (18722 . "HANGUL SYLLABLE TIKEUT YEO PHIEUPH")
    (18723 . "HANGUL SYLLABLE TIKEUT YEO HIEUH")
    (18724 . "HANGUL SYLLABLE TIKEUT YE SSANGKIYEOK")
    (18725 . "HANGUL SYLLABLE TIKEUT YE KIYEOK-SIOS")
    (18726 . "HANGUL SYLLABLE TIKEUT YE NIEUN-CIEUC")
    (18727 . "HANGUL SYLLABLE TIKEUT YE NIEUN-HIEUH")
    (18728 . "HANGUL SYLLABLE TIKEUT YE TIKEUT")
    (18729 . "HANGUL SYLLABLE TIKEUT YE RIEUL-KIYEOK")
    (18730 . "HANGUL SYLLABLE TIKEUT YE RIEUL-MIEUM")
    (18731 . "HANGUL SYLLABLE TIKEUT YE RIEUL-PIEUP")
    (18732 . "HANGUL SYLLABLE TIKEUT YE RIEUL-SIOS")
    (18733 . "HANGUL SYLLABLE TIKEUT YE RIEUL-THIEUTH")
    (18734 . "HANGUL SYLLABLE TIKEUT YE RIEUL-PHIEUPH")
    (18735 . "HANGUL SYLLABLE TIKEUT YE RIEUL-HIEUH")
    (18736 . "HANGUL SYLLABLE TIKEUT YE PIEUP-SIOS")
    (18737 . "HANGUL SYLLABLE TIKEUT YE SSANGSIOS")
    (18738 . "HANGUL SYLLABLE TIKEUT YE CIEUC")
    (18739 . "HANGUL SYLLABLE TIKEUT YE CHIEUCH")
    (18740 . "HANGUL SYLLABLE TIKEUT YE KHIEUKH")
    (18741 . "HANGUL SYLLABLE TIKEUT YE THIEUTH")
    (18742 . "HANGUL SYLLABLE TIKEUT YE PHIEUPH")
    (18743 . "HANGUL SYLLABLE TIKEUT YE HIEUH")
    (18744 . "HANGUL SYLLABLE TIKEUT O NIEUN-CIEUC")
    (18745 . "HANGUL SYLLABLE TIKEUT O NIEUN-HIEUH")
    (18746 . "HANGUL SYLLABLE TIKEUT O RIEUL-KIYEOK")
    (18747 . "HANGUL SYLLABLE TIKEUT O RIEUL-THIEUTH")
    (18748 . "HANGUL SYLLABLE TIKEUT O RIEUL-PHIEUPH")
    (18749 . "HANGUL SYLLABLE TIKEUT O PIEUP-SIOS")
    (18750 . "HANGUL SYLLABLE TIKEUT O SSANGSIOS")
    (18751 . "HANGUL SYLLABLE TIKEUT WA SSANGKIYEOK")
    (18752 . "HANGUL SYLLABLE TIKEUT WA KIYEOK-SIOS")
    (18753 . "HANGUL SYLLABLE TIKEUT WA NIEUN-CIEUC")
    (18754 . "HANGUL SYLLABLE TIKEUT WA NIEUN-HIEUH")
    (18755 . "HANGUL SYLLABLE TIKEUT WA TIKEUT")
    (18756 . "HANGUL SYLLABLE TIKEUT WA RIEUL-KIYEOK")
    (18757 . "HANGUL SYLLABLE TIKEUT WA RIEUL-MIEUM")
    (18758 . "HANGUL SYLLABLE TIKEUT WA RIEUL-PIEUP")
    (18759 . "HANGUL SYLLABLE TIKEUT WA RIEUL-SIOS")
    (18760 . "HANGUL SYLLABLE TIKEUT WA RIEUL-THIEUTH")
    (18761 . "HANGUL SYLLABLE TIKEUT WA RIEUL-PHIEUPH")
    (18762 . "HANGUL SYLLABLE TIKEUT WA RIEUL-HIEUH")
    (18763 . "HANGUL SYLLABLE TIKEUT WA PIEUP-SIOS")
    (18764 . "HANGUL SYLLABLE TIKEUT WA CIEUC")
    (18765 . "HANGUL SYLLABLE TIKEUT WA CHIEUCH")
    (18766 . "HANGUL SYLLABLE TIKEUT WA KHIEUKH")
    (18767 . "HANGUL SYLLABLE TIKEUT WA THIEUTH")
    (18768 . "HANGUL SYLLABLE TIKEUT WA PHIEUPH")
    (18769 . "HANGUL SYLLABLE TIKEUT WA HIEUH")
    (18770 . "HANGUL SYLLABLE TIKEUT WAE KIYEOK")
    (18771 . "HANGUL SYLLABLE TIKEUT WAE SSANGKIYEOK")
    (18772 . "HANGUL SYLLABLE TIKEUT WAE KIYEOK-SIOS")
    (18773 . "HANGUL SYLLABLE TIKEUT WAE NIEUN-CIEUC")
    (18774 . "HANGUL SYLLABLE TIKEUT WAE NIEUN-HIEUH")
    (18775 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-KIYEOK")
    (18776 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-MIEUM")
    (18777 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-PIEUP")
    (18778 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-SIOS")
    (18779 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-THIEUTH")
    (18780 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-PHIEUPH")
    (18781 . "HANGUL SYLLABLE TIKEUT WAE RIEUL-HIEUH")
    (18782 . "HANGUL SYLLABLE TIKEUT WAE PIEUP-SIOS")
    (18783 . "HANGUL SYLLABLE TIKEUT WAE IEUNG")
    (18784 . "HANGUL SYLLABLE TIKEUT WAE CIEUC")
    (18785 . "HANGUL SYLLABLE TIKEUT WAE CHIEUCH")
    (18786 . "HANGUL SYLLABLE TIKEUT WAE KHIEUKH")
    (18787 . "HANGUL SYLLABLE TIKEUT WAE THIEUTH")
    (18788 . "HANGUL SYLLABLE TIKEUT WAE PHIEUPH")
    (18789 . "HANGUL SYLLABLE TIKEUT WAE HIEUH")
    (18790 . "HANGUL SYLLABLE TIKEUT OE SSANGKIYEOK")
    (18791 . "HANGUL SYLLABLE TIKEUT OE KIYEOK-SIOS")
    (18792 . "HANGUL SYLLABLE TIKEUT OE NIEUN-CIEUC")
    (18793 . "HANGUL SYLLABLE TIKEUT OE NIEUN-HIEUH")
    (18794 . "HANGUL SYLLABLE TIKEUT OE RIEUL-KIYEOK")
    (18795 . "HANGUL SYLLABLE TIKEUT OE RIEUL-MIEUM")
    (18796 . "HANGUL SYLLABLE TIKEUT OE RIEUL-PIEUP")
    (18797 . "HANGUL SYLLABLE TIKEUT OE RIEUL-THIEUTH")
    (18798 . "HANGUL SYLLABLE TIKEUT OE RIEUL-PHIEUPH")
    (18799 . "HANGUL SYLLABLE TIKEUT OE RIEUL-HIEUH")
    (18800 . "HANGUL SYLLABLE TIKEUT OE PIEUP-SIOS")
    (18801 . "HANGUL SYLLABLE TIKEUT OE CIEUC")
    (18802 . "HANGUL SYLLABLE TIKEUT OE CHIEUCH")
    (18803 . "HANGUL SYLLABLE TIKEUT OE KHIEUKH")
    (18804 . "HANGUL SYLLABLE TIKEUT OE THIEUTH")
    (18805 . "HANGUL SYLLABLE TIKEUT OE PHIEUPH")
    (18806 . "HANGUL SYLLABLE TIKEUT OE HIEUH")
    (18807 . "HANGUL SYLLABLE TIKEUT YO SSANGKIYEOK")
    (18808 . "HANGUL SYLLABLE TIKEUT YO KIYEOK-SIOS")
    (18809 . "HANGUL SYLLABLE TIKEUT YO NIEUN-CIEUC")
    (18810 . "HANGUL SYLLABLE TIKEUT YO NIEUN-HIEUH")
    (18811 . "HANGUL SYLLABLE TIKEUT YO RIEUL-KIYEOK")
    (18812 . "HANGUL SYLLABLE TIKEUT YO RIEUL-MIEUM")
    (18813 . "HANGUL SYLLABLE TIKEUT YO RIEUL-PIEUP")
    (18814 . "HANGUL SYLLABLE TIKEUT YO RIEUL-SIOS")
    (18815 . "HANGUL SYLLABLE TIKEUT YO RIEUL-THIEUTH")
    (18816 . "HANGUL SYLLABLE TIKEUT YO RIEUL-PHIEUPH")
    (18817 . "HANGUL SYLLABLE TIKEUT YO RIEUL-HIEUH")
    (18818 . "HANGUL SYLLABLE TIKEUT YO PIEUP-SIOS")
    (18819 . "HANGUL SYLLABLE TIKEUT YO SSANGSIOS")
    (18820 . "HANGUL SYLLABLE TIKEUT YO CIEUC")
    (18821 . "HANGUL SYLLABLE TIKEUT YO CHIEUCH")
    (18822 . "HANGUL SYLLABLE TIKEUT YO KHIEUKH")
    (18823 . "HANGUL SYLLABLE TIKEUT YO THIEUTH")
    (18824 . "HANGUL SYLLABLE TIKEUT YO PHIEUPH")
    (18825 . "HANGUL SYLLABLE TIKEUT U SSANGKIYEOK")
    (18826 . "HANGUL SYLLABLE TIKEUT U KIYEOK-SIOS")
    (18827 . "HANGUL SYLLABLE TIKEUT U NIEUN-CIEUC")
    (18828 . "HANGUL SYLLABLE TIKEUT U NIEUN-HIEUH")
    (18829 . "HANGUL SYLLABLE TIKEUT U RIEUL-SIOS")
    (18830 . "HANGUL SYLLABLE TIKEUT U RIEUL-THIEUTH")
    (18831 . "HANGUL SYLLABLE TIKEUT U RIEUL-PHIEUPH")
    (18832 . "HANGUL SYLLABLE TIKEUT U RIEUL-HIEUH")
    (18833 . "HANGUL SYLLABLE TIKEUT U PIEUP-SIOS")
    (18834 . "HANGUL SYLLABLE TIKEUT U SSANGSIOS")
    (18835 . "HANGUL SYLLABLE TIKEUT U CIEUC")
    (18836 . "HANGUL SYLLABLE TIKEUT U CHIEUCH")
    (18837 . "HANGUL SYLLABLE TIKEUT U KHIEUKH")
    (18838 . "HANGUL SYLLABLE TIKEUT U THIEUTH")
    (18839 . "HANGUL SYLLABLE TIKEUT U HIEUH")
    (18840 . "HANGUL SYLLABLE TIKEUT WEO SSANGKIYEOK")
    (18841 . "HANGUL SYLLABLE TIKEUT WEO KIYEOK-SIOS")
    (18842 . "HANGUL SYLLABLE TIKEUT WEO NIEUN-CIEUC")
    (18843 . "HANGUL SYLLABLE TIKEUT WEO NIEUN-HIEUH")
    (18844 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-KIYEOK")
    (18845 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-MIEUM")
    (18846 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-PIEUP")
    (18847 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-SIOS")
    (18848 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-THIEUTH")
    (18849 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-PHIEUPH")
    (18850 . "HANGUL SYLLABLE TIKEUT WEO RIEUL-HIEUH")
    (18851 . "HANGUL SYLLABLE TIKEUT WEO PIEUP-SIOS")
    (18852 . "HANGUL SYLLABLE TIKEUT WEO CIEUC")
    (18853 . "HANGUL SYLLABLE TIKEUT WEO CHIEUCH")
    (18854 . "HANGUL SYLLABLE TIKEUT WEO KHIEUKH")
    (18855 . "HANGUL SYLLABLE TIKEUT WEO THIEUTH")
    (18856 . "HANGUL SYLLABLE TIKEUT WEO PHIEUPH")
    (18857 . "HANGUL SYLLABLE TIKEUT WEO HIEUH")
    (18858 . "HANGUL SYLLABLE TIKEUT WE SSANGKIYEOK")
    (18859 . "HANGUL SYLLABLE TIKEUT WE KIYEOK-SIOS")
    (18860 . "HANGUL SYLLABLE TIKEUT WE NIEUN-CIEUC")
    (18861 . "HANGUL SYLLABLE TIKEUT WE NIEUN-HIEUH")
    (18862 . "HANGUL SYLLABLE TIKEUT WE RIEUL-KIYEOK")
    (18863 . "HANGUL SYLLABLE TIKEUT WE RIEUL-MIEUM")
    (18864 . "HANGUL SYLLABLE TIKEUT WE RIEUL-PIEUP")
    (18865 . "HANGUL SYLLABLE TIKEUT WE RIEUL-SIOS")
    (18866 . "HANGUL SYLLABLE TIKEUT WE RIEUL-THIEUTH")
    (18867 . "HANGUL SYLLABLE TIKEUT WE RIEUL-PHIEUPH")
    (18868 . "HANGUL SYLLABLE TIKEUT WE RIEUL-HIEUH")
    (18869 . "HANGUL SYLLABLE TIKEUT WE PIEUP-SIOS")
    (18870 . "HANGUL SYLLABLE TIKEUT WE CIEUC")
    (18871 . "HANGUL SYLLABLE TIKEUT WE CHIEUCH")
    (18872 . "HANGUL SYLLABLE TIKEUT WE KHIEUKH")
    (18873 . "HANGUL SYLLABLE TIKEUT WE THIEUTH")
    (18874 . "HANGUL SYLLABLE TIKEUT WE PHIEUPH")
    (18875 . "HANGUL SYLLABLE TIKEUT WE HIEUH")
    (18876 . "HANGUL SYLLABLE TIKEUT WI SSANGKIYEOK")
    (18877 . "HANGUL SYLLABLE TIKEUT WI KIYEOK-SIOS")
    (18878 . "HANGUL SYLLABLE TIKEUT WI NIEUN-CIEUC")
    (18879 . "HANGUL SYLLABLE TIKEUT WI NIEUN-HIEUH")
    (18880 . "HANGUL SYLLABLE TIKEUT WI RIEUL-KIYEOK")
    (18881 . "HANGUL SYLLABLE TIKEUT WI RIEUL-MIEUM")
    (18882 . "HANGUL SYLLABLE TIKEUT WI RIEUL-PIEUP")
    (18883 . "HANGUL SYLLABLE TIKEUT WI RIEUL-SIOS")
    (18884 . "HANGUL SYLLABLE TIKEUT WI RIEUL-THIEUTH")
    (18885 . "HANGUL SYLLABLE TIKEUT WI RIEUL-PHIEUPH")
    (18886 . "HANGUL SYLLABLE TIKEUT WI RIEUL-HIEUH")
    (18887 . "HANGUL SYLLABLE TIKEUT WI PIEUP-SIOS")
    (18888 . "HANGUL SYLLABLE TIKEUT WI CIEUC")
    (18889 . "HANGUL SYLLABLE TIKEUT WI CHIEUCH")
    (18890 . "HANGUL SYLLABLE TIKEUT WI KHIEUKH")
    (18891 . "HANGUL SYLLABLE TIKEUT WI THIEUTH")
    (18892 . "HANGUL SYLLABLE TIKEUT WI PHIEUPH")
    (18893 . "HANGUL SYLLABLE TIKEUT WI HIEUH")
    (18894 . "HANGUL SYLLABLE TIKEUT YU SSANGKIYEOK")
    (18895 . "HANGUL SYLLABLE TIKEUT YU KIYEOK-SIOS")
    (18896 . "HANGUL SYLLABLE TIKEUT YU NIEUN-CIEUC")
    (18897 . "HANGUL SYLLABLE TIKEUT YU NIEUN-HIEUH")
    (18898 . "HANGUL SYLLABLE TIKEUT YU TIKEUT")
    (18899 . "HANGUL SYLLABLE TIKEUT YU RIEUL-KIYEOK")
    (18900 . "HANGUL SYLLABLE TIKEUT YU RIEUL-MIEUM")
    (18901 . "HANGUL SYLLABLE TIKEUT YU RIEUL-PIEUP")
    (18902 . "HANGUL SYLLABLE TIKEUT YU RIEUL-SIOS")
    (18903 . "HANGUL SYLLABLE TIKEUT YU RIEUL-THIEUTH")
    (18904 . "HANGUL SYLLABLE TIKEUT YU RIEUL-PHIEUPH")
    (18905 . "HANGUL SYLLABLE TIKEUT YU RIEUL-HIEUH")
    (18906 . "HANGUL SYLLABLE TIKEUT YU PIEUP-SIOS")
    (18907 . "HANGUL SYLLABLE TIKEUT YU SSANGSIOS")
    (18908 . "HANGUL SYLLABLE TIKEUT YU CIEUC")
    (18909 . "HANGUL SYLLABLE TIKEUT YU CHIEUCH")
    (18910 . "HANGUL SYLLABLE TIKEUT YU KHIEUKH")
    (18911 . "HANGUL SYLLABLE TIKEUT YU THIEUTH")
    (18912 . "HANGUL SYLLABLE TIKEUT YU PHIEUPH")
    (18913 . "HANGUL SYLLABLE TIKEUT YU HIEUH")
    (18914 . "HANGUL SYLLABLE TIKEUT EU SSANGKIYEOK")
    (18915 . "HANGUL SYLLABLE TIKEUT EU KIYEOK-SIOS")
    (18916 . "HANGUL SYLLABLE TIKEUT EU NIEUN-CIEUC")
    (18917 . "HANGUL SYLLABLE TIKEUT EU NIEUN-HIEUH")
    (18918 . "HANGUL SYLLABLE TIKEUT EU RIEUL-THIEUTH")
    (18919 . "HANGUL SYLLABLE TIKEUT EU RIEUL-PHIEUPH")
    (18920 . "HANGUL SYLLABLE TIKEUT EU RIEUL-HIEUH")
    (18921 . "HANGUL SYLLABLE TIKEUT EU PIEUP-SIOS")
    (18922 . "HANGUL SYLLABLE TIKEUT EU SSANGSIOS")
    (18923 . "HANGUL SYLLABLE TIKEUT EU CIEUC")
    (18924 . "HANGUL SYLLABLE TIKEUT EU CHIEUCH")
    (18925 . "HANGUL SYLLABLE TIKEUT EU KHIEUKH")
    (18926 . "HANGUL SYLLABLE TIKEUT EU THIEUTH")
    (18927 . "HANGUL SYLLABLE TIKEUT EU PHIEUPH")
    (18928 . "HANGUL SYLLABLE TIKEUT EU HIEUH")
    (18929 . "HANGUL SYLLABLE TIKEUT YI SSANGKIYEOK")
    (18930 . "HANGUL SYLLABLE TIKEUT YI KIYEOK-SIOS")
    (18931 . "HANGUL SYLLABLE TIKEUT YI NIEUN-CIEUC")
    (18932 . "HANGUL SYLLABLE TIKEUT YI NIEUN-HIEUH")
    (18933 . "HANGUL SYLLABLE TIKEUT YI RIEUL-KIYEOK")
    (18934 . "HANGUL SYLLABLE TIKEUT YI RIEUL-MIEUM")
    (18935 . "HANGUL SYLLABLE TIKEUT YI RIEUL-PIEUP")
    (18936 . "HANGUL SYLLABLE TIKEUT YI RIEUL-SIOS")
    (18937 . "HANGUL SYLLABLE TIKEUT YI RIEUL-THIEUTH")
    (18938 . "HANGUL SYLLABLE TIKEUT YI RIEUL-PHIEUPH")
    (18939 . "HANGUL SYLLABLE TIKEUT YI RIEUL-HIEUH")
    (18940 . "HANGUL SYLLABLE TIKEUT YI PIEUP-SIOS")
    (18941 . "HANGUL SYLLABLE TIKEUT YI SSANGSIOS")
    (18942 . "HANGUL SYLLABLE TIKEUT YI CIEUC")
    (18943 . "HANGUL SYLLABLE TIKEUT YI CHIEUCH")
    (18944 . "HANGUL SYLLABLE TIKEUT YI KHIEUKH")
    (18945 . "HANGUL SYLLABLE TIKEUT YI THIEUTH")
    (18946 . "HANGUL SYLLABLE TIKEUT YI PHIEUPH")
    (18947 . "HANGUL SYLLABLE TIKEUT YI HIEUH")
    (18948 . "HANGUL SYLLABLE TIKEUT I SSANGKIYEOK")
    (18949 . "HANGUL SYLLABLE TIKEUT I KIYEOK-SIOS")
    (18950 . "HANGUL SYLLABLE TIKEUT I NIEUN-CIEUC")
    (18951 . "HANGUL SYLLABLE TIKEUT I NIEUN-HIEUH")
    (18952 . "HANGUL SYLLABLE TIKEUT I RIEUL-KIYEOK")
    (18953 . "HANGUL SYLLABLE TIKEUT I RIEUL-MIEUM")
    (18954 . "HANGUL SYLLABLE TIKEUT I RIEUL-PIEUP")
    (18955 . "HANGUL SYLLABLE TIKEUT I RIEUL-SIOS")
    (18956 . "HANGUL SYLLABLE TIKEUT I RIEUL-THIEUTH")
    (18957 . "HANGUL SYLLABLE TIKEUT I RIEUL-PHIEUPH")
    (18958 . "HANGUL SYLLABLE TIKEUT I RIEUL-HIEUH")
    (18959 . "HANGUL SYLLABLE TIKEUT I PIEUP-SIOS")
    (18960 . "HANGUL SYLLABLE TIKEUT I KHIEUKH")
    (18961 . "HANGUL SYLLABLE SSANGTIKEUT A KIYEOK-SIOS")
    (18962 . "HANGUL SYLLABLE SSANGTIKEUT A NIEUN-CIEUC")
    (18963 . "HANGUL SYLLABLE SSANGTIKEUT A NIEUN-HIEUH")
    (18964 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-SIOS")
    (18965 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-THIEUTH")
    (18966 . "HANGUL SYLLABLE SSANGTIKEUT A RIEUL-PHIEUPH")
    (18967 . "HANGUL SYLLABLE SSANGTIKEUT A PIEUP-SIOS")
    (18968 . "HANGUL SYLLABLE SSANGTIKEUT A CIEUC")
    (18969 . "HANGUL SYLLABLE SSANGTIKEUT A CHIEUCH")
    (18970 . "HANGUL SYLLABLE SSANGTIKEUT A KHIEUKH")
    (18971 . "HANGUL SYLLABLE SSANGTIKEUT A THIEUTH")
    (18972 . "HANGUL SYLLABLE SSANGTIKEUT A PHIEUPH")
    (18973 . "HANGUL SYLLABLE SSANGTIKEUT AE KIYEOK-SIOS")
    (18974 . "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN-CIEUC")
    (18975 . "HANGUL SYLLABLE SSANGTIKEUT AE NIEUN-HIEUH")
    (18976 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-KIYEOK")
    (18977 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-MIEUM")
    (18978 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-PIEUP")
    (18979 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-SIOS")
    (18980 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-THIEUTH")
    (18981 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-PHIEUPH")
    (18982 . "HANGUL SYLLABLE SSANGTIKEUT AE RIEUL-HIEUH")
    (18983 . "HANGUL SYLLABLE SSANGTIKEUT AE PIEUP-SIOS")
    (18984 . "HANGUL SYLLABLE SSANGTIKEUT AE CIEUC")
    (18985 . "HANGUL SYLLABLE SSANGTIKEUT AE CHIEUCH")
    (18986 . "HANGUL SYLLABLE SSANGTIKEUT AE KHIEUKH")
    (18987 . "HANGUL SYLLABLE SSANGTIKEUT AE THIEUTH")
    (18988 . "HANGUL SYLLABLE SSANGTIKEUT AE PHIEUPH")
    (18989 . "HANGUL SYLLABLE SSANGTIKEUT YA SSANGKIYEOK")
    (18990 . "HANGUL SYLLABLE SSANGTIKEUT YA KIYEOK-SIOS")
    (18991 . "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN-CIEUC")
    (18992 . "HANGUL SYLLABLE SSANGTIKEUT YA NIEUN-HIEUH")
    (18993 . "HANGUL SYLLABLE SSANGTIKEUT YA TIKEUT")
    (18994 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL")
    (18995 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-KIYEOK")
    (18996 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-MIEUM")
    (18997 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-PIEUP")
    (18998 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-SIOS")
    (18999 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-THIEUTH")
    (19000 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-PHIEUPH")
    (19001 . "HANGUL SYLLABLE SSANGTIKEUT YA RIEUL-HIEUH")
    (19002 . "HANGUL SYLLABLE SSANGTIKEUT YA MIEUM")
    (19003 . "HANGUL SYLLABLE SSANGTIKEUT YA PIEUP")
    (19004 . "HANGUL SYLLABLE SSANGTIKEUT YA PIEUP-SIOS")
    (19005 . "HANGUL SYLLABLE SSANGTIKEUT YA SIOS")
    (19006 . "HANGUL SYLLABLE SSANGTIKEUT YA SSANGSIOS")
    (19007 . "HANGUL SYLLABLE SSANGTIKEUT YA CIEUC")
    (19008 . "HANGUL SYLLABLE SSANGTIKEUT YA CHIEUCH")
    (19009 . "HANGUL SYLLABLE SSANGTIKEUT YA KHIEUKH")
    (19010 . "HANGUL SYLLABLE SSANGTIKEUT YA THIEUTH")
    (19011 . "HANGUL SYLLABLE SSANGTIKEUT YA PHIEUPH")
    (19012 . "HANGUL SYLLABLE SSANGTIKEUT YA HIEUH")
    (19013 . "HANGUL SYLLABLE SSANGTIKEUT YAE KIYEOK")
    (19014 . "HANGUL SYLLABLE SSANGTIKEUT YAE SSANGKIYEOK")
    (19015 . "HANGUL SYLLABLE SSANGTIKEUT YAE KIYEOK-SIOS")
    (19016 . "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN-CIEUC")
    (19017 . "HANGUL SYLLABLE SSANGTIKEUT YAE NIEUN-HIEUH")
    (19018 . "HANGUL SYLLABLE SSANGTIKEUT YAE TIKEUT")
    (19019 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL")
    (19020 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-KIYEOK")
    (19021 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-MIEUM")
    (19022 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-PIEUP")
    (19023 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-SIOS")
    (19024 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-THIEUTH")
    (19025 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-PHIEUPH")
    (19026 . "HANGUL SYLLABLE SSANGTIKEUT YAE RIEUL-HIEUH")
    (19027 . "HANGUL SYLLABLE SSANGTIKEUT YAE MIEUM")
    (19028 . "HANGUL SYLLABLE SSANGTIKEUT YAE PIEUP")
    (19029 . "HANGUL SYLLABLE SSANGTIKEUT YAE PIEUP-SIOS")
    (19030 . "HANGUL SYLLABLE SSANGTIKEUT YAE SIOS")
    (19031 . "HANGUL SYLLABLE SSANGTIKEUT YAE SSANGSIOS")
    (19032 . "HANGUL SYLLABLE SSANGTIKEUT YAE IEUNG")
    (19033 . "HANGUL SYLLABLE SSANGTIKEUT YAE CIEUC")
    (19034 . "HANGUL SYLLABLE SSANGTIKEUT YAE CHIEUCH")
    (19035 . "HANGUL SYLLABLE SSANGTIKEUT YAE KHIEUKH")
    (19036 . "HANGUL SYLLABLE SSANGTIKEUT YAE THIEUTH")
    (19037 . "HANGUL SYLLABLE SSANGTIKEUT YAE PHIEUPH")
    (19038 . "HANGUL SYLLABLE SSANGTIKEUT YAE HIEUH")
    (19039 . "HANGUL SYLLABLE SSANGTIKEUT EO SSANGKIYEOK")
    (19040 . "HANGUL SYLLABLE SSANGTIKEUT EO KIYEOK-SIOS")
    (19041 . "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN-CIEUC")
    (19042 . "HANGUL SYLLABLE SSANGTIKEUT EO NIEUN-HIEUH")
    (19043 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-KIYEOK")
    (19044 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-SIOS")
    (19045 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-THIEUTH")
    (19046 . "HANGUL SYLLABLE SSANGTIKEUT EO RIEUL-PHIEUPH")
    (19047 . "HANGUL SYLLABLE SSANGTIKEUT EO PIEUP-SIOS")
    (19048 . "HANGUL SYLLABLE SSANGTIKEUT EO CIEUC")
    (19049 . "HANGUL SYLLABLE SSANGTIKEUT EO CHIEUCH")
    (19050 . "HANGUL SYLLABLE SSANGTIKEUT EO KHIEUKH")
    (19051 . "HANGUL SYLLABLE SSANGTIKEUT EO THIEUTH")
    (19052 . "HANGUL SYLLABLE SSANGTIKEUT EO PHIEUPH")
    (19053 . "HANGUL SYLLABLE SSANGTIKEUT E SSANGKIYEOK")
    (19054 . "HANGUL SYLLABLE SSANGTIKEUT E KIYEOK-SIOS")
    (19055 . "HANGUL SYLLABLE SSANGTIKEUT E NIEUN-CIEUC")
    (19056 . "HANGUL SYLLABLE SSANGTIKEUT E NIEUN-HIEUH")
    (19057 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-KIYEOK")
    (19058 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-MIEUM")
    (19059 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-PIEUP")
    (19060 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-SIOS")
    (19061 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-THIEUTH")
    (19062 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-PHIEUPH")
    (19063 . "HANGUL SYLLABLE SSANGTIKEUT E RIEUL-HIEUH")
    (19064 . "HANGUL SYLLABLE SSANGTIKEUT E PIEUP-SIOS")
    (19065 . "HANGUL SYLLABLE SSANGTIKEUT E CIEUC")
    (19066 . "HANGUL SYLLABLE SSANGTIKEUT E CHIEUCH")
    (19067 . "HANGUL SYLLABLE SSANGTIKEUT E KHIEUKH")
    (19068 . "HANGUL SYLLABLE SSANGTIKEUT E THIEUTH")
    (19069 . "HANGUL SYLLABLE SSANGTIKEUT E PHIEUPH")
    (19070 . "HANGUL SYLLABLE SSANGTIKEUT E HIEUH")
    (19071 . "HANGUL SYLLABLE SSANGTIKEUT YEO SSANGKIYEOK")
    (19072 . "HANGUL SYLLABLE SSANGTIKEUT YEO KIYEOK-SIOS")
    (19073 . "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN-CIEUC")
    (19074 . "HANGUL SYLLABLE SSANGTIKEUT YEO NIEUN-HIEUH")
    (19075 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL")
    (19076 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-KIYEOK")
    (19077 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-MIEUM")
    (19078 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-PIEUP")
    (19079 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-SIOS")
    (19080 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-THIEUTH")
    (19081 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-PHIEUPH")
    (19082 . "HANGUL SYLLABLE SSANGTIKEUT YEO RIEUL-HIEUH")
    (19083 . "HANGUL SYLLABLE SSANGTIKEUT YEO PIEUP-SIOS")
    (19084 . "HANGUL SYLLABLE SSANGTIKEUT YEO CIEUC")
    (19085 . "HANGUL SYLLABLE SSANGTIKEUT YEO CHIEUCH")
    (19086 . "HANGUL SYLLABLE SSANGTIKEUT YEO KHIEUKH")
    (19087 . "HANGUL SYLLABLE SSANGTIKEUT YEO THIEUTH")
    (19088 . "HANGUL SYLLABLE SSANGTIKEUT YEO PHIEUPH")
    (19089 . "HANGUL SYLLABLE SSANGTIKEUT YEO HIEUH")
    (19090 . "HANGUL SYLLABLE SSANGTIKEUT YE KIYEOK")
    (19091 . "HANGUL SYLLABLE SSANGTIKEUT YE SSANGKIYEOK")
    (19092 . "HANGUL SYLLABLE SSANGTIKEUT YE KIYEOK-SIOS")
    (19093 . "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN-CIEUC")
    (19094 . "HANGUL SYLLABLE SSANGTIKEUT YE NIEUN-HIEUH")
    (19095 . "HANGUL SYLLABLE SSANGTIKEUT YE TIKEUT")
    (19096 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL")
    (19097 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-KIYEOK")
    (19098 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-MIEUM")
    (19099 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-PIEUP")
    (19100 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-SIOS")
    (19101 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-THIEUTH")
    (19102 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-PHIEUPH")
    (19103 . "HANGUL SYLLABLE SSANGTIKEUT YE RIEUL-HIEUH")
    (19104 . "HANGUL SYLLABLE SSANGTIKEUT YE MIEUM")
    (19105 . "HANGUL SYLLABLE SSANGTIKEUT YE PIEUP-SIOS")
    (19106 . "HANGUL SYLLABLE SSANGTIKEUT YE SIOS")
    (19107 . "HANGUL SYLLABLE SSANGTIKEUT YE SSANGSIOS")
    (19108 . "HANGUL SYLLABLE SSANGTIKEUT YE IEUNG")
    (19109 . "HANGUL SYLLABLE SSANGTIKEUT YE CIEUC")
    (19110 . "HANGUL SYLLABLE SSANGTIKEUT YE CHIEUCH")
    (19111 . "HANGUL SYLLABLE SSANGTIKEUT YE KHIEUKH")
    (19112 . "HANGUL SYLLABLE SSANGTIKEUT YE THIEUTH")
    (19113 . "HANGUL SYLLABLE SSANGTIKEUT YE PHIEUPH")
    (19114 . "HANGUL SYLLABLE SSANGTIKEUT YE HIEUH")
    (19115 . "HANGUL SYLLABLE SSANGTIKEUT O KIYEOK-SIOS")
    (19116 . "HANGUL SYLLABLE SSANGTIKEUT O NIEUN-CIEUC")
    (19117 . "HANGUL SYLLABLE SSANGTIKEUT O NIEUN-HIEUH")
    (19118 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-PIEUP")
    (19119 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-SIOS")
    (19120 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-THIEUTH")
    (19121 . "HANGUL SYLLABLE SSANGTIKEUT O RIEUL-PHIEUPH")
    (19122 . "HANGUL SYLLABLE SSANGTIKEUT O PIEUP-SIOS")
    (19123 . "HANGUL SYLLABLE SSANGTIKEUT O SSANGSIOS")
    (19124 . "HANGUL SYLLABLE SSANGTIKEUT O CIEUC")
    (19125 . "HANGUL SYLLABLE SSANGTIKEUT O CHIEUCH")
    (19126 . "HANGUL SYLLABLE SSANGTIKEUT O KHIEUKH")
    (19127 . "HANGUL SYLLABLE SSANGTIKEUT O THIEUTH")
    (19128 . "HANGUL SYLLABLE SSANGTIKEUT O PHIEUPH")
    (19129 . "HANGUL SYLLABLE SSANGTIKEUT O HIEUH")
    (19130 . "HANGUL SYLLABLE SSANGTIKEUT WA SSANGKIYEOK")
    (19131 . "HANGUL SYLLABLE SSANGTIKEUT WA KIYEOK-SIOS")
    (19132 . "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN-CIEUC")
    (19133 . "HANGUL SYLLABLE SSANGTIKEUT WA NIEUN-HIEUH")
    (19134 . "HANGUL SYLLABLE SSANGTIKEUT WA TIKEUT")
    (19135 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-KIYEOK")
    (19136 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-MIEUM")
    (19137 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-PIEUP")
    (19138 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-SIOS")
    (19139 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-THIEUTH")
    (19140 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-PHIEUPH")
    (19141 . "HANGUL SYLLABLE SSANGTIKEUT WA RIEUL-HIEUH")
    (19142 . "HANGUL SYLLABLE SSANGTIKEUT WA MIEUM")
    (19143 . "HANGUL SYLLABLE SSANGTIKEUT WA PIEUP")
    (19144 . "HANGUL SYLLABLE SSANGTIKEUT WA PIEUP-SIOS")
    (19145 . "HANGUL SYLLABLE SSANGTIKEUT WA SIOS")
    (19146 . "HANGUL SYLLABLE SSANGTIKEUT WA SSANGSIOS")
    (19147 . "HANGUL SYLLABLE SSANGTIKEUT WA CIEUC")
    (19148 . "HANGUL SYLLABLE SSANGTIKEUT WA CHIEUCH")
    (19149 . "HANGUL SYLLABLE SSANGTIKEUT WA KHIEUKH")
    (19150 . "HANGUL SYLLABLE SSANGTIKEUT WA THIEUTH")
    (19151 . "HANGUL SYLLABLE SSANGTIKEUT WA PHIEUPH")
    (19152 . "HANGUL SYLLABLE SSANGTIKEUT WA HIEUH")
    (19153 . "HANGUL SYLLABLE SSANGTIKEUT WAE SSANGKIYEOK")
    (19154 . "HANGUL SYLLABLE SSANGTIKEUT WAE KIYEOK-SIOS")
    (19155 . "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN-CIEUC")
    (19156 . "HANGUL SYLLABLE SSANGTIKEUT WAE NIEUN-HIEUH")
    (19157 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-KIYEOK")
    (19158 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-MIEUM")
    (19159 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-PIEUP")
    (19160 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-SIOS")
    (19161 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-THIEUTH")
    (19162 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-PHIEUPH")
    (19163 . "HANGUL SYLLABLE SSANGTIKEUT WAE RIEUL-HIEUH")
    (19164 . "HANGUL SYLLABLE SSANGTIKEUT WAE PIEUP-SIOS")
    (19165 . "HANGUL SYLLABLE SSANGTIKEUT WAE IEUNG")
    (19166 . "HANGUL SYLLABLE SSANGTIKEUT WAE CIEUC")
    (19167 . "HANGUL SYLLABLE SSANGTIKEUT WAE CHIEUCH")
    (19168 . "HANGUL SYLLABLE SSANGTIKEUT WAE KHIEUKH")
    (19169 . "HANGUL SYLLABLE SSANGTIKEUT WAE THIEUTH")
    (19170 . "HANGUL SYLLABLE SSANGTIKEUT WAE PHIEUPH")
    (19171 . "HANGUL SYLLABLE SSANGTIKEUT WAE HIEUH")
    (19172 . "HANGUL SYLLABLE SSANGTIKEUT OE SSANGKIYEOK")
    (19173 . "HANGUL SYLLABLE SSANGTIKEUT OE KIYEOK-SIOS")
    (19174 . "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN-CIEUC")
    (19175 . "HANGUL SYLLABLE SSANGTIKEUT OE NIEUN-HIEUH")
    (19176 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-KIYEOK")
    (19177 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-MIEUM")
    (19178 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-PIEUP")
    (19179 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-SIOS")
    (19180 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-THIEUTH")
    (19181 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-PHIEUPH")
    (19182 . "HANGUL SYLLABLE SSANGTIKEUT OE RIEUL-HIEUH")
    (19183 . "HANGUL SYLLABLE SSANGTIKEUT OE PIEUP-SIOS")
    (19184 . "HANGUL SYLLABLE SSANGTIKEUT OE CIEUC")
    (19185 . "HANGUL SYLLABLE SSANGTIKEUT OE CHIEUCH")
    (19186 . "HANGUL SYLLABLE SSANGTIKEUT OE KHIEUKH")
    (19187 . "HANGUL SYLLABLE SSANGTIKEUT OE THIEUTH")
    (19188 . "HANGUL SYLLABLE SSANGTIKEUT OE PHIEUPH")
    (19189 . "HANGUL SYLLABLE SSANGTIKEUT OE HIEUH")
    (19190 . "HANGUL SYLLABLE SSANGTIKEUT YO SSANGKIYEOK")
    (19191 . "HANGUL SYLLABLE SSANGTIKEUT YO KIYEOK-SIOS")
    (19192 . "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN")
    (19193 . "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN-CIEUC")
    (19194 . "HANGUL SYLLABLE SSANGTIKEUT YO NIEUN-HIEUH")
    (19195 . "HANGUL SYLLABLE SSANGTIKEUT YO TIKEUT")
    (19196 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL")
    (19197 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-KIYEOK")
    (19198 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-MIEUM")
    (19199 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-PIEUP")
    (19200 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-SIOS")
    (19201 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-THIEUTH")
    (19202 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-PHIEUPH")
    (19203 . "HANGUL SYLLABLE SSANGTIKEUT YO RIEUL-HIEUH")
    (19204 . "HANGUL SYLLABLE SSANGTIKEUT YO MIEUM")
    (19205 . "HANGUL SYLLABLE SSANGTIKEUT YO PIEUP")
    (19206 . "HANGUL SYLLABLE SSANGTIKEUT YO PIEUP-SIOS")
    (19207 . "HANGUL SYLLABLE SSANGTIKEUT YO SIOS")
    (19208 . "HANGUL SYLLABLE SSANGTIKEUT YO SSANGSIOS")
    (19209 . "HANGUL SYLLABLE SSANGTIKEUT YO IEUNG")
    (19210 . "HANGUL SYLLABLE SSANGTIKEUT YO CIEUC")
    (19211 . "HANGUL SYLLABLE SSANGTIKEUT YO CHIEUCH")
    (19212 . "HANGUL SYLLABLE SSANGTIKEUT YO KHIEUKH")
    (19213 . "HANGUL SYLLABLE SSANGTIKEUT YO THIEUTH")
    (19214 . "HANGUL SYLLABLE SSANGTIKEUT YO PHIEUPH")
    (19215 . "HANGUL SYLLABLE SSANGTIKEUT YO HIEUH")
    (19216 . "HANGUL SYLLABLE SSANGTIKEUT U SSANGKIYEOK")
    (19217 . "HANGUL SYLLABLE SSANGTIKEUT U KIYEOK-SIOS")
    (19218 . "HANGUL SYLLABLE SSANGTIKEUT U NIEUN-CIEUC")
    (19219 . "HANGUL SYLLABLE SSANGTIKEUT U NIEUN-HIEUH")
    (19220 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-KIYEOK")
    (19221 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-SIOS")
    (19222 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-THIEUTH")
    (19223 . "HANGUL SYLLABLE SSANGTIKEUT U RIEUL-PHIEUPH")
    (19224 . "HANGUL SYLLABLE SSANGTIKEUT U PIEUP-SIOS")
    (19225 . "HANGUL SYLLABLE SSANGTIKEUT U SSANGSIOS")
    (19226 . "HANGUL SYLLABLE SSANGTIKEUT U CIEUC")
    (19227 . "HANGUL SYLLABLE SSANGTIKEUT U CHIEUCH")
    (19228 . "HANGUL SYLLABLE SSANGTIKEUT U KHIEUKH")
    (19229 . "HANGUL SYLLABLE SSANGTIKEUT U PHIEUPH")
    (19230 . "HANGUL SYLLABLE SSANGTIKEUT U HIEUH")
    (19231 . "HANGUL SYLLABLE SSANGTIKEUT WEO KIYEOK")
    (19232 . "HANGUL SYLLABLE SSANGTIKEUT WEO SSANGKIYEOK")
    (19233 . "HANGUL SYLLABLE SSANGTIKEUT WEO KIYEOK-SIOS")
    (19234 . "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN-CIEUC")
    (19235 . "HANGUL SYLLABLE SSANGTIKEUT WEO NIEUN-HIEUH")
    (19236 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL")
    (19237 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-KIYEOK")
    (19238 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-MIEUM")
    (19239 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-PIEUP")
    (19240 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-SIOS")
    (19241 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-THIEUTH")
    (19242 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-PHIEUPH")
    (19243 . "HANGUL SYLLABLE SSANGTIKEUT WEO RIEUL-HIEUH")
    (19244 . "HANGUL SYLLABLE SSANGTIKEUT WEO MIEUM")
    (19245 . "HANGUL SYLLABLE SSANGTIKEUT WEO PIEUP")
    (19246 . "HANGUL SYLLABLE SSANGTIKEUT WEO PIEUP-SIOS")
    (19247 . "HANGUL SYLLABLE SSANGTIKEUT WEO SSANGSIOS")
    (19248 . "HANGUL SYLLABLE SSANGTIKEUT WEO IEUNG")
    (19249 . "HANGUL SYLLABLE SSANGTIKEUT WEO CIEUC")
    (19250 . "HANGUL SYLLABLE SSANGTIKEUT WEO CHIEUCH")
    (19251 . "HANGUL SYLLABLE SSANGTIKEUT WEO KHIEUKH")
    (19252 . "HANGUL SYLLABLE SSANGTIKEUT WEO THIEUTH")
    (19253 . "HANGUL SYLLABLE SSANGTIKEUT WEO PHIEUPH")
    (19254 . "HANGUL SYLLABLE SSANGTIKEUT WEO HIEUH")
    (19255 . "HANGUL SYLLABLE SSANGTIKEUT WE KIYEOK")
    (19256 . "HANGUL SYLLABLE SSANGTIKEUT WE SSANGKIYEOK")
    (19257 . "HANGUL SYLLABLE SSANGTIKEUT WE KIYEOK-SIOS")
    (19258 . "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN-CIEUC")
    (19259 . "HANGUL SYLLABLE SSANGTIKEUT WE NIEUN-HIEUH")
    (19260 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-KIYEOK")
    (19261 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-MIEUM")
    (19262 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-PIEUP")
    (19263 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-SIOS")
    (19264 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-THIEUTH")
    (19265 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-PHIEUPH")
    (19266 . "HANGUL SYLLABLE SSANGTIKEUT WE RIEUL-HIEUH")
    (19267 . "HANGUL SYLLABLE SSANGTIKEUT WE PIEUP-SIOS")
    (19268 . "HANGUL SYLLABLE SSANGTIKEUT WE SIOS")
    (19269 . "HANGUL SYLLABLE SSANGTIKEUT WE CIEUC")
    (19270 . "HANGUL SYLLABLE SSANGTIKEUT WE CHIEUCH")
    (19271 . "HANGUL SYLLABLE SSANGTIKEUT WE KHIEUKH")
    (19272 . "HANGUL SYLLABLE SSANGTIKEUT WE THIEUTH")
    (19273 . "HANGUL SYLLABLE SSANGTIKEUT WE PHIEUPH")
    (19274 . "HANGUL SYLLABLE SSANGTIKEUT WE HIEUH")
    (19275 . "HANGUL SYLLABLE SSANGTIKEUT WI KIYEOK")
    (19276 . "HANGUL SYLLABLE SSANGTIKEUT WI SSANGKIYEOK")
    (19277 . "HANGUL SYLLABLE SSANGTIKEUT WI KIYEOK-SIOS")
    (19278 . "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN-CIEUC")
    (19279 . "HANGUL SYLLABLE SSANGTIKEUT WI NIEUN-HIEUH")
    (19280 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-KIYEOK")
    (19281 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-MIEUM")
    (19282 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-PIEUP")
    (19283 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-SIOS")
    (19284 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-THIEUTH")
    (19285 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-PHIEUPH")
    (19286 . "HANGUL SYLLABLE SSANGTIKEUT WI RIEUL-HIEUH")
    (19287 . "HANGUL SYLLABLE SSANGTIKEUT WI PIEUP-SIOS")
    (19288 . "HANGUL SYLLABLE SSANGTIKEUT WI SSANGSIOS")
    (19289 . "HANGUL SYLLABLE SSANGTIKEUT WI CIEUC")
    (19290 . "HANGUL SYLLABLE SSANGTIKEUT WI CHIEUCH")
    (19291 . "HANGUL SYLLABLE SSANGTIKEUT WI KHIEUKH")
    (19292 . "HANGUL SYLLABLE SSANGTIKEUT WI THIEUTH")
    (19293 . "HANGUL SYLLABLE SSANGTIKEUT WI PHIEUPH")
    (19294 . "HANGUL SYLLABLE SSANGTIKEUT WI HIEUH")
    (19295 . "HANGUL SYLLABLE SSANGTIKEUT YU SSANGKIYEOK")
    (19296 . "HANGUL SYLLABLE SSANGTIKEUT YU KIYEOK-SIOS")
    (19297 . "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN")
    (19298 . "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN-CIEUC")
    (19299 . "HANGUL SYLLABLE SSANGTIKEUT YU NIEUN-HIEUH")
    (19300 . "HANGUL SYLLABLE SSANGTIKEUT YU TIKEUT")
    (19301 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-KIYEOK")
    (19302 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-MIEUM")
    (19303 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-PIEUP")
    (19304 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-SIOS")
    (19305 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-THIEUTH")
    (19306 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-PHIEUPH")
    (19307 . "HANGUL SYLLABLE SSANGTIKEUT YU RIEUL-HIEUH")
    (19308 . "HANGUL SYLLABLE SSANGTIKEUT YU PIEUP")
    (19309 . "HANGUL SYLLABLE SSANGTIKEUT YU PIEUP-SIOS")
    (19310 . "HANGUL SYLLABLE SSANGTIKEUT YU SIOS")
    (19311 . "HANGUL SYLLABLE SSANGTIKEUT YU SSANGSIOS")
    (19312 . "HANGUL SYLLABLE SSANGTIKEUT YU CIEUC")
    (19313 . "HANGUL SYLLABLE SSANGTIKEUT YU CHIEUCH")
    (19314 . "HANGUL SYLLABLE SSANGTIKEUT YU KHIEUKH")
    (19315 . "HANGUL SYLLABLE SSANGTIKEUT YU THIEUTH")
    (19316 . "HANGUL SYLLABLE SSANGTIKEUT YU PHIEUPH")
    (19317 . "HANGUL SYLLABLE SSANGTIKEUT YU HIEUH")
    (19318 . "HANGUL SYLLABLE SSANGTIKEUT EU SSANGKIYEOK")
    (19319 . "HANGUL SYLLABLE SSANGTIKEUT EU KIYEOK-SIOS")
    (19320 . "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN-CIEUC")
    (19321 . "HANGUL SYLLABLE SSANGTIKEUT EU NIEUN-HIEUH")
    (19322 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-KIYEOK")
    (19323 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-SIOS")
    (19324 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-THIEUTH")
    (19325 . "HANGUL SYLLABLE SSANGTIKEUT EU RIEUL-PHIEUPH")
    (19326 . "HANGUL SYLLABLE SSANGTIKEUT EU PIEUP-SIOS")
    (19327 . "HANGUL SYLLABLE SSANGTIKEUT EU SSANGSIOS")
    (19328 . "HANGUL SYLLABLE SSANGTIKEUT EU CIEUC")
    (19329 . "HANGUL SYLLABLE SSANGTIKEUT EU CHIEUCH")
    (19330 . "HANGUL SYLLABLE SSANGTIKEUT EU KHIEUKH")
    (19331 . "HANGUL SYLLABLE SSANGTIKEUT EU THIEUTH")
    (19332 . "HANGUL SYLLABLE SSANGTIKEUT EU PHIEUPH")
    (19333 . "HANGUL SYLLABLE SSANGTIKEUT EU HIEUH")
    (19334 . "HANGUL SYLLABLE SSANGTIKEUT YI KIYEOK")
    (19335 . "HANGUL SYLLABLE SSANGTIKEUT YI SSANGKIYEOK")
    (19336 . "HANGUL SYLLABLE SSANGTIKEUT YI KIYEOK-SIOS")
    (19337 . "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN-CIEUC")
    (19338 . "HANGUL SYLLABLE SSANGTIKEUT YI NIEUN-HIEUH")
    (19339 . "HANGUL SYLLABLE SSANGTIKEUT YI TIKEUT")
    (19340 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-KIYEOK")
    (19341 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-MIEUM")
    (19342 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-PIEUP")
    (19343 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-SIOS")
    (19344 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-THIEUTH")
    (19345 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-PHIEUPH")
    (19346 . "HANGUL SYLLABLE SSANGTIKEUT YI RIEUL-HIEUH")
    (19347 . "HANGUL SYLLABLE SSANGTIKEUT YI PIEUP-SIOS")
    (19348 . "HANGUL SYLLABLE SSANGTIKEUT YI SIOS")
    (19349 . "HANGUL SYLLABLE SSANGTIKEUT YI SSANGSIOS")
    (19350 . "HANGUL SYLLABLE SSANGTIKEUT YI CIEUC")
    (19351 . "HANGUL SYLLABLE SSANGTIKEUT YI CHIEUCH")
    (19352 . "HANGUL SYLLABLE SSANGTIKEUT YI KHIEUKH")
    (19353 . "HANGUL SYLLABLE SSANGTIKEUT YI THIEUTH")
    (19354 . "HANGUL SYLLABLE SSANGTIKEUT YI PHIEUPH")
    (19355 . "HANGUL SYLLABLE SSANGTIKEUT YI HIEUH")
    (19356 . "HANGUL SYLLABLE SSANGTIKEUT I SSANGKIYEOK")
    (19357 . "HANGUL SYLLABLE SSANGTIKEUT I KIYEOK-SIOS")
    (19358 . "HANGUL SYLLABLE SSANGTIKEUT I NIEUN-CIEUC")
    (19359 . "HANGUL SYLLABLE SSANGTIKEUT I NIEUN-HIEUH")
    (19360 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-KIYEOK")
    (19361 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-MIEUM")
    (19362 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-PIEUP")
    (19363 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-SIOS")
    (19364 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-THIEUTH")
    (19365 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-PHIEUPH")
    (19366 . "HANGUL SYLLABLE SSANGTIKEUT I RIEUL-HIEUH")
    (19367 . "HANGUL SYLLABLE SSANGTIKEUT I PIEUP-SIOS")
    (19368 . "HANGUL SYLLABLE SSANGTIKEUT I SSANGSIOS")
    (19369 . "HANGUL SYLLABLE SSANGTIKEUT I CIEUC")
    (19370 . "HANGUL SYLLABLE SSANGTIKEUT I CHIEUCH")
    (19371 . "HANGUL SYLLABLE SSANGTIKEUT I KHIEUKH")
    (19372 . "HANGUL SYLLABLE SSANGTIKEUT I THIEUTH")
    (19373 . "HANGUL SYLLABLE SSANGTIKEUT I PHIEUPH")
    (19374 . "HANGUL SYLLABLE RIEUL A SSANGKIYEOK")
    (19375 . "HANGUL SYLLABLE RIEUL A KIYEOK-SIOS")
    (19376 . "HANGUL SYLLABLE RIEUL A NIEUN-CIEUC")
    (19377 . "HANGUL SYLLABLE RIEUL A NIEUN-HIEUH")
    (19378 . "HANGUL SYLLABLE RIEUL A RIEUL-KIYEOK")
    (19379 . "HANGUL SYLLABLE RIEUL A RIEUL-MIEUM")
    (19380 . "HANGUL SYLLABLE RIEUL A RIEUL-PIEUP")
    (19381 . "HANGUL SYLLABLE RIEUL A RIEUL-THIEUTH")
    (19382 . "HANGUL SYLLABLE RIEUL A RIEUL-PHIEUPH")
    (19383 . "HANGUL SYLLABLE RIEUL A RIEUL-HIEUH")
    (19384 . "HANGUL SYLLABLE RIEUL A PIEUP-SIOS")
    (19385 . "HANGUL SYLLABLE RIEUL AE SSANGKIYEOK")
    (19386 . "HANGUL SYLLABLE RIEUL AE KIYEOK-SIOS")
    (19387 . "HANGUL SYLLABLE RIEUL AE NIEUN-CIEUC")
    (19388 . "HANGUL SYLLABLE RIEUL AE NIEUN-HIEUH")
    (19389 . "HANGUL SYLLABLE RIEUL AE RIEUL-KIYEOK")
    (19390 . "HANGUL SYLLABLE RIEUL AE RIEUL-MIEUM")
    (19391 . "HANGUL SYLLABLE RIEUL AE RIEUL-PIEUP")
    (19392 . "HANGUL SYLLABLE RIEUL AE RIEUL-SIOS")
    (19393 . "HANGUL SYLLABLE RIEUL AE RIEUL-THIEUTH")
    (19394 . "HANGUL SYLLABLE RIEUL AE RIEUL-PHIEUPH")
    (19395 . "HANGUL SYLLABLE RIEUL AE RIEUL-HIEUH")
    (19396 . "HANGUL SYLLABLE RIEUL AE PIEUP-SIOS")
    (19397 . "HANGUL SYLLABLE RIEUL AE CIEUC")
    (19398 . "HANGUL SYLLABLE RIEUL AE CHIEUCH")
    (19399 . "HANGUL SYLLABLE RIEUL AE HIEUH")
    (19400 . "HANGUL SYLLABLE RIEUL YA SSANGKIYEOK")
    (19401 . "HANGUL SYLLABLE RIEUL YA KIYEOK-SIOS")
    (19402 . "HANGUL SYLLABLE RIEUL YA NIEUN-CIEUC")
    (19403 . "HANGUL SYLLABLE RIEUL YA NIEUN-HIEUH")
    (19404 . "HANGUL SYLLABLE RIEUL YA RIEUL-KIYEOK")
    (19405 . "HANGUL SYLLABLE RIEUL YA RIEUL-MIEUM")
    (19406 . "HANGUL SYLLABLE RIEUL YA RIEUL-PIEUP")
    (19407 . "HANGUL SYLLABLE RIEUL YA RIEUL-SIOS")
    (19408 . "HANGUL SYLLABLE RIEUL YA RIEUL-THIEUTH")
    (19409 . "HANGUL SYLLABLE RIEUL YA RIEUL-PHIEUPH")
    (19410 . "HANGUL SYLLABLE RIEUL YA RIEUL-HIEUH")
    (19411 . "HANGUL SYLLABLE RIEUL YA PIEUP-SIOS")
    (19412 . "HANGUL SYLLABLE RIEUL YA SSANGSIOS")
    (19413 . "HANGUL SYLLABLE RIEUL YA CIEUC")
    (19414 . "HANGUL SYLLABLE RIEUL YA CHIEUCH")
    (19415 . "HANGUL SYLLABLE RIEUL YA KHIEUKH")
    (19416 . "HANGUL SYLLABLE RIEUL YA THIEUTH")
    (19417 . "HANGUL SYLLABLE RIEUL YA PHIEUPH")
    (19418 . "HANGUL SYLLABLE RIEUL YA HIEUH")
    (19419 . "HANGUL SYLLABLE RIEUL YAE KIYEOK")
    (19420 . "HANGUL SYLLABLE RIEUL YAE SSANGKIYEOK")
    (19421 . "HANGUL SYLLABLE RIEUL YAE KIYEOK-SIOS")
    (19422 . "HANGUL SYLLABLE RIEUL YAE NIEUN")
    (19423 . "HANGUL SYLLABLE RIEUL YAE NIEUN-CIEUC")
    (19424 . "HANGUL SYLLABLE RIEUL YAE NIEUN-HIEUH")
    (19425 . "HANGUL SYLLABLE RIEUL YAE TIKEUT")
    (19426 . "HANGUL SYLLABLE RIEUL YAE RIEUL")
    (19427 . "HANGUL SYLLABLE RIEUL YAE RIEUL-KIYEOK")
    (19428 . "HANGUL SYLLABLE RIEUL YAE RIEUL-MIEUM")
    (19429 . "HANGUL SYLLABLE RIEUL YAE RIEUL-PIEUP")
    (19430 . "HANGUL SYLLABLE RIEUL YAE RIEUL-SIOS")
    (19431 . "HANGUL SYLLABLE RIEUL YAE RIEUL-THIEUTH")
    (19432 . "HANGUL SYLLABLE RIEUL YAE RIEUL-PHIEUPH")
    (19433 . "HANGUL SYLLABLE RIEUL YAE RIEUL-HIEUH")
    (19434 . "HANGUL SYLLABLE RIEUL YAE MIEUM")
    (19435 . "HANGUL SYLLABLE RIEUL YAE PIEUP")
    (19436 . "HANGUL SYLLABLE RIEUL YAE PIEUP-SIOS")
    (19437 . "HANGUL SYLLABLE RIEUL YAE SSANGSIOS")
    (19438 . "HANGUL SYLLABLE RIEUL YAE IEUNG")
    (19439 . "HANGUL SYLLABLE RIEUL YAE CIEUC")
    (19440 . "HANGUL SYLLABLE RIEUL YAE CHIEUCH")
    (19441 . "HANGUL SYLLABLE RIEUL YAE KHIEUKH")
    (19442 . "HANGUL SYLLABLE RIEUL YAE THIEUTH")
    (19443 . "HANGUL SYLLABLE RIEUL YAE PHIEUPH")
    (19444 . "HANGUL SYLLABLE RIEUL YAE HIEUH")
    (19445 . "HANGUL SYLLABLE RIEUL EO SSANGKIYEOK")
    (19446 . "HANGUL SYLLABLE RIEUL EO KIYEOK-SIOS")
    (19447 . "HANGUL SYLLABLE RIEUL EO NIEUN-CIEUC")
    (19448 . "HANGUL SYLLABLE RIEUL EO RIEUL-KIYEOK")
    (19449 . "HANGUL SYLLABLE RIEUL EO RIEUL-MIEUM")
    (19450 . "HANGUL SYLLABLE RIEUL EO RIEUL-PIEUP")
    (19451 . "HANGUL SYLLABLE RIEUL EO RIEUL-SIOS")
    (19452 . "HANGUL SYLLABLE RIEUL EO RIEUL-THIEUTH")
    (19453 . "HANGUL SYLLABLE RIEUL EO RIEUL-PHIEUPH")
    (19454 . "HANGUL SYLLABLE RIEUL EO RIEUL-HIEUH")
    (19455 . "HANGUL SYLLABLE RIEUL EO CHIEUCH")
    (19456 . "HANGUL SYLLABLE RIEUL EO THIEUTH")
    (19457 . "HANGUL SYLLABLE RIEUL E SSANGKIYEOK")
    (19458 . "HANGUL SYLLABLE RIEUL E KIYEOK-SIOS")
    (19459 . "HANGUL SYLLABLE RIEUL E NIEUN-CIEUC")
    (19460 . "HANGUL SYLLABLE RIEUL E NIEUN-HIEUH")
    (19461 . "HANGUL SYLLABLE RIEUL E RIEUL-KIYEOK")
    (19462 . "HANGUL SYLLABLE RIEUL E RIEUL-MIEUM")
    (19463 . "HANGUL SYLLABLE RIEUL E RIEUL-PIEUP")
    (19464 . "HANGUL SYLLABLE RIEUL E RIEUL-SIOS")
    (19465 . "HANGUL SYLLABLE RIEUL E RIEUL-THIEUTH")
    (19466 . "HANGUL SYLLABLE RIEUL E RIEUL-PHIEUPH")
    (19467 . "HANGUL SYLLABLE RIEUL E RIEUL-HIEUH")
    (19468 . "HANGUL SYLLABLE RIEUL E PIEUP-SIOS")
    (19469 . "HANGUL SYLLABLE RIEUL E CIEUC")
    (19470 . "HANGUL SYLLABLE RIEUL E CHIEUCH")
    (19471 . "HANGUL SYLLABLE RIEUL E KHIEUKH")
    (19472 . "HANGUL SYLLABLE RIEUL YEO SSANGKIYEOK")
    (19473 . "HANGUL SYLLABLE RIEUL YEO KIYEOK-SIOS")
    (19474 . "HANGUL SYLLABLE RIEUL YEO NIEUN-CIEUC")
    (19475 . "HANGUL SYLLABLE RIEUL YEO NIEUN-HIEUH")
    (19476 . "HANGUL SYLLABLE RIEUL YEO RIEUL-KIYEOK")
    (19477 . "HANGUL SYLLABLE RIEUL YEO RIEUL-MIEUM")
    (19478 . "HANGUL SYLLABLE RIEUL YEO RIEUL-PIEUP")
    (19479 . "HANGUL SYLLABLE RIEUL YEO RIEUL-THIEUTH")
    (19480 . "HANGUL SYLLABLE RIEUL YEO RIEUL-PHIEUPH")
    (19481 . "HANGUL SYLLABLE RIEUL YEO RIEUL-HIEUH")
    (19482 . "HANGUL SYLLABLE RIEUL YEO PIEUP-SIOS")
    (19483 . "HANGUL SYLLABLE RIEUL YEO CIEUC")
    (19484 . "HANGUL SYLLABLE RIEUL YEO CHIEUCH")
    (19485 . "HANGUL SYLLABLE RIEUL YEO THIEUTH")
    (19486 . "HANGUL SYLLABLE RIEUL YEO HIEUH")
    (19487 . "HANGUL SYLLABLE RIEUL YE SSANGKIYEOK")
    (19488 . "HANGUL SYLLABLE RIEUL YE KIYEOK-SIOS")
    (19489 . "HANGUL SYLLABLE RIEUL YE NIEUN-CIEUC")
    (19490 . "HANGUL SYLLABLE RIEUL YE NIEUN-HIEUH")
    (19491 . "HANGUL SYLLABLE RIEUL YE RIEUL-KIYEOK")
    (19492 . "HANGUL SYLLABLE RIEUL YE RIEUL-MIEUM")
    (19493 . "HANGUL SYLLABLE RIEUL YE RIEUL-PIEUP")
    (19494 . "HANGUL SYLLABLE RIEUL YE RIEUL-SIOS")
    (19495 . "HANGUL SYLLABLE RIEUL YE RIEUL-THIEUTH")
    (19496 . "HANGUL SYLLABLE RIEUL YE RIEUL-PHIEUPH")
    (19497 . "HANGUL SYLLABLE RIEUL YE RIEUL-HIEUH")
    (19498 . "HANGUL SYLLABLE RIEUL YE PIEUP-SIOS")
    (19499 . "HANGUL SYLLABLE RIEUL YE SSANGSIOS")
    (19500 . "HANGUL SYLLABLE RIEUL YE CIEUC")
    (19501 . "HANGUL SYLLABLE RIEUL YE CHIEUCH")
    (19502 . "HANGUL SYLLABLE RIEUL YE KHIEUKH")
    (19503 . "HANGUL SYLLABLE RIEUL YE THIEUTH")
    (19504 . "HANGUL SYLLABLE RIEUL YE PHIEUPH")
    (19505 . "HANGUL SYLLABLE RIEUL YE HIEUH")
    (19506 . "HANGUL SYLLABLE RIEUL O SSANGKIYEOK")
    (19507 . "HANGUL SYLLABLE RIEUL O KIYEOK-SIOS")
    (19508 . "HANGUL SYLLABLE RIEUL O NIEUN-CIEUC")
    (19509 . "HANGUL SYLLABLE RIEUL O NIEUN-HIEUH")
    (19510 . "HANGUL SYLLABLE RIEUL O RIEUL-KIYEOK")
    (19511 . "HANGUL SYLLABLE RIEUL O RIEUL-MIEUM")
    (19512 . "HANGUL SYLLABLE RIEUL O RIEUL-PIEUP")
    (19513 . "HANGUL SYLLABLE RIEUL O RIEUL-THIEUTH")
    (19514 . "HANGUL SYLLABLE RIEUL O RIEUL-PHIEUPH")
    (19515 . "HANGUL SYLLABLE RIEUL O PIEUP-SIOS")
    (19516 . "HANGUL SYLLABLE RIEUL O SSANGSIOS")
    (19517 . "HANGUL SYLLABLE RIEUL O CIEUC")
    (19518 . "HANGUL SYLLABLE RIEUL O CHIEUCH")
    (19519 . "HANGUL SYLLABLE RIEUL O THIEUTH")
    (19520 . "HANGUL SYLLABLE RIEUL WA SSANGKIYEOK")
    (19521 . "HANGUL SYLLABLE RIEUL WA KIYEOK-SIOS")
    (19522 . "HANGUL SYLLABLE RIEUL WA NIEUN-CIEUC")
    (19523 . "HANGUL SYLLABLE RIEUL WA NIEUN-HIEUH")
    (19524 . "HANGUL SYLLABLE RIEUL WA TIKEUT")
    (19525 . "HANGUL SYLLABLE RIEUL WA RIEUL-KIYEOK")
    (19526 . "HANGUL SYLLABLE RIEUL WA RIEUL-MIEUM")
    (19527 . "HANGUL SYLLABLE RIEUL WA RIEUL-PIEUP")
    (19528 . "HANGUL SYLLABLE RIEUL WA RIEUL-SIOS")
    (19529 . "HANGUL SYLLABLE RIEUL WA RIEUL-THIEUTH")
    (19530 . "HANGUL SYLLABLE RIEUL WA RIEUL-PHIEUPH")
    (19531 . "HANGUL SYLLABLE RIEUL WA RIEUL-HIEUH")
    (19532 . "HANGUL SYLLABLE RIEUL WA PIEUP-SIOS")
    (19533 . "HANGUL SYLLABLE RIEUL WA CIEUC")
    (19534 . "HANGUL SYLLABLE RIEUL WA CHIEUCH")
    (19535 . "HANGUL SYLLABLE RIEUL WA KHIEUKH")
    (19536 . "HANGUL SYLLABLE RIEUL WA THIEUTH")
    (19537 . "HANGUL SYLLABLE RIEUL WA PHIEUPH")
    (19538 . "HANGUL SYLLABLE RIEUL WA HIEUH")
    (19539 . "HANGUL SYLLABLE RIEUL WAE KIYEOK")
    (19540 . "HANGUL SYLLABLE RIEUL WAE SSANGKIYEOK")
    (19541 . "HANGUL SYLLABLE RIEUL WAE KIYEOK-SIOS")
    (19542 . "HANGUL SYLLABLE RIEUL WAE NIEUN-CIEUC")
    (19543 . "HANGUL SYLLABLE RIEUL WAE NIEUN-HIEUH")
    (19544 . "HANGUL SYLLABLE RIEUL WAE RIEUL-KIYEOK")
    (19545 . "HANGUL SYLLABLE RIEUL WAE RIEUL-MIEUM")
    (19546 . "HANGUL SYLLABLE RIEUL WAE RIEUL-PIEUP")
    (19547 . "HANGUL SYLLABLE RIEUL WAE RIEUL-SIOS")
    (19548 . "HANGUL SYLLABLE RIEUL WAE RIEUL-THIEUTH")
    (19549 . "HANGUL SYLLABLE RIEUL WAE RIEUL-PHIEUPH")
    (19550 . "HANGUL SYLLABLE RIEUL WAE RIEUL-HIEUH")
    (19551 . "HANGUL SYLLABLE RIEUL WAE MIEUM")
    (19552 . "HANGUL SYLLABLE RIEUL WAE PIEUP")
    (19553 . "HANGUL SYLLABLE RIEUL WAE PIEUP-SIOS")
    (19554 . "HANGUL SYLLABLE RIEUL WAE IEUNG")
    (19555 . "HANGUL SYLLABLE RIEUL WAE CIEUC")
    (19556 . "HANGUL SYLLABLE RIEUL WAE CHIEUCH")
    (19557 . "HANGUL SYLLABLE RIEUL WAE KHIEUKH")
    (19558 . "HANGUL SYLLABLE RIEUL WAE THIEUTH")
    (19559 . "HANGUL SYLLABLE RIEUL WAE PHIEUPH")
    (19560 . "HANGUL SYLLABLE RIEUL WAE HIEUH")
    (19561 . "HANGUL SYLLABLE RIEUL OE SSANGKIYEOK")
    (19562 . "HANGUL SYLLABLE RIEUL OE KIYEOK-SIOS")
    (19563 . "HANGUL SYLLABLE RIEUL OE NIEUN-HIEUH")
    (19564 . "HANGUL SYLLABLE RIEUL OE RIEUL-KIYEOK")
    (19565 . "HANGUL SYLLABLE RIEUL OE RIEUL-MIEUM")
    (19566 . "HANGUL SYLLABLE RIEUL OE RIEUL-PIEUP")
    (19567 . "HANGUL SYLLABLE RIEUL OE RIEUL-SIOS")
    (19568 . "HANGUL SYLLABLE RIEUL OE RIEUL-THIEUTH")
    (19569 . "HANGUL SYLLABLE RIEUL OE RIEUL-PHIEUPH")
    (19570 . "HANGUL SYLLABLE RIEUL OE RIEUL-HIEUH")
    (19571 . "HANGUL SYLLABLE RIEUL OE PIEUP-SIOS")
    (19572 . "HANGUL SYLLABLE RIEUL OE CIEUC")
    (19573 . "HANGUL SYLLABLE RIEUL OE CHIEUCH")
    (19574 . "HANGUL SYLLABLE RIEUL OE KHIEUKH")
    (19575 . "HANGUL SYLLABLE RIEUL OE THIEUTH")
    (19576 . "HANGUL SYLLABLE RIEUL OE PHIEUPH")
    (19577 . "HANGUL SYLLABLE RIEUL OE HIEUH")
    (19578 . "HANGUL SYLLABLE RIEUL YO SSANGKIYEOK")
    (19579 . "HANGUL SYLLABLE RIEUL YO KIYEOK-SIOS")
    (19580 . "HANGUL SYLLABLE RIEUL YO NIEUN-CIEUC")
    (19581 . "HANGUL SYLLABLE RIEUL YO NIEUN-HIEUH")
    (19582 . "HANGUL SYLLABLE RIEUL YO RIEUL-KIYEOK")
    (19583 . "HANGUL SYLLABLE RIEUL YO RIEUL-MIEUM")
    (19584 . "HANGUL SYLLABLE RIEUL YO RIEUL-PIEUP")
    (19585 . "HANGUL SYLLABLE RIEUL YO RIEUL-SIOS")
    (19586 . "HANGUL SYLLABLE RIEUL YO RIEUL-THIEUTH")
    (19587 . "HANGUL SYLLABLE RIEUL YO RIEUL-PHIEUPH")
    (19588 . "HANGUL SYLLABLE RIEUL YO RIEUL-HIEUH")
    (19589 . "HANGUL SYLLABLE RIEUL YO PIEUP-SIOS")
    (19590 . "HANGUL SYLLABLE RIEUL YO SSANGSIOS")
    (19591 . "HANGUL SYLLABLE RIEUL YO CIEUC")
    (19592 . "HANGUL SYLLABLE RIEUL YO CHIEUCH")
    (19593 . "HANGUL SYLLABLE RIEUL YO KHIEUKH")
    (19594 . "HANGUL SYLLABLE RIEUL YO THIEUTH")
    (19595 . "HANGUL SYLLABLE RIEUL YO PHIEUPH")
    (19596 . "HANGUL SYLLABLE RIEUL YO HIEUH")
    (19597 . "HANGUL SYLLABLE RIEUL U SSANGKIYEOK")
    (19598 . "HANGUL SYLLABLE RIEUL U KIYEOK-SIOS")
    (19599 . "HANGUL SYLLABLE RIEUL U NIEUN-CIEUC")
    (19600 . "HANGUL SYLLABLE RIEUL U NIEUN-HIEUH")
    (19601 . "HANGUL SYLLABLE RIEUL U RIEUL-KIYEOK")
    (19602 . "HANGUL SYLLABLE RIEUL U RIEUL-MIEUM")
    (19603 . "HANGUL SYLLABLE RIEUL U RIEUL-SIOS")
    (19604 . "HANGUL SYLLABLE RIEUL U RIEUL-THIEUTH")
    (19605 . "HANGUL SYLLABLE RIEUL U RIEUL-PHIEUPH")
    (19606 . "HANGUL SYLLABLE RIEUL U RIEUL-HIEUH")
    (19607 . "HANGUL SYLLABLE RIEUL U PIEUP-SIOS")
    (19608 . "HANGUL SYLLABLE RIEUL U SSANGSIOS")
    (19609 . "HANGUL SYLLABLE RIEUL U CIEUC")
    (19610 . "HANGUL SYLLABLE RIEUL U CHIEUCH")
    (19611 . "HANGUL SYLLABLE RIEUL U THIEUTH")
    (19612 . "HANGUL SYLLABLE RIEUL U HIEUH")
    (19613 . "HANGUL SYLLABLE RIEUL WEO SSANGKIYEOK")
    (19614 . "HANGUL SYLLABLE RIEUL WEO KIYEOK-SIOS")
    (19615 . "HANGUL SYLLABLE RIEUL WEO NIEUN-CIEUC")
    (19616 . "HANGUL SYLLABLE RIEUL WEO NIEUN-HIEUH")
    (19617 . "HANGUL SYLLABLE RIEUL WEO RIEUL-KIYEOK")
    (19618 . "HANGUL SYLLABLE RIEUL WEO RIEUL-MIEUM")
    (19619 . "HANGUL SYLLABLE RIEUL WEO RIEUL-PIEUP")
    (19620 . "HANGUL SYLLABLE RIEUL WEO RIEUL-SIOS")
    (19621 . "HANGUL SYLLABLE RIEUL WEO RIEUL-THIEUTH")
    (19622 . "HANGUL SYLLABLE RIEUL WEO RIEUL-PHIEUPH")
    (19623 . "HANGUL SYLLABLE RIEUL WEO RIEUL-HIEUH")
    (19624 . "HANGUL SYLLABLE RIEUL WEO PIEUP-SIOS")
    (19625 . "HANGUL SYLLABLE RIEUL WEO CIEUC")
    (19626 . "HANGUL SYLLABLE RIEUL WEO CHIEUCH")
    (19627 . "HANGUL SYLLABLE RIEUL WEO KHIEUKH")
    (19628 . "HANGUL SYLLABLE RIEUL WEO THIEUTH")
    (19629 . "HANGUL SYLLABLE RIEUL WEO PHIEUPH")
    (19630 . "HANGUL SYLLABLE RIEUL WEO HIEUH")
    (19631 . "HANGUL SYLLABLE RIEUL WE SSANGKIYEOK")
    (19632 . "HANGUL SYLLABLE RIEUL WE KIYEOK-SIOS")
    (19633 . "HANGUL SYLLABLE RIEUL WE NIEUN-CIEUC")
    (19634 . "HANGUL SYLLABLE RIEUL WE NIEUN-HIEUH")
    (19635 . "HANGUL SYLLABLE RIEUL WE TIKEUT")
    (19636 . "HANGUL SYLLABLE RIEUL WE RIEUL-KIYEOK")
    (19637 . "HANGUL SYLLABLE RIEUL WE RIEUL-MIEUM")
    (19638 . "HANGUL SYLLABLE RIEUL WE RIEUL-PIEUP")
    (19639 . "HANGUL SYLLABLE RIEUL WE RIEUL-SIOS")
    (19640 . "HANGUL SYLLABLE RIEUL WE RIEUL-THIEUTH")
    (19641 . "HANGUL SYLLABLE RIEUL WE RIEUL-PHIEUPH")
    (19642 . "HANGUL SYLLABLE RIEUL WE RIEUL-HIEUH")
    (19643 . "HANGUL SYLLABLE RIEUL WE PIEUP-SIOS")
    (19644 . "HANGUL SYLLABLE RIEUL WE CIEUC")
    (19645 . "HANGUL SYLLABLE RIEUL WE CHIEUCH")
    (19646 . "HANGUL SYLLABLE RIEUL WE KHIEUKH")
    (19647 . "HANGUL SYLLABLE RIEUL WE THIEUTH")
    (19648 . "HANGUL SYLLABLE RIEUL WE PHIEUPH")
    (19649 . "HANGUL SYLLABLE RIEUL WE HIEUH")
    (19650 . "HANGUL SYLLABLE RIEUL WI SSANGKIYEOK")
    (19651 . "HANGUL SYLLABLE RIEUL WI KIYEOK-SIOS")
    (19652 . "HANGUL SYLLABLE RIEUL WI NIEUN-CIEUC")
    (19653 . "HANGUL SYLLABLE RIEUL WI NIEUN-HIEUH")
    (19654 . "HANGUL SYLLABLE RIEUL WI RIEUL-KIYEOK")
    (19655 . "HANGUL SYLLABLE RIEUL WI RIEUL-MIEUM")
    (19656 . "HANGUL SYLLABLE RIEUL WI RIEUL-PIEUP")
    (19657 . "HANGUL SYLLABLE RIEUL WI RIEUL-SIOS")
    (19658 . "HANGUL SYLLABLE RIEUL WI RIEUL-THIEUTH")
    (19659 . "HANGUL SYLLABLE RIEUL WI RIEUL-PHIEUPH")
    (19660 . "HANGUL SYLLABLE RIEUL WI RIEUL-HIEUH")
    (19661 . "HANGUL SYLLABLE RIEUL WI PIEUP-SIOS")
    (19662 . "HANGUL SYLLABLE RIEUL WI CIEUC")
    (19663 . "HANGUL SYLLABLE RIEUL WI CHIEUCH")
    (19664 . "HANGUL SYLLABLE RIEUL WI KHIEUKH")
    (19665 . "HANGUL SYLLABLE RIEUL WI THIEUTH")
    (19666 . "HANGUL SYLLABLE RIEUL WI PHIEUPH")
    (19667 . "HANGUL SYLLABLE RIEUL WI HIEUH")
    (19668 . "HANGUL SYLLABLE RIEUL YU SSANGKIYEOK")
    (19669 . "HANGUL SYLLABLE RIEUL YU KIYEOK-SIOS")
    (19670 . "HANGUL SYLLABLE RIEUL YU NIEUN-CIEUC")
    (19671 . "HANGUL SYLLABLE RIEUL YU NIEUN-HIEUH")
    (19672 . "HANGUL SYLLABLE RIEUL YU RIEUL-KIYEOK")
    (19673 . "HANGUL SYLLABLE RIEUL YU RIEUL-MIEUM")
    (19674 . "HANGUL SYLLABLE RIEUL YU RIEUL-PIEUP")
    (19675 . "HANGUL SYLLABLE RIEUL YU RIEUL-SIOS")
    (19676 . "HANGUL SYLLABLE RIEUL YU RIEUL-THIEUTH")
    (19677 . "HANGUL SYLLABLE RIEUL YU RIEUL-PHIEUPH")
    (19678 . "HANGUL SYLLABLE RIEUL YU RIEUL-HIEUH")
    (19679 . "HANGUL SYLLABLE RIEUL YU PIEUP-SIOS")
    (19680 . "HANGUL SYLLABLE RIEUL YU SSANGSIOS")
    (19681 . "HANGUL SYLLABLE RIEUL YU CIEUC")
    (19682 . "HANGUL SYLLABLE RIEUL YU CHIEUCH")
    (19683 . "HANGUL SYLLABLE RIEUL YU KHIEUKH")
    (19684 . "HANGUL SYLLABLE RIEUL YU THIEUTH")
    (19685 . "HANGUL SYLLABLE RIEUL YU PHIEUPH")
    (19686 . "HANGUL SYLLABLE RIEUL YU HIEUH")
    (19687 . "HANGUL SYLLABLE RIEUL EU SSANGKIYEOK")
    (19688 . "HANGUL SYLLABLE RIEUL EU KIYEOK-SIOS")
    (19689 . "HANGUL SYLLABLE RIEUL EU NIEUN-CIEUC")
    (19690 . "HANGUL SYLLABLE RIEUL EU NIEUN-HIEUH")
    (19691 . "HANGUL SYLLABLE RIEUL EU RIEUL-MIEUM")
    (19692 . "HANGUL SYLLABLE RIEUL EU RIEUL-PIEUP")
    (19693 . "HANGUL SYLLABLE RIEUL EU RIEUL-THIEUTH")
    (19694 . "HANGUL SYLLABLE RIEUL EU RIEUL-PHIEUPH")
    (19695 . "HANGUL SYLLABLE RIEUL EU RIEUL-HIEUH")
    (19696 . "HANGUL SYLLABLE RIEUL EU PIEUP-SIOS")
    (19697 . "HANGUL SYLLABLE RIEUL EU SSANGSIOS")
    (19698 . "HANGUL SYLLABLE RIEUL EU KHIEUKH")
    (19699 . "HANGUL SYLLABLE RIEUL YI SSANGKIYEOK")
    (19700 . "HANGUL SYLLABLE RIEUL YI KIYEOK-SIOS")
    (19701 . "HANGUL SYLLABLE RIEUL YI NIEUN-CIEUC")
    (19702 . "HANGUL SYLLABLE RIEUL YI NIEUN-HIEUH")
    (19703 . "HANGUL SYLLABLE RIEUL YI TIKEUT")
    (19704 . "HANGUL SYLLABLE RIEUL YI RIEUL-KIYEOK")
    (19705 . "HANGUL SYLLABLE RIEUL YI RIEUL-MIEUM")
    (19706 . "HANGUL SYLLABLE RIEUL YI RIEUL-PIEUP")
    (19707 . "HANGUL SYLLABLE RIEUL YI RIEUL-SIOS")
    (19708 . "HANGUL SYLLABLE RIEUL YI RIEUL-THIEUTH")
    (19709 . "HANGUL SYLLABLE RIEUL YI RIEUL-HIEUH")
    (19710 . "HANGUL SYLLABLE RIEUL YI PIEUP-SIOS")
    (19711 . "HANGUL SYLLABLE RIEUL YI SSANGSIOS")
    (19712 . "HANGUL SYLLABLE RIEUL YI CIEUC")
    (19713 . "HANGUL SYLLABLE RIEUL YI CHIEUCH")
    (19714 . "HANGUL SYLLABLE RIEUL YI KHIEUKH")
    (19715 . "HANGUL SYLLABLE RIEUL YI THIEUTH")
    (19716 . "HANGUL SYLLABLE RIEUL YI PHIEUPH")
    (19717 . "HANGUL SYLLABLE RIEUL YI HIEUH")
    (19718 . "HANGUL SYLLABLE RIEUL I SSANGKIYEOK")
    (19719 . "HANGUL SYLLABLE RIEUL I KIYEOK-SIOS")
    (19720 . "HANGUL SYLLABLE RIEUL I NIEUN-CIEUC")
    (19721 . "HANGUL SYLLABLE RIEUL I NIEUN-HIEUH")
    (19722 . "HANGUL SYLLABLE RIEUL I RIEUL-KIYEOK")
    (19723 . "HANGUL SYLLABLE RIEUL I RIEUL-MIEUM")
    (19724 . "HANGUL SYLLABLE RIEUL I RIEUL-PIEUP")
    (19725 . "HANGUL SYLLABLE RIEUL I RIEUL-SIOS")
    (19726 . "HANGUL SYLLABLE RIEUL I RIEUL-THIEUTH")
    (19727 . "HANGUL SYLLABLE RIEUL I RIEUL-PHIEUPH")
    (19728 . "HANGUL SYLLABLE RIEUL I RIEUL-HIEUH")
    (19729 . "HANGUL SYLLABLE RIEUL I CIEUC")
    (19730 . "HANGUL SYLLABLE RIEUL I CHIEUCH")
    (19731 . "HANGUL SYLLABLE RIEUL I THIEUTH")
    (19732 . "HANGUL SYLLABLE RIEUL I HIEUH")
    (19733 . "HANGUL SYLLABLE MIEUM A KIYEOK-SIOS")
    (19734 . "HANGUL SYLLABLE MIEUM A NIEUN-CIEUC")
    (19735 . "HANGUL SYLLABLE MIEUM A RIEUL-PIEUP")
    (19736 . "HANGUL SYLLABLE MIEUM A RIEUL-THIEUTH")
    (19737 . "HANGUL SYLLABLE MIEUM A RIEUL-PHIEUPH")
    (19738 . "HANGUL SYLLABLE MIEUM A RIEUL-HIEUH")
    (19739 . "HANGUL SYLLABLE MIEUM A PIEUP-SIOS")
    (19740 . "HANGUL SYLLABLE MIEUM AE SSANGKIYEOK")
    (19741 . "HANGUL SYLLABLE MIEUM AE KIYEOK-SIOS")
    (19742 . "HANGUL SYLLABLE MIEUM AE NIEUN-CIEUC")
    (19743 . "HANGUL SYLLABLE MIEUM AE NIEUN-HIEUH")
    (19744 . "HANGUL SYLLABLE MIEUM AE RIEUL-MIEUM")
    (19745 . "HANGUL SYLLABLE MIEUM AE RIEUL-PIEUP")
    (19746 . "HANGUL SYLLABLE MIEUM AE RIEUL-SIOS")
    (19747 . "HANGUL SYLLABLE MIEUM AE RIEUL-THIEUTH")
    (19748 . "HANGUL SYLLABLE MIEUM AE RIEUL-PHIEUPH")
    (19749 . "HANGUL SYLLABLE MIEUM AE RIEUL-HIEUH")
    (19750 . "HANGUL SYLLABLE MIEUM AE PIEUP-SIOS")
    (19751 . "HANGUL SYLLABLE MIEUM AE KHIEUKH")
    (19752 . "HANGUL SYLLABLE MIEUM AE HIEUH")
    (19753 . "HANGUL SYLLABLE MIEUM YA SSANGKIYEOK")
    (19754 . "HANGUL SYLLABLE MIEUM YA KIYEOK-SIOS")
    (19755 . "HANGUL SYLLABLE MIEUM YA NIEUN-CIEUC")
    (19756 . "HANGUL SYLLABLE MIEUM YA NIEUN-HIEUH")
    (19757 . "HANGUL SYLLABLE MIEUM YA TIKEUT")
    (19758 . "HANGUL SYLLABLE MIEUM YA RIEUL-KIYEOK")
    (19759 . "HANGUL SYLLABLE MIEUM YA RIEUL-MIEUM")
    (19760 . "HANGUL SYLLABLE MIEUM YA RIEUL-PIEUP")
    (19761 . "HANGUL SYLLABLE MIEUM YA RIEUL-SIOS")
    (19762 . "HANGUL SYLLABLE MIEUM YA RIEUL-THIEUTH")
    (19763 . "HANGUL SYLLABLE MIEUM YA RIEUL-PHIEUPH")
    (19764 . "HANGUL SYLLABLE MIEUM YA RIEUL-HIEUH")
    (19765 . "HANGUL SYLLABLE MIEUM YA PIEUP-SIOS")
    (19766 . "HANGUL SYLLABLE MIEUM YA SSANGSIOS")
    (19767 . "HANGUL SYLLABLE MIEUM YA CIEUC")
    (19768 . "HANGUL SYLLABLE MIEUM YA CHIEUCH")
    (19769 . "HANGUL SYLLABLE MIEUM YA THIEUTH")
    (19770 . "HANGUL SYLLABLE MIEUM YA PHIEUPH")
    (19771 . "HANGUL SYLLABLE MIEUM YA HIEUH")
    (19772 . "HANGUL SYLLABLE MIEUM YAE KIYEOK")
    (19773 . "HANGUL SYLLABLE MIEUM YAE SSANGKIYEOK")
    (19774 . "HANGUL SYLLABLE MIEUM YAE KIYEOK-SIOS")
    (19775 . "HANGUL SYLLABLE MIEUM YAE NIEUN")
    (19776 . "HANGUL SYLLABLE MIEUM YAE NIEUN-CIEUC")
    (19777 . "HANGUL SYLLABLE MIEUM YAE NIEUN-HIEUH")
    (19778 . "HANGUL SYLLABLE MIEUM YAE TIKEUT")
    (19779 . "HANGUL SYLLABLE MIEUM YAE RIEUL")
    (19780 . "HANGUL SYLLABLE MIEUM YAE RIEUL-KIYEOK")
    (19781 . "HANGUL SYLLABLE MIEUM YAE RIEUL-MIEUM")
    (19782 . "HANGUL SYLLABLE MIEUM YAE RIEUL-PIEUP")
    (19783 . "HANGUL SYLLABLE MIEUM YAE RIEUL-SIOS")
    (19784 . "HANGUL SYLLABLE MIEUM YAE RIEUL-THIEUTH")
    (19785 . "HANGUL SYLLABLE MIEUM YAE RIEUL-PHIEUPH")
    (19786 . "HANGUL SYLLABLE MIEUM YAE RIEUL-HIEUH")
    (19787 . "HANGUL SYLLABLE MIEUM YAE MIEUM")
    (19788 . "HANGUL SYLLABLE MIEUM YAE PIEUP")
    (19789 . "HANGUL SYLLABLE MIEUM YAE PIEUP-SIOS")
    (19790 . "HANGUL SYLLABLE MIEUM YAE SIOS")
    (19791 . "HANGUL SYLLABLE MIEUM YAE SSANGSIOS")
    (19792 . "HANGUL SYLLABLE MIEUM YAE IEUNG")
    (19793 . "HANGUL SYLLABLE MIEUM YAE CIEUC")
    (19794 . "HANGUL SYLLABLE MIEUM YAE CHIEUCH")
    (19795 . "HANGUL SYLLABLE MIEUM YAE KHIEUKH")
    (19796 . "HANGUL SYLLABLE MIEUM YAE THIEUTH")
    (19797 . "HANGUL SYLLABLE MIEUM YAE PHIEUPH")
    (19798 . "HANGUL SYLLABLE MIEUM YAE HIEUH")
    (19799 . "HANGUL SYLLABLE MIEUM EO SSANGKIYEOK")
    (19800 . "HANGUL SYLLABLE MIEUM EO NIEUN-CIEUC")
    (19801 . "HANGUL SYLLABLE MIEUM EO NIEUN-HIEUH")
    (19802 . "HANGUL SYLLABLE MIEUM EO RIEUL-PIEUP")
    (19803 . "HANGUL SYLLABLE MIEUM EO RIEUL-SIOS")
    (19804 . "HANGUL SYLLABLE MIEUM EO RIEUL-THIEUTH")
    (19805 . "HANGUL SYLLABLE MIEUM EO RIEUL-PHIEUPH")
    (19806 . "HANGUL SYLLABLE MIEUM EO RIEUL-HIEUH")
    (19807 . "HANGUL SYLLABLE MIEUM EO PIEUP-SIOS")
    (19808 . "HANGUL SYLLABLE MIEUM EO KHIEUKH")
    (19809 . "HANGUL SYLLABLE MIEUM EO THIEUTH")
    (19810 . "HANGUL SYLLABLE MIEUM EO PHIEUPH")
    (19811 . "HANGUL SYLLABLE MIEUM E SSANGKIYEOK")
    (19812 . "HANGUL SYLLABLE MIEUM E KIYEOK-SIOS")
    (19813 . "HANGUL SYLLABLE MIEUM E NIEUN-CIEUC")
    (19814 . "HANGUL SYLLABLE MIEUM E NIEUN-HIEUH")
    (19815 . "HANGUL SYLLABLE MIEUM E RIEUL-KIYEOK")
    (19816 . "HANGUL SYLLABLE MIEUM E RIEUL-MIEUM")
    (19817 . "HANGUL SYLLABLE MIEUM E RIEUL-PIEUP")
    (19818 . "HANGUL SYLLABLE MIEUM E RIEUL-THIEUTH")
    (19819 . "HANGUL SYLLABLE MIEUM E RIEUL-PHIEUPH")
    (19820 . "HANGUL SYLLABLE MIEUM E RIEUL-HIEUH")
    (19821 . "HANGUL SYLLABLE MIEUM E PIEUP-SIOS")
    (19822 . "HANGUL SYLLABLE MIEUM E KHIEUKH")
    (19823 . "HANGUL SYLLABLE MIEUM E PHIEUPH")
    (19824 . "HANGUL SYLLABLE MIEUM E HIEUH")
    (19825 . "HANGUL SYLLABLE MIEUM YEO SSANGKIYEOK")
    (19826 . "HANGUL SYLLABLE MIEUM YEO KIYEOK-SIOS")
    (19827 . "HANGUL SYLLABLE MIEUM YEO NIEUN-CIEUC")
    (19828 . "HANGUL SYLLABLE MIEUM YEO NIEUN-HIEUH")
    (19829 . "HANGUL SYLLABLE MIEUM YEO RIEUL-KIYEOK")
    (19830 . "HANGUL SYLLABLE MIEUM YEO RIEUL-PIEUP")
    (19831 . "HANGUL SYLLABLE MIEUM YEO RIEUL-SIOS")
    (19832 . "HANGUL SYLLABLE MIEUM YEO RIEUL-THIEUTH")
    (19833 . "HANGUL SYLLABLE MIEUM YEO RIEUL-PHIEUPH")
    (19834 . "HANGUL SYLLABLE MIEUM YEO RIEUL-HIEUH")
    (19835 . "HANGUL SYLLABLE MIEUM YEO PIEUP-SIOS")
    (19836 . "HANGUL SYLLABLE MIEUM YEO CIEUC")
    (19837 . "HANGUL SYLLABLE MIEUM YEO KHIEUKH")
    (19838 . "HANGUL SYLLABLE MIEUM YEO PHIEUPH")
    (19839 . "HANGUL SYLLABLE MIEUM YEO HIEUH")
    (19840 . "HANGUL SYLLABLE MIEUM YE SSANGKIYEOK")
    (19841 . "HANGUL SYLLABLE MIEUM YE KIYEOK-SIOS")
    (19842 . "HANGUL SYLLABLE MIEUM YE NIEUN-CIEUC")
    (19843 . "HANGUL SYLLABLE MIEUM YE NIEUN-HIEUH")
    (19844 . "HANGUL SYLLABLE MIEUM YE RIEUL-KIYEOK")
    (19845 . "HANGUL SYLLABLE MIEUM YE RIEUL-MIEUM")
    (19846 . "HANGUL SYLLABLE MIEUM YE RIEUL-PIEUP")
    (19847 . "HANGUL SYLLABLE MIEUM YE RIEUL-SIOS")
    (19848 . "HANGUL SYLLABLE MIEUM YE RIEUL-THIEUTH")
    (19849 . "HANGUL SYLLABLE MIEUM YE RIEUL-PHIEUPH")
    (19850 . "HANGUL SYLLABLE MIEUM YE RIEUL-HIEUH")
    (19851 . "HANGUL SYLLABLE MIEUM YE PIEUP-SIOS")
    (19852 . "HANGUL SYLLABLE MIEUM YE SSANGSIOS")
    (19853 . "HANGUL SYLLABLE MIEUM YE CIEUC")
    (19854 . "HANGUL SYLLABLE MIEUM YE KHIEUKH")
    (19855 . "HANGUL SYLLABLE MIEUM YE HIEUH")
    (19856 . "HANGUL SYLLABLE MIEUM O SSANGKIYEOK")
    (19857 . "HANGUL SYLLABLE MIEUM O NIEUN-CIEUC")
    (19858 . "HANGUL SYLLABLE MIEUM O NIEUN-HIEUH")
    (19859 . "HANGUL SYLLABLE MIEUM O RIEUL-PIEUP")
    (19860 . "HANGUL SYLLABLE MIEUM O RIEUL-THIEUTH")
    (19861 . "HANGUL SYLLABLE MIEUM O RIEUL-PHIEUPH")
    (19862 . "HANGUL SYLLABLE MIEUM O RIEUL-HIEUH")
    (19863 . "HANGUL SYLLABLE MIEUM O PIEUP-SIOS")
    (19864 . "HANGUL SYLLABLE MIEUM O SSANGSIOS")
    (19865 . "HANGUL SYLLABLE MIEUM O CIEUC")
    (19866 . "HANGUL SYLLABLE MIEUM O KHIEUKH")
    (19867 . "HANGUL SYLLABLE MIEUM O THIEUTH")
    (19868 . "HANGUL SYLLABLE MIEUM O PHIEUPH")
    (19869 . "HANGUL SYLLABLE MIEUM WA SSANGKIYEOK")
    (19870 . "HANGUL SYLLABLE MIEUM WA KIYEOK-SIOS")
    (19871 . "HANGUL SYLLABLE MIEUM WA NIEUN-CIEUC")
    (19872 . "HANGUL SYLLABLE MIEUM WA NIEUN-HIEUH")
    (19873 . "HANGUL SYLLABLE MIEUM WA RIEUL-KIYEOK")
    (19874 . "HANGUL SYLLABLE MIEUM WA RIEUL-MIEUM")
    (19875 . "HANGUL SYLLABLE MIEUM WA RIEUL-PIEUP")
    (19876 . "HANGUL SYLLABLE MIEUM WA RIEUL-SIOS")
    (19877 . "HANGUL SYLLABLE MIEUM WA RIEUL-THIEUTH")
    (19878 . "HANGUL SYLLABLE MIEUM WA RIEUL-PHIEUPH")
    (19879 . "HANGUL SYLLABLE MIEUM WA RIEUL-HIEUH")
    (19880 . "HANGUL SYLLABLE MIEUM WA PIEUP-SIOS")
    (19881 . "HANGUL SYLLABLE MIEUM WA CIEUC")
    (19882 . "HANGUL SYLLABLE MIEUM WA CHIEUCH")
    (19883 . "HANGUL SYLLABLE MIEUM WA KHIEUKH")
    (19884 . "HANGUL SYLLABLE MIEUM WA THIEUTH")
    (19885 . "HANGUL SYLLABLE MIEUM WA PHIEUPH")
    (19886 . "HANGUL SYLLABLE MIEUM WA HIEUH")
    (19887 . "HANGUL SYLLABLE MIEUM WAE KIYEOK")
    (19888 . "HANGUL SYLLABLE MIEUM WAE SSANGKIYEOK")
    (19889 . "HANGUL SYLLABLE MIEUM WAE KIYEOK-SIOS")
    (19890 . "HANGUL SYLLABLE MIEUM WAE NIEUN-CIEUC")
    (19891 . "HANGUL SYLLABLE MIEUM WAE NIEUN-HIEUH")
    (19892 . "HANGUL SYLLABLE MIEUM WAE TIKEUT")
    (19893 . "HANGUL SYLLABLE MIEUM WAE RIEUL-KIYEOK")
    (19894 . "HANGUL SYLLABLE MIEUM WAE RIEUL-MIEUM")
    (19895 . "HANGUL SYLLABLE MIEUM WAE RIEUL-PIEUP")
    (19896 . "HANGUL SYLLABLE MIEUM WAE RIEUL-SIOS")
    (19897 . "HANGUL SYLLABLE MIEUM WAE RIEUL-THIEUTH")
    (19898 . "HANGUL SYLLABLE MIEUM WAE RIEUL-PHIEUPH")
    (19899 . "HANGUL SYLLABLE MIEUM WAE RIEUL-HIEUH")
    (19900 . "HANGUL SYLLABLE MIEUM WAE MIEUM")
    (19901 . "HANGUL SYLLABLE MIEUM WAE PIEUP-SIOS")
    (19902 . "HANGUL SYLLABLE MIEUM WAE SIOS")
    (19903 . "HANGUL SYLLABLE MIEUM WAE IEUNG")
    (19904 . "HANGUL SYLLABLE MIEUM WAE CIEUC")
    (19905 . "HANGUL SYLLABLE MIEUM WAE CHIEUCH")
    (19906 . "HANGUL SYLLABLE MIEUM WAE KHIEUKH")
    (19907 . "HANGUL SYLLABLE MIEUM WAE THIEUTH")
    (19908 . "HANGUL SYLLABLE MIEUM WAE PHIEUPH")
    (19909 . "HANGUL SYLLABLE MIEUM WAE HIEUH")
    (19910 . "HANGUL SYLLABLE MIEUM OE SSANGKIYEOK")
    (19911 . "HANGUL SYLLABLE MIEUM OE KIYEOK-SIOS")
    (19912 . "HANGUL SYLLABLE MIEUM OE NIEUN-CIEUC")
    (19913 . "HANGUL SYLLABLE MIEUM OE NIEUN-HIEUH")
    (19914 . "HANGUL SYLLABLE MIEUM OE RIEUL-KIYEOK")
    (19915 . "HANGUL SYLLABLE MIEUM OE RIEUL-MIEUM")
    (19916 . "HANGUL SYLLABLE MIEUM OE RIEUL-PIEUP")
    (19917 . "HANGUL SYLLABLE MIEUM OE RIEUL-SIOS")
    (19918 . "HANGUL SYLLABLE MIEUM OE RIEUL-THIEUTH")
    (19919 . "HANGUL SYLLABLE MIEUM OE RIEUL-PHIEUPH")
    (19920 . "HANGUL SYLLABLE MIEUM OE RIEUL-HIEUH")
    (19921 . "HANGUL SYLLABLE MIEUM OE PIEUP-SIOS")
    (19922 . "HANGUL SYLLABLE MIEUM OE SSANGSIOS")
    (19923 . "HANGUL SYLLABLE MIEUM OE CIEUC")
    (19924 . "HANGUL SYLLABLE MIEUM OE CHIEUCH")
    (19925 . "HANGUL SYLLABLE MIEUM OE KHIEUKH")
    (19926 . "HANGUL SYLLABLE MIEUM OE THIEUTH")
    (19927 . "HANGUL SYLLABLE MIEUM OE PHIEUPH")
    (19928 . "HANGUL SYLLABLE MIEUM OE HIEUH")
    (19929 . "HANGUL SYLLABLE MIEUM YO SSANGKIYEOK")
    (19930 . "HANGUL SYLLABLE MIEUM YO KIYEOK-SIOS")
    (19931 . "HANGUL SYLLABLE MIEUM YO NIEUN-CIEUC")
    (19932 . "HANGUL SYLLABLE MIEUM YO NIEUN-HIEUH")
    (19933 . "HANGUL SYLLABLE MIEUM YO RIEUL-KIYEOK")
    (19934 . "HANGUL SYLLABLE MIEUM YO RIEUL-MIEUM")
    (19935 . "HANGUL SYLLABLE MIEUM YO RIEUL-PIEUP")
    (19936 . "HANGUL SYLLABLE MIEUM YO RIEUL-SIOS")
    (19937 . "HANGUL SYLLABLE MIEUM YO RIEUL-THIEUTH")
    (19938 . "HANGUL SYLLABLE MIEUM YO RIEUL-PHIEUPH")
    (19939 . "HANGUL SYLLABLE MIEUM YO RIEUL-HIEUH")
    (19940 . "HANGUL SYLLABLE MIEUM YO PIEUP-SIOS")
    (19941 . "HANGUL SYLLABLE MIEUM YO SSANGSIOS")
    (19942 . "HANGUL SYLLABLE MIEUM YO CIEUC")
    (19943 . "HANGUL SYLLABLE MIEUM YO CHIEUCH")
    (19944 . "HANGUL SYLLABLE MIEUM YO KHIEUKH")
    (19945 . "HANGUL SYLLABLE MIEUM YO THIEUTH")
    (19946 . "HANGUL SYLLABLE MIEUM YO PHIEUPH")
    (19947 . "HANGUL SYLLABLE MIEUM YO HIEUH")
    (19948 . "HANGUL SYLLABLE MIEUM U KIYEOK-SIOS")
    (19949 . "HANGUL SYLLABLE MIEUM U NIEUN-CIEUC")
    (19950 . "HANGUL SYLLABLE MIEUM U RIEUL-PIEUP")
    (19951 . "HANGUL SYLLABLE MIEUM U RIEUL-THIEUTH")
    (19952 . "HANGUL SYLLABLE MIEUM U RIEUL-PHIEUPH")
    (19953 . "HANGUL SYLLABLE MIEUM U RIEUL-HIEUH")
    (19954 . "HANGUL SYLLABLE MIEUM U PIEUP-SIOS")
    (19955 . "HANGUL SYLLABLE MIEUM U SSANGSIOS")
    (19956 . "HANGUL SYLLABLE MIEUM U CIEUC")
    (19957 . "HANGUL SYLLABLE MIEUM U KHIEUKH")
    (19958 . "HANGUL SYLLABLE MIEUM U PHIEUPH")
    (19959 . "HANGUL SYLLABLE MIEUM WEO SSANGKIYEOK")
    (19960 . "HANGUL SYLLABLE MIEUM WEO KIYEOK-SIOS")
    (19961 . "HANGUL SYLLABLE MIEUM WEO NIEUN-CIEUC")
    (19962 . "HANGUL SYLLABLE MIEUM WEO NIEUN-HIEUH")
    (19963 . "HANGUL SYLLABLE MIEUM WEO RIEUL-KIYEOK")
    (19964 . "HANGUL SYLLABLE MIEUM WEO RIEUL-MIEUM")
    (19965 . "HANGUL SYLLABLE MIEUM WEO RIEUL-PIEUP")
    (19966 . "HANGUL SYLLABLE MIEUM WEO RIEUL-SIOS")
    (19967 . "HANGUL SYLLABLE MIEUM WEO RIEUL-THIEUTH")
    (19968 . "CJK UNIFIED IDEOGRAPH-%4x")
    (63744 . "CJK COMPATIBILITY IDEOGRAPH-%4x")
    (64256 . "LATIN SMALL LIGATURE FF")
    (64257 . "LATIN SMALL LIGATURE FI")
    (64258 . "LATIN SMALL LIGATURE FL")
    (64259 . "LATIN SMALL LIGATURE FFI")
    (64260 . "LATIN SMALL LIGATURE FFL")
    (64261 . "LATIN SMALL LIGATURE LONG S T")
    (64262 . "LATIN SMALL LIGATURE ST")
    (64275 . "ARMENIAN SMALL LIGATURE MEN NOW")
    (64276 . "ARMENIAN SMALL LIGATURE MEN ECH")
    (64277 . "ARMENIAN SMALL LIGATURE MEN INI")
    (64278 . "ARMENIAN SMALL LIGATURE VEW NOW")
    (64279 . "ARMENIAN SMALL LIGATURE MEN XEH")
    (64286 . "HEBREW POINT JUDEO-SPANISH VARIKA")
    (64287 . "HEBREW LIGATURE YIDDISH YOD YOD PATAH")
    (64288 . "HEBREW LETTER ALTERNATIVE AYIN")
    (64289 . "HEBREW LETTER WIDE ALEF")
    (64290 . "HEBREW LETTER WIDE DALET")
    (64291 . "HEBREW LETTER WIDE HE")
    (64292 . "HEBREW LETTER WIDE KAF")
    (64293 . "HEBREW LETTER WIDE LAMED")
    (64294 . "HEBREW LETTER WIDE FINAL MEM")
    (64295 . "HEBREW LETTER WIDE RESH")
    (64296 . "HEBREW LETTER WIDE TAV")
    (64297 . "HEBREW LETTER ALTERNATIVE PLUS SIGN")
    (64298 . "HEBREW LETTER SHIN WITH SHIN DOT")
    (64299 . "HEBREW LETTER SHIN WITH SIN DOT")
    (64300 . "HEBREW LETTER SHIN WITH DAGESH AND SHIN DOT")
    (64301 . "HEBREW LETTER SHIN WITH DAGESH AND SIN DOT")
    (64302 . "HEBREW LETTER ALEF WITH PATAH")
    (64303 . "HEBREW LETTER ALEF WITH QAMATS")
    (64304 . "HEBREW LETTER ALEF WITH MAPIQ")
    (64305 . "HEBREW LETTER BET WITH DAGESH")
    (64306 . "HEBREW LETTER GIMEL WITH DAGESH")
    (64307 . "HEBREW LETTER DALET WITH DAGESH")
    (64308 . "HEBREW LETTER HE WITH MAPIQ")
    (64309 . "HEBREW LETTER VAV WITH DAGESH")
    (64310 . "HEBREW LETTER ZAYIN WITH DAGESH")
    (64312 . "HEBREW LETTER TET WITH DAGESH")
    (64313 . "HEBREW LETTER YOD WITH DAGESH")
    (64314 . "HEBREW LETTER FINAL KAF WITH DAGESH")
    (64315 . "HEBREW LETTER KAF WITH DAGESH")
    (64316 . "HEBREW LETTER LAMED WITH DAGESH")
    (64318 . "HEBREW LETTER MEM WITH DAGESH")
    (64320 . "HEBREW LETTER NUN WITH DAGESH")
    (64321 . "HEBREW LETTER SAMEKH WITH DAGESH")
    (64323 . "HEBREW LETTER FINAL PE WITH DAGESH")
    (64324 . "HEBREW LETTER PE WITH DAGESH")
    (64326 . "HEBREW LETTER TSADI WITH DAGESH")
    (64327 . "HEBREW LETTER QOF WITH DAGESH")
    (64328 . "HEBREW LETTER RESH WITH DAGESH")
    (64329 . "HEBREW LETTER SHIN WITH DAGESH")
    (64330 . "HEBREW LETTER TAV WITH DAGESH")
    (64331 . "HEBREW LETTER VAV WITH HOLAM")
    (64332 . "HEBREW LETTER BET WITH RAFE")
    (64333 . "HEBREW LETTER KAF WITH RAFE")
    (64334 . "HEBREW LETTER PE WITH RAFE")
    (64335 . "HEBREW LIGATURE ALEF LAMED")
    (64336 . "ARABIC LETTER ALEF WASLA ISOLATED FORM")
    (64337 . "ARABIC LETTER ALEF WASLA FINAL FORM")
    (64338 . "ARABIC LETTER BEEH ISOLATED FORM")
    (64339 . "ARABIC LETTER BEEH FINAL FORM")
    (64340 . "ARABIC LETTER BEEH INITIAL FORM")
    (64341 . "ARABIC LETTER BEEH MEDIAL FORM")
    (64342 . "ARABIC LETTER PEH ISOLATED FORM")
    (64343 . "ARABIC LETTER PEH FINAL FORM")
    (64344 . "ARABIC LETTER PEH INITIAL FORM")
    (64345 . "ARABIC LETTER PEH MEDIAL FORM")
    (64346 . "ARABIC LETTER BEHEH ISOLATED FORM")
    (64347 . "ARABIC LETTER BEHEH FINAL FORM")
    (64348 . "ARABIC LETTER BEHEH INITIAL FORM")
    (64349 . "ARABIC LETTER BEHEH MEDIAL FORM")
    (64350 . "ARABIC LETTER TTEHEH ISOLATED FORM")
    (64351 . "ARABIC LETTER TTEHEH FINAL FORM")
    (64352 . "ARABIC LETTER TTEHEH INITIAL FORM")
    (64353 . "ARABIC LETTER TTEHEH MEDIAL FORM")
    (64354 . "ARABIC LETTER TEHEH ISOLATED FORM")
    (64355 . "ARABIC LETTER TEHEH FINAL FORM")
    (64356 . "ARABIC LETTER TEHEH INITIAL FORM")
    (64357 . "ARABIC LETTER TEHEH MEDIAL FORM")
    (64358 . "ARABIC LETTER TTEH ISOLATED FORM")
    (64359 . "ARABIC LETTER TTEH FINAL FORM")
    (64360 . "ARABIC LETTER TTEH INITIAL FORM")
    (64361 . "ARABIC LETTER TTEH MEDIAL FORM")
    (64362 . "ARABIC LETTER VEH ISOLATED FORM")
    (64363 . "ARABIC LETTER VEH FINAL FORM")
    (64364 . "ARABIC LETTER VEH INITIAL FORM")
    (64365 . "ARABIC LETTER VEH MEDIAL FORM")
    (64366 . "ARABIC LETTER PEHEH ISOLATED FORM")
    (64367 . "ARABIC LETTER PEHEH FINAL FORM")
    (64368 . "ARABIC LETTER PEHEH INITIAL FORM")
    (64369 . "ARABIC LETTER PEHEH MEDIAL FORM")
    (64370 . "ARABIC LETTER DYEH ISOLATED FORM")
    (64371 . "ARABIC LETTER DYEH FINAL FORM")
    (64372 . "ARABIC LETTER DYEH INITIAL FORM")
    (64373 . "ARABIC LETTER DYEH MEDIAL FORM")
    (64374 . "ARABIC LETTER NYEH ISOLATED FORM")
    (64375 . "ARABIC LETTER NYEH FINAL FORM")
    (64376 . "ARABIC LETTER NYEH INITIAL FORM")
    (64377 . "ARABIC LETTER NYEH MEDIAL FORM")
    (64378 . "ARABIC LETTER TCHEH ISOLATED FORM")
    (64379 . "ARABIC LETTER TCHEH FINAL FORM")
    (64380 . "ARABIC LETTER TCHEH INITIAL FORM")
    (64381 . "ARABIC LETTER TCHEH MEDIAL FORM")
    (64382 . "ARABIC LETTER TCHEHEH ISOLATED FORM")
    (64383 . "ARABIC LETTER TCHEHEH FINAL FORM")
    (64384 . "ARABIC LETTER TCHEHEH INITIAL FORM")
    (64385 . "ARABIC LETTER TCHEHEH MEDIAL FORM")
    (64386 . "ARABIC LETTER DDAHAL ISOLATED FORM")
    (64387 . "ARABIC LETTER DDAHAL FINAL FORM")
    (64388 . "ARABIC LETTER DAHAL ISOLATED FORM")
    (64389 . "ARABIC LETTER DAHAL FINAL FORM")
    (64390 . "ARABIC LETTER DUL ISOLATED FORM")
    (64391 . "ARABIC LETTER DUL FINAL FORM")
    (64392 . "ARABIC LETTER DDAL ISOLATED FORM")
    (64393 . "ARABIC LETTER DDAL FINAL FORM")
    (64394 . "ARABIC LETTER JEH ISOLATED FORM")
    (64395 . "ARABIC LETTER JEH FINAL FORM")
    (64396 . "ARABIC LETTER RREH ISOLATED FORM")
    (64397 . "ARABIC LETTER RREH FINAL FORM")
    (64398 . "ARABIC LETTER KEHEH ISOLATED FORM")
    (64399 . "ARABIC LETTER KEHEH FINAL FORM")
    (64400 . "ARABIC LETTER KEHEH INITIAL FORM")
    (64401 . "ARABIC LETTER KEHEH MEDIAL FORM")
    (64402 . "ARABIC LETTER GAF ISOLATED FORM")
    (64403 . "ARABIC LETTER GAF FINAL FORM")
    (64404 . "ARABIC LETTER GAF INITIAL FORM")
    (64405 . "ARABIC LETTER GAF MEDIAL FORM")
    (64406 . "ARABIC LETTER GUEH ISOLATED FORM")
    (64407 . "ARABIC LETTER GUEH FINAL FORM")
    (64408 . "ARABIC LETTER GUEH INITIAL FORM")
    (64409 . "ARABIC LETTER GUEH MEDIAL FORM")
    (64410 . "ARABIC LETTER NGOEH ISOLATED FORM")
    (64411 . "ARABIC LETTER NGOEH FINAL FORM")
    (64412 . "ARABIC LETTER NGOEH INITIAL FORM")
    (64413 . "ARABIC LETTER NGOEH MEDIAL FORM")
    (64414 . "ARABIC LETTER NOON GHUNNA ISOLATED FORM")
    (64415 . "ARABIC LETTER NOON GHUNNA FINAL FORM")
    (64416 . "ARABIC LETTER RNOON ISOLATED FORM")
    (64417 . "ARABIC LETTER RNOON FINAL FORM")
    (64418 . "ARABIC LETTER RNOON INITIAL FORM")
    (64419 . "ARABIC LETTER RNOON MEDIAL FORM")
    (64420 . "ARABIC LETTER HEH WITH YEH ABOVE ISOLATED FORM")
    (64421 . "ARABIC LETTER HEH WITH YEH ABOVE FINAL FORM")
    (64422 . "ARABIC LETTER HEH GOAL ISOLATED FORM")
    (64423 . "ARABIC LETTER HEH GOAL FINAL FORM")
    (64424 . "ARABIC LETTER HEH GOAL INITIAL FORM")
    (64425 . "ARABIC LETTER HEH GOAL MEDIAL FORM")
    (64426 . "ARABIC LETTER HEH DOACHASHMEE ISOLATED FORM")
    (64427 . "ARABIC LETTER HEH DOACHASHMEE FINAL FORM")
    (64428 . "ARABIC LETTER HEH DOACHASHMEE INITIAL FORM")
    (64429 . "ARABIC LETTER HEH DOACHASHMEE MEDIAL FORM")
    (64430 . "ARABIC LETTER YEH BARREE ISOLATED FORM")
    (64431 . "ARABIC LETTER YEH BARREE FINAL FORM")
    (64432 . "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE ISOLATED FORM")
    (64433 . "ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM")
    (64467 . "ARABIC LETTER NG ISOLATED FORM")
    (64468 . "ARABIC LETTER NG FINAL FORM")
    (64469 . "ARABIC LETTER NG INITIAL FORM")
    (64470 . "ARABIC LETTER NG MEDIAL FORM")
    (64471 . "ARABIC LETTER U ISOLATED FORM")
    (64472 . "ARABIC LETTER U FINAL FORM")
    (64473 . "ARABIC LETTER OE ISOLATED FORM")
    (64474 . "ARABIC LETTER OE FINAL FORM")
    (64475 . "ARABIC LETTER YU ISOLATED FORM")
    (64476 . "ARABIC LETTER YU FINAL FORM")
    (64477 . "ARABIC LETTER U WITH HAMZA ABOVE ISOLATED FORM")
    (64478 . "ARABIC LETTER VE ISOLATED FORM")
    (64479 . "ARABIC LETTER VE FINAL FORM")
    (64480 . "ARABIC LETTER KIRGHIZ OE ISOLATED FORM")
    (64481 . "ARABIC LETTER KIRGHIZ OE FINAL FORM")
    (64482 . "ARABIC LETTER KIRGHIZ YU ISOLATED FORM")
    (64483 . "ARABIC LETTER KIRGHIZ YU FINAL FORM")
    (64484 . "ARABIC LETTER E ISOLATED FORM")
    (64485 . "ARABIC LETTER E FINAL FORM")
    (64486 . "ARABIC LETTER E INITIAL FORM")
    (64487 . "ARABIC LETTER E MEDIAL FORM")
    (64488 . "ARABIC LETTER UIGHUR KAZAKH KIRGHIZ ALEF MAKSURA INITIAL FORM")
    (64489 . "ARABIC LETTER UIGHUR KAZAKH KIRGHIZ ALEF MAKSURA MEDIAL FORM")
    (64490 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF ISOLATED FORM")
    (64491 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF FINAL FORM")
    (64492 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH AE ISOLATED FORM")
    (64493 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH AE FINAL FORM")
    (64494 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH WAW ISOLATED FORM")
    (64495 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH WAW FINAL FORM")
    (64496 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH U ISOLATED FORM")
    (64497 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH U FINAL FORM")
    (64498 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH OE ISOLATED FORM")
    (64499 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH OE FINAL FORM")
    (64500 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YU ISOLATED FORM")
    (64501 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YU FINAL FORM")
    (64502 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E ISOLATED FORM")
    (64503 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E FINAL FORM")
    (64504 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH E INITIAL FORM")
    (64505 . "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA ISOLATED FORM")
    (64506 . "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA FINAL FORM")
    (64507 . "ARABIC LIGATURE UIGHUR KIRGHIZ YEH WITH HAMZA ABOVE WITH ALEF MAKSURA INITIAL FORM")
    (64508 . "ARABIC LETTER FARSI YEH ISOLATED FORM")
    (64509 . "ARABIC LETTER FARSI YEH FINAL FORM")
    (64510 . "ARABIC LETTER FARSI YEH INITIAL FORM")
    (64511 . "ARABIC LETTER FARSI YEH MEDIAL FORM")
    (64512 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH JEEM ISOLATED FORM")
    (64513 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HAH ISOLATED FORM")
    (64514 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM ISOLATED FORM")
    (64515 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF MAKSURA ISOLATED FORM")
    (64516 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YEH ISOLATED FORM")
    (64517 . "ARABIC LIGATURE BEH WITH JEEM ISOLATED FORM")
    (64518 . "ARABIC LIGATURE BEH WITH HAH ISOLATED FORM")
    (64519 . "ARABIC LIGATURE BEH WITH KHAH ISOLATED FORM")
    (64520 . "ARABIC LIGATURE BEH WITH MEEM ISOLATED FORM")
    (64521 . "ARABIC LIGATURE BEH WITH ALEF MAKSURA ISOLATED FORM")
    (64522 . "ARABIC LIGATURE BEH WITH YEH ISOLATED FORM")
    (64523 . "ARABIC LIGATURE TEH WITH JEEM ISOLATED FORM")
    (64524 . "ARABIC LIGATURE TEH WITH HAH ISOLATED FORM")
    (64525 . "ARABIC LIGATURE TEH WITH KHAH ISOLATED FORM")
    (64526 . "ARABIC LIGATURE TEH WITH MEEM ISOLATED FORM")
    (64527 . "ARABIC LIGATURE TEH WITH ALEF MAKSURA ISOLATED FORM")
    (64528 . "ARABIC LIGATURE TEH WITH YEH ISOLATED FORM")
    (64529 . "ARABIC LIGATURE THEH WITH JEEM ISOLATED FORM")
    (64530 . "ARABIC LIGATURE THEH WITH MEEM ISOLATED FORM")
    (64531 . "ARABIC LIGATURE THEH WITH ALEF MAKSURA ISOLATED FORM")
    (64532 . "ARABIC LIGATURE THEH WITH YEH ISOLATED FORM")
    (64533 . "ARABIC LIGATURE JEEM WITH HAH ISOLATED FORM")
    (64534 . "ARABIC LIGATURE JEEM WITH MEEM ISOLATED FORM")
    (64535 . "ARABIC LIGATURE HAH WITH JEEM ISOLATED FORM")
    (64536 . "ARABIC LIGATURE HAH WITH MEEM ISOLATED FORM")
    (64537 . "ARABIC LIGATURE KHAH WITH JEEM ISOLATED FORM")
    (64538 . "ARABIC LIGATURE KHAH WITH HAH ISOLATED FORM")
    (64539 . "ARABIC LIGATURE KHAH WITH MEEM ISOLATED FORM")
    (64540 . "ARABIC LIGATURE SEEN WITH JEEM ISOLATED FORM")
    (64541 . "ARABIC LIGATURE SEEN WITH HAH ISOLATED FORM")
    (64542 . "ARABIC LIGATURE SEEN WITH KHAH ISOLATED FORM")
    (64543 . "ARABIC LIGATURE SEEN WITH MEEM ISOLATED FORM")
    (64544 . "ARABIC LIGATURE SAD WITH HAH ISOLATED FORM")
    (64545 . "ARABIC LIGATURE SAD WITH MEEM ISOLATED FORM")
    (64546 . "ARABIC LIGATURE DAD WITH JEEM ISOLATED FORM")
    (64547 . "ARABIC LIGATURE DAD WITH HAH ISOLATED FORM")
    (64548 . "ARABIC LIGATURE DAD WITH KHAH ISOLATED FORM")
    (64549 . "ARABIC LIGATURE DAD WITH MEEM ISOLATED FORM")
    (64550 . "ARABIC LIGATURE TAH WITH HAH ISOLATED FORM")
    (64551 . "ARABIC LIGATURE TAH WITH MEEM ISOLATED FORM")
    (64552 . "ARABIC LIGATURE ZAH WITH MEEM ISOLATED FORM")
    (64553 . "ARABIC LIGATURE AIN WITH JEEM ISOLATED FORM")
    (64554 . "ARABIC LIGATURE AIN WITH MEEM ISOLATED FORM")
    (64555 . "ARABIC LIGATURE GHAIN WITH JEEM ISOLATED FORM")
    (64556 . "ARABIC LIGATURE GHAIN WITH MEEM ISOLATED FORM")
    (64557 . "ARABIC LIGATURE FEH WITH JEEM ISOLATED FORM")
    (64558 . "ARABIC LIGATURE FEH WITH HAH ISOLATED FORM")
    (64559 . "ARABIC LIGATURE FEH WITH KHAH ISOLATED FORM")
    (64560 . "ARABIC LIGATURE FEH WITH MEEM ISOLATED FORM")
    (64561 . "ARABIC LIGATURE FEH WITH ALEF MAKSURA ISOLATED FORM")
    (64562 . "ARABIC LIGATURE FEH WITH YEH ISOLATED FORM")
    (64563 . "ARABIC LIGATURE QAF WITH HAH ISOLATED FORM")
    (64564 . "ARABIC LIGATURE QAF WITH MEEM ISOLATED FORM")
    (64565 . "ARABIC LIGATURE QAF WITH ALEF MAKSURA ISOLATED FORM")
    (64566 . "ARABIC LIGATURE QAF WITH YEH ISOLATED FORM")
    (64567 . "ARABIC LIGATURE KAF WITH ALEF ISOLATED FORM")
    (64568 . "ARABIC LIGATURE KAF WITH JEEM ISOLATED FORM")
    (64569 . "ARABIC LIGATURE KAF WITH HAH ISOLATED FORM")
    (64570 . "ARABIC LIGATURE KAF WITH KHAH ISOLATED FORM")
    (64571 . "ARABIC LIGATURE KAF WITH LAM ISOLATED FORM")
    (64572 . "ARABIC LIGATURE KAF WITH MEEM ISOLATED FORM")
    (64573 . "ARABIC LIGATURE KAF WITH ALEF MAKSURA ISOLATED FORM")
    (64574 . "ARABIC LIGATURE KAF WITH YEH ISOLATED FORM")
    (64575 . "ARABIC LIGATURE LAM WITH JEEM ISOLATED FORM")
    (64576 . "ARABIC LIGATURE LAM WITH HAH ISOLATED FORM")
    (64577 . "ARABIC LIGATURE LAM WITH KHAH ISOLATED FORM")
    (64578 . "ARABIC LIGATURE LAM WITH MEEM ISOLATED FORM")
    (64579 . "ARABIC LIGATURE LAM WITH ALEF MAKSURA ISOLATED FORM")
    (64580 . "ARABIC LIGATURE LAM WITH YEH ISOLATED FORM")
    (64581 . "ARABIC LIGATURE MEEM WITH JEEM ISOLATED FORM")
    (64582 . "ARABIC LIGATURE MEEM WITH HAH ISOLATED FORM")
    (64583 . "ARABIC LIGATURE MEEM WITH KHAH ISOLATED FORM")
    (64584 . "ARABIC LIGATURE MEEM WITH MEEM ISOLATED FORM")
    (64585 . "ARABIC LIGATURE MEEM WITH ALEF MAKSURA ISOLATED FORM")
    (64586 . "ARABIC LIGATURE MEEM WITH YEH ISOLATED FORM")
    (64587 . "ARABIC LIGATURE NOON WITH JEEM ISOLATED FORM")
    (64588 . "ARABIC LIGATURE NOON WITH HAH ISOLATED FORM")
    (64589 . "ARABIC LIGATURE NOON WITH KHAH ISOLATED FORM")
    (64590 . "ARABIC LIGATURE NOON WITH MEEM ISOLATED FORM")
    (64591 . "ARABIC LIGATURE NOON WITH ALEF MAKSURA ISOLATED FORM")
    (64592 . "ARABIC LIGATURE NOON WITH YEH ISOLATED FORM")
    (64593 . "ARABIC LIGATURE HEH WITH JEEM ISOLATED FORM")
    (64594 . "ARABIC LIGATURE HEH WITH MEEM ISOLATED FORM")
    (64595 . "ARABIC LIGATURE HEH WITH ALEF MAKSURA ISOLATED FORM")
    (64596 . "ARABIC LIGATURE HEH WITH YEH ISOLATED FORM")
    (64597 . "ARABIC LIGATURE YEH WITH JEEM ISOLATED FORM")
    (64598 . "ARABIC LIGATURE YEH WITH HAH ISOLATED FORM")
    (64599 . "ARABIC LIGATURE YEH WITH KHAH ISOLATED FORM")
    (64600 . "ARABIC LIGATURE YEH WITH MEEM ISOLATED FORM")
    (64601 . "ARABIC LIGATURE YEH WITH ALEF MAKSURA ISOLATED FORM")
    (64602 . "ARABIC LIGATURE YEH WITH YEH ISOLATED FORM")
    (64603 . "ARABIC LIGATURE THAL WITH SUPERSCRIPT ALEF ISOLATED FORM")
    (64604 . "ARABIC LIGATURE REH WITH SUPERSCRIPT ALEF ISOLATED FORM")
    (64605 . "ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF ISOLATED FORM")
    (64606 . "ARABIC LIGATURE SHADDA WITH DAMMATAN ISOLATED FORM")
    (64607 . "ARABIC LIGATURE SHADDA WITH KASRATAN ISOLATED FORM")
    (64608 . "ARABIC LIGATURE SHADDA WITH FATHA ISOLATED FORM")
    (64609 . "ARABIC LIGATURE SHADDA WITH DAMMA ISOLATED FORM")
    (64610 . "ARABIC LIGATURE SHADDA WITH KASRA ISOLATED FORM")
    (64611 . "ARABIC LIGATURE SHADDA WITH SUPERSCRIPT ALEF ISOLATED FORM")
    (64612 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH REH FINAL FORM")
    (64613 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ZAIN FINAL FORM")
    (64614 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM FINAL FORM")
    (64615 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH NOON FINAL FORM")
    (64616 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH ALEF MAKSURA FINAL FORM")
    (64617 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH YEH FINAL FORM")
    (64618 . "ARABIC LIGATURE BEH WITH REH FINAL FORM")
    (64619 . "ARABIC LIGATURE BEH WITH ZAIN FINAL FORM")
    (64620 . "ARABIC LIGATURE BEH WITH MEEM FINAL FORM")
    (64621 . "ARABIC LIGATURE BEH WITH NOON FINAL FORM")
    (64622 . "ARABIC LIGATURE BEH WITH ALEF MAKSURA FINAL FORM")
    (64623 . "ARABIC LIGATURE BEH WITH YEH FINAL FORM")
    (64624 . "ARABIC LIGATURE TEH WITH REH FINAL FORM")
    (64625 . "ARABIC LIGATURE TEH WITH ZAIN FINAL FORM")
    (64626 . "ARABIC LIGATURE TEH WITH MEEM FINAL FORM")
    (64627 . "ARABIC LIGATURE TEH WITH NOON FINAL FORM")
    (64628 . "ARABIC LIGATURE TEH WITH ALEF MAKSURA FINAL FORM")
    (64629 . "ARABIC LIGATURE TEH WITH YEH FINAL FORM")
    (64630 . "ARABIC LIGATURE THEH WITH REH FINAL FORM")
    (64631 . "ARABIC LIGATURE THEH WITH ZAIN FINAL FORM")
    (64632 . "ARABIC LIGATURE THEH WITH MEEM FINAL FORM")
    (64633 . "ARABIC LIGATURE THEH WITH NOON FINAL FORM")
    (64634 . "ARABIC LIGATURE THEH WITH ALEF MAKSURA FINAL FORM")
    (64635 . "ARABIC LIGATURE THEH WITH YEH FINAL FORM")
    (64636 . "ARABIC LIGATURE FEH WITH ALEF MAKSURA FINAL FORM")
    (64637 . "ARABIC LIGATURE FEH WITH YEH FINAL FORM")
    (64638 . "ARABIC LIGATURE QAF WITH ALEF MAKSURA FINAL FORM")
    (64639 . "ARABIC LIGATURE QAF WITH YEH FINAL FORM")
    (64640 . "ARABIC LIGATURE KAF WITH ALEF FINAL FORM")
    (64641 . "ARABIC LIGATURE KAF WITH LAM FINAL FORM")
    (64642 . "ARABIC LIGATURE KAF WITH MEEM FINAL FORM")
    (64643 . "ARABIC LIGATURE KAF WITH ALEF MAKSURA FINAL FORM")
    (64644 . "ARABIC LIGATURE KAF WITH YEH FINAL FORM")
    (64645 . "ARABIC LIGATURE LAM WITH MEEM FINAL FORM")
    (64646 . "ARABIC LIGATURE LAM WITH ALEF MAKSURA FINAL FORM")
    (64647 . "ARABIC LIGATURE LAM WITH YEH FINAL FORM")
    (64648 . "ARABIC LIGATURE MEEM WITH ALEF FINAL FORM")
    (64649 . "ARABIC LIGATURE MEEM WITH MEEM FINAL FORM")
    (64650 . "ARABIC LIGATURE NOON WITH REH FINAL FORM")
    (64651 . "ARABIC LIGATURE NOON WITH ZAIN FINAL FORM")
    (64652 . "ARABIC LIGATURE NOON WITH MEEM FINAL FORM")
    (64653 . "ARABIC LIGATURE NOON WITH NOON FINAL FORM")
    (64654 . "ARABIC LIGATURE NOON WITH ALEF MAKSURA FINAL FORM")
    (64655 . "ARABIC LIGATURE NOON WITH YEH FINAL FORM")
    (64656 . "ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF FINAL FORM")
    (64657 . "ARABIC LIGATURE YEH WITH REH FINAL FORM")
    (64658 . "ARABIC LIGATURE YEH WITH ZAIN FINAL FORM")
    (64659 . "ARABIC LIGATURE YEH WITH MEEM FINAL FORM")
    (64660 . "ARABIC LIGATURE YEH WITH NOON FINAL FORM")
    (64661 . "ARABIC LIGATURE YEH WITH ALEF MAKSURA FINAL FORM")
    (64662 . "ARABIC LIGATURE YEH WITH YEH FINAL FORM")
    (64663 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH JEEM INITIAL FORM")
    (64664 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HAH INITIAL FORM")
    (64665 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH KHAH INITIAL FORM")
    (64666 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM INITIAL FORM")
    (64667 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HEH INITIAL FORM")
    (64668 . "ARABIC LIGATURE BEH WITH JEEM INITIAL FORM")
    (64669 . "ARABIC LIGATURE BEH WITH HAH INITIAL FORM")
    (64670 . "ARABIC LIGATURE BEH WITH KHAH INITIAL FORM")
    (64671 . "ARABIC LIGATURE BEH WITH MEEM INITIAL FORM")
    (64672 . "ARABIC LIGATURE BEH WITH HEH INITIAL FORM")
    (64673 . "ARABIC LIGATURE TEH WITH JEEM INITIAL FORM")
    (64674 . "ARABIC LIGATURE TEH WITH HAH INITIAL FORM")
    (64675 . "ARABIC LIGATURE TEH WITH KHAH INITIAL FORM")
    (64676 . "ARABIC LIGATURE TEH WITH MEEM INITIAL FORM")
    (64677 . "ARABIC LIGATURE TEH WITH HEH INITIAL FORM")
    (64678 . "ARABIC LIGATURE THEH WITH MEEM INITIAL FORM")
    (64679 . "ARABIC LIGATURE JEEM WITH HAH INITIAL FORM")
    (64680 . "ARABIC LIGATURE JEEM WITH MEEM INITIAL FORM")
    (64681 . "ARABIC LIGATURE HAH WITH JEEM INITIAL FORM")
    (64682 . "ARABIC LIGATURE HAH WITH MEEM INITIAL FORM")
    (64683 . "ARABIC LIGATURE KHAH WITH JEEM INITIAL FORM")
    (64684 . "ARABIC LIGATURE KHAH WITH MEEM INITIAL FORM")
    (64685 . "ARABIC LIGATURE SEEN WITH JEEM INITIAL FORM")
    (64686 . "ARABIC LIGATURE SEEN WITH HAH INITIAL FORM")
    (64687 . "ARABIC LIGATURE SEEN WITH KHAH INITIAL FORM")
    (64688 . "ARABIC LIGATURE SEEN WITH MEEM INITIAL FORM")
    (64689 . "ARABIC LIGATURE SAD WITH HAH INITIAL FORM")
    (64690 . "ARABIC LIGATURE SAD WITH KHAH INITIAL FORM")
    (64691 . "ARABIC LIGATURE SAD WITH MEEM INITIAL FORM")
    (64692 . "ARABIC LIGATURE DAD WITH JEEM INITIAL FORM")
    (64693 . "ARABIC LIGATURE DAD WITH HAH INITIAL FORM")
    (64694 . "ARABIC LIGATURE DAD WITH KHAH INITIAL FORM")
    (64695 . "ARABIC LIGATURE DAD WITH MEEM INITIAL FORM")
    (64696 . "ARABIC LIGATURE TAH WITH HAH INITIAL FORM")
    (64697 . "ARABIC LIGATURE ZAH WITH MEEM INITIAL FORM")
    (64698 . "ARABIC LIGATURE AIN WITH JEEM INITIAL FORM")
    (64699 . "ARABIC LIGATURE AIN WITH MEEM INITIAL FORM")
    (64700 . "ARABIC LIGATURE GHAIN WITH JEEM INITIAL FORM")
    (64701 . "ARABIC LIGATURE GHAIN WITH MEEM INITIAL FORM")
    (64702 . "ARABIC LIGATURE FEH WITH JEEM INITIAL FORM")
    (64703 . "ARABIC LIGATURE FEH WITH HAH INITIAL FORM")
    (64704 . "ARABIC LIGATURE FEH WITH KHAH INITIAL FORM")
    (64705 . "ARABIC LIGATURE FEH WITH MEEM INITIAL FORM")
    (64706 . "ARABIC LIGATURE QAF WITH HAH INITIAL FORM")
    (64707 . "ARABIC LIGATURE QAF WITH MEEM INITIAL FORM")
    (64708 . "ARABIC LIGATURE KAF WITH JEEM INITIAL FORM")
    (64709 . "ARABIC LIGATURE KAF WITH HAH INITIAL FORM")
    (64710 . "ARABIC LIGATURE KAF WITH KHAH INITIAL FORM")
    (64711 . "ARABIC LIGATURE KAF WITH LAM INITIAL FORM")
    (64712 . "ARABIC LIGATURE KAF WITH MEEM INITIAL FORM")
    (64713 . "ARABIC LIGATURE LAM WITH JEEM INITIAL FORM")
    (64714 . "ARABIC LIGATURE LAM WITH HAH INITIAL FORM")
    (64715 . "ARABIC LIGATURE LAM WITH KHAH INITIAL FORM")
    (64716 . "ARABIC LIGATURE LAM WITH MEEM INITIAL FORM")
    (64717 . "ARABIC LIGATURE LAM WITH HEH INITIAL FORM")
    (64718 . "ARABIC LIGATURE MEEM WITH JEEM INITIAL FORM")
    (64719 . "ARABIC LIGATURE MEEM WITH HAH INITIAL FORM")
    (64720 . "ARABIC LIGATURE MEEM WITH KHAH INITIAL FORM")
    (64721 . "ARABIC LIGATURE MEEM WITH MEEM INITIAL FORM")
    (64722 . "ARABIC LIGATURE NOON WITH JEEM INITIAL FORM")
    (64723 . "ARABIC LIGATURE NOON WITH HAH INITIAL FORM")
    (64724 . "ARABIC LIGATURE NOON WITH KHAH INITIAL FORM")
    (64725 . "ARABIC LIGATURE NOON WITH MEEM INITIAL FORM")
    (64726 . "ARABIC LIGATURE NOON WITH HEH INITIAL FORM")
    (64727 . "ARABIC LIGATURE HEH WITH JEEM INITIAL FORM")
    (64728 . "ARABIC LIGATURE HEH WITH MEEM INITIAL FORM")
    (64729 . "ARABIC LIGATURE HEH WITH SUPERSCRIPT ALEF INITIAL FORM")
    (64730 . "ARABIC LIGATURE YEH WITH JEEM INITIAL FORM")
    (64731 . "ARABIC LIGATURE YEH WITH HAH INITIAL FORM")
    (64732 . "ARABIC LIGATURE YEH WITH KHAH INITIAL FORM")
    (64733 . "ARABIC LIGATURE YEH WITH MEEM INITIAL FORM")
    (64734 . "ARABIC LIGATURE YEH WITH HEH INITIAL FORM")
    (64735 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH MEEM MEDIAL FORM")
    (64736 . "ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH HEH MEDIAL FORM")
    (64737 . "ARABIC LIGATURE BEH WITH MEEM MEDIAL FORM")
    (64738 . "ARABIC LIGATURE BEH WITH HEH MEDIAL FORM")
    (64739 . "ARABIC LIGATURE TEH WITH MEEM MEDIAL FORM")
    (64740 . "ARABIC LIGATURE TEH WITH HEH MEDIAL FORM")
    (64741 . "ARABIC LIGATURE THEH WITH MEEM MEDIAL FORM")
    (64742 . "ARABIC LIGATURE THEH WITH HEH MEDIAL FORM")
    (64743 . "ARABIC LIGATURE SEEN WITH MEEM MEDIAL FORM")
    (64744 . "ARABIC LIGATURE SEEN WITH HEH MEDIAL FORM")
    (64745 . "ARABIC LIGATURE SHEEN WITH MEEM MEDIAL FORM")
    (64746 . "ARABIC LIGATURE SHEEN WITH HEH MEDIAL FORM")
    (64747 . "ARABIC LIGATURE KAF WITH LAM MEDIAL FORM")
    (64748 . "ARABIC LIGATURE KAF WITH MEEM MEDIAL FORM")
    (64749 . "ARABIC LIGATURE LAM WITH MEEM MEDIAL FORM")
    (64750 . "ARABIC LIGATURE NOON WITH MEEM MEDIAL FORM")
    (64751 . "ARABIC LIGATURE NOON WITH HEH MEDIAL FORM")
    (64752 . "ARABIC LIGATURE YEH WITH MEEM MEDIAL FORM")
    (64753 . "ARABIC LIGATURE YEH WITH HEH MEDIAL FORM")
    (64754 . "ARABIC LIGATURE SHADDA WITH FATHA MEDIAL FORM")
    (64755 . "ARABIC LIGATURE SHADDA WITH DAMMA MEDIAL FORM")
    (64756 . "ARABIC LIGATURE SHADDA WITH KASRA MEDIAL FORM")
    (64757 . "ARABIC LIGATURE TAH WITH ALEF MAKSURA ISOLATED FORM")
    (64758 . "ARABIC LIGATURE TAH WITH YEH ISOLATED FORM")
    (64759 . "ARABIC LIGATURE AIN WITH ALEF MAKSURA ISOLATED FORM")
    (64760 . "ARABIC LIGATURE AIN WITH YEH ISOLATED FORM")
    (64761 . "ARABIC LIGATURE GHAIN WITH ALEF MAKSURA ISOLATED FORM")
    (64762 . "ARABIC LIGATURE GHAIN WITH YEH ISOLATED FORM")
    (64763 . "ARABIC LIGATURE SEEN WITH ALEF MAKSURA ISOLATED FORM")
    (64764 . "ARABIC LIGATURE SEEN WITH YEH ISOLATED FORM")
    (64765 . "ARABIC LIGATURE SHEEN WITH ALEF MAKSURA ISOLATED FORM")
    (64766 . "ARABIC LIGATURE SHEEN WITH YEH ISOLATED FORM")
    (64767 . "ARABIC LIGATURE HAH WITH ALEF MAKSURA ISOLATED FORM")
    (64768 . "ARABIC LIGATURE HAH WITH YEH ISOLATED FORM")
    (64769 . "ARABIC LIGATURE JEEM WITH ALEF MAKSURA ISOLATED FORM")
    (64770 . "ARABIC LIGATURE JEEM WITH YEH ISOLATED FORM")
    (64771 . "ARABIC LIGATURE KHAH WITH ALEF MAKSURA ISOLATED FORM")
    (64772 . "ARABIC LIGATURE KHAH WITH YEH ISOLATED FORM")
    (64773 . "ARABIC LIGATURE SAD WITH ALEF MAKSURA ISOLATED FORM")
    (64774 . "ARABIC LIGATURE SAD WITH YEH ISOLATED FORM")
    (64775 . "ARABIC LIGATURE DAD WITH ALEF MAKSURA ISOLATED FORM")
    (64776 . "ARABIC LIGATURE DAD WITH YEH ISOLATED FORM")
    (64777 . "ARABIC LIGATURE SHEEN WITH JEEM ISOLATED FORM")
    (64778 . "ARABIC LIGATURE SHEEN WITH HAH ISOLATED FORM")
    (64779 . "ARABIC LIGATURE SHEEN WITH KHAH ISOLATED FORM")
    (64780 . "ARABIC LIGATURE SHEEN WITH MEEM ISOLATED FORM")
    (64781 . "ARABIC LIGATURE SHEEN WITH REH ISOLATED FORM")
    (64782 . "ARABIC LIGATURE SEEN WITH REH ISOLATED FORM")
    (64783 . "ARABIC LIGATURE SAD WITH REH ISOLATED FORM")
    (64784 . "ARABIC LIGATURE DAD WITH REH ISOLATED FORM")
    (64785 . "ARABIC LIGATURE TAH WITH ALEF MAKSURA FINAL FORM")
    (64786 . "ARABIC LIGATURE TAH WITH YEH FINAL FORM")
    (64787 . "ARABIC LIGATURE AIN WITH ALEF MAKSURA FINAL FORM")
    (64788 . "ARABIC LIGATURE AIN WITH YEH FINAL FORM")
    (64789 . "ARABIC LIGATURE GHAIN WITH ALEF MAKSURA FINAL FORM")
    (64790 . "ARABIC LIGATURE GHAIN WITH YEH FINAL FORM")
    (64791 . "ARABIC LIGATURE SEEN WITH ALEF MAKSURA FINAL FORM")
    (64792 . "ARABIC LIGATURE SEEN WITH YEH FINAL FORM")
    (64793 . "ARABIC LIGATURE SHEEN WITH ALEF MAKSURA FINAL FORM")
    (64794 . "ARABIC LIGATURE SHEEN WITH YEH FINAL FORM")
    (64795 . "ARABIC LIGATURE HAH WITH ALEF MAKSURA FINAL FORM")
    (64796 . "ARABIC LIGATURE HAH WITH YEH FINAL FORM")
    (64797 . "ARABIC LIGATURE JEEM WITH ALEF MAKSURA FINAL FORM")
    (64798 . "ARABIC LIGATURE JEEM WITH YEH FINAL FORM")
    (64799 . "ARABIC LIGATURE KHAH WITH ALEF MAKSURA FINAL FORM")
    (64800 . "ARABIC LIGATURE KHAH WITH YEH FINAL FORM")
    (64801 . "ARABIC LIGATURE SAD WITH ALEF MAKSURA FINAL FORM")
    (64802 . "ARABIC LIGATURE SAD WITH YEH FINAL FORM")
    (64803 . "ARABIC LIGATURE DAD WITH ALEF MAKSURA FINAL FORM")
    (64804 . "ARABIC LIGATURE DAD WITH YEH FINAL FORM")
    (64805 . "ARABIC LIGATURE SHEEN WITH JEEM FINAL FORM")
    (64806 . "ARABIC LIGATURE SHEEN WITH HAH FINAL FORM")
    (64807 . "ARABIC LIGATURE SHEEN WITH KHAH FINAL FORM")
    (64808 . "ARABIC LIGATURE SHEEN WITH MEEM FINAL FORM")
    (64809 . "ARABIC LIGATURE SHEEN WITH REH FINAL FORM")
    (64810 . "ARABIC LIGATURE SEEN WITH REH FINAL FORM")
    (64811 . "ARABIC LIGATURE SAD WITH REH FINAL FORM")
    (64812 . "ARABIC LIGATURE DAD WITH REH FINAL FORM")
    (64813 . "ARABIC LIGATURE SHEEN WITH JEEM INITIAL FORM")
    (64814 . "ARABIC LIGATURE SHEEN WITH HAH INITIAL FORM")
    (64815 . "ARABIC LIGATURE SHEEN WITH KHAH INITIAL FORM")
    (64816 . "ARABIC LIGATURE SHEEN WITH MEEM INITIAL FORM")
    (64817 . "ARABIC LIGATURE SEEN WITH HEH INITIAL FORM")
    (64818 . "ARABIC LIGATURE SHEEN WITH HEH INITIAL FORM")
    (64819 . "ARABIC LIGATURE TAH WITH MEEM INITIAL FORM")
    (64820 . "ARABIC LIGATURE SEEN WITH JEEM MEDIAL FORM")
    (64821 . "ARABIC LIGATURE SEEN WITH HAH MEDIAL FORM")
    (64822 . "ARABIC LIGATURE SEEN WITH KHAH MEDIAL FORM")
    (64823 . "ARABIC LIGATURE SHEEN WITH JEEM MEDIAL FORM")
    (64824 . "ARABIC LIGATURE SHEEN WITH HAH MEDIAL FORM")
    (64825 . "ARABIC LIGATURE SHEEN WITH KHAH MEDIAL FORM")
    (64826 . "ARABIC LIGATURE TAH WITH MEEM MEDIAL FORM")
    (64827 . "ARABIC LIGATURE ZAH WITH MEEM MEDIAL FORM")
    (64828 . "ARABIC LIGATURE ALEF WITH FATHATAN FINAL FORM")
    (64829 . "ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM")
    (64830 . "ORNATE LEFT PARENTHESIS")
    (64831 . "ORNATE RIGHT PARENTHESIS")
    (64848 . "ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM")
    (64849 . "ARABIC LIGATURE TEH WITH HAH WITH JEEM FINAL FORM")
    (64850 . "ARABIC LIGATURE TEH WITH HAH WITH JEEM INITIAL FORM")
    (64851 . "ARABIC LIGATURE TEH WITH HAH WITH MEEM INITIAL FORM")
    (64852 . "ARABIC LIGATURE TEH WITH KHAH WITH MEEM INITIAL FORM")
    (64853 . "ARABIC LIGATURE TEH WITH MEEM WITH JEEM INITIAL FORM")
    (64854 . "ARABIC LIGATURE TEH WITH MEEM WITH HAH INITIAL FORM")
    (64855 . "ARABIC LIGATURE TEH WITH MEEM WITH KHAH INITIAL FORM")
    (64856 . "ARABIC LIGATURE JEEM WITH MEEM WITH HAH FINAL FORM")
    (64857 . "ARABIC LIGATURE JEEM WITH MEEM WITH HAH INITIAL FORM")
    (64858 . "ARABIC LIGATURE HAH WITH MEEM WITH YEH FINAL FORM")
    (64859 . "ARABIC LIGATURE HAH WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64860 . "ARABIC LIGATURE SEEN WITH HAH WITH JEEM INITIAL FORM")
    (64861 . "ARABIC LIGATURE SEEN WITH JEEM WITH HAH INITIAL FORM")
    (64862 . "ARABIC LIGATURE SEEN WITH JEEM WITH ALEF MAKSURA FINAL FORM")
    (64863 . "ARABIC LIGATURE SEEN WITH MEEM WITH HAH FINAL FORM")
    (64864 . "ARABIC LIGATURE SEEN WITH MEEM WITH HAH INITIAL FORM")
    (64865 . "ARABIC LIGATURE SEEN WITH MEEM WITH JEEM INITIAL FORM")
    (64866 . "ARABIC LIGATURE SEEN WITH MEEM WITH MEEM FINAL FORM")
    (64867 . "ARABIC LIGATURE SEEN WITH MEEM WITH MEEM INITIAL FORM")
    (64868 . "ARABIC LIGATURE SAD WITH HAH WITH HAH FINAL FORM")
    (64869 . "ARABIC LIGATURE SAD WITH HAH WITH HAH INITIAL FORM")
    (64870 . "ARABIC LIGATURE SAD WITH MEEM WITH MEEM FINAL FORM")
    (64871 . "ARABIC LIGATURE SHEEN WITH HAH WITH MEEM FINAL FORM")
    (64872 . "ARABIC LIGATURE SHEEN WITH HAH WITH MEEM INITIAL FORM")
    (64873 . "ARABIC LIGATURE SHEEN WITH JEEM WITH YEH FINAL FORM")
    (64874 . "ARABIC LIGATURE SHEEN WITH MEEM WITH KHAH FINAL FORM")
    (64875 . "ARABIC LIGATURE SHEEN WITH MEEM WITH KHAH INITIAL FORM")
    (64876 . "ARABIC LIGATURE SHEEN WITH MEEM WITH MEEM FINAL FORM")
    (64877 . "ARABIC LIGATURE SHEEN WITH MEEM WITH MEEM INITIAL FORM")
    (64878 . "ARABIC LIGATURE DAD WITH HAH WITH ALEF MAKSURA FINAL FORM")
    (64879 . "ARABIC LIGATURE DAD WITH KHAH WITH MEEM FINAL FORM")
    (64880 . "ARABIC LIGATURE DAD WITH KHAH WITH MEEM INITIAL FORM")
    (64881 . "ARABIC LIGATURE TAH WITH MEEM WITH HAH FINAL FORM")
    (64882 . "ARABIC LIGATURE TAH WITH MEEM WITH HAH INITIAL FORM")
    (64883 . "ARABIC LIGATURE TAH WITH MEEM WITH MEEM INITIAL FORM")
    (64884 . "ARABIC LIGATURE TAH WITH MEEM WITH YEH FINAL FORM")
    (64885 . "ARABIC LIGATURE AIN WITH JEEM WITH MEEM FINAL FORM")
    (64886 . "ARABIC LIGATURE AIN WITH MEEM WITH MEEM FINAL FORM")
    (64887 . "ARABIC LIGATURE AIN WITH MEEM WITH MEEM INITIAL FORM")
    (64888 . "ARABIC LIGATURE AIN WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64889 . "ARABIC LIGATURE GHAIN WITH MEEM WITH MEEM FINAL FORM")
    (64890 . "ARABIC LIGATURE GHAIN WITH MEEM WITH YEH FINAL FORM")
    (64891 . "ARABIC LIGATURE GHAIN WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64892 . "ARABIC LIGATURE FEH WITH KHAH WITH MEEM FINAL FORM")
    (64893 . "ARABIC LIGATURE FEH WITH KHAH WITH MEEM INITIAL FORM")
    (64894 . "ARABIC LIGATURE QAF WITH MEEM WITH HAH FINAL FORM")
    (64895 . "ARABIC LIGATURE QAF WITH MEEM WITH MEEM FINAL FORM")
    (64896 . "ARABIC LIGATURE LAM WITH HAH WITH MEEM FINAL FORM")
    (64897 . "ARABIC LIGATURE LAM WITH HAH WITH YEH FINAL FORM")
    (64898 . "ARABIC LIGATURE LAM WITH HAH WITH ALEF MAKSURA FINAL FORM")
    (64899 . "ARABIC LIGATURE LAM WITH JEEM WITH JEEM INITIAL FORM")
    (64900 . "ARABIC LIGATURE LAM WITH JEEM WITH JEEM FINAL FORM")
    (64901 . "ARABIC LIGATURE LAM WITH KHAH WITH MEEM FINAL FORM")
    (64902 . "ARABIC LIGATURE LAM WITH KHAH WITH MEEM INITIAL FORM")
    (64903 . "ARABIC LIGATURE LAM WITH MEEM WITH HAH FINAL FORM")
    (64904 . "ARABIC LIGATURE LAM WITH MEEM WITH HAH INITIAL FORM")
    (64905 . "ARABIC LIGATURE MEEM WITH HAH WITH JEEM INITIAL FORM")
    (64906 . "ARABIC LIGATURE MEEM WITH HAH WITH MEEM INITIAL FORM")
    (64907 . "ARABIC LIGATURE MEEM WITH HAH WITH YEH FINAL FORM")
    (64908 . "ARABIC LIGATURE MEEM WITH JEEM WITH HAH INITIAL FORM")
    (64909 . "ARABIC LIGATURE MEEM WITH JEEM WITH MEEM INITIAL FORM")
    (64910 . "ARABIC LIGATURE MEEM WITH KHAH WITH JEEM INITIAL FORM")
    (64911 . "ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM")
    (64914 . "ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM")
    (64915 . "ARABIC LIGATURE HEH WITH MEEM WITH JEEM INITIAL FORM")
    (64916 . "ARABIC LIGATURE HEH WITH MEEM WITH MEEM INITIAL FORM")
    (64917 . "ARABIC LIGATURE NOON WITH HAH WITH MEEM INITIAL FORM")
    (64918 . "ARABIC LIGATURE NOON WITH HAH WITH ALEF MAKSURA FINAL FORM")
    (64919 . "ARABIC LIGATURE NOON WITH JEEM WITH MEEM FINAL FORM")
    (64920 . "ARABIC LIGATURE NOON WITH JEEM WITH MEEM INITIAL FORM")
    (64921 . "ARABIC LIGATURE NOON WITH JEEM WITH ALEF MAKSURA FINAL FORM")
    (64922 . "ARABIC LIGATURE NOON WITH MEEM WITH YEH FINAL FORM")
    (64923 . "ARABIC LIGATURE NOON WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64924 . "ARABIC LIGATURE YEH WITH MEEM WITH MEEM FINAL FORM")
    (64925 . "ARABIC LIGATURE YEH WITH MEEM WITH MEEM INITIAL FORM")
    (64926 . "ARABIC LIGATURE BEH WITH KHAH WITH YEH FINAL FORM")
    (64927 . "ARABIC LIGATURE TEH WITH JEEM WITH YEH FINAL FORM")
    (64928 . "ARABIC LIGATURE TEH WITH JEEM WITH ALEF MAKSURA FINAL FORM")
    (64929 . "ARABIC LIGATURE TEH WITH KHAH WITH YEH FINAL FORM")
    (64930 . "ARABIC LIGATURE TEH WITH KHAH WITH ALEF MAKSURA FINAL FORM")
    (64931 . "ARABIC LIGATURE TEH WITH MEEM WITH YEH FINAL FORM")
    (64932 . "ARABIC LIGATURE TEH WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64933 . "ARABIC LIGATURE JEEM WITH MEEM WITH YEH FINAL FORM")
    (64934 . "ARABIC LIGATURE JEEM WITH HAH WITH ALEF MAKSURA FINAL FORM")
    (64935 . "ARABIC LIGATURE JEEM WITH MEEM WITH ALEF MAKSURA FINAL FORM")
    (64936 . "ARABIC LIGATURE SEEN WITH KHAH WITH ALEF MAKSURA FINAL FORM")
    (64937 . "ARABIC LIGATURE SAD WITH HAH WITH YEH FINAL FORM")
    (64938 . "ARABIC LIGATURE SHEEN WITH HAH WITH YEH FINAL FORM")
    (64939 . "ARABIC LIGATURE DAD WITH HAH WITH YEH FINAL FORM")
    (64940 . "ARABIC LIGATURE LAM WITH JEEM WITH YEH FINAL FORM")
    (64941 . "ARABIC LIGATURE LAM WITH MEEM WITH YEH FINAL FORM")
    (64942 . "ARABIC LIGATURE YEH WITH HAH WITH YEH FINAL FORM")
    (64943 . "ARABIC LIGATURE YEH WITH JEEM WITH YEH FINAL FORM")
    (64944 . "ARABIC LIGATURE YEH WITH MEEM WITH YEH FINAL FORM")
    (64945 . "ARABIC LIGATURE MEEM WITH MEEM WITH YEH FINAL FORM")
    (64946 . "ARABIC LIGATURE QAF WITH MEEM WITH YEH FINAL FORM")
    (64947 . "ARABIC LIGATURE NOON WITH HAH WITH YEH FINAL FORM")
    (64948 . "ARABIC LIGATURE QAF WITH MEEM WITH HAH INITIAL FORM")
    (64949 . "ARABIC LIGATURE LAM WITH HAH WITH MEEM INITIAL FORM")
    (64950 . "ARABIC LIGATURE AIN WITH MEEM WITH YEH FINAL FORM")
    (64951 . "ARABIC LIGATURE KAF WITH MEEM WITH YEH FINAL FORM")
    (64952 . "ARABIC LIGATURE NOON WITH JEEM WITH HAH INITIAL FORM")
    (64953 . "ARABIC LIGATURE MEEM WITH KHAH WITH YEH FINAL FORM")
    (64954 . "ARABIC LIGATURE LAM WITH JEEM WITH MEEM INITIAL FORM")
    (64955 . "ARABIC LIGATURE KAF WITH MEEM WITH MEEM FINAL FORM")
    (64956 . "ARABIC LIGATURE LAM WITH JEEM WITH MEEM FINAL FORM")
    (64957 . "ARABIC LIGATURE NOON WITH JEEM WITH HAH FINAL FORM")
    (64958 . "ARABIC LIGATURE JEEM WITH HAH WITH YEH FINAL FORM")
    (64959 . "ARABIC LIGATURE HAH WITH JEEM WITH YEH FINAL FORM")
    (64960 . "ARABIC LIGATURE MEEM WITH JEEM WITH YEH FINAL FORM")
    (64961 . "ARABIC LIGATURE FEH WITH MEEM WITH YEH FINAL FORM")
    (64962 . "ARABIC LIGATURE BEH WITH HAH WITH YEH FINAL FORM")
    (64963 . "ARABIC LIGATURE KAF WITH MEEM WITH MEEM INITIAL FORM")
    (64964 . "ARABIC LIGATURE AIN WITH JEEM WITH MEEM INITIAL FORM")
    (64965 . "ARABIC LIGATURE SAD WITH MEEM WITH MEEM INITIAL FORM")
    (64966 . "ARABIC LIGATURE SEEN WITH KHAH WITH YEH FINAL FORM")
    (64967 . "ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM")
    (65008 . "ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM")
    (65009 . "ARABIC LIGATURE QALA USED AS KORANIC STOP SIGN ISOLATED FORM")
    (65010 . "ARABIC LIGATURE ALLAH ISOLATED FORM")
    (65011 . "ARABIC LIGATURE AKBAR ISOLATED FORM")
    (65012 . "ARABIC LIGATURE MOHAMMAD ISOLATED FORM")
    (65013 . "ARABIC LIGATURE SALAM ISOLATED FORM")
    (65014 . "ARABIC LIGATURE RASOUL ISOLATED FORM")
    (65015 . "ARABIC LIGATURE ALAYHE ISOLATED FORM")
    (65016 . "ARABIC LIGATURE WASALLAM ISOLATED FORM")
    (65017 . "ARABIC LIGATURE SALLA ISOLATED FORM")
    (65018 . "ARABIC LIGATURE SALLALLAHOU ALAYHE WASALLAM")
    (65019 . "ARABIC LIGATURE JALLAJALALOUHOU")
    (65056 . "COMBINING LIGATURE LEFT HALF")
    (65057 . "COMBINING LIGATURE RIGHT HALF")
    (65058 . "COMBINING DOUBLE TILDE LEFT HALF")
    (65059 . "COMBINING DOUBLE TILDE RIGHT HALF")
    (65072 . "PRESENTATION FORM FOR VERTICAL TWO DOT LEADER")
    (65073 . "PRESENTATION FORM FOR VERTICAL EM DASH")
    (65074 . "PRESENTATION FORM FOR VERTICAL EN DASH")
    (65075 . "PRESENTATION FORM FOR VERTICAL LOW LINE")
    (65076 . "PRESENTATION FORM FOR VERTICAL WAVY LOW LINE")
    (65077 . "PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS")
    (65078 . "PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS")
    (65079 . "PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET")
    (65080 . "PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET")
    (65081 . "PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET")
    (65082 . "PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET")
    (65083 . "PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET")
    (65084 . "PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET")
    (65085 . "PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET")
    (65086 . "PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET")
    (65087 . "PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET")
    (65088 . "PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET")
    (65089 . "PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET")
    (65090 . "PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET")
    (65091 . "PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET")
    (65092 . "PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET")
    (65097 . "DASHED OVERLINE")
    (65098 . "CENTERLINE OVERLINE|CENTRELINE OVERLINE")
    (65099 . "WAVY OVERLINE")
    (65100 . "DOUBLE WAVY OVERLINE")
    (65101 . "DASHED LOW LINE")
    (65102 . "CENTERLINE LOW LINE|CENTRELINE LOW LINE")
    (65103 . "WAVY LOW LINE")
    (65104 . "SMALL COMMA")
    (65105 . "SMALL IDEOGRAPHIC COMMA")
    (65106 . "SMALL PERIOD|SMALL FULL STOP")
    (65108 . "SMALL SEMICOLON")
    (65109 . "SMALL COLON")
    (65110 . "SMALL QUESTION MARK")
    (65111 . "SMALL EXCLAMATION MARK")
    (65112 . "SMALL EM DASH")
    (65113 . "SMALL LEFT PARENTHESIS")
    (65114 . "SMALL RIGHT PARENTHESIS")
    (65115 . "SMALL LEFT CURLY BRACKET")
    (65116 . "SMALL RIGHT CURLY BRACKET")
    (65117 . "SMALL LEFT TORTOISE SHELL BRACKET")
    (65118 . "SMALL RIGHT TORTOISE SHELL BRACKET")
    (65119 . "SMALL NUMBER SIGN")
    (65120 . "SMALL AMPERSAND")
    (65121 . "SMALL ASTERISK")
    (65122 . "SMALL PLUS SIGN")
    (65123 . "SMALL HYPHEN-MINUS")
    (65124 . "SMALL LESS-THAN SIGN")
    (65125 . "SMALL GREATER-THAN SIGN")
    (65126 . "SMALL EQUALS SIGN")
    (65128 . "SMALL REVERSE SOLIDUS")
    (65129 . "SMALL DOLLAR SIGN")
    (65130 . "SMALL PERCENT SIGN")
    (65131 . "SMALL COMMERCIAL AT")
    (65136 . "ARABIC FATHATAN ISOLATED FORM")
    (65137 . "ARABIC TATWEEL WITH FATHATAN ABOVE")
    (65138 . "ARABIC DAMMATAN ISOLATED FORM")
    (65140 . "ARABIC KASRATAN ISOLATED FORM")
    (65142 . "ARABIC FATHA ISOLATED FORM")
    (65143 . "ARABIC FATHA MEDIAL FORM")
    (65144 . "ARABIC DAMMA ISOLATED FORM")
    (65145 . "ARABIC DAMMA MEDIAL FORM")
    (65146 . "ARABIC KASRA ISOLATED FORM")
    (65147 . "ARABIC KASRA MEDIAL FORM")
    (65148 . "ARABIC SHADDA ISOLATED FORM")
    (65149 . "ARABIC SHADDA MEDIAL FORM")
    (65150 . "ARABIC SUKUN ISOLATED FORM")
    (65151 . "ARABIC SUKUN MEDIAL FORM")
    (65152 . "ARABIC LETTER HAMZA ISOLATED FORM")
    (65153 . "ARABIC LETTER ALEF WITH MADDA ABOVE ISOLATED FORM")
    (65154 . "ARABIC LETTER ALEF WITH MADDA ABOVE FINAL FORM")
    (65155 . "ARABIC LETTER ALEF WITH HAMZA ABOVE ISOLATED FORM")
    (65156 . "ARABIC LETTER ALEF WITH HAMZA ABOVE FINAL FORM")
    (65157 . "ARABIC LETTER WAW WITH HAMZA ABOVE ISOLATED FORM")
    (65158 . "ARABIC LETTER WAW WITH HAMZA ABOVE FINAL FORM")
    (65159 . "ARABIC LETTER ALEF WITH HAMZA BELOW ISOLATED FORM")
    (65160 . "ARABIC LETTER ALEF WITH HAMZA BELOW FINAL FORM")
    (65161 . "ARABIC LETTER YEH WITH HAMZA ABOVE ISOLATED FORM")
    (65162 . "ARABIC LETTER YEH WITH HAMZA ABOVE FINAL FORM")
    (65163 . "ARABIC LETTER YEH WITH HAMZA ABOVE INITIAL FORM")
    (65164 . "ARABIC LETTER YEH WITH HAMZA ABOVE MEDIAL FORM")
    (65165 . "ARABIC LETTER ALEF ISOLATED FORM")
    (65166 . "ARABIC LETTER ALEF FINAL FORM")
    (65167 . "ARABIC LETTER BEH ISOLATED FORM")
    (65168 . "ARABIC LETTER BEH FINAL FORM")
    (65169 . "ARABIC LETTER BEH INITIAL FORM")
    (65170 . "ARABIC LETTER BEH MEDIAL FORM")
    (65171 . "ARABIC LETTER TEH MARBUTA ISOLATED FORM")
    (65172 . "ARABIC LETTER TEH MARBUTA FINAL FORM")
    (65173 . "ARABIC LETTER TEH ISOLATED FORM")
    (65174 . "ARABIC LETTER TEH FINAL FORM")
    (65175 . "ARABIC LETTER TEH INITIAL FORM")
    (65176 . "ARABIC LETTER TEH MEDIAL FORM")
    (65177 . "ARABIC LETTER THEH ISOLATED FORM")
    (65178 . "ARABIC LETTER THEH FINAL FORM")
    (65179 . "ARABIC LETTER THEH INITIAL FORM")
    (65180 . "ARABIC LETTER THEH MEDIAL FORM")
    (65181 . "ARABIC LETTER JEEM ISOLATED FORM")
    (65182 . "ARABIC LETTER JEEM FINAL FORM")
    (65183 . "ARABIC LETTER JEEM INITIAL FORM")
    (65184 . "ARABIC LETTER JEEM MEDIAL FORM")
    (65185 . "ARABIC LETTER HAH ISOLATED FORM")
    (65186 . "ARABIC LETTER HAH FINAL FORM")
    (65187 . "ARABIC LETTER HAH INITIAL FORM")
    (65188 . "ARABIC LETTER HAH MEDIAL FORM")
    (65189 . "ARABIC LETTER KHAH ISOLATED FORM")
    (65190 . "ARABIC LETTER KHAH FINAL FORM")
    (65191 . "ARABIC LETTER KHAH INITIAL FORM")
    (65192 . "ARABIC LETTER KHAH MEDIAL FORM")
    (65193 . "ARABIC LETTER DAL ISOLATED FORM")
    (65194 . "ARABIC LETTER DAL FINAL FORM")
    (65195 . "ARABIC LETTER THAL ISOLATED FORM")
    (65196 . "ARABIC LETTER THAL FINAL FORM")
    (65197 . "ARABIC LETTER REH ISOLATED FORM")
    (65198 . "ARABIC LETTER REH FINAL FORM")
    (65199 . "ARABIC LETTER ZAIN ISOLATED FORM")
    (65200 . "ARABIC LETTER ZAIN FINAL FORM")
    (65201 . "ARABIC LETTER SEEN ISOLATED FORM")
    (65202 . "ARABIC LETTER SEEN FINAL FORM")
    (65203 . "ARABIC LETTER SEEN INITIAL FORM")
    (65204 . "ARABIC LETTER SEEN MEDIAL FORM")
    (65205 . "ARABIC LETTER SHEEN ISOLATED FORM")
    (65206 . "ARABIC LETTER SHEEN FINAL FORM")
    (65207 . "ARABIC LETTER SHEEN INITIAL FORM")
    (65208 . "ARABIC LETTER SHEEN MEDIAL FORM")
    (65209 . "ARABIC LETTER SAD ISOLATED FORM")
    (65210 . "ARABIC LETTER SAD FINAL FORM")
    (65211 . "ARABIC LETTER SAD INITIAL FORM")
    (65212 . "ARABIC LETTER SAD MEDIAL FORM")
    (65213 . "ARABIC LETTER DAD ISOLATED FORM")
    (65214 . "ARABIC LETTER DAD FINAL FORM")
    (65215 . "ARABIC LETTER DAD INITIAL FORM")
    (65216 . "ARABIC LETTER DAD MEDIAL FORM")
    (65217 . "ARABIC LETTER TAH ISOLATED FORM")
    (65218 . "ARABIC LETTER TAH FINAL FORM")
    (65219 . "ARABIC LETTER TAH INITIAL FORM")
    (65220 . "ARABIC LETTER TAH MEDIAL FORM")
    (65221 . "ARABIC LETTER ZAH ISOLATED FORM")
    (65222 . "ARABIC LETTER ZAH FINAL FORM")
    (65223 . "ARABIC LETTER ZAH INITIAL FORM")
    (65224 . "ARABIC LETTER ZAH MEDIAL FORM")
    (65225 . "ARABIC LETTER AIN ISOLATED FORM")
    (65226 . "ARABIC LETTER AIN FINAL FORM")
    (65227 . "ARABIC LETTER AIN INITIAL FORM")
    (65228 . "ARABIC LETTER AIN MEDIAL FORM")
    (65229 . "ARABIC LETTER GHAIN ISOLATED FORM")
    (65230 . "ARABIC LETTER GHAIN FINAL FORM")
    (65231 . "ARABIC LETTER GHAIN INITIAL FORM")
    (65232 . "ARABIC LETTER GHAIN MEDIAL FORM")
    (65233 . "ARABIC LETTER FEH ISOLATED FORM")
    (65234 . "ARABIC LETTER FEH FINAL FORM")
    (65235 . "ARABIC LETTER FEH INITIAL FORM")
    (65236 . "ARABIC LETTER FEH MEDIAL FORM")
    (65237 . "ARABIC LETTER QAF ISOLATED FORM")
    (65238 . "ARABIC LETTER QAF FINAL FORM")
    (65239 . "ARABIC LETTER QAF INITIAL FORM")
    (65240 . "ARABIC LETTER QAF MEDIAL FORM")
    (65241 . "ARABIC LETTER KAF ISOLATED FORM")
    (65242 . "ARABIC LETTER KAF FINAL FORM")
    (65243 . "ARABIC LETTER KAF INITIAL FORM")
    (65244 . "ARABIC LETTER KAF MEDIAL FORM")
    (65245 . "ARABIC LETTER LAM ISOLATED FORM")
    (65246 . "ARABIC LETTER LAM FINAL FORM")
    (65247 . "ARABIC LETTER LAM INITIAL FORM")
    (65248 . "ARABIC LETTER LAM MEDIAL FORM")
    (65249 . "ARABIC LETTER MEEM ISOLATED FORM")
    (65250 . "ARABIC LETTER MEEM FINAL FORM")
    (65251 . "ARABIC LETTER MEEM INITIAL FORM")
    (65252 . "ARABIC LETTER MEEM MEDIAL FORM")
    (65253 . "ARABIC LETTER NOON ISOLATED FORM")
    (65254 . "ARABIC LETTER NOON FINAL FORM")
    (65255 . "ARABIC LETTER NOON INITIAL FORM")
    (65256 . "ARABIC LETTER NOON MEDIAL FORM")
    (65257 . "ARABIC LETTER HEH ISOLATED FORM")
    (65258 . "ARABIC LETTER HEH FINAL FORM")
    (65259 . "ARABIC LETTER HEH INITIAL FORM")
    (65260 . "ARABIC LETTER HEH MEDIAL FORM")
    (65261 . "ARABIC LETTER WAW ISOLATED FORM")
    (65262 . "ARABIC LETTER WAW FINAL FORM")
    (65263 . "ARABIC LETTER ALEF MAKSURA ISOLATED FORM")
    (65264 . "ARABIC LETTER ALEF MAKSURA FINAL FORM")
    (65265 . "ARABIC LETTER YEH ISOLATED FORM")
    (65266 . "ARABIC LETTER YEH FINAL FORM")
    (65267 . "ARABIC LETTER YEH INITIAL FORM")
    (65268 . "ARABIC LETTER YEH MEDIAL FORM")
    (65269 . "ARABIC LIGATURE LAM WITH ALEF WITH MADDA ABOVE ISOLATED FORM")
    (65270 . "ARABIC LIGATURE LAM WITH ALEF WITH MADDA ABOVE FINAL FORM")
    (65271 . "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA ABOVE ISOLATED FORM")
    (65272 . "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA ABOVE FINAL FORM")
    (65273 . "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW ISOLATED FORM")
    (65274 . "ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW FINAL FORM")
    (65275 . "ARABIC LIGATURE LAM WITH ALEF ISOLATED FORM")
    (65276 . "ARABIC LIGATURE LAM WITH ALEF FINAL FORM")
    (65279 . "ZERO WIDTH NO-BREAK SPACE")
    (65281 . "FULLWIDTH EXCLAMATION MARK")
    (65282 . "FULLWIDTH QUOTATION MARK")
    (65283 . "FULLWIDTH NUMBER SIGN")
    (65284 . "FULLWIDTH DOLLAR SIGN")
    (65285 . "FULLWIDTH PERCENT SIGN")
    (65286 . "FULLWIDTH AMPERSAND")
    (65287 . "FULLWIDTH APOSTROPHE")
    (65288 . "FULLWIDTH LEFT PARENTHESIS")
    (65289 . "FULLWIDTH RIGHT PARENTHESIS")
    (65290 . "FULLWIDTH ASTERISK")
    (65291 . "FULLWIDTH PLUS SIGN")
    (65292 . "FULLWIDTH COMMA")
    (65293 . "FULLWIDTH HYPHEN-MINUS")
    (65294 . "FULLWIDTH PERIOD|FULLWIDTH FULL STOP")
    (65295 . "FULLWIDTH SOLIDUS")
    (65296 . "FULLWIDTH DIGIT ZERO")
    (65297 . "FULLWIDTH DIGIT ONE")
    (65298 . "FULLWIDTH DIGIT TWO")
    (65299 . "FULLWIDTH DIGIT THREE")
    (65300 . "FULLWIDTH DIGIT FOUR")
    (65301 . "FULLWIDTH DIGIT FIVE")
    (65302 . "FULLWIDTH DIGIT SIX")
    (65303 . "FULLWIDTH DIGIT SEVEN")
    (65304 . "FULLWIDTH DIGIT EIGHT")
    (65305 . "FULLWIDTH DIGIT NINE")
    (65306 . "FULLWIDTH COLON")
    (65307 . "FULLWIDTH SEMICOLON")
    (65308 . "FULLWIDTH LESS-THAN SIGN")
    (65309 . "FULLWIDTH EQUALS SIGN")
    (65310 . "FULLWIDTH GREATER-THAN SIGN")
    (65311 . "FULLWIDTH QUESTION MARK")
    (65312 . "FULLWIDTH COMMERCIAL AT")
    (65313 . "FULLWIDTH LATIN CAPITAL LETTER A")
    (65314 . "FULLWIDTH LATIN CAPITAL LETTER B")
    (65315 . "FULLWIDTH LATIN CAPITAL LETTER C")
    (65316 . "FULLWIDTH LATIN CAPITAL LETTER D")
    (65317 . "FULLWIDTH LATIN CAPITAL LETTER E")
    (65318 . "FULLWIDTH LATIN CAPITAL LETTER F")
    (65319 . "FULLWIDTH LATIN CAPITAL LETTER G")
    (65320 . "FULLWIDTH LATIN CAPITAL LETTER H")
    (65321 . "FULLWIDTH LATIN CAPITAL LETTER I")
    (65322 . "FULLWIDTH LATIN CAPITAL LETTER J")
    (65323 . "FULLWIDTH LATIN CAPITAL LETTER K")
    (65324 . "FULLWIDTH LATIN CAPITAL LETTER L")
    (65325 . "FULLWIDTH LATIN CAPITAL LETTER M")
    (65326 . "FULLWIDTH LATIN CAPITAL LETTER N")
    (65327 . "FULLWIDTH LATIN CAPITAL LETTER O")
    (65328 . "FULLWIDTH LATIN CAPITAL LETTER P")
    (65329 . "FULLWIDTH LATIN CAPITAL LETTER Q")
    (65330 . "FULLWIDTH LATIN CAPITAL LETTER R")
    (65331 . "FULLWIDTH LATIN CAPITAL LETTER S")
    (65332 . "FULLWIDTH LATIN CAPITAL LETTER T")
    (65333 . "FULLWIDTH LATIN CAPITAL LETTER U")
    (65334 . "FULLWIDTH LATIN CAPITAL LETTER V")
    (65335 . "FULLWIDTH LATIN CAPITAL LETTER W")
    (65336 . "FULLWIDTH LATIN CAPITAL LETTER X")
    (65337 . "FULLWIDTH LATIN CAPITAL LETTER Y")
    (65338 . "FULLWIDTH LATIN CAPITAL LETTER Z")
    (65339 . "FULLWIDTH LEFT SQUARE BRACKET")
    (65340 . "FULLWIDTH REVERSE SOLIDUS")
    (65341 . "FULLWIDTH RIGHT SQUARE BRACKET")
    (65342 . "FULLWIDTH CIRCUMFLEX ACCENT")
    (65343 . "FULLWIDTH LOW LINE")
    (65344 . "FULLWIDTH GRAVE ACCENT")
    (65345 . "FULLWIDTH LATIN SMALL LETTER A")
    (65346 . "FULLWIDTH LATIN SMALL LETTER B")
    (65347 . "FULLWIDTH LATIN SMALL LETTER C")
    (65348 . "FULLWIDTH LATIN SMALL LETTER D")
    (65349 . "FULLWIDTH LATIN SMALL LETTER E")
    (65350 . "FULLWIDTH LATIN SMALL LETTER F")
    (65351 . "FULLWIDTH LATIN SMALL LETTER G")
    (65352 . "FULLWIDTH LATIN SMALL LETTER H")
    (65353 . "FULLWIDTH LATIN SMALL LETTER I")
    (65354 . "FULLWIDTH LATIN SMALL LETTER J")
    (65355 . "FULLWIDTH LATIN SMALL LETTER K")
    (65356 . "FULLWIDTH LATIN SMALL LETTER L")
    (65357 . "FULLWIDTH LATIN SMALL LETTER M")
    (65358 . "FULLWIDTH LATIN SMALL LETTER N")
    (65359 . "FULLWIDTH LATIN SMALL LETTER O")
    (65360 . "FULLWIDTH LATIN SMALL LETTER P")
    (65361 . "FULLWIDTH LATIN SMALL LETTER Q")
    (65362 . "FULLWIDTH LATIN SMALL LETTER R")
    (65363 . "FULLWIDTH LATIN SMALL LETTER S")
    (65364 . "FULLWIDTH LATIN SMALL LETTER T")
    (65365 . "FULLWIDTH LATIN SMALL LETTER U")
    (65366 . "FULLWIDTH LATIN SMALL LETTER V")
    (65367 . "FULLWIDTH LATIN SMALL LETTER W")
    (65368 . "FULLWIDTH LATIN SMALL LETTER X")
    (65369 . "FULLWIDTH LATIN SMALL LETTER Y")
    (65370 . "FULLWIDTH LATIN SMALL LETTER Z")
    (65371 . "FULLWIDTH LEFT CURLY BRACKET")
    (65372 . "FULLWIDTH VERTICAL LINE")
    (65373 . "FULLWIDTH RIGHT CURLY BRACKET")
    (65374 . "FULLWIDTH TILDE")
    (65377 . "HALFWIDTH IDEOGRAPHIC PERIOD|HALFWIDTH IDEOGRAPHIC FULL STOP")
    (65378 . "HALFWIDTH LEFT CORNER BRACKET")
    (65379 . "HALFWIDTH RIGHT CORNER BRACKET")
    (65380 . "HALFWIDTH IDEOGRAPHIC COMMA")
    (65381 . "HALFWIDTH KATAKANA MIDDLE DOT")
    (65382 . "HALFWIDTH KATAKANA LETTER WO")
    (65383 . "HALFWIDTH KATAKANA LETTER SMALL A")
    (65384 . "HALFWIDTH KATAKANA LETTER SMALL I")
    (65385 . "HALFWIDTH KATAKANA LETTER SMALL U")
    (65386 . "HALFWIDTH KATAKANA LETTER SMALL E")
    (65387 . "HALFWIDTH KATAKANA LETTER SMALL O")
    (65388 . "HALFWIDTH KATAKANA LETTER SMALL YA")
    (65389 . "HALFWIDTH KATAKANA LETTER SMALL YU")
    (65390 . "HALFWIDTH KATAKANA LETTER SMALL YO")
    (65391 . "HALFWIDTH KATAKANA LETTER SMALL TU")
    (65392 . "HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK")
    (65393 . "HALFWIDTH KATAKANA LETTER A")
    (65394 . "HALFWIDTH KATAKANA LETTER I")
    (65395 . "HALFWIDTH KATAKANA LETTER U")
    (65396 . "HALFWIDTH KATAKANA LETTER E")
    (65397 . "HALFWIDTH KATAKANA LETTER O")
    (65398 . "HALFWIDTH KATAKANA LETTER KA")
    (65399 . "HALFWIDTH KATAKANA LETTER KI")
    (65400 . "HALFWIDTH KATAKANA LETTER KU")
    (65401 . "HALFWIDTH KATAKANA LETTER KE")
    (65402 . "HALFWIDTH KATAKANA LETTER KO")
    (65403 . "HALFWIDTH KATAKANA LETTER SA")
    (65404 . "HALFWIDTH KATAKANA LETTER SI")
    (65405 . "HALFWIDTH KATAKANA LETTER SU")
    (65406 . "HALFWIDTH KATAKANA LETTER SE")
    (65407 . "HALFWIDTH KATAKANA LETTER SO")
    (65408 . "HALFWIDTH KATAKANA LETTER TA")
    (65409 . "HALFWIDTH KATAKANA LETTER TI")
    (65410 . "HALFWIDTH KATAKANA LETTER TU")
    (65411 . "HALFWIDTH KATAKANA LETTER TE")
    (65412 . "HALFWIDTH KATAKANA LETTER TO")
    (65413 . "HALFWIDTH KATAKANA LETTER NA")
    (65414 . "HALFWIDTH KATAKANA LETTER NI")
    (65415 . "HALFWIDTH KATAKANA LETTER NU")
    (65416 . "HALFWIDTH KATAKANA LETTER NE")
    (65417 . "HALFWIDTH KATAKANA LETTER NO")
    (65418 . "HALFWIDTH KATAKANA LETTER HA")
    (65419 . "HALFWIDTH KATAKANA LETTER HI")
    (65420 . "HALFWIDTH KATAKANA LETTER HU")
    (65421 . "HALFWIDTH KATAKANA LETTER HE")
    (65422 . "HALFWIDTH KATAKANA LETTER HO")
    (65423 . "HALFWIDTH KATAKANA LETTER MA")
    (65424 . "HALFWIDTH KATAKANA LETTER MI")
    (65425 . "HALFWIDTH KATAKANA LETTER MU")
    (65426 . "HALFWIDTH KATAKANA LETTER ME")
    (65427 . "HALFWIDTH KATAKANA LETTER MO")
    (65428 . "HALFWIDTH KATAKANA LETTER YA")
    (65429 . "HALFWIDTH KATAKANA LETTER YU")
    (65430 . "HALFWIDTH KATAKANA LETTER YO")
    (65431 . "HALFWIDTH KATAKANA LETTER RA")
    (65432 . "HALFWIDTH KATAKANA LETTER RI")
    (65433 . "HALFWIDTH KATAKANA LETTER RU")
    (65434 . "HALFWIDTH KATAKANA LETTER RE")
    (65435 . "HALFWIDTH KATAKANA LETTER RO")
    (65436 . "HALFWIDTH KATAKANA LETTER WA")
    (65437 . "HALFWIDTH KATAKANA LETTER N")
    (65438 . "HALFWIDTH KATAKANA VOICED SOUND MARK")
    (65439 . "HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK")
    (65440 . "HALFWIDTH HANGUL FILLER")
    (65441 . "HALFWIDTH HANGUL LETTER KIYEOK")
    (65442 . "HALFWIDTH HANGUL LETTER SSANGKIYEOK")
    (65443 . "HALFWIDTH HANGUL LETTER KIYEOK-SIOS")
    (65444 . "HALFWIDTH HANGUL LETTER NIEUN")
    (65445 . "HALFWIDTH HANGUL LETTER NIEUN-CIEUC")
    (65446 . "HALFWIDTH HANGUL LETTER NIEUN-HIEUH")
    (65447 . "HALFWIDTH HANGUL LETTER TIKEUT")
    (65448 . "HALFWIDTH HANGUL LETTER SSANGTIKEUT")
    (65449 . "HALFWIDTH HANGUL LETTER RIEUL")
    (65450 . "HALFWIDTH HANGUL LETTER RIEUL-KIYEOK")
    (65451 . "HALFWIDTH HANGUL LETTER RIEUL-MIEUM")
    (65452 . "HALFWIDTH HANGUL LETTER RIEUL-PIEUP")
    (65453 . "HALFWIDTH HANGUL LETTER RIEUL-SIOS")
    (65454 . "HALFWIDTH HANGUL LETTER RIEUL-THIEUTH")
    (65455 . "HALFWIDTH HANGUL LETTER RIEUL-PHIEUPH")
    (65456 . "HALFWIDTH HANGUL LETTER RIEUL-HIEUH")
    (65457 . "HALFWIDTH HANGUL LETTER MIEUM")
    (65458 . "HALFWIDTH HANGUL LETTER PIEUP")
    (65459 . "HALFWIDTH HANGUL LETTER SSANGPIEUP")
    (65460 . "HALFWIDTH HANGUL LETTER PIEUP-SIOS")
    (65461 . "HALFWIDTH HANGUL LETTER SIOS")
    (65462 . "HALFWIDTH HANGUL LETTER SSANGSIOS")
    (65463 . "HALFWIDTH HANGUL LETTER IEUNG")
    (65464 . "HALFWIDTH HANGUL LETTER CIEUC")
    (65465 . "HALFWIDTH HANGUL LETTER SSANGCIEUC")
    (65466 . "HALFWIDTH HANGUL LETTER CHIEUCH")
    (65467 . "HALFWIDTH HANGUL LETTER KHIEUKH")
    (65468 . "HALFWIDTH HANGUL LETTER THIEUTH")
    (65469 . "HALFWIDTH HANGUL LETTER PHIEUPH")
    (65470 . "HALFWIDTH HANGUL LETTER HIEUH")
    (65474 . "HALFWIDTH HANGUL LETTER A")
    (65475 . "HALFWIDTH HANGUL LETTER AE")
    (65476 . "HALFWIDTH HANGUL LETTER YA")
    (65477 . "HALFWIDTH HANGUL LETTER YAE")
    (65478 . "HALFWIDTH HANGUL LETTER EO")
    (65479 . "HALFWIDTH HANGUL LETTER E")
    (65482 . "HALFWIDTH HANGUL LETTER YEO")
    (65483 . "HALFWIDTH HANGUL LETTER YE")
    (65484 . "HALFWIDTH HANGUL LETTER O")
    (65485 . "HALFWIDTH HANGUL LETTER WA")
    (65486 . "HALFWIDTH HANGUL LETTER WAE")
    (65487 . "HALFWIDTH HANGUL LETTER OE")
    (65490 . "HALFWIDTH HANGUL LETTER YO")
    (65491 . "HALFWIDTH HANGUL LETTER U")
    (65492 . "HALFWIDTH HANGUL LETTER WEO")
    (65493 . "HALFWIDTH HANGUL LETTER WE")
    (65494 . "HALFWIDTH HANGUL LETTER WI")
    (65495 . "HALFWIDTH HANGUL LETTER YU")
    (65498 . "HALFWIDTH HANGUL LETTER EU")
    (65499 . "HALFWIDTH HANGUL LETTER YI")
    (65500 . "HALFWIDTH HANGUL LETTER I")
    (65504 . "FULLWIDTH CENT SIGN")
    (65505 . "FULLWIDTH POUND SIGN")
    (65506 . "FULLWIDTH NOT SIGN")
    (65507 . "FULLWIDTH MACRON")
    (65508 . "FULLWIDTH BROKEN BAR")
    (65509 . "FULLWIDTH YEN SIGN")
    (65510 . "FULLWIDTH WON SIGN")
    (65512 . "HALFWIDTH FORMS LIGHT VERTICAL")
    (65513 . "HALFWIDTH LEFTWARDS ARROW")
    (65514 . "HALFWIDTH UPWARDS ARROW")
    (65515 . "HALFWIDTH RIGHTWARDS ARROW")
    (65516 . "HALFWIDTH DOWNWARDS ARROW")
    (65517 . "HALFWIDTH BLACK SQUARE")
    (65518 . "HALFWIDTH WHITE CIRCLE")
    (65533 . "REPLACEMENT CHARACTER")
    )
  "An assoc list of ISO10646 numeric character entities.")

(defun w3-resolve-numeric-entity (ent)
  "Return a string representing the numeric entity ENT (&#ENT;)"
  (let ((val (cdr-safe (assq ent w3-numeric-entities))))
    (cond
     ((null val) (char-to-string ent))
     ((stringp val) val)
     ((integerp val) (char-to-string val))
     ((listp val) (apply 'concat val))
     ((symbolp val) (get 'w3-entities val))
     (t (message "Invalid data type in `w3-numeric-entities' for %d" ent)))))

(provide 'w3-10646)
